/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.io.InputStream;
import java.net.URI;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.Version;
import com.unitesk.requality.core.attribute.VersionsStorage;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;

public abstract class AbstractTreeStorage implements ITreeStorage {
  private IStorageDeltaListener deltaTarget;
  //Here to be stored version between transactions
  //for Struct and Attrs to allow proper handling of undo-redo
  protected Map<UUID, Map<String, Version>> versstorage = new HashMap<>();
  protected CacheStorage cs = null;

//  protected int chRow = 0;
//
//  public int startChangeRow() {
//    return chRow++;
//  }
//
//  public int stopChangeRow() {
//    return chRow--;
//  }

  public AbstractTreeStorage() {
  }

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#setDeltaTarget(com.unitesk.requality.core.IStorageDeltaListener)
   */
  @Override
  public void setDeltaTarget(IStorageDeltaListener deltaTarget) {
    this.deltaTarget = deltaTarget;
  }

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getDeltaTarget()
   */
  @Override
  public IStorageDeltaListener getDeltaTarget() {
    return deltaTarget;
  }

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#createNode(java.util.UUID, java.util.UUID, java.lang.String, java.lang.String)
   */
  @Deprecated
  @Override
  public void createNode(UUID parentNode, final UUID node, String type, String id) {
    if (parentNode == null) {
      type = "DBRoot";
    }
    final String ftype = type;
    Map<String, AttributeDefinition> attr = new HashMap<>();
    attr.put(
        TreeNode.ATTR_TYPE,
        new AttributeDefinition(node, TreeNode.ATTR_TYPE, new ConstValueProvider(AttributeType.STRING, ftype),
            Flag.HIDDEN)); 
    createNode(parentNode, node, id, attr);
  }

  @Override
  public String getNodeRawType(UUID uuid) {
    return AttributeStorage.getNodeRawTypeByAD(uuid, getAttributes(uuid));
  }
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getNodeType(java.util.UUID)
   */

  public Map<String, Version> getVersions(UUID uuid) {
    Map<String, Version> base = new HashMap<>(VersionsStorage.getDefaults()),
        stored = versstorage.get(uuid);
    if (stored == null) stored = Collections.<String, Version>emptyMap();
    for (String s : stored.keySet()) base.put(s, stored.get(s));
    return base;
  }

  public void saveVersions(UUID uuid, ChangesSet<Version> changes) {
    if (getDeltaTarget() != null) getDeltaTarget().versionChanged(uuid, changes, Moment.BEFORE,  true);
    Map<String,Version> nmap = new HashMap<>(changes.getNewMap()); 
    versstorage.put(uuid, nmap);
    for (String s : nmap.keySet()) 
      nmap.put(s, nmap.get(s).clone());
    if (getDeltaTarget() != null) getDeltaTarget().versionChanged(uuid, changes, Moment.AFTER, true);
  }

  public ITreeStorage getRootStorage() {
    ITreeStorage base = this;
    while (base.getBaseStorage() != null) base = base.getBaseStorage();
    return base;
  }

//  protected Map<UUID, Map<String, Map<String, Map<String,Object>>>> cacheStore =
//      new HashMap<>();


  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getRootNode()
   */
  @Override
  public abstract UUID getRootNode();
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getParent(java.util.UUID)
   */
  @Override
  public abstract UUID getParent(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getChildren(java.util.UUID)
   */
  @Override
  public abstract Collection<UUID> getChildren(UUID uuid);
  protected abstract void createStorageNode(
      UUID parentUUId, UUID uuid, String id, Map<String, AttributeDefinition> attributes);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#createNode(java.util.UUID, java.util.UUID, java.lang.String, java.util.Collection)
   */
  @Override
  public final void createNode(
      UUID parentUUId, UUID uuid, String id, Map<String, AttributeDefinition> attributes){
    attributes.put(TreeNode.ATTR_ID, TreeNode.formIdAttribute(uuid, id));
    createStorageNode(parentUUId, uuid, id, attributes);
  }
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#remove(java.util.UUID)
   */
  @Override
  public abstract void remove(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getNodeId(java.util.UUID)
   */
  @Override
  public abstract String getNodeId(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#changeNodeId(java.util.UUID, java.lang.String)
   */
  @Override
  public abstract void changeNodeId(UUID uuid, String id);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getAttributes(java.util.UUID)
   */
  public Map<String, AttributeDefinition> getAttributes(UUID uuid){
    Map<String, AttributeDefinition> fatr = getStorageAttributes(uuid);
    if(fatr!=null)
      fatr.put(TreeNode.ATTR_ID,TreeNode.formIdAttribute(uuid, getNodeId(uuid)));
    return fatr;
  }
  public abstract Map<String, AttributeDefinition> getStorageAttributes(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#saveAttributes(java.util.UUID, java.util.Collection)
   */
  @Override
  public void saveAttributes(
      UUID uuid, ChangesSet<Attribute> params){
    Attribute id = params.getNew(TreeNode.ATTR_ID);
    if(id!=null)
      changeNodeId(uuid, id.getGValue());
    saveStorageAttributes(uuid, params);
  }
  public abstract void saveStorageAttributes(
      UUID uuid, ChangesSet<Attribute> params);
  
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#moveNodeTo(java.util.UUID, java.util.UUID)
   */
  @Override
  public abstract void moveNodeTo(UUID uuid, UUID parent);

  @Override
  public abstract boolean canLoad(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#isResourceExists(java.util.UUID, java.lang.String)
   */
  @Override
  public abstract boolean isResourceExists(UUID uuid, String resName);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getNodeResourcePath(java.util.UUID)
   */
  @Override
  public abstract URI getNodeResourcePath(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getResourceContents(java.util.UUID, java.lang.String)
   */
  @Override
  public abstract InputStream getResourceContents(UUID uuid, String resName);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#setResourceContents(java.util.UUID, java.lang.String, java.io.InputStream)
   */
  @Override
  public abstract void setResourceContents(UUID uuid, String resName, InputStream in, Object data);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#listResources(java.util.UUID)
   */
  @Override
  public abstract Set<String> listResources(UUID uuid);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#deleteResource(java.util.UUID, java.lang.String)
   */
  @Override
  public abstract void deleteResource(UUID uuid, String resName);

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#createStorage(java.lang.String)
   */
  @Override
  public abstract void createStorage(String name);
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#deleteStorage()
   */
  @Override
  public abstract void deleteStorage();

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ITreeStorage#getName()
   */
  @Override
  public abstract String getName();

  public Map<UUID, Map<String, Version>> getVersionsStorage() {
    return versstorage;
  }

  public void clearCachedCache(UUID uid) {
    //TODO crear
    versstorage.remove(uid);
  }

  public void clearCaches() {
    //TODO def clear
    versstorage.clear();
  }

  
  public CacheStorage getCacheStorage()
  {
    return cs;
  }

}
