package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.TextNode;

public class AttributesVersioningRule implements IVersioningRule<Attribute> {
  private boolean all = false;
  private Map<String,Set<String>> names = null;
  public static String ATTR_NODE_VERSION = "_nodeversion";
  private static final String PREFERENCE_ATTR_POLICY = "_com.unitesk.requality.ver.policy",
      PREFERENCE_EN_ALL = "_com.unitesk.requality.ver.enall",
      PREFERENCE_IS_ENABLED= "_com.unitesk.requality.ver.isen";
  public static final String VALUE_ILLEGAL = "NONE";
  private static Set<String> legaltypes = new HashSet<>();
  static{
    legaltypes.add(Comment.TYPE_NAME);
    legaltypes.add(Requirement.TYPE_NAME);
    legaltypes.add(TextNode.TYPE_NAME);
    legaltypes.add(TestPurpose.TYPE_NAME);
  }
  
  public AttributesVersioningRule(Map<String, Set<String>> nname, boolean all){
    this.all = all;
    this.names = nname;
  }
  
  public boolean isEnAll(){
    return all;
  }
  
  public Set<String> getNames(String type){
    Set<String> ret = names.get(type);
    if(ret!=null)
      return ret;
    return Collections.emptySet();
  }
  @Override
  public boolean match(String type, ChangesSet<Attribute> cs) {
    if(all){
      if(legaltypes.contains(type))
        for(String as:cs.getAll())
        {
          if(as.equals(AttributesVersioningRule.ATTR_NODE_VERSION))
            continue;//skip version attribute 
          Attribute n = cs.getNew(as),
                    o = cs.getOld(as);
          if(n!=null && !n.isVirtual() || o!=null && !o.isVirtual())
            return true;
        }
    }else
    {
      Set<String> typed = names.get(type);
      Set<String> al = cs.getAll(), mrg = new HashSet<>(al);
      if(typed!=null){
        mrg.addAll(typed);
        if(mrg.size()<(typed.size()+al.size()))
          return true;
      }
    }
    return false;
  }

  public Map<String, Set<String>> getTypedNames() {
    return names;
  }

  public static AttributesVersioningRule load(TreeNode nodeToStore) {
    boolean b = nodeToStore.attrs().getBooleanAttributeValue(PREFERENCE_IS_ENABLED,false);
    AttributesVersioningRule avr = null;
    if(b){
      boolean ena = nodeToStore.attrs().getBooleanAttributeValue(PREFERENCE_EN_ALL,true);
      if(ena)
        avr = new AttributesVersioningRule(null, ena);
      else{
        Map<String, Set<String>> names = null;
        try{
          names = loadMap(nodeToStore.attrs().getListAttributeValue(PREFERENCE_ATTR_POLICY));
        }
        catch(Exception e){
          RLogger.logError(e.getMessage()+":"+CoreUtils.getTrace(e));
        }
        avr = new AttributesVersioningRule(names, false);
      }
    }
    return avr;
  }
  private static Map<String, Set<String>> loadMap(List<Object> listAttributeValue) {
    Map<String,Set<String>> amap = new HashMap<>();
    for(Object o : listAttributeValue){
      if(!(o instanceof ListValue))
        throw new RuntimeException("cannot load attribute versioning settings - incorrect value");
      ListValue lv = (ListValue)o;
      List<Object> lvo = lv.getValues();
      Object fv = lvo.get(0);
      if(fv == null)
        throw new RuntimeException("incorrect value in stored settings");
      String tp = fv.toString();
      amap.put(tp, new HashSet<>());
      for(int i = 1;i<lvo.size();i++)
        amap.get(tp).add(lvo.get(i).toString());
    }
    return amap;
  }
  private static ListValue saveMap(Map<String, Set<String>> mp) {
    List<Object> lvl1 = new LinkedList<>();
    for(Entry<String, Set<String>> en:mp.entrySet())
    {
      List<Object> lvl2 = new ArrayList<>();
      lvl2.add(en.getKey());
      for(String s:en.getValue())
        lvl2.add(s);
      lvl1.add(new ListValue(AttributeType.STRING, lvl2));
    }
    return new ListValue(AttributeType.LIST, lvl1);
  }
  public static void save(TreeNode nodeToStore, AttributesVersioningRule avr) {
    if(avr == null){
      nodeToStore.attrs().removeAttribute(PREFERENCE_IS_ENABLED);
      nodeToStore.attrs().removeAttribute(PREFERENCE_EN_ALL);
      nodeToStore.attrs().removeAttribute(PREFERENCE_ATTR_POLICY);
    }else{
      if(avr.isEnAll()){
        nodeToStore.attrs().setConstAttribute(PREFERENCE_IS_ENABLED, true, Flag.HIDDEN);
        nodeToStore.attrs().removeAttribute(PREFERENCE_EN_ALL);
        nodeToStore.attrs().removeAttribute(PREFERENCE_ATTR_POLICY);
      }else{
        nodeToStore.attrs().setConstAttribute(PREFERENCE_IS_ENABLED, true, Flag.HIDDEN);
        nodeToStore.attrs().setConstAttribute(PREFERENCE_EN_ALL, false, Flag.HIDDEN);
        nodeToStore.attrs().setConstAttribute(PREFERENCE_ATTR_POLICY, saveMap(avr.getTypedNames()), Flag.HIDDEN);
      }
    }
    nodeToStore.commitAttributes();
  }

  public static boolean isLegal(String type) {
    return legaltypes.contains(type);
  }
}
