package com.unitesk.requality.core;

import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class CacheStorage implements IUUIDsCached{
  String connectionPath, path;
  ITreeStorage storage;
  Connection c = null;
  String trans = null;
  private String prefix="";
  private static final String LIST_PREFIX = "L";
  ReadWriteLock rwl = new ReentrantReadWriteLock();
  Lock rl = rwl.readLock();
  Lock wl = rwl.writeLock();
  private void readLock(){
    rl.lock();
  }
  private void readUnlock(){
    rl.unlock();
  }
  private void writeLock(){
    wl.lock();
  }
  private void writeUnlock(){
    wl.unlock();
  }
  private static Map<String, Connection> smap = new HashMap<>();
  public Connection getConnection(){
    if(c!=null)
      return c;
    if(smap.containsKey(connectionPath)){
      c = smap.get(connectionPath);
      try {
        if(!c.isClosed())
          return c;
      } catch (SQLException e) {
        e.printStackTrace();
      }
    }
    try {
      Class.forName("org.sqlite.JDBC");
    } catch (ClassNotFoundException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    }
    try {
      File f = new File(path);
      if(!f.exists())
        try {
          f.createNewFile();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      c = DriverManager.getConnection("jdbc:sqlite:"+connectionPath);
      smap.put(connectionPath, c);
      return c;
    } catch (SQLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }
  public CacheStorage(ITreeStorage storage,String prefix) {
    this(storage, CoreUtils.getCacheDir(storage.getRootStorage().getLocation()),
        storage.getName(), prefix);
  }
  public CacheStorage(ITreeStorage storage,java.nio.file.Path path2, String name,String prefix) {
    this.storage = storage;
    this.prefix = prefix;
    try {
      java.nio.file.Path p = path2;
      File pf = p.toFile();
      connectionPath = path = pf.toPath()
          .resolve(""+name+".db")
          .toString();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

//  
//  /* (non-Javadoc)
//   * @see com.unitesk.requality.core.ICacheOperator#setValue(java.util.UUID, java.lang.String, java.lang.String, java.lang.Object)
//   */
//  @Override
  public void setCacheValue(String uid, String section, String key, Object value) {
    initSection(prefix+section);
    removeByKey(uid, prefix+section, key);
    Connection c = getConnection();
    writeLock();
    try{
      //Statement s = c.createStatement();
      String exp = "INSERT INTO '"+prefix+section+"'(uid,key,key2,value) VALUES (?,?,?,?)";
      PreparedStatement ps = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(1, uid.toString());
        ps.setString(2, key.toString());
        ps.setObject(3, value.hashCode()+"");
        ps.setObject(4, value);
        ps.executeUpdate();
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }
    finally{
      writeUnlock();
    }
    //setValue(db.getStorage(), uid, section, key, value);
  }
  public void setCacheValue(String uid, String section, String key,String key2, Object value) {
    initSection(prefix+section);
    removeByKey(uid, prefix+section, key, key2);
    Connection c = getConnection();
    writeLock();
    try {
      //Statement s = c.createStatement();
      String exp = "INSERT INTO '"+prefix+section+"'(uid,key,key2,value) VALUES (?,?,?,?)";
      PreparedStatement ps = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(1, uid.toString());
        ps.setString(2, key.toString());
        ps.setObject(3, key2.toString());
        ps.setObject(4, value);
        ps.executeUpdate();
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }
    finally{
      writeUnlock();
    }
    //setValue(db.getStorage(), uid, section, key, value);
  }
//
//  public <T extends Object> Map<String, T> getValues(String uid, String section, Set<String> keys) {
//    return getValues(db.getStorage(), uid, section, keys);
//  }

//  @SuppressWarnings("unchecked")
//  public static <T extends Object> Map<String, T> getValues(
//      ITreeStorage base, String uid, String section, Set<String> keys) {
//    
//    if (base != null) { // && base instanceof ResourceStorage){
//      // ResourceStorage rs = (ResourceStorage)base;
//      Map<String, String> vals = base.getCacheVars(uid, section, keys);
//      Map<String, T> ret = new HashMap<>();
//      Map<String,Map<Integer,String>> addvals = new HashMap<>();
//      Set<String> ak = new HashSet<>();
//      for(String s:keys)
//        if(!vals.containsKey(s)||vals.get(s)==null||vals.get(s).isEmpty())
//          ak.add(s+"~"+1);
//      Map<String, String> avals = base.getCacheVars(uid, section, ak);
//      while(ak.size()>0){
//        ak.clear();
//        for(String s:avals.keySet()){
//          ak.add(s);
//          String basek = s.substring(0,s.lastIndexOf("~"));
//          if(!vals.containsKey(basek))
//            vals.put(basek, avals.get(s));
//          else
//            vals.put(basek, vals.get(basek)+avals.get(s));
//        }
//        avals = base.getCacheVars(uid, section, ak);
//      }
//      if(vals!=null)
//        for (String s : vals.keySet()) {
//          ret.put(s, (T) fromString(vals.get(s)));
//        }
//      return ret;
//    }
//    return null;
//  }

//  /* (non-Javadoc)
//   * @see com.unitesk.requality.core.ICacheOperator#getValue(java.util.UUID, java.lang.String, java.lang.String)
//   */
//  @Override
//  public <T extends Object> T getCacheValue(String uid, String section, String key) {
//    return getValue(this, uid, prefix+section, key);
//  }
//
//  @SuppressWarnings("unchecked")
//  public static <T extends Object> T getValue(
//      CacheStorage cs, String uid, String section, String key) {
//    if (cs != null) { // && base instanceof ResourceStorage){
//      cs.initSection(cs.prefix+section);
//      Connection c = cs.getConnection();
//      //Statement s = c.createStatement();
//      String exp = "SELECT value FROM '"+cs.prefix+section+"' WHERE uid = ? AND key = ?";
//      PreparedStatement ps = null;
//      try{
//        ps = c.prepareStatement(exp);
//        ps.setString(2, key.toString());
//        ps.setObject(1, uid.toString());
//        ResultSet rs = ps.executeQuery();
//        if(rs.next())
//            return rs.getObject(1);
//      }catch(Exception e){
//        e.printStackTrace();
//      }
//      finally{
//        try{
//          if(ps!=null)
//            ps.close();
//        }catch(Exception e){
//          e.printStackTrace();
//        }
//        try {
//          if(cs.c==null)c.close();
//        } catch (SQLException e1) {
//          // TODO Auto-generated catch block
//          e1.printStackTrace();
//        }
//      }
//      return null;
////      // ResourceStorage rs = (ResourceStorage)base;
////      T val = (T) fromString(cs.getCacheVar(uid, section, key));
////      if(val==null){
////        StringBuilder sb = new StringBuilder();
////        String s1 = cs.getCacheVar(uid, section, key+"~1");
////        if(s1!=null){
////          sb.append(s1);
////          int i = 2;
////          String ts;
////          while((ts = cs.getCacheVar(uid, section, key+"~"+i))!=null){
////            sb.append(ts);
////            i++;
////          }
////          val = (T) fromString(cs.getCacheVar(uid, section, sb.toString()));
////        }
////      }
////      return val;
//    }
//    return null;
//  }
//   public static Object getValue(
//      CacheStorage cs, String uid, String section, String key,String key2) {
//    if (cs != null) { // && base instanceof ResourceStorage){
//      cs.initSection(cs.prefix+section);
//      Connection c = cs.getConnection();
//      //Statement s = c.createStatement();
//      String exp = "SELECT value FROM '"+cs.prefix+section+"' WHERE uid = ? AND key = ? AND key2 = ?";
//      PreparedStatement ps = null;
//      try{
//        ps = c.prepareStatement(exp);
//        ps.setString(3, key2.toString());
//        ps.setString(2, key.toString());
//        ps.setObject(1, uid.toString());
//        ResultSet rs = ps.executeQuery();
//        if(rs.next())
//            return rs.getObject(1);
//      }catch(Exception e){
//        e.printStackTrace();
//      }
//      finally{
//        try{
//          if(ps!=null)
//            ps.close();
//        }catch(Exception e){
//          e.printStackTrace();
//        }
//        try {
//          if(cs.c==null)c.close();
//        } catch (SQLException e1) {
//          // TODO Auto-generated catch block
//          e1.printStackTrace();
//        }
//      }
//      return null;
//    }
//    return null;
//  }
//
//  public static Object fromString(String persistent) {
//    if (persistent == null) return null;
//    int pos = persistent.indexOf("|");
//    if (pos == -1) return persistent; // Tends to be string
//    String tp = persistent.substring(0, pos);
//    AttributeType at = AttributeType.valueOf(tp);
//    persistent = persistent.substring(pos + 1);
//    if (at == AttributeType.LIST) {
//      pos = persistent.indexOf("|");
//      if (pos < -1) return null;
//      tp = persistent.substring(0, pos);
//      if (tp.length() > 0) at = AttributeType.valueOf(tp);
//      else return null;
//      persistent = persistent.substring(pos + 1);
//      String[] vals = persistent.split("\\-\\|\\-");
//      List<Object> lvals = new ArrayList<>();
//      for (String s : vals) if (s.length() > 0) lvals.add(s); // TODO something other than string
//      return new ListValue(at, lvals);
//    }
//    if (at == AttributeType.STRING) return persistent;
//    return null;
//  }
//
//  private static String toString(Object value) {
//    if (value == null) return null;
//    if (value instanceof ListValue) {
//      ListValue lv = (ListValue) value;
//      StringBuffer sb = new StringBuffer();
//      sb.append(AttributeType.LIST.name());
//      sb.append("|");
//      sb.append(lv.getType().name());
//      sb.append("|");
//      for (Object o : lv.getValues()) {
//        if (sb.charAt(sb.length() - 1) != |) sb.append("-|-");
//        sb.append(o.toString());
//      }
//      return sb.toString();
//    }
//    value = value.toString();
//    if (value instanceof String) {
//      StringBuffer sb = new StringBuffer();
//      sb.append(AttributeType.STRING.name());
//      sb.append("|");
//      sb.append(value);
//      return sb.toString();
//    }
//    // TODO Auto-generated method stub
//    return null;
//  }
  public void removeByKey(String uuId, String section, String key, String key2) {
    initSection(prefix+section);
    Connection c = getConnection();
    writeLock();
    try{
      //Statement s = c.createStatement();\
      String exp = "DELETE FROM '"+prefix+section+"' WHERE uid = ? AND key = ? AND key2 = ?";
      PreparedStatement ps = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(1, uuId.toString());
        ps.setString(2, key);
        ps.setString(3, key2);
        ps.executeUpdate();
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      writeUnlock();
    }
  }
  public void removeByKey(String uuId, String section, String key) {
    initSection(prefix+section);
    Connection c = getConnection();
    writeLock();
    try{
      //Statement s = c.createStatement();\
      String exp = "DELETE FROM '"+prefix+section+"' WHERE uid = ? AND key = ?";
      PreparedStatement ps = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(1, uuId.toString());
        ps.setString(2, key);
        ps.executeUpdate();
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }
    finally{
      writeUnlock();
    }
  }
  public void removeByID(String uuId, String section) {
    removeByKeys(uuId, section, null);
  }
  public void removeByKeys(String uuId, String section, Collection<String> keys) {
    initSection(prefix+section);
    Connection c = getConnection();
    //Statement s = c.createStatement();\
    //String exp = "DELETE FROM '"+section+"' WHERE uid = ? AND key in ?";
    writeLock();
    try {
      StringBuilder sb = new StringBuilder("DELETE FROM \"");
      sb.append(prefix+section);
      sb.append("\" WHERE uid = ?");
      if(keys!=null){
        sb.append(" AND key");
        if(keys.size()>1){
          sb.append(" in (");
          for(int i = 0;i<keys.size();i++){
            if(i>0)
              sb.append(",");
            sb.append("?");
          }
          sb.append(")");
        }else
          sb.append(" = ?");
      }
      PreparedStatement ps = null;
      try{
        ps = c.prepareStatement(sb.toString());
        int i = 2;
        ps.setString(1, uuId.toString());
        if(keys!=null)
          for(String s:keys)
            ps.setString(i++,s);
        ps.executeUpdate();
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      writeUnlock();
    }
  }
  private void removeByKeys(String section, Map<String, Map<String, Map<String, Object>>> tos) {
    initSection(prefix+section);
    Connection c = getConnection();
    //Statement s = c.createStatement();\
    //String exp = "DELETE id1 + id2 as FullKey FROM '"+section+"' WHERE FullKey in ?";
    writeLock();
    try{
      StringBuilder sb = new StringBuilder("DELETE FROM ");
      sb.append(prefix+section);
      sb.append(" WHERE EXISTS ( SELECT (uid || ' ' || key) AS expr FROM ");
      sb.append(prefix+section);
      sb.append("  WHERE expr in (");
      int l = 0;
      for(String uid:tos.keySet())
        l+=tos.get(uid).size();
      for(int i = 0;i<l;i++){
        if(i>0)
          sb.append(",");
        sb.append("?");
      }
      sb.append("))");
      PreparedStatement ps = null;
      try{
        boolean b = false;
        if(!inTransaction()||ac==0){
          c.setAutoCommit(false);
          b = true;
        }
        ps = c.prepareStatement(sb.toString());
        int i = 1;
        for(String s:tos.keySet())
          for(String s2:tos.get(s).keySet()){
            ps.setString(i,s+" "+s2);
            //ps.setString(i*2+2,s2);
            ps.addBatch();
            i++;
          }
        ps.executeBatch();
        if(b || this.ac == 1)c.commit();
        if(b)c.setAutoCommit(true);
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
        
      }
    }
    finally{
      writeUnlock();
    }
  }
  private Set<String> inited = new HashSet<>();
  public static int DEFAULT = 0, COMMIT_SINGLE = 1, COMMIT_ALL = 2;
  private int ac = DEFAULT;
  //prefix needs to be 
  private void initAltSection(String section) {
    if(inited.contains(section))
      return;
    readLock();
    try{
      Connection c = getConnection();
      ResultSet tables = null;
      PreparedStatement s = null;
      try{
        DatabaseMetaData dbm = c.getMetaData();
        // check if "employee" table is there
        boolean exists = false;
        try{
          tables = dbm.getTables(null, null, section, null);
          if (tables.next())
            exists = true;
        }
        catch(Exception e){
          e.printStackTrace();
        }
        if(!exists) {
          readUnlock();
          writeLock();
          try{
            StringBuilder sql = new StringBuilder("CREATE TABLE '");
            sql.append(section);
            sql.append("' ");
            sql.append("(uid VARCHAR(36) not NULL, ");
            sql.append(" key VARCHAR(255), "); 
            sql.append(" value BLOB, ");
            sql.append("UNIQUE(uid, key) ON CONFLICT REPLACE)");
            s = c.prepareStatement(sql.toString());
            s.execute();
          }
          finally{
            writeUnlock();
            readLock();
          }
        }
      }
      catch(Exception e){
        e.printStackTrace();
      }
      finally{
        if(s!=null)
          try {
            s.close();
          } catch (SQLException e2) {
            // TODO Auto-generated catch block
            e2.printStackTrace();
          }
        if(tables!=null)
          try {
            tables.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
//        try {
//          c.close();
//        } catch (SQLException e) {
//          // TODO Auto-generated catch block
//          e.printStackTrace();
//        }
      }
      inited.add(section);
    }
    finally{
      readUnlock();
    }
  }
  private void initSection(String section) {
    if(inited.contains(section))
      return;
    readLock();
    try{
      Connection c = getConnection();
      ResultSet tables = null;
      PreparedStatement s = null;
      try{
        DatabaseMetaData dbm = c.getMetaData();
        // check if "employee" table is there
        boolean exists = false;
        try{
          tables = dbm.getTables(null, null, section, null);
          if (tables.next())
            exists = true;
        }
        catch(Exception e){
          e.printStackTrace();
        }
        if(!exists) {
          readUnlock();
          writeLock();
          try{
            StringBuilder sql = new StringBuilder("CREATE TABLE '");
            sql.append(section);
            sql.append("' ");
            sql.append("(uid VARCHAR(36) not NULL, ");
            sql.append(" key VARCHAR(255), "); 
            sql.append(" key2 VARCHAR(255) DEFAULT '', "); 
            sql.append(" value BLOB, ");
            sql.append("UNIQUE(uid, key, key2) ON CONFLICT REPLACE)");
            s = c.prepareStatement(sql.toString());
            s.execute();
          }
          finally{
            writeUnlock();
            readLock();
          }
        }
      }
      catch(Exception e){
        e.printStackTrace();
      }
      finally{
        if(s!=null)
          try {
            s.close();
          } catch (SQLException e2) {
            // TODO Auto-generated catch block
            e2.printStackTrace();
          }
        if(tables!=null)
          try {
            tables.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
//        try {
//          c.close();
//        } catch (SQLException e) {
//          // TODO Auto-generated catch block
//          e.printStackTrace();
//        }
      }
      inited.add(section);
    }
    finally{
      readUnlock();
    }
  }

//  /* (non-Javadoc)
//   * @see com.unitesk.requality.core.ICacheOperator#getMap(java.util.UUID, java.lang.String, java.lang.String)
//   */
//  @Override
//  public <T extends Object> Map<String, T> getCacheMap(String uuId, String section, String key) {
//    return getMap(this, uuId, prefix+section, key);
//  }

//  /* (non-Javadoc)
//   * @see com.unitesk.requality.core.ICacheOperator#getMapByValue(java.util.UUID, java.lang.String, java.lang.String)
//   */
//  @Override
//  public <T extends Object> Map<String, List<T>> getCacheMapByValue(String uuId, String string, String string2) {
//    return getMapByValue(this, uuId, string,string2);
//  }
//  public static <T extends Object> Map<String, List<T>> getMapByValue(
//      CacheStorage cs,String uuId, String section, String key) {
//    Map<String,List<T>> ret = new HashMap<>();
//    if (cs != null) { // && base instanceof ResourceStorage){
//      cs.initSection(cs.prefix+section);
//      Connection c = cs.getConnection();
//      //Statement s = c.createStatement();
//      String exp = "SELECT key2, uid FROM '"+cs.prefix+section+"' WHERE value = ? AND key = ?";
//      PreparedStatement ps = null;
//      try{
//        ps = c.prepareStatement(exp);
//        ps.setString(2, key.toString());
//        ps.setObject(1, uuId.toString());
//        ResultSet rs = ps.executeQuery();
//        while(rs.next()){
//          String key2 = rs.getString(1);
//          Object val = rs.getObject(2);
//          if(!ret.containsKey(key2))
//            ret.put(key2, new LinkedList<>());
//          ret.get(key2).add((T) val);
//        }
//      }catch(Exception e){
//        e.printStackTrace();
//      }
//      finally{
//        try{
//          if(ps!=null)
//            ps.close();
//        }catch(Exception e){
//          e.printStackTrace();
//        }
//        try {
//          if(cs.c==null)c.close();
//        } catch (SQLException e1) {
//          // TODO Auto-generated catch block
//          e1.printStackTrace();
//        }
//      }
//    }
//    return ret;
//  }
  public Map<String,Map<String, List<Object>>> getCacheMapsByValue(
    String uuId, String section, Collection<String> keys) {
    Map<String, Map<String, List<Object>>> ret = new HashMap<>();
    initSection(prefix+section);
    Connection c = getConnection();
    readLock();
    try{
      //Statement s = c.createStatement();
      StringBuilder sb = new StringBuilder();
      sb.append("SELECT key2, uid, key FROM '");
      sb.append(prefix);
      sb.append(section);
      sb.append("' WHERE value = ?");
      if(keys!=null){
        sb.append(" AND key ");
        if(keys.size() == 1){
          sb.append(" = ?");
        }else{
          sb.append("in (");
          for(int i = 0;i<keys.size();i++){
            if(i>0)
              sb.append(",");
            sb.append("?");
          }
          sb.append(")");
        }
      }
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(sb.toString());
        ps.setObject(1, uuId.toString());
        int j = 2;
        if(keys!=null)
          for(String s:keys)
            ps.setString(j++, s);
        rs = ps.executeQuery();
        while(rs.next()){
          String key = rs.getString(3);
          if(!ret.containsKey(key))
            ret.put(key, new HashMap<>());
          String key2 = rs.getString(1);
          Object val = rs.getObject(2);
          if(!ret.get(key).containsKey(key2))
            ret.get(key).put(key2, new LinkedList<>());
          ret.get(key).get(key2).add(val);
        }
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }
    finally{
      readUnlock();
    }
    return ret;
  }
  public static Map<String, Object> getMap(
      CacheStorage cs,String uuId, String section, String key) {
    Map<String,Object> ret = new HashMap<>();
    if (cs != null) { // && base instanceof ResourceStorage){
      cs.initSection(cs.prefix+section);
      PreparedStatement ps = null;
      cs.readLock();
      try{
        Connection c = cs.getConnection();
        //Statement s = c.createStatement();
        String exp = "SELECT key2, value FROM '"+cs.prefix+section+"' WHERE uid = ? AND key = ?";
        ResultSet rs = null;
        try{
          ps = c.prepareStatement(exp);
          ps.setString(2, key.toString());
          ps.setObject(1, uuId.toString());
          rs = ps.executeQuery();
          while(rs.next()){
            String key2 = rs.getString(1);
            Object val = rs.getObject(2);
            ret.put(key2, val);
          }
        }catch(Exception e){
          e.printStackTrace();
        }
        finally{
          try{
            if(ps!=null){
              ps.close();
            }
          }catch(Exception e){
            e.printStackTrace();
          }
          if(rs!=null)
            try {
              rs.close();
            } catch (SQLException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
        }
      }finally{
        cs.readUnlock();
      }
    }
    return ret;
//    ps.setString(2, key.toString());
//
//      
//      
//    String size = getValue(st, uuId, section, key + "|_keys");
//    if (size == null) return null;
//    try {
//      String[] keys = size.split("\\|");
//      // for(int i =0;i<len;i++)
//      Set<String> ks = new HashSet<>();
//      for (String s : keys) ks.add(key + "|" + s);
//      // keys.add(key+"|"+i);
//      Map<String, String> vals = getValues(st, uuId, section, ks);
//      Map<String, T> ret = new HashMap<>();
//      for (String s : keys) ret.put(s, (T) fromString(vals.get(key + "|" + s)));
//      return ret;
//    } catch (Exception e) {
//      e.printStackTrace();
//    }
//    return new HashMap<>();
  }
  
  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ICacheOperator#getMaps(java.util.UUID, java.lang.String, java.util.Collection)
   */
  public Map<String, Map<String, Object>> getCacheMaps(String uuId, String section) {
    return getCacheMaps(uuId, section, null);
  }
  
  public Map<String, Map<String, Object>> getCacheMaps(String uuId, String section,Collection<String> keys) {
    return getMaps(this, uuId, prefix+section, keys);
  }

  public static Map<String, Map<String,Object>> getMaps(
      CacheStorage cs,String uuId, String section, Collection<String> keys) {
    if(uuId == null)
      return new HashMap<>();
    Map<String, Map<String, Object>> ret = new HashMap<>();
    if (cs != null) { // && base instanceof ResourceStorage){
      cs.initSection(cs.prefix+section);
      Connection c = cs.getConnection();
      cs.readLock();
      try {
        //cs.removeByKey(uuId, section, key);
        //Statement s = c.createStatement();
        StringBuilder sb = new StringBuilder();
        sb.append("SELECT key, key2, value FROM ");
        sb.append(cs.prefix+section);
        sb.append(" WHERE uid = ?");
        if(keys!=null){
          sb.append(" AND key in (");
          for(int i = 0;i<keys.size();i++){
            if(i>0)
              sb.append(",");
            sb.append("?");
          }
          sb.append(")");
        }
        ResultSet rs = null;
        PreparedStatement ps = null;
        try{
          ps = c.prepareStatement(sb.toString());
          ps.setObject(1, uuId.toString());
          int i = 2;
          if(keys!=null)
            for(String s:keys)
              ps.setString(i++, s);
          rs = ps.executeQuery();
          while(rs.next()){
            String key = rs.getString(1);
            String key2 = rs.getString(2);
            Object val = rs.getObject(3);
            if(!ret.containsKey(key))
              ret.put(key, new HashMap<>());
            ret.get(key).put(key2, val);
          }
        }catch(Exception e){
          e.printStackTrace();
        }
        finally{
          try{
            if(ps!=null)
              ps.close();
          }catch(Exception e){
            e.printStackTrace();
            RLogger.logInfo(cs.connectionPath+" error "+CoreUtils.getTrace(e));
          }
          if(rs!=null)
            try {
              rs.close();
            } catch (SQLException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
        }
      }finally{
        cs.readUnlock();
      }
    }
    return ret;
//    ps.setString(2, key.toString());
//
//      
//      
//    String size = getValue(st, uuId, section, key + "|_keys");
//    if (size == null) return null;
//    try {
//      String[] keys = size.split("\\|");
//      // for(int i =0;i<len;i++)
//      Set<String> ks = new HashSet<>();
//      for (String s : keys) ks.add(key + "|" + s);
//      // keys.add(key+"|"+i);
//      Map<String, String> vals = getValues(st, uuId, section, ks);
//      Map<String, T> ret = new HashMap<>();
//      for (String s : keys) ret.put(s, (T) fromString(vals.get(key + "|" + s)));
//      return ret;
//    } catch (Exception e) {
//      e.printStackTrace();
//    }
//    return new HashMap<>();
  }
  public boolean saveUUIDs() {
    while (storage != null && !(storage instanceof IUUIDsCached))
      storage = storage.getBaseStorage();
    if (storage != null) {
      if (((IUUIDsCached) storage).areUUIDsChanged()) {
        ((IUUIDsCached) storage).saveUUIDs();
        return true;
      }
    }
    return false;
  }
  @Override
  public boolean areUUIDsChanged() {
    while (storage != null && !(storage instanceof IUUIDsCached))
      storage = storage.getBaseStorage();
    if (storage != null) {
      return ((IUUIDsCached)storage).areUUIDsChanged();
    }
    return false;
  }


  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ICacheOperator#setMaps(java.util.UUID, java.lang.String, java.util.Map)
   */
  public void setCacheMaps(String uuid, String section, Map<String, Map<String, Object>> vars) {
    initSection(prefix+section);
    Connection c = getConnection();
    writeLock();
    try{
      Map<String, Map<String, Object>> old = null;
      //if(this.c == null)
      //old = getCacheMaps(uuid, section, vars.keySet());//TODO is it proper in not commit version?
      removeByKeys(uuid, prefix+section, vars.keySet());
      //Statement s = c.createStatement();
      String exp = "INSERT INTO '"+prefix+section+"'(uid,key,key2,value) VALUES (?,?,?,?)";
      PreparedStatement ps = null;
        try{
          ps = c.prepareStatement(exp);
          boolean dis = false;
          if(!inTransaction()||this.ac==0){
            c.setAutoCommit(false);
            dis = true;
          }
  //          for (String s : map.keySet())
  //            vals.put(key + "|" + s, map.get(s));
          boolean changes = false;
          for(String key:vars.keySet()){
            Map<String,Object>m = vars.get(key);
            if(m!=null)
              for(String key2:m.keySet()){
                ps.setString(1, uuid.toString());
                ps.setString(2, key);
                ps.setString(3, key2);
                ps.setObject(4, vars.get(key).get(key2));
                ps.addBatch();
                changes = true;
              }
          }
          if(changes)
            ps.executeBatch();
          if(dis||this.ac==1)
            c.commit();
          if(dis){
            c.setAutoCommit(true);
          }
        }catch(Exception e){
          e.printStackTrace();
        }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }
    finally{
      writeUnlock();
    }
  }
    // TODO Auto-generated method stub
    
  

//  public static void setValue(
//      ITreeStorage base, String uid, String section, String key, Object value) {
//    if (base != null) {
//      String nw = toString(value);
//      String ov = base.getCacheVar(uid, section, key);
//      if(ov==null || !nw.equals(ov))
//        base.setCacheVar(uid, section, key, nw);
//    }
//  }

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ICacheOperator#clearCache(java.util.UUID)
   */
  public void clearCachedCache(String uuId) {
    Connection c = getConnection();
    List<PreparedStatement> pss = new LinkedList<>();
    writeLock();
    try {
      DatabaseMetaData md;
      ResultSet rs = null;
      PreparedStatement p = null;
      try {
        md = c.getMetaData();
        rs = md.getTables(null, null, "%", null);
        boolean ac = false;
        if(!inTransaction()||this.ac==0){
          c.setAutoCommit(false);
          ac = true;
        }
        while(rs.next()){
          p = c.prepareStatement("DELETE FROM '"+prefix+rs.getString(3)+"' WHERE uid = '"+uuId.toString()+"'");
          pss.add(p);
          p.execute();
        }
        if(ac || this.ac==1)
          c.commit();
        //if(this.c==null)c.commit();
        if(ac)c.setAutoCommit(true);
      } catch (SQLException e) {
        e.printStackTrace();
      }
      finally{
        if(p!=null)
          try {
            p.close();
          } catch (SQLException e2) {
            // TODO Auto-generated catch block
            e2.printStackTrace();
          }
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
       for(PreparedStatement pa:pss)
        try {
          pa.close();
        } catch (SQLException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }finally{
      wl.unlock();
    }
  }
  public String startCacheTransaction(String name,int ac) {
    if(trans != null)
      return "Transaction "+name+" is already started";
    Connection c = getConnection();
    if(this.ac!=0)
      return "Commit already started";
    writeLock();
    try {
      trans = name;
      this.ac = ac;
      if(ac>0)
        try {
          c.setAutoCommit(false);
        } catch (SQLException e){ 
          e.printStackTrace();
        }
  //    try {
  //      //c.setAutoCommit(false);
  //    } catch (Exception e) {
  //      e.printStackTrace();
  //      return false;
  //    }
      
    }finally{
      wl.unlock();
    }
    return null;
  }
  
  public void commitCache(){
    Connection c = getConnection();
    writeLock();
    try {
      if(ac==2){
        try {
          c.commit();
        } catch (SQLException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
        try {
          c.setAutoCommit(true);
        } catch (SQLException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
      trans = null;
      ac = 0;
    }finally{
      writeUnlock();
    }
//    try {
//      c.commit();
//    } catch (Exception e) {
//      e.printStackTrace();
//    }
//    try {
//      c.close();
//    } catch (Exception e) {
//      // TODO Auto-generated catch block
//      e.printStackTrace();
//    }
//    c = null;
  }

  /* (non-Javadoc)
   * @see com.unitesk.requality.core.ICacheOperator#clear()
   */
  public void remove(){
    try{
      if(c!=null)
        c.close();
      ac = 0;
      trans = null;
      new File(path).delete();
      smap.remove(connectionPath);
    }
    catch(Exception e){
      e.printStackTrace();
    }
  }
  public void clearCached() {
    Connection c = getConnection();
    writeLock();
    try {
      DatabaseMetaData md;
      ResultSet rs = null;
      try {
        md = c.getMetaData();
        rs = md.getTables(null, null, "%", null);
        boolean ac = false;
        if(!inTransaction()||this.ac==0){
          c.setAutoCommit(false);
          ac = true;
        }
        while(rs.next()){
          c.prepareStatement("DELETE FROM '"+prefix+rs.getString(3)+"'").execute();
        }
        if(ac||this.ac==1)
          c.commit();
        if(ac)c.setAutoCommit(true);
      } catch (SQLException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
      finally{
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
      }
    }finally{
      writeUnlock();
    }
  }
  public void addCacheListValue(String uuId, String section, String key, String added) {
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Map<String, Object>> old = getCacheMaps(uuId, section, l);
    if(!old.containsKey(key))
      old.put(key, new HashMap<>());
    String key2 = LIST_PREFIX+added.hashCode();
    //old.get(key).remove(key2);
    old.get(key).put(key2, added);
    setCacheMaps(uuId, section, old);
  }

  public void addCacheListValues(String uuId, String section, String key, List<String> list) {
    Map<String, Object> old = getCacheMap(uuId, section, key);
    if(old == null)
      old = new HashMap<>();
    for(String added:list){
      String key2 = LIST_PREFIX+added.hashCode();
      old.put(key2, added);
    }
    setCacheMap(uuId, section,key, old);
  }

  public void removeCacheListValue(String uuId, String section, String key, String value) {
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Object> old = getCacheMap(uuId, section, key);
    if(old!=null && old.containsKey(key)){
       String key2 = LIST_PREFIX+value.hashCode();
       old.remove(key2);
    }
    setCacheMap(uuId, section, key, old);
  }

  public void removeCacheListValues(String uuId, String section, String key, List<String> toda) {
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Object> old = getCacheMap(uuId, section, key);
    if(old!=null){
      for(String value:toda){
        String key2 = LIST_PREFIX+value.hashCode();
        old.remove(key2);
      }
      if(old.isEmpty())
        old = null;
    }
    setCacheMap(uuId, section, key, old);
  }
  public void setCacheListValue(String uuId, String section, String key, List<String> tosave) {
    Map<String, Map<String, Object>> old = new HashMap<>();
    old.put(key, new HashMap<>());
    for(String s:tosave){
      String key2 = LIST_PREFIX+s.hashCode();
      old.get(key).put(key2, s);
    }
    setCacheMaps(uuId, section, old);
  }

  public List<String> getCacheListValues(String uuId, String section, String key){
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Map<String, Object>> old = getCacheMaps(uuId, section, l);
    if(old == null)
      old = new HashMap<>();
    if(old.containsKey(key)){
      List<String> ret = new LinkedList<>();
      for(String s:old.keySet())
      {
        Object v = old.get(s);
        if(s.equals(LIST_PREFIX+v.toString().hashCode())){//get only list values
          ret.add(s);
        }
      }
      return ret;
    }
    return new LinkedList<>();
  }

  public Map<String, Object> getCacheMap(String uuId, String section, String key) {
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Map<String, Object>> old = getCacheMaps(uuId, section, l);
    if(old!=null)
      return old.get(key);
    return new HashMap<>();
  }

  public Map<String, List<Object>> getCacheMapByValue(String uuId, String string, String key) {
    List<String> l = new LinkedList<>();
    l.add(key);
    Map<String, Map<String, List<Object>>> old =
        getCacheMapsByValue(uuId,string,l);
    if(old!=null)
      return old.get(key);
    return new HashMap<>();
  }

  public Object getCacheValue(String uid, String section, String key) {
    //if (cs != null) { // && base instanceof ResourceStorage){
    initSection(prefix+section);
    Connection c = getConnection();
    readLock();
    try{
      //Statement s = c.createStatement();
      String exp = "SELECT value FROM '"+prefix+section+"' WHERE uid = ? AND key = ?";
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(2, key.toString());
        ps.setObject(1, uid.toString());
        rs = ps.executeQuery();
        if(rs.next())
            return rs.getObject(1);
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      readUnlock();
    }
    return null;
  }
  public List<String> getCacheKey2(String uid, String section, String key, String value) {
    initSection(prefix+section);
    Connection c = getConnection();
    List<String> ret = new LinkedList<>();
    readLock();
    try{
      //Statement s = c.createStatement();
      String exp = "SELECT key2 FROM '"+prefix+section+"' WHERE uid = ? AND key = ? AND value = ?";
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(3, value.toString());
        ps.setString(2, key.toString());
        ps.setObject(1, uid.toString());
        rs = ps.executeQuery();
        while(rs.next())
          ret.add(rs.getString(1));
        return ret;
      }catch(Exception e){
        e.printStackTrace();
      }
      finally{
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      readUnlock();
    }
    return new LinkedList<>();
  }
  public Object getCacheValue(String uid, String section, String key, String key2) {
    initSection(prefix+section);
    Connection c = getConnection();
    readLock();
    try{
      //Statement s = c.createStatement();
      String exp = "SELECT value FROM '"+prefix+section+"' WHERE uid = ? AND key = ? AND key2 = ?";
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setString(3, key2.toString());
        ps.setString(2, key.toString());
        ps.setObject(1, uid.toString());
        rs = ps.executeQuery();
        if(rs.next())
            return rs.getObject(1);
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{

        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      readUnlock();
    }
    return null;
  }


  public void setCacheMap(
     String uuId, String section, String key, Map<String, Object> hashMap) {
    Map<String, Map<String, Object>> old = new HashMap<>();
    old.put(key, hashMap);
    setCacheMaps(uuId, section, old);
  }

  public void addCacheMapValue(String uuid, String section, String key, String key2, Object value) {
    if(value == null)
      throw new RuntimeException("null value on setValue");
    initSection(prefix+section);
    //if(this.c == null)
    //old = getCacheMaps(uuid, section, vars.keySet());//TODO is it proper in not commit version?
    removeByKey(uuid, prefix+section, key,key2);
    Connection c = getConnection();
    writeLock();
    try{
      //Statement s = c.createStatement();
      String exp = "INSERT INTO '"+prefix+section+"'(uid,key,key2,value) VALUES (?,?,?,?)";
      PreparedStatement ps = null;
        try{
          ps = c.prepareStatement(exp);
  //        for (String s : map.keySet())
  //          vals.put(key + "|" + s, map.get(s));
          //for(String key2:vars.get(key).keySet()){
            ps.setString(1, uuid.toString());
            ps.setString(2, key);
            ps.setString(3, key2);
            ps.setObject(4, value);
            ps.executeUpdate();
          //}
          //ps.executeBatch();
        }catch(Exception e){
          e.printStackTrace();
          RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
        }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      writeUnlock();
    }
  }
  public boolean inTransaction() {
    return trans!=null;
  }
  public List<String> getAltSet(String section, String id) {
    initSection(prefix+section);
    Connection c = getConnection();
    List<String> ret = new LinkedList<>();
    readLock();
    try{
      //Statement s = c.createStatement();
      String exp = "SELECT key, value FROM '"+prefix+section+"' WHERE uid = ?";
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(exp);
        ps.setObject(1, id);
        rs = ps.executeQuery();
        while(rs.next()){
          ret.add(rs.getString(2));
        }
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{

        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      readUnlock();
    }
    return ret;
  }
  public void setAltSet(String section, String id, String[] vals) {
    initAltSection(prefix+section);
    Connection c = getConnection();
    writeLock();
    try{
      //Statement s = c.createStatement();
      removeByID(id, section);
      String exp = "INSERT INTO '"+prefix+section+"'(uid,key,value) VALUES (?,?,?)";
      PreparedStatement ps = null;
        try{
          ps = c.prepareStatement(exp);
  //        for (String s : map.keySet())
  //          vals.put(key + "|" + s, map.get(s));
          //for(String key2:vars.get(key).keySet()){
          for(String s:vals){
            ps.setString(1, id);
            ps.setString(2, "L"+s.hashCode());
            ps.setObject(3, s);
            ps.addBatch();
          }
          ps.executeBatch();
          //}
          //ps.executeBatch();
        }catch(Exception e){
          e.printStackTrace();
          RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
        }
      finally{
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      writeUnlock();
    }
  }
  public Map<String,List<String>> findSetsByValue(String section, List<String> ls) {
    if(ls.size() == 0)
      return Collections.emptyMap();
    initSection(prefix+section);
    Connection c = getConnection();
    Map<String,List<String>> ret = new HashMap<>();
    readLock();
    try{
      //Statement s = c.createStatement();
      StringBuilder sb = new StringBuilder("SELECT uid, value FROM '");
      sb.append(prefix);
      sb.append(section);
      sb.append("' WHERE value ");
      if(ls.size()>1){
        sb.append("in (");
        java.util.Iterator<String> it = ls.iterator();
        while(it.hasNext())
        {
          String next = it.next();
          sb.append("?");
          if(it.hasNext())
            sb.append(",");
        }
        sb.append(")");
      }else
        sb.append("= ?");
      PreparedStatement ps = null;
      ResultSet rs = null;
      try{
        ps = c.prepareStatement(sb.toString());
        int i = 1;
        for(String s:ls)
          ps.setObject(i++, s);
        rs = ps.executeQuery();
        while(rs.next()){
          String uid = rs.getString(1);
          Object obj = rs.getObject(2);
          String objs = obj.toString();
          if(!ret.containsKey(objs))
            ret.put(objs, new LinkedList<>());
          ret.get(objs).add(uid);
        }
      }catch(Exception e){
        e.printStackTrace();
        RLogger.logInfo(connectionPath+" error "+CoreUtils.getTrace(e));
      }
      finally{
        if(rs!=null)
          try {
            rs.close();
          } catch (SQLException e1) {
            e1.printStackTrace();
          }
        try{
          if(ps!=null)
            ps.close();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    }finally{
      readUnlock();
    }
    return ret;
  }
}
