package com.unitesk.requality.core;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ChangesSet <T>{
  private Set<String> changed,removed,added;
  private Map<String,T> olda;// = new HashMap<>();
  private Map<String,T> newa;// = new HashMap<>();
  private List<String> order = null;
  public ChangesSet(){
    changed = new HashSet<>();
    removed = new HashSet<>();
    added = new HashSet<>();
  }
  public ChangesSet(ChangesSet<T> changesb) {
    changed = new HashSet<>(changesb.changed);
    removed = new HashSet<>(changesb.removed);
    added = new HashSet<>(changesb.added);
    if(changesb.olda != null)
      olda = new HashMap<>(changesb.olda);
    if(changesb.newa != null)
      newa = new HashMap<>(changesb.newa);
    if(changesb.order!=null)
      order = new LinkedList<>(changesb.getOrder());
  }
  public List<String> getOrder() {
    return order;
  }
  public void setOrder(List<String> ord){
    assert order == null;
    order = ord;
  }
  public ChangesSet(Set<String> toadd, Set<String> tomodify, Set<String> todelete,Map<String,T>old,
      Map<String,T>newa,
      List<String> order) {
    changed = tomodify;
    if(changed == null)
      changed = new HashSet<>();
    removed = todelete;
    if(removed == null)
      removed = new HashSet<>();
    added = toadd;
    if(added == null)
      added = new HashSet<>();
    this.order = order;
    if(old!=null)
      this.olda = new HashMap<>(old);
    if(newa!=null)
      this.newa = new HashMap<>(newa);
  }
  private void addChange(String name, T o, T n){
    assert order == null;
    if(olda == null)
      olda = new HashMap<>();
    if(!olda.containsKey(name))
      olda.put(name, o);
    if(newa == null)
      newa = new HashMap<>();
    newa.put(name, n);
    if(removed.contains(name))
      removed.remove(name);
    if(added.contains(name))
      return;
    changed.add(name);
  }
  private void addCreate(String name, T o, T n){
    assert order == null;
    if(olda == null)
      olda = new HashMap<>();
    if(!olda.containsKey(name))
      olda.put(name, o);
    if(newa == null)
      newa = new HashMap<>();
    newa.put(name, n);
    if(removed.contains(name)){
      removed.remove(name);
      changed.add(name);
    }else{
      if(!changed.contains(name))
        added.add(name);
    }
  }
  private void addRemoved(String key, T o) {
    assert order == null;
    if(olda == null)
      olda = new HashMap<>();
    if(!olda.containsKey(key))
      olda.put(key, o);
    if(newa == null)
      newa = new HashMap<>();
    newa.put(key, null);
    added.remove(key);
    changed.remove(key);
    removed.add(key);
  }
  public boolean contains(String s) {
    if(newa == null)
      return false;
    return newa.containsKey(s);
  }
  public Set<String> getRemoved() {
    return removed;
  }
  public Set<String> getChanged() {
    return changed;
  }
  public Set<String> getAdded() {
    return added;
  }
  public boolean isEmpty() {
    return changed.isEmpty()&&added.isEmpty()&&removed.isEmpty();
  }
  public void clear() {
    changed.clear();
    added.clear();
    removed.clear();
    if(olda!=null)
      olda.clear();
    if(newa!=null)
      newa.clear();
    order = null;
  }
  public ChangesSet<T> merge(ChangesSet<T> changesb) {
    ChangesSet<T> ns = new ChangesSet<>();
    LinkedList<String> order = new LinkedList<>(changesb.getOrder());
    for(String s:this.order)
      if(!order.contains(s))
        order.add(s);
    for(String s:order){
      T old = this.getOld(s);
      if(old == null)
        old = changesb.getOld(s);
      T n = changesb.getNew(s);
      if(n == null)
        n = this.getNew(s);
      ns.addSomeChange(s, old, n);
    }
    ns.setOrder(order);
    return ns; 
  }
  public void addSomeChange(String s, T old, T n) {
//    if(olda==null)
//      olda = new HashMap<>();
//    if(newa == null)
//      newa
//    if(olda.containsKey(s))
//      old = olda.get(s);
//    olda.put(s, old);
//    newa.put(s, n);
    if(old == null && n!=null)
      addCreate(s, old, n);
    else
      if(n == null)
        addRemoved(s, old);
      else
        addChange(s,old,n);
    
  }
  //  public void addChanges(Set<String> rel, Set<String> set) {
//    for(String s:rel)
//      if(!set.contains(s))
//        addChange(s);
//  }
  public Set<String> getAll() {
    return newa.keySet();
  }
  public ChangesSet<T> getReverted() {
    List<String> or = order;
    if(order!=null)
      Collections.reverse(or);
    return new ChangesSet<>(new HashSet<>(removed),new HashSet<>(changed),
        new HashSet<>(added),newa,olda,or);
  }
  public T getOld(String key) {
    if(olda == null)
      return null;
    return olda.get(key);
  }

  public T getNew(String key) {
    if(newa == null)
      return null;
    return newa.get(key);
  }
  public Map<String, T> getOldMap() {
    return olda;
  }
  public Map<String, T> getNewMap() {
    return newa;
  }
  public void removeKey(String key) {
    removed.remove(key);
    changed.remove(key);
    added.remove(key);
    olda.remove(key);
    newa.remove(key);
    order.remove(key);
  }
  
}
