package com.unitesk.requality.core;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;
import java.util.regex.Pattern;
import org.apache.commons.httpclient.URIException;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Document.OutputSettings;
import org.jsoup.nodes.Entities.EscapeMode;
import org.jsoup.safety.Whitelist;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.VirtualNode;

public class CoreUtils {
  public static final String CACHE_FOLDER_NAME = ".reqcache";
  public static int DEBUG_LEVEL = 0;

  public static String getParentQualifiedId(String QId) {
    return QId.substring(0, QId.lastIndexOf("/"));
  }
  private static int countRTNChilds(TreeNode node) {
    int cnt = 0;
    for(TreeNode tn:node.getActiveWorkChildren())
      if(tn.getType().equals(Requirement.TYPE_NAME)||tn.getType().equals(TextNode.TYPE_NAME))
        cnt++;
    return cnt;
  }
  
  public static boolean isHeader(TreeNode node){
    String ctype = node.getType(); 
    boolean isReq = ctype.equals(Requirement.TYPE_NAME);
    boolean isTxt = ctype.equals(TextNode.TYPE_NAME);
    boolean isVN = ctype.equals(VirtualNode.TYPE_NAME);
    boolean isTS = ctype.equals("TestStep");
    if(isReq || isTxt || isVN || isTS){
      //  var child = getNodeByUUID(uuid);
        if (isTS || (isReq || isVN) && node.getActiveWorkChildren().length>0 && 
        (countRTNChilds(node))>0 
                || node.hasHeaderAttribute())
                    return true;
    }
    return false;
  }
  public static String splitWords(String s) {
    if (s == null) return null;
    for (int i = 1; i < s.length(); i++) // s.length() is variable
    if (s.charAt(i - 1) >= 'a'
          && s.charAt(i - 1) <= 'z'
          && s.charAt(i) >= 'A'
          && s.charAt(i) <= 'Z') {
        s = s.substring(0, i) + " " + s.substring(i);
      }
    return s;
  }

  public static String URLEncode(String s) {
    try {
      return org.apache.commons.httpclient.util.URIUtil.encodeQuery(s);
    } catch (URIException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  public static String htmlEscape(String str) {
    if (str == null) return null;
    return htmlSpaces(str.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;"));
  }

  public static String adaptTables(String str, String width, String targetwidth) {
    int ind = str.indexOf("<table");
    int last = 0;
    while(ind>-1)
    {
      last = ind;
      int end = str.indexOf(">",ind);
      int w = str.indexOf("width");
      if(w<end){
        int c = w+6;
        while(Character.isWhitespace(str.charAt(c)))
          c++;
        char ch = str.charAt(c);
        char br = ' ';
        if(!Character.isDigit(ch))
          br = ch;
        else 
          c--;
        int wend =  Math.min(str.indexOf(br,c+1), end)-1;
        String wdth = str.substring(c+1,wend);
        if(wdth.contains(width))
          str = str.substring(last, c)+targetwidth+str.substring(wend);
      }
      ind = str.indexOf("<table",ind+10);
    }
    return str;
  }
  public static String htmlSpaces(String str) {
    if (str == null) return null;
    return str.replace("  ", (char) 160 + "" + (char) 160);
  }

  public static String listToString(List<String> list, String separator) {
    return listToString(list.toArray(), separator);
  }

  public static String listToString(Object[] list, String separator) {
    return listToString(list, separator, false);
  }

  public static String listToString(Object[] list, String separator, boolean b) {
    String r = "";
    String qu = b ? "'" : "";
    if (list == null) list = new Object[0];
    for (int i = 0; i < list.length; i++)
      if (b && !list[i].toString().startsWith("'")) r += qu + list[i] + qu + separator;
      else r += list[i] + separator;
    if (r.equals("")) return r;
    return r.substring(0, r.length() - separator.length());
  }

  private static Pattern tagFind = Pattern.compile("\\<[/]?([a-zA-Z]+)(?:[^>]*)\\>");
  private static Map<String, Stack<String>> gtags = new HashMap<>(),
      gclosed = new HashMap<>(),
      gclosedNames = new HashMap<>(),
      gtagsBasic = new HashMap<>();
  public static String htmlEncapsulate(String newDescr) {
    return htmlEncapsulate(newDescr, false);
  }
  /**
   * Method to close all opened tags and fix some problems with tags overlap. Function only
   * considering general structure of tags without deep analysis. Single tags needs to be provide in
   * accordance with XHTML style like "<br>
   * " instead of "<br>
   * ". Otherwise closing tag will be added. Example: for string "<b>one<i> more</b><u><img/> <br>
   * " function will return "<b>one<i> more</i></b><u><img/> <br>
   * </br> </u>". "<br>
   * " is not comply with XHTML standard and therefore it is provided by closing tag.
   *
   * @param xhtml markup with possible unclosed and overlapping tags
   * @return correct xhtml markup
   */
  public static String htmlEncapsulate(String newDescr, boolean replamp) {
    if (newDescr == null) return null;
    Whitelist wl = Whitelist.relaxed();
    wl.addTags("svg");
//    wl.addTags("img");
//    wl.addAttributes("img", "alt");
//    wl.addAttributes("img", "src");
//    wl.addProtocols("a", "href", "requality");
//    wl.addTags("table", "td", "tr");
    OutputSettings os = new OutputSettings();
    os.syntax(Document.OutputSettings.Syntax.xml);
    os.escapeMode(EscapeMode.xhtml);
    os.prettyPrint(false);
    os.charset(Charset.forName("UTF-8"));
    Document doc = 
        Jsoup.parse(newDescr);
    doc.outputSettings(os);
    String ret = doc.body().html();
//    if(replamp)
//      ret = replAmp(ret);
    return ret;
  }

  private static String replAmp(String string) {
    int ind = string.indexOf("&");
    while (ind > -1) {
      int pos = ind + 1;
      int end = -1;
      while (pos < string.length() && end == -1) {
        char c = string.charAt(pos);
        if (!Character.isLetterOrDigit(c))
          if (c == ';') end = pos;
          else end = -2;
        pos++;
      }
      if (end > -1) {
        String substr = string.substring(ind + 1, end);
        String repl = HTMLSpecChars.getCode(substr);
        if (repl != null) {
          repl = "&#" + repl;
          string = string.substring(0, ind) + repl + string.substring(end);
          ind += repl.length() - substr.length();
          ;
        }
      }
      ind = string.indexOf("&", ind + 1);
    }
    return string;
  }

  private static String replaceBracket(String substring) {
    int bpos = substring.indexOf("<");
    while (bpos > -1) {
      substring = substring.substring(0, bpos) + "&lt;" + substring.substring(bpos + 1);
      bpos = substring.indexOf("<", bpos + 1);
    }
    bpos = substring.indexOf(">");
    while (bpos > -1) {
      substring = substring.substring(0, bpos) + "&gt;" + substring.substring(bpos + 1);
      bpos = substring.indexOf(">", bpos + 1);
    }
    return substring;
  }

  public static String htmlClear(String content) {
    int lpos = content.indexOf("<");
    while (lpos > -1) {
      int rpos = content.indexOf(">", lpos);
      content = content.substring(0, lpos) + content.substring(rpos + 1);
      lpos = content.indexOf("<", lpos * 2 - rpos + 1);
    }
    return content;
  }

  public static int calcDist(String s1, String s2) {
    if (s1 == null) return Integer.MAX_VALUE;
    if (s2 == null) return Integer.MAX_VALUE;
    int m = s1.length(), n = s2.length();
    int[] d1;
    int[] d2 = new int[n + 1];

    for (int i = 0; i <= n; i++) d2[i] = i;

    for (int i = 1; i <= m; i++) {
      d1 = d2;
      d2 = new int[n + 1];
      for (int j = 0; j <= n; j++) {
        if (j == 0) d2[j] = i;
        else {
          int cost = (s1.charAt(i - 1) != s2.charAt(j - 1)) ? 1 : 0;
          if (d2[j - 1] < d1[j] && d2[j - 1] < d1[j - 1] + cost) d2[j] = d2[j - 1] + 1;
          else if (d1[j] < d1[j - 1] + cost) d2[j] = d1[j] + 1;
          else d2[j] = d1[j - 1] + cost;
        }
      }
    }
    return d2[n];
  }

  public static String TABLES_REFS = "relates-table";

  public static String reworkLibreDesc(
      TreeNode node,
      String desc,
      boolean replace,
      Set<TreeNode> tosave,
      Map<UUID, String> doorsids) {
    int rpos = desc.indexOf("[");
    while (rpos > -1) {
      int min = Math.max(0, rpos - 200);
      int apos = -1;
      int hpos = desc.substring(min, rpos).lastIndexOf("href");
      if (hpos > -1) {
        hpos = hpos + min;
        apos = desc.substring(min, rpos).lastIndexOf("<a") + min;
        char quote = '"';
        int mode = 0;
        while (mode < 3) {
          char c = desc.charAt(hpos);
          if (c == '=' && mode == 0) mode = 1;
          else if ((c == '\'' || c == '"') && (mode == 1)) {
            quote = c;
            mode = 2;
          } else if (mode == 2) {
            mode = 3;
            String link = desc.substring(hpos, desc.indexOf(quote, hpos)).trim();
            if (link.startsWith("#")) link = link.substring(1);
            if (link.indexOf("|") > -1) link = link.substring(0, link.indexOf("|"));
            List<TreeNode> rnode = null;
            try {
              rnode = node.getTreeDB()
                 .getLinkResolver()
                 .getNotDisposedResolvePath(
                    node,
                    null,
                    node.getTreeDB(),
                    link,
                    Path.XPATH | Path.QID | Path.UUID | Path.UV);
            } catch (NodeNotFoundException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
            // TODO warn on not found db
            if (rnode == null || rnode.size() == 0) {
              rnode = new LinkedList<>();
              rnode.add(node.getTreeDB().getByTags(node, link));
            }
            if (rnode.size() > 0)
              if (rnode.iterator().next() == null) rnode = null;
              else link = rnode.iterator().next().getQualifiedId();

            if (rnode != null
                && rnode.size() > 0
                && replace
                && doorsids != null
                && doorsids.containsKey(rnode.iterator().next().getUUId())) {
              desc =
                  desc.substring(0, apos)
                      + doorsids.get(rnode.iterator().next().getUUId())
                      + desc.substring(desc.indexOf("</a>", apos));
              rpos = apos;
            }
          }

          hpos++;
        }
      }
      rpos = desc.indexOf("[", rpos + 1);
    }
    return desc;
  }

  public static String removeTag(String name, String s) {
    int pos = s.indexOf("<" + name);
    while (pos > -1) {
      s = s.substring(0, pos) + s.substring(s.indexOf(">", pos) + 1, s.length());
      pos = s.indexOf("<" + name);
    }
    return s;
  }

  public static String getTrace(Exception ex) {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    PrintWriter pw = new PrintWriter(baos);
    ex.printStackTrace(pw);
    pw.close();
    try {
      baos.close();
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    try {
      return baos.toString("UTF-8");
    } catch (UnsupportedEncodingException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return "Incorrect encoding";
  }

  private static java.nio.file.Path dir = null;
  public static java.nio.file.Path getCacheDir(java.nio.file.Path p) {
    File f = p.toFile();
    File nf = f.toPath().resolve(CACHE_FOLDER_NAME).toFile();
    
    if(!nf.exists())
      nf.mkdirs();
    File fi = f.toPath().resolve(".gitignore").toFile();
    try {
      addIgnore(fi, CACHE_FOLDER_NAME+"/");
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return nf.toPath();
  }
  private static void addIgnore(File fi,String st) throws IOException {
    if(!fi.exists())
      try {
        fi.createNewFile();
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    FileReader fr = new FileReader(fi);
    BufferedReader br = new BufferedReader(fr);
    boolean found = false;
    boolean single = false;
    try{
      while(br.ready()){
        String s = br.readLine();
        single = true;
        if(s.equals(st)){
          found = true;
          break;
        }
      }
    }finally{
      br.close();
      fr.close();
    }
    if(!found)
    {
      FileWriter fw = new FileWriter(fi,true);
      BufferedWriter bw = new BufferedWriter(fw);
      try{
        if(single)
          bw.write("\n");
        bw.append(st);
      }
      finally{
        bw.close();
        fw.close();
      }
    }
  }
  public static java.nio.file.Path getTempDir() {
    if(dir == null)
      try {
        dir = Files.createTempDirectory("requality");
        File df = dir.toFile();
        if(!df.exists())
          df.mkdirs();
        df.deleteOnExit();
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    return dir;
  }
  public static String getTrace(StackTraceElement[] stackTrace) {
    StringBuilder sb = new StringBuilder();
    for(StackTraceElement s:stackTrace){
      sb.append(s);
      sb.append("\n");
    }
    return sb.toString();
  }
  //Pattern.compile("[_]{0,2}[%]{0,1}[A-Za-z0-9а-яА-Я_[$].,-]+[A-Za-z0-9а-яА-Я,._[$] -]*");
  private static Set<Character> allowedIdChars = new HashSet<>();
  static{
    allowedIdChars.add('_');
    allowedIdChars.add('$');
    allowedIdChars.add(',');
    allowedIdChars.add('.');
    allowedIdChars.add('-');
    
  }
  public static String filterNodeId(String nameOrId) {
    StringBuilder sb = new StringBuilder();
    for(int i = 0;i<nameOrId.length();i++){
      char c = nameOrId.charAt(i);
      if(Character.isLetterOrDigit(c)||allowedIdChars.contains(c)||Character.isWhitespace(c))
        sb.append(c);
    }
    return sb.toString();
  }
}
