package com.unitesk.requality.core;

import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.ErrorNode;
import com.unitesk.requality.nodetypes.InstancesFolder;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.RestrictedFolder;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.ToDoItem;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.report.Template;

public class FilesTracker implements IResTracker {
  private static final String projEx =
      "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<projectDescription>\n<name>{name}</name>\n<comment></comment>\n<projects>\n</projects>\n"
          + "<buildSpec>\n<buildCommand>\n<name>com.unitesk.requality.checkers</name>\n<arguments>\n</arguments>\n</buildCommand>\n</buildSpec><natures><nature>com.unitesk.requality.natures.reqcatalogue</nature>\n"
          + "</natures>\n</projectDescription>";
  private String path = null, templpath = null;;
  private java.nio.file.Path ppath = null;
  private Map<String, TreeDB> dbs;
  private Map<TreeDB, String> dbnames;
  private Map<TreeDB, java.nio.file.Path> locations;

  @Override
  public TreeDB[] getTreeDBs() {
    return dbs.values().toArray(new TreeDB[0]);
  }

  @Override
  public TreeDB createProject(String name) {
    File f = new File(URI.create(path + "/" + name));
    if (f.exists())
      throw new RuntimeException("Project with name '" + name + "' is already exists");
    f.mkdirs();
    TreeDB.create(name, new FSStorage(f.toURI().toString()));
    emulateDotProjectCreate(name, new File(URI.create(path + "/" + name + "/" + ".project")));
    loadDB(f);
    return dbs.get(name);
  }

  private void emulateDotProjectCreate(String name, File file) {
    try {
      FileWriter fw = new FileWriter(file);
      BufferedWriter bw = new BufferedWriter(fw);
      bw.write(projEx.replace("{name}", name));
      bw.close();
      fw.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private static FileFilter ff =
      new FileFilter() {

        @Override
        public boolean accept(File pathname) {
          return pathname.getName().equals(".project");
        }
      };

  private static FileFilter root =
      new FileFilter() {

        @Override
        public boolean accept(File pathname) {
          return pathname.getName().equals("root.json");
        }
      };

  @Override
  public void init(String... params) {
    if (params.length != 2)
      throw new RuntimeException(
          "Two parameters should be provided - path to workspace and path to templates folder");
    path = params[0];
    templpath = params[1];
    dbs = new HashMap<>();
    dbnames = new HashMap<>();
    locations = new HashMap<>();
    File f = new File(URI.create(path));
    if (!f.exists()) f.mkdirs();
    ppath = f.toPath();
    for (File ch : f.listFiles()) {
      if (ch.isDirectory()) {
        File[] fa = ch.listFiles(ff);
        if (fa != null && fa.length == 1)
          if (findString(fa[0], "com.unitesk.requality.natures.reqcatalogue")) {
            File[] rt = ch.listFiles(root);
            if (rt.length == 1) {
              loadDB(ch);
            }
          }
      }
    }
    File sidep = f.toPath().resolve(".metadata/.plugins/org.eclipse.core.resources/.projects").toFile();
    if(sidep.exists())
      for(File pp : sidep.listFiles()){
        if(pp.exists() && pp.isDirectory()){
          File loc = pp.toPath().resolve(".location").toFile();
          if(loc.exists()){
            byte[] trash = new byte[16];
            try{
              FileInputStream fis = new FileInputStream(loc);
              DataInputStream dis = new DataInputStream(fis);
              dis.read(trash);
              String sos = dis.readUTF();
              dis.close();
              fis.close();
              trackTreeDB(URI.create(sos.substring(5)));
            }
            catch(Exception e){
              e.printStackTrace();
            }
          }
        }
      }
  }

  private void loadDB(File ch) {
    TreeDB db = TreeDB.load(new FSStorage(ch.toURI().toString()));
    initClasses(db);
    dbs.put(ch.getName(), db);
    dbnames.put(db, ch.getName());
    locations.put(db, ch.toPath());//TODO check
    db.setIsLoaded();
    Instantiator.removeTempInstances(db);
    db.getLinkResolver().initCaches();
    db.getTermsStorage().initCaches();
    Instantiator.loadDBs(db);
  }

  private boolean findString(File fa, String string) {
    List<String> ln = null;
    try {
      ln = Files.readAllLines(fa.toPath());
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    if(ln == null)
      return false;
    StringBuffer sb = new StringBuffer();
    boolean found = false;
    Iterator<String> it = ln.iterator();
    while (!found && it.hasNext()) {
      String l = it.next();
      l = l.trim();
      sb.append(l);
      if (sb.indexOf(string) > -1) {
        found = true;
      }
      if (sb.length() > string.length()) sb.delete(0, sb.length() - string.length());
    }
    return found;
  }

  @Override
  public String getName(TreeDB db) {
    return dbnames.get(db);
  }

  @Override
  public TreeDB getTreeDB(String name) {
    return dbs.get(name);
  }

  public void recDelete(File f) {
    if (f.isDirectory()) for (File f2 : f.listFiles()) recDelete(f2);
    f.delete();
  }

  @Override
  public void remove(String name) {
    TreeDB db = dbs.remove(name);
    if (db != null) dbnames.remove(db);
    File f = new File(URI.create(path + "/" + name));
    if (f.exists()) recDelete(f);
    // TODO look
  }

  @Override
  public String getLocation(TreeDB treeDB) {
    java.nio.file.Path location = locations.get(treeDB);
    if(location == null){
      location = ppath.resolve(getName(treeDB));
      locations.put(treeDB, location);
    }
    return location.toUri().toString();
  }

  @Override
  public Map<String, Template> getReportTemplates() {
    return null;
  }

  @Override
  public void clearCache(TreeDB db) {
    String name = dbnames.remove(db);
    if (name != null) dbs.remove(name);
    // TODO Auto-generated method stub

  }

  @Override
  public void trackProject(String name, TreeDB db) {
    dbs.put(name, db);
    dbnames.put(db, name);
  }

  @Override
  public void reOpen(TreeDB treeDB) {
    // TODO maybe add close-open analog
  }

  @Override
  public void initClasses(TreeDB db) {
    db.addNodeType(Comment.TYPE_NAME, Comment.class);
    db.addNodeType(DocFolder.TYPE_NAME, DocFolder.class);
    db.addNodeType(Document.TYPE_NAME, Document.class);
    db.addNodeType(ErrorNode.TYPE_NAME, ErrorNode.class);
    db.addNodeType(InstancesFolder.TYPE_NAME, InstancesFolder.class);
    db.addNodeType(Location.TYPE_NAME, Location.class);
    db.addNodeType(ReportData.TYPE_NAME, ReportData.class);
    db.addNodeType(ReportSettings.TYPE_NAME, ReportSettings.class);
    db.addNodeType(Requirement.TYPE_NAME, Requirement.class);
    db.addNodeType(RestrictedFolder.TYPE_NAME, RestrictedFolder.class);
    db.addNodeType(TestPurpose.TYPE_NAME, TestPurpose.class);
    db.addNodeType(TextNode.TYPE_NAME, TextNode.class);
    db.addNodeType(ToDoItem.TYPE_NAME, ToDoItem.class);
    db.addNodeType(VirtualNode.TYPE_NAME, VirtualNode.class);
  }

  @Override
  public TreeDB trackTreeDB(URI path) throws IOException {
    return trackTreeDB(path, null);
  }

  @Override
  public TreeDB trackTreeDB(URI path, String name) throws IOException {
    File f = new File(path);
    if(!f.exists())
      throw new FileNotFoundException("Directory on path " +path+" is not exist");
    if(!f.isDirectory())
      throw new java.io.IOException("Object on path "+path+" is file instead of directory");
    File[] fs = f.listFiles(ff);
    if(fs.length==0)
      throw new FileNotFoundException(".project file is not found in given dir "+path);
    if (!findString(fs[0], "com.unitesk.requality.natures.reqcatalogue"))
      throw new RuntimeException("Project on given path is not a Requality project "+path);
    List<String> ln = Files.readAllLines(fs[0].toPath());
    if(name == null){
      name = f.getName();
      for(String s : ln){
        int i = s.indexOf("<name>");
        if(i > 0){
          name = s.substring(i+6, s.indexOf("</name>",i+1));
          break;
        }
      }
    }
    TreeDB db = TreeDB.load(new FSStorage(path.toString()));
    initClasses(db);
    dbs.put(name, db);
    dbnames.put(db, name);
    db.setIsLoaded();
    Instantiator.removeTempInstances(db);
    Instantiator.loadDBs(db);
    db.getLinkResolver().initCaches();
    db.getTermsStorage().initCaches();
    return db;
  }

  @Override
  public Path getRootLocation() {
    return ppath;
  }

  @Override
  public Path getGlobalCacheLocation() {
    return ppath;
  }
}
