package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.UUID;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.InvalidXPathException;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.domproxy.RequalityElement;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;

public class LinkResolver {
  public static final int RESOLVE_MODE = LinksStorage.RESOLVE_MODE;
  TreeDB db;
  CacheStorage cs;

  public LinkResolver(CacheStorage cs) {
    this.cs = cs;
    //if (db != null) db.addListener(changeListener);
  }

  
  public LinkResolver(TreeDB db) {
    this.db = db;
    if (db != null) db.addListener(changeListener);
  }

  protected CacheStorage getCacheStorage(){
    if(db!=null)
      return db.getCacheStorage();
    return cs;
  }
  
  public void close() {
    if(db!=null)
      db.removeListener(changeListener);
  }

  private LinksStorage ls = null;
  
  public LinksStorage getStorage() {
    //if (db == null || !db.isVirtual()) {
      if (ls == null) ls = new LinksStorage(this);
      return ls;
//    }
//    if (db.isVirtual()) {
//      if (Instantiator.getBaseTree(db) == null) {
//        return null;
//      }
//      TreeDB base = Instantiator.getBaseTree(db);
//      if (base != null) return base.getLinkResolver().getStorage();
//      return ls;
//    }
  //  return ls;
  }

  private static Map<RequalityElement, Document> proxy = new HashMap<>();
  //private HashMap<UUID, Attribute> vers = new HashMap<>();
//  private HashMap<String, Set<UUID>> namecache = new HashMap<>();
//  private HashMap<UUID, String> backnamecache = new HashMap<>();

  private Map<String, Set<LateResolveRequest>> lates = new HashMap<>();

  private void registerLateResolve(String name, TreeNode node, String key) {
    if (!lates.containsKey(name)) lates.put(name, new HashSet<LateResolveRequest>());
    lates.get(name).add(new LateResolveRequest(node, key));
    saveLates();
  }

  public void loadLates() {
    lates.clear();
    if (db.getRootNode() == null) return;
    List<String> lv =
        getCacheStorage()
            .getCacheListValues("lateLinksCache", "lateLinksCache", "lateLinksCache");
    if (lv != null)
      for (String s : lv) {
        if (s == null) continue;
        String[] spl = s.split("\\|");
        String name = spl[0];
        String uid = spl[1];
        String key = spl[2];
        if (!name.isEmpty() && !uid.isEmpty() && !key.isEmpty()) {
          if (!lates.containsKey(name)) lates.put(name, new HashSet<LateResolveRequest>());
          try {
            TreeNode node = db.getNode(UUID.fromString(uid));
            if (node != null) lates.get(name).add(new LateResolveRequest(node, key));
          } catch (Exception e) {
            e.printStackTrace();
          }
        }
      }
  }

  public void saveLates() {
    List<String> tosave = new LinkedList<>();
    for (String n : lates.keySet())
      for (LateResolveRequest r : lates.get(n)) {
        StringBuilder sb = new StringBuilder();
        sb.append(n);
        sb.append("|");
        if (r.node != null) sb.append(r.node.getUUId());
        sb.append("|");
        if (r.key != null) sb.append(r.key);
        tosave.add(sb.toString());
      }
    getCacheStorage()
        .setCacheListValue(
            "lateLinksCache", "lateLinksCache", "lateLinksCache", tosave);
  }

  public NodeDesc[] getNodesByUFName(NodeDesc db, String path, Map<UUID, NodeDesc> map) {
    return getNodesByUFName(db, path, null, map, null);
  }

  /**
   * Method to get all possible elements with given User-Visible name Method to get Array of UUId's
   * of elements with given User Visible Id
   *
   * @param path array of uuids
   */
  @Deprecated
  public TreeNode[] getNodesByUFName(TreeDB db, String path) {
    return getNodesByUFName(db, path, null);
  }
  /**
   * Method to get Array of UUId's of elements with given User Visible Id
   *
   * @param path array of uuids
   */
  @Deprecated
  public TreeNode[] getNodesByUFName(TreeDB db, String path, List<String> types) {
    int pos = path.indexOf("::");
    TreeDB bdb = db;
    if (pos > -1) {
      String name = path.substring(0, pos);
      db = ResManager.getManager().getTreeDB(name);
      if (db != null) path = path.substring(pos + 2);
    }
    if (db != null) return db.getLinkResolver().getNodesByUFName(db, path, types);
    Set<TreeNode> curnodes = new HashSet<>();
    StringTokenizer st = new StringTokenizer(path, "/");
    if (!st.hasMoreTokens()) return new TreeNode[0];
    String nodeId = st.nextToken();
    //if we have node with name nodeId in requirement then use it
    TreeNode rnode =
        bdb.getRootNode()
            .findSingleWorkChildPlusName("Requirements")
            .findSingleWorkChildPlusName(nodeId);
    if (rnode != null) curnodes.add(rnode);

    rnode = bdb.getRootNode().findSingleWorkChildPlusName(nodeId);
    if (rnode != null) curnodes.add(rnode);
    //if cache of Requiremens names contains nodeId
    //then try to get appropriate node
    curnodes.addAll(findByName(bdb, nodeId));
    return findNodes(curnodes, st);
  }

  private Set<TreeNode> findByName(TreeDB db, String nodeId) throws NodeNotFoundException {
    Set<TreeNode> ret = new HashSet<>();
    if (db.getLinkResolver().getStorage().containsName(nodeId)) {
      Set<UUID> nodes = db.getLinkResolver().getStorage().getByName(nodeId);
      if (nodes.size() < 1) return ret;
      for (UUID uid : new HashSet<>(nodes)) {
        try{
          TreeNode nde = db.getNodeExc(uid);
          if (!ret.contains(nde) && nde != null) ret.add(nde);
        }
        catch(NodeNotFoundException nfe){
        }
      }
    }
    return ret;
  }

  private TreeNode[] findNodes(Set<TreeNode> curnodes, StringTokenizer st) {
    if (curnodes.size() == 0) {
      return new TreeNode[0];
    }

    String[] tokens = new String[st.countTokens()];
    int i = 0;
    while (st.hasMoreTokens()) tokens[i++] = st.nextToken();

    Iterator<TreeNode> it = curnodes.iterator();
    LinkedList<TreeNode> ll = new LinkedList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      List<TreeNode> uids = findNodePlusName(node, tokens);
      if (uids != null) ll.addAll(uids);
    }
    return ll.toArray(new TreeNode[0]);
  }

  //	    private List<TreeNode> findRecNodePlusName(List<TreeNode> cur,String s){
  //	    	List<TreeNode> ret = new ArrayList<TreeNode>();
  //	    	Stack<TreeNode> tohandle = new Stack<TreeNode>();
  //			tohandle.addAll(cur);
  //			while(!tohandle.isEmpty()){
  //				TreeNode nde = tohandle.pop();
  //				if(VirtualNode.TYPE_NAME.equals(nde.getType())){
  //					if(((VirtualNode)nde).isCollapsed())
  //						tohandle.addAll(nde.get)
  //				}
  //				if(s.equals(nde.getName())||nde.getId().equals(s))
  //					ret.add(nde);
  //			}
  //}
  public List<TreeNode> findNodesPlusName(TreeNode nde, String s) {
    List<TreeNode> ret = new ArrayList<>();
    for (TreeNode node : nde.getShownChildren())
      //coolapsed vnodes tends to be hidden in shownchildren
      if (node.getId().equals(s)
          || s.equals(node.getName())
          || node.getRawId().equals(s)
          || node.getUUId().toString().equals(s)) ret.add(node);
    return ret;
  }

  public List<TreeNode> findNodePlusName(TreeNode node, String[] tokens) {
    List<TreeNode> cur = new ArrayList<>();
    cur.add(node);
    for (String s : tokens) {
      List<TreeNode> newgen = new ArrayList<>();
      for (TreeNode node2 : cur) newgen.addAll(findNodesPlusName(node2, s));
      cur = newgen;
      if (cur.isEmpty()) return cur;
    }
    return cur;
  }

  private String getName(NodeDesc nd) {
    if (nd != null) {
      AttributeDefinition al = nd.getAttributes().get(TreeNode.ATTR_NAME);
      if (al != null) {
        Object val = al.getRawValue().getValue();
        return val == null ? null : val.toString();
      }
    }
    return null;
  }

  private LinkedList<NodeDesc> formCurChilds(
      List<UUID> childs, String nodeId, Map<UUID, NodeDesc> nds) {
    LinkedList<NodeDesc> curnodes = new LinkedList<>();
    if (childs != null)
      for (UUID chi : childs) {
        NodeDesc rnode = null;
        NodeDesc nd = nds.get(chi);
        if (nd.getId().equals(nodeId)) rnode = nd;
        String name = getName(nd);
        if (name != null && name.equals(nodeId)) rnode = nd;
        if (rnode != null) curnodes.add(rnode);
        if (VirtualNode.TYPE_NAME.equals(nd.getType())) {
          if (nd.getAttributes().containsKey(VirtualNode.ATTR_COLLAPSEVNODE)){
            AttributeDefinition ad = nd.getAttributes().get(VirtualNode.ATTR_COLLAPSEVNODE);
            if (ad!=null && ad.getRawValue()!=null && (boolean) ad.getRawValue().getValue()
                == true) {
              curnodes.addAll(
                  formCurChilds(nd.getChildrenUUId(), nodeId, nds)); //look in hidden vnode
            }
          //TODO check
          }
        }
      }
    return curnodes;
  }

  public NodeDesc[] getNodesByUFName(
      NodeDesc root,
      String path,
      List<String> types,
      Map<UUID, NodeDesc> nds,
      Map<UUID, LinkedList<UUID>> childfunc) {
    LinkedList<NodeDesc> curnodes = new LinkedList<>();
    TreeDB db = null;
    int pos = path.indexOf("::");
    if (pos > -1) {
      String name = path.substring(0, pos);
      db = ResManager.getManager().getTreeDB(name);
      if (db != null) path = path.substring(pos + 2);
      else throw new RuntimeException("Project '" + name + "' was not found or can not be loaded");
    }
    if (db != null) return db.getLinkResolver().getNodesByUFName(root, path, types, nds, childfunc);
    StringTokenizer st = new StringTokenizer(path, "/");
    if (!st.hasMoreTokens()) return new NodeDesc[] {root}; //TODO test
    String nodeId = st.nextToken();
    //if we have node with name nodeId in requirement then use it

    List<UUID> childs = null;
    if (childfunc != null) childs = childfunc.get(root.getUUId());
    else childs = root.getChildrenUUId();
    curnodes = formCurChilds(childs, nodeId, nds);
    //	        if(rnode != null)
    //	        	curnodes.add(rnode);
    //if cache of Requiremens names contains nodeId
    //then try to get appropriate node
    if (getStorage().containsName(nodeId)) {
      Set<UUID> nodes = getStorage().getByName(nodeId);
      if (nodes.size() < 1) return null;
      for (UUID uid : nodes) {
        NodeDesc nde = nds.get(uid);
        if (!curnodes.contains(nde)) curnodes.add(nde);
      }
    }
    if (curnodes.size() == 0) {
      return new NodeDesc[0];
    }

    String[] tokens = new String[st.countTokens()];
    int i = 0;
    while (st.hasMoreTokens()) tokens[i++] = st.nextToken();

    Iterator<NodeDesc> it = curnodes.iterator();
    LinkedList<NodeDesc> ll = new LinkedList<>();
    while (it.hasNext()) {
      NodeDesc node = it.next();
      NodeDesc uid = findNodeDesc(node, tokens, nds, childfunc);
      if (uid != null) ll.add(uid);
    }
    return ll.toArray(new NodeDesc[0]);
  }
  /**
   * Method to update cached requirement name.
   *
   * @param node requirement to update name
   */
  public void updateName(UUID uid, String newName) {
    getStorage().updateName(uid,newName);
  }

  //	public List<TreeNode> findNodesPlusName(TreeNode node, String[]tokens ){
  //	   	TreeNode temp = node;
  //	   	List<UUID> res = new ArrayList<UUID>();
  //	   	for(int i = 0;i<tokens.length&&temp!=null;i++){
  //	   		List<>
  //	   		temp = temp.findChildPlusName(tokens[i]);
  //	   	}
  //	  	return temp == null?null:temp.getUUId();
  //	}
  public NodeDesc findNodeDesc(
      NodeDesc node,
      String[] tokens,
      Map<UUID, NodeDesc> nds,
      Map<UUID, LinkedList<UUID>> childfunc) {
    NodeDesc temp = node;
    for (int i = 0; i < tokens.length && temp != null; i++) {
      if(tokens[i].isEmpty())
        continue;
      Collection<UUID> chi = null;
      if (childfunc != null) chi = childfunc.get(temp.getUUId());
      else chi = temp.getChildrenUUId();
      if (chi == null) return null;
      temp = null;
      for (UUID uid : chi) {
        NodeDesc nd = nds.get(uid);
        if (nd.getId().equals(tokens[i])) {
          temp = nd;
          continue;
        }
        String name = getName(nd);
        if (name != null && name.equals(tokens[i])) {
          temp = nd;
          continue;
        }
        if (node.getUUId().toString().equals(tokens[i])) {
          temp = nd;
          continue;
        }
      }
      if (temp == null) return null;
    }
    return temp;
  }
  
  
  public List<TreeNode> getNotDisposedResolvePath(TreeNode node, String akey, TreeDB db, String qId, int mode) throws NodeNotFoundException {
    return resolvePath(node, akey, db, qId, mode, true);
  }

  public List<TreeNode> resolvePath(TreeNode node, String akey, TreeDB db, String qId, int mode) throws NodeNotFoundException {
    return resolvePath(node, akey, db, qId, mode, false);
  }
  public List<TreeNode> resolvePath(TreeNode node, String akey, TreeDB db, String qId, int mode,boolean denydisposed) throws NodeNotFoundException {
    //TODO multimethod
    if (qId == null) return null;
    if (node != null){
      db = node.getTreeDB();
      if(node.isVRoot())
        db = node.getVirtualDB();
    }
    int pos = qId.indexOf("::");
    if (pos > -1) {
      String name = qId.substring(0, pos);
      TreeDB adb = null;
      adb = ResManager.getManager().getTreeDB(name);
      if (adb != null) {
        //adb.addListener(changeListener);
        qId = qId.substring(pos + 2);//.replace(".", "/");
        db = adb;
      } else 
        return Collections.emptyList();//db is not loaded
    }
    Set<TreeNode> ret = new HashSet<>();
    TreeNode root = db.getRootNode();
    if ((mode & Path.UUID) == Path.UUID) {
      try {
        UUID uid = UUID.fromString(qId);
        if(uid!=null){
          TreeNode nde = db.getNodeExc(uid);
          if (nde != null && (!denydisposed||!nde.isDisposed())){
            LinkedList<TreeNode >ll = new LinkedList<>();
            ll.add(nde);
            return ll;
          }
          return Collections.emptyList();
        }
      } catch (IllegalArgumentException ie) {
        //incorrect uid
      }
    }
    if (((mode & Path.QID) == Path.QID || (mode & Path.UV) == Path.UV)
        && (qId.equals("/") || qId.isEmpty())) {
      return Arrays.asList(new TreeNode[] {root});
    }
    TreeNode[] tnodes = null;
    try{
        if ((mode & Path.XPATH) == Path.XPATH) tnodes = resolveRefPath(db.getRootNode(), qId);
    }
    catch(Exception e){
        //incorrect xpath, //TODO message
    }
    if (tnodes != null && tnodes.length > 0)
      for(TreeNode nde:tnodes)
        if(!denydisposed || !nde.isDisposed())
          ret.add(nde);
        //ret.addAll(Arrays.asList(tnodes));
    qId = qId.replace((char) 160, ' ');
    if (((mode & Path.UV) == Path.UV || (mode & Path.QID) == Path.QID)) {
      StringTokenizer st = new StringTokenizer(qId, "/");
      String step = st.nextToken();
      Set<TreeNode> initSet = new HashSet<>();
      TreeNode rt = db.getRootNode();
      if (rt != null) {
        initSet.add(db.getRootNode());
        rt =
            db.getRootNode().findSingleWorkChildPlusName(Requirement.getTypeRootQId().substring(1));
        if (rt != null && (!denydisposed||!rt.isDisposed())) initSet.add(rt);
      }
      initSet = findStep(db, initSet, step, (mode & Path.CREATE) == Path.CREATE);
      if ((mode & Path.UV) != 0) initSet.addAll(findByName(db, step));
      while (st.hasMoreTokens()) {
        step = st.nextToken();
        initSet = findStep(db, initSet, step, (mode & Path.CREATE) == Path.CREATE);
      }
      for(TreeNode nde:initSet)
        if(!denydisposed||!nde.isDisposed())
          ret.add(nde);
    }
//    if (db != this.db) {
//      DeepFirstTreeWalker dftw =
//          new DeepFirstTreeWalker(db.getRootNode(), false, false, true, true, false);
//      Iterator<TreeNode> it = dftw.iterator();
//      while (it.hasNext()) {
//        TreeNode nde = it.next();
//        if (nde != null && nde.getId().equals(qId) && (!denydisposed||!nde.isDisposed())) ret.add(nde);
//      }
//    }
    if ((mode & Path.TAG) == Path.TAG) {
      //TODO look for tag
    }
    return new LinkedList<>(ret);
  }

  private Set<TreeNode> findStep(TreeDB db, Set<TreeNode> initSet, String step, boolean create) {
    Set<TreeNode> next = new HashSet<>();
    for (TreeNode cur : initSet) {
      if (cur != null) {
        List<TreeNode> next2 = cur.findWorkChildPlusName(step);
        if (next2 != null && next2.size() > 0) next.addAll(next2);
        else if (create
            && (initSet.size() == 1
                || initSet.contains(db.getRootNode())
                    && initSet.size() == 2
                    && cur.getQualifiedId().equals(Requirement.getTypeRootQId())))
          next.add(cur.createChild(step));
      }
    }
    return next;
  }

  private TreeNode[] resolveRefPath(TreeNode targ, String path) {
    //		try{
    //			TreeNode sres = targ.getTreeDB().getNode(path);
    //			if(sres!=null)
    //				return new TreeNode[]{sres};
    //		}
    //		catch(Exception e){}
    TreeNode root = targ.getTreeDB().getRootNode();
    RequalityElement rt = RequalityElement.toElement(root.getRoot().getUUId(), root);
    synchronized (proxy) {
      if (!proxy.containsKey(rt)) {
        if (!proxy.containsValue(rt.getDocument()))
          proxy.put(rt, DocumentHelper.createDocument(rt));
        else proxy.put(rt, rt.getDocument());
      }
    }
    RequalityElement.toElement(targ.getRoot().getUUId(), targ.getTreeDB().getRootNode())
        .setDocument(proxy.get(rt));
    RequalityElement rer = RequalityElement.toElement(targ.getRoot().getUUId(), targ);
    org.dom4j.XPath xpath;
    try {
      xpath = proxy.get(rt).createXPath(path);
    } catch (InvalidXPathException icpe) {
      return null;
    }
    Object result = xpath.evaluate(rer);
    if (result instanceof Collection<?>) {
      List<TreeNode> res = new ArrayList<>();
      @SuppressWarnings("unchecked")
      Iterator<Element> it = ((Collection<Element>) result).iterator();
      while (it.hasNext()) {
        Element re = it.next();
        res.add(((RequalityElement) re).getWrappedTreeNode());
      }
      if (res.size() == 0) return null;

      return res.toArray(new TreeNode[0]);
    }
    if (result instanceof RequalityElement) {
      return new TreeNode[] {((RequalityElement) result).getWrappedTreeNode()};
    }
    return null;
  }

  public boolean verify(TreeNode src, TreeNode trg, String path) {
    //TreeNode targ = db.getNode(targuid);
    if (path == null) return false;
    try {
      TreeNode ntarg = db.getNode(path, Path.QID | Path.UV | Path.UUID); //TODO crossproject
      if (ntarg != null) return ntarg.getUUId().equals(trg.getUUId());
    } catch (Exception e) {
    }
    TreeNode root = src.getTreeDB().getRootNode();
    RequalityElement re = RequalityElement.toElement(root.getRoot().getUUId(), root);
    synchronized (proxy) {
      if (!proxy.containsKey(re)) {
        if (re.getDocument() != null) proxy.put(re, re.getDocument());
        else proxy.put(re, DocumentHelper.createDocument(re));
      }
    }
    RequalityElement.toElement(root.getRoot().getUUId(), db.getRootNode())
        .setDocument(proxy.get(re));
    //RequalityElement.toElement(root.getRoot().getUUId(),targ);
    org.dom4j.XPath xpath;
    try {
      xpath = proxy.get(re).createXPath(path);
    } catch (InvalidXPathException icpe) {
      return false;
    }
    //TODO crossproject
    if (trg == null
        || trg.getRoot() == null
        || RequalityElement.toElement(trg.getRoot().getUUId(), trg) == null) return false;
    return xpath.matches(RequalityElement.toElement(trg.getRoot().getUUId(), trg));
  }

  public Set<UUID> getNameCache(String nid) {
    return getStorage().getByName(nid);
  }

  /**
   * Get list of nodes linked to current by linkto: relation
   *
   * @return list of linked nodes
   */
  //	public List<TreeNode> getLinkedNodes(TreeNode rnode){
  //		List<TreeNode> nodes = new ArrayList<TreeNode>();
  //		if(rnode == null)
  //			 return nodes;
  //		Attribute links = rnode.attrs().getAttribute(TreeNode.ATTR_LINKS);
  //        if(links != null){
  //            @SuppressWarnings("unchecked")
  //			ArrayList<String> linksarray = (ArrayList<String>)links.getValue(rnode.attrs());
  //            for(String link : linksarray){
  //                if(link.startsWith("*"))
  //                    link = link.substring(1);
  //                String uid = link.substring(0,link.indexOf('_'));
  //                TreeNode node = db.getNode(link.substring(uid.length()+1), Path.QID|Path.UUID|Path.UV|Path.XPATH);
  //                if(node!=null)
  //                	nodes.add(node);
  //            }
  //        }
  //		return nodes;
  //	}
  public ReferenceValue[] getRefValues(TreeNode node, String key) {
    if (node == null) return new ReferenceValue[]{new ReferenceValue(key, "Source not found!")};
    return getRefValues(node, key, null);
    //		TreeNode node = attrs().getLinkAttributeValue(key);
    //		return node==null?"Missing reference: '"+attrs().getStringAttributeValue(key)+"'":node.getUserVisibleName(false);
  }

  public static ReferenceValue[] getRefValues(TreeNode targ, String key, String label) {
//    List<TreeNode> nde =
//        targ.getTreeDB()
//            .getLinkResolver()
//            .getStorage()
//            .resolveLink(targ.getTreeDB().getLinkResolver(), targ, key, null, label, true, true);
    Attribute a = targ.attrs().getAttribute(key);
    if (a!=null){
      Object val = a.getValue();
      if(val!=null)
      {
        if(a.getType()==AttributeType.REFERENCE)
        {
          ReferenceValue v = (ReferenceValue) val;
//          if(v.startsWith(ReferenceValueProvider.ERROR_CHAR))
//            return new Object[]{v.substring(1)};
//          TreeNode node = targ.getTreeDB().getNode(v);
//          if(node == null)
//            return new Object[]{ReferenceValueProvider.MISSING_MESSAGE.replace("{val}",v).substring(1)};
          if(v.hasErrorsOnResolve(targ.attrs()))
            return new ReferenceValue[]{v};
          v.getNodes(targ.attrs());
          return new ReferenceValue[]{v};
//            ret.add(node);
//          return ret.toArray(new Object[0]);
        }else
        {
          List<ReferenceValue> ret = new LinkedList<>();
          ListValue lv = (ListValue) a.getValue();
          for(Object o : lv.getValues()){
            if(o == null)
              ret.add(new ReferenceValue(key, ReferenceValueProvider.NULL_MESSAGE.substring(1)));
            else{
              ReferenceValue v = (ReferenceValue) o;
              if(v.hasErrorsOnResolve(targ.attrs()))
                ret.add(v);
              else
                ret.add(v);
            }
          }
          return ret.toArray(new ReferenceValue[0]);
        }
      }
      else
        return new ReferenceValue[]{new ReferenceValue(key, ReferenceValueProvider.NULL_MESSAGE)};
    }
//    if (label == null) label = targ.attrs().getStringAttributeValue(key);
//    return (nde == null || nde.size() == 0)
//        ? ("Missing reference: '" + label + "'")
//        : nde.iterator().next().getUserVisibleName(false);
    return new ReferenceValue[]{new ReferenceValue(key, ReferenceValueProvider.NULL_MESSAGE)};
  }

  public Map<String, Set<TreeNode>> getRevertLinks(TreeNode rnode) {
    return getRevertLinks(rnode, false);
  }

  private Map<String, Set<TreeNode>> getRevertLinks(TreeNode rnode, final boolean validate) {
    if (rnode == null) return null;
    return getStorage().getRevLinks(rnode,validate);
  }

  private void removeIncorrectLinks(TreeNode node) {
//    ReadLock rl = getStorage().rrwl.readLock();
//    rl.lock();
////    if (!getStorage().isInited(node)) {
//      rl.unlock();
//      initLinksCache(node);
//      rl.lock();
//    }
    Map<String, Set<TreeNode>> rev = getStorage().getLinks(node);
    //rl.unlock();
    if (rev != null && rev.size() > 0) {
      for (String key : new HashSet<>(rev.keySet()))
        for (TreeNode to : new HashSet<>(rev.get(key)))
          if (!checkIsCorrect(node, to, key)) {
            removeLink(node, to, key);
            to.revCacheChanged(node, key);
          }
    }
  }
  private void removeLink(TreeNode from, TreeNode to, String key) {
    ReadLock rl = getStorage().rrwl.readLock();
    boolean ru = true;
    rl.lock();
    try {
      if (getStorage().links.containsKey(from)) { //tends to be loaded if rev is loaded
//        Set<TreeNode> vals = getStorage().linkCache.get(from).get(key);
//        if (vals != null) {
//          vals.remove(to);
//        }
//        if (vals == null || vals.isEmpty()) {
//          getStorage().linkCache.get(from).remove(key);
//        } else {
//          getStorage().linkCache.get(from).put(key, vals);
//        }
        getStorage().links.get(from).remove(key);
      }
    } finally {
      if (ru) rl.unlock();
    }
    getStorage().removeRevLink(from, to, key, true);
    to.revCacheChanged(to, key);
  }

  private boolean checkIsCorrect(TreeNode from, TreeNode to, String savedName) {
    if (from == null || from.isDisposed()) return false;
    if (to == null || to.isDisposed()) return false;
    int ind = savedName.indexOf("["), pos = -1;
    if (ind > -1) {
      int end = savedName.indexOf("]", ind + 1);
      if (end > -1) {
        pos = Integer.parseInt(savedName.substring(ind + 1, end));
        savedName = savedName.substring(0, ind) + savedName.substring(end + 1);
      }
    }
    //if(from!=null){
      Attribute a = from.attrs().getAttribute(savedName);
      if (a == null || !a.isReference()) return false;
      int i = 0;
      TypedValue tv = a.getTValue();
      if(tv.getType() == AttributeType.REFERENCE){
        ReferenceValue rv = tv.getGValue();
        UUID[] uids = rv.getNodeUUIDs(from.attrs());
        if(uids != null)
          for(UUID node:uids)
            if(node.equals(to.getUUId())||to.getCloneParentUUId()!=null && to.getCloneParentUUId().equals(node))
          return true;
      }else{
        ListValue lv = a.getGValue();
        for (Object s : lv.getValues()) {
          ReferenceValue rv = (ReferenceValue)s;
          UUID[] uids = rv.getNodeUUIDs(from.attrs());
          if(uids!=null)
            for(UUID node:uids)
              if(node.equals(to.getUUId())||to.getCloneParentUUId()!=null && to.getCloneParentUUId().equals(node))
            return true;
          i++;
        }
      }
   // }
    return false;
  }
  public static void registerLates(AttributeStorage where, String key, AttributeType type, Object value) {
    if(value == null)
      return;
    if(type == AttributeType.REFERENCE){
      ReferenceValue rv = (ReferenceValue)value;
      Object raw = rv.getRaw();
      if(raw == null)
        return;
      String sref = raw.toString();
      int ind = sref.indexOf("::");
      if(ind>-1)
      {
        String dbname = sref.substring(0,ind);
        TreeDB db = ResManager.getManager().getTreeDB(dbname);
        if(db == null)
          TreeDB.getCPLinkResolver().registerLateResolve(dbname, where.getNode(), key);
      }
      return;
    }
    ListValue lv = (ListValue)value;
    for(Object o:lv)
      registerLates(where, key, lv.getType(), o);
  }
  public void handleRefsChanges(
      TreeNode node, Set<String> ofint,Map<String,Attribute>newa, boolean internal, boolean instance){
    if (getStorage() == null) return;
    
    Map<String,Map<String,List<String>>> toa = null,tod = null,cptoa = null,cptod = null;
    String nid = node.getUUId().toString(),
      cpnid = LinksStorage.getCPId(node);
    Map<String, Map<String, List<Object>>> vals = 
        getStorage().getDirectLinksCacheItems(node, ofint);
    for(String s:ofint){
      Set<Object> oldS = new HashSet<>();
      Map<String,List<Object>>old = null;
      if(vals.get(s)!=null)
        old=vals.get(s);
      if(old!=null)
        for(String k : old.keySet())
          oldS.addAll(old.get(k));
      Attribute a = newa.get(s);
      if (a != null && a.isReference()) {
        if(!instance)
          registerLates(node.attrs(), s, a.getType(),a.getValue());
        List<TreeNode>list = ReferenceValue.getValues(node.attrs(), a.getType(),a.getValue(),false);
        if(list!=null)
          for(TreeNode nde:list){
            String tos = null;
            boolean local = !LinksStorage.isCPRelation(node, nde);
            if(!local)
              tos = LinksStorage.getCPId(nde);
            else
              tos = nde.getUUId().toString();
            if(!oldS.remove(tos)){
              if(local){
                if(toa == null)
                  toa = new HashMap<>();
                addValue(toa, tos, s, nid);
              }else{
                if(cptoa == null)
                  cptoa = new HashMap<>();
                addValue(cptoa, tos, s, cpnid);
              }
            }
          }
      }

      if(!oldS.isEmpty()){
        for(Object s2:oldS){
          String s2s = s2.toString();
          if(s2s.contains("::"))
          {
            if(cptod == null)
              cptod = new HashMap<>();
            addValue(cptod, s2s, s, cpnid);
          }else{
            if(tod == null)
              tod = new HashMap<>();
            addValue(tod, s2s, s, nid);
          }
        }
      }
    }
    
    if(toa!=null || tod!=null){
      String t = getCacheStorage().startCacheTransaction("Resolve links "+node.toString(),CacheStorage.COMMIT_SINGLE);
      if(tod!=null)
        getStorage().removeRevLinks(tod);
      if(toa!=null)
        getStorage().addRevLinks(toa);
      if (t==null)
        getCacheStorage().commitCache();
    }
    CacheStorage cps = TreeDB.getCPCacheStorage();
    synchronized (cps) {
      if(cptoa!=null || cptod!=null){
        LinkResolver clr = TreeDB.getCPLinkResolver();
        String t = cps.startCacheTransaction("Resolve links "+node.toString(),CacheStorage.COMMIT_SINGLE);
        if(cptod!=null)
          clr.getStorage().removeRevLinks(cptod);
        if(cptoa!=null)
          clr.getStorage().addRevLinks(cptoa);
        if (t==null)
          cps.commitCache();
      }
    }
  }
  
  //TODO move to propogator
  public void handleAttrChanges(
      TreeNode node, ChangesSet<Attribute> params, boolean internal) {
    if (getStorage() == null) return;
   
    Set<String> ofint = null;
    //Map<String, Attribute> oldA = params.getOldMap();
    for (String s:params.getAll()) {
      Attribute olda = params.getOld(s);
      Attribute a = params.getNew(s);
      if (a == null
          || a.isReference()
          || olda != null && olda.isReference()){
        if(ofint == null)
          ofint = new HashSet<>();
        ofint.add(s); 
      }
    }
    if(ofint!=null)
      handleRefsChanges(node,ofint,params.getNewMap(),internal, false);
  }

  private static void addValue(
      Map<String, Map<String, List<String>>> toa, String tos, String s, String nid) {
    if(!toa.containsKey(tos))
      toa.put(tos, new HashMap<>());
    if(!toa.get(tos).containsKey(s))
      toa.get(tos).put(s, new LinkedList<>());
    toa.get(tos).get(s).add(nid);
  }


  private ITreeChangeListener changeListener =
      new DefaultTreeChangeListener() {

        @Override
        public void moved(TreeNode node, TreeNode oldParent, TreeNode newParent, boolean internal) {
          if (getStorage() == null || node.getTreeDB()!=db) return;
          recCheck(node);
        }

        @Override
        public void deleted(TreeNode nodeDesc, boolean internal, Moment m) {
          if (getStorage() == null || m == Moment.AFTER || nodeDesc.getTreeDB()!=db) return;
          Set<TreeNode> nds = new HashSet<>();
          recAddTreeNodes(nodeDesc, nds);
          String t = getCacheStorage().startCacheTransaction("Delete node "+nodeDesc.getUUId(), CacheStorage.COMMIT_ALL);
          for (TreeNode node : nds) {
            removeIncorrectRevLinks(node);
            removeIncorrectLinks(node);
            //getStorage().removeCache(node.getUUId().toString());
            //TODO remove?
//            getStorage().linkCache.remove(node);
//            getStorage().revLinksCahce.remove(node);
          }
          if(t == null)
            getCacheStorage().commitCache();


          //LinksStorage.removeRevKey(getStorage(),nodeDesc.getUUId().toString());
        }

        @Override
        public void created(TreeNode node, boolean internal) {
          if (getStorage() == null || node.getTreeDB()!=db) return;
          String t = getCacheStorage().startCacheTransaction("resolve created links",CacheStorage.COMMIT_SINGLE);
          Set<String> ofint = null;
          for (String s : node.attrs().getLocalAttributeKeys()) {
            Attribute a = node.attrs().getAttribute(s);
            if (a == null || a.isReference())
            {
              if(ofint == null)
                ofint = new HashSet<>();
              ofint.add(s);
            }
          }
          if(ofint!=null && !ofint.isEmpty())
            handleRefsChanges(node, ofint, node.attrs().getLocalAttributes(), internal, false);
          //if(getStorage().getStoredRevKeys().contains(node.getUUId().toString())){
          removeIncorrectLinks(node);
          removeIncorrectRevLinks(node);
          if(t == null)
            getCacheStorage().commitCache();
          //}else
            //correctRevLinks(node);
        };

        

        private void removeIncorrectRevLinks(TreeNode node) {
          Map<String, Set<TreeNode>> rev = getRevertLinks(node, true);
          if (rev != null && rev.size() > 0) {
            for (String key : new HashSet<>(rev.keySet()))
              for (TreeNode from : new HashSet<>(rev.get(key)))
                if (!checkIsCorrect(from, node, key)) {
                  removeLink(from, node, key);
                }
          }
        }

        private void recAddTreeNodes(TreeNode nd, Set<TreeNode> uids) {
          if (nd == null) return;
          TreeNode node = db.getNode(nd.getUUId());
          if (node != null) uids.add(node);
          for (TreeNode ndc : nd.getWorkChildren(null,true,false,true,false)) recAddTreeNodes(ndc, uids);
        }

        
        @Override
        public void changed(
            TreeNode node, ChangesSet<Attribute> params, boolean internal) {
          if(node.getTreeDB()!=db)
            return;
          LinkResolver.this.handleAttrChanges(node,params,internal);
//          if (getStorage() == null) return;
//          Map<String, Attribute> newA = params.getNewMap();
//          for (Entry<String, Attribute> en : newA.entrySet()) {
//            Attribute olda = params.getOld(en.getKey());
//            Attribute a = en.getValue();
//            if (a == null
//                || a.isReference()
//                || olda != null && olda.isReference()) {
////              Set<TreeNode> oldL = new HashSet<TreeNode>();
////              if (getStorage().linkCache.containsKey(node)
////                  && getStorage().linkCache.get(node).get(s) != null)
////                oldL.addAll(getStorage().linkCache.get(node).get(s))              
//              if (a != null && a.isReference()) {
//                getStorage()
//                    .resolveLink(
//                        node.getTreeDB().getLinkResolver(),
//                        node,
//                        en.getKey(),
//                        null,
//                        en.getKey(),
//                        true); //TODO look
//              }
////              Set<TreeNode> newL = new HashSet<TreeNode>();
////              if (getStorage().linkCache.containsKey(node)
////                  && getStorage().linkCache.get(node).get(s) != null)
////                newL.addAll(getStorage().linkCache.get(node).get(s));
////              for (TreeNode tn : new HashSet<>(newL))
////                if (oldL.contains(tn)) {
////                  newL.remove(tn);
////                  oldL.remove(tn);
////                }
////              oldL.addAll(newL);
//            }
//          }
//          removeIncorrectLinks(node);
        }

        private void recCheck(TreeNode source) {
          removeIncorrectRevLinks(source);
          for (TreeNode node : source.getWorkChildren()) recCheck(node);
        }

        @Override
        public void changed(List<TreeNodeChangeEvent> changes) {
          if (getStorage() == null) return;
          for (TreeNodeChangeEvent ce : changes)
            switch (ce.getType()) {
              case CHANGE:
                {
                  TreeNode node = ce.getTreeNodeTarget(); 
                  if(node.getTreeDB()==db)
                    changed(
                        node,
                        (ChangesSet<Attribute>) ce.getArguments()[1],
                        (Boolean) ce.getArguments()[0]);
                }
                break;
              case MOVE:{
                TreeNode node = ce.getTreeNodeTarget(); 
                if(node.getTreeDB()==db)
                  moved(
                      node,
                    (TreeNode) ce.getArguments()[1],
                    (TreeNode) ce.getArguments()[2],
                    (Boolean) ce.getArguments()[0]);
              }break;
              case CREATE:{
                TreeNode node = ce.getTreeNodeTarget(); 
                if(node.getTreeDB()==db)
                  created(node, (Boolean) ce.getArguments()[0]);
              }break;
              case DELETE:{
                TreeNode node = ce.getTreeNodeTarget(); 
                if(node.getTreeDB()==db)
                  deleted(node, (Boolean) ce.getArguments()[0], (Moment)ce.getArguments()[2]);
              }
              default:
                break;
            }
        }

//        @Override
//        public void cacheChanged(
//            TreeNode node,
//            String section,
//            java.util.Set<String> keys,
//            Map<String, Map<String,Object>> oldvs,
//            Map<String, Map<String,Object>> newvs,
//            boolean forceInternal) {
//          if (getStorage() == null) return;
//          TreeDB db = Instantiator.getRootTree(node.getTreeDB());
//          if (section.startsWith("revert")) {
//            //We need to maintain cache changes due to problem with reverse of node deletion
//            //Use case: we have node to be delete with has some links pointing to it
//            //If we delete it then all rev links that points to it will be deleted too.
//            //Reversing of this delete will force us to look for all places there this node
//            //can be referenced. This situation is similar to create of node that meets
//            //conditions of some reference that may point to it.
//            //
//            if (node.isRoot()) {
//              int cnt = 0;
//              Set<String> allk = new HashSet<>();
//              for (String s : keys) {
//                if (s.contains("_revlinks")) {
//                  if (newvs != null) {
//                    Map<String,Object> sv = newvs.get(s);
//                    if (sv != null) {
//                      String sva = sv.get("").toString();
//                      if (!s.endsWith("size")) allk.add(sva);
//                      cnt++;
//                    }
//                  }
//                }
//              }
//              ReadLock rl = getStorage().rrwl.readLock();
//              WriteLock wl = getStorage().rrwl.writeLock();
//              boolean ru = false;
//              rl.lock();
//              ru = true;
//              try {
//                //allk.removeAll(getStorage().revKeys);
//                rl.unlock();
//                ru = false;
//                boolean tos = false;
//                for (String s : allk) {
//                  try{
//                    TreeNode ode = db.getNode(UUID.fromString(s));
//                    if (ode != null) {
//                      //correctLinks(ode);
//                      correctRevLinks(ode);
//                      ode.revCacheChanged(ode, s);
//                    }
//                  }catch(Exception e){
//                    
//                  }
//                }
//              } finally {
//                if (ru) rl.unlock();
//              }
//            } else {
//              
//              Map<String,Set<TreeNode>> old = getRevertLinks(node);
//              Collection<String> vals = Collections.emptySet();
//              if(section.endsWith("_revItem"))
//                vals = newvs.get(section).keySet();
//              for(String k:vals){
//                String val = newvs.get("_revItem").get(k).toString();
//                if(val!=null){
//                  val = val.substring(val.indexOf("|")+1);
//                  if(val.indexOf("|")<(val.length()-1)&&val.indexOf("|")>-1){
//                    val = val.substring(val.indexOf("|")+1);
//                    String[] avals = val.split("\\|");
//                    for(String d:avals){
//                      if(!d.isEmpty()){
//                        TreeNode nde = db.getNode(d);
//                        if(nde!=null){
//                          getStorage().addLink(nde, node, k, false);
//                          node.revCacheChanged(nde, k);
//                        }
//                      }
//                    }
//                  }
//                }
//              }
//            }
//          }
//        };
      };
  /***
   * Initiates revert links by direct links if this operation was not performed before.
   * The state is determined by flag in cache storage. If flag is not set then the whole
   * tree wil loaded and all reference
   * Also will cause name cache calculation
   * Also handles late reference calls for multi-project refs
   */
  public void initCaches() {
    initCaches(false);
  }
  public void initCaches(boolean instance) {
    String name = ResManager.getManager().getName(db);
    db.removeListener(changeListener);
    getStorage().initNamesCache();
    getStorage().initRevLinks(instance);
    db.addListener(changeListener);
    if(!instance){
      loadLates();
    //for (TreeDB db : ResManager.getManager().getTreeDBs()) {
      //String fname = ResManager.getManager().getName(db);
      synchronized (TreeDB.getCPCacheStorage()) {
        LinkResolver lr = TreeDB.getCPLinkResolver();
        if (!lr.equals(this) && lr.lates.containsKey(name)) {
          getStorage().rrwl.writeLock().lock();
          try{
            Set<LateResolveRequest> lrr = lr.lates.remove(name);
            lr.saveLates();
            Map<TreeNode,Set<String>> toch = new HashMap<>();
            for (LateResolveRequest l : lrr) {
              if (!l.node.isDisposed()){
                if(!toch.containsKey(l.node))
                  toch.put(l.node, new HashSet<>());
                toch.get(l.node).add(l.key);
              }
            }
            for(TreeNode node:toch.keySet()){
              Set<String>links = toch.get(node);
              for(String s:links){
                Attribute a = node.attrs().getAttribute(s);
                if(a!=null && a.isReference())
                  ReferenceValue.resetAttribute(node.attrs(),a.getType(),a.getValue());
              }
              if(!toch.get(node).isEmpty())
                node.getTreeDB().getLinkResolver().handleRefsChanges(
                    node, toch.get(node), node.attrs().getLocalAttributes(), false, instance);
            }
          }finally{
            getStorage().rrwl.writeLock().unlock();
          }
        }
      }
    }
  }

  public void removeCache(String string) {
    getStorage().removeCache(string);
  }
//
//  public List<TreeNode> resolveLink(
//      LinkResolver lr,
//      TreeNode where,
//      String key,
//      TreeNode def,
//      final String savedName,
//      boolean lock,
//      boolean save) {
//    return getStorage().resolveLink(lr, where, key, def, savedName, lock, save);
//  }

  public Map<String, Set<TreeNode>> getLinkedNodes(TreeNode node) {
    return getStorage().getLinks(node);
  }


  public void initNames() {
    getStorage().initNamesCache();
  }

  //	private Map<TreeNode,Map<String,Set<TreeNode>>> getAllRevLinks(){
  //		if(revLinkCahce == null)
  //			loadRevertLinks();
  //		return revLinkCahce;
  //	}
  //public Map<String, Set<TreeNode>> getRevertLinks(TreeNode node1){
  //    	Map<String, Set<TreeNode>> map = getAllRevLinks().get(node1);
  //    	if(map!=null)
  //    		return map;
  //    	return new HashMap<String,Set<TreeNode>>();
  //    }
  //
  //	private void registerRevertLink(TreeNode node1, TreeNode node, String key) {
  //		//TODO to root
  //		if(!revLinkCahce.keySet().contains(node1))
  //			revLinkCahce.put(node1, new HashMap<String,Set<TreeNode>>());
  //		if(!revLinkCahce.get(node1).containsKey(key))
  //			revLinkCahce.get(node1).put(key,new HashSet<TreeNode>());
  //		revLinkCahce.get(node1).get(key).add(node);
  //		if(!linkCahce.containsKey(node))
  //			linkCahce.put(node, new HashMap<String,Set<TreeNode>>());
  //		if(!linkCahce.get(node).containsKey(key))
  //			linkCahce.get(node).put(key, new HashSet<TreeNode>());
  //		linkCahce.get(node).get(key).add(node1);//adds registered direct link
  //	}
  //
  //	public void getAllRevertLinks() {
  //		getAllRevLinks();
  //	}

}
