package com.unitesk.requality.core;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class LinksStorage {
  public static final int RESOLVE_MODE = Path.QID | Path.UV | Path.UUID | Path.XPATH;
  ChangesSet<String> revKeysChanges = new ChangesSet<>();
  ReentrantReadWriteLock rrwl = new ReentrantReadWriteLock();
  Map<TreeNode, Map<String, Attribute>> links = new HashMap<>();

  private HashMap<String, Set<UUID>> namecache = new HashMap<>();
  boolean saveAllowed = true;
  private LinkResolver lr = null;

  public LinksStorage(LinkResolver linkResolver) {
    lr = linkResolver;
  }

  Map<String, Map<String, Object>> getRevertLinksCacheItems(
      TreeNode node, Set<String> ofint) {
      
    Map<String, Map<String, Object>> local = 
        lr.getCacheStorage().getCacheMaps(node.getUUId().toString(), "revert_revItem",null);
        
    synchronized (TreeDB.getCPCacheStorage()) {
        Map<String, Map<String, Object>> cp = TreeDB.getCPCacheStorage().getCacheMaps(
            LinksStorage.getCPId(node), "revert_revItem", ofint);
      for(String s:cp.keySet())
        if(!local.containsKey(s)){
          local.put(s, cp.get(s));
        }else
          for(String s2:cp.get(s).keySet())
            if(!local.get(s).containsKey(s2))
              local.get(s).put(s2, cp.get(s).get(s2));
    }
    return local;
  }
  Map<String, Map<String, List<Object>>> getDirectLinksCacheItems(
      TreeNode node, Set<String> ofint) {
      
    Map<String,Map<String,List<Object>>> local = 
        lr.getCacheStorage().getCacheMapsByValue(node.getUUId().toString(), "revert_revItem", ofint);
        
    synchronized (TreeDB.getCPCacheStorage()) {
        Map<String,Map<String,List<Object>>> cp = TreeDB.getCPCacheStorage().getCacheMapsByValue(
            LinksStorage.getCPId(node), "revert_revItem", ofint);
      for(String s:cp.keySet())
      if(!local.containsKey(s)){
        local.put(s, cp.get(s));
      }else
        for(String s2:cp.get(s).keySet())
          if(!local.get(s).containsKey(s2))
            local.get(s).put(s2, cp.get(s).get(s2));
          else
            local.get(s).get(s2).addAll(cp.get(s).get(s2));
    }
    return local;
  }
//  public List<TreeNode> resolveLinks(
//      TreeNode where,
//      Set<String> keys,
//      boolean lock) {
//    String[] vals = null;
//    
//   // boolean process = true;
//    List<TreeNode> ret = null;
//    Map<String, Map<String, List<Object>>> lvs = getCacheItems(where,keys);
//    for(Entry<String, Map<String, List<Object>>> en:lvs.entrySet()){
//    //if (process)
//      if ((!links.containsKey(where)
//          || !links.get(where).containsKey(key)
//          || (links.get(where).get(key)!=(a)))) {
//        String t = lr.getCacheStorage().startCacheTransaction("Resolve links",CacheStorage.COMMIT_SINGLE);
//        if (lv!=null && lv.containsKey(key))//called only on start or create so it can be normal
//          for (Object node : lv.get(key)) removeRevLink(where, node!=null?node.toString():null, key, lock);
//        if (vals != null) { 
//          for (String val : vals) {
//            List<TreeNode> nds = Collections.emptyList();
//            try {
//              nds = lr.getNotDisposedResolvePath(where, key, lr.db, val, RESOLVE_MODE);
//            } catch (NodeNotFoundException e) {
//              
//            }
//            if(nds != null){
//              for (TreeNode nd : nds) {
//                addLink(where, nd, key, lock);
//                nd.revCacheChanged(nd, key);
//                //if (save) nd.getTreeDB().getLinkResolver().getStorage().saveRevLinks(nd);
//              }
//
//              if(ret == null)
//                ret = new LinkedList<>();
//              ret.addAll(nds);
//            }
//          }
//        }
//        if(t==null)
//          lr.getCacheStorage().commitCache();
//        where.attrs().updateAttribute(key);
//      }
//      else if (lv!=null){
//        List<Object> uo = lv.get(key);
//        if(uo!=null){
//          ret = new LinkedList<>();
//          for(Object o:uo)
//            ret.add(where.getTreeDB().getNode(o.toString()));
//        }
//      }
//    }
//   if(ret!=null)
//     return ret;
//   return Collections.emptyList();
//  }
  /*
   * if(save){
              String t = lr.getCacheStorage().startCacheTransaction("Resolve links",CacheStorage.COMMIT_SINGLE);
              for (TreeNode nd : nds) {
                addLink(where, nd, key, lock);
                nd.revCacheChanged(nd, key);
                //if (save) nd.getTreeDB().getLinkResolver().getStorage().saveRevLinks(nd);
              }
              if(t==null)
                lr.getCacheStorage().commitCache();
              where.attrs().updateAttribute(key);
            }
   */
  static String getCPId(TreeNode where){
    return where.getTreeDB().getResName()+"::"+where.getUUId().toString();
  }
  static boolean isCPRelation(TreeNode where, TreeNode nodeToSave){
    return !Instantiator.getRootTree(where.getTreeDB()).equals(Instantiator.getRootTree(nodeToSave.getTreeDB()));
  }
//  protected void addLink(final TreeNode from, TreeNode to, final String key, boolean lock) {
//    if(from == null || to == null)
//      return;
//    WriteLock wl = null;
//    boolean wu = false;
//    try {
//      if (lock) {
//        wl = rrwl.writeLock();
//        wl.lock();
//        wu = true;
//      }
//      //TODO rewrite here
//      to.getTreeDB().getCacheStorage().addCacheListValue(to.getUUId().toString(), "revert_revItem", key,
//          nodeToSave(to,from));
//      if (!links.containsKey(from)) links.put(from, new HashMap<String, Attribute>());
//
//      if (wu && wl!=null) {
//        wl.unlock();
//        wu = false;
//      }
//      Attribute a = from.attrs().getAttribute(key);
//      if (!wu && wl!=null) {
//        wl.lock();
//        wu = true;
//      }
//      links.get(from).put(key, a);
//    } finally {
//      if (wl!=null && wu) wl.unlock();
//    }
//  }

//  protected void removeRevLink(TreeNode from, String to, String key, boolean lock) {
//    removeRevLink(from, lr.db.getNode(UUID.fromString(to)), key, lock);
//  }
//  protected void removeRevLink(String from, String to, String key, boolean lock) {
//    removeRevLink(lr.db.getNode(UUID.fromString(from)), lr.db.getNode(UUID.fromString(to)), key, lock);
//  }

  public void removeRevLinks(Map<String, Map<String, List<String>>> tod) {
    for(String key:tod.keySet())
      for(String s:tod.get(key).keySet()){
        List<String> toda = tod.get(key).get(s);
        if(toda!=null)
          lr.getCacheStorage().removeCacheListValues(key, "revert_revItem", s, toda);
        else
          lr.getCacheStorage().removeByKey(key, "revert_revItem", s);

      if(links.containsKey(s) && links.get(s).isEmpty())
        links.remove(s);
  //      if(links.containsKey(node))
  //        links.get(node).remove(s);
      }
//    if(links.containsKey(node) && links.get(node).isEmpty())
//      links.remove(node);
  }
  public void addRevLinks(Map<String, Map<String, List<String>>> tod) {
    for(String key:tod.keySet())
      for(String s:tod.get(key).keySet()){
        lr.getCacheStorage().addCacheListValues(key, "revert_revItem", s, tod.get(key).get(s));
      }
  }
  protected void removeRevLink(TreeNode from, TreeNode to, String key, boolean lock) {
    if(from==null || to == null)
      return;
    String k = to.getUUId().toString();
    lr.getCacheStorage().removeCacheListValue(k, "revert_revItem", key,from.getUUId().toString());
    from.attrs().updateAttribute(key);
  }

  public void removeCache(String s) {
    lr.getCacheStorage().clearCachedCache(s);
    //linkCache.remove(uuId);
  }

  public void initNamesCache() {
    UUID ruid = lr.db.getRootNodeUUID();
    if(ruid == null)
        return;
    //UUID ruid = root.getUUId();
    TreeNode root = lr.db.getRootNode();
    if(root == null)
      return;
    String val =
        (String) lr.getCacheStorage().getCacheValue(ruid.toString(), "names", "_namesInited");
    if (val == null || !val.equals("true")) {
      WriteLock wl = rrwl.writeLock();
      wl.lock();
      try {
        namecache = new HashMap<>();
      } finally {
        wl.unlock();
      }
      DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(lr.db.getRootNode(), false, false, true, false, false);
      Iterator<TreeNode> it = dftw.iterator();
//      saveRevLinks(null, true);
      long l = System.currentTimeMillis();
      String vala = lr.getCacheStorage().startCacheTransaction("init names",CacheStorage.COMMIT_ALL);
      while (it.hasNext()) {
        TreeNode node = it.next();
        String name = node.getName();
        if(name!=null){
          if(!namecache.containsKey(name))
            namecache.put(name, new HashSet<>());
          namecache.get(name).add(node.getUUId());
          //init is called during loading
  //        root.getTreeDB().getCacheStorage().setCacheValue(lr.db.getRootNode().getUUId(), "names","name", node.getUUId().toString(), name);
        }
      }
      lr.getCacheStorage().setCacheValue(ruid.toString(), "names", "_namesInited","true");
      if(vala == null)
        lr.getCacheStorage().commitCache();
      if(CoreUtils.DEBUG_LEVEL>0)
        RLogger.logInfo("rslv names " + (System.currentTimeMillis()-l) +" "+lr.db.getName()); 
//      saveRevLinks(null, true);
    }else
    {
      namecache = new HashMap<>();
    }
  }
  public void setRevInited(String root, String val){
    String s = (String) lr.getCacheStorage().getCacheValue(root, "revert", "_revInited");
    if(!val.equals(s))
      lr.getCacheStorage().setCacheValue(root, "revert", "_revInited", val);
  }
  public void initRevLinks() {
    initRevLinks(false);
  }
  public void initRevLinks(boolean instance) {
    TreeNode root = null;
    if(lr.db!=null)
      root = lr.db.getRootNode();
    if (root == null) return;
    UUID rootuuid = root.getUUId();
    if(instance){
      return;
    }
    String val =
        (String) root.getTreeDB().getCacheStorage().getCacheValue(rootuuid.toString(), "revert", "_revInited");
    if (val == null || !val.equals("true")) {
      WriteLock wl = rrwl.writeLock();
      wl.lock();
      try {
//        revKeys = new HashSet<>();
//        revLinksCahce = new HashMap<>();
//        linkCache = new HashMap<>();
        links = new HashMap<>();
      } finally {
        wl.unlock();
      }
      DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(root, false, false, true, false, false);
      Iterator<TreeNode> it = dftw.iterator();
//      saveRevLinks(null, true);
      long l = System.currentTimeMillis();
      String vala = root.getTreeDB().getCacheStorage().startCacheTransaction("init links",CacheStorage.COMMIT_ALL);
      Set<String> ofint = new HashSet<>();
      while (it.hasNext()) {
        TreeNode node = it.next();
        for (String key : new HashSet<>(node.attrs().getLocalAttributeKeys())) {
          Attribute a = node.attrs().getAttribute(key);
          if (a != null)
            if (a.isReference())
              ofint.add(key);
        }
        if(!ofint.isEmpty())
          node.getTreeDB().getLinkResolver().handleRefsChanges(
              node, ofint, node.attrs().getLocalAttributes(), false, instance);
        ofint.clear();
      }
      setRevInited(rootuuid.toString(), "true");
      if(vala == null)
        root.getTreeDB().getCacheStorage().commitCache();
      if(CoreUtils.DEBUG_LEVEL>0)
        RLogger.logInfo("rslv " + (System.currentTimeMillis()-l) +" "+root.getTreeDB().getName()); 
//      saveRevLinks(null, true);
    }else
    {
      links = new HashMap<>();
    }
  }

  public Map<String, Set<TreeNode>> getRevLinks(TreeNode node, boolean validate) {
    Map<String, Map<String, Object>> st = 
        getRevertLinksCacheItems(node, null);
    Map<String,Set<TreeNode>> revs = null;
    if(st!=null)
    for (String key : st.keySet()) {
      //String val = stored.get(key);
      //if (val == null) continue;
      //String[] paths = val.split("\\|");
      TreeDB db = node.getTreeDB();
      for (Object patho : st.get(key).values()) {
        if(patho == null)
          continue;
        String path = patho.toString();
        TreeNode node2 = null;
        TreeDB adb = db;
        int pos = path.indexOf("::");
        if (pos > -1) {
          String name = path.substring(0, pos);
          ResManager tm = ResManager.getManager(); //TODO load?
          adb = tm.getTreeDB(name);
          if (adb != null) {
            path = path.substring(pos + 2);
          }
        }
        if(adb == null)
          adb = db;
        try {
          node2 = adb.getNodeExc(UUID.fromString(path));
        } catch (NodeNotFoundException e) {
        }
        if (node2 != null) {
          Attribute a = node2.attrs().getAttribute(key);
          if (a != null && node2.attrs().isReference(key)) {
            boolean isvalid = !validate;
            String akey = key;
            boolean mult = false;
            int al = 0;
            ;
            if(validate){
              if (a.getType() == AttributeType.REFERENCE) {
                isvalid =
                    node2
                        .getTreeDB()
                        .getLinkResolver()
                        .verify(node2, node, ((ReferenceValue)a.getValue()).getRaw().toString());
              } else {
                List<Object> list = ((ListValue) a.getValue()).getValues();
                int l = 0;
                mult = list.size() > 1;
                for (Object o2 : list) {
                  if (node2
                      .getTreeDB()
                      .getLinkResolver()
                      .verify(node2, node, ((ReferenceValue)o2).getRaw().toString())) {
                    isvalid = true;
                    al = l;
                  }
                  l++;
                }
              }
            }
            if (isvalid) {
              if(revs == null)
                revs = new HashMap<>();
              if(!revs.containsKey(key))
                revs.put(key, new HashSet<>());
              revs.get(key).add(node2);
//              getStorage().addLink(node2, node, key, true);
//              getStorage().saveRevLinks(node);
            }
          }
        }
      }
    }
//    Map<String,List<TreeNode>> ret = Collections.emptyMap(); 
//    if(st!=null&&!st.isEmpty()){
//      ret = new HashMap<>();
//      for(String s:st.keySet()){
//        ret.put(s, new LinkedList<>());
//        for(String s2:st.get(s))
//        {//TODO autocorrection
//          TreeNode nde = lr.db.getNode(UUID.fromString(s2));
//          if(nde!=null)
//            ret.get(s).add(nde);
//        }
//      }
//    }
    if(revs == null)
      return Collections.emptyMap();
    return revs;
  }
  public Map<String, Set<TreeNode>> getLinks(TreeNode node) {
    Map<String, Map<String, List<Object>>> st = 
        lr.getCacheStorage().getCacheMapsByValue(node.getUUId().toString(), "revert_revItem",null);
    Map<String,Set<TreeNode>> revs = null;
    if(st!=null)
    for (String key : st.keySet()) {
      //String val = stored.get(key);
      //if (val == null) continue;
      //String[] paths = val.split("\\|");
      TreeDB db = node.getTreeDB();
      Attribute a = node.attrs().getAttribute(key);
      for(List<Object> lo:st.get(key).values())
        if(lo!=null)
          for (Object pathO : lo) {
            if(pathO == null)
              continue;
            String path = pathO.toString();
            TreeNode node2 = null;
            TreeDB adb = db;
            int pos = path.indexOf("::");
            if (pos > -1) {
              String name = path.substring(0, pos);
              ResManager tm = ResManager.getManager(); //TODO load?
              adb = tm.getTreeDB(name);
              if (adb != null) {
                path = path.substring(pos + 2);
                break;
              }
            }
            if(adb == null)
              adb = db;
            try {
              node2 = adb.getNode(UUID.fromString(path));
            } catch (Exception e) {
            }
            if (node2 != null && !node2.isDisposed()) {
                 String akey = key;
                boolean mult = false;
                int al = 0;
                ;
                if(revs == null)
                  revs = new HashMap<>();
                if(!revs.containsKey(key))
                  revs.put(key, new HashSet<>());
                revs.get(key).add(node2);
  //              getStorage().addLink(node2, node, key, true);
  //              getStorage().saveRevLinks(node);
              
            }
          }
    }
    if(revs == null)
      return Collections.emptyMap();
    return revs;
  }
  public boolean isRevLinksInited(String uid) {
    return lr.getCacheStorage().getCacheValue(uid, "revert", "_revInited") != null;
  }

  public boolean containsName(String nodeId) {
    Set<UUID> bn = getByName(nodeId);
    return bn!=null && bn.size()>0;
  }

  public void updateName(UUID uid, String newName) {
    if(lr.db == null)
      return;//cross-project
    String rid = lr.db.getRootNodeUUID().toString();
    //TreeDB db = lr.db;
    
    Object old = lr.getCacheStorage().getCacheValue(rid
        , "names","name", uid.toString());
    boolean eq = old!=null&&old.equals(newName) || old == null && newName == null;
    if(!eq){
      if(old!=null){
        Set<UUID> set = namecache.get(old.toString());
        if(set!=null)
          set.remove(uid);
        lr.getCacheStorage().removeByKey(
            rid, "names","name", uid.toString());
      }
      
      if(newName!=null){
        if(!namecache.containsKey(newName))
          namecache.put(newName, new HashSet<>());
        namecache.get(newName).add(uid);
        lr.getCacheStorage().setCacheValue(
            rid, "names","name", uid.toString(), newName);
      }
    }
    
  }

  public Set<UUID> getByName(String nodeId) {
    if(lr.db == null)
      return Collections.emptySet();//cross-project
  if(namecache.containsKey(nodeId))
      return namecache.get(nodeId);
    //TreeDB db = lr.db;
    UUID root = lr.db.getRootNodeUUID();
    if(root == null)
      return Collections.emptySet();
    //Map<String, List<Object>> store = db.getCacheStorage().getCacheMapByValue(lr.db.getRootNode().getUUId(), "names", "names");
    List<String> val = lr.getCacheStorage().getCacheKey2(root.toString(), "names", "name", nodeId);
    if(val!=null && val.size()>0){
      namecache.put(nodeId, new HashSet<>());
      for(String s:val)
        namecache.get(nodeId).add(UUID.fromString(s));
    }
    return namecache.get(nodeId);
  }

}
