/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import com.unitesk.requality.core.attribute.AppliedEnumValue;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.EnumDefinitionValue;
import com.unitesk.requality.core.attribute.EnumValue;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.IValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.VPFactory;

public class NodeMapper {
  private NodeMapper() {}

  public static UUID getUUId(String nodeString) throws JSONException {
    if (nodeString == null) {
      return null;
    }
    //fast vers
    nodeString = nodeString.trim();
    if (nodeString.endsWith("}")) {
      String temp = nodeString.substring(nodeString.length() - 47);
      if (temp.startsWith("\"uuid\"")) return UUID.fromString(temp.substring(9, temp.length() - 2));
      temp = nodeString.substring(nodeString.length() - 48);
      if (temp.startsWith("\"uuid\"")) return UUID.fromString(temp.substring(9, temp.length() - 3));
    }
    JSONObject obj = new JSONObject(nodeString);
    return UUID.fromString(obj.getString("uuid"));
  }

  public static Map<String, AttributeDefinition> getAttributes(
      UUID uuid, String nodeString) throws JSONException {
    return getAttributes(uuid, nodeString, true);
  }
  public static Map<String, AttributeDefinition> getAttributes(
      UUID uuid, String nodeString, boolean yeld) throws JSONException {
    if (nodeString == null) {
      return null;
    }
    List<AttributeDefinition> result = new LinkedList<>();
    JSONObject obj = null;
    try {
      obj = new JSONObject(nodeString);
      JSONObject attrObj = obj.getJSONObject("attributes");
      loadNewFormat(uuid, result, attrObj);
    } catch (Exception e) {
      //        	e.printStackTrace();
      //      	RLogger.logInfo(nodeString);
      if(yeld){
        e.printStackTrace();
        RLogger.logInfo(nodeString);
      }
    }
    Map<String, AttributeDefinition> ret = new HashMap<>();
    for (AttributeDefinition a : result) {
      //ret.put(a.getKey(),AttributeStorage.addToList(ret,ret.get(a.getKey()), a, uuid).);
      if (!ret.containsKey(a.getKey())) ret.put(a.getKey(), a);
      else {
        AttributeDefinition b = ret.get(a);
        if (a.isMoreVisible(null, b, uuid, true)) ret.put(a.getKey(), a);
      }
      //ret.put(a.getKey(),AttributeStorage.addToList(ret,ret.get(a.getKey()), a, uuid));
      //ret.get(a.getKey()).add(a);
      //TODO check order

    }
    return ret;
  }

  private static void loadNewFormat(
      UUID uuid, List<AttributeDefinition> result, JSONObject attr)
      throws JSONException {
    String[] jnames = JSONObject.getNames(attr);
    if (jnames == null) return; //TODO return throw?
    for (String jkey : jnames) {
      JSONObject a = attr.getJSONObject(jkey);
      UUID origin = uuid;
      String key = jkey;
      if (a.has("key")) {
        key = a.getString("key");
      }
      if (a.has("origin")) {
        String or = a.getString("origin");
        origin = UUID.fromString(or);
      }

      AttributeAvailability pa = AttributeAvailability.LOCAL;
      if (a.has("availability")) {
        String avail = a.getString("availability");
        pa = AttributeAvailability.valueOf(avail);
      }
      int flag = 0;
      if(a.has("flag"))
        flag = a.getInt("flag");
      AttributeDefinition rattr;
      IValueProvider vp = jsonToValueProvider(uuid, a.getJSONObject("value"));
      rattr = new AttributeDefinition(origin,key,pa,vp,flag);
      result.add(rattr);
    }
  }
  public static void saveValue(JSONObject setts, TypedValue tv) throws JSONException{
    saveValue(setts, tv.getType(),tv.getValue());
  }
  public static void saveValue(JSONObject setts, AttributeType at, Object val) throws JSONException{
    if(at!=AttributeType.STRING)
      setts.put("atype", at).toString();
    if(val!=null)
      valueToJSON(setts, at, val, "value");
  }

  public static TypedValue loadValue(JSONObject setts) throws JSONException {
    AttributeType type = AttributeType.STRING;
    if(setts.has("atype"))
      try {
        type = AttributeType.valueOf(setts.getString("atype"));
      } catch (JSONException e) {
        e.printStackTrace();
      }
    Object value = valueFromJSON(type, setts);
    return new TypedValue(type, value);
  }
  private static Object valueFromJSON(AttributeType type, JSONObject a) throws JSONException {
    Object value;
    if (type == AttributeType.BOOL) {
      if (!a.has("value")) return null;
      String v = a.getString("value");
      if ("true".equalsIgnoreCase(v) && !"false".equalsIgnoreCase(v)) {
        value = Boolean.valueOf(v);
      } else {
        value = v;
      }
    } else if (type == AttributeType.LIST) {
      value = listFromJSON(a);
    } else if (type == AttributeType.ENUM_DEFINITION) {
      value = enumDefinitionFromJSON(a);
    } else if (type == AttributeType.REFERENCE) {
      if (!a.has("value")) return null;
      value = a.getString("value");
    } else if (type == AttributeType.ENUM) {
      value = appliedEnumFromJSON(a);
    } else {
      if (!a.has("value")) return null;
      // TODO: all attribute types support
      try {
        if (!a.has("value")) value = AttributeType.getDefaultValue(type);
        else value = type.fromString(a.getString("value"));
      } catch (Exception e) {
        value = a.getString("value");
      }
    }
    return value;
  }

  private static AppliedEnumValue appliedEnumFromJSON(JSONObject a) throws JSONException {
    if (!a.has("enum_key")) return null;
    AttributeType at = AttributeType.STRING;
    if(a.has("elements_type"))
        at = AttributeType.valueOf(a.getString("elements_type"));
    EnumValue ev = enumFromJSON(at, a);
    return new AppliedEnumValue(ev.getValue(), ev.getComment(), at, a.getString("enum_key"));
  }

  private static EnumValue enumFromJSON(AttributeType at, JSONObject a) throws JSONException {
    Object val = valueFromJSON(at, a);
    String comm = null;
    if (a.has("comment")) comm = a.getString("comment");
    return new EnumValue(val, comm);
  }

  public static ListValue listFromJSON(JSONObject a) throws JSONException {
    if (!a.has("value")) return null;
    AttributeType ltype = AttributeType.STRING;
    if (a.has("elements_type")) ltype = AttributeType.valueOf(a.getString("elements_type"));
    ArrayList<Object> attrLst = new ArrayList<>();
    JSONArray vals = a.getJSONArray("value");
    for (int j = 0; j < vals.length(); j++) {
      Object o = vals.get(j);
      if (o instanceof JSONObject) attrLst.add(valueFromJSON(ltype, vals.getJSONObject(j)));
      else attrLst.add(o); //backward comp
    }
    return new ListValue(ltype, attrLst);
  }

  private static EnumDefinitionValue enumDefinitionFromJSON(JSONObject a) throws JSONException {
    if (!a.has("value")) return null;
    JSONArray vals = a.getJSONArray("value");

    AttributeType ltype = AttributeType.STRING;
    if (a.has("elements_type")) ltype = AttributeType.valueOf(a.getString("elements_type"));

    EnumDefinitionValue edv = new EnumDefinitionValue(ltype, new ArrayList<EnumValue>());

    for (int j = 0; j < vals.length(); j++)
      edv.addEnumValue(enumFromJSON(ltype, vals.getJSONObject(j)));

    //        if(ltype != AttributeType.LIST&&ltype != AttributeType.ENUM_DEFINITION)
    //        	for (int j = 0; j < vals.length(); j++) {
    //        		attrLst.put(ltype.fromString(vals.getString(j)),valsComments==null?null:valsComments.getString(j).equals("")?null:valsComments.getString(j));
    //        	}
    //        else
    //        {
    //        	if(ltype == AttributeType.LIST)
    //        		for (int j = 0; j < vals.length(); j++) {
    //        			attrLst.put(listFromJSON(vals.getJSONObject(j)),valsComments==null?null:valsComments.getString(j).equals("")?null:valsComments.getString(j));
    //        		}
    //
    //        	if(ltype == AttributeType.ENUM_DEFINITION)
    //        		for (int j = 0; j < vals.length(); j++) {
    //        			attrLst.put(enumDefinitionFromJSON(vals.getJSONObject(j)),valsComments==null?null:valsComments.getString(j).equals("")?null:valsComments.getString(j));
    //        		}
    //        }

    return edv;
  }

  public static final JSONObject toJSON(TreeNode node, boolean ignoreVirtual) throws JSONException {
    return toJSON(node.getUUId(), AttributeStorage.toDefMap(node.attrs().getLMap()).values(),ignoreVirtual); //TODO
  }

  public static final JSONObject toJSON(UUID uuid, Collection<AttributeDefinition> collection,boolean ignoreVirtual)
      throws JSONException {
    return toJSON(uuid, collection, false,ignoreVirtual);
  }

  public static final JSONObject toJSON(UUID uuid, Collection<AttributeDefinition> collection, boolean kdup, boolean ignoreVirtual)
      throws JSONException {
    JSONObject result = null;
    result = new JSONObject();
    result.put("uuid", uuid);
    JSONObject attrs = new JSONObject();
    for (AttributeDefinition a : collection) {
      if (ignoreVirtual && (a.isVirtual()||a.getKey().equals(TreeNode.ATTR_ID)))
        continue;
      if (!a.getOrigin().equals(uuid) && !a.isHidden()) continue;
      //if (a.getKey().startsWith("_%id")) continue; //ignore virt attr
      JSONObject attrObj = new JSONObject();
      //attrObj.put("key", a.getKey());
      //if (a.getType() != AttributeType.STRING) attrObj.put("type", a.getType());
      if (!a.getOrigin().equals(uuid)) attrObj.put("origin", a.getOrigin());
      //attrObj.put("version", a.getVersion().toString());
      //if (a.getRawValue() instanceof ArrayList) RLogger.logInfo("problem!");
      //Object val = a.getGValue();
      IValueProvider vp = a.getValueProvider();
      valueProviderToJSON(attrObj,a,vp);
      if(a.getFlags()!=0)
        attrObj.put("flag", a.getFlags());
      //TODO make valueProvider-specific save
      //valueToJSON(attrObj, a.getType(), a.getType(), vp.getRawValue());
      if (!a.getAvailability().equals(AttributeAvailability.LOCAL))
        attrObj.put("availability", a.getAvailability().name());
      
//      AttributeGenerator ag = null;
//      if(vp instanceof GeneratorValueProvider)
//        ag = ((GeneratorValueProvider)vp).getGenerator();
//      if (ag != null) {
//        JSONObject genObj = new JSONObject();
//        genObj.put("uuid", ag.getUUID());
//        genObj.put("attribute_type", ag.getAttributeType().ordinal());
//        if (ag.getAttributeListType() != null)
//          genObj.put("atribute_list_type", ag.getAttributeListType().ordinal());
//        genObj.put("generator_type", ag.getGeneratorType().getId());
//        JSONObject genAttrs = new JSONObject();
//        Map<String, Object> setts = ag.getSettings();
//        for (String s : setts.keySet())
//          if (setts.get(s) instanceof ListValue) {
//            JSONObject val = new JSONObject();
//            valueToJSON(val, AttributeType.LIST, AttributeType.LIST, (ListValue) setts.get(s));
//            genAttrs.put(s, val);
//          } else genAttrs.put(s, setts.get(s));
//        genObj.put("settings", genAttrs);
//        attrObj.put("generator", genObj);
//      }
      attrs.put(kdup ? a.getKey() : (a.getKey()), attrObj);
    }
    result.put("attributes", attrs);
    return result;
  }
  private static IValueProvider jsonToValueProvider(
      UUID or, JSONObject attrObj) throws JSONException {
    //JSONObject vpo = new JSONObject();
    //attrObj.put("valueProvider", vpo);
    String vt = ConstValueProvider.TYPE;
    if(attrObj.has("type")){
      JSONObject tp = attrObj.optJSONObject("type");
      if(tp != null)
        vt = tp.getString("value");
      else
        vt = attrObj.getString("type");
    }
    //if(attrObj.has("atype"))
      //  at = AttributeType.valueOf(attrObj.getString("atype"));
    ////Object o = valueFromJSON(at, attrObj);
    //Map<String, Object> setts = new HashMap<>();
    //JSONObject set = new JSONObject();
//    if(attrObj.has("settings"))
//      set = attrObj.getJSONObject("settings");
////    JSONArray names = set.names();
//    if (names != null)
//      for (int i = 0; i < names.length(); i++) {
//        try {
//          JSONObject jo = set.getJSONObject(names.getString(i));
//          setts.put(names.getString(i), listFromJSON(jo));
//        } catch (JSONException e) {
//          setts.put(names.getString(i), set.get(names.getString(i)));
//        }
//      }
    return VPFactory.initProvider(vt, or, attrObj);
  }
  private static void valueProviderToJSON(
      JSONObject attrObj, AttributeDefinition a, IValueProvider vp) throws JSONException {
    JSONObject vpo = new JSONObject();
    attrObj.put("value", vpo);
//    TypedValue tv = vp.getRawValue();
//    if(tv!=null){
//      //if(tv.getType()!=AttributeType.STRING)
//        //vpo.put("atype", tv.getType());//TODO mb remove?
//      valueToJSON(vpo, tv.getType(), tv.getValue(),"value");
//    }
//    Map<String,Object> setts = vp.getRawStores();
//    JSONObject genAttrs = new JSONObject();
//    if(setts!=null&&!setts.isEmpty()){
//      for (String s : setts.keySet())
//        if (setts.get(s) instanceof ListValue) {
//          JSONObject val = new JSONObject();
//          valueToJSON(val, AttributeType.LIST, setts.get(s),"value");
//          genAttrs.put(s, val);
//        } else genAttrs.put(s, setts.get(s));
//      vpo.put("settings", genAttrs);
//    }
    vp.saveJSON(vpo);
    if(!vp.getType().equals(ConstValueProvider.TYPE))
      vpo.put("type", vp.getType());
  }

  public static void valueToJSON(
      JSONObject attrObj, AttributeType type, Object rawValue,String key)
      throws JSONException {
    if (type.equals(AttributeType.ENUM)) {
      AppliedEnumValue aev = (AppliedEnumValue) rawValue;
      if (aev == null) return;
      if(aev.getType()!=AttributeType.STRING)
        attrObj.put("elements_type", aev.getType());
      attrObj.put("enum_key", aev.getBaseKey());
      String c = aev.getComment();
      if(c!=null && !c.isEmpty())
      attrObj.put("comment", c);
      valueToJSON(attrObj, aev.getType(), aev.getValue(),key);
      // 	        if(rawValue!=null)
      // 	        	attrObj.put("rawvalue",aev.toStored());
    } else if (type == AttributeType.LIST) {
      JSONArray lst = new JSONArray();
      ListValue lv = (ListValue) rawValue;
      if (rawValue == null) return;

      //Object[]val = (Object[])rawValue;
      AttributeType at = lv.getType();
      if(at!=AttributeType.STRING)
        attrObj.put("elements_type", at);
      for (Object o : lv) {
        JSONObject jso = new JSONObject();
        valueToJSON(jso, at, o,key);
        lst.put(jso);
      }
      attrObj.put(key, lst);
      //TODO new value
    } else if (type == AttributeType.ENUM_DEFINITION) {
      JSONArray lst = new JSONArray();
      EnumDefinitionValue edv = (EnumDefinitionValue) rawValue;
      if (edv == null) return;
      if(edv.getType()!=AttributeType.STRING)
        attrObj.put("elements_type", edv.getType());
      for (EnumValue ev : edv.getValues()) {
        JSONObject jo = enumValueToJSON(edv.getType(), ev);
        if (jo != null) lst.put(jo);
      }
      attrObj.put(key, lst);
      //TODO new value
    }else if (type == AttributeType.REFERENCE && rawValue instanceof ReferenceValue) {
        ReferenceValue edv = (ReferenceValue) rawValue;
        attrObj.put(key, edv.getRaw().toString());
      //TODO new value
    } else {
      
      if (rawValue == null) return;
//      if(node!=null&&(rawValue instanceof String))
//        rawValue = rawValue.toString().replace("{node.resURL}", node);
        
      attrObj.put(key, rawValue);
    }
  }

  private static JSONObject enumValueToJSON(AttributeType at, EnumValue ev) throws JSONException {
    if (ev == null) return null;
    JSONObject jo = new JSONObject();
    if (ev.getComment() != null) jo.put("comment", ev.getComment());
    valueToJSON(jo, at, ev.getValue(),"value");
    return jo;
  }

  public static JSONObject toJSON(UUID uid, Map<String, AttributeDefinition> map,boolean ignoreVirtual) throws JSONException {
    List<AttributeDefinition> coll = new LinkedList<>();
    for (String s : map.keySet())
      if (map.get(s) != null)
        //			for(Attribute a:map.get(s))
        coll.add(map.get(s));
    return toJSON(uid, coll, ignoreVirtual);
  }

  public static JSONObject toJSJSON(UUID uuid, LinkedList<Attribute> collection,TreeNode node) throws JSONException {
    //for editors js
    JSONObject result = null;
    result = new JSONObject();
    result.put("uuid", uuid);
    JSONObject attrs = new JSONObject();
    for (Attribute a : collection) {
      //if (!a.getOrigin().equals(uuid)) continue;
      JSONObject attrObj = new JSONObject();
      //attrObj.put("key", a.getKey());
      attrObj.put("type", a.getType());
      if (!a.getOrigin().equals(uuid)) attrObj.put("origin", a.getOrigin());
      //attrObj.put("version", a.getVersion().toString());
      //newval?
      TypedValue nw = a.getTValue(), rw = a.getDefinition().getRawValue();
        valueToJSON(attrObj, nw.getType(), nw.getValue(), "newvalue");
        if(rw != null)
          valueToJSON(attrObj, rw.getType(), rw.getValue(), "value"); 
      if (!a.getAvailability().equals(AttributeAvailability.LOCAL))
        attrObj.put("availability", a.getAvailability().name());
      attrs.put(a.getKey(), attrObj);
    }
    result.put("attributes", attrs);
    return result;
  }

	public static Map<String, TypedValue> convertToMap(JSONObject set) {
	  Map<String, TypedValue> setts = new HashMap<>();
	  JSONArray names = set.names();
	  if (names != null)
	    for (int i = 0; i < names.length(); i++) {
	      try {
	        JSONObject jo = set.getJSONObject(names.getString(i));
	        setts.put(names.getString(i), new TypedValue(AttributeType.LIST,listFromJSON(jo)));
	      } catch (JSONException e) {
	        try {
				setts.put(names.getString(i), Attribute.getTypedValue(set.get(names.getString(i))));
			} catch (JSONException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
	      }
	    }
	  return setts;
	}

}
