package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class TermsStorage extends DefaultTreeChangeListener{
  public static String TERMS_INITED = "terms-inited";
  public static String DEF_TERM = "def-term",
      USE_TERM = "use-term",
      USE_TERM_ALT = "usedef",
      TERM_REF_NAME = "term-";
  private CacheStorage cs; 
  private TreeDB db;
  public TermsStorage(CacheStorage cs, TreeDB db){
    this.cs = cs;
    this.db = db;
    db.addListener(this);
  }
//  public boolean isInited(){
//    //cs.getCacheValue(db.getRootNode().toString(), "", key)
//  }
  @Override
  public void changed(TreeNode node, ChangesSet<Attribute> what) {
    addRemoveUse(node, what.getOld(USE_TERM),what.getNew(USE_TERM));
    addRemoveUse(node, what.getOld(USE_TERM_ALT),what.getNew(USE_TERM_ALT));
    addRemoveDef(node, what.getOld(DEF_TERM),what.getNew(DEF_TERM));
  }

  private void addRemoveDef(TreeNode node, Attribute old, Attribute new1) {
    if(old != null || new1 != null){
      if(old != null)
        if(new1 == null)
          removeDefTerm(node, node.attrs().toStringArray(old, null));
        else
        {
          String[] tadd = node.attrs().toStringArray(new1, null);
          setDefTerm(node, tadd);
        }
      else
        if(new1!=null)
          setDefTerm(node, node.attrs().toStringArray(new1, null));
    }
  }

  private void addRemoveUse(TreeNode node, Attribute old, Attribute new1) {
    if(old != null || new1 != null){
      if(old != null)
        if(new1 == null)
          removeUseTerm(node, node.attrs().toStringArray(old, null));
        else
        {
          setUseTerm(node, node.attrs().toStringArray(new1, null));
        }
      else
        if(new1!=null)
          setUseTerm(node, node.attrs().toStringArray(new1, null));
    }
  }

  @Override
  public void created(TreeNode node) {
    String[]use1 = node.attrs().getStringArrayAttributeValue(USE_TERM);
    String[]use2 = node.attrs().getStringArrayAttributeValue(USE_TERM_ALT);
    String[]defterm = node.attrs().getStringArrayAttributeValue(DEF_TERM);
    if(use1!=null)
      setUseTerm(node, use1);
    if(use2!=null)
      setUseTerm(node, use2);
    if(defterm!=null)
      setDefTerm(node, defterm);
  }

  private void setUseTerm(TreeNode node, String[] use2) {
    if(use2 == null || use2.length == 0)
      return;
    String id = toIdentifier(node);
//    List<String> set = cs.getAltSet("useterm", id);
//    set.addAll(Arrays.asList(use2));
    cs.setAltSet("useterm", id, use2);
    updateRelated(getDirectTerms(node), true);
  }
  
  private List<String> getUseTerm(TreeNode node) {
    if(node == null)
      return null;
    String id = toIdentifier(node);
    List<String> set = cs.getAltSet("useterm", id);
    return set;
  }
  private List<String> getDefTerm(TreeNode node) {
    if(node == null)
      return null;
    String id = toIdentifier(node);
    List<String> set = cs.getAltSet("defterm", id);
    return set;
  }
  private void setDefTerm(TreeNode node, String[] defterm) {
    if(defterm == null || defterm.length == 0)
      return;
    String id = toIdentifier(node);
    cs.setAltSet("defterm", id, defterm);
    updateRelated(getRevertTerms(node),false);
  }

  private void updateRelated(Map<String, Set<TreeNode>> revertTerms, boolean rev) {
    Set<TreeNode> up = new HashSet<>();
    for(Entry<String, Set<TreeNode>> en:revertTerms.entrySet())
      up.addAll(en.getValue());
    for(TreeNode nde:up)
      nde.attrs().updateAttribute(rev?TreeNode.ATTR_TERMS_REV:TreeNode.ATTR_TERMS);
  }
  @Override
  public void deleted(TreeNode node) {
    String[]use1 = node.attrs().getStringArrayAttributeValue(USE_TERM);
    String[]use2 = node.attrs().getStringArrayAttributeValue(USE_TERM_ALT);
    String[]defterm = node.attrs().getStringArrayAttributeValue(DEF_TERM);
    if(use1!=null)
      removeUseTerm(node, use1);
    if(use2!=null)
      removeUseTerm(node, use2);
    if(defterm!=null)
      removeDefTerm(node, defterm);
  }

  private void removeDefTerm(TreeNode node, String[] defterm) {
    if(defterm == null || defterm.length == 0)
      return;
    Map<String, Set<TreeNode>> k = getRevertTerms(node);
    cs.removeByID(toIdentifier(node), "defterm");
    updateRelated(k,false);
  }

  private void removeUseTerm(TreeNode node, String[] use1) {
    if(use1 == null || use1.length == 0)
      return;
    Map<String, Set<TreeNode>> k = getDirectTerms(node);
    cs.removeByID(toIdentifier(node), "useterm");
    updateRelated(k,true);
  }
  private String toIdentifier(TreeNode node){
    return node.getUUId().toString();
  }
  public void initCaches(){
    initCaches(false);
  }
  public void initCaches(boolean ints) {
    db.removeListener(this);
    try{
      Object o = cs.getCacheValue(db.getRootNode().getUUId().toString(), "init", TERMS_INITED);
      if(o == null){
        String s = cs.startCacheTransaction("init terms cache ", CacheStorage.COMMIT_SINGLE);
        DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(db.getRootNode());
        java.util.Iterator<TreeNode> it = dftw.iterator();
        while(it.hasNext()){
          TreeNode node = it.next();
          created(node);
        }
        cs.setCacheValue(db.getRootNode().getUUId().toString(), "init", TERMS_INITED, true);
        if(s!=null)
          cs.commitCache();
      }
    }finally{
      db.addListener(this);
    }
  }
  public Map<String,Set<TreeNode>> getRevertTerms(TreeNode node) {
    Map<String,Set<TreeNode>> map = Collections.emptyMap();
    List<String> ls = getDefTerm(node);
    Map<String, List<String>> uses = null;
    if(ls==null||ls.isEmpty())
      return Collections.emptyMap();
    
    uses = getTermsUses(ls);
    TreeDB iroot = Instantiator.getRootTree(node.getTreeDB());
    List<String> als = new LinkedList<>();
    for(String s:ls)
      als.add(iroot.getResName()+"::"+s);
    
    for(TreeDB db2:ResManager.getManager().getTreeDBs()){
      if(iroot != db2){
        Map<String,List<String>>uses2 = db2.getTermsStorage().getTermsUses(als);
        if(uses2!=null && uses2.size()>0){
          if(uses == null || uses.isEmpty())
            uses = new HashMap<>();
          for(Entry<String, List<String>> en:uses2.entrySet())
          {
            String ak = en.getKey().substring(en.getKey().indexOf("::")+2);
            if(!uses.containsKey(ak))
              uses.put(ak, new LinkedList<>());
            for(String s:en.getValue())
              uses.get(ak).add(db2.getResName()+"::"+s);
          }
        }
      }
    }
    if(uses!=null && uses.size()>0){
      map = new HashMap<>();
      for(String s:uses.keySet()){
        String ps = /*TERM_REF_NAME+*/s;
        map.put(ps, new HashSet<>());
        for(String s2:uses.get(s)){
          TreeNode nde = node.getTreeDB().getNode(s2);
          if(nde!=null)
            map.get(ps).add(nde);
        }
      }
    }
    return map;
  }
  public Map<String,Set<TreeNode>> getDirectTerms(TreeNode node) {
    Map<String,Set<TreeNode>> map = Collections.emptyMap();
    List<String> ls = getUseTerm(node);
    if(ls!=null){
      Map<String, List<String>> uses = getUsedTerms(ls);
      Map<String,List<String>> fids = null;
      for(String s:ls)
      {
        int pos = s.indexOf("::");
        if(pos>-1)
        {
          if(fids == null) 
            fids = new HashMap<>();
          String did = s.substring(0,pos);
          s = s.substring(pos+2);
          if(!fids.containsKey(did))
            fids.put(did, new ArrayList<>());
          fids.get(did).add(s);
        }
      }
      if(fids!=null){
        for(Entry<String, List<String>> en:fids.entrySet()){
          TreeDB fdb = ResManager.getManager().getTreeDB(en.getKey());
          if(fdb!=null)
          {
            if(uses == null || uses.size() == 0)
              uses = new HashMap<>();
            Map<String, List<String>> use2 = fdb.getTermsStorage().getUsedTerms(en.getValue());
            if(use2!=null && !use2.isEmpty())
              for(Entry<String, List<String>> en2:use2.entrySet()){
                String nk = en.getKey()+"::"+en2.getKey();
                if(!uses.containsKey(nk))
                  uses.put(nk, new LinkedList<>());
                for(String s:en2.getValue())
                  uses.get(nk).add(en.getKey()+"::"+s);
              }
              
          }
        }
      }
      if(uses!=null && uses.size()>0){
        map = new HashMap<>();
        for(String s:uses.keySet()){
          String ps = /*TERM_REF_NAME+*/s;
          map.put(ps, new HashSet<>());
          for(String s2:uses.get(s)){
            TreeNode nde = node.getTreeDB().getNode(s2);
            if(nde!=null)
              map.get(ps).add(nde);
          }
        }
      }
    }
    return map;
  }
  private Map<String, List<String>> getTermsUses(List<String> ls) {
    return cs.findSetsByValue("useterm", ls);
  }
  private Map<String, List<String>> getUsedTerms(List<String> ls) {
    return cs.findSetsByValue("defterm", ls);
  }
  public void close() {
    db.removeListener(this);
  }
}
