package com.unitesk.requality.core;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.InstancesFolder;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.ReportFolder;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;

public class TreeNodeComparator {
  private static Map<SortingSettings, Comparator<TreeNode>> comparators =
      new HashMap<>();
  private static Map<SortingSettings, Comparator<TreeNode>> gcomparators =
      new HashMap<>();
  
  public static Comparator<TreeNode> getGlobalComparator(TreeDB db) {
    if (db.isVirtual()) db = Instantiator.getRootTree(db); // fix
    // TODO think about it
    SortingSettings st = db.getSortingSettings();
    return getComparator(st);
  }
  
  public static Comparator<TreeNode> getComparator(TreeDB db) {
    if (db.isVirtual()) db = Instantiator.getRootTree(db); // fix
    // TODO think about it
    SortingSettings st = db.getSortingSettings();
    return getComparator(st);
  }

  private static Map<String, Integer> order = new HashMap<>();

  static {
    order.put(DocFolder.TYPE_NAME, 1);
    order.put(InstancesFolder.TYPE_NAME, 2);
    order.put(ReportFolder.TYPE_NAME, 3);
    order.put(Requirement.TYPE_NAME, 4);
    order.put("TestProcedure", 5);
  }

  public static Comparator<TreeNode> getComparator(final SortingSettings settings) {
    if (!comparators.containsKey(settings)) {
      comparators.put(
          settings,
          new Comparator<TreeNode>() {
            @Override
            public int compare(TreeNode n1, TreeNode n2) {
              if (n1 == null) return -1;
              if (n2 == null) return 1;
              if (n1.getUUId().equals(n2.getUUId())) return 0;
              if (n1.isRoot() && n2.isRoot()) {
                return n1.getTreeDB().getName().compareTo(n2.getTreeDB().getName());
              }
              if (n1.getType() == null) return -1;
              if (n2.getType() == null) return 1;
              if (n1.getWorkParent() != null
                  && n1.getWorkParent().isRoot()
                  && n2.getWorkParent() != null
                  && n2.getWorkParent().isRoot()) {
                Integer i1 = order.get(n1.getType()), i2 = order.get(n2.getType());
                if (i1 != null && i2 != null)
                  if (i1 != i2) return i1.compareTo(i2);
                  else ;
                else if (i1 != null) return -1;
                else if (i2 != null) return 1;
              }
              if (settings !=null && settings.getSeparateByTypesFlag()
                  && (n1.getType() != null)
                  && (n2.getType() != null)
                  && (n1.getType().compareTo(n2.getType()) != 0)) {
                return n1.getType().compareTo(n2.getType());
              }
              if (n1.getType().equals(Location.TYPE_NAME)
                  && n2.getType().equals(Location.TYPE_NAME)) {
                int li = ((Location) n1).compareIndexTo((Location) n2);
                if (li != 0) return li;
              } // fix to sort text nodes in proper way
              if (settings !=null && settings.getMethod() == SortingMethod.BY_LOCATIONS
                  && mayHaveLocations(n1)
                  && mayHaveLocations(n2)) {
                if (!(n1 instanceof Requirement) || !(n2 instanceof Requirement))
                  RLogger.logInfo("error type loading! " + n1.toString() + " " + n2.toString());
                else {
                  int ci = ((Requirement) n1).compareLocationIndexes((Requirement) n2);
                  if (ci != 0) return ci;
                }
              }
              int k = 0;
              if (settings!=null && (settings.getMethod() == SortingMethod.BY_LOCATIONS // &&
                  // n1.getTreeDB().getTreeLogic().usesIndexByDefault(n1.getType())
                  || settings.getMethod() == SortingMethod.BY_INDEX))
                k = n1.getIndex().compareTo(n2.getIndex());
              if (k == 0) {
                String s1 = n1.getNameOrId(), s2 = n2.getNameOrId();
                return compareStrings(s1, s2);
              }
              return k;
            }

            private boolean mayHaveLocations(TreeNode n1) {
              return Requirement.TYPE_NAME.equals(n1.getType())
                  || TextNode.TYPE_NAME.equals(n1.getType());
            }
          });
    }
    return comparators.get(settings);
  }
  public static Comparator<TreeNode> getGComparator(final SortingSettings settings) {
    if (!gcomparators.containsKey(settings)) {
      gcomparators.put(
          settings,
          new Comparator<TreeNode>() {
            @Override
            public int compare(TreeNode n1, TreeNode n2) {
              if (n1 == null) return -1;
              if (n2 == null) return 1;
              if (n1.getUUId().equals(n2.getUUId())) return 0;
              TreeNode wp1=n1.getWorkParent(), wp2 = n2.getWorkParent();
              if(!wp1.getUUId().equals(wp2.getUUId())){
                int d1 = 0, d2 = 0;
                while(wp1!=null){
                  d1++;
                  wp1 = wp1.getWorkParent();
                }
                while(wp2!=null){
                  d2++;
                  wp2 = wp2.getWorkParent();
                }
                wp1 = n1.getWorkParent();
                wp2 = n2.getWorkParent();
                boolean isp = true;
                while (d1 != d2){
                  if (d1 > d2){
                    n1 = wp1;
                    wp1 = n1.getWorkParent();
                    d1--;
                  }else{
                    isp = false;
                    n2 = wp2;
                    wp2 = n2.getWorkParent();
                    d2--;
                  }
                }
                if(wp1!=null&&wp2!=null)
                  while(wp1!=null&&!wp1.getUUId().equals(wp2.getUUId())){
                    n1 = wp1;
                    wp1 = n1.getWorkParent();
                    n2 = wp2;
                    wp2 = n2.getWorkParent();
                  }
                if(n1.getUUId().equals(n2.getUUId()))//special case when we compare node and one of its successors
                  return isp?1:-1;//if n2 is parent of n1 then ret -1 else 1
              }
              return getComparator(settings).compare(n1, n2);
            }

            private boolean mayHaveLocations(TreeNode n1) {
              return Requirement.TYPE_NAME.equals(n1.getType())
                  || TextNode.TYPE_NAME.equals(n1.getType());
            }
          });
    }
    return gcomparators.get(settings);
  }
  private static int compareNumbers(String in1, String in2) {
    while (in1.length() > in2.length()) in2 = "0" + in2;
    while (in1.length() < in2.length()) in1 = "0" + in1;
    int i = 0;
    while (in1.charAt(i) == in2.charAt(i) && i < (in1.length() - 1)) i++;
    char c1 = in1.charAt(i), c2 = in2.charAt(i);
    if (c1 > c2) return 1;
    if (c1 < c2) return -1;
    return 0;
  }

  public static int compareStrings(String s1, String s2) {
    int k = 0;
    StringBuffer l1 = new StringBuffer(),
        l2 = new StringBuffer(),
        nu1 = new StringBuffer(),
        nu2 = new StringBuffer();
    int i1 = 0, i2 = 0;
    if(s1 == null)
      return -1;
    if(s2 == null)
      return 1;
    for (; i1 < s1.length() && i2 < s2.length(); ) {
      while (i1 < s1.length() && Character.isDigit(s1.charAt(i1))) {
        nu1.append(s1.charAt(i1));
        i1++;
      }
      while (i2 < s2.length() && Character.isDigit(s2.charAt(i2))) {
        nu2.append(s2.charAt(i2));
        i2++;
      }
      String in1 = "0", in2 = "0";
      if (nu1.length() > 0) in1 = nu1.toString();
      if (nu2.length() > 0) in2 = nu2.toString();
      if (nu1.length() > 0)
        if (nu2.length() > 0) k = compareNumbers(in1, in2);
        else k = -1;
      else if (nu2.length() > 0) k = 1;
      if (k != 0) return k;
      if (i1 < s1.length() && i2 < s2.length()) {
        while (i1 < s1.length() && !Character.isDigit(s1.charAt(i1))) {
          l1.append(s1.charAt(i1));
          i1++;
        }
        while (i2 < s2.length() && !Character.isDigit(s2.charAt(i2))) {
          l2.append(s2.charAt(i2));
          i2++;
        }
        k = l1.toString().compareToIgnoreCase(l2.toString());
        if (k != 0) return k;
      }
    }
    return s1.substring(i1).compareToIgnoreCase(s2.substring(i2));
  }
}
