package com.unitesk.requality.core.api;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.unitesk.requality.core.descriptors.CoverageStorageSourceDescriptor;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.tests.ICoverageStorageSource;

public class API {
  private static List<IAPISource> apis = new LinkedList<>();

  static {
    register(new DefaultAPISource());
  }

  private static MassOppDescriptor descrs[];

  public static void register(IAPISource source) {
    apis.add(source);
  }

  public static MassOppDescriptor[] getMassOppDescriptors() {
    if (descrs != null) return descrs;
    List<MassOppDescriptor> descrs = new LinkedList<>();
    for (IAPISource as : apis) {
      Collection<MassOppDescriptor> d = as.loadMassOpsDescriptors();
      if (d != null) descrs.addAll(d);
    }
    API.descrs = descrs.toArray(new MassOppDescriptor[0]);
    return API.descrs;
  }

  public static Map<String, CoverageStorageSourceDescriptor> getCoverageSources() {
    Map<String, CoverageStorageSourceDescriptor> ret = new HashMap<>();
    for (IAPISource api : apis) {
      Map<String, CoverageStorageSourceDescriptor> t = api.getCoverageSources();
      if (t != null) ret.putAll(t);
    }
    return ret;
  }

  public static ICoverageStorageSource getCoverageSource(String type) {
    CoverageStorageSourceDescriptor desc = getCoverageSources().get(type);
    if (desc != null) return desc.getStorageInput();
    return null;
  }

  public static Map<String, Template> getReportTemplates() {
    Map<String, Template> ret = new HashMap<>();
    for (IAPISource api : apis) {
      Map<String, Template> t = api.getReportTemplates();
      if (t != null) ret.putAll(t);
    }
    return ret;
  }
}
