package com.unitesk.requality.core.api;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URI;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.commons.httpclient.URIException;
import org.apache.commons.httpclient.util.URIUtil;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.operations.GenerateReportHandler;
import com.unitesk.requality.core.operations.ReportGenResult;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class ReqRESTMethods {
  ResManager tt;;

  public ReqRESTMethods(ResManager tracker) {
    tt = tracker;
  }

  private TreeDB getTreeDB(String name) {
    for (TreeDB db : tt.getTreeDBs()) if (name.equals(db.getName())) return db;
    return null;
  }

  private TreeNode getNode(TreeDB db, String path) {
    if (db == null) return null;
    try {
      return db.getNode(path, Path.QID | Path.UUID | Path.UV);
    } catch (RuntimeException re) {
      re.printStackTrace();
    }
    return null;
  }

  // projects subtree
  public boolean createProject(String name) {
    Set<String> ret = new HashSet<>();
    for (TreeDB db : tt.getTreeDBs()) ret.add(db.getName());
    if (ret.contains(name)) return false;
    TreeDB db = tt.createProject(name);
    if (db != null) return true;
    return false;
  }

  public boolean removeProject(String name) {
    for (TreeDB db : tt.getTreeDBs()) {
      if (name.equals(db.getName())) {
        try {
          ResManager.getManager() /* TODO params */.remove(db.getName());
          return true;
        } catch (Exception e) {
          e.printStackTrace();
        }
        return false;
      }
    }
    return false;
  }

  public String[] getProjectNames() {
    String[] ret = new String[tt.getTreeDBs().length];
    int i = 0;
    for (TreeDB db : tt.getTreeDBs()) ret[i++] = db.getName();
    return ret;
  }

  // requirements subtree
  public String[] listRequirements(String project) {
    TreeNode reqroot = getNode(getTreeDB(project), Requirement.getTypeRootQId());
    if (reqroot == null) return null;
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(reqroot, false, true, true);
    Iterator<TreeNode> it = dftw.iterator();
    List<String> l = new LinkedList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      l.add(node.getQualifiedId());
    }
    return l.toArray(new String[0]);
  }

  public boolean createRequirement(String project, String parent, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), parent);
    if (reqroot == null) return false;
    if (id != null && !id.isEmpty() && reqroot.findWorkChild(id) != null
        || !reqroot.getTreeDB().getTreeLogic().canBeChildOf(Requirement.TYPE_NAME, reqroot))
      return false; // already exists
    if (id == null || id.isEmpty())
      id = RequalityCLI.getNextId(reqroot, Requirement.TYPE_NAME, null);
    TreeNode chi = reqroot.createChild(id, Requirement.TYPE_NAME);
    return chi != null;
  }

  public boolean removeRequirement(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return false;
    try {
      reqroot.delete();
      return true;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }

  // can be called on any element
  public String getType(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    return reqroot.getType();
  }

  public String getDescription(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    try {
      return reqroot.getDescription();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public String getUVId(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    try {
      return reqroot.getUserVisibleName();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  // attributes subtree
  // работа со свойствами, может в принципе вызываться над всеми элементами
  public String[] listAttributeKeys(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    Set<String> keys = new HashSet<>();
    for (String s : reqroot.attrs().getAllAttributeKeys()) if (!s.startsWith("_")) keys.add(s);
    return keys.toArray(new String[0]);
  }

  public String getAttributeValue(String project, String id, String key) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    Attribute a = reqroot.attrs().getAttribute(key);
    if (a == null) return null;
    return Attribute.valueToString(reqroot.attrs(), a.getType(), a.getValue()); // TODO enum, list
  }

  public boolean setAttributeValue(String project, String id, String key, String value) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return false;
    reqroot.attrs().setAttribute(new AttributeDefinition(reqroot.getUUId(),
        key,AttributeAvailability.LOCAL,AttributeDefinition.detectValueProvider(key, AttributeType.STRING, value),
        AttributeDefinition.detectFlags(key)));
    reqroot.commitAttributes();
    return true;
  }

  public boolean removeAttribute(String project, String id, String key) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return false;
    Attribute a = reqroot.attrs().removeAttribute(key);
    if (a != null) reqroot.commitAttributes();
    return a != null;
  }

  // documents subtree
  public String[] listDocuments(String project) {
    TreeNode reqroot = getNode(getTreeDB(project), "/Documents");
    if (reqroot == null) return null;
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(reqroot, false, true, true);
    Iterator<TreeNode> it = dftw.iterator();
    List<String> l = new LinkedList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      if (Document.TYPE_NAME.equals(node.getType())) l.add(node.getQualifiedId());
    }
    return l.toArray(new String[0]);
  }

  // получение содержимого документа
  public String getDocumentContents(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    if (!Document.TYPE_NAME.equals(reqroot.getType())) return null;
    Document d = (Document) reqroot;
    org.w3c.dom.Document dom = null;
    try {
      dom = d.getDOM();
    } catch (Exception e1) {
      e1.printStackTrace();
    }
    if (dom == null) return null;
    try {
      ByteArrayOutputStream newXml = new ByteArrayOutputStream();
      Writer out = new OutputStreamWriter(newXml, "UTF-8");
      Result result = new StreamResult(out);
      Transformer t = TransformerFactory.newInstance().newTransformer();
      t.setOutputProperty(OutputKeys.METHOD,"xml");
      t.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
      t.transform(new DOMSource(dom), result);
      out.close();
      return newXml.toString();

    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public boolean removeDocument(String project, String id) {
    // TODO check is it proper
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null || !Document.TYPE_NAME.equals(reqroot.getType())) return false;
    try {
      reqroot.delete();
      return true;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }
  // reports subtree

  public String[] listReports(String project) {
    TreeNode reqroot = getNode(getTreeDB(project), "/Reports");
    if (reqroot == null) return null;
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(reqroot, false, true, true);
    Iterator<TreeNode> it = dftw.iterator();
    List<String> l = new LinkedList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      if (ReportData.TYPE_NAME.equals(node.getType())) l.add(node.getQualifiedId());
    }
    return l.toArray(new String[0]);
  }

  public String[] listReportSettings(String project) {
    TreeNode reqroot = getNode(getTreeDB(project), "/Reports");
    if (reqroot == null) return null;
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(reqroot, false, true, true);
    Iterator<TreeNode> it = dftw.iterator();
    List<String> l = new LinkedList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      if (ReportSettings.TYPE_NAME.equals(node.getType())) l.add(node.getQualifiedId());
    }
    return l.toArray(new String[0]);
  }

  public String getReportContents(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return null;
    if (!ReportData.TYPE_NAME.equals(reqroot.getType())) return null;

    ReportData rdata = (ReportData) reqroot;
    Template tpl = API.getReportTemplates().get(rdata.getTemplateId());
    URI savePath = rdata.getSaveDir();
    URI href = null;
    if (tpl == null) {
      if (rdata.getTemplateId() != null) {
        String lastGen = rdata.attrs().getStringAttributeValue(ReportData.ATTR_LAST_GENERATED);
        if (lastGen != null) {
          tpl = API.getReportTemplates().get(rdata.getTemplateId());
          // href = CoreUtils.getUrl(savePath + lastGen).toString();
        } // else we have new template and it is ridiculous
        return null;
      } else {
        tpl = API.getReportTemplates().get(Template.DEFAULT_ID);
      }
    }
    if (tpl != null)
      try {
        href =
            URI.create(
                savePath.toString()
                    + "/"
                    + URIUtil.encodeQuery(
                            tpl.getTemplates().get(0).getOutputFileName(reqroot.getTreeDB())));
      } catch (URIException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    if (href == null) return null;
    try {
      File f = new File(href);
      if (!f.exists() || !f.canRead()) return null;
      FileReader fr = new FileReader(f);
      BufferedReader br = new BufferedReader(fr);
      StringBuilder sb = new StringBuilder();
      while (br.ready()) {
        sb.append(br.readLine());
        sb.append("\n");
      }
      br.close();
      fr.close();
      return sb.toString();

    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public boolean generateReport(String project, String id) {
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null) return false;
    if (!ReportSettings.TYPE_NAME.equals(reqroot.getType())) return false;
    ReportSettings rs = (ReportSettings) reqroot;
    try{
      ReportGenResult rgr = GenerateReportHandler.getOne().generateReport(
        rs,null);
      return rgr!=null;
    }catch(Exception e){
      RLogger.logError(CoreUtils.getTrace(e));
    }
    return false;
  }

  public boolean removeReport(String project, String id) {
    // TODO check is it proper
    TreeNode reqroot = getNode(getTreeDB(project), id);
    if (reqroot == null || !ReportData.TYPE_NAME.equals(reqroot.getType())) return false;
    try {
      reqroot.delete();
      return true;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return false;
  }
}
