package com.unitesk.requality.core.attribute;

import java.util.Iterator;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.GeneratorValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.IValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;

public class AttributeDefinition {
//  //raw value is the value of attribute before all replaces or value in
//  //Requality specific object
//  private Object rawvalue;
//  //value with replaces or value in "java" format (Lists for LIST type and
//  //so on)
//  private Object newvalue;
  private final IValueProvider valueProvider;
  private final UUID origin;
  private final String key;
  private final AttributeAvailability availability;
  private final int flags;
  public AttributeDefinition(UUID origin, String key, IValueProvider vp, int flag){
    this(origin,key,AttributeAvailability.LOCAL,vp,flag);
  }
  public AttributeDefinition(UUID origin, String key, IValueProvider vp){
    this(origin,key,vp,detectFlags(key));
  }
  public static byte detectFlags(String key){
    byte res = 0;
    if(key.startsWith("_"))
      res |= Flag.HIDDEN;
    if(TreeNode.isVirtName(null,key)){
      res |= Flag.VIRTUAL;
      res |= Flag.HIDDEN;
    }
    return res;
  }
  public AttributeDefinition(UUID origin, String key, AttributeAvailability av, IValueProvider vp, int flags){
    this.origin = origin;
    this.availability = av;
    this.valueProvider = vp;
    this.flags = flags;
    this.key = key;
  }

  public AttributeDefinition(UUID origin, String key, AttributeType type, Object value) {
    this(origin, key, new ConstValueProvider(type, value));
  }
  
  boolean isMoreAVVisibleThan(AttributeDefinition rag2) {
    if (rag2 == null) return true;
    AttributeAvailability pa1 = this.getAvailability(), pa2 = rag2.getAvailability();
    if (pa1.equals(AttributeAvailability.LOCAL)) return !pa2.equals(AttributeAvailability.LOCAL);
    else if (pa1.equals(AttributeAvailability.DIRECT_CHILDREN))
      return pa2.equals(AttributeAvailability.SUBTREE);
    else return false;
  }

  boolean isSameAVVisibleWith(AttributeDefinition rag2) {
    if (rag2 == null) return true;
    AttributeAvailability pa1 = this.getAvailability(), pa2 = rag2.getAvailability();
    return pa1.equals(pa2);
  }

  public boolean isMoreVisible(TreeDB db, AttributeDefinition a, UUID where, boolean samemore) {
    if (a == null) return true;
    if (where.equals(getOrigin())) {
      if (!where.equals(a.getOrigin())) return true;
    } else {
      if (where.equals(a.getOrigin())) return false;
    }
    if (isGenerated()) {
      if (a.isGenerated()) {
      } else return false;
    } else if (a.isGenerated()) {
      return true;
    }
    return this.isMoreAVVisibleThan(a)
        || (this.isSameAVVisibleWith(a)); // )
  }
  public Set<String> getDependencies(AttributeStorage as) {
    if(valueProvider == null)
      return IValueProvider.NULLDEP;
    return valueProvider.getDependencies(as);
  }
  public AttributeAvailability getAvailability() {
    return availability;
  }
  public String getKey() {
    return key;
  }
  public TypedValue getRawValue() {
    return valueProvider.getRawValue();
  }
  TypedValue getTValue(AttributeStorage as) {
    assert as!=null;
    return valueProvider.getTypedValue(getKey(),as);
  }
  public UUID getOrigin() {
    return origin;
  }
  public boolean isGenerated() {
    // TODO Auto-generated method stub
    return valueProvider.getType().equals(GeneratorValueProvider.TYPE);
  }
  public AttributeType getType() {
    return valueProvider.getAttributeType();
  }
  public boolean isVirtual() {
    return (flags&Flag.VIRTUAL)==Flag.VIRTUAL;
  }
//  @Override
//  protected AttributeDefinition clone(){
//    return new AttributeDefinition(origin, key, type, availability, valueProvider.clone(), flags);
//  }
  public IValueProvider getValueProvider() {
    return valueProvider;
  }
  public Object getValue(AttributeStorage as) {
    TypedValue tv = getTValue(as);
    if(tv == null)
      return null;
    return tv.getValue();
  }
  public boolean weakEquals(Object obj) {
    if(obj == null)
      return false;
    if (super.equals(obj)) return true;
    if (!(obj instanceof Attribute)) {
      return false;
    }
    AttributeDefinition attr = (AttributeDefinition) obj;
    if (!key.equals(attr.getKey())) {
      return false;
    }
//    if (type != null && attr.getType() != null && !type.equals(attr.getType())) // TODO
//      // rec
//      return false;
    if (!availability.equals(attr.getAvailability())) return false;
    if (!getValueProvider().getRawValue().equals(attr.getValueProvider().getRawValue()))
      return false;
    return true;
  }
  public AttributeDefinition copy(UUID uuId) {
    if(uuId!=null && !uuId.equals(origin))
      return new AttributeDefinition(uuId, key, availability, valueProvider, flags);
    return this;
  }
  public Attribute toAttribute(TreeNode node) {
    return toAttribute(node.attrs());
  }
  public Attribute toAttribute(AttributeStorage as) {
    return new Attribute(this, as);
  }
  public int getFlags() {
    return flags;
  } 
  public static IValueProvider detectValueProvider(String key, AttributeType type, Object val) {
    return detectValueProvider(key, type, detectFlags(key), val);
  }
  public static IValueProvider detectValueProvider(String key, AttributeType type, int flag, Object val) {
    if(Attribute.getHTMLNames().contains(key))
      type = AttributeType.HTML;
    if(type == AttributeType.REFERENCE||type == AttributeType.LIST && 
        val!=null && ((ListValue)val).getType()==AttributeType.REFERENCE){
      return new ReferenceValueProvider(new TypedValue(type, val));
    }
    if(type == AttributeType.STRING||type == AttributeType.HTML){
      String av = null;
      if(val!=null)
        av = val.toString();
      if(type == AttributeType.HTML)
        return new HtmlValueProvider(type, av);
      if(av!=null){
        int s = av.indexOf("{");
        if(s>-1&&av.indexOf("}",s)>-1)
          return new TemplValueProvider(type, av);
      }
    }
    if(type == AttributeType.LIST && val != null){
      if(isTemplStringList(((ListValue)val).getType(),val))
        return new TemplValueProvider(type, val);
    }
    return new ConstValueProvider(type, val);
  }
  private static boolean isTemplStringList(AttributeType type2, Object val) {
    if(type2==AttributeType.STRING||type2==AttributeType.HTML){
      ListValue lv = (ListValue)val;
      for(Object o : lv.getValues())
      {
        String av = o.toString();
        int s = av.indexOf("{");
        if(s>-1&&av.indexOf("}",s)>-1)
          return true;
      }
    }
    if(type2==AttributeType.LIST){
      ListValue lv = (ListValue)val;
      Iterator<ListValue> it = lv.giterator();
      while(it.hasNext())
      {
        ListValue i = it.next();
        if(isTemplStringList(i.getType(), i))
          return true;
      }
    }
    return false;
  }
  public AttributeGenerator getGenerator() {
    if(valueProvider instanceof GeneratorValueProvider)
      return ((GeneratorValueProvider)valueProvider).getGenerator();
    return null;
  }
  public boolean isHidden() {
    return (flags&Flag.HIDDEN)==Flag.HIDDEN;
  }
  public boolean isReadOnly() {
    return (flags&Flag.READ_ONLY)==Flag.READ_ONLY;
  }
  public boolean isAvaiableOn(TreeNode node) { // does not check av correctness
    // TODO look into
    if (node == null) return false;
    if (getOrigin() == null) return true; // Attr to be added
    if (getOrigin().equals(node.getUUId())) return true;//if it is local then available
    TreeNode or = node.getTreeDB().getNode(getOrigin());
    if (or == null) return false;
    Attribute ora = or.attrs().getAttribute(getKey());
    if (ora == null || !ora.getOrigin().equals(or.getUUId())) // removed
        return false; // then required one was removed
    AttributeAvailability aa = ora.getAvailability();
    UUID paruid = node.getWorkParentUUID();
    if (!aa.equals(AttributeAvailability.LOCAL)) {
      if (paruid!=null && paruid.equals(getOrigin())) return true;
    }
    TreeNode par = node.getWorkParent();
    while(par!=null&&!par.getUUId().equals(or.getUUId())){
      par = par.getWorkParent();
    }
    // if subtree var is visible then all is correct
    return (or.isRoot()
                || or.isVRoot()
                || par!=null&&par.getUUId().equals(or.getUUId()))
            && aa.equals(AttributeAvailability.SUBTREE)
        || isAvaiableOn(node.getCloneParent());
  }
  public String toString(AttributeStorage as) {
    StringBuilder sb = new StringBuilder(key);
    sb.append('[');
    sb.append(getType());
    sb.append(']');
    sb.append('=');
    sb.append(Attribute.valueToString(as, getType(), getRawValue().getValue()));
    if(origin!=null)
    {
      sb.append('(');
      sb.append(origin);
      sb.append(')');
    }
    return sb.toString();
  }
//  public boolean isReference(AttributeStorage as) {
//    return getType().equals(AttributeType.REFERENCE)
//        || getType().equals(AttributeType.LIST) && 
//            getValue(as)!=null && 
//            ((ListValue)getValue(as)).getType().equals(AttributeType.REFERENCE);
//  }
}
