package com.unitesk.requality.core.attribute;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.LinkResolver;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;

public class ReferenceValue implements IContextStringProvider{
  private AttributeType at = null;
  private Object raw;
  private boolean isError = false;
  private String key = null, stringValue = null;
  private UUID[] nodes;
  private Map<UUID, String>trees;
  public ReferenceValue(AttributeType type, Object o, String key) {
    this.at = type;
    this.raw = o;
    if(raw instanceof ReferenceValue)
      throw new RuntimeException("Ref value where it does not supposed to be");
    this.key = key;
  }
  public ReferenceValue(String key, String error) {
    this.at = AttributeType.REFERENCE;
    this.raw = "";
    this.key = key;
    stringValue = error;
    isError = true;
  }
  public ReferenceValue(String key, Set<TreeNode> set, AttributeStorage where) {
    this.at = AttributeType.REFERENCE;
    this.raw = "";
    this.key = key;
    stringValue = "resolved term";//TODO mb rewrite
    nodes = new UUID[set.size()];
    int i = 0;
    for(TreeNode node:set){
      if(!node.isSameTree(where.getNode())){
        if(trees == null)
          trees = new HashMap<>();
        trees.put(node.getUUId(), node.getTreeDB().getResName());
      }
      nodes[i++] = node.getUUId();
    }
  }
  public UUID[] getNodeUUIDs(AttributeStorage where) {
    if(stringValue == null)
      resolve(where, true);
    return nodes;
  }
  public TreeNode[] getNodes(AttributeStorage where) {
    return getNodes(where, true);
  }
  public TreeNode[] getNodes(AttributeStorage where, boolean persist) {
    UUID[]nds = nodes;
    if(stringValue == null || !persist)
      nds = resolve(where, persist);
    if(nds == null)
      return null;
    TreeNode[]anodes = new TreeNode[nds.length];
    int i = 0;
    for(UUID uid:nds){
      TreeDB db = where.getNode().getTreeDB();
      if(trees!=null){
        String aname = trees.get(uid);
        if(aname!=null)
          db = ResManager.getManager().getTreeDB(aname);
      }
      if(db!=null){
        TreeNode node = db.getNode(uid);
        if(node!=null)
          anodes[i++] = node;
      }
    }
    if(i<anodes.length)
      return Arrays.copyOf(anodes, i);
    return anodes;
  }
  public UUID[] resolve(AttributeStorage where,boolean persist){
    isError = false;
    LinkResolver lr = where.getNode().getTreeDB().getLinkResolver();
    if(at==AttributeType.REFERENCE){
      Object val = raw;
      if(val == null){
        if(persist){
          stringValue = ReferenceValueProvider.NULL_MESSAGE;
          isError = true;
        }
      }else{
        if(val instanceof String && ((String)val).startsWith(ReferenceValueProvider.ERROR_CHAR)){
          if(persist){
            stringValue = val.toString();
            isError = true;
          }
        }else{
          List<TreeNode> lvars = Collections.emptyList();
          try {
            lvars = lr.getNotDisposedResolvePath(where.getNode(), key, where.getNode().getTreeDB(), raw.toString(), LinkResolver.RESOLVE_MODE);
          } catch (NodeNotFoundException e) {
          }
          if(lvars.size() == 0){
            if(persist){
              stringValue = ReferenceValueProvider.MISSING_MESSAGE.replace("{val}", raw.toString());
              isError = true;
            }
            return null;
          }
          UUID[]ret = null;
          if(lvars.size() == 1)
          {
            TreeNode node = lvars.iterator().next();
            ret = new UUID[]{node.getUUId()};
            if(persist){
              nodes = ret;
              stringValue = node.getUserVisibleName();
            }
            if(!node.isSameTree(where.getNode())){
              if(trees == null)
                trees = new HashMap<>();
              trees.put(node.getUUId(), node.getTreeDB().getResName());
              if(persist)
                stringValue = node.getTreeDB().getResName()+"::"+stringValue;
            }
          }else
          {
            ret = new UUID[lvars.size()];
            int i = 0;
            StringBuilder sb = new StringBuilder();
            if(persist)
              trees = new HashMap<>();
            for(TreeNode node:lvars){
              ret[i++] = node.getUUId();
              if(persist){
                if(sb.length()>0)
                  sb.append(", ");
                if(!node.isSameTree(where.getNode())){
                  sb.append(node.getTreeDB().getResName());
                  sb.append("::");
                  trees.put(node.getUUId(),node.getTreeDB().getResName());
                }
                sb.append(node.getUserVisibleName());
              }
            }
            if(persist){
              stringValue = "["+sb.toString()+"]";
              nodes = ret;
            }
            //nodes = lvars.toArray(new TreeNode[0]);
          }
          return ret;
        }
      }
    }
    return null;
  }
  
  @Override
  public String toString() {
    return stringValue!=null ? stringValue : "Unresolved "+raw;

  }
  public Object getRaw() {
    return raw;
  }
  public AttributeType getType() {
    return at;
  }
  public String getKey() {
    return key;
  }
  @Override
  public String toString(AttributeStorage where) {
    if(stringValue == null)
      resolve(where, true);
    return stringValue;
  }
  public boolean hasErrorsOnResolve(AttributeStorage where) {
    if(stringValue == null)
      resolve(where, true);
    return isError;
  }
  public static List<TreeNode> getValues(AttributeStorage where, AttributeType type, Object value, boolean persist) {
    if(value == null)
      return Collections.emptyList();
    if(type == AttributeType.REFERENCE){
      TreeNode[]nodes = ((ReferenceValue)value).getNodes(where);
      if(nodes == null)
        return null;
      return Arrays.asList(nodes);
    }
    LinkedList<TreeNode> ret = new LinkedList<>();
    ListValue lv = (ListValue)value;
    for(Object o:lv){
      List <TreeNode> conv = getValues(where, lv.getType(), o,persist);
      if(conv!=null)
        ret.addAll(conv);
    }
    return ret;
  }
  public static void resetAttribute(AttributeStorage as , AttributeType at, Object val) {
    if(at.equals(AttributeType.REFERENCE)){
      ReferenceValue rv = (ReferenceValue)val;
      rv.nodes = null;
      rv.isError = false;
      rv.stringValue = null;
      rv.trees = null;
    }else if(at == AttributeType.LIST){
      ListValue lv = (ListValue)val;
      for(Object o:lv.getValues())
        resetAttribute(as, lv.getType(), o);
    }
  }
}
