package com.unitesk.requality.core.attribute.valueproviders;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.tools.RequalityCLI;

public class HtmlValueProvider extends TemplValueProvider {
  public static String TYPE="HTMLValueProvider";
  private File dir = null;
  private List<File> added;
  public HtmlValueProvider(){
    
  }
  public HtmlValueProvider createAppended(AttributeType html,Object ds){
    return new HtmlValueProvider(html,ds,null,dir,added);
  }
  public HtmlValueProvider(AttributeType html, Object string) {
    this(html,string,null,null,null);
  }
  public HtmlValueProvider(
      AttributeType html, Object ds, Set<String> deps, File dir2, List<File> added2) {
    this.dir = dir2;
    this.added = added2;
    if(added == null)
      added = new LinkedList<>();
    if(dir==null){
      dir = CoreUtils.getTempDir().resolve("htmlvp"+UUID.randomUUID().toString()).toFile();
      if(!dir.exists())
        dir.mkdirs();
      //dir = Files.createTempDirectory("htmlvp"+UUID.randomUUID().toString()).toFile();
      dir.deleteOnExit();
    }
    init(html,ds,deps);
  }
  @Override
  public String getType() {
    return TYPE;
  }
  @Override
  protected void init(AttributeType at2, Object val2, Set<String> dps) {
    if(added == null)
      added = new LinkedList<>();
    if(dir==null){
      dir = CoreUtils.getTempDir().resolve("htmlvp"+UUID.randomUUID().toString()).toFile();
      if(!dir.exists())
        dir.mkdirs();
      //dir = Files.createTempDirectory("htmlvp"+UUID.randomUUID().toString()).toFile();
      dir.deleteOnExit();
    }
    Object v3 = recHndl(at2,val2);
    if(v3!=null)
      val2 = v3;
    super.init(at2,val2,dps);
    
  }
  public Object recHndl(AttributeType at, Object val){
    if(at == AttributeType.LIST){
      ListValue lv = (ListValue)val;
      List<Object>nval = new ArrayList<>();
      boolean changed = false;
      if(lv.getType()==AttributeType.LIST)
        for(Object o:lv.getValues()){
          Object r = (recHndl(lv.getType(), o));
          if(r != null){
            nval.add(r);
            changed = true;
          }
          else 
            nval.add(o);
        }
      else
        if(lv.getType()==AttributeType.STRING||lv.getType()==AttributeType.HTML){
          for(Object o : lv.getValues()){
            String s = o.toString();
            Object r = recHndl(lv.getType(),s);
            if(r != null){
              nval.add(r);
              changed = true;
            }else
              nval.add(o);
          }
        }
      if(changed)
        return new ListValue(at, nval);
      else
        return lv;
    }else
      if(at == AttributeType.STRING||at == AttributeType.HTML){
        return hndlString(val!=null?val.toString():null);
      }
    return null;
  }
  private Object hndlString(String s) {
    Object odesc = s, raw = s;
    // if(raw!=null){
    // if desc is null then we also needs to delete all images
    String desc = raw == null ? null : raw.toString();
    desc =
        RequalityCLI.updateImages(dir.toURI(), desc, added).replace((char) 160, ' ');
    if (raw != null) {
      try {
        // TODO can be removed?
        desc = desc.replace(dir.toURI().toString(), "{node.resURL}");
      } catch (Exception e) { // unsop
      }
      //return new AttributeDefinition(nlist.getOrigin(),
        //  nlist.getKey(), new TemplValueProvider(AttributeType.HTML, desc));
    }
    if(!desc.equals(odesc))
      return desc;
    return null;
  }
  private static Map<String,String> getContext(String key, AttributeStorage as){
    Map<String,String> ret = new HashMap<>(1);
    StringBuilder sb = new StringBuilder();
    URI path = as.getNode().getNodeResourcesPath(key);
    if(path!=null)
      sb.append(path.toString());
    ret.put("node.resURL", sb.toString());
    return ret;
  }
  
  @Override
  public TypedValue getTypedValue(String key, AttributeStorage as) {
    TypedValue tv = new TypedValue(at, calculateValue(as, at, val ,getContext(key, as)));
    //TypedValue tv = super.getTypedValue(key, as);
    if(as!=null){
      Set<String> res = as.getNode().listResources();
      for(File f:added){
        java.nio.file.Path p = dir.toPath().relativize(f.toPath());
        String name = p.toString();
        if(!res.contains(name)){
          FileInputStream fis = null;
          try {
            fis = new FileInputStream(f);
            as.getNode().setResourceContent(key+"/"+name, fis);
          } catch (FileNotFoundException e) {
            e.printStackTrace();
          }
          finally{
            if(fis!=null)
              try {
                fis.close();
              } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
              }
          }
        }
      }
      added.clear();
    }
    return tv;
  }
  public List<File> getAdded() {
    return added;
  };
}
