package com.unitesk.requality.core.attribute.valueproviders;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.json.JSONException;
import org.json.JSONObject;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.attribute.TypedValue;

public class ReferenceValueProvider implements IValueProvider {
  public static final String ERROR_CHAR = "*";
  public static final String MISSING_MESSAGE = ERROR_CHAR+"Missing: '{val}'";
  public static final String NULL_MESSAGE =  ERROR_CHAR+"Null value";
  private TypedValue raw;
  public static String TYPE="ReferenceValueProvider";
  public ReferenceValueProvider(){
  }
  
  public ReferenceValueProvider(TypedValue tv) {
    raw = tv;
  }
  public ReferenceValueProvider(AttributeType at,Object val) {
    raw = new TypedValue(at, val);
  }

  @Override
  public TypedValue getTypedValue(String key, AttributeStorage as) {
    if(raw.getType()!=AttributeType.LIST){
      if(raw.getGValue() instanceof ReferenceValue)
        throw new RuntimeException("Ref value is where it does not supposed to be");
      return new TypedValue(AttributeType.REFERENCE, new ReferenceValue(AttributeType.REFERENCE, raw.getGValue(),key));
    }
    else{
      ListValue lv = raw.getGValue();
      List<Object> v = new LinkedList<>();
      for(Object o:lv.getValues())
        if(o instanceof ReferenceValue)
          v.add(o);
        else
          v.add(new ReferenceValue(AttributeType.REFERENCE, o,key));
      return new TypedValue(AttributeType.LIST, new ListValue(AttributeType.REFERENCE, v));
    }
//    List<TreeNode> res = lr.resolveLink(lr, as.getNode(), key, null, key, true, true);
//    if(res.size()>0){
//      List<Object> ret = new ArrayList<>();
//      for(TreeNode node:res)
//        ret.add(node.getUUId().toString());
//      return new TypedValue(AttributeType.LIST,new ListValue(AttributeType.REFERENCE, ret));
//    }
//    return new TypedValue(AttributeType.REFERENCE, res.isEmpty()?null:res.iterator().next().getUUId().toString());
  }

  @Override
  public Set<String> getDependencies(AttributeStorage as) {
    return NULLDEP;
  }


  @Override
  public String getType() {
    return TYPE;
  }

  @Override
  public TypedValue getRawValue() {
    return raw;
  }

  @Override
  public Map<String, Object> getRawStores() {
    return null;
  }
  
  @Override
  public void attach(TreeNode node) {
  }

  @Override
  public void detach(TreeNode node) {
  }

  @Override
  public void saveJSON(JSONObject vpo) throws JSONException {
    NodeMapper.saveValue(vpo, raw);
  }
  @Override
  public AttributeType getAttributeType() {
    return raw.getType();
  }
}
