package com.unitesk.requality.core.attribute.valueproviders;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.TypedValue;

public class VPFactory {
  private static Map<String,IVPCreator> constrs = new HashMap<>();
  private static VPFactory ft = new VPFactory();
  static{
    constrs.put(ConstValueProvider.TYPE, new IVPCreator(){
      @Override
      public IValueProvider create(JSONObject object) throws JSONException {
        return new ConstValueProvider(NodeMapper.loadValue(object));
      }
      
    });
    constrs.put(ReferenceValueProvider.TYPE, new IVPCreator(){
      @Override
      public IValueProvider create(JSONObject object) throws JSONException {
        return new ReferenceValueProvider(NodeMapper.loadValue(object));
      }
      
    });
    constrs.put(TemplValueProvider.TYPE, new IVPCreator(){
      @Override
      public IValueProvider create(JSONObject object) throws JSONException {
        Set<String> deps = new HashSet<>();
        if(object.has("deps")){
          JSONArray ar = object.getJSONArray("deps");
          for(int i = 0;i<ar.length();i++)
            deps.add(ar.getString(i));
        }
        TypedValue tv = NodeMapper.loadValue(object);
        return new TemplValueProvider(tv,deps);
      }
    });

    constrs.put(HtmlValueProvider.TYPE,new IVPCreator(){
    
      @Override
      public IValueProvider create(JSONObject object) throws JSONException {
        Set<String> deps = new HashSet<>();
        if(object.has("deps")){
          JSONArray ar = object.getJSONArray("deps");
          for(int i = 0;i<ar.length();i++)
            deps.add(ar.getString(i));
        }
        TypedValue tv = NodeMapper.loadValue(object);
        File f = CoreUtils.getTempDir().resolve("htmlvp"+UUID.randomUUID().toString()).toFile();
        f.deleteOnExit();
        return new HtmlValueProvider(tv.getType(),
            tv.getValue(),
            deps,
            f,
            new ArrayList<>());
      }
    });
    constrs.put(GeneratorValueProvider.TYPE, new IVPCreator(){
      @Override
      public IValueProvider create(JSONObject object) throws JSONException {
        UUID uid = UUID.fromString(object.getString("uuid"));
        Set<String> keys = new HashSet<>();
        Map<String,TypedValue>setts = new HashMap<>();
        java.util.Iterator<Object> it = object.keys();
        while(it.hasNext()){
          Object o = it.next();
          if(o!=null){
            JSONObject jo = object.optJSONObject(o.toString());
            if(jo!=null)
              setts.put(o.toString(), NodeMapper.loadValue(jo));
            else
              setts.put(o.toString(), Attribute.getTypedValue(object.getString(o.toString())));
          }
        }
        
        return new GeneratorValueProvider(uid, setts);
      }
    });
    //constrs.put(ReferenceValueProvider.TYPE, ReferenceValueProvider.class);
  }
  public static IValueProvider initProvider(String type, UUID origin, JSONObject set){
    IVPCreator c = constrs.get(type);
    if(c==null)
      return null;
    IValueProvider cs = null;
    try {
      cs = c.create(set);
    } catch (JSONException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    if(cs==null)
      return null;
    return cs;
  }
  public static VPFactory getOne() {
    // TODO Auto-generated method stub
    return ft;
  }
}
