package com.unitesk.requality.core.checkers;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.ErrorNode;

public class CheckerContainer {
  private ArrayList<IChecker> checkers = new ArrayList<>();
  private TreeNode curNode;

  public IMark[] check(TreeNode node) {
    return check(node, false);
  }

  private Set<TreeNode> checked = new HashSet<>();

  public IMark[] check(TreeNode node, boolean batch) {
    return doCheck(node, batch);
  }

  protected IMark[] doCheck(TreeNode node, boolean batch) {
    if (node.isDisposed() ) // deleted
    return null;
    checked.add(node);
    curNode = node;
    ArrayList<IMark> result = new ArrayList<>(); 
    node.getTreeDB().startChangeRow();
    try{
      if (node instanceof ErrorNode) {
        result.add(new CheckerProblem(node, "Corrupted element"));
      }
      for (IChecker checker : new ArrayList<>(checkers))
        //temporary do not check virt nodes
        if (!node.isVirtual()&&!checker.check(node, result)) checked.remove(node);
      applyProblemsInfo(result);
    }finally{
      node.getTreeDB().stopChangeRow(); 
    }
    return result.toArray(new CheckerProblem[0]);
  }

  public TreeNode getNode() {
    return curNode;
  }

  protected void applyProblemsInfo(ArrayList<IMark> result) {
    for (IMark problem : result) {
      RLogger.logInfo(problem);
    }
  }

  public void addChecker(IChecker checker) {
    checkers.add(checker);
  }

  public void clearChecked() {
    checked.clear();
  }

  public boolean isChecked(TreeNode node) {
    return checked.contains(node);
  }

  public void markChecked(TreeNode node) {
    checked.add(node);
  }

  public void markUnChecked(TreeNode node) {
    checked.remove(node);
  }
}
