package com.unitesk.requality.core.checkers;

import java.util.Arrays;
import java.util.List;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;

public class LocationRefsChecker implements IChecker {
  public LocationRefsChecker() {}

  @Override
  public boolean check(TreeNode node, List<IMark> result) {
    // null type is an error
    if (node.isDisposed()
        || node.getType() == null
        || (!node.getType().equals(Requirement.TYPE_NAME)
            && !node.getType().equals(TextNode.TYPE_NAME)
            && !node.getType().equals(Location.TYPE_NAME))) {
      return true;
    }
    // if(node.getType().equals(Document.TYPE_NAME)){
    //
    // Document d = (Document)node;
    // //TreeNode[]reqs = l.getWorkChildren();
    // for(UUID u:d.getLocations()){
    // TreeNode l = (TreeNode) node.getTreeDB().getNode(u);
    // if(l == null)
    // result.add(new CheckerProblem(l.getWorkParent(), "Location with UUID:
    // " + u
    // + " was missing"));
    // }
    //
    // }else
    if (node.getType().equals(Location.TYPE_NAME)) {
      if (Location.isLoading(Instantiator.getRootTree(node.getTreeDB()))) return false;

      Location l = (Location) node;
      Requirement[] reqs = l.getRequirements(true,false);
      if (reqs == null) return false; // Loading
      if (reqs.length == 0 || reqs[0] == null) {
        reqs = l.getRequirements(false,false);
        result.add(
            new CheckerProblem(
                l.getWorkParent(), "Location with QId: " + l.getQualifiedId() + " was unlinked"));
      }
    } else
      try {
        Requirement req = (Requirement) node;
        IMark[] probs = req.getTreeDB().getCheckerProblems(req);
        if (probs != null) {
          result.addAll(Arrays.asList(probs));
          // Instantiator.getRootTree(req.getTreeDB());
          List<ReferenceValue> locQIds = req.getLocationQIds();
          for (ReferenceValue qid : locQIds) {
            TreeDB db = req.getTreeDB();
            if(qid.hasErrorsOnResolve(req.attrs()))
              continue;
            try {
              TreeNode location = qid.getNodes(req.attrs())[0];
              if (location == null) {
                String fqid = qid.getRaw().toString();
                if (fqid.contains("/")) fqid = fqid.substring(0, fqid.lastIndexOf("/"));
                TreeNode doc = db.getNode(fqid, Path.QID | Path.UUID);
                if (doc != null)
                  result.add(
                      new CheckerProblem(
                          doc,
                          "Location with QId: "
                              + qid
                              + " was not found in "
                              + req.getQualifiedId()));
                else
                  result.add(
                      new CheckerProblem(
                          req,
                          "Location with QId: "
                              + qid
                              + " and related document were not found in "
                              + req.getQualifiedId()));
              } else if (!location.getType().equals(Location.TYPE_NAME)) {
                result.add(
                    new CheckerProblem(
                        req,
                        "Location with QId: " + qid + " has wrong type " + location.getType()));
              }
            } catch (Exception e) {
              e.printStackTrace();
            }
          }
        }
      } catch (ClassCastException e) {
        result.add(new CheckerProblem(node, "Cannot load class for " + node.getQualifiedId()));
      }
    return true;
  }
}
