package com.unitesk.requality.core.checkers;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.TypedValue;

public class NodeAttribChecker implements IChecker {

  public NodeAttribChecker() {}

  @Override
  public boolean check(TreeNode node, List<IMark> result) {
    // Set<String> attrs = node.getAttributes().getAttributeKeys();
    if (node.isRoot()) // ignore attrs stored in root
    return true;
    Collection<Attribute> tocheck = new LinkedList<>(node.attrs().getLMap().values());
    for (Attribute attr : tocheck) {
      if (attr == null) {
        result.add(new CheckerProblem(node, "Attribute \"" + "\" is null"));
        continue;
      }
      String key = attr.getKey();
      if (!Attribute.checkNameIsValid(key, true)) {
        result.add(new CheckerProblem(node, "Attribute \"" + key + "\" has incorrect key"));
        continue;
      }

      TypedValue tv = attr.getTValue();
      Object val = tv.getValue();
      if (val == null) {
        if (node.getUUId().equals(attr.getOrigin()))
          result.add(new CheckerProblem(node, "Attribute \"" + key + "\" has null value"));
        continue;
      }
      String value = val.toString();
      //String value = attr.getRawValue().toString();
      // TODO check another non-result problems
      switch (tv.getType()) {
        case INT:
          try {
            Integer.parseInt(value);
          } catch (NumberFormatException e) {
            result.add(
                new CheckerProblem(
                    node,
                    "Attribute \"" + key + "\" of type INT has wrong value: \"" + value + "\""));
          }
          break;
        case FLOAT:
          try {
            Float.parseFloat(value);
          } catch (NumberFormatException e) {
            result.add(
                new CheckerProblem(
                    node,
                    "Attribute \"" + key + "\" of type FLOAT has wrong value: \"" + value + "\""));
          }
          break;
        case BOOL:
          if (!"true".equalsIgnoreCase(value) && !"false".equalsIgnoreCase(value)) {
            result.add(
                new CheckerProblem(
                    node,
                    "Attribute \"" + key + "\" of type BOOL has wrong value: \"" + value + "\""));
          }
          break;
        case STRING:
          // TODO add string check?
          // if(!Attribute.checkNameIsValid(key, true))
          // result.add(new CheckerProblem(node, "Attribute \"" + key +
          // "\" of type STRING has wrong value: \"" + value + "\""));
        default:
          break;
      }
    }
    // ResourceStorage rs =
    // EclipseTracker.getResourceStorage(node.getTreeDB());
    // if(rs!=null){
    // IFile f = rs.getResource(node.getUUId());
    // if(f!=null){
    // UUID uid = rs.getFileUUID(f);
    // if(uid!=null && !uid.equals(node.getUUId())){
    // node.saveAttributes();
    // result.add(new CheckerProblem(node, "UUID for \"" +
    // node.getQualifiedId() + "\"("+node.getUUId()+
    // ") was not equal to stored one: \"" + uid + "\""));
    // }
    // }
    // }
    return true;
  }
}
