package com.unitesk.requality.core.instancer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.DefaultUUIDGenerator;
import com.unitesk.requality.core.IStorageSource;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;
import com.unitesk.requality.nodetypes.InstancesFolder;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.RestrictedFolder;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.VirtualNode;

public class Instantiator {
  private static final String PARENT_ROOT = "_%parentRoot";
  private static final String INST_NAME = "_%instName";
  public static final String ATTR_TEMP = "_temporary";
  public static String DEFAULT_INSTANCE_NAME = "_default";
  private static Map<String, TreeDB> rootToDBMap = new HashMap<>(); // TODO
  // move
  // to
  // treedb
  // load
  // mech?
  private static Map<String, TreeInstancer> rootToTI = new HashMap<>(); // TODO
  // move
  // to
  // treedb
  // load
  // mech?
  private static Map<TreeDB, TreeDB> iDBToRootDB = new HashMap<>();
  private static Set<String> initiated = new HashSet<>();
  // private static Map<UUID,TreeDB> iRootToInstanceTreeDBs = new
  // HashMap<UUID, TreeDB>();
  private static ReentrantReadWriteLock instLock = new ReentrantReadWriteLock();
  private static ReadLock rl = instLock.readLock();
  private static WriteLock wl = instLock.writeLock();
  private static Map<String, Map<String, String>> rootToInstances =
      new HashMap<>();

  private static void recClear(String rid) {
    writeLock();
    try{
      if (rootToInstances.containsKey(rid)) {
        for (String name : rootToInstances.get(rid).keySet())
          recClear(rootToInstances.get(rid).get(name));
        rootToInstances.remove(rid);
      }
      initiated.remove(rid);
      
      rootToTI.remove(rid);
      //initiated.remove(rid);
      TreeDB db = rootToDBMap.remove(rid);
      if (db != null) iDBToRootDB.remove(db);
  
      for(String s:new HashSet<>(rootToTI.keySet()))
        if(s.startsWith(rid.substring(0,rid.indexOf("|"))))
          recClear(s);//TODO rewrite
    }
    finally{
      writeUnlock();
    }
  }

  public static void clearCache(TreeDB db) {
    try {
      TreeNode node = db.getRootNode();
      if(node == null)
        return;
      UUID root = node.getUUId();
      String rid = getId(root, db);
      recClear(rid);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public static String getId(UUID uid, TreeDB tree) {
    return getId(uid, tree, false);
  }

  public static String getId(UUID uid, TreeDB tree, boolean isroot) {
    if (uid == null) return null;
    if (!isroot) tree = getRootTree(tree);
    return tree.getName() + "|" + uid.toString();
  }

  public static TreeDB getRootTree(TreeDB db) {
    return getRootTree(db, TreeDB.DEFAULT_DEPTH);
  }

  public static TreeDB getRootTree(TreeDB db, int deep) {
    if (db == null || !db.isLoaded() || deep < 1) return db;
    // String dbid = getId(db.getRootNodeUUID(),db);
    if (!iDBToRootDB.containsKey(db)) {
      String qid = getId(db.getRootNodeUUID(), db, true);
      rootToDBMap.put(qid, db);
      return db;
    }
    return iDBToRootDB.get(db);
  // return
    // rootToDBMap.get(getRoo(getId(db.getRootNodeUUID(),db),deep-1));
  }
//
//  private static boolean initIfNeeded(TreeDB db) {
//    return initIfNeeded(db, false);
//  }
//
//  private static boolean initIfNeeded(TreeDB db, boolean isroot) {
//    db = getRootTree(db, TreeDB.DEFAULT_DEPTH);
//    if (!db.isVirtual()
//        && db.isLoaded()
//        && !initiated.contains(getId(db.getRootNodeUUID(), db, isroot))) {
//      writeLock();
//      try {
//        // TreeNode node = db.getNode("/Instances", Path.QID);
//        UUID root = db.getRootNodeUUID();
//        initiated.add(getId(db.getRootNodeUUID(), db, isroot));
//        // rootToInstances.put(root,new HashMap<String, UUID>());
//        UUID node = null;
//        for (UUID uid : db.getStorage().getChildren(root))
//          if (db.getStorage().getNodeId(uid, true).equals("Instances")) {
//            node = uid;
//            break;
//          }
//        if (node != null) {
//          for (UUID chi : db.getStorage().getChildren(node)) {
//            if (chi != null && db.isNodeExist(chi)) {
//              Map<String, AttributeDefinition> attrs = db.getStorage().getAttributes(chi);
//              AttributeDefinition temp = attrs.get(INST_NAME);
//              String name = null;
//              if (temp != null && temp.getRawValue() != null) {
//                Object o = temp.getRawValue().getValue();
//                if (o != null) name = o.toString();
//              }
//              temp = attrs.get(PARENT_ROOT);
//              String droot = null;
//              if (temp != null && temp.getRawValue() != null) {
//                Object o = temp.getRawValue().getValue();
//                if (o != null) droot = o.toString();
//              }
//              if (droot != null) {
//                UUID dr = getUUID(droot);
//                String idr = getId(dr, db, isroot), ichi = getId(chi, db, isroot);
//                if (!rootToInstances.containsKey(idr))
//                  rootToInstances.put(idr, new HashMap<String, String>());
//                rootToInstances.get(idr).put(name, ichi);
//              }
//            }
//          }
//        }
//        return true;
//      } catch (Exception e) {
//        e.printStackTrace();
//      }
//      finally{
//        writeUnlock(); 
//      }
//    }
//    return false;
//  }

  // 1 step upper
  public static TreeDB getBaseTree(TreeDB fakedb) {
    readLock();
    try{
      if (!fakedb.isVirtual()) return fakedb;
      // initIfNeeded(fakedb);
      // UUID uid = fakedb.getRootNodeUUID();
      // String iuid = getId(uid,fakedb);
      TreeDB root = iDBToRootDB.get(fakedb);
      if (root != null) return root;
      return null;
    }
    finally{
      readUnlock();
    }
      // return revInstanceTreeDBs.get(fakedb);
  }

  // private static TreesTracker getTracker(TreeDB db){
  // if(db!=null && !(db.getStorage() instanceof FSStorage)){
  // ITreeStorage ts = VirtualNode.getRootTree(db).getStorage();
  // while(ts instanceof TransactionStorage)
  // ts = ((TransactionStorage)ts).getStorage();
  // if(!(ts instanceof FSStorage))
  // return TreesTracker.getTracker();
  // }
  // return null;
  //
  // }
//  public static TreeDB getVirtualTreeDB(TreeDB db, String name) {
//    return getVirtualTreeDB(db, name, null, null, false);
//  }
//
//  public static TreeDB getVirtualTreeDB(TreeDB db, String name, boolean force) {
//    return getVirtualTreeDB(db, name, null, null, force);
//  }

  private static List<TreeDB> instLoading = new ArrayList<>();

  public static boolean haveVirtualTreeDB(TreeDB db, UUID uuId) {
    readLock();
    try{
      db = getRootTree(db);
      //UUID uroot = db.getRootNodeUUID();
      //if (uroot == null) return false;
      if (uuId == null) return true; // def//getExistVirtualTreeDB(db, DEFAULT_INSTANCE_NAME,
      // null, null)!=null;
      //initIfNeeded(db);
      // Map<String,String> map=rootToInstances.get(getId(uroot,db));
      String id = getId(uuId, db);//, whid = getId(uuId, db);
      TreeDB name = getExistVirtualTreeDBByRoot(db, id);
      if (name != null) {
        return true;
      }
      // TODO check
      return false;
    }
    finally{
      readUnlock();
    }
  }

  private static String getVTreeName(String where, String what) {
    readLock();
    try{
      Map<String, String> map = rootToInstances.get(where);
      if (map != null) {
        for (String s : map.keySet()) if (map.get(s).equals(what)) return s;
        for (String s : map.keySet()) {
          String r = getVTreeName(map.get(s), what);
          if (r != null) return r;
        }
      }
    }
    finally{
      readUnlock();
    }
    return null;
  }
//
//  private static TreeDB getVTreeDB(TreeDB db, UUID what) {
//    if (db == null || db.getStorage() == null) return null;
//    initIfNeeded(db);
//    readLock();
//    try{
//      UUID where = db.getRootNodeUUID();
//      String wid = getId(where, db), whid = getId(what, db);
//      Map<String, String> map = rootToInstances.get(wid);
//      if (map != null) {
//        for (String s : map.keySet())
//          if (map.get(s).equals(whid)){
//            readUnlock();
//            try{
//              return getVirtualTreeDB(db, s, null, null, true);
//            }
//            finally{
//              readLock();
//            }
//          }
//        for (String s : map.keySet()) {
//          readUnlock();
//          try{
//            if (getVirtualTreeDB(db, s, null, null, true) == db) continue;
//            TreeDB r = getVTreeDB(getVirtualTreeDB(db, s, null, null, true), what);
//            if (r != null) return r;
//          }finally{
//            readLock();
//          }
//        }
//      }
//      return null;
//    }
//    finally{
//      readUnlock();
//    }
//  }

  public static TreeDB getExistVirtualTreeDB(TreeDB db, UUID uuId) {
    // if(!db.isLoaded())
    // return null;
    // TODO look
    TreeDB root = getRootTree(db);
    UUID uroot = db.getRootNodeUUID();
    if (uroot == null) return root;
    if (uuId == null || !root.isNodeExist(uuId)) return db;
    // return getVirtualTreeDB(db, DEFAULT_INSTANCE_NAME, null, null);
    // if(canReach(uroot,uuId))
    //initIfNeeded(db);
    String whid = getId(uuId, db);
    TreeDB rch = getExistVirtualTreeDBByRoot(db, whid);
    if (rch != null) return rch;
    return getRootTree(db);
  }
//
//  public static TreeDB getVirtualTreeDB(
//      TreeDB db, String name, final TreeInstancer prevInst, Map<String, Object> initPar) {
//    return getVirtualTreeDB(db, name, prevInst, initPar, false);
//  }

//  public static TreeDB getVirtualTreeDB(
//      TreeDB db,
//      String name,
//      final TreeInstancer prevInst,
//      Map<String, Object> initPar,
//      boolean force) {
//    // TODO avoid transaction?
//  }

  private static ITreeChangeListener delListener(TreeDB db, final TreeDB instTree) {
    return new DefaultTreeChangeListener() {

      @Override
      public void deleted(TreeNode node, boolean internal, Moment m) {
        // delete(node.getUUId(),node.getWorkParentUUID());
        if(m == Moment.BEFORE)
          return;
        if (isVRoot(node)) {
          TreeDB root = Instantiator.getRootTree(node.getTreeDB());
//          boolean ex = haveVirtualTreeDB(instTree, node.getUUId());
//          if (ex)
            removeInstance(instTree, node.getUUId());
            TreeNode ir = root.getNode("/Instances");
            if(ir!=null&&ir.getWorkChildrenUUIDs(true).size()==0)
              ir.delete();
        }
      }
    };
  }

  private static TreeInstancer loadOrCreateTi(
      UUID instR, final TreeDB db, final TreeInstancer prevInst, Map<String, Object> initPar) {
    String insti = getId(instR, db);
    if (rootToTI.containsKey(insti)) return rootToTI.get(insti);
    TreeInstancer ti = null;
    // root storage to create raw root
    TreeDB rootT = getRootTree(db);
    String rt = rootT.getStorage().getNodeId(instR); // root.getNode(instR);
    if (prevInst != null) {
      // is it proper??
      IStorageSource ms =
          new IStorageSource() {

            @Override
            public ITreeStorage getBaseStorage() {
              return prevInst;
              // MemoryStorage ms = new MemoryStorage(finstR);
              // ms.setDeltaTarget(fdb.getNullDeltaListener());
              // return ms;
            }
          };
      // TODO rename?
      ti = new TreeInstancer(db, instR, rt, ms, prevInst);
    } else {
      // may be wrong

      IStorageSource ms =
          new IStorageSource() {
            @Override
            public ITreeStorage getBaseStorage() {
              return db.getStorage();
            }
          };
      ti = new TreeInstancer(db, instR, rt, ms);
    }
    rootToTI.put(insti, ti);
    return ti;
  }
//
//  public static TreeDB getExistVirtualTreeDB(
//      TreeDB db, String name, TreeInstancer prevInst, Map<String, Object> initPar) {
//    // TODO avoid transaction?
//    // if(!db.isLoaded())
//    // return null;
//    if (db.isVirtual() && db.getLevelOfAbstraction() >= 3) return db;
//    UUID uroot = db.getRootNodeUUID();
//    if (uroot == null) return null;
//    initIfNeeded(db);
//    String iuroot = getId(uroot, db);
//    Map<String, String> map = rootToInstances.get(iuroot);
//
//    if (map == null) map = new HashMap<>();
//    // String nme = db.getName();
//    // if(nme == null)
//    String nme = "";
//    String instR = map.get(nme + name);
//    if (instR == null && !instLoading.contains(db)) {
//      return null;
//    }
//    TreeDB instTree = rootToDBMap.get(instR);
//    if (instTree == null && instR != null) instTree = loadDB(db, prevInst, getUUID(instR), initPar);
//    return instTree;
//  }

  private static UUID getUUID(String instR) {
    int pos = instR.lastIndexOf("|");
    if (pos > -1) return UUID.fromString(instR.substring(pos + 1));
    return null;
  }

  public static void loadDBs(
      TreeDB db) {
    // root storage to create raw root
    if(db == null)
      return;
    TreeDB root = getRootTree(db);
    String id = getId(db.getRootNodeUUID(), db);
    if(initiated.contains(id))
      return;
    writeLock();
    try{
      //TreeNode rt = root.getNode("/Instances");
      UUID rid = root.getRootNodeUUID();
      initiated.add(getId(rid, db));
      UUID node = null;
      for (UUID uid : db.getStorage().getChildren(rid))
        if (db.getStorage().getNodeId(uid).equals("Instances")) {
         node = uid;
         break;
        }
      if (node != null) {
        for (UUID chi : db.getStorage().getChildren(node)) {
          if (chi != null && db.isNodeExist(chi)) {
            Map<String, AttributeDefinition> attrs = db.getStorage().getAttributes(chi);
            AttributeDefinition temp = attrs.get(INST_NAME);
            String name = null;
            if (temp != null && temp.getRawValue() != null) {
              Object o = temp.getRawValue().getValue();
              if (o != null) name = o.toString();
            }
            temp = attrs.get(PARENT_ROOT);
            String droot = null;
            if (temp != null && temp.getRawValue() != null) {
              Object o = temp.getRawValue().getValue();
              if (o != null) droot = o.toString();
            }
            if (droot != null) {
              UUID dr = getUUID(droot);
              String idr = getId(dr, db), ichi = getId(chi, db);
              if (!rootToInstances.containsKey(idr))
                rootToInstances.put(idr, new HashMap<String, String>());
              rootToInstances.get(idr).put(name, ichi);
            }
            
          }
        }
      }
      Map<String,String> inst = rootToInstances.get(id);
      if(inst!=null)
        for(String s:inst.keySet())
          recInitInst(db, rootToInstances.get(id).get(s));
      //if (dinst.containsKey(db)) return dinst.get(db);
      TreeNode r = db.getNode(Requirement.getTypeRootQId());
      if (node == null) r = db.getRootNode(); // TODO bug
      // return null;
      if(r == null)
        return;
      UUID instR = r.getUUId();
      TreeInstancer ti = loadOrCreateTi(instR, db, null, null);
      
      if (instR == null) RLogger.logInfo("instr null");
      dinst.put(db, ti);
      //return ti;
    } catch (Exception e) {
      e.printStackTrace();
    }finally{
      writeUnlock();
    }
  }

  private static void recInitInst(TreeDB db, String fid) {
    UUID rid = getUUID(fid);
    TreeInstancer ti = loadOrCreateTi(rid, db, null, null);
    TreeDB instTree = TreeDB.load(ti, new DefaultUUIDGenerator(), db.getLevelOfAbstraction() + 1, db);
    
    Map<String,String> inst = rootToInstances.get(fid);
    if(inst!=null)
      for(String s:inst.keySet())
        recInitInst(instTree, rootToInstances.get(fid).get(s));
  }

  // TODO bottleneck
  public static Map<UUID, Set<VirtualNode>> reuses = new HashMap<>();
  public static Map<VirtualNode, UUID> revreuses = new HashMap<>();

  public static Collection<UUID> getClonesAndVnodes(UUID uuId, TreeDB db) {
    db = Instantiator.getRootTree(db);
    Collection<UUID> cl = getClones(uuId, db);
    if (reuses.containsKey(uuId)) {
      for (VirtualNode vn : reuses.get(uuId)) cl.add(vn.getUUId());
    }
    cl.remove(uuId);
    return cl;
    // was reused, clear vnodes
  }

  public static Collection<UUID> getClones(UUID uuId, TreeDB db2) {
    // if(!db.isVirtual())//TODO look into
    // db = getVirtualTreeDB(db,DEFAULT_INSTANCE_NAME);
    db2 = Instantiator.getRootTree(db2);
    HashSet<UUID> cl = new HashSet<>();
    cl.addAll(TreeInstancer.getClones(db2, db2.getStorage(), uuId, null));
    return cl;
  }

  private static boolean isCloneEquals(TreeDB db, UUID uid, UUID what) {
    Set<VirtualNode> ruses = reuses.get(uid);
    if (ruses != null) for (VirtualNode v : ruses) if (v.getUUId().equals(what)) return true;
    if (uid == null) return false;
    if (uid.equals(what)) return true;
    Object cpv =
        AttributeStorage.getAttributeDefValue(
            db, db.getStorage().getAttributes(uid), uid, TreeNode.ATTR_CLONE_PARENT);
    if (cpv != null) {
      String cp = cpv.toString();
      if (cp.equals(uid.toString())) return false; // selfcloneparent
      return isCloneEquals(db, UUID.fromString(cp), what);
    }
    return false;
  }

  public static TreeNode getChildOrClone(TreeDB db, TreeNode node, UUID fromString) {
    List<UUID> childs = node.getActiveWorkChildrenUUIDs();
    if (childs != null)
      for (UUID uid : childs) if (isCloneEquals(db, uid, fromString)) return db.getNode(uid);
    return null;
  }

  protected static boolean isVRoot(TreeNode node) {
    //Map<String, AttributeDefinition> attrs = node.getAttributes();
    //if (attrs != null) {
      Attribute la = node.attrs().getAttribute(TreeNode.ATTR_VIRT_STATUS);
      if (la != null && la.getValue() != null) {
        Object o = la.getValue();
        if (o != null && o.equals(node.getUUId().toString()))
          return true;
      }
    //}
    return false;
  }

  private static Map<TreeDB, TreeInstancer> dinst = new HashMap<>();

  public static TreeInstancer getVirtualStorage(
      TreeDB pdb, String name){//, Map<String, Object> initpar, final TreeInstancer oldti) {
    TreeDB db = getExistVirtualTreeDBByInstName(pdb, name);
    if (db != null){
      TreeInstancer ti = gatherVirtualStorage(db);
      if(ti == null)
        ti = loadOrCreateTi(db.getRootNodeUUID(), db, null, null);
      return ti;
    }
    return null;
  }

  public static TreeInstancer gatherVirtualStorage(TreeDB rdb) {
    if (rdb != null){
      if(rdb.isVirtual())
        return rdb.getVStorage();
      if(!rdb.isVirtual())
        for(TreeDB db:dinst.keySet())
          if(db!=null && db.getResName()!=null && db.getResName().equals(rdb.getResName()))
            return dinst.get(db);
    }
    return null;
  }

  public static void removeInstance(TreeDB instTree, UUID iroot) {
    writeLock();
    //TODO actual delete?
    
    try{
      TreeDB root = getRootTree(instTree);
      String iiroot = getId(iroot, instTree);
      TreeDB droot = iDBToRootDB.get(instTree);
      instTree.close();
      // TreeDB rootDB = getRootTree(instTree);
      rootToDBMap.remove(iiroot);
      iDBToRootDB.remove(instTree);
      rootToInstances.remove(iiroot);
      // TODO rec
      String torem = null;
      if (droot != null) {
        instTree.getStorage().deleteStorage();
        String drid = getId(droot.getRootNodeUUID(), droot);
        if (rootToInstances.get(drid) != null) {
          for (String s : rootToInstances.get(drid).keySet())
            if (s != null && rootToInstances.get(drid).get(s).equals(iiroot)) {
              torem = s;
  
              break;
            }
          if (torem != null) rootToInstances.get(drid).remove(torem);
        }
      }
    }
    finally{
      writeUnlock();
    }
  }

  private static void writeLock() {
    wl.lock();
  }
  private static void writeUnlock() {
    wl.unlock();
  }


  public static void readLock() {
    rl.lock();
  }
  public static void readUnlock() {
    rl.unlock();
  }
  public static List<TreeNode> getClones(TreeDB db, TreeNode node) {
    readLock();
    try{
      TreeDB vdb = getRootTree(db);
      Collection<UUID> uids = getClones(node.getUUId(), vdb);
      List<TreeNode> nodes = new LinkedList<>();
      for (UUID uid : uids) {
        try {
          TreeNode nde = vdb.getCachedNodeOrVirt(uid);
          if (nde != null) nodes.add(nde);
        } catch (Exception e) {
        }
      }
      return nodes;
    }finally{
      readUnlock();
    }
  }

  public static TreeDB getVirtualTreeDB(TreeInstancer treeInstancer) {
    readLock();
    try{
      for (String s : rootToTI.keySet())
        if (rootToTI.get(s) != null
            && rootToTI.get(s).getRootNode().equals(treeInstancer.getRootNode())) // possible
          // source
          // of
          // misstargeting
          // whene
          // there
          // are
          // clones
          // of
          // projects
          return rootToDBMap.get(s);
      // getId(treeInstancer.getRootNode(), tree)
      // TreeDB db = rootToDBMap.get();
      return null;
    }finally{
      readUnlock();
    }
  }
  public static boolean isVRoot(TreeDB db, UUID uuid, Map<String, Attribute> attrs) {
    if (iDBToRootDB.containsKey(db)) return true;
    Object o = AttributeStorage.getAttributeValue(db, attrs, uuid, TreeNode.ATTR_VIRT_STATUS);
    if (o != null && uuid.equals(o)) return true;
    return false;
  }
  public static boolean isDefVRoot(TreeDB db, UUID uuid, Map<String, AttributeDefinition> attrs) {
    if (iDBToRootDB.containsKey(db)) return true;
    Object o = AttributeStorage.getAttributeDefValue(db, attrs, uuid, TreeNode.ATTR_VIRT_STATUS);
    if (o != null && uuid.equals(o)) return true;
    return false;
  }
  public static Collection<? extends TreeDB> getVirtualTreeDBs(TreeDB treeDB) {
    //initIfNeeded(treeDB);
    readLock();
    try{
      Set<TreeDB> vtrees = new HashSet<>();
      String iroot = getId(treeDB.getRootNodeUUID(), treeDB);
      
      Map<String, String> insts = rootToInstances.get(iroot);
      if (insts != null) {
        for (String uid : insts.values())
          if (rootToDBMap.containsKey(uid)) {
            vtrees.add(rootToDBMap.get(uid));
            vtrees.addAll(getVirtualTreeDBs(rootToDBMap.get(uid)));
          }
      }
      return vtrees;
    }
    finally{
      readUnlock();
    }
  }
  public static TreeDB getExistVirtualTreeDBByRoot(TreeDB db, String name){
    if (db == null) return null;
    if (!db.isLoaded() || db.getStorage() == null) return null;
    readLock();
    try{
      if(rootToDBMap.containsKey(name))
        return rootToDBMap.get(name);
    }finally{
      readUnlock();
    }
    return null;
  }
  public static TreeDB getExistVirtualTreeDBByInstName(TreeDB db, String name){
    if (db == null) return null;
    if (!db.isLoaded() || db.getStorage() == null) return null;
    readLock();
    try{
      String uroot = getId(db.getRootNodeUUID(), db);
      Map<String,String> map = rootToInstances.get(uroot);
      if(name.equals(DEFAULT_INSTANCE_NAME))
        return db;
      if(map == null)
        return null;
      if(map.containsKey(name))
        return rootToDBMap.get(map.get(name));
    }finally{
      readUnlock();
    }
    return null;
  }
//init tree from 0
  public static TreeDB initVirtualTreeDB(
      TreeDB db, String name, Object object, Map<String, Object> initPar,boolean temp) {
    if (db == null) return null;
    if (!db.isLoaded()  || db.getStorage() == null) return null;
    String uroot = getId(db.getRootNodeUUID(), db);
    if (uroot == null) return null;
    if (name == null || name.equals(DEFAULT_INSTANCE_NAME)) return db;
    if (db.getStorage().getName().equals(name)) throw new RuntimeException("Name is already used!");//TODO wtf?
    TreeDB root = getRootTree(db, TreeDB.DEFAULT_DEPTH);
    writeLock();
    String instR = null;
    Map<String,String> map = rootToInstances.get(uroot);
    if(map == null)
      map = new HashMap<String, String>();
     try{
        //boolean ld = false;
        if (!instLoading.contains(db)) {
          instLoading.add(db);
          //initIfNeeded(root);
          TreeNode node = root.getRootNode();
          if (node == null) return null; // loading
          TreeNode
            insts = root.getRootNode().findWorkChild("Instances");
          if(insts == null)
            insts = root.getRootNode().createChild("Instances",InstancesFolder.TYPE_NAME);
          
//          for (UUID uid : root.getStorage().getChildren(root.getStorage().getRootNode()))
//            if (root.getStorage().getNodeId(uid, true).equals("Instances")) {
//              insts = uid;
//              break;
//            }
        //  TreeNode insts = node.findWorkChild("Instances");
//          if (insts == null) insts = node.createChild("Instances", InstancesFolder.TYPE_NAME).getUUId();
//          for (UUID uid : insts.getActiveWorkChildrenUUIDs()) {
//            if (root.isNodeExist(uid)) {
//              Map<String, AttributeDefinition> attrs = root.getStorage().getAttributes(uid);
//              AttributeDefinition temp = attrs.get(INST_NAME);
//              String iname = null;
//              if (temp != null && temp.getRawValue() != null) {
//                Object o = temp.getRawValue().getValue();
//                if (o != null) iname = o.toString();
//              }
//              if (iname != null && iname.equals(name)
//                  || root.getStorage().getNodeId(uid, true).equals(name)) {
//                instR = getId(uid, db);
//                break;
//              }
//            }
//          }
          //if (instR == null) {
          //{  
            Map<String, AttributeDefinition> attrs = new HashMap<>();
            // inst =
            // root.getNode(insts).createChild(nme+name,InstancesFolder.TYPE_NAME);
            UUID instRU = UUID.randomUUID();
            instR = getId(instRU, db);
            // make it silent
            AttributeDefinition l =
                new AttributeDefinition(instRU, TreeNode.ATTR_VIRT_STATUS, AttributeAvailability.LOCAL,
                    new ConstValueProvider(AttributeType.STRING, instRU.toString()),Flag.HIDDEN);
            attrs.put(TreeNode.ATTR_VIRT_STATUS, l);
            l = new AttributeDefinition(instRU, INST_NAME, AttributeAvailability.LOCAL,
                new ConstValueProvider(AttributeType.STRING, name),Flag.HIDDEN);
            attrs.put(INST_NAME, l);
            
            attrs.put(ATTR_TEMP, new AttributeDefinition(instRU, ATTR_TEMP, 
                new ConstValueProvider(AttributeType.BOOL, true),Flag.HIDDEN));
            l = new AttributeDefinition(instRU, PARENT_ROOT, AttributeAvailability.LOCAL,
                new ConstValueProvider(AttributeType.STRING, uroot.toString()),Flag.HIDDEN);
            attrs.put(PARENT_ROOT, l);
            l = new AttributeDefinition(instRU, TreeNode.ATTR_TYPE, AttributeAvailability.LOCAL,
                new ConstValueProvider(AttributeType.STRING, InstancesFolder.TYPE_NAME),Flag.HIDDEN);
            attrs.put(TreeNode.ATTR_TYPE, l);
            if (initPar != null) {
              // attrs = new HashMap<String,Attribute>();
              for (String s : initPar.keySet()) {
                if (!s.startsWith("_")) {
                  AttributeDefinition a =
                      new AttributeDefinition(instRU,s,AttributeAvailability.SUBTREE,new ConstValueProvider(AttributeType.STRING, initPar.get(s).toString()),AttributeDefinition.detectFlags(s));
    //                  new Attribute(
    //                      Version.getUnset(),
    //                      instRU,
    //                      AttributeType.STRING,
    //                      s,
    //                      initPar.get(s).toString());
                 // a.setAvailability(AttributeAvailability.SUBTREE);
                  attrs.put(s, a);
                }
              }
            }
            TreeNode old = insts.findWorkChild(name);
            if(old!=null)
              old.delete();//TODO mb more complex
            if (!rootToDBMap.containsKey(uroot)) rootToDBMap.put(uroot, db);
            map.put(name, instR);
            rootToInstances.put(uroot, map);
            // TreeInstancer ti = loadOrCreateTi(instR, fdb, prevInst, initPar);
            instLoading.remove(db);
            insts.createChild(insts.getUUId(), instRU, name, attrs);
            // root.getDeltaTarget().deleted(uuid, true);
            // root.removeCache(inst.getUUId());
          //}
          // iRootToDRoot.put(instR,uroot);
          //ld = true;
        }
       // if (instLoading.contains(db)) return null;
        TreeDB instTree = rootToDBMap.get(instR);
        //if (ld || instTree == null && instR != null) {
        final UUID finstR = getUUID(instR);
        TreeInstancer ti = loadOrCreateTi(finstR, db, null, initPar);
        instTree = TreeDB.load(ti, new DefaultUUIDGenerator(), db.getLevelOfAbstraction() + 1, db);
        instTree.setVirtualState(true);
        iDBToRootDB.put(instTree, db);
        // root.replaceCache(rt.getUUId(),instTree.getRootNode());
        TreeNode[] reqs = db.getRootNode().getWorkChildren(Requirement.TYPE_NAME, true, true, true);
        if (reqs == null || reqs.length == 0) {
          reqs = db.getRootNode().getWorkChildren(TextNode.TYPE_NAME, true, true, true);
          if (reqs == null || reqs.length == 0)
            reqs = db.getRootNode().getWorkChildren(null, true, true, true); // not
          // recommended
        }
  
        if (instR == null) RLogger.logInfo("instr null");
        else rootToDBMap.put(instR, instTree);
        for (TreeNode rootreq : reqs) {
          if (rootreq.getType().equals(RestrictedFolder.TYPE_NAME)
              || rootreq.getType().equals(InstancesFolder.TYPE_NAME)) continue;
          Collection<UUID> childs = ti.getChildren(finstR);
          UUID iroot = null;
          if (childs != null && childs.iterator().hasNext()) iroot = childs.iterator().next();
          TreeNode nroot = null;
          if (iroot == null) { // rec
            iroot = UUID.randomUUID();
            instTree.setIsLoaded(true);
            iroot = ti.cloneUUID(instTree, iroot, rootreq, finstR, rootreq.getId());
            try {
              nroot = instTree.getNode(iroot, TreeDB.DEFAULT_DEPTH, true);
            } catch (NodeNotFoundException e) {
              e.printStackTrace();
            }
            instTree.updateNodeChildren(nroot);
          } else {
            // nroot = instTree.getNode(iroot,false,false,false);
            try {
              nroot = instTree.getNode(iroot, TreeDB.DEFAULT_DEPTH, false);
            } catch (NodeNotFoundException e) {
              e.printStackTrace();
            }
          }
        }
  
        // instTree.getNode(iroot,false,false,false);
        // if(up){
        // Stack<UUID>stack= new Stack<UUID>();
        // stack.add(iroot);
        // while(!stack.isEmpty())
        // {
        // UUID uid = stack.pop();
        // TreeNode node = instTree.getNode(uid);
        // if(Version.lessThen(nroot.getCPVersion(),rootreq.getVersion()))
        // nroot.updateCPChildren();
        // stack.addAll(node.getRawChildrenUUIDs());
        // RLogger.logInfo(node.getQualifiedId()+"
        // "+node.getUserVisibleName()+" "+Requirement.getFullName(node,
        // instTree));
        // //if(!node.getVersion().equals(node.getCPVersion()))
        // //db.updateNodeChildren(node);
        // }
        // }
        // instTree.addListener(getOldToNewDBListener(db,instTree));
        instTree.addListener(delListener(db, instTree));
        instTree.setIsLoaded();
        //}
     // Свойства тут не переустанавливаются
        // Дополнительно - в req explorer не отображается поддерево требований
        // для отчета, но работает
        if (initPar != null) {
          boolean changes = false;
          Set<String> torem = new HashSet<>();
          for (Entry<String, Attribute> en : instTree.getRootNode().attrs().getLocalAttributes().entrySet()) {
            if (!en.getKey().startsWith("_")) {
              Attribute a = en.getValue();
              if (instTree.getRootNode() != null
                  && a != null
                  && a.getAvailability() == AttributeAvailability.SUBTREE
                  && a.getOrigin().equals(instTree.getRootNodeUUID())) {
                torem.add(en.getKey());
                changes = true;
              }
            }
          }
          for (Entry<String, Object> en : initPar.entrySet())
            if (!en.getKey().startsWith("_")) {
              torem.remove(en.getKey());
              changes = true;
              AttributeDefinition ad = new AttributeDefinition(instTree
                  .getRootNode().getUUId(), en.getKey(), AttributeAvailability.SUBTREE,new TemplValueProvider(AttributeType.STRING,en.getValue().toString()),
                  0);
              instTree.getRootNode().attrs().setAttribute(ad);
            }
          for (String s : torem) instTree.getRootNode().attrs().removeAttribute(s);
          if (changes) instTree.getRootNode().commitAttributes();
        }

        if (instR != null) return instTree;
        return null;
      }
      catch(Exception e){
        RLogger.logError("init instance error: "+CoreUtils.getTrace(e));
      }
      finally{
        writeUnlock();
      }
    return null;
    }
 
    public static void removeTempInstances(TreeDB db) {
      TreeNode node = db.getNode("/Instances");
      if(node!=null)
      {
        for(TreeNode chi:node.getWorkChildren())
        {
          if(chi.attrs().getBooleanAttributeValue(ATTR_TEMP,false)){
            removeInstance(db, chi.getUUId());
            chi.delete();
          }
        }
        if(node.sizeWorkChildren()==0)
          node.delete();
      }
    }
  }
