package com.unitesk.requality.core.operations;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class CommonRemoveLocation {

  private static CommonRemoveLocation clr;

  public static void init(CommonRemoveLocation clre) {
    clr = clre;
  }

  public static CommonRemoveLocation getOne() {
    if (clr == null) init(new CommonRemoveLocation());
    return clr;
  }

  private void doRemoveLocation(Location loc, Requirement req,String locid) {
    req.removeLocation(locid, loc);
    req.commitAttributes();
  }
  private void doRemoveLocation(String loc, Requirement req) {
    req.removeLocationByPath(loc);
    req.commitAttributes();
  }

  protected void doRemoveLocations(Map<Requirement, Set<Location>> locs) {
    for (Requirement req : locs.keySet()) {
      boolean changed = false;
      for (Location loc : locs.get(req)) {
        req.removeLocation(loc.getUUId().toString(), loc);
        changed = true;
      }
      if (changed&&!req.isDisposed()) req.commitAttributes();
    }
  }

  public void removeLocation(
      Location loc, List<Requirement> reqs, /* ReqMarker reqMarker, */ boolean batch) {
    removeLocation(loc, reqs /* ,reqMarker */, batch, null, null);
  }

  // parameterized removing locations:
  // if req != null, then from requirement req, else from everywhere;
  // if reqMarker != null, then from Markup Editor reqMarker, else search any
  // Markup Editor
  // if batch, then asks, else TRANSACTION MUST BE STARTED!
  public void removeFragments(Map<Location, List<Requirement>> locs, boolean b) {
    Map<Document, List<String>> torem = new HashMap<>();
    for(Location loc : locs.keySet()){
      List<Requirement> reqs = locs.get(loc);
      if (reqs == null || loc.getRequirements().length == 0 || Arrays.equals(loc.getRequirements(),reqs.toArray())) {
        Document doc = (Document) loc.getWorkParent();
        //else doc = (Document) db.getNode(locID.substring(0, locID.lastIndexOf("/")), Path.QID);
        if (doc != null) {
          if(!torem.containsKey(doc))
            torem.put(doc, new LinkedList<>());
          torem.get(doc).add(loc.getUUId().toString());
          //String[] rem = new String[] {loc.getUUId().toString(), loc.getRawId()};
          
          // }
        }
      }
    }
    for(Document doc:torem.keySet()){
      try {
        doc.getDOM();
      } catch (Exception e) {
        e.printStackTrace();
      }
      String js = doc.removeLocationRegions(new HashSet<>(torem.get(doc)));
      doc.saveDOM(js, doc.getXmlDoc());
      doc.reloadLocationUUIdToIndex();
    }
  }
  public void removeFragment(
      Location loc,
      List<Requirement> reqs /* , ReqMarker reqMarker */,
      boolean batch,
      String locID /* for deleted */,
      TreeDB db) {
    if (reqs == null || loc.getRequirements().length == 0|| Arrays.equals(loc.getRequirements(),reqs.toArray())) {
      Document doc = null;
      if (loc != null) doc = (Document) loc.getWorkParent();
      else doc = (Document) db.getNode(locID.substring(0, locID.lastIndexOf("/")), Path.QID);
      if (doc != null) {
        try {
          doc.getDOM();
        } catch (Exception e) {
          e.printStackTrace();
        }
        String js = "";
        if (loc == null) js = doc.removeLocationRegion(locID.substring(locID.lastIndexOf("/") + 1));
        else {
          String[] rem = new String[] {loc.getUUId().toString()};
          js = doc.removeLocationRegions(new HashSet<>(Arrays.asList(rem)));
        }
        doc.saveDOM(js, doc.getXmlDoc());
        doc.reloadLocationUUIdToIndex();
        
        // }
      }
    }
  }

  public boolean openConfirm(String title, String message) {
    // to be override
    return true;
  }

  public void openError(String title, String message) {
    // to be override
  }

  public void removeLocation(
      Location loc,
      List<Requirement> reqs,
      /* ReqMarker reqMarker, */ boolean batch,
      String locID /* for deleted */,
      TreeDB db) {
    if (loc == null && locID == null) {
      return;
    }
    if (loc != null) db = loc.getTreeDB();
    if (!batch) {
      if (!openConfirm("Delete Location", "Delete selected Location?")) {
        return;
      }
      if (!db.startTransaction("Delete Location", true)) {
        openError("Delete Location Error", "Some actions are already performing");
        return;
      }
    }
    removeRequirementLocation(loc, reqs, batch, locID);
    if (reqs == null || loc!=null && loc.getRequirements(true,false).length == 0) {
      removeFragment(loc, reqs, /* reqMarker, */ batch, locID, db);
      if (loc != null) loc.delete();
    }
    if (!batch) {
      if (db != null) db.commit();
    }
  }

  public void removeRequirementLocation(
      Location loc, List<Requirement> reqs, boolean batch, String locID) {
    if (reqs != null) {
      for (Requirement r : reqs) doRemoveLocation(loc, r, locID);
    } else {
      if (loc != null) {
        Requirement[] recs = loc.getRequirements();
        for (Requirement _req : recs) doRemoveLocation(loc, _req, locID);
      }
    }
  }
  public void removeRequirementLocations(
      Map<Location, List<Requirement>> map, boolean batch) {
    Map<Requirement, Set<Location>>revm = new HashMap<>();
    for(Entry<Location, List<Requirement>> l:map.entrySet())
    {
      List<Requirement>req = l.getValue();
      if(req == null) req = Arrays.asList(l.getKey().getRequirements());
      for(Requirement r:req){
        if(!revm.containsKey(r))
          revm.put(r, new HashSet<>());
        revm.get(r).add(l.getKey());
      }
      
    }
    doRemoveLocations(revm);
  }
//
//  public void RemoveLocations(
//      Map<Requirement, Set<Location>> locs, /* ReqMarker reqMarker, */ boolean batch) {
//    if (locs == null) {
//      return;
//    }
//    TreeDB db = null;
//    if (!locs.isEmpty()) db = locs.keySet().iterator().next().getTreeDB();
//    if (!batch) {
//      if (!openConfirm("Delete Locations", "Delete selected Locations?")) {
//        return;
//      }
//      if (db!=null && !db.startTransaction("Delete Locations", true)) {
//        openError("Delete Locations Error", "Some actions are already performing");
//        return;
//      }
//    }
//
//    doRemoveLocations(locs);
//    String jsData = null;
//    Map<Document, Set<String>> toDel = new HashMap<>();
//    Map<Location, Set<Requirement>> stdel = new HashMap<>();
//    for (Requirement req : locs.keySet())
//      for (Location loc : locs.get(req)) {
//        if (!stdel.containsKey(loc)) stdel.put(loc, new HashSet<Requirement>());
//        stdel.get(loc).add(req);
//      }
//    for (Location l : stdel.keySet())
//      if (stdel.get(l).iterator().next() == null || l.getRequirements().length == 0) {
//        Document doc = (Document) l.getWorkParent();
//        if (doc != null) {
//          if (!toDel.containsKey(doc)) toDel.put(doc, new HashSet<String>());
//          toDel.get(doc).add(l.getUUId().toString());
//          toDel.get(doc).add(l.getRawId());
//        }
//      }
//
//    for (Document doc : toDel.keySet()) {
//      /*
//       * if (reqMarker == null) { IEditorReference eRef = null; try { eRef
//       * = RequalityPlugin.findMarkEditorForDocument(
//       * doc.getQualifiedId()); } catch (PartInitException e) {
//       * e.printStackTrace(); } if (eRef != null && eRef.getEditor(true)
//       * != null && (eRef.getEditor(true) instanceof ReqMarker)) {
//       * reqMarker = (ReqMarker) eRef.getEditor(true); } } if (reqMarker
//       * != null) { try { Set<String>locidmap = new HashSet<String>();
//       * for(String lc:toDel.get(doc)) locidmap.add(lc); jsData =
//       * reqMarker.getBrowser().deleteLocations(locidmap);
//       * reqMarker.saveDocument(jsData); } catch (CoreException e) {
//       * e.printStackTrace(); } } else
//       */ {
//        try {
//          doc.getDOM();
//        } catch (Exception e) {
//          e.printStackTrace();
//        }
//        String data = doc.removeLocationRegions(toDel.get(doc));
//        doc.saveDOM(data, doc.getXmlDoc());
//      }
//    }
//    for (Location l : stdel.keySet()) {
//      if (stdel.get(l).iterator().next() == null || l.getRequirements().length == 0) l.delete();
//    }
//    if (!batch && db!=null) {
//      db.commit();
//    }
//  }

  public void removeFragments(List<Location> locs, Requirement fnode, boolean b) {
    Map<Location,List<Requirement>> tor = new HashMap<>();
    for(Location l:locs){
      tor.put(l, new LinkedList<>());
      tor.get(l).add(fnode);
    }
    removeFragments(tor, b);
  }

  public void removeLocations(Map<Location, List<Requirement>> map, boolean batch) {
    TreeDB db = null;
    for(Location loc:map.keySet()){
      if (loc == null) 
        continue;
      db = loc.getTreeDB();
    }
    if(db == null&&!batch)
      throw new RuntimeException("Tries to delete empty locs");
    
    if (!batch) {
      if (!openConfirm("Delete Locations", "Delete selected Locations?")) {
        return;
      }
      if (db!=null && !db.startTransaction("Delete Locations", true)) {
        openError("Delete Locations Error", "Some actions are already performing");
        return;
      }
    }
    removeRequirementLocations(map, batch);
    Map<Location,List<Requirement>> tor= new HashMap<>();
    for(Entry<Location, List<Requirement>> l:map.entrySet()){
      if (l == null || l.getKey() == null)
        continue;
      if (l.getValue() == null || l.getKey()!=null && l.getKey().getRequirements(true,false).length == 0) {
        tor.put(l.getKey(), null);
      }
    }
    if (!tor.isEmpty()) {
      removeFragments(tor, batch);
      for(Location loc:tor.keySet())
        if (loc != null) loc.delete();
    }
    if (!batch) {
      if (db != null) db.commit();
    }
  }

}
