/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core.operations;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URI;
import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import org.apache.commons.httpclient.util.URIUtil;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.api.API;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.documents.IProgressMonitor;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.report.ReportGenerator;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

// TODO report undo
public class GenerateReportHandler implements IGenerateReportHandler {
  /*
   * (non-Javadoc)
   *
   * @see
   * com.unitesk.requality.core.operations.IGenerateReportHandler#execute(java
   * .util.List, java.util.Map)
   */
  @Override
  public Object execute(List<TreeNode> nodes, Map<?, ?> par, IProgressMonitor monitor) throws Exception {
    if (nodes == null && par != null) return null;
    if (nodes == null) {
      nodes = new LinkedList<>();
      TreeNode t = (TreeNode) par.get("target");
      if (t != null) nodes.add(t);
    }

    // nodes.get(0).getTreeDB().startTransaction("Generate Reports");
    for (TreeNode node : nodes) {
      ReportSettings report = null;
      if (node instanceof ReportSettings) {
        report = (ReportSettings) node;
        generateReport(report, monitor);
      }
    }
    // nodes.get(0).getTreeDB().commit();
    return null;
  }

  protected void openInformation(String title, String message) {
    RLogger.logInfo("I:" + title + ": " + message);
  }

  protected void openError(String title, String message) {
    System.err.println("E:" + title + ": " + message);
  }

  /*
   * (non-Javadoc)
   *
   * @see com.unitesk.requality.core.operations.IGenerateReportHandler#
   * generateReport(com.unitesk.requality.nodetypes.ReportSettings)
   */
  @Override
  public ReportGenResult generateReport(ReportSettings reportS, IProgressMonitor monitor) {
    TreeDB db = reportS.getTreeDB();
    if(reportS.isGenerating())
      throw new RuntimeException("Some generation is already performing");
    reportS.setGenerating(true);
    reportS.commitAttributes();
    try{
      NodeDesc nd = reportS.toNodeDesc(true, false);
      String date = new Date(System.currentTimeMillis()).toString();
      nd.setAttribute(new AttributeDefinition(nd.getUUId(), 
          TreeNode.ATTR_TYPE, AttributeAvailability.LOCAL,new ConstValueProvider(AttributeType.STRING, ReportData.TYPE_NAME),
          Flag.HIDDEN));
  //    nd.setAttribute(
  //        new Attribute(
  //            Version.getUnset(), nd.getUUId(), AttributeType.STRING, "_type", ReportData.TYPE_NAME));
      URI href = null;
      URI savePath = null;
      TreeInstancer oldti = null, ti = null;
      TreeDB instTree = null, olddb = null;
      ReportData report = null;
      boolean started = false;
      //String st = db.getCacheStorage().startCacheTransaction("generate", CacheStorage.COMMIT_ALL);
      WriteLock wl = null;
      try {
        Template tpl = null;
        monitor.beginTask("init template", 8);
        tpl = getTemplate(reportS.getTemplateId());
        if (tpl == null) {
          if (reportS.getTemplateId() != null) {
            openError(
                "Report generation",
                "Report template '" + reportS.getTemplateId() + "' for this report is not avalible.");
            return null;
          }
          tpl = API.getReportTemplates().get(Template.DEFAULT_ID);
        }
        monitor.worked(8);monitor.done();
        // nd.setId((reportS.getId()+" ["+tpl.getName()+"]
        // ("+date+")").replace(' ', '_').replace(':', '-'));
        monitor.beginTask("init root node", 8);
        SimpleDateFormat dt = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");
        nd.setId((reportS.getId() + "(" + dt.format(new Date(System.currentTimeMillis()))) + ")");
        report = (ReportData) nd.createNode(db, reportS.getWorkParentUUID(), true, false);
        Map<String, Object> objMap = new HashMap<>();
        //TODO mb merge with getData
        for (Entry<String, Attribute> en : report.attrs().getLocalAttributes().entrySet()) { // TODO null pointer
          if (en.getValue().getValue() != null)
            objMap.put(en.getKey(), en.getValue().getValue().toString());
        }
        objMap.put(VirtualNode.ATTR_COLLAPSEVNODE, true);
        olddb = db; // Instantiator.getVirtualTreeDB(db,Instantiator.DEFAULT_INSTANCE_NAME);
        // oldti = olddb.getVStorage();
        // s oldti.setReadOnly(true);
        monitor.worked(8);monitor.done();
        monitor.beginTask("init treedb", 8);
        
        instTree = Instantiator.initVirtualTreeDB(olddb, reportS.getTemplateId(), null, objMap,true);
  //      if (instTree != null) {
  //        UUID root = instTree.getStorage().getRootNode();
  //        instTree.getDeltaTarget().deleted(root, Moment.BEFORE, true);
  //        instTree.getDeltaTarget().deleted(root, Moment.AFTER, true);
  //        TreeDB.clearAllCaches(instTree, root);
  //        Instantiator.removeInstance(instTree, root);
  //      }
  
  //      instTree = Instantiator.getExistVirtualTreeDB(olddb, reportS.getTemplateId());
        ti = instTree.getVStorage();
        // ti.initInitialAttrs(objMap);//reinit attrs for reloaded db
        // ti.setReadOnly(true);
        // instTree = TreeDB.load(ti);
        monitor.worked(8);monitor.done();
        instTree.setVirtualState(true);
        monitor.beginTask("init rev links", 8);
        //instTree.getLinkResolver().initRevLinks(false);
        refinesLinks(instTree, db);//loads and updates links on instance
        //instTree.getLinkResolver().initRevLinks();
        monitor.worked(8);monitor.done();
        // RequalityPlugin
        // .getAdditionalTypes();
        // for (String name : nts.keySet()) {
        // instTree.addNodeType(name, nts.get(name).getClazz());
        // }
        try {
          started = true;
          olddb.startChangeRow();
          monitor.beginTask("get root", 8);
          TreeNode newroot =
              instTree.getNode(
                  report.getStartRequirementQId(),
                  com.unitesk.requality.core.Path.QID | com.unitesk.requality.core.Path.UV);
          monitor.worked(8);monitor.done();
          //instTree.updateRequirementsNames(newroot);
        } catch (RuntimeException re) {
          re.printStackTrace();
          openInformation("Report generator error", "Target Requirement does not exist.");
          return null;
        }
        savePath = report.getNodeResourcesPath();
        // TODO look
        /*
         * instTree.setReadOnlyState(false);
         * hideInstancers(instTree.getNode(node .getStartRequirementQId()));
         * instTree.setReadOnlyState(true);
         */
        Map<TemplateFile, Writer> bufs = new HashMap<>();
        Map<TemplateFile, ByteArrayOutputStream> baos =
            new HashMap<>();
        for (TemplateFile tf : tpl.getTemplates()) {
          baos.put(tf, new java.io.ByteArrayOutputStream());
          bufs.put(tf, new OutputStreamWriter(baos.get(tf), Charset.forName("UTF-8")));
        }
        try{
          Requirement r =
              (Requirement)
                  instTree.getNode(
                      report.getStartRequirementQId(), com.unitesk.requality.core.Path.QID);
          if (r == null)
            for (UUID uid : instTree.getRootNode().getWorkChildrenUUIDs()) {
              TreeNode n = instTree.getNode(uid);
              if (Requirement.TYPE_NAME.equals(n.getType())) r = (Requirement) n;
            }
          if (r == null)
            r =
                (Requirement)
                    instTree.getNode(
                        report.getStartRequirementQId(), com.unitesk.requality.core.Path.QID);
          monitor.beginTask("init main generation", 8);
          ReportGenerator.generateReport(
              r, tpl, bufs, savePath.toString(), report.getData());
          for (TemplateFile tf : tpl.getTemplates()) {
            InputStream is = new ByteArrayInputStream(baos.get(tf).toByteArray());
            report.setResourceContent( tf.getOutputFileName(reportS.getTreeDB()),
                is);
            is.close();
          }
          monitor.worked(8);monitor.done();
          // TODO: index file selection
          href =
              URI.create(
                  savePath.toString()
                      + URIUtil.encodeQuery(
                          '/'+
                              tpl.getTemplates().get(0).getOutputFileName(reportS.getTreeDB())));
    
          File folder =
              new File((savePath).relativize(URI.create(ResManager.getManager().getLocation(db))));
          //
          // IProject project = EclipseTracker.getTracker().getProject(db);
          // IPath path = Path.fromOSString(savePath).makeRelativeTo(
          // project.getLocation());
    
          // IFolder folder = project.getFolder(path);
          // folder.refreshLocal(IResource.DEPTH_INFINITE, null);
          // TODO report undo
          // node.getTreeDB().startTransaction("Generate Report");
          if (report.attrs().getAttributeValue(ReportSettings.ATTR_TEMPLATE) == null)
            report.attrs().setConstAttribute(ReportSettings.ATTR_TEMPLATE, tpl.getId(),Flag.HIDDEN);
          report.attrs().setConstAttribute(ReportSettings.ATTR_DATE, date, 0);
          report.attrs().setConstAttribute(ReportSettings.ATTR_LAST_GENERATED, new File(href).getName(),Flag.HIDDEN);
          report.commitAttributes();
          String s = ti.getErrors();
          if (!s.equals("")) openInformation("Report Generator", "Warnings: " + s);
    
          // node.getTreeDB().commit();
        }
        catch(Exception e){
          e.printStackTrace();
        }
        finally{
//          if(st == null)
//            db.getCacheStorage().commitCache();
        }
        // instTree.stopChangeRow();
      } catch (Exception e) {
        e.printStackTrace();
      } finally {
        monitor.beginTask("clearing", 8);
        if (instTree != null) {
          db.startChangeRow();
          try{
            //instTree.getRootNode().delete();;
            Instantiator.removeInstance(instTree, instTree.getRootNode().getUUId());
            instTree.getRootNode().delete();
          }finally{
            db.stopChangeRow();
          }
        }
        monitor.worked(8);monitor.done();
        monitor.beginTask("send events ", 8);
        if (started&& olddb!=null) olddb.stopChangeRow();
        monitor.worked(8);monitor.done();
      }
      if(report != null)
        openReport(report);
      return new ReportGenResult(href, report, savePath);
    }finally{
      reportS.setGenerating(false);
      reportS.commitAttributes();
    }
  }

  private void refinesLinks(TreeDB instDB, TreeDB db) {
    instDB.getLinkResolver().initCaches(true);
    instDB.getTermsStorage().initCaches(true);
    DeepFirstTreeWalker adftw = new DeepFirstTreeWalker(instDB.getRootNode());
    Iterator<TreeNode>it = adftw.iterator();
    List<AttributeDefinition> ch = new LinkedList<>();
    List<Object> nset = new LinkedList<>();
    Map<String,Map<String,List<String>>> rev = new HashMap<>();
    while(it.hasNext()){
      ch.clear();
      TreeNode itn = it.next();
      for(String s:itn.attrs().getLocalAttributeKeys()){
        if(s.equals(TreeNode.ATTR_CLONE_PARENT))
          continue;//ignore clone-cloneparent relation
        Attribute a = itn.attrs().getAttribute(s);
        if(a!=null && a.isReference()){
          boolean isList = a.getType() == AttributeType.LIST;
          List<TreeNode> links = ReferenceValue.getValues(itn.attrs(), a.getType(), a.getValue(), false);
          if(links!=null){
            nset.clear();
            boolean cha = true;
            for(TreeNode node:links)
            {
              if(node.getTreeDB()==db){//link is on node in base tree
                String qid = node.getQualifiedId();
                if(qid.startsWith(Requirement.getTypeRootQId())){
                  TreeNode node2 = instDB.getNode(qid);
                  if(node2!=null){
                    nset.add(node2.getUUId().toString());//add virtual analog instead
                    cha = true;
                  }
                  else
                    RLogger.logInfo("Node "+node.getQualifiedId()+" was not found for report");
                }
              }
              //if(!cha){
                if(node.getTreeDB()==instDB)
                  nset.add(node.getUUId().toString());
                else
                  nset.add(node.getTreeDB().getResName()+"::"+node.getUUId().toString());
              //}
            }
            if(cha){
              if(isList){
                List<Object> rv = new LinkedList<>();
                for(Object o:nset)
                  rv.add(o.toString());
                ch.add(new AttributeDefinition(itn.getUUId(), s, new ReferenceValueProvider(AttributeType.LIST, new ListValue(AttributeType.REFERENCE, rv))));
              }else
                if(nset.size()>0){
                  ch.add(new AttributeDefinition(itn.getUUId(), s, new ReferenceValueProvider(AttributeType.REFERENCE, nset.get(0).toString())));  
                }
            }
          }
        }
      }
      if(!ch.isEmpty()){
        for(AttributeDefinition ad:ch)
          itn.attrs().setAttribute(ad);
        itn.commitAttributes();
      }
      Map<String, Set<TreeNode>> revs = TreeDB.getCPLinkResolver().getRevertLinks(itn.getCloneParent());
      if(revs!=null && !revs.isEmpty()){
        for(Entry<String, Set<TreeNode>> en:revs.entrySet()){
          String sid = itn.getUUId().toString();
          if(!rev.containsKey(sid))
            rev.put(sid, new HashMap<>());
          List<String> tni = new LinkedList<>();
          for(TreeNode tn2:en.getValue())
            tni.add(tn2.getTreeDB().getResName()+"::"+tn2.getUUId().toString());
          if(!rev.get(sid).containsKey(en.getKey()))
            rev.get(sid).put(en.getKey(), tni);
          else
            rev.get(sid).get(en.getKey()).addAll(tni);
        }
      }
    }
    instDB.getLinkResolver().getStorage().addRevLinks(rev);
  }

  protected void openReport(ReportData res) {
    // OpenReportHandler.openReport(report);
    // TODO some open
  }

  public static Template getTemplate(ReportSettings report) {
    String templateId = report.getTemplateId();
    Template tpl = null;
    if (templateId == null) {
      tpl = API.getReportTemplates().get(Template.DEFAULT_ID);
      report.attrs().setConstAttribute(ReportSettings.ATTR_TEMPLATE, tpl.getId(),
          Flag.HIDDEN);
      report.commitAttributes();
    } else {
      tpl = API.getReportTemplates().get(templateId);
    }
    return tpl;
  }

  public static Template getTemplate(String tid) {
    return API.getReportTemplates().get(tid);
  }

  // remove old image from report before new generation
  @Override
  public void clearOldImages(ReportSettings nde) {
    if (!nde.getType().equals(ReportSettings.TYPE_NAME)) return;
    ReportSettings node = nde;
    Template tpl = null;
    try {
      tpl = getTemplate(node.getTemplateId());
      if (tpl == null) {
        if (node.getTemplateId() != null) {
          return;
        }
        tpl = API.getReportTemplates().values().iterator().next();
      }
      URI savePath = node.getSaveDir();
      for (TemplateFile t : tpl.getTemplates()) {
        File outFile =
            new File(
                new URI(
                    savePath.toString()
                        + URIUtil.encodeQuery(
                            "/" +
                                t.getOutputFileName(node.getTreeDB()))));
        ReportGenerator.clearOldImages(outFile, savePath + "/");
      }
      // TODO: index file selection
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /*
   * (non-Javadoc)
   *
   * @see
   * com.unitesk.requality.core.operations.IGenerateReportHandler#isEnabled(
   * java.util.List<Object>)
   */
  @Override
  public boolean isEnabled(List<TreeNode> nodes) {
    if (nodes == null) return false;
    for (TreeNode node : nodes) if (!node.getType().equals(ReportSettings.TYPE_NAME)) return false;
    return true;
  }

  private static IGenerateReportHandler gen;

  public static void init(IGenerateReportHandler g) {
    gen = g;
  }

  public static IGenerateReportHandler getOne() {
    if (gen == null) init(new GenerateReportHandler());
    return gen;
  }
}
