package com.unitesk.requality.core.operations.mass;

import java.util.HashMap;
import java.util.Map;
import com.unitesk.requality.core.TreeNode;

public abstract class AbstractNodeOperation {
  String defcomment = null;
  private boolean ignoreTrans = false;
  private Map<String, String> settings = new HashMap<>();

  public AbstractNodeOperation(String comment) {
    defcomment = comment;
  }

  public abstract void applySettings(Map<String, String> args);

  public boolean execute(TreeNode root) {
    return execute(root, defcomment);
  }

  public boolean execute(TreeNode root, String comment) {
    if (ignoreTrans
        || root.getTreeDB()
            .startTransaction(
                getName() + " on '" + root.getUserVisibleName(true) + "'" + (comment != null
                    ? " " + comment
                    : ""))) {
      try {
        performAction(root);
        if (!ignoreTrans) root.getTreeDB().commit();
      } catch (Exception e) {
        e.printStackTrace();
        if (!ignoreTrans) root.getTreeDB().rollback();
        return false;
      }
    } else return false;
    return true;
  }

  public abstract String getName();

  public abstract void performAction(TreeNode root);

  public void setIgnoreTransaction(boolean b) {
    ignoreTrans = b;
  }
}
