package com.unitesk.requality.core.operations.mass;

import java.util.HashMap;
import java.util.Map;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;

public class MassAttributeTypeChange extends AbstractMassOperation {
  private String aname = null;
  private String newAName = null;
  private AttributeType atype = AttributeType.STRING;

  public MassAttributeTypeChange() {
    super(null);
  }

  public MassAttributeTypeChange(String attrName, String newAttrName, AttributeType at) {
    this();
    Map<String, String> map = new HashMap<>();
    map.put(MassOppsArguments.Argument.NEW_ATTRIBUTE_NAME.name(), newAttrName);
    map.put(MassOppsArguments.Argument.ATTRIBUTE_NAME.name(), attrName);
    map.put(MassOppsArguments.Argument.ATTRIBUTE_TYPE.name(), at.name());
    applySettings(map);
  }

  public void performSingleAction(TreeNode root) {
    Attribute a = root.attrs().getAttribute(aname);
    if (a != null) {
      AttributeDefinition newA;
      boolean changes = false;
      if (newAName.equals(aname) || newAName == null) newA = null;
      else {
        newA = new AttributeDefinition(a.getOrigin(), newAName, a.getAvailability(),
            a.getDefinition().getValueProvider(),a.getDefinition().getFlags());
//            root.attrs()
//                .createAttribute(
//                    a.getOrigin(),
//                    a.getRawType(),
//                    newAName,
//                    null,
//                    a.getRawValue(),
//                    a.getGenerator());
//        newA.setAvailability(a.getAvailability());
        root.attrs().removeAttribute(aname);
        root.attrs().setAttribute(newA);
        changes = true;
      }
      if (atype != null && (newA==null || a.getType() != atype)) { // &&
        // root.getUUId().equals(a.getOrigin())){
        Object val = null;
        try {
          val =
              Attribute.convertValue(
                  root, a.getType(), atype, a.getValue(), null);
        } catch (Exception e) {
          e.printStackTrace();
        }
        if (val == null) {
          // TODO error
          RLogger.logInfo(
              "Cannot convert '"
                  + aname
                  + "' on '"
                  + root.getUserVisibleName(true)
                  + "' to "
                  + atype);
          return;
        }
        root.attrs().setAttribute(new AttributeDefinition(a.getOrigin(), newAName, a.getAvailability(),AttributeDefinition.detectValueProvider(newAName, a.getType(), val),
            a.getDefinition().getFlags()));
        //root.attrs().setAttribute(a.getOrigin(), atype, newA.getKey(), val);
        changes = true;
      }
      if (changes) root.commitAttributes();
    }
  }

  @Override
  public String getName() {
    return "Change attribute type of " + aname + " to " + atype;
  }

  @Override
  public void applySettings(Map<String, String> args) {
    atype = AttributeType.valueOf(args.get(MassOppsArguments.Argument.ATTRIBUTE_TYPE.toString()));
    aname = args.get(MassOppsArguments.Argument.ATTRIBUTE_NAME.toString());
    newAName = args.get(MassOppsArguments.Argument.NEW_ATTRIBUTE_NAME.toString());
  }
}
