package com.unitesk.requality.core.operations.mass;

import java.util.Map;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.IValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;

public class MassAttributeValueChange extends AbstractMassOperation {
  private String aname = null;
  private AttributeType atype = AttributeType.STRING;
  private Object newVal;

  public MassAttributeValueChange() {
    super("MassAttributeValueChange");
  }

  @Override
  public void performSingleAction(TreeNode root) {
    Attribute a = root.attrs().getAttribute(aname);
    if (a != null && (atype == null || a.getType() == atype)) { // &&
      // root.getUUId().equals(a.getOrigin())){
      //Object val = newVal;
      if (atype == null) atype = a.getType(); // TODO more proper
      IValueProvider vp = a.getDefinition().getValueProvider();
      Class<? extends IValueProvider> c = vp.getClass();
      boolean changed = false;
      if(c.getTypeName().equals(ConstValueProvider.TYPE)){
        vp = new ConstValueProvider(atype, newVal);
        changed = true;
      }
      else
        if(c.getTypeName().equals(TemplValueProvider.TYPE)){
          vp = new TemplValueProvider(atype, newVal);
          changed = true;
        }
      else
        if(c.getTypeName().equals(HtmlValueProvider.TYPE)){
          vp = new HtmlValueProvider(atype, newVal);
          changed = true;
        }
      if(changed){
        root.attrs().setAttribute(new AttributeDefinition(root.getUUId(),aname, a.getAvailability(),vp, a.getDefinition().getFlags()));
        root.commitAttributes();
      }
    }
  }

  @Override
  public String getName() {
    return "Change attribute value of " + aname + "[" + atype + "] to " + newVal;
  }

  @Override
  public void applySettings(Map<String, String> args) {
    String s = args.get(MassOppsArguments.Argument.ATTRIBUTE_TYPE.toString());
    atype = s == null ? null : AttributeType.valueOf(s);
    aname = args.get(MassOppsArguments.Argument.ATTRIBUTE_NAME.toString());
    newVal = args.get(MassOppsArguments.Argument.NEW_VALUE.toString());
  }
}
