package com.unitesk.requality.core.operations.mass;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.regex.Pattern;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class MassIDRenames extends AbstractNodeOperation {

  public MassIDRenames() {
    super("");
  }

  @Override
  public String getName() { // first transaction name
    return "Change numerical ids to temporary";
  }

  Pattern p = Pattern.compile("[0-9]*[_]?[0-9]*");

  @Override
  public void performAction(TreeNode root) {
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(root, true, true, true);
    Map<UUID, String> newIds = new HashMap<>();
    for (TreeNode node : dftw) {
      TreeNode[] sch = node.getWorkChildren();
      Set<String> ids = new HashSet<>();
      for (int i = 0; i < sch.length; i++) {
        String id = sch[i].getId();
        if (p.matcher(id).matches()) {
          String tempid = sch[i].getId() + "_temp";
          int cnt = 0;
          while (node.findWorkChild(tempid + cnt) != null) cnt++;
          newIds.put(sch[i].getUUId(), RequalityCLI.getNextId(ids));
          ids.add(newIds.get(sch[i].getUUId()));
          sch[i].setId(tempid + cnt);
          sch[i].commitAttributes();
        } // cannot be non-num id breaking all order?
      }
    }
    root.getTreeDB().commit();
    root.getTreeDB().startTransaction("Change numerical ids to new generated"); // second
    // transaction
    // name
    for (TreeNode node : dftw) {
      if (newIds.containsKey(node.getUUId())){
        node.setId(newIds.get(node.getUUId()));
        node.commitAttributes();
      }
    }
  }

  @Override
  public void applySettings(Map<String, String> args) {
    // TODO Auto-generated method stub

  }
}
