package com.unitesk.requality.core.operations.mass;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.search.SearchMethod;
import com.unitesk.requality.core.search.SearchRequest;

public class NodeSearch extends AbstractMassOperation {
  private SearchRequest sr;
  public NodeSearch() {
    this(null);
  }
  public NodeSearch(SearchRequest sr) {
    super("SearchNode");
    this.sr =sr;
  }

  @Override
  public String getName() {
    return "Search Node";
  }

  @Override
  public boolean execute(TreeNode root) {
     matches.clear();
     if(sr == null)
       return false;
     if(sr.getMethod()==SearchMethod.BY_PATH){
       List<TreeNode> nds = Collections.emptyList();
       try {
         nds=root.getTreeDB().getLinkResolver().resolvePath(root, null, 
         root.getTreeDB(),sr.get("path"), Path.QID|Path.UUID|Path.UV|Path.XPATH);
       } catch (NodeNotFoundException e){}
       matches.addAll(nds);
//       if(nds!=null)
//         for(TreeNode nd:nds)
//           if(nd != null && nd.getUUId().equals(root.getUUId()))
//             return true;
       return true;
     }
     else
       return super.execute(root);
  }

  Set<TreeNode> matches = new HashSet<>();

  @Override
  public void performSingleAction(TreeNode node) {
    if(sr.test(node))
      matches.add(node);
  }

  @Override
  public void applySettings(Map<String, String> args) {
    SearchRequest sr = SearchRequest.restore(args);
    if(sr != null)
      this.sr = sr;
  }
  public Collection<? extends TreeNode> getMatches() {
    return matches;
  }
}
