package com.unitesk.requality.core.propagators;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;

public class ParentAttrsPropagator implements IAttributeChangePropagator {

  @Override
  public boolean shouldBeApplied(
      TreeNode node, ChangesSet<Attribute> cs,  Moment m) {
    if(m==Moment.AFTER)
      return false;
    for(String s:cs.getOrder()){
      Attribute olda=cs.getOld(s);
      Attribute newa=cs.getNew(s);
      if(olda!=null&&olda.getAvailability()!=AttributeAvailability.LOCAL||
          newa!=null && newa.getAvailability()!=AttributeAvailability.LOCAL)
        return true;
    }
    return false;
  }

  @Override
  public boolean apply(TreeNode node, ChangesSet<Attribute> changes) {
    //Map<String,AttributeDefinition> oldMap = old;
    //getDeltaTarget().changeAttributes(uuid, what, ad);
    LinkedList<AttributeDefinition> poolRemove=new LinkedList<>(),poolAdd=new LinkedList<>();
    //LinkedList<String> order = new LinkedList<>();
    //ChangesSet chVis = new ChangesSet();
    //TreeDB.addAttrsRec(node, changes, chVis, order);
    for(String s:changes.getOrder()){
      Attribute olda = changes.getOld(s);
      Attribute newa = changes.getNew(s);
      if(olda!=null && olda.getAvailability()!=AttributeAvailability.LOCAL)
        poolRemove.add(olda.getDefinition());
      if(newa!=null && newa.getAvailability()!=AttributeAvailability.LOCAL){
        poolAdd.add(newa.getDefinition());
      }
    }
    LinkedList<TreeNode> toh = new LinkedList<>();
    for(TreeNode s:node.getActiveWorkChildren())
      toh.add(s);
    int i = 0,j=toh.size();
    Map<String,Attribute> torem = new HashMap<>();
    while(!toh.isEmpty()){
      torem.clear();
      TreeNode n = toh.pollFirst();
      i++;
      Iterator<AttributeDefinition> itr = poolRemove.iterator();
      boolean changed = false;
      while(itr.hasNext())
      {
        AttributeDefinition ada = itr.next();
        if(ada.getAvailability()==AttributeAvailability.DIRECT_CHILDREN&&!ada.isAvaiableOn(n))//TODO test
          itr.remove();
        else{
          Attribute aa = n.attrs().getAttribute(ada.getKey());
          if(aa==null || aa.getDefinition().getOrigin().equals(ada.getOrigin())){
            torem.put(ada.getKey(), aa);
            changed = true;
          }
        }
      }
      itr = poolAdd.iterator();
      while(itr.hasNext())
      {
        AttributeDefinition ada = itr.next();
        if(ada.getAvailability()==AttributeAvailability.DIRECT_CHILDREN&&!ada.isAvaiableOn(n))//TODO test
          itr.remove();
        else{
          Attribute olda = torem.remove(ada.getKey());
          Attribute b = n.attrs().getAttribute(ada.getKey());
          if(olda == null && b!=null&&b.getDefinition().isMoreVisible(node.getTreeDB(), ada, n.getUUId(), false)){
            if(ada.getAvailability()==AttributeAvailability.SUBTREE)
              changed = true;//continue to propogate...
          }
          else{
            n.attrs().setAttribute(ada);
            changed = true;
          }
        }
      }
      for(String s:torem.keySet())
        n.attrs().removeAttribute(s);
      if(changed)
        n.commitAttributes();
    }
    return true;
  }
  static IAttributeChangePropagator one = new ParentAttrsPropagator();
  public static IAttributeChangePropagator getOne() {
    return one;
  }

  @Override
  public int getPriority() {
    return 40;
  }
}
