package com.unitesk.requality.core.selection;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.ISelectionProvider;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.SharedConsts;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;

public class SelectionManager {
  private static long SEL_TL = 200;//200ms tl for selection changes
  private Map<String, Map<String, List<TreeNode>>> activeNodes =
      new HashMap<>();
  private static SelectionManager sm;
  private String lastES = EventSource.GENERAL;
  private Set<ISelectionProvider> prov = new HashSet<>();

  private SelectionManager() {}

  public void registerProvider(ISelectionProvider p) {
    if (p != null) prov.add(p);
  }

  public void unregisterProvider(ISelectionProvider p) {
    prov.remove(p);
  }
  // UI state change listener
  private Map<String, List<IUIStateChangeListener>> listeners =
      new HashMap<>();
  private INodeChangeListener nodeRemListener =
      new INodeChangeListener() {

        @Override
        public void deleted(TreeNode source) {
          if (source == null) return;
          for (String es : EventSource.values()) {
            //Map<String, List<TreeNode>> map = getActiveNodes(es, false);
            Map<String, List<TreeNode>> map = new HashMap<>();
            if (map.containsKey(source.getType())){
              Iterator<TreeNode> it = map.get(source.getType()).iterator();
              boolean rem = false;
              while(it.hasNext())
              {
                TreeNode node = it.next();
                if((source.getQualifiedId(true)+"/").contains(node.getQualifiedId(true)+"/")){
                  rem = true;
                  it.remove();
                }
              }
              if(rem){
                TreeNode node = null;
                try {
                  node = source.getTreeDB().getNode(source.getWorkParentUUID());
                } catch (Exception e) {
                  //TODO node not found
                }
                if (node == null || node.isDisposed()) continue;
                if (!map.containsKey(node.getType()))
                  map.put(node.getType(), new ArrayList<TreeNode>());
                map.get(node.getType()).add(node);
                setActiveNodes(map, map, es);
              }
            }
          }
        }

        @Override
        public void attributeChange(TreeNode source, String what, Attribute old) {
          // TODO Auto-generated method stub

        }

        @Override
        public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          // TODO Auto-generated method stub

        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };

  public void setSource(String src) {
    if(ignoreSet==-1 || ignoreSet<System.currentTimeMillis()){
      ignoreSet = -1;
//      if(!lastES.equals(src) || !getActiveNodes(EventSource.GENERAL).equals(getActiveNodes(src))){
      setActiveNodes(getActiveNodes(src),getActiveNodes(src), src);
    }
  }

  public static SelectionManager getOne() {
    if (sm == null) sm = new SelectionManager();
    return sm;
  }

  public void setActiveNode(TreeNode node, String type, String es) {
    LinkedList<TreeNode> nodes = new LinkedList<>();
    if (node != null) nodes.add(node);
    setActiveNodes(nodes,nodes, es);
  }

  public static Map<String, List<TreeNode>> toMap(List<TreeNode> lnodes) {
    Map<String, List<TreeNode>> nodes = new TreeMap<>();
    if (lnodes != null)
      for (TreeNode node : lnodes) {
        if (!nodes.containsKey(node.getType()))
          nodes.put(node.getType(), new LinkedList<TreeNode>());
        nodes.get(node.getType()).add(node);
      }
    return nodes;
  }

  public void setActiveNodes(List<TreeNode> lnodes,List<TreeNode> expnodes, String es) {
    setActiveNodes(toMap(lnodes),toMap(expnodes), es);
  }

//  public void setActiveNodes(Map<String, List<TreeNode>> nodes,Map<String, List<TreeNode>> expnodes, String es) {
//    setActiveNodes(nodes,expnodes, es, true);
//  }
//  public void setActiveNodes(
//      Map<String, List<TreeNode>> nodes,
//      Map<String, List<TreeNode>> expnodes,String es, boolean first) {
//    setActiveNodes(nodes, expnodes, es, new HashSet<>() ,first);
//  }
  private long ignoreSet = -1;
  public void sourcesSetNodeAction(Map<String, List<TreeNode>> nodes,
      Map<String, List<TreeNode>> expnodes,String es){
    boolean ch = checkChanged(nodes, es);
    if(ch){
      Map<String, List<TreeNode>>an = getActiveNodes(es, true);
      for (String s : an.keySet())
        for (TreeNode node : an.get(s))
          if (!node.isDisposed()) node.removeGuiChangeListener(nodeRemListener);
      an = new HashMap<>();
      List<TreeNode> allNodes = new ArrayList<>();
      for (String s : nodes.keySet()) {
        an.put(s, nodes.get(s));
        allNodes.addAll(nodes.get(s));
        for (TreeNode node : nodes.get(s)) node.addGuiChangeListener(nodeRemListener);
      }
      activeNodes.put(es, an);
      List<IUIStateChangeListener> list = listeners.get(es);
      if (list != null)
        for (IUIStateChangeListener listener : list) {
          listener.newActiveNodes(allNodes);
        }
    }
  }
  private boolean checkChanged(Map<String, List<TreeNode>> nodes, String es) {
    Map<String, List<TreeNode>> an = getActiveNodes(es, true);
    Set<String> keys = new HashSet<>(nodes.keySet());
    keys.addAll(an.keySet());
    boolean eq = true;
    for(String s:keys){
      List<TreeNode> o = an.get(s), n = nodes.get(s);
      if(o==null||!o.equals(n)){
        eq = false;
        break;
      }
    }
    return !eq;
  }

  public void setActiveNodes(
      Map<String, List<TreeNode>> nodes,
      Map<String, List<TreeNode>> expnodes,String es) {
    long lig = ignoreSet;
//    if(lig!=-1&&System.currentTimeMillis()<=lig){
//      RLogger.logDebug("IGNORE: "+CoreUtils.getTrace(Thread.currentThread().getStackTrace()), 2); 
//      return;
//    }
    //if(ignoreSet == -1)
    ignoreSet = System.currentTimeMillis()+SEL_TL;
    RLogger.logDebug("Set source(active: "+lastES+") and nodes: "+es+" "+Arrays.toString(nodes.keySet().toArray()), 2); 
    if(checkChanged(nodes, es)||!es.equals(lastES)){
        RLogger.logDebug( "\tsucceed",2); 
        lastES = es;
        if (!es.equals(EventSource.EXPLORER))
          if ((!es.equals(EventSource.EDITOR) || SharedConsts.isNavigatorLinkingEnabled))
            sourcesSetNodeAction(expnodes, expnodes, EventSource.EXPLORER);
        if (!es.equals(EventSource.EDITOR))
          if ((!es.equals(EventSource.EXPLORER) || SharedConsts.isNavigatorLinkingEnabled))
            sourcesSetNodeAction(expnodes, expnodes, EventSource.EDITOR);
        if (!es.equals(EventSource.LCNF)) sourcesSetNodeAction(expnodes, expnodes, EventSource.LCNF);
        if (!es.equals(EventSource.PROPERTIES)) sourcesSetNodeAction(expnodes, expnodes, EventSource.PROPERTIES);
        sourcesSetNodeAction(nodes,expnodes, es);
        if (!es.equals(EventSource.GENERAL)) sourcesSetNodeAction(nodes,expnodes, EventSource.GENERAL);
    //  if(es.equals(lastES)||es.equals(EventSource.GENERAL)){
        
      //}
    }
  }

  public synchronized List<TreeNode> getSelectedNodes(String es, boolean inner) {
    List<TreeNode> sel = new ArrayList<>();
    Map<String, List<TreeNode>> list = getActiveNodes(es, inner);
    if (list != null) for (String s : list.keySet()) sel.addAll(list.get(s));
    return sel;
  }

  public synchronized List<TreeNode> getSelectedNodes(String es) {
    return getSelectedNodes(es, false);
  }

  public synchronized List<TreeNode> getSelectedNodes(boolean inner) {
    return getSelectedNodes(lastES, inner);
  }

  public synchronized List<TreeNode> getSelectedNodes() {
    return getSelectedNodes(false);
  }

  public synchronized Map<String, List<TreeNode>> getActiveNodes(String es) {
    return getActiveNodes(es, false);
  }

  public synchronized Map<String, List<TreeNode>> getActiveNodes(String es, boolean inner) {
//    if (!inner)
//      for (ISelectionProvider p : prov) {
//        String cat = p.getGroupId();
//        Map<String, List<TreeNode>> sel = p.getSelection();
//        if (!equals(sel, activeNodes.get(es))) {
//          setActiveNodes(sel, sel, cat);
//        }
//      }
    Map<String, List<TreeNode>> list = activeNodes.get(es);
    if (list != null) return list;
    return new HashMap<>();
  }

  private boolean equals(Map<String, List<TreeNode>> sel, Map<String, List<TreeNode>> map) {
    if (map == null) return false;
    Set<String> set = new HashSet<>(sel.keySet());
    set.addAll(map.keySet());
    for (String s : set) {
      List<TreeNode> l1 = sel.get(s), l2 = map.get(s);
      if (l1 == null || l2 == null) return false;
      if (l1.size() != l2.size()) return false;
      int cnt = 0;
      for (TreeNode node : l1) {
        if (l2.contains(node)) cnt++;
      }
      return cnt == l1.size();
    }
    return false;
  }

  public synchronized List<TreeNode> getActiveNodes(String string, String es, boolean inner) {
    Map<String, List<TreeNode>> list = getActiveNodes(es, inner);
    if (list == null) return null;
    return list.get(string);
  }

  public synchronized boolean addChangeListener(String es, IUIStateChangeListener listener) {
    if (!listeners.containsKey(es)) listeners.put(es, new ArrayList<IUIStateChangeListener>());
    return listeners.get(es).add(listener);
  }

  public synchronized boolean removeChangeListener(String es, IUIStateChangeListener listener) {
    List<IUIStateChangeListener> list = listeners.get(es);
    if (list != null) return list.remove(listener);
    return false;
  }
}
