package com.unitesk.requality.core.transaction;

import java.util.Arrays;
import java.util.Map;
import java.util.UUID;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;

public class SaveAttributes extends TreeOperation {
  private UUID uuid;
  //private Map<String, AttributeDefinition> oldAttributes;
  //private Map<String, AttributeDefinition> newAttributes;
  private ChangesSet<Attribute> changed;

  public SaveAttributes(
      UUID uuid,
      ChangesSet<Attribute> altered) {
    this.uuid = uuid;
    this.changed = new ChangesSet<>(altered);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("Save attributes ");
    sb.append(uuid);
    sb.append(" ");
    sb.append(Arrays.toString(changed.getAll().toArray()));
    for(String s:changed.getAll()){
      Attribute oa = changed.getOld(s);
      Attribute na = changed.getNew(s);
      if(oa!=null)
        sb.append(oa.toString());
      else
        sb.append("null");
      sb.append("->");
      if(na!=null)
        sb.append(na.toString());
      else
        sb.append("null");
      sb.append(";");
    }
    return sb.toString();
    // return super.toString();
  }

  @Override
  public void apply(ITreeStorage storage) {
    if(storage.getDeltaTarget()!=null && 
        !storage.getDeltaTarget().changeAttributes(uuid, changed))
      storage.saveAttributes(uuid, changed);
  }

  @Override
  public void rollback(ITreeStorage storage) {
    Map<String,AttributeDefinition> old = AttributeStorage.toDefMap(changed.getOldMap());
    ChangesSet<Attribute> cs = changed.getReverted();
    if(storage.getDeltaTarget()!=null && 
        !storage.getDeltaTarget().changeAttributes(uuid, cs))
      storage.saveAttributes(uuid, cs);
  }
}
