package com.unitesk.requality.core.transaction;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import org.apache.commons.httpclient.URIException;
import org.apache.commons.httpclient.util.URIUtil;
import com.unitesk.requality.core.AbstractTreeStorage;
import com.unitesk.requality.core.CacheStorage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.Version;
import com.unitesk.requality.core.instancer.TreeInstancer;

public class TransactionStorage extends AbstractTreeStorage {
  public class OperationData {
    private TreeOperation operation;
    private Object data;

    public OperationData(TreeOperation operation, Object data) {
      this.operation = operation;
      this.data = data;
    }

    public TreeOperation getOperation() {
      return operation;
    }

    public Object getData() {
      return data;
    }
  }

  @Override
  public Map<String, Version> getVersions(UUID uuid) {
    Map<String, Version> base = new HashMap<>(storage.getVersions(uuid)),
        vers = super.getVersions(uuid);
    if (!vers.containsKey(uuid)) return base;
    return vers;
  }

  @Override
  public void saveVersions(UUID uuid, ChangesSet<Version> changes) {
    Map<String, Version> // oldAttributes = new
        // HashMap<String,Version>(getVersions(uuid)),
        newA = new HashMap<>(changes.getNewMap());
    for (String s : newA.keySet()) newA.put(s, newA.get(s).clone());
    Map<String, Version> old = new HashMap<>(getVersions(uuid));
    for (String s : old.keySet()) old.put(s, old.get(s).clone());

    getDeltaTarget().versionChanged(uuid, changes, Moment.BEFORE, true);
    super.saveVersions(uuid,changes);
    SaveVersions saveOperation = new SaveVersions(uuid, changes, old, newA);
    operations.add(saveOperation);
    // is it needed?
    getDeltaTarget().versionChanged(uuid, changes, Moment.AFTER, true);
    // super.saveVersions(uuid, attributes);
  }

  private static final String TMP_RESOURCE_PREFIX = "requality";
  private ITreeStorage storage;
  private TreeInstancer vstorage;
  private Map<UUID, String> newId = new HashMap<>();
  private Map<UUID, String> oldId = new HashMap<>();
  private Map<UUID, Collection<UUID>> newChildren = new ConcurrentHashMap<>();
  private Map<UUID, Collection<UUID>> oldChildren = new ConcurrentHashMap<>();
  private Map<UUID, Map<String, AttributeDefinition>> newAttr =
      new ConcurrentHashMap<>();
  private Map<UUID, Map<String, AttributeDefinition>> oldAttr =
      new ConcurrentHashMap<>();
  private Map<UUID, HashMap<String, String>> resourсes =
      new LinkedHashMap<>();

  private ConcurrentLinkedQueue<TreeOperation> operations =
      new ConcurrentLinkedQueue<>();
  private LinkedList<OperationData> operationsData = new LinkedList<>();

  private List<ITransactionListener> tListeners;

  private String transactionName = "complex operation";
  private boolean execute = true;
  private boolean verbose = false;

  public TransactionStorage(
      ITreeStorage storage, TreeInstancer vstorage, List<ITransactionListener> tListeners) {
    this.storage = storage;
    this.vstorage = vstorage;
    this.tListeners = tListeners;
    cs = storage.getCacheStorage();
  }

  public String getName() {
    return transactionName;
  }

  public void setName(String name) {
    transactionName = name;
  }

  @Override
  public UUID getRootNode() {
    return storage.getRootNode();
  }

  @Override
  public UUID getParent(UUID uuid) {
    for (UUID pUUId : newChildren.keySet()) {
      if (pUUId != null
          && newChildren.get(pUUId) != null
          && newChildren.get(pUUId).contains(uuid)) {
        return pUUId;
      }
    }
    UUID oldParentUUID = getOldParent(uuid);
    if (oldParentUUID != null) { // if node was deleted
      return oldParentUUID;
    }
    UUID pid = storage.getParent(uuid);
    if (pid != null) return pid;
    // if(vstorage == null)
    // return null;
    return null;
    // return vstorage.getStoredParent(uuid);
  }

  @Override
  public String getNodeId(UUID uuid) {
    String id = null;
    if (newId.containsKey(uuid)) {
      id = newId.get(uuid);
    }
    if (id == null) id = storage.getNodeId(uuid);
    // if(id == null && vstorage!=null)
    // id = vstorage.getNodeId(uuid);
    return id;
  }

  @Override
  public Map<String, AttributeDefinition> getStorageAttributes(UUID uuid) {
    return currentAttrs(uuid);
  }

  @Override
  public Collection<UUID> getChildren(UUID uuid) {
    return currentChildren(uuid);
  }

  @Override
  public boolean canLoad(UUID uuid) {
    if (uuid.equals(getRootNode())) return true;
    if (getParent(uuid) == null) {
      return false;
    }
    return currentChildren(getParent(uuid)).contains(uuid);
  }

  @Override
  public URI getNodeResourcePath(UUID uuId){
    URI path = null;
    URI s = storage.getNodeResourcePath(uuId);
    if (s != null) path = s;
    if (path == null) {
      UUID par = getParent(uuId);
      if (par == null) return null;
      s = getNodeResourcePath(par);
      URI ppath = null;
      if (s != null) ppath = new File(s).getParentFile().toURI();
      if (ppath != null) {
        try {
          return new URI(
              ppath
                  + URIUtil.encodeQuery(
                      "/"
                          + getNodeId(getParent(uuId))
                          + "/"
                          + getNodeId(uuId)
                          + "_resources"));
        } catch (URIException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        } catch (URISyntaxException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
    return path;
  }

  @Override
  public Set<String> listResources(UUID uuid) {
    return new HashSet<>(currentResources(uuid).keySet());
  }

  public ITreeStorage getStorage() {
    return storage;
  }

  private UUID getOldParent(UUID uuid) {
    for (UUID pUUId : oldChildren.keySet()) {
      if (oldChildren.get(pUUId).contains(uuid)) {
        return pUUId;
      }
    }
    return null;
  }

  @Override
  public void createStorageNode(UUID parentUUId, UUID uuid, String id, Map<String, AttributeDefinition> attrs) {
    if (getDeltaTarget() != null) getDeltaTarget().beforeCreated(uuid, parentUUId, attrs, true);

    newId.put(uuid, id);
    oldId.put(uuid, null);
    currentChildren(parentUUId).add(uuid);

    newAttr.put(uuid, new HashMap<>(attrs));
    oldAttr.put(uuid, new HashMap<String, AttributeDefinition>());

    CreateNode createOp =
        new CreateNode(parentUUId, uuid, id, new HashMap<>(attrs));
    operations.add(createOp);

    if (getDeltaTarget() != null) getDeltaTarget().created(uuid, true);
  }

  @Override
  public void remove(UUID uuid) {
    if (getDeltaTarget() != null) getDeltaTarget().deleted(uuid, Moment.BEFORE, true);
    UUID pUUID = getParent(uuid);
    String id = getNodeId(uuid);
    Map<String, AttributeDefinition> attrs = getAttributes(uuid);

    newId.remove(uuid);
    oldId.put(uuid, id);
    if (pUUID != null) currentChildren(pUUID).remove(uuid);

    // TODO: transaction remove resources
    Map<String, String> resources = currentResources(uuid);
    for (String key : resources.keySet()) {
      deleteResource(pUUID, key);
    }
    DeleteNode deleteOperation = new DeleteNode(pUUID, uuid, id, new HashMap<>(attrs));
    operations.add(deleteOperation);
    if (getDeltaTarget() != null) getDeltaTarget().deleted(uuid, Moment.AFTER, true);
  }

  @Override
  public void changeNodeId(UUID uuid, String newNodeId) {
    String oldNodeId = getNodeId(uuid);
//    if (oldId != null && !oldId.equals(newId)) {
//      //TODO add id change instead?
////      getDeltaTarget()
////          .changed(
////              uuid,
////              AttributeStorage.formChangesSet(uuid, TreeNode.ATTR_ID, oldNodeId),
////              AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, newNodeId)),
////              Moment.BEFORE,
////              true);
//      getDeltaTarget().moved(uuid, getParent(uuid), getParent(uuid), Moment.BEFORE, true);
//    }
    oldId.put(uuid, oldNodeId);
    newId.put(uuid, newNodeId);
    RenameNode moveOperation = new RenameNode(uuid, oldNodeId, newNodeId);
    operations.add(moveOperation);
    // TODO is it needed?
//    if (oldId != null && !oldId.equals(newId)) {
////      getDeltaTarget()
////          .changed(
////              uuid,
////              AttributeStorage.formChangesSet(uuid, TreeNode.ATTR_ID, oldNodeId),
////              AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, newNodeId)),
////              Moment.AFTER,
////              true);
//      getDeltaTarget().moved(uuid, getParent(uuid), getParent(uuid), Moment.AFTER, true);
//    }
  }

  @Override
  public void saveStorageAttributes(UUID uuid, ChangesSet<Attribute> changes) {
//    Map<String, AttributeDefinition> 
//          oldAttributes = new HashMap<>();
    //Map<String, AttributeDefinition> newA = new HashMap<>(newAttributes);
//    getDeltaTarget().changed(uuid, changes, newA, Moment.BEFORE, true);
    currentAttrs(uuid);
    Map<String, AttributeDefinition> ad = newAttr.get(uuid);
    if(ad!=null){
      Map<String, Attribute> newa = changes.getNewMap();
      for(Entry<String, Attribute> e:newa.entrySet())
      {
        if(e.getValue() == null)
          ad.remove(e.getKey());
        else
          ad.put(e.getKey(), e.getValue().getDefinition());
      }
    }
    SaveAttributes saveOperation =
        new SaveAttributes(
            uuid,
            changes);
    operations.add(saveOperation);
    // is it needed?

    //Map<String,Version> nva = new HashMap<String,Version>(),
      //                  va = new HashMap<String,Version>();
//    for(String s:changes.getAll()){
//      nva.put(s, newA.get(s)!=null?newA.get(s).getVersion():Version.getMinusOne());
//      va.put(s, oldAttributes.get(s)!=null?oldAttributes.get(s).getVersion():Version.getMinusOne());
//    }
//    saveVersions(uuid, changes, nva);
//    getDeltaTarget().changed(uuid, changes, newA, Moment.AFTER, true);
  }

  @Override
  public void moveNodeTo(UUID uuid, UUID newParent) {
    UUID oldParent = getParent(uuid);
    getDeltaTarget().moved(uuid, oldParent, newParent, Moment.BEFORE, true);
    if (oldParent != null) currentChildren(oldParent).remove(uuid);
    currentChildren(newParent).add(uuid);
    MoveNode moveOperation = new MoveNode(uuid, oldParent, newParent);
    operations.add(moveOperation);
    getDeltaTarget().moved(uuid, oldParent, newParent, Moment.AFTER, true);
  }

  private Collection<UUID> currentChildren(UUID uuid) {

    // TODO
    // Storage == null when mocking is active
    if (uuid == null && storage.getName() != null) {
      throw new IllegalArgumentException("uuid is null");
    }
    if (!newChildren.containsKey(uuid)) {
      TreeSet<UUID> initalData = new TreeSet<>();
      Collection<UUID> children;
      // if(vstorage.canLoad(uuid))//vnode
      // children = vstorage.getChildren(uuid, true);
      // else
      children = storage.getChildren(uuid);
      if (children != null) {
        for (UUID uid : children) if (uid != null) initalData.add(uid);
      }
      newChildren.put(uuid, new TreeSet<>(initalData));
      oldChildren.put(uuid, new TreeSet<>(initalData));
    }
    return newChildren.get(uuid);
  }

  private Map<String, AttributeDefinition> currentAttrs(UUID uuid) {
    if (!newAttr.containsKey(uuid)) {
      Map<String, AttributeDefinition> initalData = new HashMap<>();
      Map<String, AttributeDefinition> attrs = storage.getAttributes(uuid);
      if (attrs != null) {
        for (String s : attrs.keySet()) initalData.put(s, attrs.get(s));
      }
      newAttr.put(uuid, new HashMap<>(initalData));
      oldAttr.put(uuid, new HashMap<>(initalData));
    }
    return newAttr.get(uuid);
  }
  private static String OLD = "~~~old~~~|placeh";
  private Map<String, String> currentResources(UUID uuid) {
    if (!resourсes.containsKey(uuid)) {
      resourсes.put(uuid, new HashMap<String, String>());
      for (String s : storage.listResources(uuid))
        resourсes.get(uuid).put(s,OLD);
        // setResourceContents(uuid, s, null, null);
    }
    return resourсes.get(uuid);
  }

  public void apply() {
    apply(storage);
  }

  public void apply(ITreeStorage aStorage) {
    operationsData.clear();
    // aStorage.startChangeRow();
    int ops = operations.size(), i = 0;
    if (verbose) RLogger.logInfo(ops + " operations to be performed");
    LinkedList<TreeOperation> opss = new LinkedList<>();
    int j = operations.size();

    while (j > 0) {
      LinkedList<TreeOperation> q = new LinkedList<>(operations);
      j = 0;
      opss.addAll(operations);
      operations.clear();
      for (TreeOperation op : q) {
        if (!(op instanceof SaveVersions)) j++;
        i++;
        if (i % 10 == 0 && verbose)
          RLogger.logInfo(i + " operations out of " + opss.size() + " were completed");
        if (verbose) {
          RLogger.logInfo("applying operation " + op);
        }
        String b = storage.getCacheStorage().startCacheTransaction("Start db transaction",CacheStorage.COMMIT_SINGLE);
        op.apply(aStorage);
        if(b == null) storage.getCacheStorage().commitCache();
        operationsData.add(new OperationData(op, op.getData()));
      }
    }
    operations.addAll(opss);
    versstorage.clear();
    List<ITransactionListener> tmpListeners = new ArrayList<>(tListeners);
    for (ITransactionListener tl : tmpListeners) {
      if (execute) {
        tl.execute(this);
      } else {
        tl.redo(this);
      }
    }
    execute = false; // next time apply means redo
    // aStorage.stopChangeRow();
  }

  public void rollback() {
    rollback(storage);
  }

  public void rollback(ITreeStorage aStorage) {
    operationsData.clear();
    // aStorage.startChangeRow();
    LinkedList<TreeOperation> reverceOperations = new LinkedList<>(operations);

    Collections.reverse(reverceOperations);
    operations.clear();
    operations.addAll(reverceOperations);
    int ops = reverceOperations.size(), i = 0;
    if (verbose) RLogger.logInfo(ops + " operations to be reverted");
    LinkedList<TreeOperation> opss = new LinkedList<>();

    int j = operations.size();
    // TODO think
    // while(j>0){
    j = 0;
    LinkedList<TreeOperation> q = new LinkedList<>(operations);
    opss.addAll(operations);
    operations.clear();
    for (TreeOperation op : q) {
      if (!(op instanceof SaveVersions)) j++;
      i++;
      if (i % 10 == 0 && verbose)
        RLogger.logInfo(i + " operations out of " + operations.size() + " were completed");
      if (verbose) {
        RLogger.logInfo("applying operation " + op);
      }
      String b = storage.getCacheStorage().startCacheTransaction("Init db rollback",CacheStorage.COMMIT_SINGLE);
      op.rollback(aStorage);
      if(b == null) storage.getCacheStorage().commitCache();
      operationsData.add(new OperationData(op, op.getData()));
    }
    // }
    if (!(aStorage instanceof TransactionStorage)) {
      List<ITransactionListener> tmpListeners = new ArrayList<>(tListeners);
      for (ITransactionListener tl : tmpListeners) {
        tl.undo(this);
      }
    }
    reverceOperations = new LinkedList<>(opss);

    Collections.reverse(reverceOperations);
    operations.clear();
    operations.addAll(reverceOperations);
    // aStorage.stopChangeRow();
  }

  public boolean isResourceInTransaction(UUID uuid, String resName) {
    Map<String, String> nodeResources = resourсes.get(uuid);
    if (nodeResources == null) {
      return false;
    }
    if (nodeResources.containsKey(resName)&&!OLD.equals(nodeResources.get(resName))) {
      return true;
    }
    return false;
  }

  // //////////////////////
  // Resources
  @Override
  public boolean isResourceExists(UUID uuid, String resName) {
    if (isResourceInTransaction(uuid, resName)) {
      if (resourсes.get(uuid).containsKey(resName) && resourсes.get(uuid).get(resName) == null) {
        return false;
      }
      return true;
    }
    // if (storage.getNodeId(uuid,true) == null)
    // return false;
    if (storage.isResourceExists(uuid, resName)) {
      if (!resourсes.containsValue(uuid)) { // we doun't change node in
        // transaction
        return true;
      }

      if (!resourсes.get(uuid).containsKey(resName)) { // we doun't change
        // resource in
        // transaction
        return true;
      }
      if (resourсes.get(uuid).get(resName) != null) {
        return true;
      }
    }
    return false;
  }

  @Override
  public InputStream getResourceContents(UUID uuid, String resName) {
    long t_start = System.currentTimeMillis();
    if (isResourceInTransaction(uuid, resName)) {
      try {
        if (resourсes.get(uuid).get(resName) != null) {
          if(OLD.equals(resourсes.get(uuid).get(resName)))
            return storage.getResourceContents(uuid, resName);
          return new FileInputStream(resourсes.get(uuid).get(resName));
        }
        return null;
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      }
    }

    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo(
          "time TransactionStorage.getResourceContents("
              + uuid
              + ", "
              + resName
              + ") = "
              + (t_end - t_start));
    }
    return storage.getResourceContents(uuid, resName);
  }

  @Override
  public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
    long t_start = System.currentTimeMillis();
    try {
      getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
      Map<String, String> nodeResources = currentResources(uuid);

      String oldContentFile = null;

      if (isResourceExists(uuid, resName)) {
        File tmpOldFile = File.createTempFile(TMP_RESOURCE_PREFIX, "old");
        oldContentFile = tmpOldFile.getAbsolutePath();
        InputStream oldIn = getResourceContents(uuid, resName);

        OutputStream out = new FileOutputStream(tmpOldFile);
        ChangeResource.copy(oldIn, out);
        out.close();
        oldIn.close();
      }

      File tmpNewFile = File.createTempFile(TMP_RESOURCE_PREFIX, "new");
      String newContentFile = tmpNewFile.getAbsolutePath();
      OutputStream out = new FileOutputStream(tmpNewFile);
      if (in != null) {
        ChangeResource.copy(in, out);
        in.close();
      }
      out.close();
      nodeResources.put(resName, newContentFile);

      ChangeResource operation =
          new ChangeResource(uuid, resName, oldContentFile, newContentFile, data);
      operations.add(operation);

      getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
    } catch (IOException e) {
      e.printStackTrace();
    }
    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo(
          "time TransactionStorage.setResourceContents("
              + uuid
              + ", "
              + resName
              + ", ...) = "
              + (t_end - t_start));
    }
  }

  @Override
  public void deleteResource(UUID uuid, String resName) {
    try {
      getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
      File tmpOldFile = File.createTempFile(TMP_RESOURCE_PREFIX, "new");
      OutputStream out = new FileOutputStream(tmpOldFile);
      String oldContentFile = tmpOldFile.getAbsolutePath();
      InputStream in = getResourceContents(uuid, resName);
      try{
        if (in != null) ChangeResource.copy(in, out);
      }
      finally{
        if(in!=null)
          in.close();
        out.close();
      }
      Map<String, String> nodeResources = currentResources(uuid);
      nodeResources.put(resName, null);

      ChangeResource operation = new ChangeResource(uuid, resName, oldContentFile, null, null);
      operations.add(operation);

      getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public void dispose() {
    for (TreeOperation op : operations) {
      op.dispose();
    }
  }

  @Override
  public void createStorage(String name) {
    // TODO Implement me!
    throw new UnsupportedOperationException("not implemented");
  }

  @Override
  public void deleteStorage() {
    // TODO Implement me!
    throw new UnsupportedOperationException("not implemented");
  }

  public LinkedList<OperationData> getLastOperationsData() {
    return operationsData;
  }

  public ITreeStorage getVStorage() {
    return vstorage;
  }

  @Override
  public ITreeStorage getBaseStorage() {
    return storage;
  }
//
//  @Override
//  public Object getCacheVar(UUID uuid, String section, String var) {
//    Object che = storage.getCacheVar(uuid, section, var);
//    Object vers = super.getCacheVar(uuid, section, var); // local
//    // inmem
//    if (vers != null) return vers;
//    return che;
//  }
//
//  @Override
//  public void setCacheVar(UUID uuid, String section, String var, String value) {
//    Object oldAttributes = getCacheVar(uuid, section, var);
//    Map<String, Map<String,Object>> old = new HashMap<>(), newo = new HashMap<>();
//    if(oldAttributes!=null){
//      old.put(var,new HashMap<>());
//      old.get(var).put("",  oldAttributes);
//    }
//    newo.put(var, new HashMap<String,Object>());
//    
//    newo.get(var).put("", value);
//    Set<String> keys = new HashSet<>();
//    keys.add(var);
//    getDeltaTarget().cacheChanged(uuid, section, keys, old, newo, Moment.BEFORE, true);
//    super.setCacheVar(uuid, section, var, value);
//    SaveCache saveOperation = new SaveCache(uuid, section, old, newo);
//    operations.add(saveOperation);
//    // is it needed?
//    getDeltaTarget().cacheChanged(uuid, section, keys, old, newo, Moment.AFTER, true);
//  }
//
//  @Override
//  public void setCacheVars(UUID uuid, String section, Map<String, Map<String, Object>>ts) {
//    if(ts == null)
//      throw new RuntimeException("null ts for "+uuid+" "+section);
//    Map<String, Map<String, Object>> oldAttributes = getCacheVars(uuid, section, ts.keySet());
////    oldAttributes = getCacheVars(uuid, section, ts.keySet());
//       
////        getCacheVars(uuid, section, ts != null ? ts.keySet() : null);
//    Set<String> ak = new HashSet<>(ts.keySet());
//    if (oldAttributes != null) ak.addAll(oldAttributes.keySet());
//    // for(String s:ak)
//    getDeltaTarget()
//        .cacheChanged(
//            uuid,
//            section,
//            ak,
//            oldAttributes != null ? oldAttributes : null,
//            ts,
//            Moment.BEFORE,
//            true);
//    setCacheVars(uuid, section, ts);
//    SaveCache saveOperation = new SaveCache(uuid, section, oldAttributes, ts);
//    operations.add(saveOperation);
//    // is it needed?
//    getDeltaTarget()
//        .cacheChanged(
//            uuid,
//            section,
//            ak,
//            oldAttributes != null ? oldAttributes : null,
//            ts,
//            Moment.AFTER,
//            true);
//  }
//
//  @Override
//  public Map<String, Map<String, Object>> getCacheVars(UUID uuid, String section, Set<String> vars) {
//
//    Map<String, Map<String, Object>> che = storage.getCacheVars(uuid, section, vars);
//    Map<String, Map<String, Object>> vers = super.getCacheVars(uuid, section, vars); // local inmem
//    if (che != null) che = new HashMap<>(che);
//    else che = new HashMap<>();
//    if (vers != null) {
//      for (String s : vers.keySet()) che.put(s, vers.get(s));
//      return che;
//    }
//    return che;
//  }

  @Override
  public void clearCachedCache(UUID uid) {
    storage.clearCachedCache(uid);
  }

  @Override
  public void clearCaches() {
    storage.clearCaches();
  }

  @Override
  public File getResource(UUID uuid, String resName) {
    long t_start = System.currentTimeMillis();
    if (isResourceInTransaction(uuid, resName)) {
      if (resourсes.get(uuid).get(resName) != null) {
        if(OLD.equals(resourсes.get(uuid).get(resName)))
          return storage.getResource(uuid, resName);
        return new File(resourсes.get(uuid).get(resName));
      }
      return null;
    }

    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo(
          "time TransactionStorage.getResourceContents("
              + uuid
              + ", "
              + resName
              + ") = "
              + (t_end - t_start));
    }
    return storage.getResource(uuid, resName);
  }

  @Override
  public Path getLocation() {
    return storage.getLocation();
  }
}
