/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.IReqURIResolver;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

public class Location extends TreeNode implements INodeChangeListener {
  public static final String TYPE_NAME = "Location";

  //private String label;
  // private boolean isTransferred = false;

  public Location() {
    super(null, null);
  }

  public Location(TreeDB db, UUID uid) {
    super(db, uid);
    setType(TYPE_NAME);
  }

  private String getHTMLFolder() {
    TreeNode parent = getWorkParent();
    // in a case of tests
    if (parent == null) return "";
    return new File(((com.unitesk.requality.nodetypes.Document) parent).getHTMLPath()).getParent();
  }

  public Object getDocumentNodeModel() {
    Document d = getDocument();
    if (d == null) return null;
    return d.getNodeModel();
  }

  @Override
  public String getId() { // hack.. beware of files
    return getUUId().toString();
  };

  @Override
  public String getDescription(String prefix, String postfix, String basePath, IReqURIResolver rslv) {
    String desc = getDescription();
    if (desc == null) desc = "";
    try {
      return resolveURI(rslv, CoreUtils.htmlEncapsulate(
          RequalityCLI.getReportDescription(
              this, getHTMLFolder(), desc, prefix, postfix, basePath)));
    } catch (UnsupportedEncodingException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public String getQualifiedId() {
    if (getWorkParent() != null)
      return getWorkParent().getQualifiedId() + "/" + getUUId().toString();
    RLogger.logInfo("null parent!");
    return getUUId().toString();
  };

  public Requirement[] getRequirements() {
    return getRequirements(false);
  }

  public Requirement[] getRequirements(boolean sort) {
    return getRequirements(true,sort);
  }
  public Requirement[] getRequirements(boolean usestored,boolean sort) {
    Map<String, Set<TreeNode>> rev = getTreeDB().getLinkResolver().getRevertLinks(this);
    Requirement[] reqs = null;
    Set<TreeNode> reqss = new HashSet<>();
    if (rev != null) {
      for (String s : rev.keySet())
        if (s.contains(Requirement.ATTR_LOCATIONS)) reqss.addAll(rev.get(s));
    }
    int i = 0;
    reqs = new Requirement[reqss.size()];
    for (TreeNode node : reqss) if (node instanceof Requirement) reqs[i++] = (Requirement) node;
    if(sort)
      Arrays.sort(reqs, TreeNodeComparator.getComparator(getTreeDB()));
    // String locQId = getQualifiedId();
    // if(usestored)
    // reqs = getStoredReqs();
    // if(reqs == null){
    // try{
    // reqs = getRequirementsFor(getTreeDB(), locQId);
    // }catch(RuntimeException re){
    // RLogger.logInfo(re.getMessage()+"
    // "+re.getStackTrace()[0].toString());
    // return null;
    // }
    // List<Object> rs = new ArrayList<Object>();
    // if(reqs!=null){
    // for(Requirement r:reqs){
    // Attribute lv = r.attrs().getAttribute(Requirement.ATTR_LOCATIONS);
    // long v = -1;
    // if(lv!=null){
    // v = lv.getVersion().getSInt();
    // }else{
    // //TODO error
    // }
    // rs.add(r.getUUId()+"|"+v);
    // }
    // ListValue lv = new ListValue(AttributeType.STRING, rs);
    // if(getTreeDB().getCacheStorage()!=null)
    // getTreeDB().getCacheStorage().setValue(getUUId(),ATTR_REV_LOCATIONS,
    // lv);
    // saveAttributes();
    // }
    // }
    return reqs;
  }

  private static Set<TreeDB> loading = new HashSet<>();

  public static void markLocations(TreeNode r, Set<String> old) {
    if (!(r instanceof Requirement)) return;
    if (old != null) old = new HashSet<>(old);
    TreeDB db = r.getTreeDB();
    if (db.isVirtual()) db = Instantiator.getRootTree(db);
    Attribute locs = r.attrs().getAttribute(Requirement.ATTR_LOCATIONS);
    // TODO think
  }

  private static boolean hasNVirtVNParent(TreeNode r) {
    if (r == null) return false;
    if (!r.isVirtual()) return true;
    TreeNode par = r.getWorkParent();
    if (par == null || par.isRoot()) return false;
    if (!par.isVirtual()) return true;
    return hasNVirtVNParent(par);
  }

  public static Requirement[] getRequirementsFor(TreeDB db, Location loc) {
    TreeNode t = null;
    db = Instantiator.getRootTree(db);
    if (loading.contains(db)) throw new RuntimeException("Loading... " + loc);
    if (db.isVirtual()) db = Instantiator.getRootTree(db);
    try {
      t = loc; // slow
      if (t != null) {
        Requirement[] req = ((Location) t).getRequirements(true,true);
        if (req != null) return req;
      }
    } catch (Exception e) {
      e.printStackTrace();
    }

    return new Requirement[0];
  }

  private static Comparator<Requirement> reqcomparator =
      new Comparator<Requirement>() {
        @Override
        public int compare(Requirement arg0, Requirement arg1) {
          return arg0.getUserVisibleName(true).compareTo(arg1.getUserVisibleName(true));
        }
      };

  /**
   * Returns document location is related to. If there is no such document - returns null. If there
   * are two or more documents - returns first available and prints warning to console
   *
   * @return document related to location or null
   */
  public Document getDocument() {
    return (Document) getWorkParent();
  }

  public Document[] getDocuments() {
    UUID locQId = getUUId();
    return getDocumentsFor(getTreeDB(), locQId);
  }

  public static Document[] getDocumentsFor(TreeDB db, UUID locQId) {
    ArrayList<Document> result = new ArrayList<>();
    for (TreeNode node :
        new DeepFirstTreeWalker(
            db.getNode(Document.getTypeRootQId(), Path.QID | Path.UV), true, true, false)) {
      if (!node.getType().equals(Document.TYPE_NAME)) {
        continue;
      }
      Document doc = (Document) node;
      if (doc.getLocations().contains(locQId)) {
        result.add(doc);
      }
    }
    return result.toArray(new Document[0]);
  }

  private static String toString(TreeDB db, Location loc, String uid) {
    Requirement[] reqs = null;
    try {
      reqs = getRequirementsFor(db, loc);
    } catch (RuntimeException re) {
      // cache loading
      return "LOADING..." + loc;
    }
    if (reqs.length == 0) {
      return "orphan location(" + loc.getId() + ")";
    }
    String[] names = new String[reqs.length];

    int i = 0;

    Set<String> qids = new HashSet<>();
    for (Requirement curReq : reqs) {
      String qid = curReq.getQualifiedId(true);
      if (!qids.contains(qid)) {
        qids.add(qid);
        String reqName = curReq.getUserVisibleName(true);
        List<ReferenceValue> locQIds = curReq.getLocationQIds();
        int curInd = curReq.getLocationIndex(uid) + 1;
        if (locQIds.size() > 1) {
          names[i] = reqName + "#" + curInd;
        } else {
          names[i] = reqName;
        }
        i++;
      }
    }
    int n = i;
    StringBuilder sb = new StringBuilder();
    sb.append(names[0]);
    for (i = 1; i < n; i++) {
      sb.append(", ");
      sb.append(names[i]);
    }
    return sb.toString();
  }

  @Override
  public String formLabel(){
    return Location.toString(getTreeDB(), this, getUUId().toString());
  }
  
  @Override
  public Set<String> getLabelDeps() {
    // TODO Auto-generated method stub
    return Collections.emptySet();
  }

  public void resetLabel() {
    attrs().updateAttribute(TreeNode.ATTR_LABEL);
//    String nl = Location.toString(getTreeDB(), getQualifiedId(), getUUId().toString());
//    if (label == null || !nl.equals(label)) {
//      label = nl;
//      getTreeDB()
//          .fireNodeChange(
//              this,
//              new ChangesSet(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet()),
//              Collections.EMPTY_MAP,
//              true);
//      fireNodeAttributeChange(
//          new ChangesSet(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet()));
//      getTreeDB()
//        .fireNodeGUIChange(
//         this,
//         new ChangesSet(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet()),
//         Collections.EMPTY_MAP,
//         true);
//
//      fireGUINodeAttributeChange(
//          new ChangesSet(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet())
//          );
//    }
  }

  @Override
  public Index getIndex() {
    return Index.create(Integer.toString(getIndexInDocument()));
  }

  public int getIndexInDocument() {
    Document d = getDocument();
    if (d != null) return d.getLocationIndex(getUUId());
    return -1;
  }

  public int compareIndexTo(Location l2) {
    Document doc1 = getDocument(), doc2 = l2.getDocument();
    if (doc1 == null || doc2 == null) return 0;
    if (!doc1.getUUId().equals(doc2.getUUId())) return doc1.compareTo(doc2);
    return ((Integer) getIndexInDocument()).compareTo(l2.getIndexInDocument());
  }

  public static void clearCache(TreeDB db) {
    // reqCahceMap.remove(db);
    loading.remove(db);
  }

  public static void clearCache(TreeDB db, String qualifiedId) {
    // if(reqCahceMap.containsKey(db))
    // reqCahceMap.get(db).remove(qualifiedId);
  }

  @Override
  public void attributeChange(TreeNode source, String what, Attribute old) {}

  @Override
  public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {}

  @Override
  public void deleted(TreeNode source) {
    // TreeDB db = getTreeDB();
    // TODO root?
    // if(db.isVirtual())
    // db = Instantiator.getRootTree(db);
    if (!source.getUUId().equals(getUUId())) return;
    // if(reqCahceMap.containsKey(db))
    // reqCahceMap.get(db).remove(getQualifiedId()+"/"+getUUId().toString());
    // Requirement[]reqs = getRequirements();
    getDocument().removeChildUUID(getUUId()); // TODO move?
    resetRelated();
  }

  @Override
  public void moved(TreeNode node, TreeNode oldParent, boolean internal) {
    // TODO Auto-generated method stub

  }

  //
  @Override
  public boolean isHidden() {
    // TODO check parameters
    return super.isHidden();
  }
  // public void setTransferredLocation() {
  // isTransferred = true;
  // }
  //
  // public boolean isTransferred() {
  // return isTransferred;
  //
  // }

  @Override
  public void fireNodeResourceChange(TreeNode treeNode, String resName, Moment m, boolean inner) {
    // TODO Auto-generated method stub

  }

  public void removeFragments(Requirement node) {
    // updateRevLocations(node);
    Requirement[] reqs = getRequirements();
    if (reqs == null || reqs.length == 0) delete(); // TODO check
    else commitAttributes();
  }

  public static boolean isLoading(TreeDB rootTree) {
    return loading.contains(rootTree);
  }

  @Override
  public void reactAfterAttributeChange(String s, Attribute attribute) {
    super.reactAfterAttributeChange(s, attribute);
   // if (s.contains(TreeNode.ATTR_ID)) resetRelated();
  }

  @Override
  public void revCacheChanged(TreeNode node, String s) {
    super.revCacheChanged(node, s);
    //resetRelated();
  }

  public void resetRelated() {
    resetLabel();
    for (Requirement r : getRequirements())
      for (Location l : r.getLocations()) l.resetLabel();
  }
  @Override
  public TreeNode getExposedNode() {
    Requirement[] reqs = getRequirements(true);
    if(reqs!=null && reqs.length>0)
      return reqs[0];
    return super.getExposedNode();
  }
}
