/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.UUID;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Flag;

public class NodeType extends TreeNode {
  public static final String TYPE_NAME = "NodeType";
  private static final String ATTRIBUTE_CLASSNAME = "_classname";
  private static final String ATTRIBUTE_VISIBLE_CLASSNAME = "_visclassname";
  private static final String ATTRIBUTE_ICON = "_iconResource";
  private Class<? extends TreeNode> clazz;

  public NodeType(String name, TreeDB db) {
    super(db, UUID.randomUUID());
    setId(name);
    setType(TYPE_NAME);
  }

  public NodeType(TreeDB db, UUID uuid) {
    super(db, uuid);
  }

  public String getName() {
    return getId();
  }

  @Override
  public String toString() {
    return "TypeNode '" + getName() + "' (" + getUUId() + ")";
  }

  public void setVisibleClassName(Class<? extends TreeNode> clazz) {
    String name = CoreUtils.splitWords(clazz.getSimpleName());
    Method m;
    try {
      m = clazz.getMethod("getVisibleType", new Class[0]);
      if (m != null) {
        name = (String) m.invoke(clazz, new Object[0]);
      }

    } catch (SecurityException e) {
    } catch (NoSuchMethodException e) {
    } catch (IllegalArgumentException e) {
    } catch (IllegalAccessException e) {
    } catch (InvocationTargetException e) {
    }
    attrs().setConstAttribute(ATTRIBUTE_VISIBLE_CLASSNAME, name, Flag.HIDDEN);
  }

  public String getVisibleClassName() {
    return attrs().getAttribute(ATTRIBUTE_VISIBLE_CLASSNAME).getValue().toString();
  }

  public void setClassName(String className) {
    attrs().setConstAttribute(ATTRIBUTE_CLASSNAME, className,Flag.HIDDEN);
  }

  public void setClass(Class<? extends TreeNode> clazz) {
    attrs().setConstAttribute(ATTRIBUTE_CLASSNAME, clazz.getName(),Flag.HIDDEN);
    this.clazz = clazz;
    setVisibleClassName(clazz);
  }

  @SuppressWarnings("unchecked")
  public Class<? extends TreeNode> getTypeClass() {
    if (clazz != null) {
      return clazz;
    }
    String className = attrs().getStringAttributeValue(ATTRIBUTE_CLASSNAME);
    if (className == null) {
      return null;
    }
    try {
      return (Class<? extends TreeNode>) Class.forName(className);
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    return null;
  }

  public String getIconName() {
    if (attrs().getStringAttributeValue(ATTRIBUTE_ICON) == null) return getId();
    return attrs().getStringAttributeValue(ATTRIBUTE_ICON);
  }

  public boolean setName(String icon) {
    attrs().setConstAttribute(ATTRIBUTE_ICON, icon,Flag.HIDDEN);
    return true;
  }

  @Override
  public boolean equals(Object obj) {
    if (super.equals(obj)) return true;
    if (obj instanceof NodeType) {
      NodeType t = (NodeType) obj;
      if (t.getName().equals(getName())) return true;
    }
    return false;
  }

  @Override
  public boolean isRestricted() {
    return true;
  }
}
