/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;

public class ReportSettings extends TreeNode {
  public static SimpleDateFormat dataFormat =
      new SimpleDateFormat("EEE MMM dd HH:mm:ss zzz yyyy", Locale.ENGLISH);
  protected String lastTemplate = null;
  private INodeChangeListener ncl =
      new INodeChangeListener() {
        @Override
        public void deleted(TreeNode source) {}

        @Override
        public void attributeChange(TreeNode source, String what, Attribute o) {
          String old = null;
          if(o!=null && o.getValue()!=null)
            old = o.getGValue();
          if (source == ReportSettings.this
              && (what.equals(ATTR_TEMPLATE) || what.equals(TreeNode.ATTR_UNDEFINED))
              && old != null
              && (getLastTemplate() == null || !lastTemplate.equals(old))) {
            lastTemplate = old;
            // if(source.getTreeDB().getStorage() instanceof
            // ResourceStorage)
            // GenerateReportHandler.generateReport(ReportSettings.this);

          }
        }

        @Override
        public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          if (what.contains(ATTR_TEMPLATE) || what.contains(TreeNode.ATTR_UNDEFINED))
            attributeChange(source, ATTR_TEMPLATE, what.getOld(ATTR_TEMPLATE));
        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };
  public static final String TYPE_NAME = "ReportSettings";
  public static final String ATTR_ROOT_REQUIREMENT = "_root_requiremnt_qid";
  public static final String ATTR_DATE = "date";
  public static final String ATTR_TEMPLATE = "_template_id";
  public static final String ATTR_LAST_GENERATED = "_last_generated";
  public static final String ATTR_GEN = "generating_in_process";

  public ReportSettings() {
    this(null, null);
  }

  protected String getLastTemplate() {
    if (lastTemplate == null) lastTemplate = this.attrs().getStringAttributeValue(ATTR_TEMPLATE);
    return lastTemplate;
  }

  public ReportSettings(TreeDB db, UUID uuid) {
    super(db, uuid);
    this.addChangeListener(ncl);
    // lastTemplate = attrs().getLocalStringAttributeValue(ATTR_TEMPLATE);
  }

  @Override
  public String getMultiTypeLabel() {
    return getTypeLabel();
  }

  public String getStartRequirementQId() {
    return attrs().getStringAttributeValue(ATTR_ROOT_REQUIREMENT, Requirement.getTypeRootQId());
  }

  public Map<String, Object> getData() {
    //TODO remove
    Map<String, Object> result = new HashMap<>();
    for (Entry<String, Attribute> en : attrs().getLocalAttributes().entrySet()) {
      if (!en.getKey().startsWith("_")) {
        result.put(en.getKey(), en.getValue().getValue());
      }
    }
    return result;
  }

  public URI getSaveDir() {
    try {
      return getNodeResourcesPath();
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  public String getTemplateId() {
    return attrs().getStringAttributeValue(ATTR_TEMPLATE);
  }

  @Override
  public boolean canEditAttributesInTable() {
    return true;
  }

  public boolean isGenerating() {
    return attrs().getBooleanAttributeValue(ATTR_GEN);
  }

  public void setGenerating(boolean b) {
    attrs().setConstAttribute(ATTR_GEN, b, Flag.VIRTUAL| Flag.HIDDEN);
  }
}
