package com.unitesk.requality.nodetypes;

import java.net.URI;
import java.util.UUID;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.tools.RequalityCLI;

public class TestPurpose extends TreeNode {

  public static final String TYPE_NAME = "TestPurpose";
  public static final String ATTR_EXPECTED_RESULTS = "_expectedResults";
  public static final String ATTR_STATUS = "_status";
  public static final String ATTR_AUTHOR = "_author";
  public static final String ATTR_NAME = "_name";
  public static final String ATTR_TEST_METHOD = "_testMethod";
  public static final String ATTR_TEST_LOCATION = "_testLocation";

  {
    Attribute.registerStatusName(ATTR_STATUS);
    Attribute.registerHTMLNames(ATTR_EXPECTED_RESULTS);
    Attribute.registerStatusName(ATTR_TEST_METHOD);
    Attribute.registerStatusName(ATTR_TEST_LOCATION);
  }

  // Constructor for eclipse instancer
  public TestPurpose() {
    super(null, null);
  }

  private void fixAttributes() {
    String oldResults = attrs().getStringAttributeValue("expectedResults", "");
    if (oldResults == null || oldResults.isEmpty()) return;
    attrs().removeAttribute("expectedResults");
    setExpectedResults(attrs().getStringAttributeValue(ATTR_EXPECTED_RESULTS, "") + oldResults);
  }

  public TestPurpose(TreeDB db, UUID id) {
    super(db, id);
  }

  public String getTargetRequirementQId() {
    if (getShownParent() == null) return null;
    return getShownParent().getQualifiedId();
  }

  public void setTargetRequirementQId(String targetReqQId) {
    if (!targetReqQId.equals(getWorkParent().getQualifiedId())) {
      setParent(getTreeDB().getNode(targetReqQId, Path.QID | Path.UV | Path.UUID));
    }
  }

  public String getExpectedResults(String prefix, String postfix, String basePath) {
    String desc = getExpectedResults(false);
    if (desc == null) return desc;
    try {
      URI path = getNodeResourcesPath();
      if (path == null) return null;
      return RequalityCLI.getReportDescription(
          this, path.toString(), desc, prefix, postfix, basePath);
    } catch (Exception e) { // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  public String getExpectedResults(String basePath) {
    String result = getExpectedResults("<div>", "</div>", basePath);
    return result;
  }

  public String getExpectedResults(boolean process) {
    fixAttributes();
    String result = attrs().getStringAttributeValue(ATTR_EXPECTED_RESULTS, "");
    if (process) {
      result = processResourcesPath(result);
    }
    return result;
  }

  public void setExpectedResults(String expectedResults) {
    attrs().setAttribute(new AttributeDefinition(getUUId(), ATTR_EXPECTED_RESULTS, new 
        HtmlValueProvider(AttributeType.HTML, expectedResults),Flag.HIDDEN));
  }

  @Override
  public boolean canEditAttributesInTable() {
    return true;
  }
}
