/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.VersionsStorage.Type;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.core.nodeiterators.RequalityChildIterator;
import com.unitesk.requality.core.nodeiterators.RequalityIterator;
import com.unitesk.requality.core.nodeiterators.RequalityIteratorType;
import com.unitesk.requality.core.propagators.VNodePropagator;

public class VirtualNode extends TreeNode {
  public static final String TYPE_NAME = "VirtualNode";
  public static final String ATTR_CHILD_ITERATOR = "_childiterators";
  public static final String ATTR_TARGET = "_target";
  public static final String ATTR_TREECHANGES = "_changes";
  public static final String ATTR_HIDDEN_STATUS = "_%hidden";
  public static final String ATTR_COLLAPSEVNODE = "_hiddenstate";
  public static final String ATTR_TARGET_STRUCT_VERSION = "_%vstversion";

  {
    Attribute.registerStatusName(VirtualNode.ATTR_TARGET);
  }

  private TreeNode target;
  private boolean moved = false;

  private RequalityChildIterator childiterator;
  private boolean childrenLoaded = false;
  private INodeChangeListener targetChange =
      new INodeChangeListener() {
        @Override
        public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
          try {
            if (getChildIterator().checkTarget(VirtualNode.this, source) == null) {
            } else {
              setTarget(null);
              commitAttributes();
            }
          } catch (Exception e) {

          }
        }

        @Override
        public void deleted(TreeNode source) {
          try {
            if (isCollapsed()) delete();
            else {
              setTarget(null);
              commitAttributes();
            }
          } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
        }

        @Override
        public void attributeChange(TreeNode source, String what, Attribute old) {}

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          // TODO Auto-generated method stub

        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };
  private boolean onMove = false;

  @Override
  public void setParent(TreeNode parent) {
    TreeNode targ = getTarget();
    try {
      onMove = true;
      if (getChildIterator().checkTarget(VirtualNode.this, targ) == null) {
        setTarget(targ);
      } else {
        setTarget(null);
      }
      commitAttributes();
      super.setParent(parent);

    } catch (Exception e) {
      e.printStackTrace();
    } finally {
      onMove = false;
    }
  };

  public RequalityChildIterator getChildIterator() {
    //    	if(!isVirtual()){
    //    		TreeNode node = getVirtMirror();
    //    		if(node!=null)
    //    			return ((VirtualNode)node).getChildIterator();
    //    		return null;
    //    	}
    if (childiterator == null) {
      Attribute a = attrs().getAttribute(ATTR_CHILD_ITERATOR);
      if (a != null) {
        String val = (String) a.getValue();
        if (val != null) {
          String[] s = val.split("\\|(?=(?:[^\"]*[\"][^\"]*[\"])*[^\"]*$)");
          if (s.length > 1) {
            childiterator = RequalityChildIterator.fromAttributeString(null, this, val);
          } else
            childiterator =
                (RequalityChildIterator) RequalityIterator.fromAttributeString(getUUId(), val);
        }
      }
    }
    //first initiation
    if (childiterator == null) {
      childiterator =
          (RequalityChildIterator)
              RequalityChildIterator.getOne(null, getUUId(), RequalityIteratorType.CHILD);
      if (childiterator.getMasterListener(this) != null)
        addChangeListener(childiterator.getMasterListener(this));
    }
    return childiterator;
  }

  public void removeClones() {
    //    	if(!isVirtual()){
    //    		VirtualNode ivn = getVirtMirror();
    //    		if(ivn!=null)
    //    			ivn.removeClones();
    //    		return;
    //    	}
    //
    List<UUID> clones = getWorkChildrenUUIDs(true);
    if (clones != null) {
      for (UUID uid : clones) {
        try {
          getVirtualTreeDB().getNode(uid).delete();
          UUID vuid = Instantiator.revreuses.remove(this);
          if (vuid != null) {
            Instantiator.reuses.get(vuid).remove(this);
            if (Instantiator.reuses.get(vuid).isEmpty()) Instantiator.reuses.remove(vuid);
          }
        } catch (RuntimeException re) {
          re.printStackTrace();
        }
        //getVirtualTreeDB().getNode(uid).fireNodeDelete();
        //getVirtualStorage().remove(uid,true);
      }
    }
    childrenLoaded = false;
    shownChildren = null;
  }

  public void setChildIterator(RequalityChildIterator rci) {
    String childIteratorNow = attrs().getStringAttributeValue(ATTR_CHILD_ITERATOR);
    String toSave = rci.toAttributeString();
    if (childIteratorNow == null || !childIteratorNow.equals(toSave)) {
      attrs().setConstAttribute(ATTR_CHILD_ITERATOR, toSave, Flag.HIDDEN);
    }
  }
  //    private void reloadStorage(){
  //    	initStorageOver(getTarget());
  //    }
  Attribute olda = null;
  @Override
  public void reactAfterAttributeChange(String s, Attribute attribute) {
    super.reactAfterAttributeChange(s, attribute);
    boolean isTarget = s.equals(ATTR_TARGET),
        isChildIt =
            s.equals(ATTR_CHILD_ITERATOR);
//    if (s.equals(ATTR_COLLAPSEVNODE)) {
//      try {
//        TreeNode par = getShownParent();
//        getTreeDB().updateNodeChildren(par);
//      } catch (RuntimeException re) {
//        re.printStackTrace();
//        //parent not found
//      }
//    }

    if (isTarget) { //&&isVirtual()){
      if (isVirtual()) {
        //    			getTreeDB().updateNodeChildren(getCloneParent());
        return;
      }
      String value =
          (attribute == null || attribute.getValue() == null)
              ? ""
              : attribute.getValue().toString();
      TreeNode oldt = null;
      Attribute oldAttribute = olda;
      if (oldAttribute != null
          && oldAttribute.getValue() != null
          && oldAttribute.getValue().toString().length() > 0)
        try {
          oldt = getTreeDB().getNode(UUID.fromString(oldAttribute.getValue().toString()));
        } catch (Exception e) {
        }
      if (Instantiator.revreuses.containsKey(this)) {
        UUID uid = Instantiator.revreuses.remove(this);
        Instantiator.reuses.get(uid).remove(this);
        if (Instantiator.reuses.get(uid).isEmpty()) Instantiator.reuses.remove(uid);
      }
      target = null;
      try {
        target = value.equals("") ? null : getTreeDB().getNode(UUID.fromString(value));
      } catch (Exception e) {
      }
      boolean changed = false;
      if (oldt != null && target != null && oldt.getUUId().equals(target.getUUId()))
        return; //nothing to be change
      if (value.equals("") || oldt != null && !oldt.equals(target)) {
        removeClones();
        changed = true;
      }
      childrenLoaded = false;
      if (oldt == null && target != null || oldt != null && !oldt.equals(target)) {
        if (oldt != null) oldt.removeChangeListener(targetChange);
        //updateChildrenIfNeeded();
        UUID ru = Instantiator.revreuses.remove(this);
        if (ru != null) //TODO look into
        Instantiator.reuses.remove(ru);
        if (target != null) {
          target.addChangeListener(targetChange);
          TreeNode targ = getTarget();
          if (targ != null) {
            if (!Instantiator.reuses.containsKey(target.getUUId()))
              Instantiator.reuses.put(target.getUUId(), new HashSet<VirtualNode>());
            Set<VirtualNode> vns = Instantiator.reuses.get(target.getUUId());
            vns.add(this);
            Instantiator.reuses.put(target.getUUId(), vns);
            Instantiator.revreuses.put(this, target.getUUId());
          }
        }
      }
      if (updateChildrenIfNeeded() || changed) {
        getTreeDB().updateNodeChildren(this);
        //				Instantiator.getBaseTree(getTreeDB()).updateNodeChildren(getCloneParent());
        //if (isCollapsed()) getTreeDB().updateNodeChildren(getShownParent());
      }
      olda = attribute;
    }
    if (isChildIt) { //&&isVirtual()){
      if (isVirtual()) {
        //		setChildren(getCloneParent().getWorkChildrenUUIDs());
        return;
      }
      String old = null;
      if (childiterator != null && childiterator.getMasterListener(this) != null) {
        removeChangeListener(childiterator.getMasterListener(this));
        old = childiterator.toAttributeString();
      }
      childiterator = null;
      boolean changed = false;
      attribute = attrs().getAttribute(ATTR_CHILD_ITERATOR);
      if (old != null
          && (attribute == null || !old.equals(attribute.getValue().toString()))) {
        if (Instantiator.revreuses.containsKey(this)) {
          UUID uid = Instantiator.revreuses.remove(this);
          Instantiator.reuses.get(uid).remove(this);
          if (Instantiator.reuses.get(uid).isEmpty()) Instantiator.reuses.remove(uid);
        }

        removeClones();
        changed = true;
        childrenLoaded = false;
      }
      if (attribute != null) {
        Object val = attribute.getValue();
        if (val == null) {
          Attribute a = attrs().getAttribute(ATTR_CHILD_ITERATOR);
          if (a != null) val = a.getValue();
        }
        if(val!=null){
          childiterator = RequalityChildIterator.fromAttributeString(null, this, val.toString());
          childiterator.updateDependencies(this, getTreeDB());
        }
        //updateChildrenIfNeeded();
        if (childiterator != null && childiterator.getMasterListener(this) != null)
          addChangeListener(childiterator.getMasterListener(this));
      }
      if (changed) childrenLoaded = false;
      if (updateChildrenIfNeeded() || changed) {
        getTreeDB().updateNodeChildren(this);
        //Instantiator.getBaseTree(getTreeDB()).updateNodeChildren(getCloneParent());
        if (isCollapsed()) getShownParent().getTreeDB().updateNodeChildren(getShownParent());
        ;
      }
    }
  };

  @Override
  public List<UUID> getWorkChildrenUUIDs(boolean onlyloaded) {
    if (!isVirtual()) {
      if (!onlyloaded) {
        updateChildrenIfNeeded();

        //	AttributeStorage.markUp(getTreeDB(), getUUId());
        return getActiveWorkChildrenUUIDs();
      }
    }
    return super.getWorkChildrenUUIDs(onlyloaded);
  }
  //    @Override
  //    public void setChildrenUUIDs(List<UUID> children2,boolean init) {
  //  //  	if(isVirtual()){
  //			super.setChildren(children2,init);
  //	    	if(isCollapsed())
  //				getShownParent().getVS().incVersion(Type.STRUCT);
  //			return;
  ////    	}else
  ////    		if(!init)
  ////    			children2=new ArrayList<UUID>(getVirtualTreeDB().getVStorage().getChildren(getUUId()));
  ////    	Set<UUID> childs = new HashSet<UUID>(children2);
  ////    	//int cnt = childs.size();
  ////    	//childs.addAll(getVNChildren(true));
  ////    	super.setChildrenUUIDs(new ArrayList<UUID>(childs),init);
  ////    	//if(isVHidden()){
  ////    		//getTreeDB().updateNodeChildren(getParent(true));
  ////    	//}
  ////    	if(isCollapsed())
  ////			getTreeDB().updateShownNodeChildren(getShownParent(),init);
  //	}
  @Override
  public void clearShownChildren() {
    super.clearShownChildren();
    if(isCollapsed()){
      getShownParent().clearShownChildren();
    }
    //getTreeDB().updateNodeChildren(getShownParent());
    //}
  }

  public void setTarget(TreeNode node) throws Exception {
    //getVirtMirror();
    if (node != null && (getQualifiedId() + "/").contains(node.getQualifiedId() + "/"))
      throw new RuntimeException("Recursion on targeting mechanism is prohibited");
    //if(target != null)
    //target.removeChangeListener(targetChange);
    boolean changes = true;
    if (target != null) changes = target != node;
    //if(target!=null)
    //target.addChangeListener(targetChange);
    if (changes) {
      String s = attrs().getStringAttributeValue(ATTR_TARGET);
      if (node == null || !node.getUUId().toString().equals(s)) {
        attrs().setConstAttribute(ATTR_TARGET, node != null ? node.getUUId().toString() : "",
            Flag.HIDDEN);
        //	reloadStorage();
      }
    }
  }

  public TreeNode getTarget() {
    String tid = attrs().getStringAttributeValue(ATTR_TARGET);
    try {
      if (target == null && !tid.equals("")) {
        target = getTreeDB().getNode(UUID.fromString(tid));
        if (target != null) target.addChangeListener(targetChange);
      }
    } catch (Exception e) {
    }
    return target;
  }

  public int getCloneCount() {
    if (getTarget() == null) return 0;
    return getClones().size();
  }

  public boolean hide;

  public TreeDB getVirtualTreeDB() {
    TreeDB db = getTreeDB();
    if (db.isVirtual()) return db;
    UUID uid = getVRoot();
    if (uid != null) return Instantiator.getExistVirtualTreeDB(db, uid);
    return db;
    //Instantiator.getVirtualTreeDB(getTreeDB(),Instantiator.DEFAULT_INSTANCE_NAME);
  }

  public VirtualNode() {
    this(null, null);
  }

  private boolean updateChildrenIfNeeded() {
    return updateChildrenIfNeeded(false);
  }

  public static void setHidden(TreeNode node, boolean state) {
    Attribute hid = node.attrs().getAttribute(VirtualNode.ATTR_HIDDEN_STATUS);

    if (state) {
      if (hid == null) ; //what?;
      //changedAttribute.put(uid, new HashMap<String, Attribute>());
      node.attrs().setConstAttribute(VirtualNode.ATTR_HIDDEN_STATUS, true,
          Flag.HIDDEN);
    } else {
      if (hid != null) {
        node.attrs().removeAttribute(VirtualNode.ATTR_HIDDEN_STATUS);
      }
    }
  }

  @Override
  public TreeNode findWorkChild(String id) {
    for (UUID clone : getWorkChildrenUUIDs()) {
      TreeNode node = getVirtualTreeDB().getNode(clone);
      if (node.getId().equals(id) || id.equals(node.attrs().getStringAttributeValue(TreeInstancer.ATTR_CLONE_ID)))
        return node;
    }
    return super.findWorkChild(id);
  }

  @Override
  public List<TreeNode> findWorkChildPlusName(String id) {
    Set<TreeNode> nodes = new HashSet<>();
    for (UUID clone : getWorkChildrenUUIDs()) {
      TreeNode node = getVirtualTreeDB().getNode(clone);
      if (node.getId().equals(id)
          || id.equals(node.getName())
          || id.equals(node.attrs().getStringAttributeValue(TreeInstancer.ATTR_CLONE_ID))) nodes.add(node);
    }
    nodes.addAll(super.findWorkChildPlusName(id));
    return new ArrayList<>(nodes);
  };

  public boolean updateChildrenIfNeeded(boolean force) {
    //    	if(isVirtual()){
    //    		VirtualNode ivn = (VirtualNode) getCloneParent();
    //    		if(ivn!=null)
    //    			return ivn.updateChildrenIfNeeded(force);
    //    		return false;
    //    	}
    //if(getCloneParent().isVirtual())
    //	return false;//TODO more proper virt reuse
    if (onMove) return false;
    if (moved) {
      if (getChildIterator().checkTarget(VirtualNode.this, getTarget()) != null)
        try {
          setTarget(null);
          return true;
        } catch (Exception e) {
          e.printStackTrace();
        }
      moved = false;
    }
    boolean res = false;
    if (!childrenLoaded || force || getChildIterator().needUpdate(this)) {
      shownChildren = null;
      getChildIterator().needUpdate(this);
      childiterator = getChildIterator();
      childrenLoaded = true;
      TreeNode targ = getTarget();
      if (Instantiator.revreuses.containsKey(this)) {
        UUID uid = Instantiator.revreuses.remove(this);
        Instantiator.reuses.get(uid).remove(this);
        if (Instantiator.reuses.get(uid).isEmpty()) Instantiator.reuses.remove(uid);
      }
      if (targ != null) {
        if (!Instantiator.reuses.containsKey(targ.getUUId()))
          Instantiator.reuses.put(targ.getUUId(), new HashSet<VirtualNode>());
        Set<VirtualNode> vns = Instantiator.reuses.get(targ.getUUId());
        vns.add(this);
        Instantiator.reuses.put(targ.getUUId(), vns);
        Instantiator.revreuses.put(this, targ.getUUId());
      }
      if (childiterator != null) {//Change rows may cause dead locks
        //getTreeDB().startChangeRow();
        //getTreeDB().startChangeRow();
        try{
          res = getChildIterator().updateChildren(VirtualNode.this, getTreeDB());
          if (getTarget() == null) attrs().removeAttribute(ATTR_TARGET_STRUCT_VERSION);
          else
            attrs()
                .setConstAttribute(
                    ATTR_TARGET_STRUCT_VERSION,
                    getTarget().getVS().getVersion(Type.STRUCT).toString(),Flag.HIDDEN|Flag.VIRTUAL);
          getTreeDB().updateNodeChildren(this);
        }
        finally{
          //getTreeDB().stopChangeRow();
        }
      }
      //getTreeDB().fireNodeChange(this, true);
      //getTreeDB().fireNodeGUIChange(this, true);
      //fireGUINodeAttributeChange(TreeNode.ATTR_UNDEFINED);
      //getTreeDB().stopChangeRow();
    }
    return res;
  }
  //    @Override
  //    public TreeNode getChild(int i) {
  //		updateChildrenIfNeeded();
  //		//int cnt = getRawChildrenUUID().size();
  //
  //		return getVirtualTreeDB().getNode(uid);
  //    };
  @Override
  public List<UUID> getWorkChildrenUUIDs() {
    //    	if(!isVirtual()){
    //    		TreeNode ivn = getVirtMirror();
    //    		if(ivn!=null)
    //    			return ivn.getWorkChildrenUUIDs();
    //    		return new ArrayList<UUID>();
    //    	}
    //if(!raw)//&&super.getRawChildrenUUIDs(true).size()>0)
    //    	if(isVirtual() && getVS().isCloneParentVersionOutdated(Type.STRUCT) ||
    //    			!isVirtual() && updateChildrenIfNeeded())
    //    		getTreeDB().updateNodeChildren(this);
    //List<UUID> clones = getVNChildren(true);
    //if(clones!=null)
    //			clones = new ArrayList<UUID>(clones);
    //	else
    //	clones = new ArrayList<UUID>();
    //clones.addAll(super.getRawChildrenUUIDs());
    return super.getWorkChildrenUUIDs();
    //return clns;
    //return getVNChildren(raw);
  }

  public VirtualNode(TreeDB db, UUID uuid) {
    super(db, uuid);
    addAPropogator(VNodePropagator.getOne());
  }

  public void saveChildIterator() {
    setChildIterator(getChildIterator());
  }

  public boolean isCollapsed() {
    boolean hiddenLayer;
    Object value = attrs().getAttributeValue(ATTR_COLLAPSEVNODE);
    if (value instanceof Boolean) hiddenLayer = (Boolean) value;
    else if (value instanceof String) hiddenLayer = ((String) value).toLowerCase().equals("true");
    else hiddenLayer = false;
    return hiddenLayer;
  }

  public List<UUID> getClones() {
    return getWorkChildrenUUIDs();
  }
  //	public List<UUID> getIterChildren(boolean raw) {
  ////		if(!isVirtual()){
  ////    		VirtualNode ivn = getVirtMirror();
  ////    		if(ivn!=null)
  ////    			return ivn.getIterChildren(raw);
  ////    		return new ArrayList<UUID>();
  ////    	}
  //
  //		List<UUID> childs = getVirtualTreeDB().getVStorage().getRepChildren(getUUId(),raw),//TODO make it iter
  //				   tosee = new ArrayList<UUID>(childs);
  //		TreeDB db = getVirtualTreeDB();
  //		if(!raw)
  //			for(UUID uid:tosee){
  //				TreeNode node = db.getNode(uid);
  //				String s = node.attrs().evaluatePredicateString((db.getVStorage()).getPredicate(uid));
  //				if(s.equals(""))
  //					continue;
  //				if(s.startsWith("0")||s.equals("false"))
  //					childs.remove(uid);
  //			}
  //		return childs;
  //	}
  //	public List<UUID> getVNChildren(boolean raw) {
  //		TreeInstancer ti = getVirtualTreeDB().getVStorage();
  //		List<UUID> childs =  ti.getRepChildren(getUUId(),raw),
  //				   tosee = new ArrayList<UUID>(childs);
  //		TreeDB db = getVirtualTreeDB();
  //		if(!raw)
  //			for(UUID uid:tosee){
  //				TreeNode node = db.getNode(uid);
  //				String s = node.attrs().evaluatePredicateString(ti.getPredicate(uid));
  //				if(s.equals(""))
  //					continue;
  //				if(s.startsWith("0")||s.equals("false"))
  //					childs.remove(uid);
  //			}
  //    	return childs;
  //	}
  public static void updateClonesChildren(TreeNode treeNode) {
    if (treeNode == null) return;
    TreeDB db = treeNode.getTreeDB();
    if (!db.isVirtual()) {
      UUID uid = treeNode.getVRoot();
      if (uid != null) db = Instantiator.getExistVirtualTreeDB(treeNode.getTreeDB(), uid);
      else {
        db = Instantiator.getRootTree(treeNode.getTreeDB());
        //db = Instantiator.getExistVirtualTreeDB(treeNode.getTreeDB(),Instantiator.DEFAULT_INSTANCE_NAME,null,null);
      }
    }
    //TODO many instances
//    if (db != null) {
//      db.startChangeRow();
//      Collection<UUID> clones = Instantiator.getClonesAndVnodes(treeNode.getUUId(), db);
//      //TODO needed?
//      //	if(!treeNode.isVirtual())
//      //	db.getNode(treeNode.getUUId()).setChildrenUUIDs(treeNode.getRawChildrenUUIDs());
//      for (UUID uid : clones) {
//        TreeNode clone = db.getCachedNodeOrVirt(uid);
//        if (clone != null) {
//          //TODO struct eventws
////          clone.fireGUINodeAttributeChange(TreeNode.formIdEvent());
////          db.fireNodeGUIChange(clone, TreeNode.IdEvent, Collections.EMPTY_MAP, true);
//        }
//      }
//      db.stopChangeRow();
//    }
  }

  @Override
  public boolean canEditAttributesInTable() {
    return true;
  }

  public void collapse() {
    attrs().setConstAttribute(ATTR_COLLAPSEVNODE, true,Flag.HIDDEN);
  }

  public void unCollapse() {
    attrs().removeAttribute(ATTR_COLLAPSEVNODE);
  }
//
//  @Override
//  public void getUserVisibleName(boolean inner, boolean fc, StringBuilder sb) {
//    if (isCollapsed()) getShownParent().getUserVisibleName(inner, fc, sb);
//    else super.getUserVisibleName(inner, fc, sb);
//  }

  public void markMoved() {
    moved = true;
  }
}
