package com.unitesk.requality.report;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import javax.xml.namespace.QName;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class ReportUtils {
  private TreeDB db;
  private TreeNode root;
  private TemplateFile tf;
  public ReportUtils(TreeDB treeDB, TreeNode root, TemplateFile t) {
    db = treeDB;
    this.root = root;
    tf = t;
  }
  public String getTemplFileName(TreeDB db){
    return tf.getOutputFileName(db);
  }
  public DefaultReportURIResolver getURIResolver(){
    return new DefaultReportURIResolver(tf);
  }
  public static String[] getDirectTerms(TreeNode node){
    Map<String, Set<TreeNode>> map = node.getTreeDB().getTermsStorage().getDirectTerms(node);
    Set<String> nms = new HashSet<>();
    if(map!=null)
    {
      for(Entry<String, Set<TreeNode>> s:map.entrySet())
        nms.add(s.getKey());
      return nms.toArray(new String[0]);
    }
    return null;
  }
  public static TreeNode[] getDirectNodes(TreeNode node, String revname){
    Set<TreeNode> nodes = node.getLinks(revname);
    if(nodes!=null && nodes.size()>0)
      return nodes.toArray(new TreeNode[0]);
    return null;
  }
  public static String[] getRevertTerms(TreeNode node){
    Map<String, Set<TreeNode>> map = node.getTreeDB().getTermsStorage().getRevertTerms(node);
    Set<String> nms = new HashSet<>();
    if(map!=null)
    {
      for(Entry<String, Set<TreeNode>> s:map.entrySet())
        nms.add(s.getKey());
      return nms.toArray(new String[0]);
    }
    return null;
  }
  public static TreeNode[] getRevertTerms(TreeNode node, String revname){
    return getRevertNodes(node, revname);
  }
  public static TreeNode[] getRevertNodes(TreeNode node, String revname){
    Set<TreeNode> nodes = node.getRevertLinks(revname);
    if(nodes!=null && nodes.size()>0)
      return nodes.toArray(new TreeNode[0]);
    return null;
  }
  public static boolean filter(TreeNode node, List<String> filters, String key) {
    if (key.startsWith("_")) return false;
    for (String s : filters)
      if (s.startsWith("*ref")) {
        String opt = s.substring(4);
        boolean not = false;
        if (opt.startsWith("!")) {
          not = true;
          opt = opt.substring(1);
        }
        Attribute a = node.attrs().getAttribute(key);
        if (a != null
            && (a.getType() == AttributeType.REFERENCE
                || a.getListType() != null && a.getListType() == AttributeType.REFERENCE)) {
          if (opt.isEmpty()) return true;
          if (opt.startsWith("*")) return not ^ key.endsWith(opt.substring(1));
          if (opt.endsWith("*")) return not ^ key.startsWith(opt.substring(0, opt.length() - 1));
          return not ^ key.equals(opt);
        }
      } else if (s.equalsIgnoreCase(key)) return true;
    return false;
  }

  public static Map<String, Attribute> getAttributesInList(final TreeNode node) {
    //TODO move to attribute storage as getVisibleAttributes
    Map<String, Attribute> attrs = new HashMap<>();
    attrs.put("id", new Attribute(new AttributeDefinition(node.getUUId(),"id",
        new ConstValueProvider(AttributeType.STRING,node.getId())),node.attrs()));//TODO remove
    for (String s : node.attrs().getAllAttributeKeys()) {
      Attribute a = node.attrs().getLocalAttribute(s);
      if (a!=null && s.startsWith("_")) /*
									 * && !TreeNode.ATTR_PREDICATE.equals(s)&&
									 * !TreeNode.ATTR_NAME.equals(s) &&
									 * !TreeNode.ATTR_TYPE.equals(s))
									 */ continue;
      if (a!=null && a.getOrigin().equals(node.getUUId())) attrs.put(s, a);
    }
    return attrs;
  }
 
  public static String getAttributes(final TreeNode node) {
    StringBuffer sb = new StringBuffer("");
    for (Attribute a : getAttributesInList(node).values()) {
      sb.append(a.getKey());
      sb.append(": ");
      sb.append(Attribute.valueToString(node.attrs(), a.getType(), a.getValue()));//TODO enum
      sb.append("\n");
    }
    return sb.toString();
  }
  // public static String getAttributesInJSON(final TreeNode node){
  // try {
  // return NodeMapper.toJSON(node.getUUId(),
  // AttributeStorage.toCollection(getAttributesInList(node))).toString();
  // } catch (JSONException e) {
  // e.printStackTrace();
  // }
  // return "";
  // }

  private static QName styleName = new QName("class");

  // private static void recUpdate(org.w3c.dom.Node d1, String type) {
  // NodeList childs = d1.getChildNodes();
  // for(int i = 0;i<childs.getLength();i++){
  // if(childs.item(i))
  // }
  // }
  public static String brTransform(String desc) {
    return desc;//.replace("\n", "<br/>");
  }

  public static String reworkDescription(String desc, String type, String ntv) {
    ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
    Writer bw = new OutputStreamWriter(byteOut);
    XMLEventReader reader;
    LinkedList<String> openned = new LinkedList<>();
    if (desc.isEmpty()) return desc;
    try {
      if (type.startsWith(TextNode.TYPE_NAME) && ntv != null && ntv.equals("in_process"))
        type = "review";
      Reader r = new BufferedReader(new StringReader("<root>" + htmlEncapsulate(desc) + "</root>"));
      reader = XMLInputFactory.newInstance().createXMLEventReader(r);

      XMLEventWriter writer = XMLOutputFactory.newInstance().createXMLEventWriter(bw);
      writer.setDefaultNamespace("");
      writer.setPrefix("", "");
      LinkedList<String> watch = new LinkedList<>();
      XMLEventFactory eventFactory = XMLEventFactory.newInstance();
      while (reader.hasNext() && reader.peek() != null) {
        XMLEvent event = (XMLEvent) reader.next();
        boolean addattr = false;
        if (event.isStartElement()) {
          String tag = event.asStartElement().getName().getLocalPart();
          if (isStyleBreaking(tag) && (!openned.isEmpty() && openned.peekLast().equals("Sspan"))) {
            writer.add(eventFactory.createEndElement("", "", "span"));
            openned.pollLast(); // p?
          }
          if ((event.asStartElement().getName().getLocalPart().equals("p")
                  || event.asStartElement().getName().getLocalPart().equals("pre"))
              || event.asStartElement().getName().getLocalPart().equals("div")) addattr = true;
          openned.addLast(event.asStartElement().getName().getLocalPart());
        } else if (event.isEndElement()) {
          String tag = event.asEndElement().getName().getLocalPart();
          if (!watch.isEmpty() && watch.peek().equals(tag)) {
            if(openned.peek().equals("Sspan")){
              openned.poll();
              writer.add(eventFactory.createEndElement("", "", "span"));
            }
            watch.pollLast(); // span?
          }
          while (!openned.isEmpty() && !openned.peekLast().equals(tag)) {
            String ltag = openned.removeLast();
            if (ltag.equals("Sspan")) ltag = "span";
            writer.add(eventFactory.createEndElement("", "", ltag));
          }
          if (!openned.isEmpty() && openned.getLast().equals(tag)) openned.pollLast();
        } else if (event.isCharacters()) {
          if (!openned.contains("p") && !openned.contains("pre") && !openned.contains("div")) {
            if (!openned.peekLast().equals("Sspan")) { // TODO fix
              writer.add(eventFactory.createStartElement("", "", "span"));
              writer.add(eventFactory.createAttribute("class", type.toLowerCase()));
              writer.add(eventFactory.createAttribute("style", "display:inline;"));
              openned.addLast("Sspan");
            }
          }
        }
        boolean ignore = false;
        if (event.isStartElement()) {
          javax.xml.stream.events.Attribute a =
              event.asStartElement().getAttributeByName(styleName);
          StartElement se = event.asStartElement();
          if (a != null && addattr) {
            ignore = true;
            Iterator in = se.getNamespaces();
            String sa = "";
            if(in.hasNext())
              sa = in.next().toString();
            writer.add(eventFactory.createStartElement("","", se.getName().getLocalPart()));
            @SuppressWarnings("unchecked")
            Iterator<Object> i = se.getAttributes();
            String rem = null;
            while (i.hasNext()) {
              javax.xml.stream.events.Attribute e = (javax.xml.stream.events.Attribute) i.next();
              if (e.getName().getLocalPart().toLowerCase().equals("class")) {
                String s = e.getValue();
                if (!s.contains(type.toLowerCase())) {
                  rem = s;
                  s = type.toLowerCase();
                }
                writer.add(eventFactory.createAttribute("class", s));
              } else
                writer.add(eventFactory.createAttribute(e.getName().getLocalPart(), e.getValue()));
            }
            if (rem != null) {
              writer.add(eventFactory.createStartElement("", "", "span"));
              openned.addLast("Sspan");
              writer.add(eventFactory.createAttribute("class", rem));
              watch.add(se.getName().getLocalPart());
            }
          }
        }
        if (!ignore) {
          try{
            if(event.isStartDocument()){
              writer.add(eventFactory.createStartDocument());
            }else
            if(event.isStartElement())
            {
              StartElement se = (StartElement)event;
              writer.add(eventFactory.createStartElement("","", se.getName().getLocalPart()));
              @SuppressWarnings("unchecked")
              Iterator<Object> i = se.getAttributes();
              String rem = null;
              while (i.hasNext()) {
                javax.xml.stream.events.Attribute e = (javax.xml.stream.events.Attribute) i.next();
                writer.add(eventFactory.createAttribute(e.getName().getLocalPart(), e.getValue()));
              }
            }else if(event.isEndElement()){
              EndElement ee = (EndElement)event;
              writer.add(eventFactory.createEndElement("","", ee.getName().getLocalPart()));
            }else
              writer.add(event);
          }catch(XMLStreamException xse){
            writer.close();
            System.out.println(desc );
            xse.printStackTrace();
          }
          if (addattr) writer.add(eventFactory.createAttribute("class", type.toLowerCase()));
        }
      }
      writer.close();
      reader.close();
      byteOut.close();
      writer = null;
      eventFactory = null;
      return byteOut
          .toString()
          .replace("<root>", "")
          .replace("</root>", "")
          .replace("<?xml version=\"1.0\" ?>", "")
          .replace("<br></br>", "<br>");
    } catch (FactoryConfigurationError e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (XMLStreamException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return "";
  }

  private static boolean isStyleBreaking(String tag) {
    return tag.equals("li")
        || tag.equals("ol")
        || tag.equals("span")
        || tag.equals("td")
        || tag.equals("font");
  }

  public static String getLvlTire(int lvl, String s2) {
    StringBuilder sb = new StringBuilder();
    for (int i = 0; i < lvl - 1; i++) sb.append(s2);
    return sb.toString();
  }

  public static String htmlEscape(String str) {
    return CoreUtils.htmlEscape(str);
  }

  public static String htmlSpaces(String str) {
    return CoreUtils.htmlSpaces(str);
  }

  public static String htmlEncapsulate(String str) {
    if (str == null) return "";
    return CoreUtils.htmlEncapsulate(str,true);
  }

  public static String replaceSlashN(String str) {
    if (str == null) return null;
    return str;//.replace("\n", "<br>");
  }

  public Object getDocumentNodeModel(String id) {
    Document d = (Document) db.getNode(id, Path.QID);
    if (d == null) return null;
    return d.getNodeModel();
  }

  public TreeNode[] getDocuments() {
    return getDocuments(root.getQualifiedId());
  }

  private Map<String, TreeNode[]> docsmap = new HashMap<>();

  public TreeNode[] getDocuments(String qid) {
    if (qid == null)
      return db.getNode(Document.getTypeRootQId(), Path.QID)
          .getSortedWorkChildren(Document.TYPE_NAME);
    if (!docsmap.containsKey(qid)) {
      Set<TreeNode> docs = new HashSet<>();
      try {
        DeepFirstTreeWalker dftw =
            new DeepFirstTreeWalker(db.getNode(qid, Path.QID), true, true, false);
        Iterator<TreeNode> iter = dftw.iterator();
        while (iter.hasNext()) {
          TreeNode node = iter.next();
          if (node.getType().equals(Requirement.TYPE_NAME))
            for (Location loc : ((Requirement) node).getLocations()) {
              Document doc = loc.getDocument();
              if (doc != null) docs.add(loc.getDocument());
            }
        }
        docsmap.put(qid, docs.toArray(new TreeNode[0]));
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    if (docsmap.containsKey(qid)) return docsmap.get(qid);
    return null;
  }

  public TreeNode getNodeByUUID(String uid) {
    try {
      UUID uuid = UUID.fromString(uid);
      return root.getTreeDB().getNode(uuid);
    } catch (Exception e) {
      return null;
    }
  }

  public String getFullName(TreeNode node) {
    return Requirement.getFullName(node, node.getTreeDB(), false);
  }
}
