package com.unitesk.requality.report;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.UUID;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class TraceabilityDataSource implements IReportDataSource {

  private static Map<TreeDB, Map<String, String>> revert_names = null;
  private static String LINK_REL = "reference";
  private static String PARENT_REL = "parent";
  private String[] column, row;

  private Map<String, Integer> ind1;
  private HashSet<String> flist;
  private ListValue settings;
  private Map<String, Map<String, Map<String, Boolean>>> edgesList =
      new TreeMap<>();
  private Map<String, Map<String, Map<String, Boolean>>> redgesList =
      new TreeMap<>();
  private Map<String, TreeNode> edjetonode = new HashMap<>();
  private Map<String, Object> attrs;
  public static String[] defaultTemplates =
      new String[] {"<b>[1]</b> [3]", "<div style='margin-left:20px'>[2] [4]</div>"};
  public static String[][] defaultSettings =
      new String[][] {
        new String[] {"{user-visible-name}", "", "", ""},
        new String[] {
          "{user-visible-name}",
          "{q-page}#{value}",
          "<span style='color: red;'>{value}</span>",
          "{user-visible-name}"
        },
        new String[] {"{revert-attribute-name}", "{next-page}#{value}", "", ""},
        new String[] {"{revert-attribute-name}", "{next-page}#{value}", "", ""}
      };

  public TreeNode getNode(String path) {
    return edjetonode.getOrDefault(path, null);
  }

  public static void setRevertNames(TreeDB db, Map<String, String> mp) {
    if (revert_names == null) revert_names = new HashMap<>();
    revert_names.put(db, mp);
  }

  @Override
  public Map<String, Object> getData() {
    Map<String, Object> ret = new TreeMap<>();
    ret.put("trace", this);
    return ret;
  }

  public String getDirectName(boolean firstPage, TreeNode req, Object filter) {
    if (!firstPage) return getRevertName(true, req, filter);
    return filter.toString();
  }

  public String getRevertName(boolean firstPage, TreeNode req, Object filter) {
    if (!firstPage) return getDirectName(true, req, filter); // TODO without root?
    return getRevertName(Instantiator.getRootTree(req.getTreeDB()), filter.toString());
  }

  private void addEdge(String treeNode, String treeNode2, String a, Boolean correctness) {
    if (!edgesList.containsKey(treeNode))
      edgesList.put(treeNode, new TreeMap<String, Map<String, Boolean>>());
    if (!edgesList.get(treeNode).containsKey(treeNode2))
      edgesList.get(treeNode).put(treeNode2, new TreeMap<String, Boolean>());
    edgesList.get(treeNode).get(treeNode2).put(a, correctness);

    if (!redgesList.containsKey(treeNode2))
      redgesList.put(treeNode2, new TreeMap<String, Map<String, Boolean>>());
    if (!redgesList.get(treeNode2).containsKey(treeNode))
      redgesList.get(treeNode2).put(treeNode, new TreeMap<String, Boolean>());
    redgesList.get(treeNode2).get(treeNode).put(a, correctness);
  }

  private String formGroupValue(
      boolean firstPage, int num, String cname, TreeNode rootReq, Object filter) {
    TreeNode node = null;
    try {
      node = rootReq.getTreeDB().getNode(cname, Path.QID | Path.UV | Path.UUID);
    } catch (Exception e) {
    }
    String link = groups[num][1];
    String value = groups[num][0];
    if (node == null) {
      value = groups[num][2];
    } else if (!groups[num][1].equals("")) {
      // value needs to be calculated first and then ' character needs to
      // be replaced
      // {user-visible-name}","{id}","{name-or-id}","{attribute-name}"},
      // links = new
      // String[]{"","{this-page}","{next-page}","{requality-cnf}
      link = getLinkValue(node, firstPage, rootReq, filter, cname, link);
      if (link != null)
        if (link.startsWith("requality://"))
          value =
              "<a href=\"#\" onclick='selectNode(\""
                  + link.substring(12)
                  + "\")'>"
                  + value.replace("'", "&#8216;")
                  + "</a>";
        else value = "<a href='" + link + "'>" + value.replace("'", "&#8216;") + "</a>";
    }
    if (link == null && groups[num].length > 3) value = groups[num][3];
    if (value.equals("{skip-line}")) return null;
    return value;
  }

  private String getLinkValue(
      TreeNode node,
      boolean firstPage,
      TreeNode rootReq,
      Object filter,
      String cname,
      String link) {
    link =
        link.replace("{id}", node.getId())
            .replace("{user-visible-name}", node.getUserVisibleName(false))
            .replace("{attribute-name}", getDirectName(firstPage, rootReq, filter))
            .replace("{value}", cname)
            .replace("{revert-attribute-name}", getRevertName(firstPage, rootReq, filter))
            .replace("{name-or-id}", node.getNameOrId())
            .replace("{this-page}", rootReq.getTreeDB().getResName()+"-"+(firstPage ? "trace.htm" : "tracerev.htm"))
            .replace("{next-page}", rootReq.getTreeDB().getResName()+"-"+(!firstPage ? "trace.htm" : "tracerev.htm"))
            .replace("{requality-cnf}", "requality://");
    if (link.startsWith("trace.htm#") && !hasTraces(link.substring(10))) link = null;
    else if (link.startsWith("tracerev.htm#") && !hasRevertTraces(link.substring(13))) link = null;
    return link;
  }

  public String formFirstString(boolean firstPage, String rname, TreeNode rootReq, Object filter) {
    return "<a id='"
        + rname
        + "'/>"
        + handleTemplate(templatemain, firstPage, filter, rname, rootReq, filter);
  }

  public String formSecondString(boolean firstPage, String rname, TreeNode rootReq, Object filter) {
    return handleTemplate(templateadditional, firstPage, filter, rname, rootReq, filter);
  }

  private String handleTemplate(
      String templ,
      boolean firstPage,
      Object filter,
      String rname,
      TreeNode rootReq,
      Object filter2) {
    boolean stop = false;
    while (!stop) {
      stop = true;
      for (int i = 1; i <= groups.length; i++) {
        if (templ.contains("[" + i + "]")) {
          stop = false;
          String gv = formGroupValue(firstPage, i - 1, rname, rootReq, filter);
          if (gv == null) return "";
          templ = templ.replace("[" + i + "]", gv);
        }
      }
    }
    templ =
        templ
            .replace("{attribute-name}", getDirectName(firstPage, rootReq, filter))
            .replace("{value}", rname)
            .replace("{revert-attribute-name}", getRevertName(firstPage, rootReq, filter))
            .replace("{this-page}", rootReq.getTreeDB().getResName()+"-"+(firstPage ? "trace.htm" : "tracerev.htm"))
            .replace("{next-page}", rootReq.getTreeDB().getResName()+"-"+(!firstPage ? "trace.htm" : "tracerev.htm"));
    TreeNode node = null;
    try {
      node = rootReq.getTreeDB().getNode(rname, Path.QID | Path.UV | Path.UUID);
      if (node != null) {
        templ =
            templ
                .replace("{user-visible-name}", node.getUserVisibleName(false))
                .replace("{name-or-id}", node.getNameOrId())
                .replace("{id}", node.getId());
        for (String s : node.attrs().getAllAttributeKeys()) {
          if(templ.indexOf("{@" + s + "}")>-1){
            Object o = node.attrs().getAttributeValue(s);
            templ = templ.replace("{@" + s + "}", o.toString());
          }
        }
      }
    } catch (Exception e) {
    }
    if (attrs != null)
      for (String s : attrs.keySet()) {
        Object o = attrs.get(s);
        if (o != null) templ = templ.replace("{#" + s + "}", o.toString());
      }

    return templ;
  }

  public String[] getTraces(String key) {
    if (key.equals("*")) {
      Set<String> set = new HashSet<>();
      for (String s : edgesList.keySet()) set.addAll(edgesList.get(s).keySet());
      return set.toArray(new String[0]);
    }
    if (!edgesList.containsKey(key)) return new String[0];
    return edgesList.get(key).keySet().toArray(new String[0]);
  }

  public boolean getCorrectness(String key, String key2) {
    if (!edgesList.containsKey(key)) return false;
    if (!edgesList.get(key).containsKey(key2)) return false;
    boolean k = true;
    for (String s : edgesList.get(key).get(key2).keySet()) {
      k = k && edgesList.get(key).get(key2).get(s);
      if (!k) break;
    }
    return k;
  }

  public String[] getRevertTraces(String key) {
    if (key.equals("*")) {
      Set<String> vals = new HashSet<>();
      for (String s : redgesList.keySet()) vals.addAll(redgesList.get(s).keySet());
      return vals.toArray(new String[0]);
    }
    if (!redgesList.containsKey(key)) return new String[0];
    return redgesList.get(key).keySet().toArray(new String[0]);
  }

  private String templatemain = defaultTemplates[0], templateadditional = defaultTemplates[1];
  private String[][] groups = defaultSettings;

  public boolean hasReqLinked(TreeNode node, String aname, String aval) {
    Set<TreeNode> set = new HashSet<>();
    return hasReqLinked(set, node, aname, aval);
  }

  public boolean hasReqLinked(Set<TreeNode> washere, TreeNode node, String aname, String aval) {
    washere.add(node);
    for (TreeNode nde : node.getRelatedNodes(null)) {
      if (washere.contains(nde)) continue;
      Attribute a = nde.attrs().getAttribute(aname);
      if (a == null
          || a.getValue() == null
          || !a.getValue().toString().equals(aval)
          || hasReqLinked(washere, nde, aname, aval)) return true;
    }
    for (String s : node.attrs().getAllAttributeKeys()) {
      Attribute a = node.attrs().getLocalAttribute(s);
      if (a != null && a.getType().equals(AttributeType.REFERENCE)) {
        List<TreeNode> rels = node.attrs().getLinkAttributeValue(s);
        if (rels != null) {
          for (TreeNode rel : rels) {
            if (rel == null || washere.contains(rel)) continue;
            Attribute b = rel.attrs().getAttribute(aname);
            if (b == null
                || b.getValue() == null
                || !b.getValue().toString().equals(aval)
                || hasReqLinked(washere, rel, aname, aval)) return true;
          }
        }
      }
    }
    return false;
  }

  public boolean linksCorrects(TreeNode node, String aname, String aval) {
    Attribute a = node.attrs().getAttribute(aname);
    if (a != null) {
      if (a.getValue() != null
          && a.getValue().toString().toLowerCase().equals(aval.toLowerCase())) {
        return !hasReqLinked(node, aname, aval);
      }
    }
    return true;
  }

  @SuppressWarnings("unchecked")
  public void initData(
      TreeNode root,
      String type1,
      String type2,
      String what,
      Object filter,
      Object setts,
      Object p) {
    attrs = (Map<String, Object>) p;
    settings = (ListValue) setts;
    if (settings != null) {
      int i = 0;
      List<Object> l = ((ListValue)settings.getValues().get(0)).getValues();
      if (l.size() == 2) {
        templatemain = l.get(0).toString();
        templateadditional = l.get(1).toString();
        i = 1;
      }
      int k = 0;
      groups = new String[settings.getValues().size() - i][];
      for (; i < settings.getValues().size(); i++) groups[k++] = ((ListValue)(settings.getValues().get(i))).getValues().toArray(new String[0]);
    }
    edjetonode.clear();
    edgesList.clear();
    redgesList.clear();
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(root);
    Iterator<TreeNode> it = dftw.iterator();
    ind1 = new TreeMap<>();
    List<TreeNode> t1 = new ArrayList<>(), t2 = new ArrayList<>();
    Set<TreeNode> toh = new HashSet<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      String val = node.getUserVisibleName(false);
      if (type1 == null || type1.equals(node.getType())) {
        ind1.put(val, t1.size());
        t1.add(node);
      }
    }
    column = new String[t1.size()];
    row = new String[t2.size()];
    int i = 0;
    for (TreeNode s2 : t2) row[i++] = s2.getUserVisibleName(false);
    flist = null;
    if (filter != null) {
      if (filter instanceof String) {
        flist = new HashSet<>();
        flist.add((String) filter);
      }
      if (filter instanceof List) flist = new HashSet<>((List<String>) filter);
    }
    if (what.equals(PARENT_REL)) {
      flist = new HashSet<>();
      flist.add("+");
    }
    i = 0;
    if (what.equals(PARENT_REL))
      for (TreeNode s1 : t1) {
        edjetonode.put(s1.getUserVisibleName(false), s1);
        column[i] = s1.getUserVisibleName(false);
        if (t1.contains(s1.getShownParent())) {
          edjetonode.put(s1.getShownParent().getUserVisibleName(false), s1.getShownParent());
          addEdge(s1.getShownParent().getUserVisibleName(false), column[i], "+", true);
        }
        i++;
      }
    else if (what.equals(LINK_REL)) {
      Set<UUID> handled = new HashSet<>();
      LinkedList<TreeNode> handl = new LinkedList<>(t1);
      while (!handl.isEmpty()) {
        toh.clear();
        Iterator<TreeNode> sit = handl.iterator();
        while (sit.hasNext()) {
          TreeNode s1 = sit.next();
          edjetonode.put(s1.getUserVisibleName(false), s1);
          if (handled.contains(s1.getUUId())) continue;
          handled.add(s1.getUUId());
          column[i] = s1.getUserVisibleName(false);
          for (Attribute at : s1.attrs()) {
            String s = at.getKey();
            boolean goodAttribute = false;
            if (at.getType().equals(AttributeType.REFERENCE)
                && (flist == null || flist.contains(s) || flist.contains("*")))
              goodAttribute = true;
            if (at.getType() == AttributeType.LIST
                && at.getListType() == AttributeType.REFERENCE) {
              if (flist == null || flist.contains(s) || flist.contains("*")) goodAttribute = true;
            }
            if (goodAttribute) {
              List<TreeNode> res = s1.attrs().getLinkAttributeValues(s);
              if (res != null && res.size() > 0)
                for (TreeNode node : res) {
                  addEdge(column[i], node.getUserVisibleName(false), s, true);
                  edjetonode.put(node.getUserVisibleName(false), node);
                  if (Instantiator.getRootTree(node.getTreeDB())
                      != Instantiator.getRootTree(s1.getTreeDB())) {
                    if (!handled.contains(node.getUUId())) toh.add(node);
                  }
                }
              else {
                String val = null;
                if(at.getValue() instanceof ReferenceValue)
                  val = 
                    ((ReferenceValue)at.getValue()).toString(s1.attrs())
                        .toString()
                        .replace("\n", "")
                        .replace((char) 160, ' ')
                        .trim();
                else
                  val = at.getValue().toString().replace("\n", "")
                      .replace((char) 160, ' ')
                      .trim();
                if (val != null && !val.isEmpty()) addEdge(column[i], val, s, true);
              }
            }
          }
          i++;
        }
        handl = new LinkedList<>(toh);
        column = Arrays.copyOf(column, column.length + toh.size());
      }
    }
    Arrays.sort(column);
  }

  public boolean hasTrace(String p1, String p2) {
    return edgesList.containsKey(p1) && edgesList.get(p1).containsKey(p2);
  }

  public boolean hasRevertTrace(String p1, String p2) {
    return redgesList.containsKey(p1) && redgesList.get(p1).containsKey(p2);
  }

  public String[] getColumnNames() {
    return column;
  }

  public String[] getRowNames() {
    return row;
  }

  public boolean hasTraces(String name) {
    return edgesList.containsKey(name) && edgesList.get(name).size() > 0;
  }

  public boolean hasRevertTraces(String name) {
    return redgesList.containsKey(name) && redgesList.get(name).size() > 0;
  }

  public String loadNames(String s1, String s2) {
    return (flist == null || flist.size() > 1)
        ? Arrays.toString(edgesList.get(s1).get(s2).keySet().toArray())
        : "";
  }

  public static String getRevertName(TreeDB db, String name) {
    if (revert_names != null
        && revert_names.containsKey(db)
        && revert_names.get(db).containsKey(name)) return revert_names.get(db).get(name);
    return name;
  }
}
