package com.unitesk.requality.tests;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.commons.httpclient.URIException;
import org.apache.commons.httpclient.util.URIUtil;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;

public class CoverageStorage {
  private TreeDB relatedTreeDB;
  private ICoverageStorageSource source;

  @SuppressWarnings("unused")
  private String lastErrors;

  private Map<String, List<String>> testtoreqs = new HashMap<>();
  public static Map<String, TreeNodeCoverageInfo> coverageInfos =
      new HashMap<>();
  public static Map<String, TestErrorInfo[]> errorsInfos = new HashMap<>();

  public void addTest(String qid, String testPath, int cnt) {
    qid = getQualifyingId(qid);
    if (!containsInfoAbout(qid)) coverageInfos.put(qid, new TreeNodeCoverageInfo());
    TreeNodeCoverageInfo cinfo = coverageInfos.get(qid);
    cinfo.registerTest(testPath, cnt);
  }

  public Object getCoverageSourceObject() {
    return source.getSourceObject(relatedTreeDB, new HashMap<String, Object>());
  }

  public void setInputSource(ICoverageStorageSource s) {
    source = s;
    coverageInfos = null;
    errorsInfos = null;
    testtoreqs = new HashMap<>();
    lastErrors = null;
    classes = new HashMap<>();
    s.setTargetStorage(this);
  }

  public boolean loadStorage() {
    try {
      source.setTargetStorage(this);
      coverageInfos = source.getCoverage();
      errorsInfos = source.getErrors();
      if (coverageInfos == null
          || source.getLastErrors() != null && !source.getLastErrors().equals("")) return false;
      for (String s : coverageInfos.keySet()) {
        TreeNodeCoverageInfo ci = getCoverageInfo(s);
        if (ci != null)
          for (String path : ci.getTestPaths()) {
            if (!testtoreqs.containsKey(path)) testtoreqs.put(path, new ArrayList<String>());
            testtoreqs.get(path).add(s);
          }
      }
      return true;
    } catch (Exception e) {
      RLogger.logError(CoreUtils.getTrace(e));
      lastErrors = e.toString();
    }
    return false;
  }

  public boolean containsInfoAbout(String qid) {
    qid = getQualifyingId(qid);
    return coverageInfos.containsKey(qid);
  }

  public TreeNodeCoverageInfo getCoverageInfo(String qid) {
    qid = getQualifyingId(qid);
    return coverageInfos.get(qid);
  }

  private String getQualifyingId(String qid) {
    try {
      return relatedTreeDB.getNode(qid, Path.QID | Path.UV | Path.UUID).getQualifiedId(true);
    } catch (Exception e) {
      return qid;
    }
  }

  public boolean isCovered(String qid) {
    qid = getQualifyingId(qid);
    if (coverageInfos == null) return false;
    TreeNodeCoverageInfo tci = coverageInfos.get(qid);
    return tci != null && tci.getTestCounts().length > 0;
  }

  public int getCoverageTestCount(String qid) {
    qid = getQualifyingId(qid);
    TreeNodeCoverageInfo tci = coverageInfos.get(qid);
    if (tci == null) return 0;
    return tci.getTestCounts().length;
  }

  public String[] getCoveragePaths(String qid) {
    qid = getQualifyingId(qid);
    TreeNodeCoverageInfo tci = coverageInfos.get(qid);
    if (tci == null) return new String[0];
    String[] paths = tci.getTestPaths();
    return paths;
  }

  private Map<TreeNode, Integer> classes = new HashMap<>();

  private void calculateClasses(TreeNode node) {
    TreeNode[] children = node.getWorkChildren();
    if (isCovered(node.getQualifiedId())) {
      classes.put(node, 2);
      return;
    }
    int cnt = children.length, cntyellow = 0, cntred = 0;
    for (TreeNode child : children) {
      int c2 = getCoverageClass(child);
      cntred += c2 == 0 ? 1 : 0;
      cntyellow += c2 == 1 ? 1 : 0;
    }
    classes.put(
        node, cnt == 0 ? 0 : cntred > 0 ? (cnt - cntred > 0) ? 1 : 0 : cntyellow > 0 ? 1 : 2);
  }

  public int getCoverageClass(Object obj) {
    if (!(obj instanceof TreeNode)) return 0;
    TreeNode node = (TreeNode) obj;
    if (!classes.containsKey(node)) calculateClasses(node);
    return classes.get(node);
  }

  public Integer[] getCoverageCounts(String qid) {
    qid = getQualifyingId(qid);
    TreeNodeCoverageInfo tci = coverageInfos.get(qid);
    if (tci == null) return new Integer[0];
    return tci.getTestCounts();
  }

  public OperationResult saveCoverageToFile(String filename) {
    String lastErrors = "";
    try {
      File f = new File(URIUtil.encodeQuery(filename, "UTF-8"));
      if(!f.exists())
        try {
          f.mkdirs();
          f.createNewFile();
        } catch (IOException e1) {
          // TODO Auto-generated catch block
          RLogger.logError(CoreUtils.getTrace(e1));
        }
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      DOMImplementation impl = builder.getDOMImplementation();
      Document document = impl.createDocument(null, null, null);
      Element e = document.createElement("coverageInfo");
      document.appendChild(e);
      if (coverageInfos == null) loadStorage();
      if (coverageInfos != null)
        for (String s : coverageInfos.keySet()) {
          Element e2 = document.createElement("reqcoverage");
          e.appendChild(e2);
          TreeNodeCoverageInfo inf = coverageInfos.get(s);
          String[] paths = inf.getTestPaths();
          Integer[] cnts = inf.getTestCounts();
          e2.setAttribute("qid", s);
          if (paths.length == 0) {
            e2.setAttribute("hits", Integer.toString(inf.getCntAll()));
          } else {
            for (int i = 0; i < paths.length; i++) {
              Element k = document.createElement("covered_by");
              e2.appendChild(k);
              k.setAttribute("uri", paths[i]);
              k.setAttribute("hits", Integer.toString(cnts[i]));
              // TODO protocoles
              // if(localstate[i])
              // k.setAttribute("local",
              // Boolean.toString(localstate[i]));
            }
          }
        }
      if (errorsInfos != null)
        for (String s : errorsInfos.keySet()) {
          TestErrorInfo[] infs = errorsInfos.get(s);
          for (TestErrorInfo inf : infs) {
            Element e2 = document.createElement("error");
            e.appendChild(e2);
            String name = inf.getName();
            String desc = inf.getDescription();
            String path = inf.getRawTestPath();
            String[] qids = inf.getRequirementsQIDs();
            String uri = inf.getURI();
            e2.setAttribute("name", name);
            e2.setAttribute("testuri", path);
            if (uri != null && uri != "") e2.setAttribute("uri", uri);
            if (desc != null && !desc.equals("")) {
              Element d = document.createElement("description");
              e2.appendChild(d);
              d.setAttribute("format", inf.isPlainText() ? "plain" : "html");
              d.setTextContent(desc);
            }
            for (String qid : qids) {
              Element k = document.createElement("violates");
              e2.appendChild(k);
              k.setAttribute("uri", qid);
            }
          }
        }

      Source source = new DOMSource(e);
      // Prepare the output file
      Result result = new StreamResult(f.getPath());
      // Write the DOM document to the file
      Transformer xformer = TransformerFactory.newInstance().newTransformer();
      xformer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
      xformer.setOutputProperty(OutputKeys.METHOD,"xml");
      xformer.transform(source, result);
      this.lastErrors = lastErrors;
      return new OperationResult(lastErrors, true);
    } catch (ParserConfigurationException e) {
      lastErrors += CoreUtils.htmlEscape(e.getMessage()) + "<br>";
    } catch (TransformerConfigurationException e1) {
      lastErrors += CoreUtils.htmlEscape(e1.getMessage()) + "<br>";
    } catch (TransformerFactoryConfigurationError e1) {
      lastErrors += CoreUtils.htmlEscape(e1.getMessage()) + "<br>";
    } catch (TransformerException e1) {
      lastErrors += CoreUtils.htmlEscape(e1.getMessage()) + "<br>";
    } catch (URIException e3) {
      // TODO Auto-generated catch block
      e3.printStackTrace();
    } finally {
    }
    this.lastErrors = lastErrors;
    return new OperationResult(lastErrors, false);
  }

  public String getLastErrors() {
    if (source != null) return source.getLastErrors();
    return "";
  }

  public TestErrorInfo[] getErrorsForPath(String path) {
    if (!errorsInfos.containsKey(path)) return null;
    return errorsInfos.get(path);
  }

  public String[] getErrorLinksForNodeAndTest(String qid, String test) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNode(qid);
    if (tei == null) return null;
    String[] links = new String[tei.length];
    int i = 0;
    for (TestErrorInfo t : tei) {
      String uri = t.getURI();
      if (uri == null) uri = "";
      links[i++] = uri;
    }
    return links;
  }

  public String[] getErrorNamesForNode(String qid) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNode(qid);
    if (tei == null) return null;
    String[] names = new String[tei.length];
    int i = 0;
    for (TestErrorInfo t : tei) names[i++] = t.getName();
    return names;
  }

  public String[] getErrorNamesForNodeAndTest(String qid, String test) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNodeAndTest(qid, test);
    if (tei == null) return null;
    String[] names = new String[tei.length];
    int i = 0;
    for (TestErrorInfo t : tei) names[i++] = t.getName();
    return names;
  }

  public boolean hasErrors(String qid) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNode(qid);
    if (tei == null || tei.length == 0) return false;
    return tei.length > 0;
  }

  public boolean hasErrors(String qid, String test) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNodeAndTest(qid, test);
    if (tei == null || tei.length == 0) return false;
    return tei.length > 0;
  }

  public String[] getErrorDescriptionsForNode(String qid) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNode(qid);
    if (tei == null) return null;
    String[] descs = new String[tei.length];
    int i = 0;
    for (TestErrorInfo t : tei) descs[i++] = t.getDescription();
    return descs;
  }

  public String[] getErrorDescriptionsForNodeAndTest(String qid, String test) {
    qid = getQualifyingId(qid);
    TestErrorInfo[] tei = getErrorsByNodeAndTest(qid, test);
    if (tei == null) return null;
    String[] descs = new String[tei.length];
    int i = 0;
    for (TestErrorInfo t : tei) descs[i++] = t.getDescription();
    return descs;
  }

  public TestErrorInfo[] getErrorsByNodeAndTest(String qid, String tst) {
    qid = getQualifyingId(qid);
    TreeNodeCoverageInfo cnci = getCoverageInfo(qid);
    if (cnci == null) return null;
    Set<TestErrorInfo> list = new HashSet<>();
    TestErrorInfo[] infos = getErrorsForPath(tst);
    if (infos != null)
      for (TestErrorInfo tei : infos) {
        String[] reqs = tei.getRequirementsQIDs();
        for (String r : reqs) {
          TreeNode node = relatedTreeDB.getNode(r, Path.QID | Path.UV);
          if (node != null && node.getQualifiedId().equals(qid)) list.add(tei);
        }
      }
    return list.toArray(new TestErrorInfo[0]);
  }

  public TestErrorInfo[] getErrorsByNode(String qid) {
    qid = getQualifyingId(qid);
    TreeNodeCoverageInfo cnci = getCoverageInfo(qid);
    if (cnci == null) return null;
    String[] tests = cnci.getTestPaths();
    Set<TestErrorInfo> list = new HashSet<>();
    for (String path : tests) {
      TestErrorInfo[] infos = getErrorsForPath(path);
      if (infos == null) continue;
      for (TestErrorInfo tei : infos) {
        String[] reqs = tei.getRequirementsQIDs();
        for (String r : reqs) {
          TreeNode node = relatedTreeDB.getNode(r, Path.QID | Path.UV);
          if (node != null && node.getQualifiedId().equals(qid)) list.add(tei);
        }
      }
    }
    return list.toArray(new TestErrorInfo[0]);
  }

  public CoverageStorage(TreeDB where) {
    relatedTreeDB = where;
    if (where == null) defStorage = this;
    else map.put(where, this);
  }

  private static CoverageStorage defStorage;
  private static Map<TreeDB, CoverageStorage> map = new HashMap<>();

  public static CoverageStorage getStorage(TreeDB treeDB, ICoverageStorageSource src) {
    if (treeDB == null) {
      if (defStorage == null) defStorage = new CoverageStorage(treeDB);
      return defStorage;
    }
    if (!map.containsKey(treeDB)) map.put(treeDB, new CoverageStorage(treeDB));
    map.get(treeDB).setInputSource(src);
    return map.get(treeDB);
  }

  public void clearStorage() {
    classes = new HashMap<>();
    coverageInfos.clear();
    errorsInfos.clear();
  }

  public void setSettings(String[] split) {
    if (source == null) return;
    source.setSettings(split);
  }

  public String[] getSettings(TreeNode node) {
    if (source == null) return null;
    return source.getSettings(node);
  }

  public String[] getQIDsForPath(String uri) {
    if (!testtoreqs.containsKey(uri)) return null;
    return testtoreqs.get(uri).toArray(new String[0]);
  }

  public TreeDB getTreeDB() {
    return relatedTreeDB;
  }

  public ICoverageStorageSource getCoverageSource() {
    return source;
  }
}
