/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.tools.dom;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import com.unitesk.requality.tools.RequalityCLI;

public class XHTMLResourceCollector {
  private final IXHTMLNode root;
  private final String src;
  private final String dst;
  private List<String> procesedResoureces = new ArrayList<>();

  public XHTMLResourceCollector(IXHTMLNode root, String srcHtmlFolder, String dst) {
    this.root = root;
    this.src = srcHtmlFolder;
    this.dst = dst;
  }

  public Map<String, String> process() {
    return walk(root);
  }

  private Map<String, String> walk(IXHTMLNode node) {
    Map<String, String> ret = new HashMap<>();
    ret.putAll(proccess(node));
    if (node != null)
      for (IXHTMLNode child : node.getChildren()) {
        ret.putAll(walk(child));
      }
    return ret;
  }

  private Map<String, String> proccess(IXHTMLNode node) {
    Map<String, String> ret = new HashMap<>();
    try {

      if (node == null) return Collections.emptyMap();
      if (node.getName().toLowerCase().equals("img")) {
        if (node.getAttributeValue("src") != null
            && !procesedResoureces.contains(node.getAttributeValue("src"))) {
          String old = node.getAttributeValue("src");
          procesedResoureces.add(old);
          try {
            String newPath = copyResource(node.getAttributeValue("src"), "imgs");
            ret.put(old, newPath);
            node.setAttribute("src", newPath);
          } catch (Exception e) {
            System.err.println(e.getMessage());
          }
        }
        if (node.getAttributeValue("alt") != null
            && !procesedResoureces.contains(node.getAttributeValue("alt"))) {
          String old = node.getAttributeValue("alt");
          procesedResoureces.add(old);
          try {
            String newPath = copyResource(node.getAttributeValue("alt"), "imgs");
            ret.put(old, newPath);
            node.setAttribute("alt", newPath);
          } catch (Exception e) {
            System.err.println(e.getMessage());
          }
        }
        if (node.getAttributeValue("alt") != null
            && !procesedResoureces.contains(node.getAttributeValue("alt"))) {
          String old = node.getAttributeValue("alt");
          procesedResoureces.add(old);
          try {
            String newPath = copyResource(node.getAttributeValue("alt"), "imgs");
            ret.put(old, newPath);
            node.setAttribute("alt", newPath);
          } catch (Exception e) {
            System.err.println(e.getMessage());
          }
        }
      } else if (node.getName().toLowerCase().equals("object")) {
        if (node.getAttributeValue("data") != null
            && !procesedResoureces.contains(node.getAttributeValue("data"))) {
          String old = node.getAttributeValue("data");
          procesedResoureces.add(old);
          try {
            String newPath = copyResource(node.getAttributeValue("data"), "imgs");
            ret.put(old, newPath);
            node.setAttribute("data", newPath);
          } catch (Exception e) {
            System.err.println(e.getMessage());
          }
        }
      } else if (node.getName().toLowerCase().equals("link")) {
        if ("stylesheet".equals(node.getAttributeValue("rel"))
            && node.getAttributeValue("href") != null
            && !procesedResoureces.contains(node.getAttributeValue("href"))) {
          String old = node.getAttributeValue("href");
          procesedResoureces.add(old);
          String newPath;
          newPath = copyResource(node.getAttributeValue("href"), "css");
          ret.put(old, newPath);
          node.setAttribute("href", newPath);
        }
      } else if (node.getName().toLowerCase().equals("script")) {
        if (node.getAttributeValue("src") != null
            && !procesedResoureces.contains(node.getAttributeValue("src"))) {
          String old = node.getAttributeValue("src");
          procesedResoureces.add(old);
          String newPath = copyResource(node.getAttributeValue("src"), "js");
          ret.put(old, newPath);
          node.setAttribute("src", newPath);
        }
      }

    } catch (Exception e) {
      e.printStackTrace();
    }
    return ret;
  }

  private String copyResource(String encodedDocImgSrc, String subfolder) throws Exception {
    String docImgSrc = java.net.URLDecoder.decode(encodedDocImgSrc, "utf-8");
    String imgNewPath = dst + "/" + subfolder + "/";
    String dstFile = getDstFileName(imgNewPath, new File(docImgSrc).getName());
    new File(imgNewPath).mkdirs();

    if (docImgSrc.startsWith("http://") || docImgSrc.startsWith("https://")) {
      URL srcUrl = new URL(docImgSrc);
      InputStream in = srcUrl.openConnection().getInputStream();
      try{
        FileOutputStream out = new FileOutputStream(URLDecoder.decode(dstFile, "utf-8"));
        int inp;
        while ((inp = in.read()) != -1) {
          out.write(inp);
        }
        out.close();
      }
      finally{
        in.close();
      }
    } else if (docImgSrc.startsWith("file:/")) {
      RequalityCLI.copyfile(null, new URI(encodedDocImgSrc), new File(URI.create(dstFile)).toURI());
    } else {
      String sourcePath = src + "/" + docImgSrc;
      File sf = new File(sourcePath);
      if(!sf.exists())
        sf = new File(docImgSrc);
      RequalityCLI.copyfile(null, sf.toURI(), new File(URI.create(dstFile)).toURI());
    }
    return subfolder + "/" + new File(URI.create(dstFile)).getName();
  }

  private String getDstFileName(String imgNewPath, String name) {
    if (!new File(URI.create(imgNewPath + name)).exists()) {
      return imgNewPath + name;
    }
    do {
      name = "_" + name;
    } while (new File(URI.create(name)).exists());
    return imgNewPath + name;
  }
}
