/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.osgi.framework.Bundle;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.checkers.IChecker;
import com.unitesk.requality.core.descriptors.NodeTypeDescriptor;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.eclipse.core.EclipseCheckerContainer;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.editors.browser.ReqMarker;
import com.unitesk.requality.eclipse.ui.editors.JSEditor;
import com.unitesk.requality.eclipse.views.RequalityPreferences;

public class RequalityPlugin {
  // older then 3.7 support
  public static final int WEBKIT = 1 << 16;
  private static Map<String, NodeTypeDescriptor> nodeTypes;
  private static int BROWSER_TYPE;
  public static final String ID = "com.unitesk.requality";

  private static boolean enginesInitiated = false;
  private static SortedMap<Integer, String> browser_engines = new TreeMap<>();
  private static SortedMap<Integer, Short> browser_engines_state = new TreeMap<>();

  private static void initXULRunner() {
    // String
    // bundleName="org.mozilla.xulrunner."+System.getProperty("osgi.ws")+"."+System.getProperty("osgi.os")+"."+System.getProperty("osgi.arch");
    String bundleName = "org.mozilla.xulrunner";
    //                + System.getProperty("osgi.ws") + "."
    //                + System.getProperty("osgi.os") + "."
    //        		+ System.getProperty("osgi.arch");
    //        Platform.getBundles("org.mozilla.xulrunner*","*");
    Bundle bundle = Platform.getBundle(bundleName); //$NON-NLS-1$
    if (bundle != null) {
      Enumeration<URL> resourceUrl = bundle.findEntries(".", "xulrunner", false); //$NON-NLS-1$
      if (resourceUrl != null) {
        try {
          URL fileUrl = FileLocator.toFileURL(resourceUrl.nextElement());
          File file = new File(fileUrl.toURI());
          System.setProperty(
              "org.eclipse.swt.browser.XULRunnerPath", file.getAbsolutePath()); //$NON-NLS-1$
        } catch (IOException e) {
          e.printStackTrace();
        } catch (URISyntaxException e) {
          e.printStackTrace();
        }
      }
    }
  }

  public static void initBrowserEngines() {
    if (enginesInitiated) return;
    try {
      int selected = RequalityPreferences.getSelectedBrowserEngine();
      try {
        initXULRunner();
      } catch (Exception e) {
        RLogger.logInfo(CoreUtils.getTrace(e));
      }
      //cheat - it is default engine that is instantly active
      browser_engines.put(SWT.NONE, "Not supported");
      IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
      //
      try {
        URI uri = Activator.getUrl("./xulrunner");
        if (uri != null) {
          //URL fileUrl = FileLocator.toFileURL(uri);
          File file = new File(uri);
          if (file.exists())
            System.setProperty(
                "org.eclipse.swt.browser.XULRunnerPath", file.getAbsolutePath()); //$NON-NLS-1$
        }
      } catch (IOException e1) {
        RLogger.logInfo(CoreUtils.getTrace(e1));
      }
      try {
        browser_engines_state.put(SWT.MOZILLA, (short) 2);
        browser_engines.put(SWT.MOZILLA, "Mozilla engine");
        File f = new File("./mozilla");
        int old = selected;
        RequalityPreferences.setSelectedBrowserEngine(-1);
        if (!f.exists() || selected == SWT.MOZILLA) {
          try {
            browser_engines_state.put(SWT.MOZILLA, (short) 0);
            f.createNewFile();
            if (window != null)
              browser_engines_state.put(
                  SWT.MOZILLA, isBrowserAvailable(SWT.MOZILLA, window.getShell()));
            f.delete();
            RequalityPreferences.setSelectedBrowserEngine(old);
          } catch (Exception e) {
            RLogger.logInfo(CoreUtils.getTrace(e));
          }
        }
      } catch (java.lang.IllegalStateException e) {
      }
      try {
        browser_engines_state.put(WEBKIT, (short) 2);
        browser_engines.put(WEBKIT, "Webkit engine");
        File f = new File("./webkit");
        int old = selected;
        RequalityPreferences.setSelectedBrowserEngine(-1);
        if (!f.exists() || selected == WEBKIT) {
          browser_engines_state.put(WEBKIT, (short) 0);
          try {
            f.createNewFile();
            if (window != null)
              browser_engines_state.put(WEBKIT, isBrowserAvailable(WEBKIT, window.getShell()));
            f.delete();
            RequalityPreferences.setSelectedBrowserEngine(old);
          } catch (Exception e) {
          }
        }
      } catch (java.lang.IllegalStateException e) {
      }
      BROWSER_TYPE = SWT.NONE;
      if (selected > -1
          && browser_engines_state.containsKey(selected)
          && browser_engines_state.get(selected) == 1) {
        BROWSER_TYPE = selected;
      } else {
        if (window == null) {
          BROWSER_TYPE = SWT.NONE;
        } else if (browser_engines_state.get(SWT.MOZILLA) == 1) {
          BROWSER_TYPE = SWT.MOZILLA;
        } else if (browser_engines_state.get(SWT.WEBKIT) == 1) {
          BROWSER_TYPE = WEBKIT;
        }
      }
      enginesInitiated = window != null;
    } catch (java.lang.IllegalStateException e) {
      BROWSER_TYPE = SWT.NONE;
    }
  }

  // Replace with standard tools?

  public static EclipseTracker getTreesTracker() {
    return EclipseTracker.getTracker();
  }

  private static short isBrowserAvailable(int type, Composite parent) {
    try {
      Browser browser = new Browser(parent, type);
      browser.dispose();
      return 1;

    } catch (SWTError er) {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      PrintWriter pw = new PrintWriter(baos);
      er.printStackTrace(pw);
      pw.close();
      try {
        baos.close();
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
      RLogger.logInfo(baos.toString());
      return 0;
    }
  }

  public static IEditorReference findMarkEditorForDocument(String docQId) throws PartInitException {
    try {
      IWorkbench w = PlatformUI.getWorkbench();
      if (w == null) return null;
      List<IEditorReference> editors = new ArrayList<IEditorReference>();
      for (IWorkbenchWindow ww : PlatformUI.getWorkbench().getWorkbenchWindows()) {
          for (IWorkbenchPage page : ww.getPages()) {
              for (IEditorReference eRef : page.getEditorReferences()) {
                 if (!(eRef.getEditor(false) instanceof ReqMarker)) {
                  continue;
                }
                ReqMarker marker = (ReqMarker) eRef.getEditor(true);
                TreeDB db = marker.getDocument().getTreeDB();
                if (db.isVirtual()) db = Instantiator.getRootTree(db);
                //if (!project.equals(TreesTracker.getTracker().getProject(db))) {
                //  continue;
                //}
                if (marker.getDocument().getQualifiedId().equals(docQId)) {
                  return eRef;
                }
              }
          }
      }
      //IEditorReference[] eRefs = ww.getActivePage().getEditorReferences();
      
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public static IEditorReference findJSEditorForNode(String nodeQId, IProject project)
      throws PartInitException, CoreException {
    IEditorReference[] eRefs =
        PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getEditorReferences();
    for (IEditorReference eRef : eRefs) {
      if (!(eRef.getEditor(false) instanceof JSEditor)) {
        continue;
      }
      JSEditor jsEditor = (JSEditor) eRef.getEditor(true);
      if (!(jsEditor.getEditorInput() instanceof NodeEditorInput)) {
        continue;
      }
      NodeEditorInput editorInput = (NodeEditorInput) jsEditor.getEditorInput();
      IResource res = EclipseTracker.getResourceByNode(editorInput.getStorage().getNode());
      if (res == null || !project.equals(res.getProject())) {
        continue;
      }
      if (editorInput.getStorage().getNode().getQualifiedId().equals(nodeQId)) {
        return eRef;
      }
    }
    return null;
  }

  public static Map<String, NodeTypeDescriptor> getAdditionalTypes() {
    if (nodeTypes == null) {
      nodeTypes = new HashMap<>();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor(Activator.EXTENSION_NAME_NODE_TYPES);
      for (IConfigurationElement conf : configurationElements) {
        try {
          TreeNode nt = (TreeNode) conf.createExecutableExtension("typeclass");
          nodeTypes.put(
              conf.getAttribute("name"),
              new NodeTypeDescriptor(
                  conf.getAttribute("name"),
                  nt.getClass(),
                  conf.getAttribute("icon"),
                  conf.getContributor().getName()));
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }

    return nodeTypes;
  }

  private static EclipseCheckerContainer checkers = null;

  public static EclipseCheckerContainer getCheckers() {
    if (checkers == null) {
      checkers = new EclipseCheckerContainer();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor(Activator.EXTENSION_NAME_CHECKERS);
      for (IConfigurationElement conf : configurationElements) {
        try {
          IChecker checker = (IChecker) conf.createExecutableExtension("class");
          checkers.addChecker(checker);
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }
    return checkers;
  }

  public static SortedMap<Integer, String> getBrowserEngineNames() {
    RequalityPlugin.initBrowserEngines();
    return browser_engines;
  }

  public static Browser newBrowser(Composite parent, int style) {
    initBrowserEngines();
    return new Browser(parent, style);
  }

  public static SortedMap<Integer, Short> getBrowserEnginesState() {
    RequalityPlugin.initBrowserEngines();
    return browser_engines_state;
  }

  public static int getBrowserType() {
    RequalityPlugin.initBrowserEngines();
    return BROWSER_TYPE;
  }
}
