package com.unitesk.requality.eclipse.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.checkers.IChecker;
import com.unitesk.requality.core.checkers.IMark;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class EclipseChecker extends IncrementalProjectBuilder {
  public static final String BUILDER_ID = "com.unitesk.requality.checkers";

  private IResourceDeltaVisitor visitor =
      new IResourceDeltaVisitor() {

        @Override
        public boolean visit(IResourceDelta delta) throws CoreException {
          if (delta.getKind() == IResourceDelta.REMOVED) {
            return false;
          }
          if (delta.getResource().getType() != IResource.FILE) {
            return true;
          }
          TreeNode node = EclipseTracker.getNodeByResource((IFile) delta.getResource(),true);
          if (node != null
              && !node.isDisposed()
              && !node.getTreeDB().getCheckerContainer().isChecked(node)) {
            RequalityPlugin.getCheckers().check(node, false);
          }
          return false;
        }
      };
  boolean deep = false;

  public EclipseChecker() {
    this(false);
  }

  public EclipseChecker(boolean d) {
    deep = d;
  }

  @Override
  protected IProject[] build(int kind, Map<String, String> args, IProgressMonitor monitor)
      throws CoreException {
    if (kind == FULL_BUILD) {
      IProject proj = getProject();
      if (proj == null) return null;
      TreeDB db = EclipseTracker.getTracker().getTree(proj);
      if (db != null) {
        db.setCheckers(RequalityPlugin.getCheckers());
        runCheck(db, null);
      }
    } else {
//      IProject proj = getProject();
//      if (proj == null) return null;
//      if (getDelta(getProject()) != null) {
//        getDelta(getProject()).accept(visitor, false);
//      }
    }
    return null;
  }

  public void runCheck(TreeDB db, final IChecker addcheck) {
    db = Instantiator.getRootTree(db);
    if (db == null || db.isBusy()) return; //TODO delay;
    final TreeNode root = db.getRootNode();
    if (root != null) {
      final DeepFirstTreeWalker tree =
          new DeepFirstTreeWalker(root, false, false, true, true, true);
      tree.setIgnoreUnloaded(!deep);
      WorkspaceJob checkNode =
          new WorkspaceJob("Check Project " + db.getName()) {
            @Override
            public IStatus runInWorkspace(IProgressMonitor monitor) {
              Set<TreeNode> nodes = new HashSet<>();
              if (root.getTreeDB().isBusy()) return Status.CANCEL_STATUS;
              for (TreeNode node : tree) {
                RequalityPlugin.getCheckers().markUnChecked(node);
                if (monitor.isCanceled() || node.getTreeDB().isBusy()) return Status.CANCEL_STATUS;
                nodes.add(node);
              }
              for (TreeNode node : nodes) {
                if(node.isVirtual())
                  continue;
                RequalityPlugin.getCheckers().markChecked(node);
                if (monitor.isCanceled() || node.getTreeDB().isBusy()) return Status.CANCEL_STATUS;
                if (!node.isDisposed()) {
                  List<IMark> marks =
                      new ArrayList<>(
                          Arrays.asList(RequalityPlugin.getCheckers().check(node, true)));
                  if (addcheck != null) addcheck.check(node, marks);
                  node.getTreeDB().setChprobUnit(node.getUUId(), marks.toArray(new IMark[0]));
                }
              }
              return Status.OK_STATUS;
            }
          };
      checkNode.setRule(EclipseTracker.getTracker().getProject(db));
      checkNode.setPriority(Job.BUILD);
      checkNode.schedule();
    }
    //        if(rs!=null)
    //			rs.unLockProject();//useless for now
  }
}
