package com.unitesk.requality.eclipse.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.checkers.CheckerContainer;
import com.unitesk.requality.core.checkers.CheckerProblem;
import com.unitesk.requality.core.checkers.IMark;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.Document;

public class EclipseCheckerContainer extends CheckerContainer {
  private static final String REQUALITY_MARKER_TYPE = "com.unitesk.requality.marker",
      REQUALITY_MARKER_WARN_TYPE = "com.unitesk.requality.markerWarn";
  private Set<TreeNode> toup = null;
  private Map<TreeDB, ReentrantReadWriteLock> rwl = new HashMap<>();
  private Set<TreeNode> was = null;

  @Override
  public IMark[] check(final TreeNode node, final boolean batch) {
    if (batch) {
      return EclipseCheckerContainer.this.doCheck(node, batch);
    } else {
      TreeDB rdb = Instantiator.getRootTree(node.getTreeDB());
      Lock wl = null, rl = null;
      synchronized (rwl) {
        if (!rwl.containsKey(rdb)) rwl.put(rdb, new ReentrantReadWriteLock());
        rl = rwl.get(rdb).readLock();
        wl = rwl.get(rdb).writeLock();
      }
      wl.lock();
      try{
        if (toup == null) {
          toup = new HashSet<>();
          toup.add(node);
          was = new HashSet<>();
          Lock fl = rl,fwl =wl;
          WorkspaceJob checkNode =
              new WorkspaceJob("Check Node " + node.getId()) {
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) {
                  synchronized (fl) {
                    fl.lock();
                    boolean brl = true;
                    try{
                      if (toup == null) {
                        return Status.CANCEL_STATUS;
                      }
                      while (toup!=null && !toup.isEmpty()) {
                        fl.unlock();
                        brl = false;
                        fwl.lock();
                        TreeNode node = null;
                        try{
                          Iterator<TreeNode> it = toup.iterator();
                          node = it.next();
                          it.remove();
                        }finally{
                          fwl.unlock();
                        }
                        if (!was.contains(node)) {
                          try {
                            node.getTreeDB().getChprob().put(node.getUUId(), EclipseCheckerContainer.this.doCheck(node, batch));
                          } catch (Exception e) {
                            e.printStackTrace();
                            return Status.CANCEL_STATUS;
                          }
                        }
                        fl.lock();
                        brl = true;
                      }
                    }
                    finally{
                      if(brl);
                        fl.unlock();
                    }
                    fwl.lock();
                    toup = null;
                    was = null;
                    fwl.unlock();
                    return Status.OK_STATUS;
                  }
                }
              };
          //checkNode.setRule(ResourcesPlugin.getWorkspace().getRuleFactory().buildRule());
          checkNode.setPriority(Job.LONG);
          checkNode.schedule();
        } else {
          //synchronized (toup) {
          toup.add(node);
          //}
        }
      }finally{
        wl.unlock();
      }
      return null; //TODO support
    }
  }

  @Override
  public void applyProblemsInfo(final ArrayList<IMark> result) {
    IResource res = EclipseTracker.getResourceByNode(getNode());
    if (res == null || !res.exists()) {
      // FIXME: why we can be here?
      return;
    }
    applyProblemsInfo(res, result.toArray(new IMark[0]));
  }

  public static void applyProblemsInfo(IResource res, final IMark[] result) {
    try {
      if (!res.exists()) {
        RLogger.logInfo("Warn! res not exists! " + res.getFullPath().toOSString());
        return;
      }
      res.deleteMarkers(REQUALITY_MARKER_WARN_TYPE, true, IResource.DEPTH_ONE);
      res.deleteMarkers(REQUALITY_MARKER_TYPE, true, IResource.DEPTH_ONE);
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      RLogger.logInfo(e.getMessage() + " " + e.getStackTrace()[0].toString());
    }

    for (IMark problem : result) {
      TreeNode node = (TreeNode) problem.getTarget();
      IResource resource = EclipseTracker.getMarkerResourceByNode(node);
      synchronized (resource) {
        if (resource == null || !resource.exists())
          // FIXME: why we can be here?
          continue;
        try {
          IMarker marker;
          if (problem instanceof CheckerProblem)
            marker = resource.createMarker(REQUALITY_MARKER_TYPE);
          else marker = resource.createMarker(REQUALITY_MARKER_TYPE);
          //marker= resource.createMarker(REQUALITY_MARKER_WARN_TYPE);
          if (marker.exists()) { //TODO why it can not exist?

            synchronized (marker.getResource()) {
              marker.setAttribute(IMarker.MESSAGE, problem.getDescription());
              if (problem instanceof CheckerProblem)
                marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
              else marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
              if (node.getType().equals(Document.TYPE_NAME))
                marker.setAttribute(IMarker.LOCATION, ((Document) node).getHTMLPath());
              else marker.setAttribute(IMarker.LOCATION, node.getQualifiedId());
            }
          } else RLogger.logInfo("non ex. marker");
        } catch (CoreException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
  }
}
