package com.unitesk.requality.eclipse.core;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.instancer.Instantiator;

public class TreeDBBackupRepeatingJob extends RepeatingJob implements ITreeStateListener {
  private TreeDB fdb = null;
  public static long DELAY = 5 * 60000;
  public boolean forcenow = false;

  public TreeDBBackupRepeatingJob(TreeDB fdb) {
    super("Save UUIDs Cache for Project: " + fdb.getName(), DELAY);
    this.fdb = fdb;
    //new RepeatingJob("Save Cache for Project: "
    //+ project.getName(),5*60000) {
    //}
  }

  @Override
  public IStatus run(IProgressMonitor monitor) {
    IStatus st = super.run(monitor);
    if (st == Status.OK_STATUS) {
      if (fdb.inTransaction() || fdb.getChangeRowCount() > 0) pause();
      else start();
      if (!pause) {
        try {
          cacheStoreActions(fdb, false, forcenow);
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }
    return st;
  }

  public static void cacheStoreActions(TreeDB db, boolean fin, boolean forcenow) {
    TreeDB vroot = Instantiator.getRootTree(db);
    if (vroot == null || vroot.isBusy() && !fin && !forcenow) return;
    long l = System.currentTimeMillis();
    if (db.getCacheStorage().saveUUIDs()) {
      if(CoreUtils.DEBUG_LEVEL>0)
        RLogger.logInfo(
          "late changes in "
              + (System.currentTimeMillis() - l)
              + " miliseconds for "
              + db.getName()
              + (fin ? " final" : ""));
    }
    //RLogger.logInfo("late changes in " +(System.currentTimeMillis()-l)+" miliseconds for "+db.getName()+(fin?" final":""));
  }

  @Override
  public void open(TreeDB db) {
    if (db.equals(fdb)) {
      if (!running) {
        running = true;
        setPriority(Job.LONG);
        schedule(DELAY);
      } else start();
    }
  }

  @Override
  public void close(TreeDB db) {
    if (db.equals(fdb)) stop();
  }

  @Override
  public void delete(IProject proj) {
    if (proj.getName().equals(fdb.getName())) //TODO fix
    stop();
  }

  public void halt() {
    running = true;
    forcenow = true;
    run(null);
    stop();
    cancel();
    try {
      join();
    } catch (InterruptedException e) {
      e.printStackTrace();
    }
  }

  public void singleAndStop(TreeDB db) {
    //db.clearAllCaches(true);
    cacheStoreActions(db, true, true);
    stop();
    cancel();
  }
}
