package com.unitesk.requality.eclipse.cvs;

import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.team.core.RepositoryProvider;

import com.unitesk.requality.core.FileSpaceTimePoint;
import com.unitesk.requality.eclipse.Activator;

public class CVSManager {
  /** Константы */
  public static final int CVCLMANAGER_OK = 0;

  public static final int CVCLMANAGER_WORKINGPLACEFAILED = 1;
  public static final int CVCLMANAGER_HISTORYFOLDERFAILED = 2;
  public static final int CVCLMANAGER_TEMPFOLDERFAILED = 3;
  public static final int CVCLMANAGER_TEMPFILEFAILED = 4;
  public static final int CVCLMANAGER_NOTIMPLEMENTED = 5;
  public static final int CVCLMANAGER_REPOMISMATCH = 6;
  public static final int CVCLMANAGER_PACKFOLDERFAILED = 7;
  public static final int CVCLMANAGER_FAILED = 8;
  public static final int CVCLMANAGER_COMPHISTFILEFAILED = 9;
  public static final int CVCLMANAGER_NOTEAMPROVIDER = 10;
  public static final int CVCLMANAGER_REPOINTERFACEFAILED = 11;
  public static final int CVCLMANAGER_REPO_AUTH_EFAILED = 12;
  public static final int CVCLMANAGER_OK_NO_PACKAGE_FOUND = 13;
  public static final int CVCLMANAGER_OK_NO_FILE_IN_REPO_OR_NO_CONNECTION = 14;
  public static final int CVCLMANAGER_OK_REV_MISMATCH = 15;
  public static final int CVCLMANAGER_OK_COMMIT_NO_CHANGES = 16;
  public static final int CVCLMANAGER_OK_REVERT_NO_COMPONENT_IN_REPO = 17;
  public static final int CVCLMANAGER_OK_REVERT_NO_CHANGES = 18;

  public static final String[] CVCLMNGR_ERROR_MESSAGES = {
    "OK",
    "Failed to create/open .cvcl folder",
    "Failed to create/open history folder",
    "Failed to create/open temp folder",
    "Failed to create/open temp file",
    "The required feature is not implemented (either support for team provider in use is not built in CVCL manager or repository client is not comprehended)",
    "File seems to be not within the given repository (repository/file URLs mismatch, the log might help)",
    "Failed to create/open folder for package history",
    "Some error has happened (the log might help)",
    "Failed to create/open component history file",
    "There is no team provider assigned to the project, CVCL manager cannot work without it.",
    "Repository interface has failed: either initialization problems, or reading history problem, or getting file problems",
    "Repository authentication problems",
    "The demanded package has not been found in aadl program. The reason might be the fact, that this package is new and has not been commited to the repository yet. Please, commit the local copy of the file.",
    "The CVCL was not able to get revision of the file. Either there is problem with connection to the repository or file is not in the repository yet. Try to roll in file to the repostory first.",
    "The local revision of the file is different from the revision in the repository. Please, update your local copy and then try again.",
    "CVCL has not found any changes in the component, so there is nothing to commit.",
    "Component seems to be absent in the repository version of the file. Nothing to revert.",
    "CVCL has not found any changes in the component, so there is nothing to revert."
  };

  /** Поля */
  // Поля, связанные с клиентом SVC
  private ICVSClient svcClient = null; // SVC клиент

  // Поля, связанные с проектом
  private IProject project = null; // проект
  private File workingPlace =
      null; // папка внутри рабочей папки проекта, где хранятся вспомогательные папки и файлы
  private String projectAbsolutePath = null; // Абсолютный путь до рабочей папки проекта
  //private String workspaceAbsolutePath = null;

  private File historyFolder = null; // папка с файлами историй компонентов
  private File tempFolder = null; // папка для временных файлов

  private int lastError = CVCLMANAGER_OK; // Код последней ошибки

  // Другие поля
  //private UsernamePasswordDialog usrPasswdDialog = null;

  /** Методы */
  public CVSManager() {
    //teamProviderId = CVCLMANAGER_TEAM_NONE;
  }

  public Set<FileSpaceTimePoint> getRevisions(IResource node) {
    return getRevisions(node, null);
  }

  private String lasterror = null;

  public String getLastRevError() {
    return lasterror;
  }

  public Set<FileSpaceTimePoint> getRevisions(IResource node, String rev) {
    try {
      lasterror = null;
      ICVSClient man = getCVSClient(node.getProject());
      Set<FileSpaceTimePoint> s = new HashSet<>();

      if (man != null)
        if (node instanceof IFile) s.addAll(man.getResourceRevisions(node, rev, null));
        else s.addAll(man.getResourceRevisions(node, rev, null));
      //s.addAll(getRevisions(man, (IContainer)node));
      return s;
    } catch (Exception e) {
      lasterror = e.toString();
    }
    return Collections.emptySet();
  }

  public Set<FileSpaceTimePoint> getRevisions(IProject p) {
    Set<FileSpaceTimePoint> list = new HashSet<>();
    try {
      ICVSClient man = getCVSClient(p);
      if (man == null) return list;
      for (IResource r : p.members()) list.addAll(getRevisions(r));
    } catch (Exception e) {
      e.printStackTrace();
    }
    return list;
  }

  public int getLastErrorCode() {
    return lastError;
  }

  public ICVSClient getCVSClient(IProject project) throws Exception {
    RepositoryProvider teamProvider = RepositoryProvider.getProvider(project);
    if (teamProvider == null) {
      lastError = CVCLMANAGER_NOTEAMPROVIDER;

      //if(withLog) log.addLogLine("No team provider");
      lastError = CVCLMANAGER_NOTEAMPROVIDER;
      throw new CVSException(
          "There is no team provider assigned to the project, CVCL manager cannot work without it.");
    }

    svcClient = Activator.getDefault().getSVCClientForTeamProvider(teamProvider.getID());
    if (svcClient == null) {
      lastError = CVCLMANAGER_NOTIMPLEMENTED;

      throw new CVSException("There is no support for the given team provider");
    }

    //svcClient.init(repoURL);
    svcClient.init(project);
    svcClient.excludeDotReqCSVDir();
    return svcClient;
  }

  public void initManagerWithProject(IProject pr, UsernamePasswordDialog upD, boolean needLog)
      throws CVSException {
    lastError = CVCLMANAGER_OK;
    if (pr == null) {
      lastError = CVCLMANAGER_FAILED;
      throw new CVSException("Project pointer is null");
    }
    project = pr;
    //usrPasswdDialog = upD;

    //workspaceAbsolutePath = project.getWorkspace().getRoot().getLocation().toFile().toString();
    projectAbsolutePath =
        (new File(
                project.getWorkspace().getRoot().getLocation().toFile().toString()
                    + "/"
                    + project.getFullPath().toString()))
            .getAbsolutePath();

    // создаём рабочую папку
    workingPlace = new java.io.File(projectAbsolutePath + "/.reqcvs/");
    if (!workingPlace.exists())
      if (!workingPlace.mkdir()) {
        lastError = CVCLMANAGER_WORKINGPLACEFAILED;
        throw new CVSException("Cannot create folder " + workingPlace.getAbsolutePath());
      }

    // создаём папку с иторией
    historyFolder = new java.io.File(workingPlace.getAbsolutePath() + "/history/");
    if (!historyFolder.exists())
      if (!historyFolder.mkdir()) {
        lastError = CVCLMANAGER_HISTORYFOLDERFAILED;
        throw new CVSException("Cannot create folder " + historyFolder.getAbsolutePath());
      }

    // создаём папку для временных файлов
    tempFolder = new java.io.File(workingPlace.getAbsolutePath() + "/tmp/");
    if (!tempFolder.exists())
      if (!tempFolder.mkdir()) {
        lastError = CVCLMANAGER_TEMPFOLDERFAILED;
        throw new CVSException("Cannot create folder " + tempFolder.getAbsolutePath());
      }

    // борт-журнал
    //выясняем, есть ли вообще team provider
    RepositoryProvider teamProvider = RepositoryProvider.getProvider(project);
    if (teamProvider == null) {
      lastError = CVCLMANAGER_NOTEAMPROVIDER;

      //if(withLog) log.addLogLine("No team provider");
      lastError = CVCLMANAGER_NOTEAMPROVIDER;
      throw new CVSException(
          "There is no team provider assigned to the project, CVCL manager cannot work without it.");
    }

    svcClient = Activator.getDefault().getSVCClientForTeamProvider(teamProvider.getID());
    if (svcClient == null) {
      lastError = CVCLMANAGER_NOTIMPLEMENTED;

      throw new CVSException("There is no support for the given team provider");
    }

    try {
      //svcClient.init(repoURL);
      svcClient.init(project);
      svcClient.excludeDotReqCSVDir();

      //repoURL = svcClient.getRepoURL();
    } catch (Exception ex) {
      //ex.printStackTrace();

      CVSException locEx = new CVSException("Some exception during svcClient initialisation");
      locEx.initCause(ex);

      lastError = CVCLMANAGER_REPOINTERFACEFAILED;
      throw locEx;
    }

    lastError = CVCLMANAGER_OK;
  }

  void scratchYard(String filename) {
    try {
    } catch (Exception e) {
      e.printStackTrace();
    }

    /*try {
    	org.tigris.subversion.svnclientadapter.commandline.CmdLineClientAdapterFactory.setup();
    }
    catch (org.tigris.subversion.svnclientadapter.SVNClientException e) {
    	e.printStackTrace();
    }*/

  }
}
