/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.editors.browser;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.handlers.requirement.DeleteRequirementHandler;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

final class RemoveLocationFromRequirement extends Action {
  private final ReqMarker reqMarker;

  RemoveLocationFromRequirement(ReqMarker reqMarker) {
    super("Remove from Requirements");
    this.reqMarker = reqMarker;
  }

  public boolean isEnabled() {
    try {
      return reqMarker.getBrowser().isRemovable();
    } catch (Exception e) {
      e.printStackTrace();
      return false;
    }
  }

  @Override
  public void run() {
    String locQId =
        reqMarker.getDocument().getQualifiedId(true)
            + "/"
            + reqMarker.getBrowser().getLocationOnSelection();
    Location loc = (Location) reqMarker.getTreeDB().getNode(locQId, Path.QID | Path.UUID);
    Requirement[] req4check = null;
    if (loc != null) req4check = loc.getRequirements();
    if (!MessageDialog.openConfirm(
        PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
        "Delete Location",
        "Delete selected Location?")) {
      return;
    }
    if (!reqMarker.getDocument().getTreeDB().startTransaction("Delete Location", true)) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Delete Location Error",
          "Some actions are already performing");
      return;
    }
    if (loc != null) loc.delete();
    else
      reqMarker.getDocument().removeLocationRegion(reqMarker.getBrowser().getSelectedLocationId());
    //   	 try{
    //	        if(req4check.length>0){
    //	        	 CommonRemoveLocation.RemoveLocation(loc, Arrays.asList(req4check), reqMarker, true);
    //
    //	        }
    //	        else if(req4check.length==0)
    //	        	CommonRemoveLocation.RemoveLocation(loc, null, reqMarker, true,locQId,reqMarker.getDocument().getTreeDB());
    //			loc.getTreeDB().commit();
    //     }
    // 	catch(Exception e){
    // 		e.printStackTrace();
    // 		loc.getTreeDB().rollback();
    // 	}
    reqMarker.getDocument().getTreeDB().commit();
    List<TreeNode> reqs = new ArrayList<>();
    String strList = "";
    if (req4check != null)
      for (Requirement req : req4check) {
        if (req.getLocations().size() == 0 && req.sizeWorkChildren() == 0) {
          reqs.add(req);
          if (!strList.equals("")) {
            strList += ", ";
          }
          strList += req.getNameOrId();
        }
      }
    if (reqs.size() > 0) {
      //TODO message
      if (reqs.get(0).getTreeDB().inTransaction()) return;
      if (MessageDialog.openQuestion(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Delete Requirement",
          (reqs.size() == 1
                  ? "Requirement \"" + reqs.get(0).getUserVisibleName(true) + "\" has "
                  : "Requirements " + strList + " have ")
              + "no locations, delete?")) {
        List<UUID> proc = new ArrayList<>();
        if (reqs.get(0)
            .getTreeDB()
            .startTransaction("Delete Requirement" + (reqs.size() > 1 ? "s" : ""), true)) {
          try {
            new DeleteRequirementHandler().process(reqs, proc, true);
          } catch (ExecutionException e) {
            e.printStackTrace();
          }
          reqs.get(0).getTreeDB().commit();
        }
      }
    } else reqMarker.fireSelectionChange();
  }
}
