/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.editors.browser;

import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.browser.BrowserFunction;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.ISaveablePart;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.navigator.CommonNavigator;
import org.eclipse.ui.operations.UndoRedoActionGroup;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.IContributedContentsView;
import org.eclipse.ui.texteditor.ITextEditor;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.w3c.dom.DOMException;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.SharedConsts;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeLogic;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeChangeEvent;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.core.transaction.ChangeResource;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.core.transaction.TransactionStorage.OperationData;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipseCopyHandler;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipsePasteHandler;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.eclipse.ui.editors.ISearchProvider;
import com.unitesk.requality.eclipse.ui.editors.UniEditor;
import com.unitesk.requality.eclipse.views.documents.ProcessorData;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.RequalityCLI;

public class ReqMarker extends EditorPart
    implements ITabbedPropertySheetPageContributor,
        ISelectionProvider,
        ITransactionListener,
        ISaveablePart,
        ISearchProvider{
  private static final String DEFAULT_BASE_ENCODING = "UTF-8";
  public static final String ID = "com.unitesk.requality.editors.browser.ReqMarker";
  protected ReqDocBrowser browser;

  private Document doc;
  private String docQId;
  private TreeDB db;
  private Location initialLocation;
  private IProject project;
  private boolean selFlag = true;

  private ReqMarkerOutlinePage reqMarkerOutlinePage;
  public String getSelectionText() {
    return browser.getSelectionText();
  }
  public boolean findText(
      String aString, boolean aCaseSensitive, boolean aBackwards, boolean aWrapAround) {
    return browser.findText(aString, aCaseSensitive, aBackwards, aWrapAround);
  }
  // private IUndoContext undoContext;
  public IUndoContext getUndoContext() {
    // use workbench context
    return PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
    // undoContext = new ObjectUndoContext(this.getEditorInput());
    // return undoContext;
  }

  public void openDocument(URL url) {
    String path = url.toString();
    browser.setUrl(path);
  }

  public ReqDocBrowser getBrowser() {
    return browser;
  }

  private ISelectionChangedListener locChangeListener =
      new ISelectionChangedListener() {

        @Override
        public void selectionChanged(SelectionChangedEvent event) {
          if (browser != null && !browser.isDisposed() && selFlag) {
            IStructuredSelection s = (IStructuredSelection) event.getSelection();
            Object o = s.getFirstElement();
            if (o instanceof Location) {
              Location loc = (Location) s.getFirstElement();
              browser.setTargetLocation(loc);
              synchWithExplorer();
            } else if (o instanceof String) {
              //missing location
              String missid = o.toString();
              if (missid.startsWith("missing: ")) missid = missid.substring(9);
              browser.setTargetLocation(missid);
            }
          }
          selFlag = true;
        }
      };

  private void nodeUpdate() {
    browser.updateLocations();
    if (reqMarkerOutlinePage != null && !reqMarkerOutlinePage.getControl().isDisposed()) {
      reqMarkerOutlinePage.refresh();
    }
  }

  private ITreeChangeListener treeChangeListener =
      new DefaultTreeChangeListener() {
        private void NodeUpdated(UUID node) {
          getSite()
              .getShell()
              .getDisplay()
              .asyncExec(
                  new Runnable() {
                    @Override
                    public void run() {
                      if (ReqMarker.this.getBrowser().isDisposed()) return;
                      if (EclipseTracker.getTracker().getProject(db) == null) {
                        getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                        return;
                      }
                      nodeUpdate();
                    }
                  });
        }

        private boolean chDeleted(TreeNode node) {
          if (doc.getUUId().equals(node.getUUId())) return true;
          if (node.getWorkParentUUID() != null && node.getWorkParentUUID().equals(doc.getUUId())) {
            if (node.getType().equals(Location.TYPE_NAME)) return true;
          }
          if (node.getType().equals(Requirement.TYPE_NAME)) {
            //TODO remove
            if (node.attrs().containsKey(Requirement.ATTR_LOCATIONS)) {
              //
              Attribute a = node.attrs().getAttribute(Requirement.ATTR_LOCATIONS);
              
              if (a != null&&a.getDefinition().getRawValue()!=null&&a.getDefinition().getRawValue().getValue()!=null) {
                List<?> l = ((ListValue) a.getDefinition().getRawValue().getValue()).getValues();
                for (Object o : l) {
                  String s = o.toString();
                  //TreeNode loc = doc.getTreeDB().getNode(UUID.fromString(s));
                  if (doc.getLocationIndexMap().get(UUID.fromString(s))!=null) return true;
                }
              }
            }
          }
          return false;
        }

        public void deleted(TreeNode node) { //TODO: async? Job?
          if (doc.getUUId().equals(node.getUUId())) {
            getSite()
                .getShell()
                .getDisplay()
                .syncExec(
                    new Runnable() {
                      @Override
                      public void run() {
                        getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                      }
                    });
          }
          if (chDeleted(node)) NodeUpdated(node.getUUId());
        }

        @Override
        public void moved(TreeNode node, TreeNode oldParent, TreeNode newParent) {
          if (chMoved(node, oldParent, newParent)) NodeUpdated(node.getUUId());
        }

        private boolean chMoved(TreeNode node, TreeNode oldParent, TreeNode newParent) {
          if (doc.getUUId().equals(node.getUUId())) return true;
          return (newParent != null
              && newParent.getUUId().equals(doc.getUUId())
              && node.getType().equals(Location.TYPE_NAME));
        }

        private boolean chChanged(TreeNode node, ChangesSet<Attribute> what) {
          if (doc.getUUId().equals(node.getUUId())) return true;
          if (node.getWorkParentUUID() != null
              && node.getWorkParentUUID().equals(doc.getUUId())
              && node.getType().equals(Location.TYPE_NAME)) return what.contains(TreeNode.ATTR_LABEL);
//          if (node.getType().equals(Requirement.TYPE_NAME)
//              && what.contains(Requirement.ATTR_LOCATIONS)) {
//            Requirement r = (Requirement) node;
//            for (TreeNode l : r.getLocations())
//              if (l.getWorkParentUUID().equals(doc.getUUId())) return true;
//          }
          return false;
        }

        @Override
        public void changed(TreeNode node, ChangesSet<Attribute> what) {
          if (chChanged(node, what)) NodeUpdated(node.getUUId());
        }

        private boolean chCreated(TreeNode node) {
          if (doc.getUUId().equals(node.getUUId())) return true;
          if (node.getWorkParentUUID() != null
              && node.getWorkParentUUID().equals(doc.getUUId())
              && node.getType().equals(Location.TYPE_NAME)) return true;
          if (Requirement.TYPE_NAME.equals(node.getType())) {
            Requirement r = (Requirement) node;
            for (TreeNode l : r.getLocations())
              if (l.getWorkParentUUID().equals(doc.getUUId())) return true;
          }
          return false;
        }

        @Override
        public void created(TreeNode node) {
          if (chCreated(node)) NodeUpdated(node.getUUId());
        }

        @SuppressWarnings("unchecked")
        @Override
        public void changed(List<TreeNodeChangeEvent> changes) {
          for (TreeNodeChangeEvent ce : changes)
            switch (ce.getType()) {
              case CREATE:
                TreeNode node = ce.getTreeNodeTarget();
                if (chCreated(node)) {
                  NodeUpdated(node.getUUId());
                  continue;
                }
                break;
              case DELETE:
                {
                  if(ce.getArguments()[2]==Moment.BEFORE)
                    continue;
                  TreeNode nd = ce.getTreeNodeTarget();
                  if (nd != null) { //needed?
                    if (chDeleted(nd)) {
                      if (doc.getUUId().equals(nd.getUUId())) {
                        getSite()
                            .getShell()
                            .getDisplay()
                            .syncExec(
                                new Runnable() {
                                  @Override
                                  public void run() {
                                    getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                                  }
                                });
                      } else NodeUpdated(nd.getUUId());
                      return;
                    }
                  } else {
                    nd = ce.getTreeNodeTarget();
                    if (chDeleted(nd)) {
                      if (doc.getUUId().equals(nd.getUUId())) {
                        getSite()
                            .getShell()
                            .getDisplay()
                            .syncExec(
                                new Runnable() {
                                  @Override
                                  public void run() {
                                    getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                                  }
                                });

                      } else NodeUpdated(nd.getUUId());
                      continue;
                    }
                  }
                  break;
                }
              case CHANGE:
                node = ce.getTreeNodeTarget();
                if (chChanged(node, (ChangesSet<Attribute>) ce.getArguments()[1])) {
                  NodeUpdated(node.getUUId());
                  continue;
                }
                break;
              
              case MOVE:
                node = ce.getTreeNodeTarget();
                TreeNode par = (TreeNode) ce.getArguments()[1];
                TreeNode npar = (TreeNode) ce.getArguments()[2];
                if (chMoved(node, par, npar)) {
                  NodeUpdated(node.getUUId());
                  continue;
                }
                break;
              default:
                break;
            }
        }
      };

  private ITreeStateListener dbStateListener =
      new ITreeStateListener() {

        @Override
        public void open(TreeDB db) {}

        @Override
        public void close(TreeDB db) {
          getSite()
              .getShell()
              .getDisplay()
              .syncExec(
                  new Runnable() {
                    @Override
                    public void run() {
                      if(browser.isDisposed())
                        return;
                      getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                    }
                  });
        }

        @Override
        public void delete(IProject proj) {
          if (project.equals(proj)) {
            getSite()
                .getShell()
                .getDisplay()
                .syncExec(
                    new Runnable() {
                      @Override
                      public void run() {
                        getEditorSite().getPage().closeEditor(ReqMarker.this, false);
                      }
                    });
          }
        }
      };
  private boolean ignoreSetLocation = false;
  final class BrowserKeyListener implements KeyListener {
    ReqDocBrowser browser;
    public BrowserKeyListener(ReqDocBrowser browser) {
      this.browser = browser;
    }

    @Override
    public void keyReleased(KeyEvent e) {
      if ((e.stateMask & SWT.CTRL) != 0) {
        if (e.keyCode == 'r') new AddSelectionToRequirement(ReqMarker.this).run();
        if (e.keyCode == 'd') new RemoveLocationFromRequirement(ReqMarker.this).run();
        if (e.keyCode == 'i') browser.showBrowserInfo();
        if (((e.stateMask & SWT.SHIFT) != 0) && e.keyCode == 'c') {
          String textData = browser.getSelectionText(false);
          if (textData == null || textData.equals("")) return;
          Clipboard clipboard = new Clipboard(getSite().getShell().getDisplay());
          TextTransfer textTransfer = TextTransfer.getInstance();
          Transfer[] transfers = new Transfer[] {textTransfer};
          Object[] data = new Object[] {textData};
          clipboard.setContents(data, transfers, DND.CLIPBOARD);
          clipboard.dispose();
        }
      }
       if (e.keyCode == SWT.DEL) {
        new RemoveLocationFromRequirement(ReqMarker.this).run();
        //e.doit = false;
      }
    }

    @Override
    public void keyPressed(KeyEvent e) {
    }
  }

  // перехватываем более низкоуровневые события, чем
  // IMenuListener.menuAboutToShow.
  // При событии mouseDown запрещаем выполнение
  // IMenuListener.menuAboutToShow,
  // (чтобы событие Right Mouse Down было передано браузеру),
  // но затем при событии mouseUp вызываем IMenuListener.menuAboutToShow
  // искусственно
  // с помощью Menu.setVisible(true)
  // Сделано это, чтобы событие Mouse Click (Right) передавалось в
  // JavaScript браузера.
  final class BrowserMouseListener implements MouseListener {
    ReqDocBrowser browser;
    public BrowserMouseListener(ReqDocBrowser browser) {
      this.browser = browser;
    }

    @Override
    public void mouseDown(MouseEvent e) {
      if (e.button == 3) {
        allow_popup_menu = false;
      }
    }

    @Override
    public void mouseUp(MouseEvent e) {
      if (e.button == 1) {
        //Moved to javascript
        //browser.activateSelectedLocation();
        //onSelectedLocationChangeByClick();
      } else if (e.button == 3) {
        allow_popup_menu = true;
        browser.getWebBrowser().getMenu().setVisible(true);
      }
    }

    @Override
    public void mouseDoubleClick(MouseEvent e) {}
  }

  public ReqMarker() {
    super();
  }

  protected void updateTargetRequirement() {
    List<Location> locationList = getTargetRequirement().getLocations();
    if (locationList.size() == 0) return;
    Location location = locationList.get(0);
    setTargetLocation(location, true);
  }

  public void setTargetLocation(Location loc, boolean first) {
    if (browser == null || ignoreSetLocation) {
      return;
    }
    ignoreSetLocation = true;
    try{
      browser.setTargetLocation(loc);
      if (reqMarkerOutlinePage != null && !reqMarkerOutlinePage.getControl().isDisposed()) {
        reqMarkerOutlinePage.refresh();
  
        StructuredSelection ss;
        if (loc != null) ss = new StructuredSelection(loc);
        else ss = new StructuredSelection();
        reqMarkerOutlinePage.setSelection(ss);
      }
      if (first) {
        Location l = loc;
        Requirement r = null;
        if(l!=null){
          Requirement[]rqs = l.getRequirements();
          if (isLinked()&&rqs.length>0) r = rqs[0];
        }
        List<TreeNode> act = new LinkedList<>(),act2=new LinkedList<>();
        act.add(l);
        if (r != null)
          act2.add(r);
        else
          act2.add(l);
        SelectionManager.getOne().setActiveNodes(act,act2, EventSource.EDITOR);
        setFocus();//To be focused on location for delete etc
        //else 
        //  SelectionManager.getOne().setActiveNode(l, Location.TYPE_NAME, EventSource.EDITOR);
      }
    }
    finally{
      ignoreSetLocation = false;
    }
  }

  /**
   * <code>saveDocument</code> must be called in transaction
   *
   * @param jsData
   * @throws CoreException
   */
  public void saveDocument(final String jsData) throws CoreException {
    IJSCommand saveDocument =
        new IJSCommand() {
          @Override
          public void onError(Exception e) {
            e.printStackTrace();
          }

          @Override
          public void onComplete(Object result, ReqDocBrowser docBrowser) {
            String htmlContent = (String) result;
            if (htmlContent != null && htmlContent.length() > 0) {
              // TODO: transaction (subtransaction)
              // db.startTransaction();
              ByteArrayInputStream in;
              try {
                in = new ByteArrayInputStream(htmlContent.getBytes(DEFAULT_BASE_ENCODING));
                doc.setResourceContent(doc.getResourceName(), in, jsData);
                in.close();
                //                        doc.resetCache();
                doc.resetUTI();
                // db.commit();
              } catch (Exception e) {
                // db.rollback();
                e.printStackTrace();
              }
            }
          }

          @Override
          public String getJS() {
            return "return serializeDocument();";
          }

          @Override
          public boolean isEvaluate() {
            return true;
          }
        };
    pullBrowser().scheduleJSCommand(saveDocument);
  }

  @Override
  public void doSave(IProgressMonitor monitor) {
    monitor.beginTask("save reqmarker", 1);
    saveAction();
    monitor.done();
  }

  @Override
  public void doSaveAs() {
    throw new UnsupportedOperationException("Unexpected method call");
  }

  @Override
  public void init(IEditorSite site, IEditorInput input) throws PartInitException {
    try {
      setSite(site);

      if (input instanceof NodeEditorInput) {
        NodeEditorInput docInput = (NodeEditorInput) input;
        if (!docInput.getStorage().getNode().getType().equals(Document.TYPE_NAME)
            && !docInput.getStorage().getNode().getType().equals(Requirement.TYPE_NAME)) {
          throw new PartInitException(
              "Wrong input type: " + docInput.getStorage().getNode().getType());
        }
        if (docInput.getStorage().getNode() instanceof Requirement) {
          Requirement req = (Requirement) docInput.getStorage().getNode();
          if (req.getLocations().size() > 0) {
            doc = (Document) req.getLocations().get(0).getWorkParent();
            initialLocation = req.getLocations().get(0);
          } else {
            doc = (Document) docInput.getStorage().getNode();
            initialLocation = (Location) docInput.getData();
          }
        } else {
          doc = (Document) docInput.getStorage().getNode();
          initialLocation = (Location) docInput.getData();
        }
        db = docInput.getStorage().getNode().getTreeDB();
        project = RequalityPlugin.getTreesTracker().getProject(db);
        docQId = doc.getQualifiedId();
      } else if (input instanceof FileEditorInput) {
        FileEditorInput fei = (FileEditorInput) input;
        project = fei.getFile().getProject();

        TreeNode node = EclipseTracker.getNodeByResource(fei.getFile());
        doc = (Document) node;
        if (doc != null) docQId = doc.getQualifiedId();
        //                MessageDialog
        //                        .openError(getSite().getShell(),
        //                                "Unsupported editor open method",
        //                                "not yet implemented");
      }
    } catch (CoreException e) {
      e.printStackTrace();
    }
    setInput(input);
    if (doc != null) {
      setPartName(doc.getId());
      doc.addGuiChangeListener(
          new INodeChangeListener() {

            @Override
            public void moved(TreeNode node, TreeNode oldParent, boolean internal) {
              // TODO Auto-generated method stub

            }

            @Override
            public void fireNodeResourceChange(
                TreeNode treeNode, String resName, Moment m, boolean inner) {
              if(treeNode.getUUId().equals(doc.getUUId()))
                if (m == Moment.AFTER && resName.equals(doc.getResourceName())) {
                  browser.updateLocations();
                  if (reqMarkerOutlinePage != null
                      && !reqMarkerOutlinePage.getControl().isDisposed()) {
                    reqMarkerOutlinePage.refresh();
                  }
                }
            }

            @Override
            public void deleted(TreeNode source) {
              // TODO Auto-generated method stub

            }

            @Override
            public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
              // TODO Auto-generated method stub

            }

            @Override
            public void attributeChange(TreeNode source, String what, Attribute old) {
              // TODO Auto-generated method stub

            }
          });
    }
    if (db == null) return;
    db.addGUIListener(treeChangeListener);
    db.addTransactionListener(this);
    EclipseTracker.getTracker().addListener(dbStateListener);
  }

  @Override
  public boolean isDirty() {
    if (browser == null || browser.isDisposed() || !browser.isReady()) return false;
    String html = browser.lastContent;
    if (html == null) return false;
    try {
      ByteArrayInputStream in;
      in = new ByteArrayInputStream(html.getBytes(DEFAULT_BASE_ENCODING));
      return doc.getDOM().isEqualNode(Document.loadDOM(in));
    } catch (DOMException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return false;
  }

  @Override
  public boolean isSaveAsAllowed() {
    return false;
  }

  @Override
  public void createPartControl(Composite parent) {
    new UndoRedoActionGroup(getSite(), getUndoContext(), true)
        .fillActionBars(getEditorSite().getActionBars());
    parent.setLayout(new GridLayout(1, false));

    Composite browserPanel = new Composite(parent, SWT.NONE);
    browserPanel.setLayoutData(new GridData(GridData.FILL_BOTH));
    browserPanel.setLayout(new FillLayout());

    browser = new ReqDocBrowser(browserPanel, SWT.NONE);
    IHandlerService handlerService = getSite().getService(IHandlerService.class);
    handlerService.activateHandler("org.eclipse.ui.edit.copy", new EclipseCopyHandler());
    handlerService.activateHandler("org.eclipse.ui.edit.paste", new EclipsePasteHandler());
    browser.getWebBrowser().setMenu(createMenu());

    new BrowserFunction(browser.getWebBrowser(), "setInitialLocation") {
      @Override
      public Object function(Object[] arguments) {
        setTargetLocation(initialLocation, true);
        return null;
      }
    };
    //method to call location change from javascript
    new BrowserFunction(browser.getWebBrowser(), "onSelectedLocationChangeByClick") {
      @Override
      public Object function(Object[] arguments) {
        onSelectedLocationChangeByClick();
        return null;
      }
    };
    new BrowserFunction(browser.getWebBrowser(), "tryOpenInEditor") {
      private String reworkPath(String uid) {
        uid = uid.replace("/./", "/");
        int point = uid.indexOf("../");
        while (point == 0) {
          uid = uid.substring(3);
          point = uid.indexOf("../", point);
        }
        while (point > -1) {
          int prep = uid.substring(0, point - 1).lastIndexOf("/");
          if (prep > -1) uid = uid.substring(0, prep) + uid.substring(point + 2);
          point = uid.indexOf("../", point);
        }
        return uid;
      }

      @Override
      public Object function(Object[] args) {
        String puid = doc.getQualifiedId();
        puid = puid.substring(0, puid.lastIndexOf("/"));
        String uid = args[0].toString().replace(".html", ".xhtml");
        String mark = null;
        int sch = uid.indexOf("#");
        if (sch > -1) {
          if (uid.indexOf("/", sch) == -1) {
            mark = uid.substring(sch + 1);
            uid = uid.substring(0, sch);
          }
        }
        TreeNode node = doc.getTreeDB().getNode(reworkPath(uid), Path.QID | Path.UV | Path.UUID);
        if (node == null) {
          uid = puid + "/" + uid;
          node = doc.getTreeDB().getNode(reworkPath(uid), Path.QID | Path.UV | Path.UUID);
        }
        if (node != null) {
          IEditorPart part = openEditor(node);
          if (mark != null && part instanceof ReqMarker) {
            ((ReqMarker) part).browser.scrollToTag(mark);
          }
          return true;
        }
        return false;
      }
    };
    browser.getWebBrowser().addKeyListener(new BrowserKeyListener(browser));
    browser.getWebBrowser().addMouseListener(new BrowserMouseListener(browser));
    // Get path to file

    // TODO: check for resource existing
    if (doc != null) {
      URI openPath = doc.getHTMLPath();
      try {

        URL url = openPath.toURL();
        openDocument(url);
      } catch (MalformedURLException e1) {
        e1.printStackTrace();
      }
    }

    getSite().setSelectionProvider(this);
  }

  protected IEditorPart openEditor(TreeNode node) {
    String type = node.getType();
    NodeEditorInput input = new NodeEditorInput(node);
    if (type.equals(TestPurpose.TYPE_NAME)
        || type.equals(Comment.TYPE_NAME)
        || type.equals("TestStep")
        || type.equals(Location.TYPE_NAME)) input = new NodeEditorInput(node.getWorkParent(), node);
    String editorid = UniEditor.ID;
    try {
      if (node.getType().equals(Document.TYPE_NAME)) editorid = ReqMarker.ID;
      IEditorReference eref;
      if (editorid == ReqMarker.ID) {
        input = new NodeEditorInput(node);
        eref = RequalityPlugin.findMarkEditorForDocument(node.getQualifiedId());
      } else
        eref =
            RequalityPlugin.findJSEditorForNode(
                node.getQualifiedId(), EclipseTracker.getResourceByNode(node).getProject());
      IEditorPart part;
      if (eref != null) part = eref.getEditor(true);
      else {
        IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
        part = IDE.openEditor(page, input, editorid, false);
      }
      if (editorid == UniEditor.ID) ((UniEditor) part).setSelectedNode(node);
      part.getSite().getPage().bringToTop(part);
      updateTargetRequirement();
      return part;
      //IDE.openEditor(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(), input, editorid, false);
    } catch (PartInitException e) {
      e.printStackTrace();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return null;
  }

  private boolean allow_popup_menu;

  private Menu createMenu() {
    MenuManager menuMgr = new MenuManager("#PopupMenu");
    menuMgr.setRemoveAllWhenShown(true);

    menuMgr.addMenuListener(
        new IMenuListener() {
          // menuAboutToShow вызывается 2 раза:
          // при mouseDown запрещаем создание меню, при mouseUp разрешаем
          public void menuAboutToShow(IMenuManager manager) {
            if (!allow_popup_menu) return;
            manager.add(new AddSelectionToRequirement(ReqMarker.this));

            Requirement req = getTargetRequirement();
            if (req != null && !req.getQualifiedId().equals(Requirement.getTypeRootQId())) {
              manager.add(new AddSelectionToRequirement(ReqMarker.this, req));
            }
            if (db == null) throw new RuntimeException("DB is null!");
            if (req == null) {
              req = (Requirement) db.getNode(Requirement.getTypeRootQId(), Path.QID);
            }
            //add check of possibility to add requirement
            //to req
            TreeLogic l = req.getTreeDB().getTreeLogic();
            if (l.canBeChildOf(Requirement.TYPE_NAME, req)
                && !req.isVirtual()
                && !VirtualNode.TYPE_NAME.equals(req.getType()))
              manager.add(
                  new AddSelectionToRequirement(
                      ReqMarker.this,
                      req,
                      RequalityCLI.getNextId(req, Requirement.TYPE_NAME, null)));

            String newName = browser.getSelectionText();
            Location loc = getSelectedLocation();
            if (loc != null
                && browser.isRemovable()
                && !browser.isSelectable()
                && loc.getDescription() != null) {
              newName = loc.getDescription();
            }
            newName = newName.trim();
            //add check of possibility to add requirement
            //to req
            if (!req.isVirtual()
                && newName.length() > 0
                && newName.length() < 128
                && l.canBeChildOf(Requirement.TYPE_NAME, req)) {
              manager.add(
                  new AddSelectionToRequirement(
                      ReqMarker.this,
                      req,
                      RequalityCLI.getNextId(req, Requirement.TYPE_NAME, null),
                      newName));
            }

            manager.add(new RemoveLocationFromRequirement(ReqMarker.this));

            if (CoreUtils.DEBUG_LEVEL > 0) {
              manager.add(
                  new Action("Show Version") {
                    @Override
                    public void run() {
                      browser.showBrowserInfo();
                    }
                  });
            }

            manager.add(
                new Action("Export Page...") {
                  @Override
                  public void run() {
                    String htmlContent = browser.serializeDocumentRaw();
                    htmlContent = htmlContent.replace(Activator.getDefault().getJSToolsPath(), "");
                    DirectoryDialog dialog = new DirectoryDialog(getSite().getShell());
                    String path = dialog.open();
                    if (path != null) {
                      try {
                        copyPageFile(path + File.separator);
                        Writer out =
                            new BufferedWriter(
                                new OutputStreamWriter(
                                    new FileOutputStream(path + File.separator + doc.getId()),
                                    "UTF8"));
                        out.write(htmlContent);
                        out.close();
                      } catch (IOException e) {
                        e.printStackTrace();
                      }
                    }
                  }

                  private void copyPageFile(String path) throws IOException {
                    String jsPath = Activator.getDefault().getJSToolsPath();
                    new File(path + File.separator + "jquery").mkdirs();
                    copyFile(
                        new File(jsPath + File.separator + "jquery" + File.separator + "jquery.js"),
                        new File(path + File.separator + "jquery" + File.separator + "jquery.js"));
                    copyFile(new File(jsPath + "edit.js"), new File(path + "edit.js"));
                    copyFile(new File(jsPath + "edit.css"), new File(path + "edit.css"));
                  }
                });
            if (loc != null) browser.activateLocation(loc.getUUId().toString());

            // manager.update(true);
          }
        });
    Menu menu = menuMgr.createContextMenu(browser);
    return menu;
  }

  // TODO: refactor - move to helpers class
  public static void copyFile(File in, File out) throws IOException {
    FileInputStream fis = new FileInputStream(in);
    FileOutputStream fos = new FileOutputStream(out);
    try {
      byte[] buf = new byte[1024];
      int i = 0;
      while ((i = fis.read(buf)) != -1) {
        fos.write(buf, 0, i);
      }
    } catch (IOException e) {
      throw e;
    } finally {
      fis.close();
      fos.close();
    }
  }

  public void onFocusActions() {
    new UndoRedoActionGroup(getSite(), getUndoContext(), true)
        .fillActionBars(getEditorSite().getActionBars());
    onSelectedLocationChangeByClick();
  }

  @Override
  public void setFocus() {
    browser.setFocus();
    if (browser.isReady()) onFocusActions();
  }

  public TreeDB getTreeDB() {
    return db;
  }

  public Document getDocument() {
    return doc;
  }

  /**
   * <code>addSelectionTo</code> must be called in transaction
   *
   * @param req
   * @param oid
   * @return array of objects. [0] - selection description, [1] - location node
   * @throws UnsupportedEncodingException
   * @throws CoreException
   */
  public Object[] addSelectionTo(Requirement req, String oid)
      throws UnsupportedEncodingException, CoreException {
    // req.getTreeDB().startTransaction();
    // move requirement creation after
    // location creation because of errors
    // in rollback if requirement is created and
    // location is not
    UUID uid = UUID.randomUUID();
    if (oid != null) uid = UUID.fromString(oid); //may cause stranges for non uuid string
    String locId = uid.toString();
    String desc = null;
    String undo = null;
    if (oid != null) {
      locId = oid;
    } else {
      Object[] jsData = browser.createLocation(locId);
      //method to check is location is not created
      if (jsData == null) {
        return null;
      }
      desc = (String) jsData[0];
      undo = (String) jsData[1];
      saveDocument(undo);
    }
    Location loc = doc.createLocation(uid);
    if (desc == null) desc = doc.getLocationsDescriptions().get(uid);
    doc.commitAttributes();
    List<File> added = new LinkedList<>();
    String val = RequalityCLI.updateImages(doc.getNodeResourcesPath(), desc, added);
    HtmlValueProvider hvp = new HtmlValueProvider(AttributeType.HTML, val,null,new File(doc.getNodeResourcesPath()),added);
    loc.setDDescription(hvp);
    loc.commitAttributes();
    req.addLocation(loc.getUUId());
    req.commitAttributes();
   // loc.resetRelated();
    // req.getTreeDB().commit();
    //doc.saveDOM();
    return new Object[] {undo, loc};
  }

  @Override
  public void dispose() {
    //saveAction();
    if (db != null) {
      db.removeGUIListener(treeChangeListener);
      db.removeTransactionListener(this);
    }
    EclipseTracker.getTracker().removeListener(dbStateListener);
    getSite().setSelectionProvider(null);
    super.dispose();
  }

  private void saveAction() {
    if (doc != null) RLogger.logInfo("saving! " + doc.getQualifiedId());
    String htmlContent = getBrowser().lastContent;
    if (htmlContent != null && htmlContent.length() > 0) {
      ByteArrayInputStream in;
      try {
        in = new ByteArrayInputStream(htmlContent.getBytes(DEFAULT_BASE_ENCODING));
        if (!doc.isDisposed()) doc.setResourceContent(doc.getResourceName(), in, null);
        in.close();
        //doc.saveDOM(null, Document.loadDOM(in));
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  public Requirement getTargetRequirement() {
    //    List<TreeNode> nodes = Activator.getDefault().getActiveNodes(
    //          Requirement.TYPE_NAME);
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes(EventSource.EXPLORER);
    Requirement req = null;
    if (nodes == null) return null;
    for (TreeNode node : nodes) {
      if (Requirement.TYPE_NAME.equals(node.getType()))
        if (req == null) req = (Requirement) node;
        else return null;
    }
    if(req == null && nodes.size() == 0)
      req = (Requirement) doc.getTreeDB().getNode(Requirement.getTypeRootQId());
    else
      if (nodes.size() != 1) return null;
    //TODO fix it to many case
    if (req != null) {
      return req;
    }
    return null;
  }

  public Location getHighlightLocation() {
    String locId = browser.getLocationOnSelection();
    if (locId != null) {
      return doc.getLocation(locId);
    }
    return null;
  }

  public Location getSelectedLocation() {
    Location loc = null;
    String locId = browser.getLocationOnSelection();
    if (locId != null) {
      loc = doc.getLocation(locId);
    }
    return loc;
  }

  public void close(boolean save) {
    if (save) saveAction();
    getSite().getPage().closeEditor(this, save);
  }

  @SuppressWarnings("unchecked")
  @Override
  public Object getAdapter(@SuppressWarnings("rawtypes") Class required) {
    if (required.equals(this.getClass())) {
      return this;
    }
    if (required.equals(ITextEditor.class)) {
      return null;
    }
    if (IContentOutlinePage.class.equals(required)) {
      if (reqMarkerOutlinePage == null) {
        reqMarkerOutlinePage = new ReqMarkerOutlinePage(this);
        reqMarkerOutlinePage.setInput(doc);
        reqMarkerOutlinePage.addSelectionChangedListener(locChangeListener);
      }
      return reqMarkerOutlinePage;
    }

    if (required == IPropertySheetPage.class) {
      return new TabbedPropertySheetPage(this);
    }
    if (required == IContributedContentsView.class) {
      return new IContributedContentsView() {
        public IWorkbenchPart getContributingPart() {
          return PlatformUI.getWorkbench()
              .getActiveWorkbenchWindow()
              .getActivePage()
              .findView("com.unitesk.requality.views.RequalityExplorer");
        }
      };
    }

    return super.getAdapter(required);
  }

  @Override
  public String getContributorId() {
    return "com.unitesk.requality.views.RequalityPropertyContributor";
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Selection provider
  //
  // //////////////////////////////////////////////////////////////////////////
  private List<ISelectionChangedListener> selectionChangeListeners =
      new ArrayList<>();

  public void onSelectedLocationChangeByClick() {
    if (reqMarkerOutlinePage != null && !reqMarkerOutlinePage.getControl().isDisposed()) {
      selFlag = false;
      String locationId = browser.getLocationOnSelection();
      if (locationId == null || locationId.isEmpty()) {
        fireSelectionChange();
        reqMarkerOutlinePage.setSelection(StructuredSelection.EMPTY);
        return;
      }
      Location location = doc.getLocation(locationId);
      if (location == null) {
        reqMarkerOutlinePage.setSelection(new StructuredSelection("missing: " + locationId));
        return;
      }
      reqMarkerOutlinePage.setSelection(new StructuredSelection(location));
      selFlag = true;
      fireSelectionChange();
      synchWithExplorer();
      setTargetLocation(location, true);
    }
  }

  @Override
  public void addSelectionChangedListener(ISelectionChangedListener listener) {
    if (listener != null) {
      selectionChangeListeners.add(listener);
    }
  }

  @Override
  public void removeSelectionChangedListener(ISelectionChangedListener listener) {
    if (listener != null) {
      selectionChangeListeners.remove(listener);
    }
  }

  @Override
  public ISelection getSelection() {
    String locationId = null;

    if (browser.isReady()) {
      locationId = browser.getLocationOnSelection();
      if (locationId != null) {
        Location location = doc.getLocation(locationId);
        if (location != null) {
          IStructuredSelection selection = new StructuredSelection(location);
          return selection;
        } else return new StructuredSelection("missing: " + locationId);
      }
      return new StructuredSelection(doc);
    }
    ISelection selection =
        PlatformUI.getWorkbench()
            .getActiveWorkbenchWindow()
            .getSelectionService()
            .getSelection(RequalityCNF.ID);
    return selection;
  }

  @Override
  public void setSelection(ISelection selection) {
    // TODO remove
    RLogger.logInfo("setSelection: " + selection);
  }

  public void fireSelectionChange() {
    ISelection sel = this.getSelection();
    SelectionChangedEvent event = new SelectionChangedEvent(this, sel);
    for (ISelectionChangedListener listener : selectionChangeListeners) {
      listener.selectionChanged(event);
    }
  }

  public boolean isLinked() {
    IViewPart view =
        Activator.getDefault()
            .getWorkbench()
            .getActiveWorkbenchWindow()
            .getActivePage()
            .findView(RequalityCNF.ID);
    if (view instanceof CommonNavigator) {
      CommonNavigator navigator = (CommonNavigator) view;
      return navigator.isLinkingEnabled();
    }
    return false;
  }

  public void synchWithExplorer() {
    // synchronization with Requality Explorer
    IViewPart view =
        Activator.getDefault()
            .getWorkbench()
            .getActiveWorkbenchWindow()
            .getActivePage()
            .findView(RequalityCNF.ID);
    if (view instanceof CommonNavigator) {
      CommonNavigator navigator = (CommonNavigator) view;
      SharedConsts.isNavigatorLinkingEnabled = navigator.isLinkingEnabled(); 
      if (navigator.isLinkingEnabled()) {
        navigator.setLinkingEnabled(true);
      }
    }
  }
  // //////////////////////////////////////////////////////////////////////////
  //
  // Undo/Redo support
  //
  // //////////////////////////////////////////////////////////////////////////
  public boolean redo(final String jsData) throws UnsupportedEncodingException, CoreException {
    ReqDocBrowser docBrowser = pullBrowser();
    IJSCommand redo =
        new IJSCommand() {
          @Override
          public void onError(Exception e) {
            fireSelectionChange();
            e.printStackTrace();
          }

          @Override
          public void onComplete(Object result, ReqDocBrowser docBrowser) {
            fireSelectionChange();
          }

          @Override
          public String getJS() {
            return "return redoAction(" + jsData + ");";
          }

          @Override
          public boolean isEvaluate() {
            return true;
          }
        };
    docBrowser.scheduleJSCommand(redo);
    return true;
  }

  public boolean undo(final String jsData) throws UnsupportedEncodingException, CoreException {
    ReqDocBrowser docBrowser = pullBrowser();
    IJSCommand undo =
        new IJSCommand() {
          @Override
          public void onError(Exception e) {
            fireSelectionChange();
            e.printStackTrace();
          }

          @Override
          public void onComplete(Object result, ReqDocBrowser docBrowser) {
            RLogger.logInfo("res " + result);
            fireSelectionChange();
          }

          @Override
          public String getJS() {
            return "return undoAction(" + jsData + ");";
          }

          @Override
          public boolean isEvaluate() {
            return true;
          }
        };
    docBrowser.scheduleJSCommand(undo);
    return true;
  }

  private ReqDocBrowser pullBrowser() throws PartInitException {
    ReqDocBrowser docBrowser = getBrowser();
    if (docBrowser.isDisposed()) {
      ReqMarker reqMarker;
      IEditorReference eref = RequalityPlugin.findMarkEditorForDocument(docQId);
      if (eref != null) {
        reqMarker = (ReqMarker) eref.getEditor(true);
      } else {
        IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
        reqMarker = (ReqMarker) IDE.openEditor(page, getEditorInput(), ReqMarker.ID, false);
      }
      reqMarker.getSite().getPage().bringToTop(reqMarker);
      docBrowser = reqMarker.getBrowser();
    }
    docBrowser.setFocus();
    docBrowser.addFocusListener(
        new FocusListener() {

          @Override
          public void focusLost(FocusEvent e) {
            // TODO Auto-generated method stub

          }

          @Override
          public void focusGained(FocusEvent e) {
            ReqMarker.this.onFocusActions();
          }
        });
    return docBrowser;
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Transaction listener
  //
  // //////////////////////////////////////////////////////////////////////////

  @Override
  public void execute(TransactionStorage tStorage) {}

  @Override
  public void redo(TransactionStorage tStorage) {
    try {
      for (OperationData opData : tStorage.getLastOperationsData()) {
        if (opData.getOperation() instanceof ChangeResource) {
          ChangeResource op = (ChangeResource) opData.getOperation();
          if (isCurrentDocOp(op)) {
            Object op_data = op.getData();
            if (op_data instanceof ProcessorData) {
              getEditorSite().getPage().closeEditor(ReqMarker.this, false);
              doc.setXmlDoc(doc.loadDOM());
              doc.saveDOMFromXMLCache();
              IWorkbenchPage page =
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
              IDE.openEditor(page, getEditorInput(), ReqMarker.ID, false);
            } else {
              redo((String) op_data);
            }
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  @Override
  public void undo(TransactionStorage tStorage) {
    try {
      for (OperationData opData : tStorage.getLastOperationsData()) {
        if (opData.getOperation() instanceof ChangeResource) {
          ChangeResource op = (ChangeResource) opData.getOperation();
          if (isCurrentDocOp(op)) {
            Object op_data = op.getData();
            if (op_data instanceof ProcessorData) {
              getEditorSite().getPage().closeEditor(ReqMarker.this, false);
              //doc.resetCache();
              IWorkbenchPage page =
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
              IDE.openEditor(page, getEditorInput(), ReqMarker.ID, false);
            } else {
              undo((String) op_data);
            }
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private boolean isCurrentDocOp(ChangeResource op) {
    return op.getNodeUUID().equals(doc.getUUId())
        && doc.getResourceName().endsWith(op.getResourceName())
        && op.getData() != null;
  }

  @Override
  public void _dispose() {}

  public String getSelectedLocationId() {
    return browser.getSelectedLocationId();
  }

  public Document getRootNode() {
    return doc;
  }

  public Location getLocationOnSelection() {
    Location loc = null;
    String locId = browser.getLocationOnSelection();
    if (locId != null) {
      loc = doc.getLocation(locId);
    }
    return loc;
  }
}
