/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.editors.browser;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.UUID;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.Viewer;

import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;

final class ReqMarkerOutlineContentProvider implements ITreeContentProvider {

  @Override
  public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {}

  @Override
  public void dispose() {}

  @Override
  public boolean hasChildren(Object element) {
    return getChildren(element).length > 0;
  }

  @Override
  public Object getParent(Object element) {
    return null;
  }

  @Override
  public Object[] getElements(Object inputElement) {
    return getChildren(inputElement);
  }

  @Override
  public Object[] getChildren(Object parentElement) {
    if (parentElement instanceof Document) {
      final Document doc = (Document) parentElement;
      List<UUID> locs = new ArrayList<>(doc.getLocations());
      Collections.sort(
          locs,
          new Comparator<UUID>() {
            @Override
            public int compare(UUID o1, UUID o2) {
              UUID u1 = o1, u2 = o2;
              if (u1 != null && u2 != null) {
                TreeNode t1 = doc.getLocation(u1.toString()), t2 = doc.getLocation(u2.toString());
                if (t1 == null) return t2 != null ? -1 : 0;
                if (t2 == null) return 1;
                return t1.getIndex().compareTo(t2.getIndex());
              }
              return u1 == null ? -1 : 1;
            }
          });
      Object[] result = new Object[locs.size()];
      int i = 0;
      for (UUID uid : locs) {
        Location loc = doc.getLocation(uid.toString());
        if (loc != null) {
          result[i++] = loc;
        } else {
          result[i++] = "missing: " + uid.toString();
        }
      }
      return result;
    }
    if (parentElement instanceof Location) {
      return new Object[] {};
    }
    if (parentElement instanceof String) {
      return new Object[] {};
    }
    return null;
  }
}
