/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.editors.browser;

import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchPartSite;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.operations.UndoRedoActionGroup;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.nodetypes.Document;

public class ReqMarkerOutlinePage extends ContentOutlinePage
    implements ITabbedPropertySheetPageContributor, IAdaptable {
  private Document input;
  private ReqMarker reqMarker;

  private IDoubleClickListener doubleClickListener =
      new IDoubleClickListener() {
        @Override
        public void doubleClick(DoubleClickEvent event) {
          IHandlerService handlerService =
              PlatformUI.getWorkbench()
              .getActiveWorkbenchWindow()
              .getActivePage()
              .getActivePart()
              .getSite()
              .getService(IHandlerService.class);
          try {
            handlerService.executeCommand("com.unitesk.requality.commands.OpenNode", null);
          } catch (Exception e) {
            e.printStackTrace();
          }
        }
      };

  public ReqMarkerOutlinePage(ReqMarker reqMarker) {
    this.reqMarker = reqMarker;
  }

  @Override
  public void createControl(Composite parent) {
    super.createControl(parent);
    getTreeViewer().setContentProvider(new ReqMarkerOutlineContentProvider());
    getTreeViewer()
        .setLabelProvider(
            new LabelProvider() {
              ILabelProvider lp = WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider();

              @Override
              public String getText(Object element) {
                if (element instanceof String) {
                  return (String) element;
                }
                if (lp != null) {
                  return lp.getText(element);
                }
                return super.getText(element);
              }

              @Override
              public Image getImage(Object element) {
                if (lp != null) {
                  return lp.getImage(element);
                }
                return super.getImage(element);
              }
            });
    getTreeViewer().setInput(input);
    getTreeViewer().addDoubleClickListener(doubleClickListener);
    getTreeViewer()
        .getControl()
        .addKeyListener(
            new KeyListener() {
              @Override
              public void keyReleased(KeyEvent e) {}

              @Override
              public void keyPressed(KeyEvent e) {
                if (e.keyCode == SWT.DEL) {
                  new RemoveLocationFromRequirement(reqMarker).run();
                  e.doit = false;
                }
              }
            });

    if (PlatformUI.getWorkbench().getActiveWorkbenchWindow() != null) {
      IWorkbenchPart part =
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart();
      if (part != null) {
        IWorkbenchPartSite site = part.getSite();
        new UndoRedoActionGroup(site, reqMarker.getUndoContext(), true)
            .fillActionBars(getSite().getActionBars());
      }
    }
  }

  public void setInput(Object input) {
    this.input = (Document) input;
    if (getTreeViewer() != null) {
      getTreeViewer().setInput(input);
    }
  }

  public void refresh() {
    getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                //if control is disposed old code may throw exception
                if ((ReqMarkerOutlinePage.this.getControl() != null)
                    && !ReqMarkerOutlinePage.this.getControl().isDisposed()) {
                  if (!getTreeViewer().isBusy()) getTreeViewer().refresh();
                }
              }
            });
  }

  @Override
  public void dispose() {
    super.dispose();
  }

  @Override
  public String getContributorId() {
    return "com.unitesk.requality.views.RequalityPropertyContributor";
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  public Object getAdapter(Class required) {
    // if (required == IContributedContentsView.class) {
    // return new IContributedContentsView() {
    // public IWorkbenchPart getContributingPart() {
    // return PlatformUI
    // .getWorkbench()
    // .getActiveWorkbenchWindow()
    // .getActivePage()
    // .findEditor(input);
    // }
    // };
    // }
    if (required == IPropertySheetPage.class) {
      return new TabbedPropertySheetPage(this);
    }
    return Platform.getAdapterManager().getAdapter(this, required);
  }
}
