package com.unitesk.requality.eclipse.handlers;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.nodetypes.Requirement;

public class AddReferenceTo extends AbstractHandler {

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    nodes
        .get(0)
        .getTreeDB()
        .startTransaction("Add reference to requirement" + (nodes.size() != 1 ? "s" : ""));
    for (final TreeNode node : nodes) {
      TreeNode rnode = node.getTreeDB().getRefTarget();
      if (rnode != null && node.getTreeDB().getRefTargetState(rnode) > 0) {
        //TODO mb templ?
        node.attrs()
            .setAttribute(
                new AttributeDefinition(node.getUUId(), 
                  node.getTreeDB().getDefReferenceName(),
                  new ReferenceValueProvider(AttributeType.REFERENCE,rnode.getUserVisibleName(true)),
                  AttributeDefinition.detectFlags(node.getTreeDB().getDefReferenceName())));
        node.attrs().getLinkAttributeValue(node.getTreeDB().getDefReferenceName());
        node.commitAttributes();
        if (node.getTreeDB().getRefTargetState(rnode) == 1) {
          node.getTreeDB().setRefTarget(null);
          Map<String, Attribute> old = new HashMap<>();
          Attribute oa = old.put(
              node.getTreeDB().getDefReferenceName(),
              rnode.attrs().getAttribute(node.getTreeDB().getDefReferenceName()));
          node.getTreeDB()
              .fireNodeGUIChange(
                  rnode,
                  new ChangesSet<>(Collections.emptySet(),AttributeStorage.formSet(node.getTreeDB().getDefReferenceName()),Collections.emptySet(),
                      AttributeStorage.formAttributeMap(oa),
                      AttributeStorage.formAttributeMap(old.get(node.getTreeDB().getDefReferenceName())),
                      AttributeStorage.formList(node.getTreeDB().getDefReferenceName())),
                  true);
        }
      }
    }
    nodes.get(0).getTreeDB().commit();
    return null;
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (!super.isEnabled()) return false;
    if (!nodes.get(0).getType().equals(Requirement.TYPE_NAME)) return false;
    TreeDB db = nodes.get(0).getTreeDB();
    for (TreeNode node : nodes)
      if (!node.getTreeDB().equals(db) || !node.getType().equals(Requirement.TYPE_NAME))
        return false;
    TreeNode rt = db.getRefTarget();
    return db.getRefTargetState(rt) > 0;
  }
}
