package com.unitesk.requality.eclipse.handlers;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;

import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.ErrorNode;

public class DeleteErrorNode extends AbstractHandler implements IDeleteHandler {

  public void recUpdateProcessed(TreeNode node, List<UUID> toprocess) {
    toprocess.add(node.getUUId());
    for (UUID uid : node.getWorkChildrenUUIDs()) {
      TreeNode nde = node.getTreeDB().getNode(uid);
      recUpdateProcessed(nde, toprocess);
    }
  }

  @Override
  public boolean isSuitableFor(TreeNode tn) {
    return tn.getType().equals(ErrorNode.TYPE_NAME);
  }

  @Override
  public void process(List<TreeNode> nodes, List<UUID> processed, boolean multi)
      throws ExecutionException {
    for (TreeNode node : nodes) {
      if (processed.contains(node.getUUId())) continue;
      if (!node.getTreeDB().getTreeLogic().canBeDeleted(node)) continue;
      if (!isSuitableFor(node)) continue;
      List<UUID> toprocess = new ArrayList<>();
      recUpdateProcessed(node, toprocess);
      node.delete();
      processed.addAll(toprocess);
    }
  }

  @Override
  public String[] getWarnings(List<TreeNode> nodes) {
    return null;
  }

  @Override
  public String titleOverride(List<TreeNode> nodes) {
    return null;
  }

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    return null;
  }
}
