/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.UUID;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.IElementUpdater;
import org.eclipse.ui.menus.UIElement;
import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.VirtualNode;

public class DeleteSelection extends AbstractHandler implements IHandler, IElementUpdater {
  //maybe move to TreeLogic
  private static Set<String> avoidDeleteTransactionTypes = new TreeSet<>();

  static {
    avoidDeleteTransactionTypes.add(Document.TYPE_NAME);
    avoidDeleteTransactionTypes.add(DocFolder.TYPE_NAME);
  }

  @SuppressWarnings("rawtypes")
  @Override
  public void updateElement(UIElement element, Map parameters) {
    element.setText(getLabel(SelectionManager.getOne().getSelectedNodes()));
  }

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes(true);
    deleteNodes(nodes);
    return null;
  }

  public static String getLabel(List<TreeNode> nodes) {
    if (nodes == null || nodes.size() < 1) return "Delete Selection";
    boolean adds = nodes.size() != 1;
    TreeNode nd = nodes.get(0);
    boolean sim = true;
    for (TreeNode node : nodes)
      if (!node.getType().equals(nd.getType())) {
        return "Delete Selection";
      }
    SortedMap<String, IDeleteHandler> dhlist = Activator.getDefault().getDeleteHandlers();
    String overrideTitle = null;
    Iterator<String> hit = dhlist.keySet().iterator();
    while (hit.hasNext()) {
      String dhname = hit.next();
      IDeleteHandler idh = dhlist.get(dhname);
      String msg = idh.titleOverride(nodes);
      if (msg != null) overrideTitle = msg;
    }
    String name = overrideTitle;
    if (name == null)
      name =
          sim ? (adds ? nd.getMultiTypeLabel() : nd.getTypeLabel()) : "Element" + (adds ? "s" : "");
    if (name.equals("Selection")) return "Delete Selection";
    return "Delete " + name;
  }

  public static void deleteNodes(List<TreeNode> nodes) throws ExecutionException {
    if (nodes == null || nodes.size() < 1) return;
    boolean adds = nodes.size() != 1;
    boolean sim = true;
    List<String> startedids = new ArrayList<>();
    Set<String> msges = new HashSet<>();
    Iterator<TreeNode> it = nodes.iterator();
    Set<TreeNode> toadd = new HashSet<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      if (node.isVirtual() && VirtualNode.TYPE_NAME.equals(node.getWorkParent().getType())) {
        if (((VirtualNode) node.getWorkParent()).isCollapsed()
            && node.getWorkParent().getWorkChildrenUUIDs().size() == 1) {
          it.remove();
          if (!node.isVirtual())
            node = Instantiator.getRootTree(node.getTreeDB()).getNode(node.getUUId());
          toadd.add(node.getWorkParent());
        }
      }
    }
    if (!toadd.isEmpty()) {
      toadd.addAll(nodes);
      nodes.clear();
      nodes.addAll(toadd);
    }

    TreeNode nd = nodes.get(0);
    for (TreeNode node : nodes)
      if (!node.getType().equals(nd.getType())) {
        sim = false;
        break;
      }
    SortedMap<String, IDeleteHandler> dhlist = Activator.getDefault().getDeleteHandlers();
    SortedMap<String, SortedMap<String, List<TreeNode>>> list =
        new TreeMap<>();
    //format like so:
    // TreeDB1
    //      -Type
    //          -node1
    //          -node2
    //      -Type2
    //      ...
    // TreeDB2
    // ...
    String overrideTitle = null;
    Iterator<String> hit = dhlist.keySet().iterator();
    StringBuilder ln = new StringBuilder();
    while (hit.hasNext()) {
      String dhname = hit.next();
      IDeleteHandler idh = dhlist.get(dhname);
      String[] rmsg = idh.getWarnings(nodes);
      String msg = idh.titleOverride(nodes);
      if (msg != null) overrideTitle = msg;
      if (rmsg == null) continue;
      for (String m : rmsg) if (!msges.contains(m)) msges.add(m);
      for (TreeNode node : nodes) {
        Map<String, Set<TreeNode>> links = node.getTreeDB().getLinkResolver().getRevertLinks(node);
        int cnt = 0;
        if(links!=null)
          for (String s : links.keySet()) if (!s.startsWith("_")) cnt++;
        if (cnt > 0) {
          if (ln.length() > 0) ln.append(", ");
          ln.append("'");
          ln.append(node.getUserVisibleName(true));
          ln.append("'");
        }
      }
    }
    String linkNodes = ln.toString();
    if (!linkNodes.isEmpty())
      msges.add("Some of selected nodes (" + linkNodes + ") are linked to another nodes.");
    SortedMap<String, TreeDB> bases = new TreeMap<>();
    SortedMap<String, WriteLock> basesLocks = new TreeMap<>();
    SortedMap<String, List<TreeNode>> nodesbytypes = new TreeMap<>();
    for (TreeNode nc : nodes) {
      if (!nodesbytypes.containsKey(nc.getType()))
        nodesbytypes.put(nc.getType(), new LinkedList<TreeNode>());
      nodesbytypes.get(nc.getType()).add(nc);
      String dbname = nc.getTreeDB().getName();
      if (!bases.containsKey(dbname))
        bases.put(dbname, nc.isVRoot() ? Instantiator.getBaseTree(nc.getTreeDB()) : nc.getTreeDB());
      if (!list.containsKey(dbname)) list.put(dbname, new TreeMap<String, List<TreeNode>>());
      Iterator<String> dhit = dhlist.keySet().iterator();
      while (dhit.hasNext()) {
        String dhname = dhit.next();
        IDeleteHandler idh = dhlist.get(dhname);
        if (idh.isSuitableFor(nc)) {
          if (!list.get(dbname).containsKey(dhname))
            list.get(dbname).put(dhname, new ArrayList<TreeNode>());
          list.get(dbname).get(dhname).add(nc);
        }
      }
    }
    String name = overrideTitle, warns = "";
    if (name == null)
      name =
          sim ? (adds ? nd.getMultiTypeLabel() : nd.getTypeLabel()) : "Element" + (adds ? "s" : "");
    if (msges.size() > 0) for (String s : msges) warns += "- " + s + "\n";
    int tag = warns.equals("") ? MessageDialog.QUESTION : MessageDialog.WARNING;
    MessageDialog md =
        new MessageDialog(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Delete " + name,
            MessageDialog.getImage(MessageDialog.DLG_IMG_MESSAGE_INFO),
            "Delete selected "
                + name
                + "?"
                + (warns.equals("")
                    ? ""
                    : "\n\nWarning!\n"
                        + warns
                        + "\nIf you still want to proceed click OK, otherwise click Cancel."),
            tag,
            new String[] {"OK", "Cancel"},
            0);
    int k = md.open();
    if (k != 0) return;
    Iterator<String> dbnameit = list.keySet().iterator();
    List<UUID> deleted = new ArrayList<>();
    List<String> rowstarted = new LinkedList<>();
    while (dbnameit.hasNext()) {
      boolean firstrtansaction = true, commited = false;
      String dbname = dbnameit.next();
      bases.get(dbname).startChangeRow();
      rowstarted.add(dbname);
      try {
        Iterator<String> idhit = list.get(dbname).keySet().iterator();
        //first pass - node types with transactions
        while (idhit.hasNext()) {
          String idname = idhit.next();
          List<TreeNode> pnodes = list.get(dbname).get(idname);
          boolean avoidtransaction = avoidDeleteTransactionTypes.contains(pnodes.get(0).getType());
          if (!avoidtransaction) {
            if (firstrtansaction && pnodes.size() > 0) {
              bases.get(dbname).startTransaction("Delete " + name);
              startedids.add(dbname);
              firstrtansaction = false;
            }
            dhlist.get(idname).process(pnodes, deleted, nodes.size() != 1);
          }
        }
        if (firstrtansaction == false) {
          bases.get(dbname).commit();
          startedids.remove(dbname);
          commited = true;
        }
        idhit = list.get(dbname).keySet().iterator();
        //second pass - node types without transactions
        while (idhit.hasNext()) {
          String idname = idhit.next();
          List<TreeNode> pnodes = list.get(dbname).get(idname);
          boolean avoidtransaction = avoidDeleteTransactionTypes.contains(pnodes.get(0).getType());
          if (avoidtransaction) dhlist.get(idname).process(pnodes, deleted, nodes.size() != 1);
        }

      } catch (Exception e) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Delete error",
            "Operation is interrupted because of error: " + e + "\n" + e.getStackTrace()[0]);
        e.printStackTrace();
        if (!firstrtansaction) {
          if (!commited && startedids.size() > 0)
            for (String s : startedids) bases.get(s).rollback();
          else {
            IUndoContext cntx = PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
            if (PlatformUI.getWorkbench().getOperationSupport().getOperationHistory().canUndo(cntx))
              PlatformUI.getWorkbench()
                  .getOperationSupport()
                  .getOperationHistory()
                  .undo(cntx, null, null);
          }
        }
      }
    }
    for (String s : rowstarted) bases.get(s).stopChangeRow();
  }

  boolean recEnabled = false;

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null || nodes.size() < 1) return false;
    TreeDB db = nodes.get(0).getTreeDB();
    String t1 = nodes.get(0).getType();
    int cnt = 0;
    for (TreeNode node : nodes) {
      if (!node.getTreeDB().getTreeLogic().canBeDeleted(node)) {
        cnt++;
        continue;
      }
      if (db != node.getTreeDB()) return true;
      if (!t1.equals(node.getType())) return true;
    }
    if (cnt == nodes.size()) return false;
    return super.isEnabled();
  }
}
