package com.unitesk.requality.eclipse.handlers;

import java.io.File;
import java.io.FilenameFilter;
import java.net.URI;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;

public class DeleteStatisticsFileHandler extends AbstractHandler {

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    IStructuredSelection sel =
        (IStructuredSelection)
            PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow()
                .getSelectionService()
                .getSelection();
    IProject project = (IProject) sel.getFirstElement();
    TreeNode tn = EclipseTracker.getTracker().getTree(project).getRootNode();
    URI path = tn.getNodeResourcesPath();
    path = new File(path).getParentFile().toURI();

    try {
      File[] list =
          new File(path)
              .listFiles(
                  new FilenameFilter() {
                    public boolean accept(File dir, String name) {
                      return name.toLowerCase().startsWith("statistics")
                          && name.toLowerCase().endsWith(".xml");
                    }
                  });

      for (int i = 0; i < list.length; i++) {
        if (list[i].delete()) {
          RLogger.logInfo(list[i].getName() + " is deleted!");
        } else {
          RLogger.logInfo("Delete operation is failed.");
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public boolean isEnabled() {
    IStructuredSelection sel =
        (IStructuredSelection)
            PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow()
                .getSelectionService()
                .getSelection();
    if (sel.getFirstElement() instanceof IProject) {
      IProject project = (IProject) sel.getFirstElement();
      try {
        if (!project.isAccessible()) return false;
        if (project.hasNature(ReqCatalogue.ID)) {
          TreeDB db = EclipseTracker.getTracker().getTree(project);
          return !db.isVirtual() && db.getNode("/Versions", Path.QID) == null;
        }
      } catch (CoreException e) {
        e.printStackTrace();
      }
    }
    return false;
  }
}
