package com.unitesk.requality.eclipse.handlers;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.report.ReportGenerator;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;
import com.unitesk.requality.tools.LibreCLI;
import com.unitesk.requality.tools.RequalityCLI;

public class EditInLibreOffice extends AbstractHandler {
  boolean cnsl = false;

  private static void recDelete(File f, boolean force) {
    if (f.isDirectory()) for (File f2 : f.listFiles()) recDelete(f2, force);
    else if (force) f.delete();
    else f.deleteOnExit();
  }

  String librepath;

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null || nodes.size() == 0) return null;
    //nodes.get(0).getTreeDB().getNode("/Reports",com.unitesk.requality.core.Path.QID);
    ResourceStorage rs = EclipseTracker.getResourceStorage(nodes.get(0).getTreeDB());
    IFolder tempFolder = rs.getResourceFolder(nodes.get(0).getTreeDB().getRootNode().getUUId());
  
    if (!tempFolder.exists()) {
      try{
        tempFolder.refreshLocal(IResource.DEPTH_INFINITE, null);
      }catch(Exception e){
        
      }
      if (!tempFolder.exists()) {
        try {
          tempFolder.create(false, true, null);
        } catch (CoreException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
    IWorkbench w = PlatformUI.getWorkbench();
    cnsl = false;
    librepath = LibreCLI.getDefault().getPath();
    if (librepath == null || librepath.isEmpty())
      w.getDisplay()
          .syncExec(
              new Runnable() {

                @Override
                public void run() {
                  Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  InputDialog id =
                      new InputDialog(
                          shell,
                          "Enter path to libreoffice",
                          "Please enter path to libreoffice to corresponding field.",
                          librepath,
                          null);
                  if (id.open() == InputDialog.OK) {
                    librepath = id.getValue();
                    if (librepath.startsWith("\"")) librepath = librepath.substring(1);
                    if (librepath.endsWith("\""))
                      librepath = librepath.substring(0, librepath.length() - 1);

                  } else cnsl = true;
                }
              });
    if (cnsl) {
      return null;
    }

    Map<String, Object> map = new HashMap<>();
    map.put(
        "attrFilter",
        new ArrayList<String>() {
          /** */
          private static final long serialVersionUID = 5462662775364582819L;

          {
            add("Ref");
          }
        });
    map.put(VirtualNode.ATTR_COLLAPSEVNODE, true);
    try {
      File tempfl =
          new File(
              EclipseTracker.getResourceStorage(nodes.get(0).getTreeDB())
                      .getResourceFolder(nodes.get(0).getTreeDB().getRootNode().getUUId())
                      .getLocationURI()
                      .toURL()
                      .getPath()
                  + IPath.SEPARATOR
                  + "tempFolder"
                  + System.currentTimeMillis());
      tempfl.mkdirs();
      if (tempfl.exists()) recDelete(tempfl, true);
      Template t =
          Activator.getDefault()
              .getReportTemplates()
              .get("com.unitesk.requality.reports.reqSpecification");
      try {
        TreeDB db = nodes.get(0).getTreeDB(); 
        for (TemplateFile tm : t.getTemplates()) {
          File f =
              new File(
                  tempfl.getAbsolutePath()
                      + Path.SEPARATOR
                      + tm.getOutputFileName(db));
          if (f.exists()) f.delete();
        }
        ReportGenerator.generateReport(
            (Requirement) nodes.get(0), t, tempfl.toURI().toString(), map);
      } catch (Exception e) {
        e.printStackTrace();
      }

      try {
        TreeDB db = nodes.get(0).getTreeDB();
        String name =
            t.getTemplates().get(0).getOutputFileName(db);
        Date dt = new Date();
        DateFormat dft = SimpleDateFormat.getDateInstance();

        String path =  tempfl.toURI().toString() + dft.format(dt) + name;
        //File f = new File(tempfl.getAbsolutePath()+Path.SEPARATOR+dft.format(dt)+name);
        RequalityCLI.copyfile(
            URI.create(tempfl.toURI().toURL().toString() + "/" + name), URI.create(path));
        LibreCLI.getDefault().setTempPath(librepath);
        LibreCLI.getDefault().handleCommand(false,false, path);
        //	    		w.getDisplay().syncExec(new Runnable() {
        //
        //	    			@Override
        //	    			public void run() {
        //	    				Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
        //	    				MessageBox mb = new MessageBox(shell,SWT.OK|SWT.CANCEL);
        //	    				mb.setMessage("Please press 'Ok' when you done and need to apply or 'Cancel' if you need to decline changes.");
        //	    				mb.setText("Edit in LibreOffice");
        //	    				if(mb.open()==SWT.OK){
        //
        //	    				}else
        //	    					cnsl = true;
        //	    			}
        //	    		});
        if (cnsl) {
          LibreCLI.getDefault().setTempPath(null);
          return null;
        }
        //	    		DocFolder df = (DocFolder) nodes.get(0).getTreeDB().getNode(Document.getTypeRootQId(),com.unitesk.requality.core.Path.QID);
        //	    		Document d = (Document) df.createChild(
        //	    				RequalityCLI.getNextId(df,Document.TYPE_NAME,""),Document.TYPE_NAME);
        //
        //	            UpdateFromHTMLModule up = new UpdateFromHTMLModule();
        //	    		up.setShowMessages(false);
        //
        //	    		up.configure(d);
        //
        //	    		d.attrs().setAttribute("_file", "temp.xhtml");
        //	    		d.saveAttributes();
        //	    		//d.setResourceContent(d.getResourceName(), new FileInputStream(new File(path.substring(1,path.length()-1))));
        //	    		//d.attrs().saveAttributes();
        //	    		OmnivorousImporter oi = new OmnivorousImporter();
        ////	    		String dpath = d.getResourcesPath().replace("file:", "");
        //	//    		if(dpath.startsWith("/")&&dpath.charAt(2)==':')
        //	  //  			dpath = dpath.substring(1);
        //
        //
        //                File outFile = new File(tempFolder.getLocation()
        //                        .toOSString(), "temp.xhtml");
        //	    		oi.importDocument(d.getQualifiedId(), new File(path), outFile,
        //	    				df.getTreeDB(), d);
        //	    		InputStream is = new BufferedInputStream(new FileInputStream(outFile));
        //                d.setResourceContent("temp.xhtml", is);
        //                is.close();
        //	    		d.saveAttributes();
        //
        //	    		nodes.get(0).getTreeDB().startChangeRow();
        //	    		up.process(d.getQualifiedId(), nodes.get(0), null);
        //	    		d.delete();
        //	    		nodes.get(0).getTreeDB().stopChangeRow();

        //now file edited)
        //p = Runtime.getRuntime().exec(new String[] {
        //	    			//	oscmd,"-c","libreoffice --headless --convert-to htm:HTML "+path+" --outdir \""+tempfl.getAbsolutePath()+"\""});
        //				r.exec("libreoffice \""+tempfl.getAbsolutePath()+Path.SEPARATOR+(name.replace(".xml", ".odt"))+"\"");

        if (tempfl.exists()) recDelete(tempfl, false);
      } catch (IOException e) {
      }
    } catch (Exception e1) {
      e1.printStackTrace();
    }
    LibreCLI.getDefault().saveSettings();
    LibreCLI.getDefault().setTempPath(null);
    return null;
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null || nodes.size() == 0) {
      return false;
    }
    for (TreeNode node : nodes) {
      if (!Requirement.TYPE_NAME.equals(node.getType())
          && !TextNode.TYPE_NAME.equals(node.getType())) return false;
    }
    return true;
  }
}
