package com.unitesk.requality.eclipse.handlers;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.jface.action.ContributionItem;
import org.eclipse.jface.action.ToolBarManager;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.ToolItem;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.ui.RequalityLCNF;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.marker.LinksDecorator;
import com.unitesk.requality.nodetypes.Requirement;

public class LightLinks extends AbstractHandler {

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    TreeNode node = RequalityLCNF.getSelectedNode();
    updateButtonLabel();
    if (RequalityLCNF.getTreeSelectedNode() == node) {
      node = null;
    }
    ColorNode(node);
    return null;
  }

  @Override
  public boolean isEnabled() {
    TreeNode node = RequalityLCNF.getSelectedNode();

    if (node == null) {
      if (LinksDecorator.isButtonEnable()) {
        updateButtonLabel();
      }
      return false;
    }
    if (!node.getType().equals(Requirement.TYPE_NAME)) return false;

    return true;
  }

  private void updateButtonLabel() {
    Display d = Display.getCurrent();
    if (d != null)
      d.syncExec(
          new Runnable() {
            @Override
            public void run() {
              RequalityLCNF view =
                  (RequalityLCNF)
                      PlatformUI.getWorkbench()
                          .getActiveWorkbenchWindow()
                          .getActivePage()
                          .findView(RequalityLCNF.ID);
              if (view != null) {
                ToolBarManager manager =
                    (ToolBarManager) view.getViewSite().getActionBars().getToolBarManager();
                ToolItem ti = null;
                for (ToolItem iti : manager.getControl().getItems()) {
                  Object o = iti.getData();
                  if (o != null && o instanceof ContributionItem) {
                    if ("com.unitesk.requality.commands.LightLinks"
                        .equals(((ContributionItem) o).getId())) ti = iti;
                  }
                }
                if (ti != null) {
                  if (LinksDecorator.isButtonEnable()) {
                    ti.setImage(
                        (Image)
                            Activator.getImageDescriptor("/icons/light_link.png")
                                .createResource(Display.getCurrent()));
                    LinksDecorator.setButtonEnable(false);
                  } else {
                    ti.setImage(
                        (Image)
                            Activator.getImageDescriptor("/icons/link_break.png")
                                .createResource(Display.getCurrent()));
                    LinksDecorator.setButtonEnable(true);
                  }
                }
              }
            }
          });
  }

  public static void ColorNode(TreeNode node) {
    ColorNode(node, null);
  }

  public static void ColorNode(TreeNode node, TreeNode oldNode) {
    Display.getCurrent()
        .syncExec(
            new Runnable() {

              @Override
              public void run() {
                RequalityLCNF viewLCNF =
                    (RequalityLCNF)
                        PlatformUI.getWorkbench()
                            .getActiveWorkbenchWindow()
                            .getActivePage()
                            .findView(RequalityLCNF.ID);

                if (viewLCNF != null) {

                  Object[] elms = viewLCNF.getCommonViewer().getExpandedElements();

                  if (node != null && oldNode != null) {
                    LinksDecorator.isColor(true);
                    if (oldNode.equals(node)) {
                      viewLCNF.getCommonViewer().refresh(node);
                    } else {
                      viewLCNF.getCommonViewer().refresh(node);
                      viewLCNF.getCommonViewer().refresh(oldNode);
                    }
                  } else if (node != null) {
                    LinksDecorator.isColor(true);
                    viewLCNF.getCommonViewer().refresh(node);
                  } else if (oldNode != null) {
                    viewLCNF.getCommonViewer().refresh(oldNode);
                    RequalityCNF viewCNF =
                        (RequalityCNF)
                            PlatformUI.getWorkbench()
                                .getActiveWorkbenchWindow()
                                .getActivePage()
                                .findView(RequalityCNF.ID);
                    if (viewCNF != null) {
                      viewCNF.refresh(oldNode);
                    }
                  }
                  viewLCNF.getCommonViewer().setExpandedElements(elms);
                }
              }
            });
  }
}
