package com.unitesk.requality.eclipse.handlers;

import java.util.List;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.PastingIntoLeafException;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

public abstract class NewElementHandler extends AbstractHandler {
  public String getElementType() {
    String[] types = getElementTypes();
    if (types != null && types.length == 1) return types[0];
    return null;
  }

  public TreeDB tree;

  public abstract String[] getElementTypes();

  public String getShowedElementType() {
    return getShowedElementType(getElementType());
  }

  public String getShowedElementType(String tp) {
    return tree.getNodeType(tp).getVisibleClassName();
  }

  protected String getNewIdentifier(TreeNode par, String type, boolean filterId) {
    return RequalityCLI.getNextId(par, type, null);
  }

  public <type extends TreeNode> type execute(TreeNode par) {
    return execute(par, null, getElementType());
  }

  @SuppressWarnings("unchecked")
  public <type extends TreeNode> type execute(TreeNode par, String templ, String type) {
    tree = par.getTreeDB();
    TreeDB db = par.getTreeDB();
    TreeNode node = null;
    if (!db.getTreeLogic().canBeChildOf(type, par.getType())
            && db.getTreeLogic().canBeChildOf(type, par.getWorkParent().getType())
        || RequalityCNF.shiftPressed) {
      node = par;
      par = par.getWorkParent();
    }

    String id = getNewIdentifier(par, type,true),name = getNewIdentifier(par, type,false);

    if (id == null) return null;
    try {
      if (!par.getTreeDB().getTreeLogic().canBeCreatedInExc(type, id, par)) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Create Element",
            "Element of type '"
                + getShowedElementType()
                + "' cannot be created in "
                + par.getUserVisibleName(true));
        return null;
      }
    } catch (PastingIntoLeafException e) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Create Element",
          "Can't add "
              + getShowedElementType()
              + " to "
              + par.getTypeLabel()
              + " with "
              + "Test Purpose child");
      return null;
    } catch (SameIdExistsInChildren e) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Create Element",
          "Element with same id exists in " + par.getUserVisibleName(true));
      return null;
    } catch (IdEmpty e) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Create Element",
          "Element id is empty");
      return null;
    } catch (IdContainsWrongChars e) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Create Element",
          "Element id contains wrong character - " + e.getWrongCharacter());
      return null;
    }
    if (db.startTransaction("Create " + getShowedElementType(type), true)) {
      Index ind = Index.getNextIndex(type, par, node, false);
      TreeNode nn = par.createChild(id, type, templ);
      if (!ind.isUndefined() && (db.getSortingSettings().getMethod() == SortingMethod.BY_INDEX)
          || db.getTreeLogic().usesIndexByDefault(nn.getType())) {
        nn.setIndex(ind);
      }
      setElementSpecificAttributes(nn);
      if(name != null && !name.equals(id))
        nn.setIgnoredName(name);
      nn.commitAttributes();
      db.commit();
      //	if(Activator.getDefault()!=null)
      //	Activator.getDefault().setActiveNode(nn, type);
      return (type) nn;
    }
    return null;
  }

  protected abstract void setElementSpecificAttributes(TreeNode nn);

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null || nodes.size() != 1) return null;
    return execute(nodes.get(0), null, getElementType());
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodelist = SelectionManager.getOne().getSelectedNodes();
    if (nodelist == null || nodelist.size() > 1) return false;
    if (nodelist.get(0).isVirtual()) return false;
    tree = nodelist.get(0).getTreeDB();
    if (!checkElementSpecificConditions(nodelist.get(0))) return false;
    for (String s : getElementTypes())
      if (nodelist.get(0).getTreeDB().getTreeLogic().canBeCreatedIn(s, nodelist.get(0)))
        return true;
    return false;
  }

  protected boolean checkElementSpecificConditions(TreeNode treeNode) {
    return true;
  }
}
