package com.unitesk.requality.eclipse.handlers;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.operations.mass.MassLocationsFixer;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Requirement;

public class TemporaryMassLocationFixer extends AbstractHandler {

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    IStructuredSelection sel =
        (IStructuredSelection)
            PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow()
                .getSelectionService()
                .getSelection();
    if (!(sel.getFirstElement() instanceof IProject)) return null;
    IProject project = (IProject) sel.getFirstElement();
    TreeDB db = EclipseTracker.getTracker().getTree(project);
    if (db == null) return null;
    try {
      MassLocationsFixer mlf = new MassLocationsFixer();
      TreeNode node1 = db.getNode(Document.getTypeRootQId());
      mlf.execute(node1); //first fix docs
      TreeNode node2 = db.getNode(Requirement.getTypeRootQId());
      mlf.execute(node2); //then fix reqs
    } catch (Exception ex) {
      ex.printStackTrace();
    }
    return null;
  }

  @Override
  public boolean isEnabled() {
    IStructuredSelection sel =
        (IStructuredSelection)
            PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow()
                .getSelectionService()
                .getSelection();
    if (sel.getFirstElement() instanceof IProject) {
      IProject project = (IProject) sel.getFirstElement();
      try {
        if (!project.isAccessible()) return false;
        if (project.hasNature(ReqCatalogue.ID)) {
          return true;
        }
      } catch (CoreException e) {
        e.printStackTrace();
      }
    }
    return false;
  }
}
