package com.unitesk.requality.eclipse.handlers.location;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.IElementUpdater;
import org.eclipse.ui.menus.UIElement;
import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.handlers.requirement.DeleteRequirementHandler;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class RemoveLocationHandler extends AbstractHandler
    implements IDeleteHandler, IElementUpdater {

  public void recUpdateProcessed(TreeNode node, List<UUID> toprocess) {
    toprocess.add(node.getUUId());
    for (UUID uid : node.getWorkChildrenUUIDs()) {
      TreeNode nde = node.getTreeDB().getNode(uid);
      recUpdateProcessed(nde, toprocess);
    }
  }

  @SuppressWarnings("rawtypes")
  @Override
  public void updateElement(UIElement element, Map parameters) {
    List<TreeNode> sel = SelectionManager.getOne().getSelectedNodes();
    int cnt = 0;
    for (TreeNode node : sel) if (node.getType().equals(Location.TYPE_NAME)) cnt++;
    element.setText("Delete Location" + ((cnt > 1) ? "s" : ""));
  }

  boolean batch = true;

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null) return false;
    batch = false;
    process(nodes, new ArrayList<UUID>(), nodes.size() != 1);
    return null;
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null) {
      return false;
    }
    for (TreeNode node : nodes) if (!node.getType().equals(Location.TYPE_NAME)) return false;
    return true;
  }

  @Override
  public boolean isSuitableFor(TreeNode tn) {
    return tn.getType().equals(Location.TYPE_NAME);
  }

  @Override
  public void process(List<TreeNode> nodes, List<UUID> processed, boolean multi)
      throws ExecutionException {
    Set<Requirement> req4check = new HashSet<>();

    for (TreeNode node : nodes) {
      if (!isSuitableFor(node)) continue;
      if (processed.contains(node.getUUId())) continue;
      Location loc = (Location) node;
      req4check.addAll(Arrays.asList(loc.getRequirements()));
      loc.delete();
      recUpdateProcessed(node, processed);
      //            List<UUID>toprocess = new ArrayList<UUID>();
      //
      //            CommonRemoveLocation.RemoveLocation(loc, null, null, batch);
      processed.add(loc.getUUId());
    }

    List<TreeNode> reqs = new ArrayList<>();
    String strList = "";
    for (Requirement req : req4check) {
      if (req.getLocations().size() == 0 && req.sizeWorkChildren() == 0) {
        reqs.add(req);
        if (!strList.equals("")) {
          strList += ", ";
        }
        strList += req.getNameOrId();
      }
    }
    if (reqs.size() > 0) {
      //TODO message
      if (reqs.get(0).getTreeDB().inTransaction()) return;
      if (MessageDialog.openQuestion(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Delete Requirement",
          (reqs.size() == 1
                  ? "Requirement \"" + reqs.get(0).getUserVisibleName(true) + "\" has "
                  : "Requirements " + strList + " have ")
              + "no locations, delete?")) {
        List<UUID> proc = new ArrayList<>();
        if (reqs.get(0)
            .getTreeDB()
            .startTransaction("Delete Requirement" + (reqs.size() > 1 ? "s" : ""), true)) {
          try {
            //TODO replace to reqs.delete
            new DeleteRequirementHandler().process(reqs, proc, true);
          } catch (ExecutionException e) {
            e.printStackTrace();
          }
          reqs.get(0).getTreeDB().commit();
        }
      }
    }
    //reqMarker.fireSelectionChange();
  }

  @Override
  public String[] getWarnings(List<TreeNode> nodes) {
    return null;
  }

  @Override
  public String titleOverride(List<TreeNode> nodes) {
    return null;
  }
}
