/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.handlers.report;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.handlers.NewElementHandler;
import com.unitesk.requality.nodetypes.ReportFolder;
import com.unitesk.requality.tools.RequalityCLI;

public class CreateReportFolderHandler extends NewElementHandler {
  @Override
  protected String getNewIdentifier(TreeNode parent, String type,boolean filter){ 
    if(!filter)
      return null;
    InputDialog dlg =
        new InputDialog(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "New Report Folder",
            "Enter folder name",
            RequalityCLI.getNextId(parent, type, "new_folder_"),
            new IInputValidator() {
              @Override
              public String isValid(String newText) {
                newText = newText.trim();
                if (newText.length() == 0) {
                  return "enter folder name";
                }
                if (newText.lastIndexOf("/") != -1) {
                  return "folder name can't contains '/'";
                }
                if (parent.findWorkChild(newText) != null) {
                  return "child with same id is exist";
                }
                if (!parent.getTreeDB().getTreeLogic().canBeCreatedIn(type, parent))
                  return "This folder cannot contain " + newText;
                return null;
              }
            });
    if (dlg.open() == Dialog.OK) return dlg.getValue();
    return null;
  };

  @Override
  public String[] getElementTypes() {
    return new String[] {ReportFolder.TYPE_NAME};
  }

  @Override
  protected void setElementSpecificAttributes(TreeNode nn) {
    // TODO Auto-generated method stub

  }
}
