/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers.report;

import java.util.List;
import java.util.Map;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchWindow;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.operations.IGenerateReportHandler;
import com.unitesk.requality.core.operations.ReportGenResult;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.report.Template;
// TODO report undo
public class GenerateReportHandler extends AbstractHandler
    implements IHandler, IGenerateReportHandler {
  private com.unitesk.requality.core.operations.GenerateReportHandler base =
      new com.unitesk.requality.core.operations.GenerateReportHandler() {
        @Override
        protected void openInformation(String title, String message) {
          Shell s = getShell();
          if (s == null) return;
          MessageDialog.openInformation(s, title, message);
        }

        @Override
        protected void openError(String title, String message) {
          Shell s = getShell();
          if (s == null) return;
          MessageDialog.openError(s, title, message);
        }

        @Override
        protected void openReport(ReportData res) {
          OpenReportHandler.openReport(res);
        }
      };

  public static IContainer getReportFolder(TreeNode node) {
    IFile nodeFile = (IFile) EclipseTracker.getResourceByNode(node);
    IContainer res = nodeFile.getParent().getFolder(Path.fromOSString(node.getId() + "_resources"));
    return res;
  }
  public static com.unitesk.requality.documents.IProgressMonitor wrapMonitor(IProgressMonitor monitor){
    if (monitor != null)
      return
          new com.unitesk.requality.documents.IProgressMonitor() {
            @Override
            public void worked(int val) {
              monitor.worked(val);
            }

            @Override
            public void done() {
              monitor.done();
            }

            @Override
            public void beginTask(String string, int sumCountersMax) {
              monitor.beginTask(string, sumCountersMax);
            }

            @Override
            public boolean isCanceled() {
              return monitor.isCanceled();
            }
          };
    return null;
  }

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    StringBuilder sb = new StringBuilder("Generate report ");
    for(TreeNode nd:nodes)
      if(nd.getType().equals(ReportSettings.TYPE_NAME)){
        if(sb.length()>16)
          sb.append(", ");
        sb.append("(");
        sb.append(nd.getTreeDB().getResName());
        sb.append(") ");
        sb.append(nd.getUserVisibleName());
      }
    WorkspaceJob wj = new WorkspaceJob(sb.toString()){
    
      @Override
      public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
        Map<?, ?> par = event.getParameters();
        try {
          base.execute(nodes, par, wrapMonitor(monitor));
        }catch(Exception e){
          Display.getDefault().asyncExec(new Runnable(){
          
            @Override
            public void run() {
              MessageDialog.openError(org.eclipse.swt.widgets.Display.getDefault().getActiveShell(),
                  "Report generation error","Cannot generate report: "+e.getMessage());
            }
          });
        }
        return Status.OK_STATUS;
      }
    };
    wj.schedule();
    //nodes.get(0).getTreeDB().commit();
    return null;
  }

  private static Shell getShell() {
    if (Activator.getDefault() == null) return null;
    IWorkbench w = Activator.getDefault().getWorkbench();
    if (w == null) return null;
    IWorkbenchWindow win = w.getActiveWorkbenchWindow();
    if (win == null) return null;
    return win.getShell();
  }

  public ReportGenResult generateReport(ReportSettings reportS, com.unitesk.requality.documents.IProgressMonitor mon) {
    ReportGenResult res = base.generateReport(reportS,mon);
    if (res == null) return null;
    if (res.savePath != null) {
      IProject project = EclipseTracker.getTracker().getProject(reportS.getTreeDB());
      //TODO check
      IPath path = Path.fromOSString(res.savePath.toString()).makeRelativeTo(project.getLocation());

      IFolder folder = project.getFolder(path);
      try {
        folder.refreshLocal(IResource.DEPTH_INFINITE, null);
      } catch (CoreException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
    return null;
    //return new  ReportGenResult(href2, report2, savePath2);
  }

  public Template getTemplate(ReportSettings report) {
    return base.getTemplate(report);
  }

  public Template getTemplate(String tid) {
    return base.getTemplate(tid);
  }
  //remove old image from report before new generation
  @Override
  public void clearOldImages(ReportSettings nde) {
    base.clearOldImages(nde);
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    return base.isEnabled(nodes);
  }

  @Override
  public Object execute(List<TreeNode> nodes, Map<?, ?> par, com.unitesk.requality.documents.IProgressMonitor monitor) throws Exception {
    return base.execute(nodes, par, monitor);
  }

  @Override
  public boolean isEnabled(List<TreeNode> nodes) {
    return base.isEnabled(nodes);
  }
}
