/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers.report;

import java.net.URI;
import java.util.List;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Path;
import org.eclipse.swt.SWT;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.IDE;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.handlers.OpenEditorHandler;
import com.unitesk.requality.eclipse.ui.editors.ReportViewer;
import com.unitesk.requality.eclipse.views.RequalityPreferences;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.report.Template;

public class OpenReportHandler extends OpenEditorHandler {
  public Object openOperation(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null) return null;
    for (TreeNode node : nodes)
      if (node instanceof ReportData) {
        ReportData rdata = (ReportData) node;
        openReport(rdata);
      }
    return null;
  }

  public static void openReport(ReportData rdata) {
    org.eclipse.swt.widgets.Display.getDefault().asyncExec(new Runnable(){
    
      @Override
      public void run() {
        Activator a = Activator.getDefault();
        if (a != null
            && a.getPreferenceStore().getBoolean(RequalityPreferences.NOT_OPEN_EDITORS)
            && RequalityPlugin.getBrowserType() == SWT.NONE) return;
        String templateId = rdata.getTemplateId();
        Template tpl = null;
        if (templateId == null) {
          tpl = Activator.getDefault().getReportTemplates().get(Template.DEFAULT_ID);
          rdata.attrs().setConstAttribute(ReportData.ATTR_TEMPLATE, tpl.getId(), Flag.HIDDEN);
          rdata.commitAttributes();
        } else {
          tpl = Activator.getDefault().getReportTemplates().get(templateId);
        }
        URI savePath = rdata.getSaveDir();
        if (tpl != null) {
          //            IResource res = OpenReportHandler.getReportFolder(rdata)
          //                .getFile(
          //                        Path.fromOSString(tpl.getTemplates().get(0)
          //                                .getOutputFileName()));
          //if (!res.exists()) {
          //  GenerateReportHandler.generateReport(rdata);
          //}
          String href =
              savePath
                  + "/"
                  + tpl.getTemplates().get(0).getOutputFileName(rdata.getTreeDB());
          if ((href.length() == 0)) return;
        }
        
        IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();

        if (page != null) {
          try {
            NodeEditorInput input = new NodeEditorInput(rdata);
            try {
              IEditorPart e =
                  PlatformUI.getWorkbench()
                      .getActiveWorkbenchWindow()
                      .getActivePage()
                      .findEditor(input);
              if (e != null && ((NodeEditorInput) e.getEditorInput()).getStorage().getNode() == rdata)
                PlatformUI.getWorkbench()
                    .getActiveWorkbenchWindow()
                    .getActivePage()
                    .closeEditor(e, false);
            } catch (CoreException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
            IDE.openEditor(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(),
                input,
                ReportViewer.ID,
                true);
          } catch (PartInitException e) {
            e.printStackTrace();
          }
        }
      }
    });
    
  }

  public static IContainer getReportFolder(ReportData rdata) {
    IFile nodeFile = (IFile) EclipseTracker.getResourceByNode(rdata);
    IContainer res =
        nodeFile.getParent().getFolder(Path.fromOSString(rdata.getId() + "_resources"));
    return res;
  }

  @Override
  public String getEditorId() {
    return ReportViewer.ID;
  }
}
