package com.unitesk.requality.eclipse.storageformatversion;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import com.unitesk.requality.core.CacheStorage;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.IValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class Requality_old_0212 implements IFormatVersionConverter {
  private static String NODE_GROUP = "com.unitesk.nodegroup.requality";
  private OldTo0212Convert conv = new OldTo0212Convert(){
    @Override
    Map<String, AttributeDefinition> reworkAttrs(UUID uid, Map<String, AttributeDefinition> ad) {
      if(ad.containsKey("_revlinks"))
        ad.remove("_revlinks");
      if (ad.containsKey("_clones")) {
        ad.remove("_clones");
      }
      if (ad.containsKey("_%revlocations")) ad.remove("_%revlocations");
      if (ad.containsKey("_%clones")) ad.remove("_%clones");
      if (ad.containsKey("_revLocations")) {
        ad.remove("_revLocations");
      }
      if (ad.containsKey("_locations")) {
        AttributeDefinition a = ad.get("_locations");
        if(a!=null && a.getRawValue().getValue() instanceof ListValue)
          ad.put(a.getKey(), new AttributeDefinition(a.getOrigin(), a.getKey(), new 
            ReferenceValueProvider(AttributeType.LIST, new ListValue(AttributeType.REFERENCE,
                new ArrayList<>(((ListValue) a.getRawValue().getValue()).getValues()))),a.getFlags()));
      }
      if (ad.containsKey("_cloneParent")) {
        AttributeDefinition a = ad.get("_cloneParent");
        ad.put("_cloneParent", new AttributeDefinition(a.getOrigin(), a.getKey(), 
            new ReferenceValueProvider(AttributeType.REFERENCE, a.getRawValue().getValue()),a.getFlags()));
      }
      Iterator<String>it = ad.keySet().iterator();
      while(it.hasNext()){
        String s = it.next();
        AttributeDefinition a = ad.get(s);
        if(!a.getOrigin().equals(uid)&&!a.getKey().startsWith("_"))
          it.remove();
      }
      return ad;
    }
    @Override
    protected void handleFile(String path, org.eclipse.core.resources.IFile f) throws CoreException {
      if(path.equals("/root/Instances.json")){
        recDelete(f.getParent().findMember("Instances"));
        recDelete(f.getParent().findMember("Instances_resources"));
        f.delete(true, null);
      }else{
        super.handleFile(path, f);
      }
    }
    private void recDelete(IResource findMember) throws CoreException {
      if(findMember == null)
        return;
      if(findMember instanceof IFile)
        findMember.delete(true, null);
      if(findMember instanceof IContainer){
        IContainer cont = (IContainer) findMember;
        IResource [] members = cont.members();
        for (IResource member : members)
         {
          recDelete(member);
         }
        findMember.delete(true, null);
      }
    };
  };
  @Override
  public String convert(IProject proj) {
    String s =  conv.convert(proj);
    if(s==null || s.isEmpty())
    {
      EclipseTracker.getTracker().trackProject(proj,true);
      TreeDB db = ResManager.getManager().getTreeDB(proj.getName());
      String b = null;
      if(db!=null){
        b = db.getCacheStorage().startCacheTransaction("convert old project",CacheStorage.COMMIT_ALL);
        boolean changes = false;
        Set<String>keys = new HashSet<> (db.getRootNode().attrs().getLocalAttributeKeys());
        for(String k:keys){
          if(k.startsWith("@")){
            db.getRootNode().attrs().removeAttribute(k);
            changes = true;
          }
        }
        if(changes)
          db.getRootNode().commitAttributes();
        DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(db.getRootNode(),true,false,true,true,false);
        Iterator<TreeNode> it = dftw.iterator();
        while(it.hasNext())
        {
          TreeNode node = it.next();
          Set<String> nkeys = new HashSet<>(node.attrs().getAllAttributeKeys());
          boolean achanges = false;
          for(String s2:nkeys)
            if(s2.endsWith("_revlinks")||s2.endsWith("%revlinks")){
              node.attrs().removeAttribute(s2);
              achanges = true;
            }else{
              Attribute a = node.attrs().getAttribute(s2);
              if(a.getType()==AttributeType.HTML||a.getType().equals(AttributeType.STRING)){
                TypedValue tv = a.getDefinition().getRawValue();
                if(tv!=null){
                  String val = tv.getGValue();
                  if(val.indexOf("node.resURL")>-1){
                    StringBuilder sb = new StringBuilder();
                    int ind = val.indexOf("{node.resURL}");
                    int lastind = 0;
                    String aval = val;
                    while(ind>-1){
                      int min = ind;
                      if(ind>0 && val.charAt(ind-1)=='$')
                        min = ind-1;
                      sb.append(val.substring(0, min));
                      int q = val.indexOf("'",ind+1),q1 = val.indexOf("\"",ind+1);
                      if(q1!=-1&&(q1<q || q==-1))
                        q = q1;
                      if(q>-1){
                        sb.append("{node.resURL}"+val.substring(ind+13,q));
                        File f = node.getNodeResourcesFile().toPath().resolve(val.substring(ind+14,q)).toFile();
                        File nf = node.getNodeResourcesFile().toPath().resolve(s2).resolve(val.substring(ind+14,q)).toFile();
                        try {
                          RequalityCLI.copyfile(f.toURI(), nf.toURI());
                        } catch (IOException e) {
                          e.printStackTrace();
                        }
                        //val = val.replace("${node.resURL}", "{node.resURL}");
                        lastind = q;
                      }
                      ind = val.indexOf("{node.resURL}",ind+1);
                    }
                    sb.append(val.substring(lastind));
                    val = sb.toString();
                    if(!val.equals(aval)){
                      IValueProvider vp = null;
                      IValueProvider ovp = a.getDefinition().getValueProvider();
                      if(ovp instanceof ConstValueProvider)
                        vp = new ConstValueProvider(ovp.getAttributeType(), val);
                      else
                        if(ovp instanceof HtmlValueProvider)
                          vp = new HtmlValueProvider(ovp.getAttributeType(), val);
                        else
                          if(ovp instanceof TemplValueProvider)
                            vp = new TemplValueProvider(ovp.getAttributeType(), val);
                      node.attrs().setAttribute(new AttributeDefinition(a.getOrigin(),a.getKey(), a.getAvailability(),
                          vp,a.getDefinition().getFlags()));
                      changes = true;
                    }
                  } 
                }
              }
            }
          //TODO noderespath
          Attribute l = node.attrs().getAttribute(Requirement.ATTR_LOCATIONS);
          if(l!=null){
            ListValue lv = l.getDefinition().getRawValue().getGValue();
            List<Object> nl = new ArrayList<>();
            if(lv!=null){
              for(Object r:lv.getValues()){
                TreeNode ref = node.getTreeDB().getNode(r.toString(),Path.QID|Path.UUID|Path.UV);
                if(ref!=null)
                  nl.add(ref.getUUId().toString());
                else{
                  nl.add(r.toString());
                  if(CoreUtils.DEBUG_LEVEL>0)
                    RLogger.logInfo("Location "+r.toString()+" was not found!");
                }
              }
              node.attrs().setAttribute(new AttributeDefinition(l.getOrigin(), l.getKey(), 
                  new ReferenceValueProvider(AttributeType.LIST, new ListValue(AttributeType.REFERENCE, nl))));
              achanges = true;
            }
            node.commitAttributes();
          }
        }
        if(b == null)
            db.getCacheStorage().commitCache();
      }
    }
    return s;
  }

  @Override
  public boolean isAppliable(IProject proj) {
    if(proj.getFile("root/Requirements.json").exists()){
      List<String> vers = StoredFormatVersioning.getFormatsVersions(proj);
      if(vers == null)
        return true;
      for(String s:vers){
        if(s.startsWith(NODE_GROUP)){
          return s.equals(NODE_GROUP + "|0.21.1")|| s.equals(NODE_GROUP + "|0.21")|| s.equals(NODE_GROUP + "|0.15");
        }
      }
    }
    return false;
  }
}
