package com.unitesk.requality.eclipse.storageformatversion;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;

public class StoredFormatVersioning {
  public static String getFullVersionByGroupId(String gid) {
    if (!versionsMap.containsKey(gid)) return null;
    return gid + "|" + versionsMap.get(gid);
  }

  public static String getVersionByGroupId(String gid) {
    if (!versionsMap.containsKey(gid)) return "undefined";
    return versionsMap.get(gid);
  }

  public static List<String> getFormatsVersions(IProject proj) {
    if (proj == null) return null;
    IFile f = proj.getFile("/.reqversions");
    if (!f.exists()) return null;
    List<String> vers = new ArrayList<>();
    try {
      //proj.touch(null);
      f.refreshLocal(0, null);
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      //e.printStackTrace();
    }
    try {
      BufferedReader br = new BufferedReader(new java.io.InputStreamReader(f.getContents()));
      Map<String, String> avers = new HashMap<>();
      while (br.ready()) {
        String s = br.readLine();
        if (s != null) {
          int ind = s.indexOf("|");
          if (ind > -1) {
            String name = s.substring(0, ind), v = s.substring(ind + 1);
            boolean replace = !avers.containsKey(name);
            if (!replace) {
              if (betterVers(v, avers.get(name))) {
                replace = true;
                vers.remove(name + "|" + avers.get(name));
              }
            }
            if (replace) {
              avers.put(name, v);
              vers.add(s);
            }
          }
        }
      }
      br.close();
    } catch (IOException e) {
      e.printStackTrace();
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return vers;
  }

  private static boolean betterVers(String v, String string) {
    v = v.trim();
    string = string.trim();
    int s1 = 0, s2 = 0;
    ;
    int i = 0, j = 0;
    while (i < v.length() || j < string.length()) {
      int n1 = 0, n2 = 0;
      while (i < v.length() && Character.isDigit(v.charAt(i))) i++;
      while (j < string.length() && Character.isDigit(string.charAt(j))) j++;
      String i1 = v.substring(s1, i), i2 = string.substring(s2, j);
      if (i <= v.length() && !i1.isEmpty()) n1 = Integer.parseInt(i1);
      if (j <= string.length() && !i2.isEmpty()) n2 = Integer.parseInt(i2);
      if (n1 > n2) return true;
      while (i < v.length() && !Character.isDigit(v.charAt(i))) i++;
      while (j < string.length() && !Character.isDigit(string.charAt(j))) j++;
      s1 = i;
      s2 = j;
    }
    return false;
  }

  private static final String separator = System.getProperty("line.separator");

  public static void setFormatsVersions(IProject proj, List<String> vers) {
    IFile f = proj.getFile("/.reqversions");
    String cont = "";
    for (String s : vers) cont += s + separator;
    try {
      if (!f.getParent().exists()) ((IFolder) f.getParent()).create(true, false, null);
      if (f.exists()) f.delete(true, null);
      f.create(new ByteArrayInputStream(cont.getBytes()), true, null);
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private static List<String> versions = null;
  private static SortedMap<String, String> versionsMap = new TreeMap<>();
  //vers
  //	-old\new
  //		-conv
  public static List<VersionUpdateInfo> collectUpdates(IProject proj) {
    List<VersionUpdateInfo> vui = new ArrayList<>();
    List<String> vers = getFormatsVersions(proj);
    //special case of old projects
    List<String> avers = getActiveVersions();
    if (vers == null) {
      for (String s : avers) {
        IFormatVersionConverter conv = getConverter(getGroupIdPart(s), "", getVersionPart(s));
        if (conv != null) {
          if (conv.isAppliable(proj)) {
            vui.add(
                new VersionUpdateInfo(
                    getGroupIdPart(s), "pre_0.15", getVersionPart(s), "update", conv));
          }
        }
      }
    } else {
      List<String> miss = getMissed(vers);
      for (String m : miss) {
        boolean fnd = false;
        for (String s : avers)
          if (hasSameGroup(m, s)) {
            if (!hasSameVersion(m, s)) {
              IFormatVersionConverter conv =
                  getConverter(getGroupIdPart(s), getVersionPart(m), getVersionPart(s));
              if (conv != null && conv.isAppliable(proj)) {
                fnd = true;
                vui.add(
                    new VersionUpdateInfo(
                        getGroupIdPart(s), getVersionPart(m), getVersionPart(s), "update", conv));
              }
            } else fnd = true;
          }
        if (!fnd) {
          String gid = getGroupIdPart(m);
          String v = getVersionByGroupId(gid);
          vui.add(
              new VersionUpdateInfo(
                  gid,
                  v,
                  getVersionPart(m),
                  v.equals("undefined") ? "plugin_not_found" : "version_not_found",
                  null));
        }
      }
    }
    return vui;
  }

  public static List<VersionUpdateInfo> updateProject(IProject proj) {
    List<String> vers = getFormatsVersions(proj);
    //special case of old projects
    List<String> avers = getActiveVersions();
    List<VersionUpdateInfo> vis = new ArrayList<>();
    if (vers == null) {
      for (String s : avers) {
        IFormatVersionConverter conv = getConverter(getGroupIdPart(s), "", getVersionPart(s));
        if (conv != null) {
          if (conv.isAppliable(proj)) {
            String err = conv.convert(proj);
            if (err != null) {
              vis.add(
                  new VersionUpdateInfo(
                      getGroupIdPart(s), "pre_0.15", getVersionPart(s), "Error:\n" + err, null));
            }
          }
        }
      }
    } else {
      List<String> miss = getMissed(vers);
      int convi = 0;
      for (String m : miss) {
        boolean fnd = false;
        VersionUpdateInfo vui = null;
        for (String s : avers)
          if (hasSameGroup(m, s)) {
            if (!hasSameVersion(m, s)) {
              IFormatVersionConverter conv =
                  getConverter(getGroupIdPart(s), getVersionPart(m), getVersionPart(s));
              if (conv != null && conv.isAppliable(proj)) {
                String err = conv.convert(proj);
                if (err != null) {
                  vui =
                      new VersionUpdateInfo(
                          getGroupIdPart(s),
                          getGroupIdPart(s),
                          getVersionPart(m),
                          "Error:\n" + err,
                          null);
                  vis.add(vui);
                } else {
                  fnd = true;
                  vis.add(
                      new VersionUpdateInfo(
                          getGroupIdPart(s),
                          getGroupIdPart(s),
                          getVersionPart(m),
                          "convertion succeed",
                          conv));
                  convi++;
                }
              }
            } else {
              fnd = true;
              convi++;
            }
          }
        if (!fnd) {
          if (vui == null) {
            String gid = getGroupIdPart(m);
            String v = getVersionByGroupId(gid);
            vis.add(
                new VersionUpdateInfo(
                    gid,
                    v,
                    getVersionPart(m),
                    v.equals("undefined") ? "plugin_not_found" : "version_not_found",
                    null));
          }
        }
      }
    }
    return vis;
  }

  private static SortedMap<String, SortedMap<String, SortedMap<String, IFormatVersionConverter>>>
      rgraph = null;

  public static IFormatVersionConverter getConverter(String group, String oldv, String reqvers) {
    if (rgraph == null) {
      //load existed converters
      rgraph = new TreeMap<>();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor("com.unitesk.requality.versionconverter");
      for (IConfigurationElement conf : configurationElements) {
        try {
          IFormatVersionConverter ah =
              (IFormatVersionConverter) conf.createExecutableExtension("converter");
          IFormatVersionConverter conv = null;
          conv = ah;
          String gid = conf.getAttribute("groupId"),
              vers_old = conf.getAttribute("versionold"),
              vers_new = conf.getAttribute("versionnew");
          if (!rgraph.containsKey(gid))
            rgraph.put(gid, new TreeMap<String, SortedMap<String, IFormatVersionConverter>>());
          if (!rgraph.get(gid).containsKey(vers_old))
            rgraph.get(gid).put(vers_old, new TreeMap<String, IFormatVersionConverter>());
          rgraph.get(gid).get(vers_old).put(vers_new, conv);
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
      for (String gr : rgraph.keySet())
        for (String versold : rgraph.get(gr).keySet()) reachConvAdd(gr, versold);
    }
    if (oldv == null) oldv = "";
    if (!rgraph.containsKey(group)) return null;
    if (!rgraph.get(group).containsKey(oldv) || !rgraph.get(group).get(oldv).containsKey(reqvers))
      return null;
    return rgraph.get(group).get(oldv).get(reqvers);
  }

  private static Set<String> emptySet = new TreeSet<>();

  private static Set<String> reachConvAdd(String gr, String versold) {
    if (!rgraph.get(gr).containsKey(versold)) return emptySet;
    Set<String> convs = rgraph.get(gr).get(versold).keySet();
    for (String versnew : convs) {
      Set<String> rec = reachConvAdd(gr, versnew);
      for (String s : rec)
        if (!rgraph.get(gr).get(versold).containsKey(s))
          rgraph.get(gr).get(versold).put(s, rgraph.get(gr).get(versold).get(versnew));
    }
    return rgraph.get(gr).get(versold).keySet();
  }

  public static String getGroupIdPart(String vers) {
    return vers.substring(0, vers.lastIndexOf("|"));
  }

  public static String getVersionPart(String vers) {
    return vers.substring(vers.lastIndexOf("|") + 1);
  }

  private static boolean hasSameGroup(String v1, String v2) {
    return v1.substring(0, v1.lastIndexOf("|")).equals(v2.substring(0, v2.lastIndexOf("|")));
  }

  private static boolean hasSameVersion(String v1, String v2) {
    return v1.substring(v1.lastIndexOf("|")).equals(v2.substring(v2.lastIndexOf("|")));
  }

  public static List<String> getMissed(List<String> stored) {
    List<String> nowv = getActiveVersions();
    List<String> ret = new ArrayList<>();
    if (stored == null || stored.size() == 0) return ret;
    for (String s : stored) {
      boolean fnd = false;
      for (String s2 : nowv)
        if (hasSameGroup(s, s2)) {
          fnd = true;
          if (!hasSameVersion(s, s2)) ret.add(s);
        }
      if (!fnd) ret.add(s);
    }
    return ret;
  }

  public static List<String> getActiveVersions() {
    if (versions == null) {
      versions = new ArrayList<>();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor("com.unitesk.requality.nodegroup");
      for (IConfigurationElement conf : configurationElements) {
        try {
          try {
            String gid = conf.getAttribute("groupId"), vers = conf.getAttribute("version");
            versions.add(gid + "|" + vers);
            versionsMap.put(gid, vers);
          } catch (ClassCastException cce) {
            //so, this is bad handlers(
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }
    return versions;
  }
}
