package com.unitesk.requality.eclipse.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.viewers.DecoratingLabelProvider;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.navigator.CommonNavigator;
import org.eclipse.ui.navigator.CommonViewer;
import org.eclipse.ui.part.IShowInSource;
import org.eclipse.ui.part.ShowInContext;
import org.eclipse.ui.progress.UIJob;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.ITreeNodeProvider;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.selection.IUIStateChangeListener;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.handlers.LightLinks;
import com.unitesk.requality.eclipse.ui.cnf.RequalitySorter;
import com.unitesk.requality.eclipse.views.PrefixAndTreeNode;
import com.unitesk.requality.eclipse.views.RequalityContentLinksProvider;
import com.unitesk.requality.eclipse.views.RequalityContentLinksProvider.TestNodeLinks;
import com.unitesk.requality.eclipse.views.RequalityContentLinksProvider.TestNodeLinksGroup;
import com.unitesk.requality.marker.LinksDecorator;

public class RequalityLCNF extends CommonNavigator implements ITabbedPropertySheetPageContributor,
  ISelectionProvider{
  public static final String ID = "com.unitesk.requality.views.RequalityLinksExplorer";
  private static byte refdir = 3;
  private boolean updating = false;
  private CommonViewer cv = null;
  private LinksDecorator decorator;
  private static List<RequalityLCNF> activeNaivgators = new ArrayList<>();

  public RequalityLCNF() {
    super();
    activeNaivgators.add(this);
  }

  public static List<RequalityLCNF> getActiveNaivgators() {
    return activeNaivgators;
  }

  @Override
  protected CommonViewer createCommonViewerObject(Composite aParent) {
    cv =
        new CommonViewer(getViewSite().getId(), aParent, SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
    cv.setContentProvider(new RequalityContentLinksProvider());
    final ILabelProvider base1 = WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider();
    final ILabelProvider base = new ILabelProvider(){
    
      @Override
      public void removeListener(ILabelProviderListener listener) {
        base1.removeListener(listener);
      }
    
      @Override
      public boolean isLabelProperty(Object element, String property) {
        return base1.isLabelProperty(element, property);
      }
    
      @Override
      public void dispose() {
        base1.dispose();
      }
    
      @Override
      public void addListener(ILabelProviderListener listener) {
        base1.addListener(listener);
      }
    
      @Override
      public String getText(Object element) {
        return base1.getText(element);
      }
    
      @Override
      public Image getImage(Object element) {
        if (element instanceof PrefixAndTreeNode)
          return getImage(((PrefixAndTreeNode)element).getNode());
        return base1.getImage(element);
      }
    };
    cv.addDoubleClickListener(
        new IDoubleClickListener() {

          @Override
          public void doubleClick(DoubleClickEvent event) {
            if (updating) return;
            updating = true;
            StructuredSelection s = (StructuredSelection) event.getSelection();
            if (s != null && !s.isEmpty()) {
              Object o = s.getFirstElement();
              TreeNode node = null;
              if(o instanceof PrefixAndTreeNode)
                node = ((PrefixAndTreeNode)o).getNode();
              if (o instanceof TreeNode) {
                node = (TreeNode) o;
              } else if (o instanceof TestNodeLinks) node = ((TestNodeLinks) o).getTarget();
              if (node != null) {
                HashMap<String, List<TreeNode>> nodes = new HashMap<>();
                nodes.put(node.getType(), new ArrayList<TreeNode>());
                nodes.get(node.getType()).add(node);
                SelectionManager.getOne().setActiveNodes(nodes,nodes, EventSource.LCNF);
              }else
                SelectionManager.getOne().setActiveNodes(Collections.EMPTY_MAP,Collections.EMPTY_MAP, EventSource.LCNF);
            }
            updating = false;
          }
        });
    decorator = new LinksDecorator(node);
    cv.setLabelProvider(
        new DecoratingLabelProvider(base, decorator) {

          @Override
          public void removeListener(ILabelProviderListener listener) {
            base.removeListener(listener);
          }

          @Override
          public boolean isLabelProperty(Object element, String property) {
            return true;
          }

          @Override
          public void dispose() {
            base.dispose();
          }

          @Override
          public void addListener(ILabelProviderListener listener) {
            base.addListener(listener);
          }

          @Override
          public String getText(Object element) {
            TreeNode node = null;
            if(element instanceof TreeNode)
              node = (TreeNode) element;
            else
              if(element instanceof ITreeNodeProvider)
                node = ((ITreeNodeProvider)element).getNode();
            String ext = "";
            if(node!=null&&!node.isSameTree(RequalityLCNF.this.getTreeSelectedNode()))
              ext = node.getTreeDB().getResName()+"::";
            String baseText = base.getText(element);
            if (!baseText.equals("")) return ext+baseText;
            String els = element.toString();
            if(els.contains("[_prefix]"))
              els = els.replace("[_prefix]", ext);
            else
              if(!els.startsWith(ext))
                return ext+els;
            return els;
          }

          @Override
          public Image getImage(Object element) {
            if (element instanceof PrefixAndTreeNode)
              return base.getImage(((PrefixAndTreeNode)element).getNode());
            if (element instanceof TestNodeLinks) return null;
            if (element instanceof TestNodeLinksGroup) {
              String imgup = "arrowup", imgdown = "arrowdown";
              TestNodeLinksGroup link = (TestNodeLinksGroup) element;
              return Activator.getDefault()
                  .getImageRegistry()
                  .getDescriptor(link.isUp() ? imgup : imgdown)
                  .createImage();
            }
            return base.getImage(element);
          }
        });

    SelectionManager.getOne()
        .addChangeListener(
            EventSource.LCNF,
            list);
    // cv.setLabelProvider(base);
    return cv;
  }
  IUIStateChangeListener list = new IUIStateChangeListener() {

    @Override
    public void newActiveNodes(final List<TreeNode> nodes) {
      new UIJob("update requality link cnf") {
        @Override
        public IStatus runInUIThread(IProgressMonitor monitor) {
          if (updating)
            return Status.CANCEL_STATUS;
          if (cv.getControl().isDisposed()) {
            return Status.CANCEL_STATUS;
          }
          updating = true;
          if (nodes.size() > 0 && cv.getInput()!=nodes.get(0)){
            decorator.setNode(nodes.get(0));
            cv.setInput(nodes.get(0));
            node = nodes.get(0);
            cv.setSelection(new StructuredSelection(nodes.get(0)));
            //SelectionManager.getOne().setActiveNodes(Collections.EMPTY_MAP,Collections.EMPTY_MAP, EventSource.LCNF);
          }
          else{
            decorator.setNode(null);
            cv.setInput(null);
            node = null;
            cv.setSelection(StructuredSelection.EMPTY);
          }
          updating = false;
          return Status.OK_STATUS;
        }
      }.schedule();
    }
  };
  @Override
  public String getContributorId() {
    return "com.unitesk.requality.views.RequalityPropertyContributor";
  }

  @Override
  public void createPartControl(Composite aParent) {
    super.createPartControl(aParent);
    getCommonViewer().setComparator(new RequalitySorter());
    getCommonViewer().addSelectionChangedListener(treeSelection);
  }

  @Override
  public void dispose() {
    super.dispose();
    SelectionManager.getOne()
        .removeChangeListener(
            EventSource.GENERAL,
            list);
    activeNaivgators.remove(this);
  };

  @SuppressWarnings("unchecked")
  @Override
  public Object getAdapter(@SuppressWarnings("rawtypes") Class adapter) {
    if (adapter == IPropertySheetPage.class) {
      // TODO null pointer cause??
      return new TabbedPropertySheetPage(this);
    }else if (adapter == IShowInSource.class){
      return new IShowInSource(){
      
        @Override
        public ShowInContext getShowInContext() {
          return new ShowInContext(getInitialInput(), getSelection());
        }
      };
    }
    return super.getAdapter(adapter);
  }
  @Override
  protected Object getInitialInput() {
    if (Activator.getDefault() != null) {
      Map<String, List<TreeNode>> sel =
          SelectionManager.getOne().getActiveNodes(EventSource.LCNF);
      if (sel != null && !sel.isEmpty()) {
        Iterator<List<TreeNode>> it = sel.values().iterator();
        if (it.hasNext()) {
          List<TreeNode> l = it.next();
          if (l != null && !l.isEmpty())
            return l.iterator().next();
        }
      }
    }
    return super.getInitialInput();
  }

  public static void toggleReferenceDirection() {
    refdir++;
    if (refdir == 4) refdir = 1;
    for (RequalityLCNF cnf : activeNaivgators) {
      cnf.getCommonViewer().refresh(true);
    }
  }

  public static int getReferenceDirection() {
    return refdir;
  }

  private static Map<TreeDB, Set<String>> filter = new HashMap<>();

  public static Set<String> getFilters(TreeDB treeDB) {
    return filter.get(treeDB);
  }

  public static void setFilter(TreeDB treeDB, HashSet<String> hashSet) {
    filter.put(treeDB, hashSet);
    for (RequalityLCNF c : activeNaivgators) {
      c.getCommonViewer().refresh(true);
      c.getCommonViewer().refresh(c.getCommonViewer().getInput());
    }
  }

  public static TreeNode getTreeSelectedNode(){
	  for (RequalityLCNF cnf : activeNaivgators) {
	    Object o = cnf.getCommonViewer().getInput();
	    if(o != null){
          if(PrefixAndTreeNode.class.isInstance(o))
            o = ((PrefixAndTreeNode)o).getNode();
          if(o instanceof TreeNode){
            TreeNode node = (TreeNode) o;
            return node;
          }
	    }
	    if(!(cnf.getInitialInput() instanceof TreeNode))
	      return null;
	    return (TreeNode) cnf.getInitialInput();
  }
  return null;
}
  
  public static TreeNode getSelectedNode() {
    IWorkbenchWindow activeWorkbenchWindow = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
    if (activeWorkbenchWindow == null) {
      return null;
    }

    ISelection selection = null;
    TreeNode node1 = null;
    for (RequalityLCNF cnf : activeNaivgators){
      selection = cnf.getCommonViewer().getSelection();
      Object inp = cnf.getInitialInput();
      if(inp instanceof TreeNode)
        node1 = (TreeNode) inp;
    }

    if (!(selection instanceof IStructuredSelection)) {
      return node1;
    }

    IStructuredSelection selected = (IStructuredSelection) selection;
    if (selected.isEmpty()) {
      return node1;
    }

    List<TreeNode> nodes = new ArrayList<>();
    @SuppressWarnings("unchecked")
    List<Object> objList = selected.toList();
    for (Object obj : objList) {
      if (obj instanceof PrefixAndTreeNode)
        obj = ((PrefixAndTreeNode)obj).getNode();
      if (obj instanceof TreeNode) {
        TreeNode node = (TreeNode) obj;
        nodes.add(node);
      }
      if (nodes.size() > 0) {
        return nodes.get(0);
      }
    }
    return node1;
  }

  TreeNode node = null, oldNode = null;
  private final ISelectionChangedListener treeSelection =
      new ISelectionChangedListener() {
        @Override
        public void selectionChanged(SelectionChangedEvent event) {
          if (event.getSelection() instanceof IStructuredSelection) {
            node = getSelectedNode();
            //if selected link is selfy-node link is tantamount to node isn't choosen
            if (getTreeSelectedNode() == node){
            	node = null;
            	//compulsion dispose decoration of oldNode
            	LinksDecorator.isDispose(true);
            }         
            LightLinks.ColorNode(node, oldNode);
            oldNode = node;
          }
        }
      };

  @Override
  public void addSelectionChangedListener(ISelectionChangedListener listener) {
    cv.addSelectionChangedListener(listener);
  }

  @Override
  public ISelection getSelection() {
    TreeNode tn = getSelectedNode();
    if(tn!=null)
      return new StructuredSelection(tn);
    return StructuredSelection.EMPTY;
  }

  @Override
  public void removeSelectionChangedListener(ISelectionChangedListener listener) {
    cv.removeSelectionChangedListener(listener);
  }

  @Override
  public void setSelection(ISelection selection) {
    cv.setSelection(selection);
  }
}
