/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.ui.cnf;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.UUID;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.ui.IWorkbenchPreferenceConstants;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.navigator.CommonNavigator;
import org.eclipse.ui.navigator.CommonViewer;
import org.eclipse.ui.progress.UIJob;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.ISelectionProvider;
import com.unitesk.requality.core.ITreeNodeFilter;
import com.unitesk.requality.core.SharedConsts;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.selection.IUIStateChangeListener;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.ui.ObjectRule;
import com.unitesk.requality.eclipse.ui.cnf.filters.AbstractNodeFilter;
import com.unitesk.requality.eclipse.views.RequalityContentProvider;
import com.unitesk.requality.eclipse.views.RequalityPreferences;
import com.unitesk.requality.marker.RequalityDecorator;
import com.unitesk.requality.nodetypes.TextNode;

public class RequalityCNF extends CommonNavigator
    implements ITabbedPropertySheetPageContributor, ISelectionProvider, ISchedulingRule {
  public static Long TIME_SKIP = 50L;
  public static boolean shiftPressed = false, altPressed = false;
  public static final String ID = "com.unitesk.requality.views.RequalityExplorer";
  public static RequalityCNF activeNavigator;
  private boolean recAdd(TreeNode node, List<TreeNode> list) {
    if (list.contains(node)) return false;
    TreeNode nde = node;
    while (nde != null) {
      nde = nde.getShownParent();
      if (nde != null && list.contains(nde)) return false;
    }
    list.add(node);
    return true;
  }
  private static Set<String> shownNames = new HashSet<>();
  static{
    shownNames.add(TreeNode.ATTR_LABEL);
    shownNames.add(TreeDB.INDEX_ATTRIBUTE);
    shownNames.add(TreeDB.SORT_METHOD);
    shownNames.add(TreeNode.ATTR_NAME);
    shownNames.add(TextNode.ATTR_TNTYPE);
  }
  
  public static boolean isShown(String s) {
    if(shownNames.contains(s))
      return true;
    for(RequalityDecorator rd:RequalityDecorator.getAll())
      {
        Set<String> atrs = rd.getRelatedKeys();
        if(atrs!=null && atrs.contains(s))
          return true;
      }
    return false;
  }
  
  private void setSelection(StructuredSelection sel, boolean intern) {
    //if (intern) localToSelect.add(sel);
    RequalityCNF.this.selectReveal(sel);
    //if (intern) localToSelect.remove(sel);
  }

  private TreeDB db = null;
  private final IUIStateChangeListener uiListener =
      new IUIStateChangeListener() {
        @Override
        public void newActiveNodes(final List<TreeNode> nodes) {
          if(selectionChanged(nodes))
          {
            UIJob uj = new UIJob("update requality explorer selection"){
              @Override
              public IStatus runInUIThread(IProgressMonitor monitor) {
                if (nodes == null) {
                  setSelection(StructuredSelection.EMPTY, true);
                  return Status.CANCEL_STATUS;
                } else {
                  Object[] objs =
                      ((IStructuredSelection) getCommonViewer().getSelection()).toArray();
                  //                      if(objs!=null && objs.length == nodes.size()){
                  //                          boolean stop = false;
                  //                          for(int i = 0; i<objs.length && !stop; i++)
                  //                              stop = nodes.contains(objs[i]);
                  //                          if(stop)
                  //                              ;
                  //                      }
                  if (nodes.size() == 0) return Status.CANCEL_STATUS;
                  //if (db != null) db.removeGUIListener(guiTreeChangeListener);
                  db = Instantiator.getRootTree(nodes.get(0).getTreeDB());
                  //db.addGUIListener(guiTreeChangeListener);
                  List<TreeNode> sel = new LinkedList<>();
                  List<TreeNode> updated = new ArrayList<>();
                  for (TreeNode node : nodes) {
                    if (Instantiator.getRootTree(node.getTreeDB())
                        != Instantiator.getRootTree(db)) continue;
                    TreeNode parentElement = node.getShownParent();
                    if (parentElement == null || !recAdd(parentElement, updated)) continue;
                    //refresh(node.getParent(true), true);
                    getCommonViewer().expandToLevel(parentElement, 1);
                    sel.add(node);
                  }
                  StructuredSelection elements = new StructuredSelection(sel);
                  getCommonViewer().removeSelectionChangedListener(treeSelection);
                  setSelection(elements, true);
                  getCommonViewer().addSelectionChangedListener(treeSelection);
                }
                return Status.OK_STATUS;
              }
            };
            uj.setRule(new ObjectRule(RequalityCNF.this.getCommonViewer()));
            uj.setPriority(Job.INTERACTIVE);
            uj.schedule();
          }
        }

        private boolean selectionChanged(List<TreeNode> nodes) {
          List<TreeNode> nodes2=getSelectedNodes();
          if(nodes==null&&nodes2!=null || nodes2==null&&nodes!=null)
            return true;
          if(nodes==null||nodes2==null)
            return false;
          return !nodes2.equals(nodes);
        }
      };

  private final ISelectionChangedListener treeSelection =
      new ISelectionChangedListener() {
        @Override
        public void selectionChanged(SelectionChangedEvent event) {
          if (event.getSelection() instanceof IStructuredSelection) {
            @SuppressWarnings("unchecked")
            Iterator<Object> it = ((IStructuredSelection) event.getSelection()).iterator();
            TreeDB db2 = null;
            SortedMap<String, List<TreeNode>> nodes = new TreeMap<>();
            SortedMap<String, List<TreeNode>> nodes2 = new TreeMap<>();
            while (it.hasNext()) {
              Object obj = it.next();
              //not sure that it is possible
              if (obj == null) continue;
              if (obj instanceof TreeNode) {
                TreeNode node = (TreeNode) obj;
                if (db2 == null) db2 = Instantiator.getRootTree(node.getTreeDB());
                //possible missunderstood here
                if (db2 != Instantiator.getRootTree(node.getTreeDB())) continue;
                //if (db != null) db.removeGUIListener(guiTreeChangeListener);
                db = Instantiator.getRootTree(node.getTreeDB());
                //db.addGUIListener(guiTreeChangeListener);
                if (!nodes.containsKey(node.getType()))
                  nodes.put(node.getType(), new LinkedList<TreeNode>());
                nodes.get(node.getType()).add(node);
                TreeNode exp = node.getExposedNode();
                if(!nodes2.containsKey(exp.getType()))
                  nodes2.put(exp.getType(), new LinkedList<>());
                nodes2.get(exp.getType()).add(exp);
              }
            }
            //if(!localToRefresh.contains(event.getSelection())){
            Map<String, List<TreeNode>> oldNodes =
                SelectionManager.getOne().getActiveNodes(EventSource.EXPLORER);
            Set<String> keys = new HashSet<>(nodes.keySet());
            boolean equal = true;
            for (String s : oldNodes.keySet())
              if (keys.contains(s)) {
                for (TreeNode node : nodes.get(s))
                  if (!oldNodes.get(s).contains(node)) equal = false;
                for (TreeNode node : oldNodes.get(s))
                  if (!nodes.get(s).contains(node)) equal = false;
                keys.remove(s);
              } else equal = false;
            if (keys.size() > 0) equal = false;
            if (!equal)
              SelectionManager.getOne().setActiveNodes(nodes, nodes2, EventSource.EXPLORER);
            //    }
          }
        }
      };
  private static boolean shown = false;

  public RequalityCNF() {
    super();
//    if(activeNavigator!=null)
//      throw new RuntimeException("Only one reqcnf will exist");
    activeNavigator = this;
    //check engines
    if (!shown
        && Activator.getDefault()
            .getPreferenceStore()
            .getBoolean(RequalityPreferences.OPEN_NO_ENGINES_DIALOGS)) {
      shown = true;
      //TODO fix window
      if (RequalityPlugin.getBrowserType() == SWT.NONE) {
        MessageDialog.openInformation(
            Activator.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Browser engines are not available",
            "There are no available browser engines. Please install XulRunner or WebKit.\nFollow instructions from http://requality.org/en/general.en.html#browser_integration.");
      }
    }
    SelectionManager.getOne().registerProvider(this);
  }

  @Override
  public void dispose() {
    super.dispose();
    activeNavigator = null;
    getCommonViewer().removeSelectionChangedListener(treeSelection);
    SelectionManager.getOne().unregisterProvider(this);
    SelectionManager.getOne().removeChangeListener(EventSource.EXPLORER, uiListener);
    //if (db != null) db.removeGUIListener(guiTreeChangeListener);
  }

  KeyListener kl =
      new KeyListener() {

        @Override
        public void keyReleased(KeyEvent e) {
          shiftPressed = shiftPressed ? (e.keyCode & SWT.SHIFT) == 0 : false;
          altPressed = altPressed ? (e.keyCode & SWT.CTRL) == 0 : false;
        }

        @Override
        public void keyPressed(KeyEvent e) {
          shiftPressed = (e.keyCode & SWT.SHIFT) != 0;
          altPressed = (e.keyCode & SWT.CTRL) != 0;
        }
      };

  @Override
  public void createPartControl(Composite aParent) {
    super.createPartControl(aParent);
    SelectionManager.getOne().addChangeListener(EventSource.EXPLORER, uiListener);
    getCommonViewer().addSelectionChangedListener(treeSelection);
    getCommonViewer().setComparator(new RequalitySorter());
    updateStaticFilter(getCommonViewer());
    Display display = Display.getDefault();
    display.addFilter(
        SWT.KeyUp,
        new Listener() {
          @Override
          public void handleEvent(org.eclipse.swt.widgets.Event e) {
            shiftPressed = shiftPressed ? (e.keyCode & SWT.SHIFT) == 0 : false;
            altPressed = altPressed ? (e.keyCode & SWT.CTRL) == 0 : false;
            //RLogger.logInfo("GKR "+(shiftPressed?"S":"NS")+" "+(altPressed?"A":"NA"));

          }
        });
    getCommonViewer().getControl().addKeyListener(kl);
    SharedConsts.isNavigatorLinkingEnabled = false;
    if (!PlatformUI.getPreferenceStore()
        .getBoolean(IWorkbenchPreferenceConstants.LINK_NAVIGATOR_TO_EDITOR)){
      SharedConsts.isNavigatorLinkingEnabled = true;
      this.setLinkingEnabled(true);
    }
  }

  private void updateStaticFilter(CommonViewer cv) {
    final SortedSet<String> tps = new TreeSet<>();
    for (ViewerFilter o : cv.getFilters())
      if (o instanceof AbstractNodeFilter) {
        tps.addAll(Arrays.asList(((AbstractNodeFilter) o).getFilterTypes()));
      }
    RequalityContentProvider.setStaticFilter(
        new ITreeNodeFilter() {

          @Override
          public boolean select(TreeNode node) {
            if (node == null) return false;
            String type = node.getType();
            if (type == null) return false;
            return !(tps.contains(type));
          }
        });
  }

  // This is danger hack
  @Override
  protected CommonViewer createCommonViewerObject(Composite aParent) {
    CommonViewer cv =
        new CommonViewer(getViewSite().getId(), aParent, SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI) {
          @Override
          public void refresh() {
            if (isBusy()) //silently do nothing :)
            return;
            updateStaticFilter(this);
            super.refresh();
          }

          @Override
          public void add(Object parentElement, Object[] childElements) {
            try {
              if (isBusy()) //silently do nothing :)
              return;
              super.add(parentElement, childElements);
            } catch (RuntimeException re) {
              re.printStackTrace();
            }
          }

          @Override
          public void remove(Object[] elements) {
            try {
              if (isBusy()) //silently do nothing :)
              return;
              super.remove(elements);
            } catch (RuntimeException re) {
            }
          }
        };
    return cv;
  }

  @Override
  public String getContributorId() {
    return "com.unitesk.requality.views.RequalityPropertyContributor";
  }

  @SuppressWarnings("unchecked")
  @Override
  public Object getAdapter(@SuppressWarnings("rawtypes") Class adapter) {
    if (adapter == IPropertySheetPage.class) {
      //TODO null pointer cause??
      return new TabbedPropertySheetPage(this);
    }
    return super.getAdapter(adapter);
  }

  protected static synchronized List<TreeNode> getSelectedNodes() {
    if (activeNavigator == null) return null;
    ISelection selection =
        activeNavigator
            .getCommonViewer()
            .getSelection(); //TODO test//activeWorkbenchWindow.getSelectionService()
    //.getSelection();

    if (!(selection instanceof IStructuredSelection)) {
      return null;
    }
    IStructuredSelection selected = (IStructuredSelection) selection;

    if (selected.isEmpty()) {
      return null;
    }
    List<TreeNode> nodes = new ArrayList<>();
    @SuppressWarnings("unchecked")
    List<Object> objList = selected.toList();
    for (Object obj : objList) {
      if (obj instanceof TreeNode) {
        TreeNode node = (TreeNode) obj;
        nodes.add(node);
      } else if (obj instanceof IFile) {
        TreeNode node = EclipseTracker.getNodeByResource((IFile) obj);
        if (node != null) nodes.add(node);
      }
    }
    if (nodes.size() > 0) return nodes;
    return null;
  }

  @SuppressWarnings("unchecked")
  public static synchronized List<Object> getSelectedObj() {
    ISelection selection = activeNavigator.getCommonViewer().getSelection();

    if (!(selection instanceof IStructuredSelection)) {
      return null;
    }
    IStructuredSelection selected = (IStructuredSelection) selection;

    if (selected.isEmpty()) {
      return null;
    }
    return selected.toList();
  }

  private Map<UUID, Long> lastUpdate = new HashMap<>();

  public void refresh(TreeNode tn) {
    if (tn == null) return;
    if (this.getCommonViewer() != null) {
      if (!this.getCommonViewer().isBusy())
        try {
          //TreeDB db = tn.getTreeDB();
          //db.getStorage().getParent(tn.getUUId());
          //lastUpdate.getOrDefault(tn.getUUId(),0L);
          //	while(par!=null){
          //		        		Long n = lastUpdate.getOrDefault(par,0L);
          //		        		if(n>max)
          //		        			max = n;
          //		        		par = db.getStorage().getParent(par);
          //		        	}
          Long now = System.currentTimeMillis();
          //if(now-max<TIME_SKIP){
          lastUpdate.put(tn.getUUId(), now);
          this.getCommonViewer().refresh(tn);
          //	}
        } catch (Exception e) {
          //TODO fix unnecessary refresh
        }
    }
  }

  @Override
  protected Object getInitialInput() {
    if (super.getInitialInput() == null) {
      return ResourcesPlugin.getWorkspace().getRoot();
    }
    return super.getInitialInput();
  }

  @Override
  public Map<String, List<TreeNode>> getSelection() {
    List<TreeNode> nodes = getSelectedNodes();
    return SelectionManager.toMap(nodes);
  }

  @Override
  public String getGroupId() {
    return EventSource.EXPLORER;
  }

  @Override
  public boolean contains(ISchedulingRule rule) {
    return false;
  }

  @Override
  public boolean isConflicting(ISchedulingRule rule) {
    return rule == this;
  }
}
