package com.unitesk.requality.eclipse.ui.dialogs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class ExportToHtmlSettings extends Dialog {
  TreeNode req;
  Object filter;
  List refs;

  @Override
  protected org.eclipse.swt.widgets.Control createContents(Composite parent) {

    Composite p = new Composite(parent, SWT.NONE);
    p.setLayout(new GridLayout(1, true));
    GridData gd = new GridData();
    gd.widthHint = 250;
    gd.heightHint = 300;
    parent.setLayoutData(gd);
    p.setLayoutData(gd);
    new Label(p, SWT.NONE).setText("Select target Reference name:");

    refs = new List(p, SWT.V_SCROLL);
    gd = new GridData(GridData.FILL_BOTH);
    refs.setLayoutData(gd);
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(req, true, true, false);
    Iterator<TreeNode> iter = dftw.iterator();
    final Map<String, Integer> cnt = new HashMap<>();
    int sel = -1;
    while (iter.hasNext()) {
      TreeNode node = iter.next();
      for (Attribute a : node.attrs()) {
        if (a.getType().equals(AttributeType.REFERENCE)) {
          if (!cnt.containsKey(a.getKey())) {
            if (filter != null && filter instanceof String && filter.equals(a.getKey()))
              //TODO list
              sel = cnt.keySet().size();
            cnt.put(a.getKey(), 0);
          }
          cnt.put(a.getKey(), cnt.get(a.getKey()) + 1);
        }
      }
    }
    for (String s : cnt.keySet()) refs.add(s + "(" + cnt.get(s) + ")");
    if (sel > -1) refs.select(new ArrayList<>(cnt.keySet()).indexOf(filter));
    refs.addSelectionListener(
        new SelectionListener() {
          @Override
          public void widgetSelected(SelectionEvent e) {
            if (refs.getSelectionIndex() > -1)
              filter = cnt.keySet().toArray()[refs.getSelectionIndex()];
            //TODO list
          }

          @Override
          public void widgetDefaultSelected(SelectionEvent e) {
            // TODO Auto-generated method stub

          }
        });

    Composite bc = new Composite(p, SWT.NONE);
    bc.setLayout(new GridLayout(2, true));
    bc.setLayoutData(new GridData(GridData.VERTICAL_ALIGN_END));
    Button cancelB = new Button(bc, SWT.NONE), okB = new Button(bc, SWT.NONE);

    okB.setText("Ok");
    gd = new GridData();
    gd.widthHint = 100;
    okB.setLayoutData(gd);
    okB.addListener(
        SWT.MouseDown,
        new Listener() {
          @Override
          public void handleEvent(Event event) {
            okPressed();
          }
        });

    cancelB.setText("Cancel");
    gd = new GridData();
    gd.widthHint = 100;
    cancelB.setLayoutData(gd);
    cancelB.addListener(
        SWT.MouseDown,
        new Listener() {
          @Override
          public void handleEvent(Event event) {
            cancelPressed();
          }
        });

    parent.layout(true, true);
    return p;
  }

  TreeNode rep;
  Shell shell;

  public ExportToHtmlSettings(Shell shell, TreeNode rep) {
    super(shell);
    shell.setText("Treceability report settings");
    this.shell = shell;
    this.rep = rep;
    Attribute ft = rep.attrs().getAttribute("filter");
    if (ft != null && ft.getType() == AttributeType.STRING)
      filter = Attribute.valueToString(rep.attrs(), ft.getType(),ft.getValue());
    req = rep.getTreeDB().getNode(((ReportSettings) rep).getStartRequirementQId(), Path.QID);
  }

  @Override
  protected void okPressed() {
    try {
      if (!rep.getTreeDB().startTransaction("Change report settings")) return;
      if (filter != null && filter instanceof String)
        rep.attrs().setConstAttribute(AttributeType.STRING, "filter", filter, 0);
      rep.commitAttributes();
      rep.getTreeDB().commit();
      super.okPressed();
    } catch (Exception e) {
      e.printStackTrace();
    }
  };
}
