package com.unitesk.requality.eclipse.ui.dialogs;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.eclipse.core.resources.IProject;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Text;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.SortingSettings;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.SystemNames;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.operations.mass.NodeSearch;
import com.unitesk.requality.core.search.AttributeOperator;
import com.unitesk.requality.core.search.SearchMethod;
import com.unitesk.requality.core.search.SearchRequest;
import com.unitesk.requality.core.selection.IUIStateChangeListener;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;

/** Find dialog */
public class FindNode extends Dialog {
  private static String TITLE = "Search Node";
  private static String LBL_FIND = "Search Node:";
  private static String LBL_SB = "Search By:",
                        LBL_PATH = "Path:",
                        LBL_ANAME = "Attribute name:";

  private Text path;
  private Text aname; 
  private Text aval; 
  private TabItem tab1, tab2;
  private TabFolder folder;
  private Combo aop;
  
  RequalityCNF cnf;
  private TreeNode[] ns = null;
  int pos = 0;
  private IUIStateChangeListener list = new IUIStateChangeListener(){
    @Override
    public void newActiveNodes(List<TreeNode> nodes) {
      ns = null;
      clearNodes();
    }
  };
  //private AttributeOperator ao = null;
  private AttributeOperator getAO(){
    int c = aop.getSelectionIndex();
    AttributeOperator ao = null;
    if(c>-1)
      ao = AttributeOperator.values()[aop.getSelectionIndex()];
    return ao;
  }
  private void clearNodes(){
    ns = null;
    pos = 0;
  }
  public FindNode(Shell shell, RequalityCNF c) {
    super(shell);
    setShellStyle(SWT.DIALOG_TRIM | SWT.CLOSE | SWT.MODELESS | SWT.BORDER | SWT.TITLE);
    cnf = c;
  }
  @Override
  public boolean close() {
    try{
      return super.close();
    }finally{
      SelectionManager.getOne().removeChangeListener(EventSource.GENERAL, list);
    }
  }
  @Override
  protected Control createDialogArea(Composite parent) {
    getShell().setText(TITLE);
    Composite comp1 = new Composite(parent, SWT.NONE);
    comp1.setLayout(new GridLayout(1,true));
    comp1.setLayoutData(new GridData(GridData.FILL_BOTH));
    
    Label l = new Label(comp1, SWT.NONE);
    l.setText(LBL_SB);
    
    folder = new TabFolder(comp1, SWT.NONE);
    folder.setLayoutData(new GridData(GridData.FILL_BOTH));
    folder.setLayout(new GridLayout(1,true));
    
    
    //Tab 1
    tab1 = new TabItem(folder, SWT.NONE);
    tab1.setText("By path");
    
    Composite c = new Composite(folder, SWT.NONE);
    c.setLayout(new GridLayout(2,false));
    c.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    
    Label l2 = new Label(c, SWT.NONE);
    l2.setText(LBL_PATH);
    path = new Text(c, SWT.NONE);
    path.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    path.addModifyListener(new ModifyListener(){
      @Override
      public void modifyText(ModifyEvent e) {
        clearNodes();
      }
    });
    
    tab1.setControl(c);
    
    tab2 = new TabItem(folder, SWT.NONE);
    tab2.setText("By attributes");
    
    Composite c2 = new Composite(folder, SWT.NONE);
    c2.setLayout(new GridLayout(3,false));
    c2.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    
    Label l3 = new Label(c2, SWT.NONE);
    l3.setText(LBL_ANAME);
    
    Combo anames = new Combo(c2, SWT.READ_ONLY);
    initANames(anames, cnf);
    anames.setText("...");
    anames.addSelectionListener(new SelectionListener(){
    
      @Override
      public void widgetSelected(SelectionEvent e) {
        aname.setText(anames.getText());
      }
    
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });
    
    aname = new Text(c2, SWT.NONE);
    aname.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    aname.addModifyListener(new ModifyListener(){
    
      @Override
      public void modifyText(ModifyEvent e) {
        clearNodes();
      }
    });

    anames.select(0);//init by names
    Label l4 = new Label(c2, SWT.NONE);
    l4.setText("Method:");
    aop = new Combo(c2, SWT.READ_ONLY);
    GridData gd = new GridData(GridData.FILL_HORIZONTAL);
    gd.horizontalSpan = 2;
    aop.setLayoutData(gd);
    for(AttributeOperator sm : AttributeOperator.values())
      aop.add(sm.name());
    aop.select(0);
    
    aop.addSelectionListener(new SelectionListener(){
      @Override
      public void widgetSelected(SelectionEvent e) {
        clearNodes();
        int i = aop.getSelectionIndex();
        AttributeOperator ao = null;
        if(i>-1)
          ao = AttributeOperator.values()[aop.getSelectionIndex()];
        aval.setEnabled(true);
        if(ao == AttributeOperator.NOT_EMPTY || ao == AttributeOperator.EMPTY)
          aval.setEnabled(false);
      }
    
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });
    
    Label l5 = new Label(c2, SWT.NONE);
    l5.setText("Value:");
    
    aval = new Text(c2, SWT.NONE);
    aval.addModifyListener(new ModifyListener(){
    
      @Override
      public void modifyText(ModifyEvent e) {
        clearNodes();
      }
    });
    gd = new GridData(GridData.FILL_HORIZONTAL);
    gd.horizontalSpan = 2;
    aval.setLayoutData(gd);
    
    tab2.setControl(c2);
    SelectionManager.getOne().addChangeListener(EventSource.GENERAL, list);
    return parent;
  }

  private void initANames(Combo c, RequalityCNF cnf2) {
    for(String s: SystemNames.getVisSysNames())
      c.add(s);
  }

  protected void createButtonsForButtonBar(Composite parent) {
    createButton(parent, IDialogConstants.CANCEL_ID, "Close", false);
    Button spaceButton = createButton(parent, 3, "", false);
    spaceButton.setVisible(false);
    createButton(parent, IDialogConstants.CLIENT_ID + 1, "<<", false);
    createButton(parent, IDialogConstants.CLIENT_ID + 2, ">>", true);
    //        if (comboStrings.isEmpty()) {
//    if (combo.getText().isEmpty()) {
//      getButton(IDialogConstants.CLIENT_ID + 1).setEnabled(false);
//      getButton(IDialogConstants.CLIENT_ID + 2).setEnabled(false);
//    }
  }

  @Override
  protected void buttonPressed(int buttonId) {
    super.buttonPressed(buttonId);
    //if (combo.isDisposed()) return;
    //if (combo.getText().isEmpty()) return;
    saveTextToCombo();
    SelectionManager.getOne().removeChangeListener(EventSource.GENERAL, list);
    if (IDialogConstants.CLIENT_ID + 1 == buttonId) {
      initNodesSet();
      pos--;
      if(pos<0)
        pos = ns.length-1;
      if(pos<0)
        pos = 0;
      setSel(pos);
//      if (browser != null) {
//        browser.findText(combo.getText(), selectCaseSensitive, true, false);
//      }
      //                    selectWrapAround);
    }
    if (IDialogConstants.CLIENT_ID + 2 == buttonId) {
      initNodesSet();
      pos++;
      if(pos>=ns.length)
        pos = 0;
      setSel(pos);
//      if (browser != null) {
//        browser.findText(combo.getText(), selectCaseSensitive, false, false);
//        //                    selectWrapAround);
//      }
    }
    SelectionManager.getOne().addChangeListener(EventSource.GENERAL, list);
  }

  private void setSel(int pos2) {
    Map<String,java.util.List<TreeNode>> selm = new HashMap<>();
    if(pos2>=ns.length)
      return;
    TreeNode nd = ns[pos2];
    if(nd!=null){
      selm.put(nd.getType(), new ArrayList<>());
      selm.get(nd.getType()).add(nd);
    }   
    SelectionManager.getOne().setActiveNodes(selm,selm,EventSource.GENERAL);
  }
  private void initNodesSet() {
    if(ns!=null)
      return;
    pos = -1;
    SearchMethod sm = folder.getSelectionIndex()>0
        ?
          SearchMethod.BY_ATTRIBUTE
        :
          SearchMethod.BY_PATH;
    Map<String,TypedValue> params = new HashMap<>();
    if(sm == SearchMethod.BY_PATH)
      params.put("path", new TypedValue(AttributeType.STRING, path.getText()));
    else
    {
      params.put("sample", new TypedValue(AttributeType.STRING, aval.getText()));
      params.put("op", new TypedValue(AttributeType.STRING,getAO().name()));
      params.put("key", new TypedValue(AttributeType.STRING,aname.getText()));
      String sn = SystemNames.getSysName(aname.getText());
      if(sn != null)
        params.put("key", new TypedValue(AttributeType.STRING,sn));  
    }
    SearchRequest sr = new SearchRequest(sm, params);
    NodeSearch nsh = new NodeSearch(sr);
    
    java.util.List<TreeNode> found = new LinkedList<TreeNode>();
    java.util.List<Object> sel = cnf.getSelectedObj();
    if(sel == null || sel.isEmpty()){
      for(TreeDB db:ResManager.getManager().getTreeDBs())
      {
        nsh.execute(db.getRootNode());
        found.addAll(nsh.getMatches());
      }
    }else
      for(Object o:sel){
        if(o instanceof IProject){
          TreeDB db = EclipseTracker.getTracker().getTree((IProject) o);//TODO mv resMan
          nsh.execute(db.getRootNode());
          found.addAll(nsh.getMatches());
        }else if(o instanceof TreeNode){
          nsh.execute((TreeNode)o);
          found.addAll(nsh.getMatches());
        }
      }
    this.ns = found.toArray(new TreeNode[0]);
    Arrays.sort(this.ns, TreeNodeComparator.getGComparator(SortingSettings.init(SortingMethod.BY_INDEX, false)));
  }
  
  private void saveTextToCombo() {
//    String text = combo.getText();
//    if (comboStrings.contains(text)) {
//      comboStrings.remove(text);
//    }
//    comboStrings.add(0, text);
//    combo.setItems(comboStrings.toArray(new String[0]));
//    combo.setText(text);
  }
}
