/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.ui.dialogs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.eclipse.ui.cnf.RequalitySorter;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class ReferenceNameSelectionDialog extends Dialog {

  protected TreeDB db;
  private TreeViewer nodeTree = null;
  private Map<String, Object> attrMap;
  private static Map<TreeDB, String[]> attrSel = new HashMap<>();

  public ReferenceNameSelectionDialog(TreeDB tn, Shell shell) {
    super(shell);
    db = tn;
  }

  protected void configureShell(Shell shell) {
    super.configureShell(shell);
    shell.setText("Select reference name(s)");
  }

  protected ILabelProvider getLabelProvider() {
    return new ILabelProvider() {

      @Override
      public void removeListener(ILabelProviderListener listener) {
        // TODO Auto-generated method stub

      }

      @Override
      public boolean isLabelProperty(Object element, String property) {
        // TODO Auto-generated method stub
        return false;
      }

      @Override
      public void dispose() {
        // TODO Auto-generated method stub

      }

      @Override
      public void addListener(ILabelProviderListener listener) {
        // TODO Auto-generated method stub

      }

      @Override
      public String getText(Object element) {
        if (element instanceof Attribute) return ((Attribute) element).getKey();
        return element.toString();
      }

      @Override
      public Image getImage(Object element) {
        // TODO Auto-generated method stub
        return null;
      }
    };
  }

  @Override
  protected Control createDialogArea(Composite parent) {
    //setPageComplete(true);
    parent.setLayout(new GridLayout(1, true));
    Composite container = new Composite(parent, SWT.NULL);
    GridData gd = new GridData(GridData.FILL_BOTH);
    gd.heightHint = 350;
    gd.widthHint = 100;
    container.setLayoutData(gd);
    GridLayout layout = new GridLayout();
    container.setLayout(layout);
    layout.verticalSpacing = 9;

    gd = new GridData(GridData.FILL_BOTH);

    nodeTree = new TreeViewer(container, SWT.MULTI);
    nodeTree.getControl().setLayoutData(gd);
    nodeTree.setLabelProvider(getLabelProvider());
    nodeTree.setAutoExpandLevel(2);
    nodeTree.setComparator(new RequalitySorter());
    if (db != null) {
      nodeTree.setContentProvider(getContentProvider());
      nodeTree.setInput(db);
      if (attrSel.containsKey(db)) {
        if (attrMap != null) {
          List<Object> tosel = new ArrayList<>();
          for (String s : attrSel.get(db)) if (attrMap.containsKey(s)) tosel.add(attrMap.get(s));
          nodeTree.setSelection(new StructuredSelection(tosel.toArray()));
        }
      }
    } else {
      nodeTree.setContentProvider(getContentProvider());
    }
    nodeTree.addSelectionChangedListener(
        new ISelectionChangedListener() {
          @Override
          public void selectionChanged(SelectionChangedEvent event) {
            StructuredSelection ss = (StructuredSelection) nodeTree.getSelection();
            if (ss.isEmpty()) return;
            @SuppressWarnings("unchecked")
            Iterator<Object> i = ss.iterator();
            List<String> ts = new ArrayList<>();
            boolean clear = false;
            while (i.hasNext()) {
              Object o = i.next();
              if (o instanceof Attribute) ts.add(((Attribute) o).getKey());
              else clear = true;
            }
            if (clear) {
              if (attrSel.containsKey(db)) attrSel.remove(db);
            } else attrSel.put(db, ts.toArray(new String[0]));
          }
        });
    return container;

    //setControl( container );

  }
  /** Ensures that both text fields are set. */
  protected ITreeContentProvider getContentProvider() {
    return new ITreeContentProvider() {
      @Override
      public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {}

      @Override
      public void dispose() {}

      @Override
      public boolean hasChildren(Object element) {
        return element instanceof TreeNode;
      }

      @Override
      public Object getParent(Object element) {
        if (element instanceof Attribute) {
          return db;
        }
        return null;
      }

      @Override
      public Object[] getElements(Object inputElement) {
        if (inputElement instanceof TreeDB) {
          attrMap = new HashMap<>();
          Set<TreeNode> nodes = new HashSet<>();
          DeepFirstTreeWalker dftw =
              new DeepFirstTreeWalker(
                  db.getNode(Requirement.getTypeRootQId(), Path.QID), true, true, false);
          Iterator<TreeNode> it = dftw.iterator();
          while (it.hasNext()) nodes.add(it.next());
          for (TreeNode n : nodes) {
            for (Attribute a : n.attrs())
              if (a.getType().equals(AttributeType.REFERENCE)
                  && !attrMap.containsKey(a.getKey())
                  && a.getValue() != null 
                  && !a.getValue().toString().replace("\n", "").trim().equals(""))
                attrMap.put(a.getKey(), a);
          }
          Object[] ret = attrMap.values().toArray(new Object[attrMap.size() + 1]);
          ret[ret.length - 1] = "~Use all references~";
          return ret;
        }
        return null;
      }

      @Override
      public Object[] getChildren(Object parentElement) {
        return getElements(parentElement);
      }
    };
  }

  public static Map<TreeDB, String[]> getSavedState() {
    Map<TreeDB, String[]> ret = new HashMap<>();
    for (TreeDB d : attrSel.keySet()) ret.put(d, attrSel.get(d).clone());
    return ret;
  }

  public static void restoreSavedState(Map<TreeDB, String[]> back) {
    attrSel = back;
  }
}
