package com.unitesk.requality.eclipse.ui.editors;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.browser.BrowserFunction;
import org.eclipse.swt.browser.LocationEvent;
import org.eclipse.swt.browser.LocationListener;
import org.eclipse.swt.browser.ProgressEvent;
import org.eclipse.swt.browser.ProgressListener;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.navigator.CommonNavigator;
import org.eclipse.ui.operations.UndoRedoActionGroup;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.part.IContributedContentsView;
import org.eclipse.ui.progress.UIJob;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.PastingIntoLeafException;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.ITreeLogic.TreeLogicException;
import com.unitesk.requality.core.ITreeLogic.WrongProjectException;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.SortingSettings;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeLogic;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeChangeEvent;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.EnumValue;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.IContextStringProvider;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.attribute.Version;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.editors.browser.ReqMarker;
import com.unitesk.requality.eclipse.handlers.DeleteSelection;
import com.unitesk.requality.eclipse.handlers.copypaste.CopyHandler;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipseCopyHandler;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipsePasteHandler;
import com.unitesk.requality.eclipse.handlers.copypaste.PasteEclipseHandler;
import com.unitesk.requality.eclipse.ui.ObjectRule;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.eclipse.ui.dialogs.EditElementDialog;
import com.unitesk.requality.eclipse.ui.dialogs.SelectNodeDialog;
import com.unitesk.requality.eclipse.ui.dialogs.SelectNodesFromList;
import com.unitesk.requality.eclipse.ui.dialogs.SelectTestPurposeDialog;
import com.unitesk.requality.eclipse.ui.dialogs.TemplatesSelectionDialog;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.tests.TestDialogProvider;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

public abstract class JSEditor extends EditorPart implements ISelectionProvider,ISearchProvider {

  public static Set<String> defaultAttrs = new HashSet<>();
  public static final int EDIT_MODE_REQ_RO = 0x01;
  public static final int EDIT_MODE_TP_RO = 0x02;
  private Browser browser;
  protected File html;
  private TreeNode rootNode;
  private TreeNode currentSelection;
  private TreeDB db;
  private List<TreeDB> dbs;
  private UniEditorOutlinePage uniEditorOutlinePage;
  private boolean javaDialogOpened = false;
  private int editMode = 0;
  private int editVirtualMode = 8;
  private SortingSettings storedSortingSettings = null;
  private Set<String> additionalAttrs = new HashSet<>();
  private Set<String> systemAttrs = new HashSet<>();
  private ObjectRule or = null;

  java.util.Timer updateFreeTimer = new Timer();
  private boolean isUpdating = false;

  public String getSelectionText() {
    return getSelectionText(true);
  }

  private String getSelectionText(boolean process) {
    String res = (String) browser.evaluate("return getSelectionText();");
    if (res == null) return "";
    if (!process) return res;
    StringTokenizer st = new StringTokenizer(res, "\n ");
    res = "";
    while (st.hasMoreTokens()) {
      if (!res.isEmpty()) {
        res += " ";
      }
      res += st.nextToken();
    }
    return res;
  }

  final class BrowserKeyListener implements KeyListener {
    @Override
    public void keyReleased(KeyEvent e) {
      if ((e.stateMask & SWT.CTRL) != 0) {
        if (e.keyCode == 'c') {
          String textData = getSelectionText(false);
          if (textData == null || textData.trim().equals("")) return;
          Clipboard clipboard = new Clipboard(getSite().getShell().getDisplay());
          TextTransfer textTransfer = TextTransfer.getInstance();
          Transfer[] transfers = new Transfer[] {textTransfer};
          Object[] data = new Object[] {textData};
          clipboard.setContents(data, transfers, DND.CLIPBOARD);
          clipboard.dispose();
        }
        if (e.keyCode == 'v') {
          //                	Clipboard clipboard = new Clipboard(getSite().getShell().getDisplay());
          //                    TextTransfer textTransfer = TextTransfer.getInstance();
          //                    Object o = clipboard.getContents(textTransfer);
          //                    clipboard.dispose();
          //                    String textData = (String)o;
          //                    if(textData == null || textData.trim().equals(""))
          //                        return;
          //                    setSelectionText(CoreUtils.htmlEscape(textData));

        }
      }
      if ((e.stateMask & SWT.CTRL) != 0 && (e.stateMask & SWT.SHIFT) != 0 && e.keyCode == 'd') {
        CoreUtils.DEBUG_LEVEL = 1;
      }
      if (e.keyCode == SWT.F5) {
        oldRefresh(rootNode);
        browser.refresh();
      }
    }

    @Override
    public void keyPressed(KeyEvent e) {}
  }

  private ITreeStateListener treeStateListener =
      new ITreeStateListener() {

        @Override
        public void open(TreeDB db) {
          // TODO Auto-generated method stub
          //add tree change listener
          if (JSEditor.this.db.equals(db) && !db.getGUIListeners().contains(guiTreeChangeListener))
            db.addGUIListener(guiTreeChangeListener);
        }

        @Override
        public void close(TreeDB db) {
          //remove tree change listener

          db.removeGUIListener(guiTreeChangeListener);
          if (!JSEditor.this.db.equals(db)) {
            return;
          }
          closeEditor();
        }

        @Override
        public void delete(IProject proj) {
          IProject project = EclipseTracker.getTracker().getProject(db);
          if (!project.equals(proj)) {
            return;
          }
          closeEditor();
        }
      };

  private ISelectionChangedListener outlineSelectionChangeListener =
      new ISelectionChangedListener() {
        @Override
        public void selectionChanged(SelectionChangedEvent event) {
          IStructuredSelection selection = (IStructuredSelection) event.getSelection();
          if (selection.isEmpty()) {
            return;
          }
          TreeNode sel = (TreeNode) selection.getFirstElement();
          if(!relates(sel))
            return;
          if(currentSelection != null && sel.getUUId().equals(currentSelection.getUUId()))
            return;
          if (selection instanceof NodeEditorInput) {
            TreeNode dat = (TreeNode) ((NodeEditorInput) selection).getData();
            if (dat != null) sel = dat;
          }
          if (sendtooutline.contains(sel.getUUId())) {
            sendtooutline.remove(sel.getUUId());
            return;
          }
          setUniEditorFocus(sel.getUUId());
          currentSelection = sel;
          // synchronization with Requality Explorer
          IViewPart view =
              PlatformUI.getWorkbench()
                  .getActiveWorkbenchWindow()
                  .getActivePage()
                  .findView(RequalityCNF.ID);
          if (view instanceof CommonNavigator) {
            CommonNavigator navigator = (CommonNavigator) view;
            if (navigator.isLinkingEnabled()) {
              navigator.setLinkingEnabled(true);
            }
          }
        }
      };

  private List<UUID> sendtooutline = new ArrayList<>();

  public TreeNode getCurrentSelection() {
    if (currentSelection == null) {
      currentSelection = targetNode;
    }
    return currentSelection;
  }

  public void setSelectedNode(TreeNode node) {
    UUID nuid = null;
    if (node != null)
      nuid = node.getUUId();
//      currentSelection = node;
//      return;
//    }

    setUniEditorFocus(nuid);
    currentSelection = node;
    if(nuid!=null)
      sendtooutline.add(nuid);
    if (uniEditorOutlinePage != null && !uniEditorOutlinePage.getControl().isDisposed()) {
      StructuredSelection ss = StructuredSelection.EMPTY;
      if(nuid!=null)
        ss = new StructuredSelection(node); 
      uniEditorOutlinePage.setSelection(ss);
    }
  }

  public boolean findText(
      String aString, boolean aCaseSensitive, boolean aBackwards, boolean aWrapAround) {
    return (Boolean)
        browser.evaluate(
            "return findText('"
                + aString
                + "',"
                + aCaseSensitive
                + ","
                + aBackwards
                + ","
                + aWrapAround
                + ");");
  }

  private static String escape(String a) {
    return a.replace("\\\"", "\\\\\"").replace("\"", "\\\"").replace("\\n", "\\\\n").replace("\\r", "\\\\r").replace("\\t", "\\\\t");
    //if(a.startsWith("\\\""))
    //	a = a.substring(1,a.length()-1);
    //return a.replace("\"","\\\\\"");
  }

  private Map<UUID, NodeDesc> changedNodes;
  private Map<UUID, Integer> changedNodesCount;
  ITreeChangeListener guiTreeChangeListener =
      new DefaultTreeChangeListener() {
        private Object getMoved(TreeNode node2, TreeNode oldParent) {
          TreeNode[] nodes = new TreeNode[2];
          if (!relates(node2) || !relates(node2.getShownParent()) || !relates(oldParent))
            return null;
          if (relates(node2.getShownParent())) {
            nodes[0] = node2.getShownParent();
          }
          if (!node2.getShownParent().getUUId().equals(oldParent.getUUId()))
            if (relates(oldParent)) nodes[1] = oldParent;
          //refresh(oldParent);
          //refresh(node2);
          JSONObject jo = new JSONObject();
          try {
            jo.put("op", "moved");
            jo.put("node", node2.getUUId().toString());
            UUID old = nodes[0] == null ? null : nodes[0].getUUId();
            if (old != null) jo.put("parent", old);
            UUID opar = nodes[1] == null ? null : nodes[1].getUUId();
            if (opar != null) jo.put("oldparent", opar);
            return jo;
          } catch (JSONException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
          return null;
          //return new Object[]{"moved",node2.getUUId(),,nodes[1]==null?null:nodes[1].getUUId()};
        }
        public JSONObject getChildrenOrderChanged(TreeNode node){
          if(!relates(node))
            return null;
          JSONObject jo1 = new JSONObject();
          try{
            jo1.put("op", "updateChildrenOrder");
            jo1.put("node", node.getUUId().toString());
          }
          catch(JSONException e){
            e.printStackTrace();
          }
          return jo1;
        }
        
        @Override
        public void childrenOrderChanged(TreeNode node) {
          if(!relates(node))
            return;
          JSONObject jo = getChildrenOrderChanged(node);
          if(jo!=null)
            sendEvents(node.attrs(),jo);
            //super.childrenOrderChanged(node);
        }
        @Override
        public void moved(TreeNode node2, TreeNode oldParent, TreeNode newParent) {
          if (relates(node2)) { //TODO collapsed conflict
            JSONObject jo = new JSONObject();
            try {
              jo.put("op", "moved");
              jo.put("node", node2.getUUId().toString());
              jo.put("parent", node2.getShownParentUUID().toString());
              jo.put("oldparent", oldParent.getUUId().toString());
              sendEvents(node2.attrs(), jo);
            } catch (Exception e) {
              e.printStackTrace();
            }
          }
        }

        private Object getDeleted(TreeNode nodeDesc) {
          if (JSEditor.this.rootNode.isDisposed()) { // current node was deleted
            closeEditor();
            //May cause errors
            return null;
          }
          if (!relates(nodeDesc))
            return null;
          if (nodeDesc == null) RLogger.logInfo("null desc");
          if (nodeDesc != null) {
            JSONObject jo = new JSONObject();
            try {
              jo.put("op", "deleted");
              jo.put("node", nodeDesc.getUUId().toString());
              try{
                if(nodeDesc.getShownParentUUIDExc()!=null)
                  jo.put("parent", nodeDesc.getShownParentUUID().toString());
              }catch(Exception e){
                
              }
              return jo;
            } catch (Exception e) {
              e.printStackTrace();
            }
          }
          return null;
        }

        private Object getDeleted(NodeDesc nodeDesc) {
          if (JSEditor.this.rootNode.isDisposed()) { // current node was deleted
            closeEditor();
            //May cause errors
            return null;
          }
          if (nodeDesc != null) {
            JSONObject jo = new JSONObject();
            try {
              jo.put("op", "deleted");
              jo.put("node", nodeDesc.getUUId().toString());
              if(nodeDesc.getShownParentUUID() != null)
                jo.put("parent", nodeDesc.getShownParentUUID().toString());
              return jo;
            } catch (Exception e) {
              e.printStackTrace();
            }
          }
          return null;
        }

        @Override
        public void deleted(TreeNode node) {
          sendEvents(node.attrs(), getDeleted(node));
        }

        private Object getChanged(TreeNode nde, ChangesSet<Attribute> what) {
          //	TreeNode[] ret = new TreeNode[2];
          if (!relates(nde)) return null;
          if (nde.isRoot()) {
            //TODO reduction
            //refresh root
            if (!db.getSortingSettings().equals(storedSortingSettings)) {
              storedSortingSettings = db.getSortingSettings();
              //TODO disposed case
              //if(!browser.isDisposed())
              //browser.refresh();
              //just update and do nothing else
              return nde;
            }
          }
          TreeNode n = nde;
          //if(n == null || qid.equals(rootqid) || nde.getQualifiedId().equals(db.getRootNode().getQualifiedId()))
          //  n = nde;
          boolean changeIt = false, changeParent = false;
          Set<String> requiredchange = new HashSet<>(),
              changesA = what.getAll();
          for (String w : changesA) {
            if(additionalAttrs.contains(w)||systemAttrs.contains(w)){
              requiredchange.add(w);
              changeIt = true;
            }
          }
          //Temporary solution. tends to become new virtual attr
          changeParent = changesA.contains(TreeDB.INDEX_ATTRIBUTE)||changesA.contains(TreeNode.ATTR_LABEL)||changesA.contains(Requirement.ATTR_LOCATIONS);
          if (n.getQualifiedId(true).equals(rootNode.getQualifiedId(true))) changeParent = false;
          JSONArray changes = new JSONArray();
          //Object[] attrchanges = null;
          if (changeIt) {
            for (String s : requiredchange) {
              try {
                Attribute a = n.attrs().getAttribute(s);
                JSONObject ach = new JSONObject();
                ach.put("key", s);
                if (a != null) {
                  TypedValue rw = a.getDefinition().getRawValue();
                  if(rw!=null){
                    ach.put(
                      "value", Attribute.rawValueToJSON(rw.getType(), rw.getValue(),n.getNodeResourcesPath(s).toString()));//TODO
                  }
                  ach.put(
                      "newvalue",
                      Attribute.rawValueToJSON(
                          a.getType(), a.getValue(),(CoreUtils.URLEncode(n.getNodeResourcesPath().toString()))));
                  //            			attrchanges[i++] = new Object[]{s,escape(Attribute.rawValueToJSONString(a.getRawType(), a.getRawValue())),
                  //            				escape(Attribute.resultValueToJSONString(a.getResultType(), a.getValue()))};
                }
                changes.put(ach);
              } catch (Exception e) {
                e.printStackTrace();
              }
            }
          }
          try {
            if (changeParent && changeIt) {
              //JSONArray joa = new JSONArray();
              JSONObject jo1 = new JSONObject();
              jo1.put("op", "updateChildrenOrder");
              jo1.put("node", n.getShownParent().getUUId().toString());
              //joa.put(jo1);
              JSONObject jo2 = new JSONObject();
              jo2.put("op", "changeAttributes");
              jo2.put("node", n.getUUId().toString());
              jo2.put("values", changes);
              //	            	joa.put(jo2);
              return new JSONObject[] {jo1, jo2};
              //return new Object[]{new Object[]{"updateChildrenOrder",n.getShownParent().getUUId()},
              //				new Object[]{"changeAttributes", n.getUUId().toString(), attrchanges}};
            }
            if (changeParent && n.getShownParent() != null) {
              JSONObject jo1 = new JSONObject();
              jo1.put("op", "updateChildrenOrder");
              jo1.put("node", n.getShownParent().getUUId().toString());
              return jo1;
            }
            if (changeIt && !changeParent) {
              JSONObject jo2 = new JSONObject();
              jo2.put("op", "changeAttributes");
              jo2.put("node", n.getUUId().toString());
              jo2.put("values", changes);
              return jo2;
              //return new Object[]{"changeAttributes", n.getUUId().toString(), attrchanges};
            }
          } catch (Exception e) {
            e.printStackTrace();
          }
          return null;
        }

        @Override
        public void changed(
            final TreeNode nde, final ChangesSet<Attribute> what) {
          Object o = getChanged(nde, what);
          if (o != null) {
            if (o instanceof TreeNode) oldRefresh(rootNode);
            else sendEvents(nde.attrs(), o);
          }
          //        	if(nde2!=null)
          //        		if(nde2.isRoot()){
          //        			oldRefresh(rootNode);
          //        			if(!browser.isDisposed())
          //        				browser.refresh();
          //        		}else{
          //        			updateData(new HashMap<UUID,Map<String,Object>>(){{
          //        				put(nde.getUUId(),new HashMap<String, Object>());
          //        				for(String key:what){
          //        					Attribute a = nde.getAttribute(key);
          //        					get(nde.getUUId()).put(key, a==null?null:a.getValue());
          //        				}
          //        				}});
          //        			//targeted update
          //        			//refresh(nde2);
          //        		}
        }


        private Object getCreated(TreeNode nde) {
          if (!relates(nde)) return null;
          JSONObject jo = new JSONObject();
          try {
            jo.put("op", "created");
            jo.put("node", nde.getUUId().toString());
            if (nde.getShownParentUUID() != null)
              jo.put("parent", nde.getShownParentUUID().toString());
          } catch (Exception e) {
          }
          return jo;
        }

        @Override
        public void created(TreeNode nde) {
          Object o = getCreated(nde);
          if (o != null) {
            sendEvents(nde.attrs(), o);
          }
        }

        @Override
        public void changed(List<TreeNodeChangeEvent> changes) {
          //List<Object> ret = new LinkedList<Object>();
          StringBuilder sb = new StringBuilder();
          sb.append("[");
          for (TreeNodeChangeEvent event : changes) {
            switch (event.getType()) {
              case CHANGE:
                {
                  @SuppressWarnings("unchecked")
                  Object o =
                      getChanged(event.getTreeNodeTarget(), (ChangesSet<Attribute>) event.getArguments()[1]);
                  if (o != null && !(o instanceof TreeNode)){ //?node??
                    if(sb.length()>1)
                      sb.append(", ");
                    sb.append(JSEditor.toString(event.getTreeNodeTarget(),o));
                  }
                  //ret.add(o);
                }
                break;
              case CREATE:
                {
                  Object o = getCreated(event.getTreeNodeTarget());
                  if (o != null){
                    if(sb.length()>1)
                      sb.append(", ");
                    sb.append(JSEditor.toString(event.getTreeNodeTarget(),o));
                    
                  }
                }
                break;
              case MOVE:
                {
                  Object o =
                      getMoved(event.getTreeNodeTarget(), (TreeNode) event.getArguments()[1]);
                  if (o != null){
                    if(sb.length()>1)
                      sb.append(", ");
                    sb.append(JSEditor.toString(event.getTreeNodeTarget(),o));
                  }
                }
                break;
              
              case CHILDREN_ORDER_CHANGE:{
                TreeNode node = event.getTreeNodeTarget();
                Object o =  getChildrenOrderChanged(node);
                if(o!=null){
                  if(sb.length()>1)
                    sb.append(", ");
                  sb.append(JSEditor.toString(node,o));
                }
              }
              case DELETE:
                {
                  if(event.getArguments()[2]==Moment.BEFORE)
                    continue;
                  TreeNode nd = event.getTreeNodeTarget();
                  Object o = null;
                  if (nd != null) o = getDeleted(nd);
                  else o = getDeleted(event.getTreeNodeTarget());
                  if (o != null){
                    if(sb.length()>1)
                      sb.append(", ");
                    sb.append(JSEditor.toString(event.getTreeNodeTarget(),o));
                  }
                }
                break;
              default:
                break;
            }
          }
          if (sb.length() > 1){
            sb.append("]");
            sendEvents(sb.toString());
          }
        }

        private JSONObject getVChanges(TreeNode nde, boolean internal) {
          if (!relates(nde)) return null;
          JSONObject jo = new JSONObject();
          try {
            jo.put("op", "versionChanged");
            jo.put("node", nde.getUUId());
          } catch (Exception e) {
            e.printStackTrace();
          }
          return jo;
        }

      };
  //Check if id is valid
  private static boolean reqIdValidExc(TreeNode req, TreeNode parent, String text)
      throws SameIdExistsInChildren, IdEmpty, IdContainsWrongChars {
    if (req == null && parent == null) return false;
    TreeLogic tlogic;
    if (req == null) {
      tlogic = parent.getTreeDB().getTreeLogic();
    } else {
      tlogic = req.getTreeDB().getTreeLogic();
    }
    return tlogic.isChildIdValidExc(parent, req, text);
  }

  private static void dataToString(AttributeStorage attrs, Object obj, JSONArray ja) throws JSONException {
    if (obj instanceof JSONObject) ja.put(obj);
    else if (obj instanceof Object[]) {
      JSONArray jar = new JSONArray();
      for (Object o : ((Object[]) obj)) {
        dataToString(attrs, o, jar);
      }
      ja.put(jar);
    } else if (obj instanceof EnumValue) {
      ja.put(((EnumValue) obj).getValue());
      //sb.append();
    } else if (obj instanceof ListValue) {
      ListValue lv = (ListValue) obj;
      JSONArray jar = new JSONArray();
      for (Object o : lv.getValues()) {
        dataToString(attrs, o, jar);
      }
      ja.put(jar);
    } else {
      String rt = "";
      if (obj != null){
        if(obj instanceof IContextStringProvider)
          rt = ((IContextStringProvider)obj).toString(attrs);
        else
          rt = obj.toString();
      }
      ja.put(rt);
    }
  }

  protected void sendEvents(AttributeStorage as, final Object deleted) {
    //TODO potential rewrite
    UIJob uj = new UIJob("Update jseditor"){
      
      @Override
      public IStatus runInUIThread(IProgressMonitor monitor) {
        if (browser != null && !browser.isDisposed()) {
          synchronized (JSEditor.this) {
            JSONArray jar = new JSONArray();
            try {
              JSEditor.dataToString(as, deleted, jar);
            } catch (JSONException e) {
              e.printStackTrace();
            } 
            execute("onEvents(\"" + escape(jar.toString()) + "\");");
            //  browser.forceFocus();
          }
        }
        return Status.OK_STATUS;
      }
    };
    uj.setRule(or);
    uj.schedule(0);
  }
  protected static String toString(TreeNode node, Object o){
    JSONArray jar = new JSONArray();
    try {
      JSEditor.dataToString(node.attrs(), o, jar);
    } catch (JSONException e) {
      e.printStackTrace();
    }
    return jar.toString();
  }
  
  protected void sendEvents(String val) {
    UIJob uj = new UIJob("Update jseditor"){
    
      @Override
      public IStatus runInUIThread(IProgressMonitor monitor) {
        if (browser != null && !browser.isDisposed()) {
          try{
            execute("onEvents(\"" + escape(val) + "\");");
          }catch(Exception e){
            RLogger.logInfo(CoreUtils.getTrace(e));
          }
            //  browser.forceFocus();
        }
        return Status.OK_STATUS;
      }
    };
    uj.setRule(or);
    uj.schedule(0);
  }
  private static boolean reqIdValid(TreeNode req, TreeNode parent, String text) {
    try {
      return reqIdValidExc(req, parent, text);
    } catch (TreeLogicException exc) {
      return false;
    }
  }
  //Check if child can be added to parent
  private static boolean reqChildValidExc(TreeNode req, TreeNode parent, String name)
      throws SameIdExistsInChildren, IdEmpty, IdContainsWrongChars, WrongProjectException,
          PastingIntoLeafException {
    if (req == null && parent == null) return false;
    TreeLogic tlogic;
    if (req == null) 
      return false;
    tlogic = req.getTreeDB().getTreeLogic();
    return (name == null || name.equals("") || tlogic.isChildIdValidExc(parent, req, name))
        && tlogic.canBeParentOf(parent.getType(), req)
        && tlogic.canBeChildOfExc(req.getType(), parent);
  }

  private static boolean reqChildValid(TreeNode req, TreeNode parent, String name) {
    try {
      return reqChildValidExc(req, parent, name);
    } catch (TreeLogicException exc) {
      return false;
    }
  }

  private TreeNode targetNode;

  @Override
  public void doSave(IProgressMonitor monitor) {}

  @Override
  public void doSaveAs() {}

  private String tabLabel = "";
  /* (non-Javadoc)
   * setter for Editor's prefix in tab label
   */
  protected void setTabLabel(String value) {
    tabLabel = value;
  }
  /* (non-Javadoc)
   * Update a tab name
   */
  protected void updateTabLabel() {
    UIJob job =
        new UIJob("update label") {
          @Override
          public IStatus runInUIThread(IProgressMonitor monitor) {
            String partName;
            if (rootNode instanceof Requirement) {
              partName = ((Requirement) rootNode).getNameOrId();
            } else {
              partName = rootNode.toString();
            }
            partName += " (" + rootNode.getType() + ")";
            if (tabLabel.equals("")) {
              setPartName(partName);
            } else {
              setPartName(tabLabel + " - " + partName);
            }
            return Status.OK_STATUS;
          }
        };
    job.schedule();
  }

  @Override
  public void init(IEditorSite site, IEditorInput input) throws PartInitException {
    changedNodes = new HashMap<>();
    changedNodesCount = new HashMap<>();
    TreeNode sel = null;
    try {
      rootNode = ((NodeEditorInput) input).getStorage().getNode();
      db = rootNode.getTreeDB();
      dbs = new ArrayList<>();
      dbs.add(db);
      for (TreeDB db2 : Instantiator.getVirtualTreeDBs(db)) dbs.add(db2);
      //dbs.addAll(VirtualNode.getInstanceDBMap().keySet());
      updateTabLabel();
      sel = ((NodeEditorInput) input).getSelectionStorage().getNode();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    setSite(site);
    setInput(input);
    targetNode = sel;
    storedSortingSettings = db.getSortingSettings();
    EclipseTracker.getTracker().addListener(treeStateListener);
    getSite().setSelectionProvider(this);
    //Two handlers to deals with command's exceptions
    IHandlerService handlerService = site.getService(IHandlerService.class);
    handlerService.activateHandler("org.eclipse.ui.edit.copy", new EclipseCopyHandler());
    handlerService.activateHandler("org.eclipse.ui.edit.paste", new EclipsePasteHandler());
    //add tree change listener
    for (TreeDB db : dbs) db.addGUIListener(guiTreeChangeListener);
  }

  @Override
  public boolean isDirty() {
    return false;
  }

  @Override
  public boolean isSaveAsAllowed() {
    return false;
  }

  long cnt = 0;

  protected String node2jsonString(TreeNode n) throws JSONException {
    if (!cnter.containsKey("2json")) cnter.put("2json", 0L);
    cnter.put("2json", cnter.get("2json") + 1);
    long t = System.currentTimeMillis();
    String result;
    LinkedList<Attribute> list = new LinkedList<>();
    for (Entry<String, Attribute> en : n.attrs().getLocalAttributes().entrySet()){
      //TODO change for set of attrs
      Attribute a = en.getValue();
      if (a != null){
        if (/*Attribute.affectsDescription(n, k)
            || Attribute.affectsId(n, k)
            || Attribute.affectsStatus(n, k)*/
            //(a.getDefinition().getFlags()&Flag.HIDDEN)==Flag.HIDDEN||
            systemAttrs.contains(en.getKey())||additionalAttrs.contains(en.getKey())) {
          list.add(a);
        }
      }
    }
    JSONObject json = NodeMapper.toJSJSON(n.getUUId(), list,n);
    ;
    json.put("id", n.getId());
    JSONArray jsonChildren = new JSONArray();
    TreeNode[] children = n.getSortedShownChildren();
    for (int i = 0; i < children.length; i++) {
      jsonChildren.put(children[i].getUUId().toString());
    }
    json.put("children", jsonChildren);

    JSONObject attr = json.getJSONObject("attributes");
    String type = n.getType();
    if (type.equals(Requirement.TYPE_NAME)
        || type.equals(TextNode.TYPE_NAME)
        || type.equals(Location.TYPE_NAME)) {
      JSONObject attrObj = new JSONObject();
      attrObj.put("key", "_hidden");
      attrObj.put("type", AttributeType.STRING);
      Object origin = n.getUUId();
      if (origin == null) origin = "";
      attrObj.put("origin", origin);
      attrObj.put("version", Version.getZero());
      attrObj.put("value", n.isHidden());
      attrObj.put("newvalue", n.isHidden());
      attrObj.put("availability", AttributeAvailability.LOCAL);
      attr.put("_lhidden", attrObj);
    }
    @SuppressWarnings("rawtypes")
    Iterator keys = attr.keys();
    while (keys.hasNext()) {
      String key = (String) keys.next();
      JSONObject a = attr.getJSONObject(key);

//      if (n.getQualifiedId().equals(Requirement.getTypeRootQId())) {
//        if (key.equals(TreeNode.ATTR_TYPE)) {
//          //a.putOpt("value", null);
//        }
//      } else if (key.equals(TreeNode.ATTR_DESCRIPTION)) {
//        String descr = n.getDescription(true, false);
//        descr = RequalityCLI.resolveRelativeImagesPath(descr, n);
//        a.putOpt("newvalue", escape(descr));
//        a.putOpt("value", escape(descr));
//      } else if (key.equals("_initcond")) {
//        String descr = n.attrs().getStringAttributeValue("_initcond");
//        if (descr != null) {
//          descr = n.processResourcesPath(descr);
//          a.putOpt("value", escape(descr));
//          a.putOpt("newvalue", escape(descr));
//        }
//      } else if (TestPurpose.TYPE_NAME.equals(n.getType())
//          && key.equals(TestPurpose.ATTR_EXPECTED_RESULTS)) {
//        String desc = ((TestPurpose) n).getExpectedResults(true);
//        if (desc != null) {
//          desc = n.processResourcesPath(desc);
//          a.putOpt("value", escape(desc));
//          a.putOpt("newvalue", escape(desc));
//        }
//      }
    }
    result = json.toString(4);
    cnt += System.currentTimeMillis() - t;
    if (!cnter.containsKey("2jsonT")) cnter.put("2jsonT", 0L);
    cnter.put("2jsonT", cnter.get("2jsonT") + cnt);

    return result;
  }

  protected String nodeTree2jsonString(TreeNode nr) throws JSONException {
    DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(nr);
    Iterator<TreeNode> it = dftw.iterator();
    JSONArray ar = new JSONArray();
    while(it.hasNext()){
      TreeNode n = it.next();
      String result;
      LinkedList<Attribute> list = new LinkedList<>();
      for (Entry<String, Attribute> en : n.attrs().getLocalAttributes().entrySet()){
        //TODO change for set of attrs
        Attribute a = en.getValue();
        if (a != null){
          if (/*Attribute.affectsDescription(n, k)
              || Attribute.affectsId(n, k)
              || Attribute.affectsStatus(n, k)*/
              //(a.getDefinition().getFlags()&Flag.HIDDEN)==Flag.HIDDEN||
              systemAttrs.contains(en.getKey())||additionalAttrs.contains(en.getKey())) {
            list.add(a);
          }
        }
      }
      JSONObject json = NodeMapper.toJSJSON(n.getUUId(), list,n);
      ;
      json.put("id", n.getId());
      JSONArray jsonChildren = new JSONArray();
      TreeNode[] children = n.getSortedShownChildren();
      for (int i = 0; i < children.length; i++) {
        jsonChildren.put(children[i].getUUId().toString());
      }
      json.put("children", jsonChildren);
  
      JSONObject attr = json.getJSONObject("attributes");
      String type = n.getType();
      if (type.equals(Requirement.TYPE_NAME)
          || type.equals(TextNode.TYPE_NAME)
          || type.equals(Location.TYPE_NAME)) {
        JSONObject attrObj = new JSONObject();
        attrObj.put("key", "_hidden");
        attrObj.put("type", AttributeType.STRING);
        Object origin = n.getUUId();
        if (origin == null) origin = "";
        attrObj.put("origin", origin);
        attrObj.put("version", Version.getZero());
        attrObj.put("value", n.isHidden());
        attrObj.put("newvalue", n.isHidden());
        attrObj.put("availability", AttributeAvailability.LOCAL);
        attr.put("_lhidden", attrObj);
      }
      @SuppressWarnings("rawtypes")
      Iterator keys = attr.keys();
      while (keys.hasNext()) {
        String key = (String) keys.next();
        JSONObject a = attr.getJSONObject(key);
      }
      ar.put(json);
      //result = json.toString(4);
      //cnt += System.currentTimeMillis() - t;
    }
    return ar.toString(4);
  }

  private TreeDB getValidDB(String string) {
    for (TreeDB db : dbs)
      try {
        TreeNode node = db.getNodeExc(UUID.fromString(string));
        if(node!=null){
          if (node.isVirtual()) {
            UUID uid = node.getVRoot();
            if (uid != null) return Instantiator.getExistVirtualTreeDB(db, uid);
            return Instantiator.getRootTree(
                db); //Instantiator.getVirtualTreeDB(db,Instantiator.DEFAULT_INSTANCE_NAME);
          }
          return db;
        }
      } catch (Exception e) {
        RLogger.logError("JSEditor tree: "+CoreUtils.getTrace(e));
      }
    return null;
  }

  private IPartListener plistener =
      new IPartListener() {
        @Override
        public void partOpened(IWorkbenchPart part) {}

        @Override
        public void partDeactivated(IWorkbenchPart part) {}

        @Override
        public void partClosed(IWorkbenchPart part) {}

        @Override
        public void partBroughtToTop(IWorkbenchPart part) {}

        @Override
        public void partActivated(IWorkbenchPart part) {
          if (part != JSEditor.this)
            JSEditor.this.execute("saveEditor();");
        }
      };
  private boolean pageLoaded;
  private boolean fbloaded;
  private Map<String, Long> cnter = new HashMap<>();
  
  @Override
  public void createPartControl(Composite parent) {
    browser = RequalityPlugin.newBrowser(parent, RequalityPlugin.getBrowserType());
    browser.addMouseListener(
        new MouseListener() {
          @Override
          public void mouseUp(MouseEvent e) {}

          @Override
          public void mouseDown(MouseEvent e) {
            //JSEditor.this.setFocus();
          }

          @Override
          public void mouseDoubleClick(MouseEvent e) {}
        });
    browser.addKeyListener(new BrowserKeyListener());
    pageLoaded = false;
    fbloaded = false;
    ;
    browser.addProgressListener(
        new ProgressListener() {

          @Override
          public void completed(ProgressEvent event) {
            if (!pageLoaded) {
              setSelectedNode(currentSelection);
              pageLoaded = true;
              if (!fbloaded && CoreUtils.DEBUG_LEVEL > 0) {
                fbloaded = true;
              }
            }
          }

          @Override
          public void changed(ProgressEvent event) {}
        });
    getSite().getPage().addPartListener(plistener);

    try {
      browser.addLocationListener(new LocationListener(){
      
        @Override
        public void changing(LocationEvent event) {
          String loc = event.location;
          if(loc.startsWith("requality://")){
            event.doit = false;
            String path = loc.substring(12);
            TreeNode node = rootNode.getTreeDB().getNode(path);
            if(node!=null){
              Map<String,List<TreeNode>> sel = new HashMap<>();
              sel.put(node.getType(), new LinkedList<>());
              sel.get(node.getType()).add(node);
              SelectionManager.getOne().setActiveNodes(sel,sel,EventSource.GENERAL);
              setSelectedNode(node);
            }
          }
        }
      
        @Override
        public void changed(LocationEvent event) {
          String loc = event.location;
          if(loc.startsWith("requality://"))
            event.doit = false;
        }
      });
      browser.setUrl(html.toURI().toURL().toString());
      new BrowserFunction(browser, "isNodeHeader") {
        @Override
        public Object function(Object[] arguments) {
          TreeDB db = getValidDB((String) arguments[0]);
          if(db==null)
            return false;
          TreeNode n1 = db.getNode(UUID.fromString((String) arguments[0]));
          if(n1!=null)
            return CoreUtils.isHeader(n1);
          return false;
        }
      };
      new BrowserFunction(browser, "getChildrenOrder") {
        @Override
        public Object function(Object[] arguments) {
          try{
            TreeDB db = getValidDB((String) arguments[0]);
            TreeNode n1 = db.getNode(UUID.fromString((String) arguments[0]));
            TreeNode[] chi = n1.getSortedShownChildren();
            Object[] retchi = new Object[chi.length];
            for (int i = 0; i < chi.length; i++) retchi[i] = chi[i].getUUId().toString();
            return retchi;
          }
          catch(Exception e){
            RLogger.logError("JSEditor: "+CoreUtils.getTrace(e));
          }
          return new Object[0];
          //            		Object[] uuids = (Object[]) arguments[0];
          //            		if(uuids.length==0)
          //            			return arguments[0];
          //            		TreeDB db = getValidDB((String)uuids[0]);
          //            		if(db == null)
          //            			db = rootNode.getTreeDB();
          //            		List<TreeNode> nodes = new LinkedList<TreeNode>();
          //            		for(UUID uid:uuids)
          //            			nodes.add(db.getNode(uid));
          //            		if(arguments[0]!=null){
          //            			UUID nodeUUID1 = UUID.fromString((String) arguments[0]);
          //            			UUID nodeUUID2 = UUID.fromString((String) arguments[1]);
          //            			TreeNode n1=db.getNode(nodeUUID1),n2=db.getNode(nodeUUID2);
          //            			return TreeNodeComparator.getComparator(db).compare(n1, n2);
          //            		}
          //            		return 0;
        }
      };
      new BrowserFunction(browser, "debug_level") {
        @Override
        public Object function(Object[] arguments) {
          return CoreUtils.DEBUG_LEVEL;
        }
      };
      
      new BrowserFunction(browser, "getExtraAttrsNames") {
        @Override
        public Object function(Object[] arguments) {
          return additionalAttrs.toArray(new String[0]);
        }
      };
      new BrowserFunction(browser, "callOpenTemplates") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), (long) 1);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          if (db == null) db = rootNode.getTreeDB();
          String type = Requirement.TYPE_NAME;
          if (arguments[0] != null) {
            UUID nodeUUID = UUID.fromString((String) arguments[0]);
            type = db.getNodeType(nodeUUID);
          }
          TemplatesSelectionDialog tsd = new TemplatesSelectionDialog(browser.getShell(), db, type);
          return tsd.open() == Window.OK;
        }
      };
      new BrowserFunction(browser, "getNodeTypeByUUID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), (long) 1);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          if (db == null) db = rootNode.getTreeDB();
          String type = Requirement.TYPE_NAME;
          if (arguments[0] != null) {
            UUID nodeUUID = UUID.fromString((String) arguments[0]);
            type = db.getNodeType(nodeUUID);
          }
          return type.toString();
        }
      };
      new BrowserFunction(browser, "printToConsole") {
        @Override
        public Object function(Object[] arguments) {
          if (CoreUtils.DEBUG_LEVEL > 0) {
            RLogger.logInfo(System.currentTimeMillis() + " :" + arguments[0].toString());
          }
          return null;
        }
      };

      new BrowserFunction(browser, "selectFileDialog") {
        @Override
        public Object function(Object[] arguments) {
          if (CoreUtils.DEBUG_LEVEL > 0) {
            RLogger.logInfo(System.currentTimeMillis() + " :" + arguments[0].toString());
          }
          String exts = arguments[0].toString();
          FileDialog dialog = new FileDialog(browser.getShell(), SWT.OPEN);
          dialog.setFilterExtensions(exts.split("|"));
          //dialog.setFilterPath(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getWorkbench().getP);
          String ret = dialog.open();
          if (ret == null) return ret;
          try {
            return Activator.getUrl(ret).toString();
          } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
          return null;
        }
      };

      new BrowserFunction(browser, "getChildNumber") {
        @Override
        public Object function(Object[] arguments) {
          TreeDB db = getValidDB((String) arguments[0]);
          UUID nodeUUID = UUID.fromString((String) arguments[0]);
          TreeNode node = db.getNode(nodeUUID);
          TreeNode[]nds = node.getShownParent().getSortedShownChildren();
          int j = 0;
          for(int i =0;i<nds.length;i++){
            TreeNode chi = nds[i];
            if(CoreUtils.isHeader(chi))
              j++;
            if(chi==node)
              return j;
          }
          return 1;
        }
      };

      new BrowserFunction(browser, "listImages") {
        @Override
        public Object function(Object[] arguments) {
          TreeDB db = getValidDB((String) arguments[0]);
          UUID nodeUUID = UUID.fromString((String) arguments[0]);
          TreeNode node = db.getNode(nodeUUID);
          String[] ress = node.listResources().toArray(new String[0]);
          return ress;
        }
      };

      new BrowserFunction(browser, "loadNodeByUUID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          UUID nodeUUID = UUID.fromString((String) arguments[0]);
          String result = "";
          if (nodeUUID == null || db == null) return null;
          try {
            TreeNode n = db.getNode(nodeUUID);
            try {
              result = node2jsonString(n);
            } catch (JSONException e) {
              e.printStackTrace();
              return null;
            }
          } catch (Exception e) {
            e.printStackTrace();
          }
          return result;
        }
      };

      new BrowserFunction(browser, "loadNodeTreeByUUID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          UUID nodeUUID = UUID.fromString((String) arguments[0]);
          String result = "";
          if (nodeUUID == null || db == null) return null;
          try {
            TreeNode n = db.getNode(nodeUUID);
            try {
              result = nodeTree2jsonString(n);
            } catch (JSONException e) {
              e.printStackTrace();
              return null;
            }
          } catch (Exception e) {
            e.printStackTrace();
          }
          return result;
        }
      };
      new BrowserFunction(browser, "getNodeUUIDByQID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);
          try{
            TreeNode n =
              db.getNode(
                  ((String) arguments[0]).replace("&nbsp;", " ").replace((char) 160, ' '),
                  Path.QID | Path.UUID);
            if (n == null) return null;
            return n.getUUId().toString();
          }catch(Exception e){
            RLogger.logError(CoreUtils.getTrace(e));
          }
          return null;
        }
      };
      /*
       * Request for the root element of the selected subtree Input
       * parameters: none Return value: String value containing uuid of
       * the subtree root node
       */
      new BrowserFunction(browser, "getStartNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          return rootNode.getUUId().toString();
        }
      };
      new BrowserFunction(browser, "countChildrenOfType") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);
          try {
            TreeDB db = getValidDB((String) arguments[0]);
            UUID nodeUUID = UUID.fromString((String) arguments[0]);
            if(db == null)
              return null;
            TreeNode nde = db.getNode(nodeUUID);
            if (arguments.length == 1) return Integer.toString(nde.getShownChildren().length);
            int cntl = nde.getShownChildren(arguments[1].toString(), true).length;
            //for(TreeNode node2 :vnds)
            //cntl+=node2.getUnsortedChildren(arguments[1].toString(),true).length;
            return Integer.toString(cntl);
          } catch (Exception e) {
            e.printStackTrace();
          }
          return null;
        }
      };
      new BrowserFunction(browser, "getSelectedNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          if (currentSelection != null) return currentSelection.getUUId().toString();
          return rootNode.getUUId().toString();
        }
      };

      /*
       * Request to change the node ID Input parameters: uuid - (String)
       * uuid of the updated node new_id - (String) new ID to be set
       * Return value: Boolean value: true if operation was successful,
       * false otherwise
       */
      new BrowserFunction(browser, "changeNodeId") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
          String newId = (String) arguments[1];
          if (newId.equals("")) {
            sendEvents(
                n.attrs(),
                new Object[] {
                  "changeAttributes",
                  arguments[0],
                  new Object[] {new Object[] {"_nameOrId", n.getNameOrId(), n.getNameOrId()}}
                });
          } else {
            try {
              newId = newId.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
              if (db.getTreeLogic().isIdValidExc(n, newId)) {
                db.startTransaction("Update " + n.getTypeLabel() + " Id");
                n.setId(newId);
                n.commitAttributes();
                db.commit();
              } else {
                MessageDialog.openError(
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                    "Update " + n.getTypeLabel() + " Id",
                    "Id of element cannot be changed to " + newId);
                return false;
              }
            } catch (IllegalStateException ise) {
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Update " + n.getTypeLabel() + " Id",
                  "Some actions are already performing.");
              return false;
            } catch (SameIdExistsInChildren e) {
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Update " + n.getTypeLabel() + " Id",
                  "Element with same id exists in " + n.getWorkParent().getUserVisibleName(false));
              return false;
            } catch (IdEmpty e) {
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Update " + n.getTypeLabel() + " Id",
                  "Element id cannot be empty");
              return false;
            } catch (IdContainsWrongChars e) {
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Update " + n.getTypeLabel() + " Id",
                  "Element id cannot contains wrong character - " + e.getWrongCharacter());
              return false;
            }
          }
          return true;
        }
      };
      new BrowserFunction(browser, "callEditNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode par = db.getNode(UUID.fromString((String) arguments[0]));
          String title = "Edit '" + par.getUserVisibleName(false) + "'";

          if (!db.startTransaction("Update '" + par.getUserVisibleName(false) + "' element"))
            return false;
          EditElementDialog edd =
              new EditElementDialog(this.getBrowser().getShell(), par, title, true);
          edd.setTransactionIgnore(true);
          if (edd.open() == Dialog.CANCEL) {
            db.rollback();
            return false;
          }
          db.commit();
          return true;
        };
      };
      new BrowserFunction(browser, "getActiveTemplateLabel") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final String tp = (String) arguments[1];
          TreeNode node = db.getActiveTemplate(tp);
          if (node == null) return null;
          return node.getId();
        };
      };
      new BrowserFunction(browser, "createNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode par = db.getNode(UUID.fromString((String) arguments[0]));
          String type = arguments[1].toString();
          String id = arguments[2].toString();
          String title = "Edit " + CoreUtils.splitWords(type);
          String addid = null;
          Map<String, Object> attrs = new HashMap<>();
          if (arguments.length > 3) {
            title = arguments[3].toString();
            if (arguments.length > 4) addid = arguments[4].toString();
            if (arguments.length > 5) {
              Object[] attributes = (Object[]) arguments[5];
              for (Object o : attributes) {
                Object[] pair = (Object[]) o;
                attrs.put(pair[0].toString(), pair[1]);
              }
            }
          }
          TreeNode nd = par.getTreeDB().getActiveTemplate(type);
          title = title.replace("[template]", nd == null ? "" : (" - " + nd.getId()));
          if (id == null || id.length() == 0) {
            if (db.getTreeLogic().isAutoRenamable(type)) {
              id = RequalityCLI.getNextId(par, type, type + " ");
            } else {
              if (type.equals(TestPurpose.TYPE_NAME)) {
                String parid = par.getId();
                if (par instanceof Requirement) parid = ((Requirement) par).getNameOrId();
                id = RequalityCLI.getNextId(par, type, parid + "_T");
              } else {
                id = RequalityCLI.getNextId(par, type, null);
              }
            }
          }
          id = id.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
          TreeLogic tl = db.getTreeLogic();
          try {
            if (!tl.canBeCreatedInExc(type, id, par)) {
              javaDialogOpened = true;
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Add Element",
                  "Element of type '"
                      + type
                      + "' cannot be created in "
                      + par.getUserVisibleName(true));
              javaDialogOpened = false;
              return null;
            }
          } catch (PastingIntoLeafException e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Element",
                "Can't add "
                    + type
                    + " to element with "
                    + (type.equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                    + "TestPurpose child");
            javaDialogOpened = false;
            return null;
          } catch (SameIdExistsInChildren e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element with same id exists in " + par.getUserVisibleName(true));
            javaDialogOpened = false;
            return null;
          } catch (IdEmpty e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element id is empty");
            javaDialogOpened = false;
            return null;
          } catch (IdContainsWrongChars e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element id contains wrong character - " + e.getWrongCharacter());
            javaDialogOpened = false;
            return null;
          }
          if (type == null) type = par.getType();
          if (db.startTransaction("Create " + CoreUtils.splitWords(type))) {
            try {
              TreeNode node = par.createChild(id, type);
              boolean save = false;
              if (db.getTreeLogic().isAutoRenamable(type)) {
                Index ind = Index.createOne();
                if (addid == null) {
                  ind = (Index.getNextIndex(type, par, null, false));
                } else {
                  ind = (Index.getNextIndex(type, par, db.getNode(UUID.fromString(addid)), false));
                }
                node.setIndex(ind);
                save = true;
              }
              if (type.equals(TestPurpose.TYPE_NAME)) {
//                node.attrs()
//                    .setAttribute(TestPurpose.ATTR_AUTHOR, System.getProperty("user.name", ""));
                node.attrs().setConstAttribute(TestPurpose.ATTR_AUTHOR, System.getProperty("user.name", ""), 
                    Flag.HIDDEN);
                node.attrs().setConstAttribute(TestPurpose.ATTR_STATUS, "in process", 
                    Flag.HIDDEN);
                save = true;
              }
              //	for comments we adds default author
              if (type.equals(Comment.TYPE_NAME)) {
                node.attrs().setConstAttribute(Comment.ATTR_AUTHOR, System.getProperty("user.name", ""), 
                    Flag.HIDDEN);
                //node.setIndex(ind);
                save = true;
              }
              if (!attrs.isEmpty()) {
                for (String s : attrs.keySet())
                  node.attrs().setAttribute(new AttributeDefinition(node.getUUId(), s, 
                      AttributeDefinition.detectValueProvider(s, AttributeType.STRING, 
                          AttributeDefinition.detectFlags(s), attrs.get(s))));
                save = true;
              }
              if (save) node.commitAttributes();
              EditElementDialog edd =
                  new EditElementDialog(this.getBrowser().getShell(), node, title, true);
              edd.setTransactionIgnore(true);
              javaDialogOpened = true;
              if (edd.open() == Dialog.CANCEL) {
                javaDialogOpened = false;
                db.rollback();
                return null;
              }
              javaDialogOpened = false;
              db.commit();
              return node.getUUId().toString();
            } catch (Exception e) {
              e.printStackTrace();
              db.rollback();
            }
          }
          return null;
        }
      };
      new BrowserFunction(browser, "addTest") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
          if (node.getType().equals(Requirement.TYPE_NAME)
              || node.getType().equals(TestPurpose.TYPE_NAME))
            TestDialogProvider.fireTestCreation(
                node.getTreeDB(), node.getQualifiedId(), browser.getShell());
          return null;
        }
      };
      /*
       * Wrapping method to call Java dialog for id altering
       * First argument is UUID of parent to support
       * id validation.
       * Next three arguments are title of shown dialog, message
       * and default value for dialog.
       * Return value: String value: null if dialog was Canceled,
       * valid id otherwise
       */
      new BrowserFunction(browser, "openIdChange") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode nde = db.getNode(UUID.fromString((String) arguments[0]));
          String title = (String) arguments[1],
              message = (String) arguments[2],
              lastValue = (String) arguments[3];
          boolean idcheck = true;
          if (arguments.length > 4) idcheck = (Boolean) arguments[4];
          final boolean checkId = idcheck;
          javaDialogOpened = true;

          InputDialog id =
              new InputDialog(
                  this.getBrowser().getShell(),
                  title,
                  message,
                  lastValue,
                  new IInputValidator() {
                    @Override
                    public String isValid(String newText) {
                      if (newText.length() == 0) return "id cannot be empty";
                      if (checkId) {
                        if (!nde.getTreeDB().getTreeLogic().isChildIdValid(nde, newText))
                          return newText
                              + " is not valid id for the child of "
                              + nde.getUserVisibleName(true)
                              + "("
                              + nde.getType()
                              + ")";
                      }
                      return null;
                    }
                  });
          id.open();
          browser.forceFocus();
          javaDialogOpened = false;
          if (id.getReturnCode() == InputDialog.CANCEL) return null;
          return id.getValue();
        }
      };
      new BrowserFunction(browser, "getNextAvalibleId") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode nde = db.getNode(UUID.fromString((String) arguments[0]));
          return RequalityCLI.getNextId(nde, nde.getType(), null);
        }
      };
      new BrowserFunction(browser, "changeNodeName") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
          if (!n.getType().equals(Requirement.TYPE_NAME)) {
            return false;
          }
          String newName = (String) arguments[1];
          if (newName != null)
            newName = newName.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
          if (db.startTransaction("Change " + n.getTypeLabel() + " Name", true)) {
            ((Requirement) n).setName(newName);
            n.commitAttributes();
            db.commit();
            return true;
          }
          return false;
        }
      };

      new BrowserFunction(browser, "getUFName") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          try {
            TreeDB db = getValidDB((String) arguments[0]);
            TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
            return n.getUserVisibleName(false).replace("  ", "&nbsp;&nbsp;");
          } catch (RuntimeException re) {
            return null;
          }
        }
      };
      /*
       * Returns true if it is possible to edit this node
       */
      new BrowserFunction(browser, "canDeleteNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
          return db.getTreeLogic().canBeDeleted(n);
        }
      };
      /*
       * Check the possibility of adding some node as
       *  child of another. Also can be used to check
       *  posibily of adding a child with some name
       *
       */
      new BrowserFunction(browser, "checkName") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode n = null;
          String newName = (String) arguments[1];
          if (newName != "")
            newName = newName.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
          if (!((String) arguments[0]).equals("")) {
            n = db.getNode(UUID.fromString((String) arguments[0]));
            if (n.getType().equals(Requirement.TYPE_NAME)) {
              Attribute a = n.attrs().getAttribute(TreeNode.ATTR_NAME);
              if (a != null && a.getValue() != null && a.getValue() != "")
                return true;
            }
            if (newName.equals("")) newName = n.getId();
          }
          TreeNode parent = null;
          if (n != null) parent = n.getWorkParent();
          if (arguments.length > 2) {
            parent = db.getNode(UUID.fromString((String) arguments[2]));
          }
          //TODO maybe change to string error message
          if ((n == null) && (parent == null) || newName.equals("")) return false;
          if (n != null && db.getTreeLogic().isAutoRenamable(n.getType())) return true;
          return reqIdValid(n, parent, newName);
        }
      };
      /*
       * Check if it possible to add some node
       * as child of another one
       *
       */
      new BrowserFunction(browser, "checkChild") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode n = null;
          String newName =
              ((String) arguments[1]).replace("&nbsp;", " ").replace((char) 160, ' ').trim();
          if (!((String) arguments[0]).equals("")) {
            n = db.getNode(UUID.fromString((String) arguments[0]));
            //if(newName.equals(""))
            //newName = n.getId();
          }
          TreeNode parent = null;
          if (arguments.length > 2) {
            parent = db.getNode(UUID.fromString((String) arguments[2]));
          }
          if ((n == null) && (parent == null)) return false;
          if (n != null && db.getTreeLogic().isAutoRenamable(n.getType())) return true;
          return reqChildValid(n, parent, newName);
        }
      };

      /*
       * Request to change the value of one of the node attributes Input
       * parameters: uuid - (String) uuid of the updated node attr_name -
       * (String) name of the changed attribute new_val - (String) new
       * attribute value to be set Return value: Boolean value: true if
       * operation was successful, false otherwise
       */
      new BrowserFunction(browser, "changeNodeAttr") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          try {
            TreeDB db = getValidDB((String) arguments[0]);
            UUID uuid = UUID.fromString((String) arguments[0]);
            TreeNode n = db.getNode(uuid);
            String key = (String) arguments[1];
            String value = (String) arguments[2];
            synchronized (JSEditor.this) {
              if (key.equals(TreeNode.ATTR_NAME)) {
                if (n.getName() != null) {
                  if (value != null) value = value.replace((char) 160, ' ').trim();
                  if (value != null && value.length() > 0 && value.charAt(0) == ((char) 255)) {

                  } else {
                    if (db.startTransaction("Change Requirement Name", true)) {
                      n.setName(value);
                      n.commitAttributes();
                      db.commit();
                      return value;
                    }
                  }
                } else if (n.getName() == null) {
                  if (value != null){//TODO null val?
                    value = value.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
                    if (value.length()>0&&value.charAt(0) == ((char) 255)) {
                      sendEvents(
                          n.attrs(),
                          new Object[] {
                            "changeAttributes",
                            arguments[0],
                            new Object[] {
                              new Object[] {"_nameOrId", n.getNameOrId(), n.getNameOrId()}
                            }
                          });
                    } else {
                      try {
                        if (db.getTreeLogic().isIdValidExc(n, value)) {
                          if (db.startTransaction("Update " + n.getTypeLabel() + " Id", true)) {
                            n.setId(value);
                            n.commitAttributes();
                            db.commit();
                            //Activator.getDefault().setActiveNode(n, n.getType());
                          }
                        } else {
                          MessageDialog.openError(
                              PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                              "Update " + n.getTypeLabel() + " Id",
                              "Id of element cannot be changed to " + value);
                          return false;
                        }
                      } catch (SameIdExistsInChildren e) {
                        MessageDialog.openError(
                            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                            "Update " + n.getTypeLabel() + " Id",
                            "Element with same id exists in "
                                + n.getWorkParent().getUserVisibleName(false));
                        return false;
                      } catch (IdEmpty e) {
                        MessageDialog.openError(
                            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                            "Update " + n.getTypeLabel() + " Id",
                            "Element id cannot be empty");
                        return false;
                      } catch (IdContainsWrongChars e) {
                        MessageDialog.openError(
                            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                            "Update " + n.getTypeLabel() + " Id",
                            "Element id cannot contains wrong character - " + e.getWrongCharacter());
                        return false;
                      }
                    }
                  }
                }
              } else {
                if (db.startTransaction("Update " + n.getTypeLabel(), true)) {
                  if (key.equals(TreeNode.ATTR_DESCRIPTION)
                      || key.equals(TestPurpose.ATTR_EXPECTED_RESULTS)
                      || key.equals("_initcond")) {
                    //RequalityCLI.deleteOldImages(n.attrs().getAttribute(key), value, n);
                  }
                  Attribute attr = null;
                  if (!value.equals("")) {
                    n.attrs().setAttribute(new AttributeDefinition(n.getUUId(), key, 
                        AttributeDefinition.detectValueProvider(key, AttributeType.STRING, value),
                        AttributeDefinition.detectFlags(key)));
                    attr = n.attrs().getAttribute(key);
                  } else {
                    n.attrs().removeAttribute(key);
                  }
                  if (!changedNodes.containsKey(uuid)) {
                    changedNodes.put(uuid, n.toNodeDesc());
                    changedNodesCount.put(uuid, 1);
                  } else {
                    NodeDesc nodeDesc = changedNodes.get(uuid);
                    if (attr != null) {
                      nodeDesc.setAttribute(attr.getDefinition());
                    } else {
                      nodeDesc.removeAttribute(key);
                    }
                    changedNodesCount.put(uuid, changedNodesCount.get(uuid) + 1);
                  }
                  n.commitAttributes();
                  db.commit();
                  return value;
                }
              }
            }
          } catch (Throwable e) {
            e.printStackTrace();
          }
          return true;
        }
      };
      new BrowserFunction(browser, "getChildrenUUIDs") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
          String[] chi = new String[n.getShownChildrenUUIDs().size()];
          int i = 0;
          for (UUID uid : n.getShownChildrenUUIDs()) chi[i++] = uid.toString();
          return chi;
        }
      };
      /*
       * Request to change the move a node into different subtree Input
       * parameters: uuid - (String) uuid of the moved node
       * new_parent_uuid - (String) uuid of the new parent node Return
       * value: Boolean value: true if operation was successful, false
       * otherwise
       */
      new BrowserFunction(browser, "changeNodeParent") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[1]));
          boolean altPressed = (Boolean) arguments[2];
          Index i = null;
          TreeNode lastNode = null;
          TreeLogic l = parent.getTreeDB().getTreeLogic();
          if (!l.canBeChildOf(n.getType(), parent)
                  && l.canBeChildOf(n.getType(), parent.getShownParent())
              || altPressed) {
            lastNode = parent;
            parent = parent.getShownParent();
          }
          String torename = null;
          //if we deal with test step
          //then special actions performed
          if (n.attrs().getAttribute("_index") != null)
            i = Index.getNextIndex(n.getType(), parent, lastNode);
          try {
            //if(!l.canBeChildOf(n.getType(), parent))
            //  parent = parent.getParent();
            //TODO mo
            if (!n.getShownParent().equals(parent) || (i != null)) {
              if (!l.canBeCopiedToExc(n, parent, true)) {
                javaDialogOpened = true;
                MessageDialog.openError(
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                    "Change Element parent",
                    "Element "
                        + n.getUserVisibleName(true)
                        + " parent cannot be changed to "
                        + parent.getUserVisibleName(true));
                javaDialogOpened = false;
                return false;
              }
              if (parent == null) return false;
              if (l.isAutoRenamable(n.getType())
                  || parent.findWorkChild(n.getId()) != null
                      && !parent.findWorkChild(n.getId()).getUUId().equals(n.getUUId())) {
                String addt = "";
                if (n.getType().equals(Comment.TYPE_NAME)) addt = Comment.TYPE_NAME + " ";
                if (db.getTreeLogic().isAutoRenamable(n.getType())) {
                  torename = RequalityCLI.getNextId(parent, n.getType(), addt + "");
                } else torename = RequalityCLI.getNextId(parent, n.getType(), null);
              }
              if (torename != null && !torename.equals("") && !l.isAutoRenamable(n.getType()))
                if (!MessageDialog.openConfirm(
                    Activator.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell(),
                    "Move Element",
                    "Moved element has id conflict:\n"
                        + "\n- Element with id "
                        + n.getId()
                        + " already exists in "
                        + parent.getUserVisibleName(true)
                        + "."
                        + "\n"
                        + "\nConflicting element will be renamed. If you still want to proceed click OK, otherwise click Cancel.")) {
                  return false;
                }
              if (db.startTransaction("Move Element", true)) {
                try{
                  if (lastNode != parent) {
                    TreeNode[] sorted = parent.getSortedShownChildren();
                    for (int j = 0; j < sorted.length; j++) {
                      Index lastInd;
                      if (j > 0) lastInd = Index.create(Integer.toString(j + 1));
                      else lastInd = Index.createOne();
                      sorted[j].setIndex(lastInd);
                      sorted[j].commitAttributes();
                    }
                    i = Index.getNextIndex(n.getType(), parent, lastNode);
                  } else if (n.attrs().getAttribute("_index") != null)
                    i = Index.getNextIndex(n.getType(), parent, lastNode);
                  if (i != null) {
                    n.setIndex(i);
                    n.commitAttributes();
                  }
                  if (!n.getWorkParent().equals(parent) && l.canBeChildOf(n.getType(), parent)) {
                    if (l.isAutoRenamable(n.getType()) || parent.findWorkChild(n.getId()) != null) {
                      n.setId(
                          RequalityCLI.getNextId(
                              Arrays.asList(n.getWorkChildren()), parent, n.getType(), null));
                      n.commitAttributes();
                    }
                    n.setParent(parent);
                    if (torename != null){
                      n.setId(torename);
                      n.commitAttributes();
                    }
                  }
                  db.commit();
                  setSelectedNode(n);
  
                  //	refresh(parent);
                  return true;
                }
                catch(Exception e){
                  if (db.inTransaction()) db.rollback();
                  throw e;
                }
              }
              return false;
            }
          } catch (PastingIntoLeafException e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Change Element parent",
                "Can't change parent of "
                    + n.getType()
                    + " to "
                    + parent.getType()
                    + " with "
                    + (n.getType().equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                    + "TestPurpose child");
            javaDialogOpened = false;
            return false;
          } catch (WrongProjectException e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Change Element parent",
                "Moving elements between projects is prohibited");
            javaDialogOpened = false;
            return false;
          } catch (SameIdExistsInChildren e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Change Element parent",
                "Element with same id exists in " + parent.getUserVisibleName(true));
            javaDialogOpened = false;
            return false;
          } catch (IdEmpty e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Change Element parent",
                "Element id is empty");
            javaDialogOpened = false;
            return false;
          } catch (IdContainsWrongChars e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Change Element parent",
                "Element id contains wrong character - " + e.getWrongCharacter());
            javaDialogOpened = false;
            return false;
          }
          return false;
        }
      };

      /*
       * Request to insert a new node Input parameters: uuid - (String)
       * uuid of the new node parent_uuid - (String) uuid of the parent
       * node Return value: Boolean value: true if operation was
       * successful, false otherwise
       */
      new BrowserFunction(browser, "addNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[0]));
          String id = (String) arguments[1];
          String type = (String) arguments[2];
          TreeLogic l = db.getTreeLogic();
          if (id == null || id.length() == 0) {
            if (db.getTreeLogic().isAutoRenamable(type)) {
              id = RequalityCLI.getNextId(parent, type, type + " ");
            } else {
              if (type.equals(TestPurpose.TYPE_NAME)) {
                String parid = parent.getId();
                if (parent instanceof Requirement) parid = ((Requirement) parent).getNameOrId();
                id = RequalityCLI.getNextId(parent, type, parid + "_T");
              } else {
                id = RequalityCLI.getNextId(parent, type, null);
              }
            }
          }
          id = id.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
          try {
            if (!l.canBeCreatedInExc(type, id, parent)) {
              javaDialogOpened = true;
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Add Element",
                  "Element of type '"
                      + type
                      + "' cannot be created in "
                      + parent.getUserVisibleName(true));
              javaDialogOpened = false;
              return null;
            }
          } catch (PastingIntoLeafException e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Element",
                "Can't add "
                    + type
                    + " to element with "
                    + (type.equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                    + "TestPurpose child");
            javaDialogOpened = false;
            return null;
          } catch (SameIdExistsInChildren e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element with same id exists in " + parent.getUserVisibleName(true));
            javaDialogOpened = false;
            return null;
          } catch (IdEmpty e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element id is empty");
            javaDialogOpened = false;
            return null;
          } catch (IdContainsWrongChars e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Node",
                "Element id contains wrong character - " + e.getWrongCharacter());
            javaDialogOpened = false;
            return null;
          }
          if (db.startTransaction("Add Element", true)) {
            Index ind = null;
            if (db.getTreeLogic().isAutoRenamable(type)) {
              if (arguments.length < 4
                  || arguments[3] == null
                  || ((String) arguments[3]).length() == 0) {
                ind = (Index.getNextIndex(type, parent, null, false));
              } else {
                ind =
                    (Index.getNextIndex(
                        type, parent, db.getNode(UUID.fromString((String) arguments[3])), false));
              }
            }
            TreeNode newNode = parent.createChild(id, type);
            //	TODO: remake to commands
            if (type.equals(TestPurpose.TYPE_NAME)) {
              newNode
                  .attrs()
                  .setAttribute(new AttributeDefinition(newNode.getUUId(), 
                      TestPurpose.ATTR_AUTHOR, new ConstValueProvider(AttributeType.STRING, System.getProperty("user.name", "")),
                      Flag.HIDDEN));
              newNode
              .attrs()
              .setAttribute(new AttributeDefinition(newNode.getUUId(), 
                  TestPurpose.ATTR_STATUS, new ConstValueProvider(AttributeType.STRING, "in process"),
                  Flag.HIDDEN));
              newNode.commitAttributes();
            }
            //	for comments we adds default author
            if (type.equals(Comment.TYPE_NAME)) {

              newNode
                  .attrs()
                  .setAttribute(new AttributeDefinition(newNode.getUUId(), 
                      Comment.ATTR_AUTHOR, new ConstValueProvider(AttributeType.STRING, System.getProperty("user.name", "")),
                      Flag.HIDDEN));
              newNode.setIndex(ind);
              newNode.commitAttributes();
            }
            db.commit();
            return newNode.getUUId().toString();
          }
          return null;
        }
      };
      new BrowserFunction(browser, "canAddNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          if (db.inTransaction()) return false;
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[0]));
          String id = (String) arguments[1];
          String type = (String) arguments[2];
          TreeLogic l = db.getTreeLogic();
          if (id == null || id.length() == 0) {
            if (db.getTreeLogic().isAutoRenamable(type)) {
              id = RequalityCLI.getNextId(parent, type, type + " ");
            } else {
              if (type.equals(TestPurpose.TYPE_NAME)) {
                String parid = parent.getId();
                if (parent instanceof Requirement) parid = ((Requirement) parent).getNameOrId();
                id = RequalityCLI.getNextId(parent, type, parid + "_T");
              } else {
                id = RequalityCLI.getNextId(parent, type, null);
              }
            }
          }
          id = id.replace((char) 160, ' ').replace("&nbsp;", " ").trim();
          try {
            if (!l.canBeCreatedInExc(type, id, parent)) {
              return false;
            }
          } catch (PastingIntoLeafException e) {
            return false;
          } catch (SameIdExistsInChildren e) {
            return false;
          } catch (IdEmpty e) {
            return false;
          } catch (IdContainsWrongChars e) {
            return false;
          }
          return true;
        }
      };

      /*
       * Request to remove a node with all its decendants Input
       * parameters: uuid - (String) uuid of the removed node Return
       * value: Boolean value: true if operation was successful, false
       * otherwise
       */
      new BrowserFunction(browser, "removeNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode n = db.getNode(UUID.fromString((String) arguments[0]));

          List<TreeNode> nds = new ArrayList<>();
          nds.add(n);
          try {
            DeleteSelection.deleteNodes(nds);
            TreeNode par = n.getShownParent();
            if (par == null || par.isDisposed()) par = n.getTreeDB().getRootNode();
            setSelectedNode(par); //TODO needed?
            return true;
          } catch (ExecutionException e) {
            e.printStackTrace();
          }
          //if(!db.getTreeLogic().canBeDeleted(n))
          //return false;
          //db.startTransaction("Remove Element");
          //if (n instanceof Requirement) {
          //deleteLocations((Requirement)n);
          //}
          //n.delete();
          //db.commit();
          return false;
        }
      };
      new BrowserFunction(browser, "setRequalityFocus") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          // workaround (ignore first unieditor selection change)
          if(db == null)
            return true;
          if (currentSelection == null) {
            currentSelection =
                targetNode != null
                    ? targetNode
                    : db.getNode(UUID.fromString((String) arguments[0]));
          } else {
            currentSelection = db.getNode(UUID.fromString((String) arguments[0]));
          }
          if (uniEditorOutlinePage != null) {
            uniEditorOutlinePage.setSelection(new StructuredSelection(currentSelection));
          }
          // synchronization with Requality Explorer
          IViewPart view =
              PlatformUI.getWorkbench()
                  .getActiveWorkbenchWindow()
                  .getActivePage()
                  .findView(RequalityCNF.ID);
          if (view instanceof CommonNavigator) {
            CommonNavigator navigator = (CommonNavigator) view;
            if (navigator.isLinkingEnabled()) {
              navigator.setLinkingEnabled(true);
            }
          }
          fireSelectionChange();
          return true;
        }
      };

      /** returns absolute URL to project folder */
      new BrowserFunction(browser, "getBaseUrl") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          try {
            return RequalityPlugin.getTreesTracker()
                .getProject(db)
                .getLocation()
                .toFile()
                .toURI()
                .toURL()
                .toString();
          } catch (MalformedURLException e) {
            e.printStackTrace();
          }
          return null;
        }
      };

      new BrowserFunction(browser, "getReadOnlyMode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          return editMode;
        }
      };

      new BrowserFunction(browser, "getReadOnlyModeOnNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
          if (node.isVirtual()) return editVirtualMode;
          return 0;
        }
      };

      new BrowserFunction(browser, "getUpdatingState") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          return isUpdating;
        }
      };

      new BrowserFunction(browser, "copyNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          try {
            TreeDB db = getValidDB((String) arguments[0]);
            TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
            if (!node.getTreeDB().getTreeLogic().isCopyable(node)) return false;
            boolean cutNode = false;
            if (arguments.length > 1) cutNode = (Boolean) arguments[1];
            (new CopyHandler()).copyNodes(new TreeNode[] {node}, cutNode);
          } catch (Exception e) {
            return false;
          }
          return true;
        }
      };

      new BrowserFunction(browser, "pasteNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          try {
            TreeNode parent = db.getNode(UUID.fromString((String) arguments[0])), lastnode = null;
            boolean pasteTree = false;
            if (arguments.length > 1) pasteTree = (Boolean) arguments[1];
            if (arguments.length > 2) lastnode = db.getNode(UUID.fromString((String) arguments[2]));
            if (db.startTransaction("Paste element")) {
              TreeNode[] nodes =
                  (TreeNode[])
                      (new PasteEclipseHandler())
                          .execute(parent, lastnode, pasteTree, true, false, true);
              if (nodes.length == 1)
                for (TreeNode node : nodes) {
                  EditElementDialog edd =
                      new EditElementDialog(
                          this.getBrowser().getShell(),
                          node,
                          "Paste element '" + node.getNameOrId() + "'",
                          true);
                  edd.setTransactionIgnore(true);
                  if (edd.open() == Dialog.CANCEL) {
                    db.rollback();
                    return false;
                  }
                }
              db.commit();
            }
            //test step, temporary here
          } catch (RuntimeException e) {
            db.rollback();
            return false;
          }
          return true;
        }
      };
      /**
       * Method to get all possible nodes with given User Visible Name first argument - User Visible
       * Name returns Array of Strings (UUIDs)
       */
      new BrowserFunction(browser, "getNodesByUFName") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          if (db == null) db = JSEditor.this.db;
          List<String> filter = db.getLinkToTypeFilter();
          if (arguments.length > 2) {
            filter = new ArrayList<>();
            String[] tp = ((String) arguments[2]).split(",");
            for (String s : tp) if (!s.equals("")) filter.add(s);
          }
          String[] nd =
              db.getNodesByUFName(
                  ((String) arguments[0]).replace((char) 160, ' ').replace("&nbsp;", " ").trim(),
                  filter);
          if (nd == null) return null;
          Object[] nodes = new Object[nd.length];
          for (int i = 0; i < nd.length; i++) nodes[i] = nd[i];
          return nodes;
        }
      };
      /**
       * Show dialog of selection from list of UUIDs. Then ask for selection. argument 1 - Array of
       * UUID in form of strings returns selected nodes or null
       */
      new BrowserFunction(browser, "fireNodeSelectionFromList") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          List<TreeNode> nodes = new ArrayList<>();
          Object[] o = (Object[]) arguments[0];
          for (int i = 0; i < o.length; i++) {
            //check if node exists
            TreeNode node = db.getNode(UUID.fromString((String) o[i]));
            if (node != null) nodes.add(node);
          }
          String caption = "";
          //if we have 2 arguments, then second one
          //is caption of dialog
          if (arguments.length > 1) {
            caption = (String) arguments[1];
          }
          if (nodes.size() == 0) return null;
          javaDialogOpened = true;
          SelectNodeDialog selDialog =
              new SelectNodesFromList(nodes, getSite().getShell(), db, caption);
          int okcanc = selDialog.open();
          javaDialogOpened = false;
          if (okcanc == SelectNodeDialog.OK) {
            TreeNode res = selDialog.getResult();
            return res.getUUId().toString();
          }
          return null;
        }
      };

      new BrowserFunction(browser, "getLastChildUUID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[0]));
          TreeNode[] children = parent.getSortedShownChildren();
          if (children == null || children.length == 0) return null;
          return children[children.length - 1].getUUId().toString();
        }
      };
      new BrowserFunction(browser, "getParentUUID") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[0]));
          if (parent.getShownParentUUID() == null) return null;
          return parent.getShownParentUUID().toString();
        }
      };
      // [0] - test procedure UUID
      // [1] - previous test step UUID
      new BrowserFunction(browser, "createTestStep") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode parent = db.getNode(UUID.fromString((String) arguments[0]));
          try {
            if (!db.getTreeLogic().canBeCreatedInExc("TestStep", parent)) {
              javaDialogOpened = true;
              MessageDialog.openError(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                  "Add Element",
                  "Element of type Test Step"
                      + " cannot be created in "
                      + parent.getUserVisibleName(true));
              javaDialogOpened = false;
              return null;
            }
          } catch (PastingIntoLeafException e) {
            javaDialogOpened = true;
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Add Element",
                "Can't add Test Step to "
                    + parent.getType()
                    + " with "
                    + (rootNode.getType().equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                    + "Test Purpose child");
            javaDialogOpened = false;
            return null;
          }
          if (arguments.length < 2
              || arguments[1] == null
              || ((String) arguments[1]).length() == 0) {
            // no children
            createTestStep(parent, Index.createOne());
            return null;
          }
          TreeNode node = db.getNode(UUID.fromString((String) arguments[1]));
          TreeNode ts = createTestStep(parent, Index.getNextIndex("TestStep", parent, node));
          return ts.getUUId().toString();
        }
      };
      /**
       * Method to remove node, that key(UUID or "*"+UUID) starts with argument 2 argument 1 - UUID
       * of target node returns true or null
       */
      new BrowserFunction(browser, "removeLink") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
          if (node == null) return null;
          Attribute a = node.attrs().getAttribute(TreeNode.ATTR_LINKS);
          if (a == null) return null;
          List<Object> l = ((ListValue) a.getValue()).getValues();
          if (arguments[1] == null || ((String) arguments[1]).length() < 1) return null;
          for (Object o : l) {
            String s = o.toString();
            if (s.substring(0, s.indexOf('_')).startsWith((String) arguments[1])) {
              if (node.getTreeDB().startTransaction("remove single Link", true)) {
                l.remove(o);
                node.attrs()
                    .setAttribute(new AttributeDefinition(node.getUUId(),
                        TreeNode.ATTR_LINKS,new ConstValueProvider(AttributeType.LIST, 
                            new ListValue(AttributeType.STRING, l)),
                        Flag.HIDDEN));
                node.commitAttributes();
                node.getTreeDB().commit();
                return true;
              } else return null;
            }
          }

          return null;
        }
      };
      /**
       * Returns UUId of new test purpose, selected by dialog returns UUId in form of string or null
       */
      new BrowserFunction(browser, "addNewTestPurposeToEditor") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          //TreeNode node = db.getNode( UUID.fromString((String) arguments[0]));
          javaDialogOpened = true;
          SelectNodeDialog selDialog = new SelectTestPurposeDialog(getSite().getShell(), db);
          int okcanc = selDialog.open();
          javaDialogOpened = false;
          if (okcanc == SelectNodeDialog.OK) {
            TreeNode res = selDialog.getResult();
            return res.getUUId().toString();
          }
          return null;
        }
      };

      new BrowserFunction(browser, "canContainType") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
          return node.getTreeDB().getTreeLogic().canBeChildOf((String) arguments[1], node);
        }
      };
      new BrowserFunction(browser, "canBeChildTypeOf") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          return db.getTreeLogic().canBeChildOf((String) arguments[0], (String) arguments[1]);
        }
      };
      /** Shows dialog and then ask for choose or add. returns absolute URL to project folder */
      new BrowserFunction(browser, "addSingleTestPurposeToTestStep") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          TreeNode node = db.getNode(UUID.fromString((String) arguments[0]));
          javaDialogOpened = true;
          SelectNodeDialog selDialog = new SelectTestPurposeDialog(getSite().getShell(), db);
          int okcanc = selDialog.open();
          javaDialogOpened = false;
          if (okcanc == SelectNodeDialog.OK) {
            TreeNode res = selDialog.getResult();
            //TODO ss
            List<Object> al = node.attrs().getListAttributeValue(TreeNode.ATTR_LINKS);
            if (al == null) al = new ArrayList<>();
            Iterator<Object> i = al.iterator();
            while (i.hasNext()) {
              String str = i.next().toString();
              //if we already have hidden link
              //then remove it
              if (str.startsWith("*")) i.remove();
            }
            al.add("*" + res.getUUId().toString() + "_" + res.getUserVisibleName(true));
            if (node.getTreeDB().startTransaction("Add Single Test Purpose", true)) {
              node.attrs()
                  .setAttribute(
                      new AttributeDefinition(node.getUUId(), TreeNode.ATTR_LINKS, 
                          AttributeAvailability.LOCAL,
                          new TemplValueProvider(new ListValue(AttributeType.STRING, al)),
                          Flag.HIDDEN));
              node.commitAttributes();
              node.getTreeDB().commit();
              //refresh(node);
              //	refresh(res);
              return res.getUUId().toString();
            }
          }
          return null;
        }
      };
      /**
       * Open Editor for node. If it is not possible, then open Editor to node's parent. Otherwise,
       * do nothing
       */
      new BrowserFunction(browser, "openEditorForNode") {
        @Override
        public Object function(Object[] arguments) {
          if (!cnter.containsKey(getName())) cnter.put(getName(), 1L);
          else cnter.put(getName(), cnter.get(getName()) + 1);

          TreeDB db = getValidDB((String) arguments[0]);
          final TreeNode nde = db.getNode(UUID.fromString((String) arguments[0]));
          TreeNode node = nde;
          final TreeNode sel = nde;
          String tp = node.getType();
          if (!node.getQualifiedId().equals(Requirement.getTypeRootQId()))
            node = node.getShownParent();
          NodeEditorInput input = new NodeEditorInput(node, sel);
          String editorid = UniEditor.ID;
          try {
            if (tp.equals(Document.TYPE_NAME)) editorid = ReqMarker.ID;
            IDE.openEditor(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(),
                input,
                editorid,
                false);
          } catch (PartInitException e) {
            e.printStackTrace();
          }
          return null;
        }
      };

    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  @Override
  public void setFocus() {
    new UndoRedoActionGroup(
            getSite(), PlatformUI.getWorkbench().getOperationSupport().getUndoContext(), true)
        .fillActionBars(getEditorSite().getActionBars());
    browser.setFocus();
    fireSelectionChange();
  }

  public void dispose() {
    getSite().getPage().removePartListener(plistener);
    db.removeGUIListener(guiTreeChangeListener);
    EclipseTracker.getTracker().removeListener(treeStateListener);
    for (String s : cnter.keySet()) RLogger.logDebug(s + ": " + cnter.get(s),2);
    cnter.clear();
    super.dispose();
  }

  public JSEditor() {
    this("unieditor.html", defaultAttrs);
  }

  public JSEditor(String module,Set<String> rats) {
    super();
    try {
      html = new File(Activator.getPath("jstools/" + module));
      systemAttrs.addAll(rats);
    } catch (IOException e) {
      e.printStackTrace();
    }
    or = new ObjectRule(this);
  }

  private boolean execute(String s) {
    return browser.execute("try{" + s + "}catch(ed){alert(ed);}");
  }
  private boolean relates(TreeNode nde) {
    UUID vroot = nde.getVRoot();
    if(vroot != null){
      TreeNode root = null;
      try{
        root = nde.getTreeDB().getNodeExc(vroot);
        if(root!=null)
          return !root.getId().startsWith("com.unitesk.requality.report");//TODO flag in report generator
      }catch(Exception e){
        return false;
      }
    }
    try{
      String qid = nde.getQualifiedId(true), rootqid = rootNode.getQualifiedId(true);
      return (qid + "/").startsWith(rootqid + "/");
    }catch(NodeNotFoundException nnfe)
    {
      return false;//unloaded subtree
    }
  }
  protected void oldRefresh(final TreeNode node) {
    // Refresh nothing but root node subtree
    updateFreeTimer.cancel();
    isUpdating = true;
    if(!relates(node))
      return;
    String rootQID = this.rootNode.getQualifiedId();
    String nodeQID = node.getQualifiedId(true);
    if (!nodeQID.startsWith(rootQID + "/") && (!nodeQID.equals(rootQID))) {
      if (node.getType().equals("TestPurpose") && (this.rootNode.getType().equals("TestProcedure")))
        execute("testStepUpdate('" + node.getUUId() + "');");
      return;
    }
    if (rootQID.equals(nodeQID)) updateTabLabel();
    getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                if (browser != null && !browser.isDisposed()) {
                  synchronized (JSEditor.this) {
                    boolean torem = false;
                    if (changedNodes.containsKey(node.getUUId())) {
                      NodeDesc nd = changedNodes.get(node.getUUId());
                      if (nd.equals(node.toNodeDesc())) {
                        Integer count = changedNodesCount.get(node.getUUId());
                        if (count > 0) {
                          changedNodesCount.put(
                              node.getUUId(), changedNodesCount.get(node.getUUId()) - 1);
                          return;
                        }
                      }
                      torem = true;
                    }

                    execute("onChangeNode('" + node.getUUId() + "');");
                    if (torem) {
                      changedNodes.remove(node.getUUId());
                      changedNodesCount.remove(node.getUUId());
                    }
                    //  browser.forceFocus();
                  }
                }
              }
            });
    updateFreeTimer = new Timer();
    updateFreeTimer.schedule(
        new TimerTask() {
          @Override
          public void run() {
            isUpdating = false;
          }
        },
        300);
  }

  protected void setUniEditorFocus(final UUID uuId) {
    if(currentSelection==null || !currentSelection.getUUId().equals(uuId))
      getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                if (javaDialogOpened) return;
                if (browser != null && !browser.isDisposed()) {
                  if(uuId!=null)
                    execute("onChangeFocus('" + uuId + "');");
                  else
                    execute("onChangeFocus();");
                }
              }
            });
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  public Object getAdapter(Class required) {
    if (IContentOutlinePage.class.equals(required)) {
      if (uniEditorOutlinePage == null) {
        uniEditorOutlinePage = new UniEditorOutlinePage();
        uniEditorOutlinePage.addSelectionChangedListener(outlineSelectionChangeListener);
        uniEditorOutlinePage.setInput(this.rootNode);
      }
      return uniEditorOutlinePage;
    }
    if (required == IPropertySheetPage.class) {
      return new TabbedPropertySheetPage(
          new ITabbedPropertySheetPageContributor() {
            @Override
            public String getContributorId() {
              return "com.unitesk.requality.views.RequalityPropertyContributor";
            }
          });
    }
    if (required.equals(IContributedContentsView.class)) {
      return new IContributedContentsView() {
        public IWorkbenchPart getContributingPart() {
          return PlatformUI.getWorkbench()
              .getActiveWorkbenchWindow()
              .getActivePage()
              .findView("com.unitesk.requality.views.RequalityExplorer");
        }
      };
    }
    return super.getAdapter(required);
  }

  public Browser getBrowser() {
    return browser;
  }

  public void setEditMode(int editMode) {
    this.editMode = editMode;
  }

  private void closeEditor() {
    getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                getEditorSite().getPage().closeEditor(JSEditor.this, false);
              }
            });
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Selection provider
  //
  // //////////////////////////////////////////////////////////////////////////

  private List<ISelectionChangedListener> selectionChangeListeners =
      new ArrayList<>();

  @Override
  public void addSelectionChangedListener(ISelectionChangedListener listener) {
    selectionChangeListeners.add(listener);
  }

  @Override
  public void removeSelectionChangedListener(ISelectionChangedListener listener) {
    selectionChangeListeners.remove(listener);
  }

  @Override
  public ISelection getSelection() {
    if (currentSelection == null) {
      return StructuredSelection.EMPTY;
    }
    return new StructuredSelection(currentSelection);
  }

  @Override
  public void setSelection(ISelection selection) {}

  private void fireSelectionChange() {
    ISelection sel = this.getSelection();
    SelectionChangedEvent event = new SelectionChangedEvent(this, sel);
    for (ISelectionChangedListener listener : selectionChangeListeners) {
      listener.selectionChanged(event);
    }
  }

  private TreeNode createTestStep(TreeNode parent, Index index) {
    String type = "TestStep";
    String id = RequalityCLI.getNextId(parent, type, null);
    if (parent.findWorkChild(id) == null) {
      try {
        if (!db.getTreeLogic().canBeCreatedInExc(type, id, parent)) {
          javaDialogOpened = true;
          MessageDialog.openError(
              PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
              "Add Element",
              "Element of type 'TestStep"
                  + "' cannot be created in "
                  + parent.getUserVisibleName(true));
          javaDialogOpened = false;
          return null;
        }
      } catch (PastingIntoLeafException e) {
        javaDialogOpened = true;
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Add Element",
            "Can't add "
                + rootNode.getTypeLabel()
                + " to "
                + parent.getTypeLabel()
                + " with "
                + (rootNode.getType().equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                + "Test Purpose child");
        javaDialogOpened = false;
        return null;
      } catch (SameIdExistsInChildren e) {
        javaDialogOpened = true;
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Add Element",
            "Element with same id exists in " + parent.getUserVisibleName(true));
        javaDialogOpened = false;
        return null;
      } catch (IdEmpty e) {
        javaDialogOpened = true;
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Add Element",
            "Element id is empty");
        javaDialogOpened = false;
        return null;
      } catch (IdContainsWrongChars e) {
        javaDialogOpened = true;
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Add Element",
            "Element id contains wrong character - " + e.getWrongCharacter());
        javaDialogOpened = false;
        return null;
      }
      if (db.startTransaction("Add Test Step", true)) {
        TreeNode newNode = parent.createChild(id, type);
        newNode.setIndex(index);
        newNode.commitAttributes();
        db.commit();
        return newNode;
      }
    }
    return null;
  }

  public TreeNode getRootNode() {
    return rootNode;
  }
}
