package com.unitesk.requality.eclipse.ui.editors;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import org.apache.commons.httpclient.util.URIUtil;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.browser.BrowserFunction;
import org.eclipse.swt.browser.ProgressEvent;
import org.eclipse.swt.browser.ProgressListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.part.IContributedContentsView;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.editors.browser.ReqMarker;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipseCopyHandler;
import com.unitesk.requality.eclipse.handlers.copypaste.EclipsePasteHandler;
import com.unitesk.requality.eclipse.handlers.report.OpenReportHandler;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.tests.TestDialogProvider;

// TODO: refacatoring
public class ReportViewer extends EditorPart implements ISelectionProvider, ISearchProvider {
  public static String ID = "com.unitesk.requality.editors.browser.ReportViewer";
  private ReportData rdata;
  private Browser browser;
  private String href;

  ITreeStateListener treeStateListener =
      new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {}

        @Override
        public void delete(IProject proj) {
          if (EclipseTracker.getResourceByNode(rdata).getProject().equals(proj)) {
            getSite()
                .getShell()
                .getDisplay()
                .asyncExec(
                    new Runnable() {
                      @Override
                      public void run() {
                        getEditorSite().getPage().closeEditor(ReportViewer.this, false);
                      }
                    });
          }
        }

        @Override
        public void close(TreeDB db) {
          if (rdata.getTreeDB().equals(db)) {
            getSite()
                .getShell()
                .getDisplay()
                .asyncExec(
                    new Runnable() {
                      @Override
                      public void run() {
                        getEditorSite().getPage().closeEditor(ReportViewer.this, false);
                      }
                    });
          }
        }
      };

  INodeChangeListener nodeChangeListener =
      new INodeChangeListener() {

        @Override
        public void deleted(TreeNode source) {
          if (source.getUUId().equals(rdata.getUUId()))
            getSite()
                .getShell()
                .getDisplay()
                .asyncExec(
                    new Runnable() {
                      @Override
                      public void run() {
                        getEditorSite().getPage().closeEditor(ReportViewer.this, false);
                      }
                    });
        }

        @Override
        public void attributeChange(TreeNode node, String what, Attribute old) {
          if (what.equals(ReportSettings.ATTR_TEMPLATE) || what.equals(TreeNode.ATTR_UNDEFINED)) {
            if (!rdata.getQualifiedId().equals(node.getQualifiedId())) {
              return;
            }
            Template tpl = Activator.getDefault().getReportTemplates().get(rdata.getTemplateId());
            String path = null;
            if (tpl == null) {
              // get first
              if (rdata.getTemplateId() != null)
                MessageDialog.openError(
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                    "Report generation",
                    "Report template '"
                        + rdata.getTemplateId()
                        + "' for this report is not avalible.");
              path =
                  ((ReportData) node).getSaveDir()
                      + "/"
                      + CoreUtils.URLEncode(
                          node.attrs().getStringAttributeValue(ReportData.ATTR_LAST_GENERATED));
            } else {
              path =
                  tpl.getTemplates().get(0).getOutputFileName(node.getTreeDB());
            }
            final IResource res =
                OpenReportHandler.getReportFolder((ReportData) node)
                    .getFile(Path.fromOSString(URI.create(path).toString()));
            if (res.exists()) {
              getSite()
                  .getShell()
                  .getDisplay()
                  .asyncExec(
                      new Runnable() {
                        @Override
                        public void run() {
                          OpenReportHandler.openReport(rdata);
                          if (!browser.isDisposed()) browser.refresh();
                        }
                      });
            } else {
              getSite()
                  .getShell()
                  .getDisplay()
                  .asyncExec(
                      new Runnable() {
                        @Override
                        public void run() {
                          getEditorSite().getPage().closeEditor(ReportViewer.this, false);
                        }
                      });
            }
          }
        }

        @Override
        public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          if (what.contains(ReportSettings.ATTR_TEMPLATE) || what.contains(TreeNode.ATTR_UNDEFINED))
            attributeChange(source, ReportSettings.ATTR_TEMPLATE, what.getOld(ReportSettings.ATTR_TEMPLATE));
        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };

  private String reportName;

  public ReportViewer() {
    super();
  }

  @Override
  public void doSave(IProgressMonitor monitor) {}

  @Override
  public void doSaveAs() {}

  @Override
  public void init(IEditorSite site, IEditorInput input) throws PartInitException {
    setSite(site);
    setInput(input);
    try {
      rdata = (ReportData) ((NodeEditorInput) input).getStorage().getNode();
      reportName = rdata.getId();
      Template tpl = Activator.getDefault().getReportTemplates().get(rdata.getTemplateId());
      URI savePath = rdata.getSaveDir();
      if (tpl == null) {
        if (rdata.getTemplateId() != null) {
          String lastGen = rdata.attrs().getStringAttributeValue(ReportData.ATTR_LAST_GENERATED);
          if (lastGen != null) {
            href =
                Activator.getUrl(new URI(savePath + "/" + URIUtil.encodeQuery(lastGen))).toString();
            MessageDialog.openError(
                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                "Report generation",
                "Report template '" + rdata.getTemplateId() + "' for this report is not avalible.");
          } //else we have new template and it is ridiculous
        } else {
          tpl = Activator.getDefault().getReportTemplates().get(Template.DEFAULT_ID);
        }
      }
      if (tpl != null)
        href =
            Activator.getUrl(
                    new URI(
                        savePath
                            + "/"
                            + URIUtil.encodeQuery(
                                tpl.getTemplates().get(0).getOutputFileName(rdata.getTreeDB()).toString())))
                .toString();
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    getSite().setSelectionProvider(this);

    fireSelectionChange();
  }

  @Override
  public boolean isDirty() {
    return false;
  }

  @Override
  public boolean isSaveAsAllowed() {
    return false;
  }

  @SuppressWarnings({"unchecked", "rawtypes"})
  @Override
  public Object getAdapter(Class required) {
    if (required == IPropertySheetPage.class) {
      TabbedPropertySheetPage sheet =
          new TabbedPropertySheetPage(
              new ITabbedPropertySheetPageContributor() {
                @Override
                public String getContributorId() {
                  return "com.unitesk.requality.views.RequalityPropertyContributor";
                }
              });
      return sheet;
    }
    if (required.equals(IContributedContentsView.class)) {
      return new IContributedContentsView() {
        public IWorkbenchPart getContributingPart() {
          IWorkbenchPart part =
              PlatformUI.getWorkbench()
                  .getActiveWorkbenchWindow()
                  .getActivePage()
                  .findView("com.unitesk.requality.views.RequalityExplorer");
          if (part == null || part.getSite() == null) return null;
          part.getSite().getSelectionProvider().setSelection(new StructuredSelection(rdata));
          return part;
        }
      };
    }
    return super.getAdapter(required);
  }

  @Override
  public void createPartControl(Composite parent) {
    setPartName("Report: " + reportName);
    browser = RequalityPlugin.newBrowser(parent, RequalityPlugin.getBrowserType());
    IHandlerService handlerService = getSite().getService(IHandlerService.class);
    handlerService.activateHandler("org.eclipse.ui.edit.copy", new EclipseCopyHandler());
    handlerService.activateHandler("org.eclipse.ui.edit.paste", new EclipsePasteHandler());

    if (href != null) {
      browser.setUrl(href);
      browser.addProgressListener(
          new ProgressListener() {

            @Override
            public void completed(ProgressEvent event) {
              fireSelectionChange();
            }

            @Override
            public void changed(ProgressEvent event) {}
          });
      new BrowserFunction(browser, "findText"){
        @Override
        public Object function(Object[] arguments) {
          return browser.evaluate("return window.find( \""+arguments[0]+
                    "\", "+arguments[1]+
                    ", "+arguments[2]+
                    ", "+arguments[3]+
                    ", false"+
                    ", false"+
                    ", false"+
                    ")");
        }
      };
      new BrowserFunction(browser, "getSelectionText"){
        @Override
        public Object function(Object[] arguments) {
          return browser.evaluate(
          "var selection = window.getSelection();"+
          "var text = \"\";"+
          "for( i = 0; i < selection.rangeCount; i++ )"+
          "{"+
          "    text += selection.getRangeAt(i).toString();"+
          "}"+
          "return text;");
        }
      };
      new BrowserFunction(browser, "callAddTest") {
        @Override
        public Object function(Object[] arguments) {
          return TestDialogProvider.fireTestCreation(
              rdata.getTreeDB(),
              (String) arguments[0],
              PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell());
        }
      };
      new BrowserFunction(browser, "selectNode") {
        @Override
        public Object function(Object[] arguments) {
          String qid = (String) arguments[0];
          try {
            TreeNode node =
                rdata
                    .getTreeDB()
                    .getNode(
                        qid,
                        com.unitesk.requality.core.Path.QID | com.unitesk.requality.core.Path.UV);
            if (node == null) return false;
            Map<String, List<TreeNode>> sel = new HashMap<>();
            sel.put(node.getType(), new ArrayList<TreeNode>());
            sel.get(node.getType()).add(node);
            SelectionManager.getOne().setActiveNodes(sel, sel, EventSource.EDITOR);
          } catch (Exception e) {
            e.printStackTrace();
            return false;
          }
          return true;
        };
      };
      new BrowserFunction(browser, "openEditor") {
        @Override
        public Object function(Object[] arguments) {
          try {
            String path = (String) arguments[0];
            if (path.startsWith("file:")) {
              URI uri = URI.create(path);
              IEditorDescriptor desc =
                  PlatformUI.getWorkbench().getEditorRegistry().getDefaultEditor(path);
              if (desc == null) {
                IEditorDescriptor[] d =
                    PlatformUI.getWorkbench().getEditorRegistry().getEditors(path);
                if (d != null && d.length > 0) desc = d[0];
              }

              if (desc == null)
                desc = PlatformUI.getWorkbench().getEditorRegistry().getDefaultEditor("*.txt");
              if (desc != null)
                return IDE.openEditor(
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(),
                    uri,
                    desc.getId(),
                    true);
            } else if (path.startsWith("requality://")) {
              path = path.substring(12);
              String projname = path.substring(0, path.substring(1).indexOf("/") + 1);
              IProject proj = Activator.getProjectByName(projname);
              if (proj == null) return null;
              TreeDB db = EclipseTracker.getTracker().getTree(proj);
              if (db == null) return null;
              try {
                TreeNode node =
                    db.getNode(
                        path.substring(projname.length()),
                        com.unitesk.requality.core.Path.QID
                            | com.unitesk.requality.core.Path.UV
                            | com.unitesk.requality.core.Path.UUID);
                if (node == null) return null;
                return openEditor(node);
              } catch (Exception e) {
              }
              return true;
            } else {
              TreeNode node = null;
              try {
                node =
                    rdata
                        .getTreeDB()
                        .getNode(
                            path,
                            com.unitesk.requality.core.Path.QID
                                | com.unitesk.requality.core.Path.UV
                                | com.unitesk.requality.core.Path.UUID);
              } catch (Exception e) {
                return null;
              }
              return openEditor(node);
            }

          } catch (PartInitException e) {
            e.printStackTrace();
          }
          return null;
        };
      };
      rdata.addChangeListener(nodeChangeListener);
      EclipseTracker.getTracker().addListener(treeStateListener);
    }
  }

  protected IEditorPart openEditor(TreeNode node) {
    String type = node.getType();
    NodeEditorInput input = new NodeEditorInput(node);
    if (type.equals(TestPurpose.TYPE_NAME)
        || type.equals(Comment.TYPE_NAME)
        || type.equals("TestStep")
        || type.equals(Location.TYPE_NAME)) input = new NodeEditorInput(node.getWorkParent(), node);
    String editorid = UniEditor.ID;
    try {
      if (node.getType().equals(Document.TYPE_NAME)) editorid = ReqMarker.ID;
      IEditorReference eref;
      if (editorid == ReqMarker.ID) {
        input = new NodeEditorInput(node);
        eref = RequalityPlugin.findMarkEditorForDocument(node.getQualifiedId());
      } else
        eref =
            RequalityPlugin.findJSEditorForNode(
                node.getQualifiedId(), EclipseTracker.getResourceByNode(node).getProject());
      IEditorPart part;
      if (eref != null) part = eref.getEditor(true);
      else {
        IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
        part = IDE.openEditor(page, input, editorid, false);
      }
      if (editorid == UniEditor.ID) ((UniEditor) part).setSelectedNode(node);
      part.getSite().getPage().bringToTop(part);
      return part;
      //IDE.openEditor(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(), input, editorid, false);
    } catch (PartInitException e) {
      e.printStackTrace();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public void dispose() {
    rdata.removeChangeListener(nodeChangeListener);
    EclipseTracker.getTracker().removeListener(treeStateListener);
    super.dispose();
  }

  @Override
  public void setFocus() {
    browser.setFocus();
    fireSelectionChange();
  }

  private List<ISelectionChangedListener> selectionChangeListeners =
      new ArrayList<>();

  @Override
  public void addSelectionChangedListener(ISelectionChangedListener listener) {
    selectionChangeListeners.add(listener);
  }

  @Override
  public void removeSelectionChangedListener(ISelectionChangedListener listener) {
    selectionChangeListeners.remove(listener);
  }

  @Override
  public ISelection getSelection() {
    return new StructuredSelection(rdata);
  }

  @Override
  public void setSelection(ISelection selection) {}

  private void fireSelectionChange() {
    ISelection sel = this.getSelection();
    SelectionChangedEvent event = new SelectionChangedEvent(this, sel);
    for (ISelectionChangedListener listener : selectionChangeListeners) {
      listener.selectionChanged(event);
    }
  }

  @Override
  public boolean findText(String aString, boolean aCaseSensitive, boolean aBackwards, boolean aWrapAround) {
      return (Boolean)
          browser.evaluate(
              "return findText('"
                  + aString
                  + "',"
                  + aCaseSensitive
                  + ","
                  + aBackwards
                  + ","
                  + aWrapAround
                  + ");");
  }


  public String getSelectionText() {
    return getSelectionText(true);
  }

  private String getSelectionText(boolean process) {
    String res = (String) browser.evaluate("return getSelectionText();");
    if (res == null) return "";
    if (!process) return res;
    StringTokenizer st = new StringTokenizer(res, "\n ");
    res = "";
    while (st.hasMoreTokens()) {
      if (!res.isEmpty()) {
        res += " ";
      }
      res += st.nextToken();
    }
    return res;
  }
}
