package com.unitesk.requality.eclipse.ui.editors;

import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IWorkbenchPartSite;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.operations.UndoRedoActionGroup;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;

import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.ui.cnf.RequalitySorter;
import com.unitesk.requality.eclipse.views.RequalityContentProvider;

public class UniEditorOutlinePage extends ContentOutlinePage {
  private TreeNode input;

  @Override
  public void dispose() {
    super.dispose();
  }

  public UniEditorOutlinePage() {}

  @Override
  public void createControl(Composite parent) {
    super.createControl(parent);
    getTreeViewer().setLabelProvider(WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider());
    getTreeViewer().setContentProvider(new RequalityContentProvider(true));
    getTreeViewer().setComparator(new RequalitySorter());
    if (input != null) {
      getTreeViewer().setInput(new Object[] {input});
    }
    getTreeViewer().expandAll();
    IWorkbenchPartSite site =
        PlatformUI.getWorkbench()
            .getActiveWorkbenchWindow()
            .getActivePage()
            .getActivePart()
            .getSite();
    IUndoContext context = PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
    new UndoRedoActionGroup(site, context, true).fillActionBars(getSite().getActionBars());
  }

  public void setInput(TreeNode node) {
    input = node;
    if (getTreeViewer() != null) {
      getTreeViewer().setInput(node);
    }
  }

  public synchronized void refresh(final TreeNode node) {
    getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                //if control is disposing then old code may
                //throw exception
                if ((UniEditorOutlinePage.this.getControl() != null)
                    && !UniEditorOutlinePage.this.getControl().isDisposed()) {
                  getTreeViewer().refresh(node);
                }
              }
            });
  }
}
