/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.viewers.AbstractTreeViewer;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.progress.UIJob;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeChangeEvent;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.ui.ObjectRule;
import com.unitesk.requality.eclipse.ui.RequalityLCNF;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.report.TraceabilityDataSource;

public class RequalityContentLinksProvider implements ITreeContentProvider {
  private TreeViewer viewer;
  private ArrayList<TreeDB> dbs = new ArrayList<>();
  private byte rootInited = 0;

  public RequalityContentLinksProvider() {}

  private void refresh(TreeNode node, boolean b) {
    rootInited = 0;
    if (node != null) viewer.refresh(node, b);
    else viewer.refresh(b);
    viewer.expandToLevel(3);
  }

  private boolean isShown(String w, Attribute val) {
    if( w.equals(TreeNode.ATTR_TERMS) || 
        w.equals(TreeNode.ATTR_TERMS_REV))
      return true;
    if(RequalityCNF.isShown(w))
      return true;
    if(val == null || val.isReference())
      return true;
    return false;
  }
  private ObjectRule viewerRule;
  private Set<TreeNode>aset = new HashSet<>();
  private ReadWriteLock rwl = new ReentrantReadWriteLock(false);
  private Lock wl = rwl.writeLock();
  private Lock rl = rwl.readLock();
  private UIJob uj = null;
  private long lup = -1;
  private void addToch(TreeNode... as) {
    wl.lock();
    aset.addAll(Arrays.asList(as));
    try{
      if(uj == null){
        lup = System.currentTimeMillis();
        uj = new UIJob("update reqcnf"){
          
          @Override
          public IStatus runInUIThread(IProgressMonitor monitor) {

            boolean wlo=false,rlo=true;
            try{
              Lock wl2 = rwl.writeLock();
              Lock rl = rwl.readLock();
              // TODO Auto-generated method stub
               rl.lock();
               long d = (System.currentTimeMillis()-lup);
               boolean first = true;
               int cnt = 0, max = aset.size();
               while((d<RequalityContentProvider.LIMIT2||first)&&(cnt!=max||d>RequalityContentProvider.LIMIT)){
                first = false;
                cnt = 0;
                if(aset.size()>0){
                  Set<TreeNode> toch = new HashSet<>(aset);
                  rl.unlock();
                  rlo = false;
                  Set<TreeNode> fch = new HashSet<>();
                  Set<TreeNode> exps = new HashSet<>();
                  for(Object o : viewer.getExpandedElements()){
                    TreeNode on = RequalityContentProvider.extractNode(o);
                    if(on != null){
                      exps.add(on);
                      for(TreeNode onc:on.getShownChildren())
                        exps.add(onc);
                    }
                  }
                  for(TreeNode node:toch){
                    if(node == null)
                      continue;
                    for(TreeNode onc:exps)
                      if(onc.getUUId().equals(node.getUUId())){
                        fch.add(node);
                        break;
                      }else
                        try{
                          if(node.getShownParentExc()!=null && 
                          onc.getUUId().equals(node.getShownParentExc().getUUId())){
                            TreeNode sp = node.getShownParent();
                            if(sp!=null){
                              fch.add(node);
                              break;
                            }
                          }
                        }
                        catch(NodeNotFoundException nde){
                          
                        }  
                    cnt++;
                  }
                  final Set<TreeNode> ftoch = fch;
                  if (viewer.getControl().isDisposed()) {
                    return Status.CANCEL_STATUS;
                  }
                  //TODO null update? mb bug
                  //Object o = viewer.getInput();
                  //RLogger.logInfo(node.toString());
                  if (ftoch.contains(null))
                    viewer.refresh(true);
                  else
                  {
                    viewer.getControl().setRedraw(false);
                    try{
                      wl2.lock();
                      wlo = true;
                      for (TreeNode node : ftoch) {
                        aset.remove(node);
                        max--;
                        cnt--;
                        if (viewer.getControl().isDisposed()) {
                          return Status.CANCEL_STATUS;
                        }
                        if (node != null && (node.equals(node.getTreeDB().getRootNode()))) {
                          TreeNode nde = node;
                          if (EclipseTracker.getResourceByNode(nde) != null)
                            viewer.refresh(
                                EclipseTracker.getResourceByNode(nde).getProject(), true);
                          return Status.OK_STATUS;
                        }
                        if (node != null && !node.isDisposed()) viewer.refresh(node, true);
                      }
                    }
                    finally{
                      viewer.getControl().setRedraw(true);
                      wl2.unlock();
                      wlo = false;
                    }
                  }
                  rl.lock();
                  rlo = true;
                  max = aset.size();
                }
                d = (System.currentTimeMillis()-lup);
              }
            }finally{
              if(rlo)
                rl.unlock();
              if(wlo)
                wl.unlock();
              uj = null;
            }
          
            return Status.OK_STATUS;
          }
        };
       uj.setRule(viewerRule);
       uj.setPriority(Job.LONG);
       uj.schedule();
      }
    }
    finally{
      wl.unlock();
    }
  };
  private ITreeChangeListener treeChangeListener =
      new DefaultTreeChangeListener() {
        private List<TreeNode> dirtyNodes = new ArrayList<>();

        private void guiRefresh(final TreeNode node) {
          addToch(node);
        }

        @Override
        public void moved(TreeNode node, TreeNode oldParent, TreeNode newParent) {
          guiRefresh(oldParent);
          if (oldParent != null
              && newParent != null
              && !oldParent.getUUId().equals(newParent.getUUId())) guiRefresh(newParent);
          guiRefresh(node);
        }

        @Override
        public void deleted(TreeNode node) {
          for(Object o : viewer.getExpandedElements())
            if(o instanceof TreeNode)
              if(((TreeNode)o).getUUId().equals(node.getUUId()))
                guiRefresh((TreeNode)o);
        }

        @Override
        public void changed(TreeNode node, ChangesSet<Attribute> cs) {
          boolean changeIt = false;//, changeParent = false;
          Map<String,Attribute> newA = cs.getNewMap();
          for (Entry<String, Attribute> en:newA.entrySet()) {
            Attribute a = en.getValue();
            Attribute old = cs.getOld(en.getKey());
            if ((en.getKey().equals(TreeNode.ATTR_LABEL)||isShown(en.getKey(),a))) changeIt = true;
            //if (Attribute.affectsPosition(node, w)) changeParent = true;
            if (!changeIt) {
              if ((a != null) && (a.isReference())) changeIt = true;
              if (!changeIt && old != null && old.isReference())
                changeIt = true;
            }
            //if (changeParent) break;
          }
          //if (changeParent) guiRefresh(node.getShownParent());
          if (changeIt) guiRefresh(node);
        }

        @Override
        public void created(TreeNode node) {
          guiRefresh(node);
        }

        @Override
        public void changed(List<TreeNodeChangeEvent> changes) {
          Set<TreeNode> toch = new HashSet<>();
          for (TreeNodeChangeEvent ch : changes) {
            switch (ch.getType()) {
              case MOVE:
                {
                  TreeNode node = ch.getTreeNodeTarget();
                  toch.add(node);
                  TreeNode old = (TreeNode) ch.getArguments()[1];
                  if (old != null) toch.add(old);
                  if (old==null || !old.equals(node.getShownParent())) toch.add(node.getShownParent());

                  //moved(ch.getTreeNodeTarget(),(TreeNode)ch.getArguments()[1],(Boolean)ch.getArguments()[0]);
                }
                break;
              case CREATE:
                toch.add(ch.getTreeNodeTarget());
                //created(ch.getTreeNodeTarget(), (Boolean)ch.getArguments()[0]);
                break;

              case DELETE:
                toch.add(null);
                //deleted(ch.getNodeDescTarget(), (Boolean)ch.getArguments()[0]);
                break;
              case CACHE_CHANGE:
                {
                  toch.add(ch.getTreeNodeTarget());
                }
                break;
              case CHANGE:
                {
                  TreeNode node = ch.getTreeNodeTarget();
                  boolean changeIt = false, changeParent = false;
                  @SuppressWarnings("unchecked")
                  ChangesSet<Attribute> what = (ChangesSet<Attribute>) ch.getArguments()[1];
                  Map<String, Attribute> newA = what.getNewMap();
                  for (Entry<String, Attribute> en : newA.entrySet()) {
                    Attribute a = en.getValue();
                    Attribute old = what.getOld(en.getKey());
                    if (en.getKey().equals(TreeNode.ATTR_LABEL)|| isShown(en.getKey(),a)) changeIt = true;
                    if (Attribute.affectsPosition(node, en.getKey())) changeParent = true;
                    if (!changeIt) {
                      if ((a != null) && (a.isReference())) changeIt = true;
                      if (!changeIt
                          && old != null
                          && old.isReference()) changeIt = true;
                    }
                    if (changeParent) break;
                  }
                  if (changeParent) toch.add(node.getShownParent());
                  else if (changeIt) toch.add(node);
                }
                break;
              default:
                break;
            }
          }
          final Set<TreeNode> ftoch = toch;
          if(ftoch.size()>0)
            Display.getDefault().asyncExec(
                  new Runnable() {
                    @Override
                    public void run() {
                      if (viewer.getControl().isDisposed()) {
                        return;
                      }
                      //TODO null update? mb bug
                      Object o = viewer.getInput();
                      //System.out.println(node.toString());
                      if (o == null || !(o instanceof TreeNode) || ftoch.contains(null))
                        refresh(null, true);
                      else if (ftoch.size() != 0) refresh(null, true);
                    }
                  });
        }

//        @Override
//        public void cacheChanged(
//            TreeNode node,
//            String section,
//            java.util.Set<String> keys,
//            Map<String, Map<String,Object>> oldvs,
//            Map<String, Map<String,Object>> newvs,
//            boolean forceInternal) {
//          boolean con = false;
//          for (String key : keys)
//            if (key.contains(TreeNode.ATTR_REVERT_LINKS)) {
//              con = true;
//            }
//          if (con) guiRefresh(node);
//        };
      };

  private ITreeStateListener stateListener =
      new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
          dbs.add(db);
          db.addGUIListener(treeChangeListener);
        }

        @Override
        public void delete(IProject proj) {
          dbs.remove(EclipseTracker.getTracker().getTree(proj));
        }

        @Override
        public void close(TreeDB db) {
          dbs.remove(db);
        }
      };

  @Override
  public void dispose() {
    RequalityPlugin.getTreesTracker().removeListener(stateListener);
  }

  @Override
  public void inputChanged(final Viewer viewer, Object oldInput, Object newInput) {
    rootInited = 0;
    RequalityPlugin.getTreesTracker().removeListener(stateListener);
    if (!(viewer instanceof TreeViewer)) {
      throw new IllegalArgumentException();
    }
    this.viewer = (TreeViewer) viewer;
    viewerRule = new ObjectRule(viewer);
    for (TreeDB db : dbs) {
      db.removeGUIListener(treeChangeListener);
    }
    dbs.clear();

    RequalityPlugin.getTreesTracker().addListener(stateListener);
    if (newInput == null) {
      return;
    }
    if (newInput instanceof TreeNode) {
      dbs.add(((TreeNode) newInput).getTreeDB());
    } else if (newInput instanceof IWorkspaceRoot) {
      dbs.addAll(Arrays.asList(RequalityPlugin.getTreesTracker().getTrees()));
    } else if (newInput instanceof Object[]) {
      dbs.add(((TreeNode) ((Object[]) newInput)[0]).getTreeDB());
    } else if (newInput instanceof TreeDB) {
      dbs.add((TreeDB) newInput);
    } else {
      throw new IllegalArgumentException();
    }
    for (TreeDB db : dbs) {
      db.addGUIListener(treeChangeListener);
    }
    viewer
        .getControl()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                //TODO if control is disposing, old code will throw exception
                if ((RequalityContentLinksProvider.this.viewer.getControl() != null)
                    && !RequalityContentLinksProvider.this.viewer.getControl().isDisposed()) {
                  RequalityContentLinksProvider.this.refresh(null, true);
                  ((AbstractTreeViewer) viewer).expandAll();
                }
              }
            });
  }

  public static class TestNodeLinksGroup {
    private boolean incoming;
    private Object[] links;
    private TreeNode par;
    private String lbl;

    public TestNodeLinksGroup(TreeNode node, String lbl, boolean incoming, Object[] links) {
      this.incoming = incoming;
      this.links = links;
      this.lbl = lbl;
      par = node;
    }

    @Override
    public String toString() {
      return lbl;
//      if (incoming) return "Incoming";
//      return "Outcoming";
    }

    public TreeNode getParent() {
      return par;
    }

    public boolean isUp() {
      return incoming;
    }
  }

  public static class TestNodeLinks {
    private String name;
    private Object[] nodes;
    private TreeNode target;
    private boolean isup;
    private TestNodeLinksGroup par;

    public TestNodeLinks(
        TestNodeLinksGroup utg, boolean isup, TreeNode node, String name, Object[] linked) {
      nodes = linked;
      this.name = name;
      target = node;
      this.isup = isup;
      par = utg;
    }

    public boolean isUp() {
      return isup;
    }

    public Object[] getChildren() {
      return nodes;
    }

    @Override
    public String toString() {
      if (isup) //TODO another way?
      return TraceabilityDataSource.getRevertName(
            Instantiator.getRootTree(target.getTreeDB()), removeBrack(name));
      return name;
    }

    public TreeNode getTarget() {
      return target;
    }

    public TestNodeLinksGroup getParent() {
      return par;
    }
  }

  @Override
  public Object[] getElements(Object inputElement) {
    return getChildren(inputElement);
  }

  private TreeNode getTreeNode(Object parentElement) {
    TreeNode parent = null;
    if (parentElement instanceof IProject) {
      IProject project = (IProject) parentElement;
      if (!project.isOpen()) {
        return null;
      }
      if (EclipseTracker.getTracker().getTree(project) != null
          && EclipseTracker.getTracker().getTree(project).getRootNode() != null // Root
      // node
      // can
      // be
      // not
      // created
      // yet
      ) {
        parent = EclipseTracker.getTracker().getTree(project).getRootNode();
      }
    }
    if (parentElement instanceof TreeNode) {
      parent = (TreeNode) parentElement;
    }
    if (parentElement instanceof PrefixAndTreeNode)
    {
      parent = ((PrefixAndTreeNode)parentElement).getNode();
    }
    if (parentElement instanceof TreeDB) {
      parent = ((TreeDB) parentElement).getRootNode();
    }
    return parent;
  }

  @SuppressWarnings("unchecked")
  @Override
  public Object[] getChildren(Object parentElement) {
    if (parentElement instanceof IProject) {
      IProject project = (IProject) parentElement;
      if (!project.isOpen()) {
        return new Object[0];
      }
    }
    if (parentElement instanceof TestNodeLinksGroup) {
      return ((TestNodeLinksGroup) parentElement).links;
    }
    if (parentElement instanceof TreeNode && parentElement == getTreeNode(viewer.getInput())) {
      if (rootInited < 1) {
        rootInited++;
        TreeNode parent = getTreeNode(parentElement);
        Map<String, ReferenceValue[]> links = parent.getLinksOrErrors(false);//new HashMap<>();
        Map<String, Object> rlinks = new HashMap<>();
        Map<String, Set<TreeNode>> revlinks =
            parent.getRevertLinks(false);
        if (revlinks != null)
          for (String s : revlinks.keySet())
            for (TreeNode nd : revlinks.get(s)) {
              if (!nd.isDisposed()) {
                if (!rlinks.containsKey(s)) rlinks.put(s, new ArrayList<>());
                ((ArrayList<Object>) rlinks.get(s)).add(nd);
              }
            }
        List<Object> tnl = new ArrayList<>();
        Set<String> fl = RequalityLCNF.getFilters(parent.getTreeDB());
        List<TestNodeLinksGroup> grps = new ArrayList<>();
        TestNodeLinksGroup utg =
            new TestNodeLinksGroup(parent, "Outcoming", false, tnl.toArray());
        if ((RequalityLCNF.getReferenceDirection() & 1) == 1) {
          Map<String, Object[]> alinks = new HashMap<>();
          List<String> names = new LinkedList<>();
          for (String s : links.keySet()) {
            String key = removeBrack(s);
            if (!key.startsWith("_") && (fl == null || fl.contains(key) || fl.isEmpty())) {
              names.add(key);
              //if (!alinks.containsKey(key)) alinks.put(key, new LinkedList<>());
              ReferenceValue[] rvs = links.get(s);
              Set<Object>alo = new HashSet<>();
              for(ReferenceValue rv:rvs)
                if(rv.hasErrorsOnResolve(parent.attrs()))
                  alo.add(rv.toString());
                else
                  alo.addAll(Arrays.asList(rv.getNodes(parent.attrs())));
              alinks.put(key, alo.toArray());
            }
          }
          for (String s : alinks.keySet()) {
            if(alinks.get(s).length>0){
              if(alinks.get(s).length>1)
                tnl.add(new TestNodeLinks(utg, false, parent, s, alinks.get(s)));
              else
                tnl.add(new PrefixAndTreeNode((TreeNode) alinks.get(s)[0],s+": "));
            }
          }
        }
        utg.links = tnl.toArray();
        if (utg.links.length > 0) grps.add(utg);
        tnl.clear();

        Map<String, Set<TreeNode>> dirt = parent.getTreeDB().getTermsStorage().getDirectTerms(parent);
        utg =
            new TestNodeLinksGroup(parent, "Terms", false, tnl.toArray());
        if ((RequalityLCNF.getReferenceDirection() & 1) == 1) {
          Map<String, Object[]> alinks = new HashMap<>();
          List<String> names = new LinkedList<>();
          for (String s : dirt.keySet()) {
            String key = removeBrack(s);
            if (!key.startsWith("_") && (fl == null || fl.contains(key) || fl.isEmpty())) {
              names.add(key);
              Set<Object>alo = new HashSet<>();
              for(TreeNode node:dirt.get(s))
                alo.add(node);
              alinks.put(key, alo.toArray());
            }
          }
          for (String s : alinks.keySet()) {
            if(alinks.get(s).length>0){
              if(alinks.get(s).length>1)
                tnl.add(new TestNodeLinks(utg, false, parent, s, alinks.get(s)));
              else
                tnl.add(new PrefixAndTreeNode(true, (TreeNode) alinks.get(s)[0],s));
            }
          }
        }
        utg.links = tnl.toArray();
        if (utg.links.length > 0) grps.add(utg);
        tnl.clear();

        utg = new TestNodeLinksGroup(parent, "Incoming", true, tnl.toArray());
        if ((RequalityLCNF.getReferenceDirection() & 2) == 2) {
          Map<String, List<Object>> alinks = new HashMap<>();
          List<String> names = new LinkedList<>();
          for (String s : rlinks.keySet()) {
            String key = removeBrack(s);
            if (!key.startsWith("_") && (fl == null || fl.contains(key) || fl.isEmpty())) {
              names.add(key);
              if (!alinks.containsKey(key)) alinks.put(key, new LinkedList<>());
              alinks.get(key).addAll(((List<?>) rlinks.get(s)));
            }
          }
          for (String s : alinks.keySet()) {
            if(alinks.get(s).size()>1)
              tnl.add(new TestNodeLinks(utg, true, parent, s, alinks.get(s).toArray()));
            else
              tnl.add(new PrefixAndTreeNode(false, (TreeNode) alinks.get(s).iterator().next(),s+": "));
          }
        }
        utg.links = tnl.toArray();
        if (utg.links.length > 0) grps.add(utg);
        tnl.clear();
        
        dirt = parent.getTreeDB().getTermsStorage().getRevertTerms(parent);
        utg =
            new TestNodeLinksGroup(parent, "Term referrers", false, tnl.toArray());
        if ((RequalityLCNF.getReferenceDirection() & 1) == 1) {
          Map<String, Object[]> alinks = new HashMap<>();
          List<String> names = new LinkedList<>();
          for (String s : dirt.keySet()) {
            String key = removeBrack(s);
            if (!key.startsWith("_") && (fl == null || fl.contains(key) || fl.isEmpty())) {
              names.add(key);
              Set<Object>alo = new HashSet<>();
              for(TreeNode node:dirt.get(s))
                alo.add(node);
              alinks.put(key, alo.toArray());
            }
          }
          for (String s : alinks.keySet()) {
            if(alinks.get(s).length>0){
              if(alinks.get(s).length>1)
                tnl.add(new TestNodeLinks(utg, false, parent, s, alinks.get(s)));
              else
                tnl.add(new PrefixAndTreeNode((TreeNode) alinks.get(s)[0],s+": "));
            }
          }
        }
        utg.links = tnl.toArray();
        if (utg.links.length > 0) grps.add(utg);
        tnl.clear();
        return grps.toArray();
      }
    } else if (parentElement instanceof TestNodeLinks)
      return ((TestNodeLinks) parentElement).getChildren();
    return new Object[0];
  }

  private static String removeBrack(String s) {
    int pos = s.indexOf("[");
    while (pos > -1) {
      int end = s.indexOf("]", pos + 1);
      if (end > -1) {
        s = s.substring(0, pos) + s.substring(end + 1);
        pos = s.indexOf("[", pos + 1);
      } else pos = s.indexOf("[", pos);
    }
    return s;
  }

  @Override
  public Object getParent(Object element) {
    if (element instanceof TestNodeLinks) {
      return ((TestNodeLinks) element).getParent();
    }
    if (element instanceof TreeNode) {
      TreeNode node = (TreeNode) element;
      if (node.isDisposed()) {
        return null;
      }
      if (node.getTreeDB().getRootNode().equals(node.getShownParent())) {
        return EclipseTracker.getTracker().getProject(node.getTreeDB());
      }
      return node.getShownParent();
    }
    if (element instanceof TestNodeLinksGroup) return ((TestNodeLinksGroup) element).getParent();
    return null;
  }

  @Override
  public boolean hasChildren(Object element) {
    return getChildren(element).length > 0;
  }
}
