/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.ui.progress.UIJob;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.ITreeNodeFilter;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.NodeNotFoundException;
import com.unitesk.requality.core.SortingSettings;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeChangeEvent;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.ui.ObjectRule;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;

public class RequalityContentProvider implements ITreeContentProvider {
  private TreeViewer viewer;
  private ArrayList<TreeDB> dbs = new ArrayList<>();
  private ArrayList<ITreeNodeFilter> filters = new ArrayList<>();
  private static ITreeNodeFilter sfilter;
  private boolean checkStatic;

  public RequalityContentProvider(boolean checkStatic) {
    this.checkStatic = checkStatic;
  }

  public RequalityContentProvider() {
    this(true);
  }

  public boolean removeFilter(ITreeNodeFilter filter) {
    return filters.remove(filter);
  }

  public boolean addFilter(ITreeNodeFilter filter) {
    return filters.add(filter);
  }

  public static boolean setStaticFilter(ITreeNodeFilter filter) {
    sfilter = filter;
    return true;
  }

  private ITreeChangeListener treeChangeListener =
      new DefaultTreeChangeListener() {
        private SortingSettings storedSortingSettings = null;
        private List<TreeNode> dirtyNodes = new ArrayList<>();

        private void guiRefresh(final TreeNode node) {
          if (dirtyNodes.contains(node) || viewer.getControl().isDisposed()) {
            return;
          }
          addToch(node);
        }

        @Override
        public void moved(TreeNode node, TreeNode oldParent, TreeNode newParent) {
          addToch(newParent);
          //guiRefresh(newParent);
          if (!oldParent.getUUId().equals(newParent.getUUId())) addToch(oldParent);
          addToch(node);
        }

        @Override
        public void deleted(TreeNode node) {
          if(node == null)
            return;
          addToch(node,node.getShownParent());
        }

        @Override
        public void changed(TreeNode node, ChangesSet<Attribute> what) {
          if(rootActions(node))
            return;
          boolean changeIt = false;//, changeParent = false;
          for (String w : what.getAll()) {
            if ((w.equals(TreeNode.ATTR_LABEL)|| RequalityCNF.isShown(w))) changeIt = true;
            //if (Attribute.affectsPosition(node, w)) changeParent = true;
            //if (changeParent) break;
          }
          //if (changeParent) guiRefresh(node.getShownParent());
          if (changeIt) addToch(node);
        }

        @Override
        public void created(TreeNode node) {
          addToch(node.getShownParent());
        }
        @Override
        public void childrenOrderChanged(TreeNode node) {
          addToch(node);
        }

        @Override
        public void changed(final List<TreeNodeChangeEvent> changes) {
            Set<TreeNode> toch = new HashSet<>();
            for (TreeNodeChangeEvent ch : changes) {
              switch (ch.getType()) {
                case MOVE:
                  {
                    try{
                      TreeNode node = ch.getTreeNodeTarget();
                      toch.add(node);
                      TreeNode old = (TreeNode) ch.getArguments()[1];
                      if (old != null)toch.add(old);
                      if (old == null ||!old.equals(node.getShownParentExc())) toch.add(node.getShownParentExc());
                    }
                    catch(NodeNotFoundException nfe){
                      
                    }
                    //moved(ch.getTreeNodeTarget(),(TreeNode)ch.getArguments()[1],(Boolean)ch.getArguments()[0]);
                  }
                  break;
  
                case CHILDREN_ORDER_CHANGE:{
                  toch.add(ch.getTreeNodeTarget());
                }break;
                case CREATE:
                  try{
                    TreeNode node = ch.getTreeNodeTarget().getShownParentExc();
                    toch.add(node);
                  }catch(NodeNotFoundException nfe){
                    
                  }
                  //created(ch.getTreeNodeTarget(), (Boolean)ch.getArguments()[0]);
                  break;
                case DELETE:
                  {
                    TreeNode nd = ch.getTreeNodeTarget();
                    if(ch.getArguments()[2]==Moment.BEFORE)
                      continue;
                    try{
                      toch.add(nd.getShownParentExc());
                    }catch(NodeNotFoundException e){//ignore late del
                    }
                    //TODO add update of parent
                    //toch.add();
                  } //deleted(ch.getNodeDescTarget(), (Boolean)ch.getArguments()[0]);
                  break;
                case CHANGE:
                  {
                    TreeNode node = ch.getTreeNodeTarget();
                    if(node.isRoot())
                      if (rootActions(node))
                        return;
                    boolean changeIt = false, changeParent = false;
                    @SuppressWarnings("unchecked")
                    ChangesSet<Attribute> what = (ChangesSet<Attribute>) ch.getArguments()[1];
                    Map<String, Attribute> newA = what.getNewMap();
                    for (Entry<String, Attribute> en : newA.entrySet()) {
                      if ((en.getKey().equals(TreeNode.ATTR_LABEL)|| RequalityCNF.isShown(en.getKey()))) changeIt = true;
                      if (Attribute.affectsPosition(node, en.getKey())) changeParent = true;
                      if (changeParent) break;
                    }
                    try{
                      if (changeParent) toch.add(node.getShownParentExc());
                      else if (changeIt) toch.add(node);
                    }catch(NodeNotFoundException nfe){
                      
                    }
                  }
                  break;
                default:
                  break;
              }
            }
            addToch(toch.toArray(new TreeNode[0]));
        }
        protected boolean rootActions(TreeNode node) {
          if (node.getUUId().equals(node.getTreeDB().getRootNode().getUUId())) {
            //refresh root
            if (!node.getTreeDB().getSortingSettings().equals(storedSortingSettings)) {
              storedSortingSettings = node.getTreeDB().getSortingSettings();
              viewer
                  .getControl()
                  .getDisplay()
                  .asyncExec(
                      new Runnable() {
                        @Override
                        public void run() {
                          viewer.refresh();
                        }
                      });
              return true;
            }
          }
          return false;
        }
  };
  static TreeNode extractNode(Object o) {
    if(o instanceof TreeNode)
      return(TreeNode)o;
    if(o instanceof IProject)
    {
      TreeDB db = EclipseTracker.getTracker().getTree((IProject)o);
      if(db!=null)
        return db.getRootNode();
    }
    return null;
  }

  private ITreeStateListener stateListener =
      new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
          dbs.add(db);
          db.addGUIListener(treeChangeListener);
        }

        @Override
        public void delete(IProject proj) {
          TreeDB db = EclipseTracker.getTracker().getTree(proj);
          if(db!=null){
            close(db);
          }
        }

        @Override
        public void close(TreeDB db) {
          db.removeGUIListener(treeChangeListener);
          dbs.remove(db);
        }
      };
      

  private boolean filterNode(TreeNode node) {
    for (ITreeNodeFilter filter : filters) {
      if (!filter.select(node)) {
        return false;
      }
    }
    return staticFilterNode(node);
  }

  private boolean staticFilterNode(TreeNode node) {
    if (checkStatic) return node != null && sfilter.select(node);
    return true;
  }

  private long lup = -1;
  private UIJob uj = null;
  private ObjectRule viewerRule;
  static int LIMIT = 25;
  static int LIMIT2 = 100;
  private Set<TreeNode>aset = new HashSet<>();
  private ReadWriteLock rwl = new ReentrantReadWriteLock(false);
  private Lock wl = rwl.writeLock();
  private Lock rl = rwl.readLock();
  private void addToch(TreeNode... as) {
    wl.lock();
    aset.addAll(Arrays.asList(as));
    try{
      if(uj == null){
        lup = System.currentTimeMillis();
        uj = new UIJob("update reqcnf"){
          
          @Override
          public IStatus runInUIThread(IProgressMonitor monitor) {
            if(viewer.getControl().isDisposed())
              return Status.CANCEL_STATUS;
            boolean wlo=false,rlo=true, irl = true;
            try{
              Lock wl2 = rwl.writeLock();
              Lock rl = rwl.readLock();
              Instantiator.readLock();
              irl = true;
              // TODO Auto-generated method stub
               rl.lock();
               long d = (System.currentTimeMillis()-lup);
               boolean first = true;
               int cnt = 0, max = aset.size();
               while((d<LIMIT2||first)&&(cnt!=max||d>RequalityContentProvider.LIMIT)){
                 if(viewer.getControl().isDisposed())
                   return Status.CANCEL_STATUS;
                first = false;
                cnt = 0;
                if(aset.size()>0){
                  Set<TreeNode> toch = new HashSet<>(aset);
                  rl.unlock();
                  rlo = false;
                  Set<TreeNode> fch = new HashSet<>();
                  Set<TreeNode> exps = new HashSet<>();
                  for(Object o : viewer.getExpandedElements()){
                    TreeNode on = extractNode(o);
                    if(on != null){
                      exps.add(on);
                      for(TreeNode onc:on.getShownChildren())
                        exps.add(onc);
                    }
                  }
                  for(TreeNode node:toch){
                    if(node == null)
                      continue;
                    for(TreeNode onc:exps)
                      if(onc.getUUId().equals(node.getUUId())){
                        fch.add(node);
                        break;
                      }else
                        try{
                          if(node.getShownParentExc()!=null && 
                          onc.getUUId().equals(node.getShownParentExc().getUUId())){
                            TreeNode sp = node.getShownParent();
                            if(sp!=null){
                              fch.add(node);
                              break;
                            }
                          }
                        }
                        catch(NodeNotFoundException nde){
                          
                        }  
                    cnt++;
                  }
                  final Set<TreeNode> ftoch = fch;
                  if (viewer.getControl().isDisposed()) {
                    return Status.CANCEL_STATUS;
                  }
                  //TODO null update? mb bug
                  //Object o = viewer.getInput();
                  //RLogger.logInfo(node.toString());
                  if (ftoch.contains(null))
                    viewer.refresh(true);
                  else
                  {
                    viewer.getControl().setRedraw(false);
                    try{
                      wl2.lock();
                      wlo = true;
                      for (TreeNode node : ftoch) {
                        aset.remove(node);
                        max--;
                        cnt--;
                        if (viewer.getControl().isDisposed()) {
                          return Status.CANCEL_STATUS;
                        }
                        if (node != null && (node.equals(node.getTreeDB().getRootNode()))) {
                          TreeNode nde = node;
                          if (EclipseTracker.getResourceByNode(nde) != null)
                            viewer.refresh(
                                EclipseTracker.getResourceByNode(nde).getProject(), true);
                          return Status.OK_STATUS;
                        }
                        if (node != null && !node.isDisposed()) viewer.refresh(node, true);
                      }
                    }
                    finally{
                      viewer.getControl().setRedraw(true);
                      wl2.unlock();
                      wlo = false;
                    }
                  }
                  rl.lock();
                  rlo = true;
                  max = aset.size();
                }
                d = (System.currentTimeMillis()-lup);
              }
            }finally{
              if(rlo)
                rl.unlock();
              if(wlo)
                wl.unlock();
              uj = null;
              if(irl)
                Instantiator.readUnlock();
            }
          
            return Status.OK_STATUS;
          }
        };
       uj.setRule(viewerRule);
       uj.setPriority(Job.LONG);
       uj.schedule();
      }
    }
    finally{
      wl.unlock();
    }
  };
  @Override
  public void dispose() {
    RequalityPlugin.getTreesTracker().removeListener(stateListener);
  }

  @Override
  public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
    RequalityPlugin.getTreesTracker().removeListener(stateListener);
    if (!(viewer instanceof TreeViewer)) {
      throw new IllegalArgumentException();
    }
    this.viewer = (TreeViewer) viewer;
    this.viewerRule = new ObjectRule(viewer);
    for (TreeDB db : dbs) {
      db.removeGUIListener(treeChangeListener);
    }
    dbs.clear();

    RequalityPlugin.getTreesTracker().addListener(stateListener);
    if (newInput == null) {
      return;
    }
    if (newInput instanceof TreeNode) {
      dbs.add(((TreeNode) newInput).getTreeDB());
    } else if (newInput instanceof IWorkspaceRoot) {
      dbs.addAll(Arrays.asList(RequalityPlugin.getTreesTracker().getTrees()));
    } else if (newInput instanceof Object[]) {
      dbs.add(((TreeNode) ((Object[]) newInput)[0]).getTreeDB());
    } else if (newInput instanceof TreeDB) {
      dbs.add((TreeDB) newInput);
    } else {
      throw new IllegalArgumentException();
    }
    for (TreeDB db : dbs) {
      db.addGUIListener(treeChangeListener);
    }
    
    viewer
        .getControl()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                //TODO if control is disposing, old code will throw exception
                if ((RequalityContentProvider.this.viewer.getControl() != null)
                    && !RequalityContentProvider.this.viewer.getControl().isDisposed())
                  RequalityContentProvider.this.viewer.refresh();
              }
            });
  }

  @Override
  public Object[] getElements(Object inputElement) {
    if (inputElement instanceof IWorkspaceRoot) {
      return ResourcesPlugin.getWorkspace().getRoot().getProjects();
    }
    if (inputElement instanceof TreeNode) {
      TreeNode nodeInput = (TreeNode) inputElement;
      ArrayList<TreeNode> result = new ArrayList<>();
      for (TreeNode node : nodeInput.getSortedShownChildren())
        if (node != null && filterNode(node)) {
          result.add(node);
        }
      return result.toArray(new TreeNode[0]);
    }
    if (inputElement instanceof TreeDB) {
      return new Object[] {EclipseTracker.getTracker().getProject((TreeDB) inputElement)};
    }
    if (inputElement instanceof Object[]) {
      return (Object[]) inputElement;
    }
    return new Object[0];
  }

  private TreeNode getTreeNode(Object parentElement) {
    TreeNode parent = null;
    if (parentElement instanceof IProject) {
      IProject project = (IProject) parentElement;
      if (!project.isOpen()) {
        return null;
      }
      if (EclipseTracker.getTracker().getTree(project) != null
          && EclipseTracker.getTracker().getTree(project).getRootNode() != null // Root
      // node
      // can
      // be
      // not
      // created
      // yet
      ) {
        parent = EclipseTracker.getTracker().getTree(project).getRootNode();
      }
    }
    if (parentElement instanceof TreeNode) {
      parent = (TreeNode) parentElement;
    }
    if (parentElement instanceof TreeDB) {
      parent = ((TreeDB) parentElement).getRootNode();
    }
    return parent;
  }

  @Override
  public Object[] getChildren(Object parentElement) {
    if (parentElement instanceof IProject) {
      IProject project = (IProject) parentElement;
      if (!project.isOpen()) {
        return new Object[0];
      }
    }
    TreeNode parent = getTreeNode(parentElement);
    if (parent == null) return new Object[0];
    ArrayList<TreeNode> result = new ArrayList<>();
    for (TreeNode node : parent.getSortedShownChildren()) {
      if (filterNode(node)) {
        result.add(node);
      }
    }
    return result.toArray(new TreeNode[0]);
  }

  @Override
  public Object getParent(Object element) {
    if (element instanceof TreeNode) {
      TreeNode node = (TreeNode) element;
      if (node.isDisposed()) {
        return null;
      }
      if (node.getTreeDB().getRootNode().equals(node.getShownParent())) {
        return EclipseTracker.getTracker().getProject(node.getTreeDB());
      }
      return node.getShownParent();
    }
    return null;
  }

  @Override
  public boolean hasChildren(Object element) {
    TreeNode node = getTreeNode(element);
    if (node == null) return false;
    List<UUID> chi = node.getWorkChildrenUUIDs(); //TODO loaded?
    if (chi.size() == 0) return false;
    TreeDB db = node.getTreeDB();
    int x = chi.size();
    for (UUID o : chi) {
      try {
        TreeNode nde = db.getCachedNodeOrVirt(o);
        if (nde != null && !staticFilterNode(nde)) x--;
      } catch (RuntimeException re) {
        re.printStackTrace();
        x--;
      }
    }
    return x > 0;
  }
}
