package com.unitesk.requality.eclipse.views.actions;

import java.util.List;
import java.util.Set;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.window.Window;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.CompoundContributionItem;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.GetterAttributeNames;
import com.unitesk.requality.core.operations.mass.AbstractMassOperation;
import com.unitesk.requality.core.operations.mass.AbstractNodeOperation;
import com.unitesk.requality.core.operations.mass.LibreOfficeAfterProcessing;
import com.unitesk.requality.core.operations.mass.MassAttributeTypeChange;
import com.unitesk.requality.core.operations.mass.MassIDRenames;
import com.unitesk.requality.core.operations.mass.MassKindFixer;
import com.unitesk.requality.core.operations.mass.MassRefFixer;
import com.unitesk.requality.core.operations.mass.MassTNHeaderExplode;
import com.unitesk.requality.core.operations.mass.TermsHandler;
import com.unitesk.requality.core.operations.mass.TestMassOps;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.ui.dialogs.EnumDefExtractDialog;
import com.unitesk.requality.eclipse.ui.dialogs.MassAttributeRetype;

public class BatchDynamics extends CompoundContributionItem {

  public BatchDynamics() {
    super();
  }

  public BatchDynamics(String id) {
    super(id);
  }

  @Override
  protected IContributionItem[] getContributionItems() {
    List<TreeNode> sels = SelectionManager.getOne().getSelectedNodes();
    if (sels == null || sels.size() != 1) return new IContributionItem[0];
    IContributionItem[] items =
        new IContributionItem[] {
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Mass attribute type change") {

                @Override
                public AbstractMassOperation initBatchProcessor(TreeNode node) {
                  Set<String> names =
                      GetterAttributeNames.getFromRoot(node, EnumDefExtractDialog.filter);

                  MassAttributeRetype mar =
                      new MassAttributeRetype(
                          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                          node.getTreeDB(),
                          names);
                  if (mar.open() == Window.OK) {
                    Object[] res = mar.getResult();
                    String name = (String) res[0];
                    AttributeType at = (AttributeType) res[1]; //ENUM error?
                    String newName = name;
                    if (res[2] != null) newName = (String) res[2];
                    MassAttributeTypeChange match = new MassAttributeTypeChange(name, newName, at);
                    return match;
                  }

                  return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Mass incorrect struct text node remove") {

                @Override
                public AbstractMassOperation initBatchProcessor(TreeNode node) {
                  //new DeepFirstTreeWalker(node,true,true,false);
                  MassTNHeaderExplode mtn = new MassTNHeaderExplode();
                  return mtn;
                  //return null;

                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Mass numerical id fixer") {

                @Override
                public AbstractNodeOperation initBatchProcessor(TreeNode node) {
                  MassIDRenames mir = new MassIDRenames();
                  return mir;
                  //return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Mass 'Вид' fixer") {

                @Override
                public AbstractNodeOperation initBatchProcessor(TreeNode node) {
                  MassKindFixer mir = new MassKindFixer();
                  return mir;
                  //return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Mass reference fixer") {

                @Override
                public AbstractNodeOperation initBatchProcessor(TreeNode node) {
                  MassRefFixer mir = new MassRefFixer();
                  return mir;
                  //return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(
                  sels.get(0), "After-import processing for libreoffice html documents") {

                @Override
                public AbstractNodeOperation initBatchProcessor(TreeNode node) {
                  LibreOfficeAfterProcessing mir = new LibreOfficeAfterProcessing();
                  return mir;
                  //return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(
                  sels.get(0), "Terms handler") {

                @Override
                public AbstractNodeOperation initBatchProcessor(TreeNode node) {
                  TermsHandler mir = new TermsHandler();
                  return mir;
                  //return null;
                }
              }),
          new ActionContributionItem(
              new SingleBatchProcessor(sels.get(0), "Test op") {

                @Override
                public AbstractMassOperation initBatchProcessor(TreeNode node) {
                  //new DeepFirstTreeWalker(node,true,true,false);
                  TestMassOps ops = new TestMassOps();
                  return ops;
                  //return null;
                }
              })
        };
    return items;
  }
}
