/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.actions;

import java.util.List;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.ITreeNodeFilter;
import com.unitesk.requality.core.ITreeNodesFilter;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.ui.cnf.CNFAction;

public class NodeProcessingAction extends CNFAction {
  private ITreeNodeFilter filter;
  private ITreeNodesFilter enfilter;
  private ITreeNodeFilter actions;
  private String transactionName;

  public NodeProcessingAction(
      String name, String transactionName, ITreeNodeFilter filter, ITreeNodeFilter actions) {
    this(name, transactionName, filter, null, actions);
  }

  public NodeProcessingAction(
      String name, ITreeNodeFilter filter, ITreeNodesFilter enfilter, ITreeNodeFilter actions) {
    this(name, name, filter, enfilter, actions);
  }

  public NodeProcessingAction(
      String name,
      String transactionName,
      ITreeNodeFilter filter,
      ITreeNodesFilter enfilter,
      ITreeNodeFilter actions) {
    // TODO:
    // super(provider);
    this.enfilter = enfilter;
    this.filter = filter;
    this.actions = actions;
    this.transactionName = transactionName;
    setText(name);
  }

  @Override
  public void run() {
    final TreeDB db = getTreeDB();
    if (db == null) return;
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (transactionName != null)
      if (!db.startTransaction(transactionName, true)) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getDisplay().getActiveShell(),
            "Processing Error",
            "Some actions are already being performed");
        return;
      }
    for (final TreeNode node : nodes) {
      //            db.getNode(Requirement
      //                    .getTypeRootQId(),Path.QID);
      try {
        if (!filter.select(node)) continue;
        try {
          if (!actions.select(node)) {
            if (transactionName != null) db.rollback();
            return;
          }
        } catch (final Exception e) {
          if (transactionName != null) db.rollback();
          PlatformUI.getWorkbench()
              .getDisplay()
              .asyncExec(
                  new Runnable() {
                    public void run() {
                      //			            	PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                      MessageDialog.openError(
                          PlatformUI.getWorkbench().getDisplay().getActiveShell(),
                          "Processing Error",
                          e.getClass().getName() + ": " + e.getMessage());
                      e.printStackTrace();
                    }
                  });
          return;
        }
      } catch (final Exception e) {
        PlatformUI.getWorkbench()
            .getDisplay()
            .asyncExec(
                new Runnable() {
                  public void run() {
                    //            		    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();

                    MessageDialog.openError(
                        PlatformUI.getWorkbench().getDisplay().getActiveShell(),
                        "Processing Error",
                        e.getClass().getName() + ": " + e.getMessage());
                    e.printStackTrace();
                  }
                });
      }
    }
    if (transactionName != null) db.commit();
  }

  private TreeDB getTreeDB() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();

    if (nodes != null) {
      return nodes.get(0).getTreeDB();
    }
    return null;
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null || nodes.size() < 1) return false;
    return (enfilter == null || enfilter.select(nodes).size() > 0);
  }
}
