package com.unitesk.requality.eclipse.views.atypeeditors;

import java.util.List;
import java.util.regex.Pattern;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.viewers.DialogCellEditor;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.eclipse.ui.cnf.RequalitySorter;
import com.unitesk.requality.eclipse.views.RequalityContentProvider;
import com.unitesk.requality.eclipse.views.atypeeditors.ListValueEditor.ValueItem;

public class ReferenceValueEditor extends DialogCellEditor {
  private boolean updating;
  private TreeNode node;

  public ReferenceValueEditor(TreeNode node, Composite table, int style) {
    super(table, style);
    this.node = node;
  }

  @Override
  protected Object openDialogBox(Control cellEditorWindow) {
    Object o = ((Table) cellEditorWindow.getParent()).getSelection()[0].getData();
    ReferenceValue val = null;
    if (o instanceof ReferenceValue)
      val = (ReferenceValue)o;
    else
    if (o instanceof Attribute)
      val = (ReferenceValue) ((Attribute) o).getValue(); //TODO or not raw
    else{
      ValueItem vi = ((ValueItem) o);
      if(vi != null && vi.getValue()!= null)
        val =  (ReferenceValue) vi.getValue();
      else
        val = new ReferenceValue(AttributeType.REFERENCE, "", null);
    }
    LinkDialog ld = new LinkDialog(cellEditorWindow.getShell(), val);
    if (ld.open() == Dialog.OK) {
      return new ReferenceValue(val.getType(),ld.getResult(), val.getKey());
    }
    return null;
  }

  private class LinkDialog extends Dialog {
    private Text xpath;
    private String xpathvalue;
    private Text err;
    private TreeViewer nodeTree;
    private int choice;
    private Pattern
        pdirect =
            Pattern.compile(
                "\\w\\w\\w\\w\\w\\w\\w\\w-\\w\\w\\w\\w-\\w\\w\\w\\w-\\w\\w\\w\\w-\\w\\w\\w\\w\\w\\w\\w\\w\\w\\w\\w\\w"),
        ppath = Pattern.compile("(\\/Requirements(?:\\/(?:[\\wа-яА-Я])*)*)");

    public LinkDialog(Shell shell, ReferenceValue string) {
      super(shell);
      xpathvalue = string.getRaw().toString();
      setShellStyle(getShellStyle() | SWT.RESIZE);
    }

    public Object getResult() {
      return xpathvalue;
    }

    private Button o1;
    private Button o2;
    private Button o3;

    @Override
    protected Control createDialogArea(Composite parent) {
      Composite container = createClientContainer(parent, 1);
      container.setLayoutData(new GridData(GridData.FILL_BOTH));
      Composite topPanel = new Composite(container, SWT.NONE);
      //Flow fl = new FillLayout();
      GridLayout gl = new GridLayout(1, false);
      gl.marginHeight = gl.marginWidth = 0;

      topPanel.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
      topPanel.setLayout(gl);

      Composite checkcom = new Composite(topPanel, SWT.NONE);
      checkcom.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
      checkcom.setLayout(new GridLayout(3, false));
      o1 = new Button(checkcom, SWT.RADIO);
      o2 = new Button(checkcom, SWT.RADIO);
      o3 = new Button(checkcom, SWT.RADIO);
      //final Button o3 = new Button(checkcom, SWT.RADIO);
      SelectionListener check =
          new SelectionListener() {

            @Override
            public void widgetSelected(SelectionEvent e) {
              choice = 0;
              if (o1.getSelection()) {
                choice = 1;
              }
              if (o2.getSelection()) {
                choice = 2;
              }
              if (o3.getSelection()) {
                choice = 3;
              }
              updateXpath();
            }

            @Override
            public void widgetDefaultSelected(SelectionEvent e) {
              // TODO Auto-generated method stub

            }
          };

      o1.setText("By element");
      o1.addSelectionListener(check);
      o2.setText("By path");
      o2.addSelectionListener(check);
      o3.setText("By UVName");
      o3.addSelectionListener(check);
      //o3.setText("Manual");

      final Button b = new Button(topPanel, SWT.CHECK);
      b.setText("Advanced settings");

      final Composite c = new Composite(topPanel, SWT.NONE);
      c.setVisible(false);
      c.setLayoutData(new GridData(0, 0));
      c.setLayout(new GridLayout(1, true));

      b.addSelectionListener(
          new SelectionListener() {

            @Override
            public void widgetSelected(SelectionEvent e) {
              if (b.getSelection()) c.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
              else c.setLayoutData(new GridData(0, 0));
              c.setVisible(b.getSelection());
              c.getShell().pack();
            }

            @Override
            public void widgetDefaultSelected(SelectionEvent e) {
              // TODO Auto-generated method stub

            }
          });

      Label valueLabel = new Label(c, SWT.WRAP);
      valueLabel.setText("Expression:");
      valueLabel.setLayoutData(new GridData(GridData.BEGINNING, GridData.CENTER, true, true));
      xpath = new Text(c, SWT.NONE);
      xpath.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
      valueLabel = new Label(c, SWT.NONE);
      valueLabel.setText("Error log:");
      valueLabel.setLayoutData(new GridData(GridData.BEGINNING, GridData.CENTER, true, true));

      err = new Text(c, SWT.MULTI | SWT.BORDER | SWT.WRAP | SWT.V_SCROLL | SWT.READ_ONLY);
      GridData gd = new GridData(GridData.FILL_HORIZONTAL);
      gd.heightHint = 100;
      err.setLayoutData(gd);
      //   	        final Table basetable = new Table(container, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);

      nodeTree = new TreeViewer(container);
      GridData ld = new GridData(GridData.FILL_BOTH);
      ld.heightHint = 400;
      ld.widthHint = 300;
      nodeTree.getControl().setLayoutData(ld);
      nodeTree.setLabelProvider(WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider());
      nodeTree.setComparator(new RequalitySorter());
      nodeTree.setAutoExpandLevel(2);
      nodeTree.setComparator(new RequalitySorter());
      nodeTree.setContentProvider(new RequalityContentProvider(true));
      nodeTree.setInput(node.getTreeDB().getRootNode());
      nodeTree.addSelectionChangedListener(
          new ISelectionChangedListener() {

            @Override
            public void selectionChanged(SelectionChangedEvent event) {
              if (updating) return;
              updating = true;
              if (choice == 0) {
                choice = 1;
                o1.setSelection(true);
              }
              updateXpath();
              updating = false;
            }
          });
      xpath.addModifyListener(
          new ModifyListener() {

            @Override
            public void modifyText(ModifyEvent e) {
              xpathvalue = xpath.getText().trim();
              if (updating) return;
              choice = 0;
              err.setText("");
              //o3.setSelection(true);
              updating = true;
              try {
                int pos = xpathvalue.indexOf("::");
                if (pos > -1) {
                  String name = xpathvalue.substring(0, pos).trim();
                  if (name != null) {
                    TreeDB db = ResManager.getManager().getTreeDB(name);
                    if (db == null) err.setText("Project '" + name + "' is not found!");
                  }
                }
                List<TreeNode> nde =
                    node.getTreeDB()
                        .getLinkResolver()
                        .getNotDisposedResolvePath(
                            node,
                            null,
                            node.getTreeDB(),
                            xpathvalue,
                            Path.QID | Path.UV | Path.UUID | Path.XPATH);
                if (nde != null && nde.size() > 0)
                  nodeTree.setSelection(new StructuredSelection(nde));
                else nodeTree.setSelection(new StructuredSelection());
              } catch (Exception ee) {
                err.setText(ee.toString());
              }

              o1.setSelection(false);
              if (pdirect.matcher(xpathvalue).matches()) {
                o1.setSelection(true);
                choice = 1;
              }
              o2.setSelection(false);
              if (ppath.matcher(xpathvalue).matches()) {
                o2.setSelection(true);
                choice = 2;
              }
              o3.setSelection(false);
              try {
                if (choice != 2
                    && node.getTreeDB()
                            .getNode(xpathvalue, Path.QID | Path.UV | Path.UUID | Path.XPATH)
                        != null) {
                  o3.setSelection(true);
                  choice = 3;
                }
              } catch (Exception ex) {
              }
              updating = false;
            }
          });
      if (xpathvalue != null) xpath.setText(xpathvalue);

      //updateXpath();
      //nodeTree.addSelectionChangedListener(new NodeSelectionListener());

      //         nodeTree.getTree().addMouseListener(mouseList);
      //       nodeTree.getTree().addKeyListener(keyList);
      getShell().setText("Link editor");
      this.getShell()
          .setImage(
              PlatformUI.getWorkbench().getSharedImages().getImage(ISharedImages.IMG_OBJ_ELEMENT));
      return container;
    }

    protected void updateXpath() {
      if (nodeTree.getSelection().isEmpty() || choice == 0) return;
      updating = true;
      TreeNode node2 = (TreeNode) ((StructuredSelection) nodeTree.getSelection()).getFirstElement();
      if (choice == 1) xpath.setText(node2.getUUId().toString());
      else if (choice == 2) {
        //					TreeNode nde = node2;
        //					String path = "";
        //					while(nde!=null && !"TreeNode".equals(nde.getType())){
        //						path = nde.getType()+"[@id=\""+nde.getId()+"\"]/"+path;
        //						nde = nde.getWorkParent();
        //					}
        xpath.setText(node2.getQualifiedId());
      } else xpath.setText(node2.getUserVisibleName(true));
      updating = false;
    }

    private Composite createClientContainer(Composite parent, int span) {
      Composite container = new Composite(parent, SWT.NONE);
      container.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
      GridLayout layout = new GridLayout();
      layout.marginWidth = layout.marginHeight = 10;
      layout.verticalSpacing = 2;
      layout.numColumns = span;
      container.setLayout(layout);
      return container;
    }
  }
}
