package com.unitesk.requality.eclipse.views.compare;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.UUID;
import org.eclipse.compare.CompareConfiguration;
import org.eclipse.compare.CompareUI;
import org.eclipse.compare.IStreamContentAccessor;
import org.eclipse.compare.ResourceNode;
import org.eclipse.compare.contentmergeviewer.ContentMergeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.json.JSONException;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.MemoryStorage;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeModification;
import com.unitesk.requality.core.descriptors.NodeTypeDescriptor;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.eclipse.storageformatversion.LegalicyNodeMapper;
import com.unitesk.requality.eclipse.views.properties.ElementViewPanel;

public class TreeNodeCompareViewer extends ContentMergeViewer {
  public TreeNodeCompareViewer(Composite parent, int style, CompareConfiguration configuration) {
    super(style, ResourceBundle.getBundle("org.eclipse.compare.contentmergeviewer.TextMergeViewerResources"), configuration);
    
    buildControl(parent);
    String title= "Requality TreeNode Compare"; //$NON-NLS-1$
    getControl().setData(CompareUI.COMPARE_VIEWER_TITLE, title);
  }
  public TreeNodeCompareViewer(Composite parent, CompareConfiguration configuration) {
    this(parent, SWT.NULL, configuration);
  }
  //org.eclipse.compare.contentmergeviewer.TextMergeViewer;
  Composite canc, lc,rc;
  ElementViewPanel left, right;
  Label leftv, rightv;
  @Override
  protected void createControls(Composite composite) {
    canc = new Composite(composite, SWT.NO_FOCUS);

    Label l = new Label(canc, SWT.NONE);
    l.setText("Test");
    //canc.setLayout(new GridLayout(2,false));
    lc = new Composite(composite, SWT.NO_FOCUS);
    lc.setLayout(new GridLayout(1,false));
    leftv = new Label(lc, SWT.NONE);
    leftv.setLayoutData(new GridData(GridData.GRAB_HORIZONTAL));
    left = new ElementViewPanel(lc, 1, true);
    rc = new Composite(composite, SWT.NO_FOCUS);
    rc.setLayout(new GridLayout(1,false));
    rightv = new Label(rc, SWT.NONE);
    rightv.setLayoutData(new GridData(GridData.GRAB_HORIZONTAL));
    right = new ElementViewPanel(rc, 1, true);
    
  }

  @Override
  protected void handleResizeAncestor(int x, int y, int width, int height) {
    if (width > 0) {
      canc.setVisible(true);
      canc.setBounds(x, y, width, height);
    } else {
      canc.setVisible(false);
    }
  }

  @Override
  protected void handleResizeLeftRight(
      int x, int y, int leftWidth, int centerWidth, int rightWidth, int height) {
    lc.setBounds(x, y, leftWidth, height);
    rc.setBounds(x+leftWidth+centerWidth, y, rightWidth, height);
  }

  @Override
  protected void updateContent(Object ancestor, Object left, Object right) {
    leftv.setText("Version: NONE");
    rightv.setText("Version: NONE");
    this.left.resetColors();
    this.right.resetColors();
    IStreamContentAccessor lte = (IStreamContentAccessor)left;
    String name1 = gatherNameNoJson(left), name2 = gatherNameNoJson(right);
    IStreamContentAccessor rte = (IStreamContentAccessor)right;
    String lc = null;
    if(lte!=null)
      lc = ResourceStorage.readFile(lte);
    String rc = null;
    if(rte!=null)
      rc = ResourceStorage.readFile(rte);
    TreeDB db = TreeDB.load(new MemoryStorage());
    Map<String, NodeTypeDescriptor> nts = RequalityPlugin.getAdditionalTypes();
    for (String name : RequalityPlugin.getAdditionalTypes().keySet())
      db.addNodeType(name, nts.get(name).getClazz());
    TreeNode l = null,r = null;
    this.left.setVisible(false);
    if(lte!=null)
      try {
        l = loadTreeNode(db, lc, name1);
        this.left.setVisible(true);
        this.left.setTreeNode(l);
        leftv.setText("Version: "+l.getNodeVersion());
      } catch (JSONException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    this.right.setVisible(false);
    if(rte!=null)
      try {
        r = loadTreeNode(db, rc, name2);
        this.right.setVisible(true);
        this.right.setTreeNode(r);
        rightv.setText("Version: "+r.getNodeVersion());
      } catch (JSONException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    markDifference(l, r);
  }

  private void markDifference(TreeNode l, TreeNode r) {
    if(l == null || r == null)
      return;
    markIfChange(TreeNode.ATTR_ID, l.getId(), r.getId());
    Set<String> keys = new HashSet<>();
    keys.addAll(l.attrs().getAllAttributeKeys());
    keys.addAll(r.attrs().getAllAttributeKeys());
    for(String s:keys){
      Object o1 = l.attrs().getAttributeValue(s),
          o2 = r.attrs().getAttributeValue(s);
      markIfChange(s, o1, o2);
    }
  }
  private void markIfChange(String k, Object v1, Object v2) {
    if(v1 == null)
      if(v2 != null)
        markChange(k,null,AttributeModification.REMOVE);
      else
        return;
    else
      if(v2==null)
        markChange(k,AttributeModification.ADD,null);
      else
        if(!v1.equals(v2))
          markChange(k, AttributeModification.MODIFY, AttributeModification.MODIFY);
  }
  private void markChange(String k, AttributeModification add, AttributeModification add2) {
    if(add != null)
      left.markAttribute(k, add);
    if(add2 != null)
      right.markAttribute(k, add2);
  }
  
  private String gatherNameNoJson(Object o) {
    String n = gatherName(o);
    if(n.toLowerCase().endsWith(".json"))
      return n.substring(0,n.length()-5);
    return n;
  }
  private String gatherName(Object o) {
    if(o instanceof ResourceNode)
      return((ResourceNode)o).getName();
    if(o == null)
      return "";
    Method m;
    try {
      m = o.getClass().getMethod("getName");
      if(m!=null)
        return (String) m.invoke(o);
    } catch (NoSuchMethodException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (SecurityException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IllegalArgumentException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (InvocationTargetException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return o.toString();
  }
  private TreeNode loadTreeNode(TreeDB db, String lc2,String id) throws JSONException {
    UUID uid = NodeMapper.getUUId(lc2);
    Map<String, AttributeDefinition> now = NodeMapper.getAttributes(uid, lc2);
    if(now == null || now.isEmpty())
      now = LegalicyNodeMapper.getAttributes(uid, lc2);
    if(now == null)
      return null;
    now.put(TreeNode.ATTR_ID, TreeNode.formIdAttribute(uid, id));
    TreeNode tn = null;
    try {
      tn = db.getNodeType(now.get(TreeNode.ATTR_TYPE).getRawValue().getValue().toString()).getTypeClass().getConstructor(TreeDB.class, UUID.class)
      .newInstance(db,uid);
    } catch (InstantiationException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IllegalAccessException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (IllegalArgumentException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (InvocationTargetException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (NoSuchMethodException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (SecurityException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    if(tn == null)
      return null;
    tn.attrs().setChanged(new ChangesSet<>(new HashSet<>(now.keySet()),
        null,null,null,now,null));
    TreeDB.addAttrsRec(tn.attrs());
    return tn;
  }
  @Override
  protected void copy(boolean leftToRight) {
    // TODO Auto-generated method stub
    
  }

  @Override
  protected byte[] getContents(boolean left) {
    return null;
  }
}
