/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import org.eclipse.jface.resource.FontRegistry;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextInputListener;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.tabbed.AbstractPropertySection;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.LinkResolver;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.VersionsStorage;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;

public class DebugPropertySection extends AbstractPropertySection {
  private TextViewer source;
  private FontRegistry fontRegistry;
  private TabbedPropertySheetPage page;
  private Lock l = new ReentrantLock();

  public DebugPropertySection() {
    super();
    l.lock();
    Display.getDefault()
        .syncExec(
            new Runnable() {

              @Override
              public void run() {
                try{
                  fontRegistry = new FontRegistry(PlatformUI.getWorkbench().getDisplay());
                  fontRegistry.put(
                    "debug", new FontData[] {new FontData("Monospaced", 12, SWT.NORMAL)});
                }
                finally{
                  l.unlock();
                }
              }
            });
  }

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    l.lock();
    try{
      super.createControls(parent, aTabbedPropertySheetPage);
      page = aTabbedPropertySheetPage;
      parent.setLayout(new FillLayout());
      source = new TextViewer(parent, SWT.MULTI | SWT.READ_ONLY);
      source.getControl().setFont(fontRegistry.get("debug"));
      source.addTextInputListener(
          new ITextInputListener() {
            @Override
            public void inputDocumentChanged(IDocument oldInput, IDocument newInput) {
              if (source.getControl().isDisposed()) return;
              source.getControl().pack();
              source.getControl().redraw();
              source.getControl().getParent().pack();
              source.getControl().getParent().layout();
              if (!page.getControl().isDisposed()) {
                page.resizeScrolledComposite();
              }
            }
  
            @Override
            public void inputDocumentAboutToBeChanged(IDocument oldInput, IDocument newInput) {
              // TODO Auto-generated method stub
  
            }
          });
    }
    finally{
      l.unlock();
    }
    //source.getControl().setLayoutData(new Fi);
  }

  private boolean transupdate = false;
  INodeChangeListener ncl =
      new INodeChangeListener() {
        @Override
        public void deleted(TreeNode source) {
          if (target != null && source.getUUId().equals(target.getUUId())) {
            if (page.getControl().isDisposed()) {
              return;
            }
            if (page.getCurrentTab() != null) {
              TreeNode par = target.getShownParent();
              if (par == null) {
                setInput(getPart(), null); //TODO del autoup
                return;
              }
              if (par.isDisposed()) par = par.getTreeDB().getRootNode();
              setInput(getPart(), new StructuredSelection(par)); //TODO del autoup
            }
          }
        }

        @Override
        public void attributeChange(final TreeNode source, String what, Attribute old) {
          if (page.getControl().isDisposed()&&target == source) {
            return;
          }
          if (source.getTreeDB().inTransaction()) {
            transupdate = true;
          } else {
            openNode(source);
          }
        }

        @Override
        public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          if (page.getControl().isDisposed()&&target == source) {
            return;
          }
          if (source.getTreeDB().inTransaction()) {
            transupdate = true;
          } else {
            openNode(source);
          }
        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };
  ITransactionListener itl =
      new ITransactionListener() {

        @Override
        public void undo(TransactionStorage tStorage) {
          if (transupdate) {
            openNode(target);
            transupdate = false;
          }
        }

        @Override
        public void redo(TransactionStorage tStorage) {
          if (transupdate) {
            openNode(target);
            transupdate = false;
          }
        }

        @Override
        public void execute(TransactionStorage tStorage) {
          if (transupdate) {
            openNode(target);
            transupdate = false;
          }
        }

        @Override
        public void _dispose() {
          // TODO Auto-generated method stub

        }
      };

  @Override
  public void setInput(IWorkbenchPart part, ISelection selection) {
    super.setInput(part, selection);

    TreeNode node = TreeNodePropertiesTab.toTreeNode(selection);
    if (node == null) {
      return;
    }
    openNode(node);
  }

  private TreeNode target;

  private void openNode(final TreeNode node) {
    DebugPropertySection.this
        .getPart()
        .getSite()
        .getShell()
        .getDisplay()
        .asyncExec(
            new Runnable() {
              @Override
              public void run() {
                openNodeSafe(TreeNodePropertiesTab.toRequirementOrNode(node));
              }
            });
  }

  private void openNodeSafe(TreeNode node) {
    if (target != null) {
      target.removeGuiChangeListener(ncl);
      target.getTreeDB().removeTransactionListener(itl);
    }
    if (node == null) return;
    target = node;
    target.addGuiChangeListener(ncl);
    target.getTreeDB().addTransactionListener(itl);
    try {
      StringBuilder addtext = new StringBuilder();
      addtext.append(node.getClass().getName());
      addtext.append("\n");
      addtext.append(node.getType());
      addtext.append("\n");
      addtext.append(node.isDisposed() ? "DISPOSED " : "");
      addtext.append(node.isVirtual() ? "" : "non");
      addtext.append("virt node");
      addtext.append(", ");
      addtext.append(node.getTreeDB().isVirtual() ? "" : "non");
      addtext.append("virt TreeDB(");
      addtext.append(node.getTreeDB().getName());
      addtext.append("), uid: ");
      addtext.append(node.getUUId().toString());
      addtext.append("\nparent: ");
      if (node.getWorkParent() == null) addtext.append("NONE");
      else {
        addtext.append(node.getWorkParent().toString());
        addtext.append("(");
        addtext.append(node.getWorkParentUUID());
        addtext.append(", ");
        addtext.append(node.getWorkParent().isVirtual() ? "" : "non");
        addtext.append("virt node");
        addtext.append(", ");
        addtext.append(node.getWorkParent().getTreeDB().isVirtual() ? "" : "non");
        addtext.append("virt TreeDB");
        addtext.append(")");
      }
      addtext.append("\nwchilds: ");
      addtext.append(node.getWorkChildrenUUIDs().size());
      addtext.append(", schilds: ");
      addtext.append(node.getShownChildrenUUIDs().size());
      addtext.append("\ncloneParent: ");
      if (node.getCloneParent() == null) addtext.append("NONE");
      else {
        addtext.append(node.getCloneParent().toString());
        addtext.append("(");
        addtext.append(node.getCloneParentUUId());
        addtext.append(", ");
        addtext.append(node.getCloneParent().isVirtual() ? "" : "non");
        addtext.append("virt node");
        addtext.append(", ");
        addtext.append(node.getCloneParent().getTreeDB().isVirtual() ? "" : "non");
        addtext.append("virt TreeDB");
        addtext.append(")");
      }
      addtext.append("\n");
      VersionsStorage vs = node.attrs().getVersionsStorage();
      for (String s : node.attrs().getVersionsStorage().getKeys()) {
        addtext.append("\n");
        addtext.append(s);
        addtext.append(": ");
        addtext.append(vs.getVersion(s).toString());
      }
      addtext.append("\nrev:\n");
      Map<String,Set<TreeNode>> set = node.getTreeDB().getLinkResolver().getRevertLinks(node);
      if(set!=null)
        for(String s: set.keySet()){
          addtext.append(s+":\n");
          for(TreeNode m:node.getTreeDB().getLinkResolver().getRevertLinks(node).get(s))
            addtext.append("\t"+m.getQualifiedId()+"\n");
        }
      addtext.append("\nrelated:\n");
      LinkResolver lr = node.getTreeDB().getLinkResolver();
      Map<String,Set<TreeNode>>map = lr.getLinkedNodes(node);
      if(map!=null)
        for(String s:map.keySet()){
          addtext.append(s+":\n");
          for(TreeNode n:map.get(s))
            addtext.append("\t"+n.getQualifiedId()+"\n");
        }
      source.setInput(new Document(addtext.toString()));

    } catch (Exception e) {
      e.printStackTrace();
    }
    //page.getControl().pack();
    //source.getControl().forceFocus();
  }
}
