/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.HashSet;
import java.util.Set;
import java.util.UUID;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.VirtualNode;

public class MainPropertySection extends TreeNodePropertiesTab {

  private AttributesPanel attribPane;
  private Text textId;
  private NameEditPanel textName;
  private Label textNameLabel;
  ControlDecoration reqIdDecorator;
  private TreeNode req;
  private Combo type;
  private Composite tp;
  private GridData tps;
  protected boolean nameEnabled = true;

  public MainPropertySection() {}

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    TabbedPropertySheetWidgetFactory wf = getWidgetFactory();

    //parent.setLayoutData(new GridData(GridData.FILL_BOTH));
    parent.setLayout(new GridLayout(2, false));

    wf.createLabel(parent, "Id:").setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));
    textId = wf.createText(parent, "");
    textId.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    reqIdDecorator = createDecorator(textId);
    reqIdDecorator.hide();
    textId.addKeyListener(
        new KeyListener() {

          @Override
          public void keyReleased(KeyEvent e) {
            try {
              if (req == null) {
                return;
              }
              reqIdDecorator.show();
              if (req.getTreeDB().getTreeLogic().isIdValidExc(req, textId.getText().trim())) {
                reqIdDecorator.hide();
              } else {
                reqIdDecorator.setDescriptionText(
                    "Requirement Id is not correct and Requirement Id should not be empty");
                reqIdDecorator.show();
              }
            } catch (SameIdExistsInChildren e1) {
              reqIdDecorator.setDescriptionText("Element with the same Id already exists");
            } catch (IdEmpty e1) {
              reqIdDecorator.setDescriptionText("Requirement Id is empty");
            } catch (IdContainsWrongChars e1) {
              reqIdDecorator.setDescriptionText(
                  "Requirement Id contains wrong character - " + e1.getWrongCharacter());
            }
          }

          @Override
          public void keyPressed(KeyEvent e) {}
        });
    textId.addFocusListener(
        new FocusAdapter() {
          public void focusLost(FocusEvent e) {
            if (updating) return;
            if (reqIdDecorator.isVisible()) {
              reqIdDecorator.hide();
              textId.setText(node.getId().trim());
            }
            if (req == null) {
              return;
            }
            if (req.getId().equals(textId.getText())) {
              return;
            }
            if (!req.getQualifiedId().equals(Requirement.getTypeRootQId())) {
              String newId = textId.getText().trim();
              if (req.getTreeDB().getTreeLogic().isIdValid(req, newId)
                  && req.getTreeDB().startTransaction("Change Requirement Id", true)) {
                req.setId(newId.trim());
                req.commitAttributes();
                req.getTreeDB().commit();
              } else {
                textId.setText(req.getId());
              }
            }
          }
        });
    textNameLabel = wf.createLabel(parent, "Name:");
    textNameLabel.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));
    textNameLabel.setVisible(nameEnabled);
    textName = new NameEditPanel(parent, SWT.BORDER);
    textName.getControl().setVisible(nameEnabled);
    textName.getControl().setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    tp = new Composite(parent, SWT.NONE);
    tp.setBackground(parent.getBackground());
    tps = new GridData(GridData.FILL_HORIZONTAL);
    tps.horizontalSpan = 2;
    tp.setLayoutData(tps);
    tp.setLayout(new GridLayout(2, false));
    wf.createLabel(tp, "Type:").setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));
    type = new Combo(tp, SWT.READ_ONLY);
    addAvailiableTypes(req);
    type.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    type.addSelectionListener(
        new SelectionListener() {

          @Override
          public void widgetSelected(SelectionEvent e) {
            if (req.getType().equals(Requirement.TYPE_NAME))
              for (UUID uid : req.getWorkChildrenUUIDs())
                if (req.getTreeDB().getNode(uid).getType().equals(Requirement.TYPE_NAME)) return;
            if (req.getTreeDB().startTransaction("Change text element type")) {
              int ind = type.getSelectionIndex();
              if (ind < 2) {
                req.attrs()
                    .setConstAttribute(
                        "_textNodeType", type.getSelectionIndex() == 1 ? "header" : "text",Flag.HIDDEN);
                req.setType(TextNode.TYPE_NAME);
              } else {
                req.setType(Requirement.TYPE_NAME);
              }
              req.commitAttributes();
              req.getTreeDB().commit();
            }
          }

          @Override
          public void widgetDefaultSelected(SelectionEvent e) {
            // TODO Auto-generated method stub

          }
        });

    GridData gridData = new GridData(GridData.FILL_BOTH);
    gridData.horizontalSpan = 2;
    attribPane = new AttributesPanel(parent);
    attribPane.getControl().setLayoutData(gridData);
    if (req != null) attribPane.setNode(req);
  }

  public static TreeNode getNonVNodeParent(TreeNode nde) {
    TreeNode node = nde.getWorkParent();
    if (node != null)
      if (VirtualNode.TYPE_NAME.equals(node.getType())) return getNonVNodeParent(node);
    return node;
  }

  private void addAvailiableTypes(TreeNode req2) {
    type.removeAll();
    type.add("TextNode(Text)");
    type.add("TextNode(Header)");
    if (req2 != null
        && (Requirement.TYPE_NAME.equals(getNonVNodeParent(req2).getType())
            || getNonVNodeParent(req2).isRoot())) type.add("Requirement");
  }

  private ControlDecoration createDecorator(Text text) {
    ControlDecoration controlDecoration = new ControlDecoration(text, SWT.LEFT | SWT.TOP);
    FieldDecoration fieldDecoration =
        FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_ERROR);
    controlDecoration.setImage(fieldDecoration.getImage());
    return controlDecoration;
  }

  @Override
  protected void setTreeNode(TreeNode node) {
    reqIdDecorator.hide();
    nameEnabled = false;
    if (node == null || node.isDisposed()) {
      return;
    }
    req = node;
    if (node instanceof Location) req = toRequirement(node);
    if (req == null || textId == null || textId.isDisposed() || req.isDisposed()) {
      return;
    }
    boolean next = true;
    if (!req.getType().equals(Requirement.TYPE_NAME) && !req.getType().equals(TextNode.TYPE_NAME))
      next = false;
    if (req.getWorkParent() != null && !req.getWorkParent().getQualifiedId().equals("/")) {
      textId.setEnabled(true);
    } else {
      textId.setEnabled(false);
    }
    if (req.getType().equals(Requirement.TYPE_NAME) || req.getType().equals(TextNode.TYPE_NAME))
      nameEnabled = true;
    if (!textId.getText().equals(req.getId())) { //against blinking
      textId.setText(req.getId());
    }
    addAvailiableTypes(req);
    if (req.getType().equals(TextNode.TYPE_NAME)) {
      Attribute a = req.attrs().getAttribute("_textNodeType");
      type.select(0);
      if (a != null) type.select(a.getValue().toString().equals("header") ? 1 : 0);
    } else if (req.getType().equals(Requirement.TYPE_NAME)) type.select(2);
    type.setEnabled(true);
    if (req.getType().equals(Requirement.TYPE_NAME))
      for (TreeNode nde : req.getWorkChildren())
        if (!req.getTreeDB().getTreeLogic().canBeChildOf(nde.getType(), TextNode.TYPE_NAME))
          type.setEnabled(false);
    textName.getControl().setVisible(nameEnabled);
    textNameLabel.setVisible(nameEnabled);
    if (!nameEnabled) {
      textName.getControl().setLayoutData(new GridData(0, 0));
      textNameLabel.setLayoutData(new GridData(0, 0));
    } else {
      textName.getControl().setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
      textNameLabel.setLayoutData(new GridData(GridData.BEGINNING));
    }
    textName.getControl().getParent().layout();
    textName.getControl().getParent().redraw();
    if (nameEnabled) {
      textName.setTreeNode(req);
    }
    if (attribPane != null) attribPane.setNode(req);
    if (next != tp.isVisible()) {
      tp.setVisible(next);
      tps.exclude = !next;
      tp.getParent().getParent().layout(true, true);
    }
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return true;
  }

  private static Set<String> ids = new HashSet<>();
  static{
    ids.add(TreeNode.ATTR_DESCRIPTION);
    ids.add(TreeNode.ATTR_ID);
    ids.add(TreeNode.ATTR_NAME);
    ids.add(TextNode.ATTR_TNTYPE);
    ids.add(TreeNode.ATTR_TYPE);
  }
  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    for(String s:what.getAll())
      if(ids.contains(s) || !s.startsWith("_"))
        return true;
    return false;
  };
}
