package com.unitesk.requality.eclipse.views.properties;

import java.util.ArrayList;
import java.util.UUID;
import org.eclipse.core.resources.IProject;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbenchPropertyPage;
import org.eclipse.ui.dialogs.PropertyPage;
import com.unitesk.requality.core.SortingSettings;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.eclipse.core.EclipseTracker;

public class ProjectProperties extends PropertyPage implements IWorkbenchPropertyPage {
  private Combo sbllist, nametype;
  private TreeNode nodeToStore;
  private IProject p;
  private Text helpLabel;
  private Button mixcheck;
  private Button desctext;
  //text on top of preference page
  private static final String desciption_text = "Sort method:";
  private static final String
      LOCATIONS_TEXT =
          "Requirements are sorted by location position in document. "
              + "If there are several documents sorting depends on document names. "
              + "If requirement has several locations sorting depends only on the first location. "
              + "Requirements without locations are at the end of the list and are sorted by name.",
      NAMES_TEXT =
          "Requirements are sorted by name (first, in ascending order, then in alphabetical order). "
              + "In case of coincidence or lack of names requirements are sorted by id.",
      INDEX_TEXT =
          "Requirements are sorted by index. Initial value of index will be determined by names.";

  private static final String PREFERENCE_REPLS = "_com.unitesk.requality.reference_replaces";

  //text on top of preference page

  public ProjectProperties() {}

  @Override
  protected Control createContents(Composite parent) {

    Composite c = new Composite(parent, SWT.NONE);
    c.setLayout(new GridLayout(1, true));
    c.setLayoutData(new GridData(GridData.FILL_BOTH));

    final Group container2 = new Group(c, SWT.NONE);
    container2.setText("Sort options:");
    container2.setLayout(new GridLayout(2, false));
    container2.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

    Label l = new Label(container2, SWT.NONE);
    l.setLayoutData(new GridData());
    l.setText(desciption_text);

    SortingSettings st = TreeDB.loadSettings(nodeToStore);
    sbllist = new Combo(container2, SWT.NONE);
    for (SortingMethod sm : SortingMethod.values()) sbllist.add(SortingSettings.getSortingName(sm));
    SortingMethod[] sms = SortingMethod.values();
    for (int i = 0; i < sms.length; i++) if (sms[i] == st.getMethod()) sbllist.select(i);
    sbllist.addSelectionListener(
        new SelectionListener() {
          @Override
          public void widgetSelected(SelectionEvent e) {
            if (helpLabel != null) {
              int sel = sbllist.getSelectionIndex();
              if (sel < 0) sel = 0;
              SortingMethod sm = SortingMethod.values()[sel];
              helpLabel.setText(
                  sm == SortingMethod.BY_LOCATIONS
                      ? LOCATIONS_TEXT
                      : sm == SortingMethod.BY_INDEX ? INDEX_TEXT : NAMES_TEXT);
              container2.layout();
            }
          }

          @Override
          public void widgetDefaultSelected(SelectionEvent e) {}
        });

    sbllist.setLayoutData(new GridData());

    helpLabel = new Text(container2, SWT.WRAP);

    GridData gd = new GridData(SWT.FILL, SWT.FILL, true, false, 1, 1);
    gd.horizontalSpan = 2;
    gd.widthHint = 200;
    helpLabel.setLayoutData(gd);
    helpLabel.setBackground(container2.getBackground());
    helpLabel.setText(
        st.getMethod() == SortingMethod.BY_LOCATIONS
            ? LOCATIONS_TEXT
            : st.getMethod() == SortingMethod.BY_INDEX ? INDEX_TEXT : NAMES_TEXT);

    mixcheck = new Button(container2, SWT.CHECK);
    GridData gd4 = new GridData();
    gd4.horizontalSpan = 2;
    mixcheck.setLayoutData(gd4);
    mixcheck.setText("Sort elements types separately");
    if (nodeToStore != null) mixcheck.setSelection(st.getSeparateByTypesFlag());

    Text helpLabel2 = new Text(container2, SWT.WRAP);

    gd = new GridData(SWT.FILL, SWT.FILL, true, false, 1, 1);
    gd.horizontalSpan = 2;
    gd.widthHint = 200;
    helpLabel2.setLayoutData(gd);
    helpLabel2.setBackground(container2.getBackground());
    helpLabel2.setText(
        "Sorting can be made separately for different kinds of elements. If check box is checked then elements will be sorted separately.");

    container2.layout();
    final Group container3_5 = new Group(c, SWT.NONE);
    container3_5.setText("Full name settings:");
    container3_5.setLayout(new GridLayout(2, false));
    container3_5.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    l = new Label(container3_5, SWT.NONE);
    l.setText("Full name mode:");

    nametype = new Combo(container3_5, SWT.NONE);
    nametype.add("By Full Name");
    nametype.add("By User-Visible Name");
    nametype.select(nodeToStore.getTreeDB().getFullNameMode() ? 1 : 0);

    final Group container5 = new Group(c, SWT.NONE);
    container5.setText("Requality Explorer settings:");
    container5.setLayout(new GridLayout(2, false));
    container5.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

    l = new Label(container5, SWT.NONE);
    l.setText("Show description on elements");

    desctext = new Button(container5, SWT.CHECK);
    desctext.setSelection(!nodeToStore.getTreeDB().textIsSimple);

    return c;
  }

  Composite isc;
  ScrolledComposite sc;

  @Override
  public void setElement(org.eclipse.core.runtime.IAdaptable element) {
    if (element instanceof IProject) {
      p = (IProject) element;
      TreeDB db = EclipseTracker.getTracker().getTree(p);
      if (db != null) {
        nodeToStore = db.getRootNode();
        SortingSettings st = db.getSortingSettings();
        if (sbllist != null) {
          SortingMethod[] sms = SortingMethod.values();
          for (int i = 0; i < sms.length; i++) if (sms[i] == st.getMethod()) sbllist.select(i);
        }
        if (mixcheck != null) mixcheck.setSelection(st.getSeparateByTypesFlag());
        if (desctext != null) desctext.setSelection(!db.textIsSimple);
      }
    }
  };
  //actions when new selection is performed
  @Override
  public boolean performOk() {
    int sel = sbllist.getSelectionIndex();
    if (sel == -1) sel = 0;
    nodeToStore
        .getTreeDB()
        .setSortingMethod(
            SortingSettings.init(SortingMethod.values()[sel], mixcheck.getSelection()), false);
    //    	nodeToStore.attrs().setAttribute(TreeDB.SORT_METHOD, .toString());
    //  	nodeToStore.attrs().setAttribute(TreeDB.SORT_SEP_TYPES,);


    nodeToStore.getTreeDB().setFullNameMode((nametype.getSelectionIndex() > 0) ? true : false);
    nodeToStore.getTreeDB().textIsSimple = !desctext.getSelection();
    nodeToStore.commitAttributes();
    fireOutClear(nodeToStore);
    return true;
  }

  private void fireOutClear(TreeNode nodeToStore2) {
    for (UUID uid : nodeToStore2.getActiveWorkChildrenUUIDs()) {
      TreeNode node = nodeToStore2.getTreeDB().getCachedNodeOrVirt(uid);
      if (node != null) {
        node.attrs().updateAttribute(TreeNode.ATTR_LABEL);
        node.commitAttributes();
        fireOutClear(node);
      }
    }
    nodeToStore2.clearShownChildren();
  }

  //Changing settings to default
  @Override
  protected void performDefaults() {
    sbllist.select(1);
    helpLabel.setText(LOCATIONS_TEXT);
    helpLabel.getParent().layout();
    if (nodeToStore != null) {
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_BY_INDEX);
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_BY_LOCATIONS);
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_SEP_TYPES);
      nodeToStore
          .attrs()
          .setConstAttribute(
              PREFERENCE_REPLS, new ListValue(AttributeType.STRING, new ArrayList<>()),
              Flag.HIDDEN);
      nodeToStore.attrs().removeAttribute(TreeDB.DEFAULT_REFERENCE_NAME);
      nodeToStore.getTreeDB().setFullNameMode(false);
      nodeToStore.getTreeDB().textIsSimple = true;
      nodeToStore.commitAttributes();
      fireOutClear(nodeToStore);
    }
    nametype.select(0);
    desctext.setSelection(false);
    
    super.performDefaults();
  }
}
