package com.unitesk.requality.eclipse.views.properties;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import org.eclipse.core.resources.IProject;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchPropertyPage;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.PropertyPage;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.report.TraceabilityDataSource;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class ProjectReferenceProperties extends PropertyPage implements IWorkbenchPropertyPage {
  private TreeNode nodeToStore;
  private IProject p;
  //private Text helpLabel;
  private Text defRefName;
  //text on top of preference page
  private static final String PREFERENCE_NAMES = "_com.unitesk.requality.reference_names";
  private static final String PREFERENCE_REPLS = "_com.unitesk.requality.reference_replaces";

  private static Map<String, String> globalRenames = new HashMap<>();
  private Set<String> names = new HashSet<>();
  private Map<String, ReplPair> replacementMap = new HashMap<>();
  //text on top of preference page
  private org.eclipse.swt.widgets.Text torename, replacement;
  private org.eclipse.swt.widgets.List listname;
  private TableViewer repl;
  private Button button;

  private class ReplPair {
    private String name, repl;

    public ReplPair(String name, String repl) {
      this.name = name;
      this.repl = repl;
    }

    @Override
    public String toString() {
      return "'" + name + "' -> '" + repl + "'";
    }

    @Override
    public boolean equals(Object obj) {
      if (obj instanceof ReplPair) return name.equals(((ReplPair) obj).name);
      return super.equals(obj);
    }
  }

  public ProjectReferenceProperties() {}

  @Override
  protected Control createContents(Composite parent) {

    Composite c = new Composite(parent, SWT.NONE);
    c.setLayout(new GridLayout(1, true));
    c.setLayoutData(new GridData(GridData.FILL_BOTH));

    final Group container3 = new Group(c, SWT.NONE);
    container3.setText("References antonyms:");
    container3.setLayout(new GridLayout(2, false));
    container3.setLayoutData(new GridData(GridData.FILL_BOTH));

    initReferenceRenameComposite(container3);

    final Group container4 = new Group(c, SWT.NONE);
    container4.setText("References settings:");
    container4.setLayout(new GridLayout(2, false));
    container4.setLayoutData(new GridData(GridData.FILL_BOTH));

    Label l = new Label(container4, SWT.NONE);
    l.setText("Default reference name:");

    defRefName = new Text(container4, SWT.NONE);
    defRefName.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    defRefName.setText(nodeToStore.getTreeDB().getDefReferenceName());

    return c;
  }

  Composite isc;
  ScrolledComposite sc;

  private void initReferenceRenameComposite(Group container3) {
    Composite desccompose = new Composite(container3, SWT.FILL);
    desccompose.setLayoutData(new GridData(GridData.FILL_BOTH));
    desccompose.setLayout(new GridLayout(2, true));
    Label descl = new Label(desccompose, SWT.HORIZONTAL);
    descl.setText("Select reference name:");
    descl = new Label(desccompose, SWT.NONE);
    descl.setText("Available renames:");

    sc = new ScrolledComposite(desccompose, SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL);
    sc.setAlwaysShowScrollBars(true);
    GridData gd = new GridData(GridData.FILL_BOTH);
    gd.horizontalSpan = 2;
    sc.setLayoutData(gd);
    sc.setLayout(new GridLayout(1, true));

    isc = new Composite(sc, SWT.NONE);
    isc.setLayout(new GridLayout(2, true));
    gd = new GridData(GridData.FILL_BOTH);
    gd.widthHint = sc.getSize().x;
    isc.setLayoutData(gd);

    sc.setContent(isc);
    sc.setExpandHorizontal(true);
    sc.setExpandVertical(true);

    Button b = new Button(isc, SWT.NONE);
    b.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

    repl = new TableViewer(isc, SWT.BORDER | SWT.RESIZE | SWT.FULL_SELECTION);
    gd = new GridData(GridData.FILL_BOTH);
    gd.verticalSpan = 2;
    repl.getTable().setLayoutData(gd);

    listname = new List(isc, SWT.NONE);
    listname.setLayoutData(new GridData(GridData.FILL_BOTH));

    final TableViewerColumn column = new TableViewerColumn(repl, SWT.FULL_SELECTION);
    column.getColumn().setText("Replacement");
    column.getColumn().setWidth(200);
    column.getColumn().setResizable(true);
    column.getColumn().setMoveable(true);

    repl.getControl()
        .addControlListener(
            new ControlListener() {
              @Override
              public void controlResized(ControlEvent e) {
                column.getColumn().setWidth(repl.getControl().getSize().x - 25);
              }

              @Override
              public void controlMoved(ControlEvent e) {
                // TODO Auto-generated method stub

              }
            });

    TableViewerColumn cdel = new TableViewerColumn(repl, SWT.FULL_SELECTION);
    cdel.getColumn().setResizable(false);
    cdel.getColumn().setWidth(25);
    cdel.setEditingSupport(
        new EditingSupport(repl) {
          @Override
          protected void setValue(Object element, Object value) {}

          @Override
          protected Object getValue(Object element) {
            return "";
          }

          @Override
          protected CellEditor getCellEditor(Object element) {
            return new TextCellEditor();
          }

          @Override
          protected boolean canEdit(Object element) {
            ReplPair pair = (ReplPair) element;
            String name = pair.name;
            boolean glob = globalRenames.containsKey(name);
            if (glob && globalRenames.get(name).equals(pair.repl)) return false;
            replacementMap.remove(name);
            repl.remove(element);
            if (glob) addAcronym(name, globalRenames.get(name));
            updateComposite();
            return false;
          }
        });

    repl.setLabelProvider(
        new ITableLabelProvider() {
          @Override
          public void removeListener(ILabelProviderListener listener) {}

          @Override
          public boolean isLabelProperty(Object element, String property) {
            return false;
          }

          @Override
          public void dispose() {}

          @Override
          public void addListener(ILabelProviderListener listener) {}

          @Override
          public Image getColumnImage(Object element, int columnIndex) {
            ReplPair pair = (ReplPair) element;
            String name = pair.name;
            boolean glob = globalRenames.containsKey(name);
            if (columnIndex == 1)
              if (!glob || !globalRenames.get(name).equals(pair.repl))
                return PlatformUI.getWorkbench()
                    .getSharedImages()
                    .getImage(ISharedImages.IMG_ELCL_REMOVE);
            return null;
          }

          @Override
          public String getColumnText(Object element, int columnIndex) {
            if (columnIndex == 1) return null;
            return element.toString();
          }
        });

    if (!replacementMap.isEmpty()) for (ReplPair rp : replacementMap.values()) repl.add(rp);

    Composite edit = new Composite(desccompose, SWT.NONE);
    gd = new GridData(GridData.FILL_BOTH);
    gd.horizontalSpan = 2;
    edit.setLayoutData(gd);
    edit.setLayout(new GridLayout(3, false));

    descl = new Label(edit, SWT.NONE);
    descl.setText("Ref. name to rename");

    descl = new Label(edit, SWT.NONE);
    descl.setText("Replacement name");
    gd = new GridData();
    gd.horizontalSpan = 2;
    descl.setLayoutData(gd);

    b.setText("Reload reference names");

    ModifyListener ml =
        new ModifyListener() {
          @Override
          public void modifyText(ModifyEvent e) {
            button.setEnabled(!torename.getText().isEmpty() && !replacement.getText().isEmpty());
          }
        };

    torename = new org.eclipse.swt.widgets.Text(edit, SWT.NONE);
    torename.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    torename.addModifyListener(ml);

    replacement = new org.eclipse.swt.widgets.Text(edit, SWT.NONE);
    replacement.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    replacement.addModifyListener(ml);

    button = new Button(edit, SWT.NONE);
    button.setText("Set");
    button.setEnabled(false);
    button.addMouseListener(
        new MouseListener() {
          @Override
          public void mouseUp(MouseEvent e) {}

          @Override
          public void mouseDown(MouseEvent e) {
            addAcronym(torename.getText(), replacement.getText());
          }

          @Override
          public void mouseDoubleClick(MouseEvent e) {}
        });

    loadNames();
    listname.setItems(names.toArray(new String[0]));

    listname.addSelectionListener(
        new SelectionListener() {
          @Override
          public void widgetSelected(SelectionEvent e) {
            button.setEnabled(true);
            if (listname.getSelectionCount() < 1) return;
            String sel = listname.getSelection()[0];
            if (replacementMap.containsKey(sel)) replacement.setText(replacementMap.get(sel).repl);
            torename.setText(sel);
          }

          @Override
          public void widgetDefaultSelected(SelectionEvent e) {
            // TODO Auto-generated method stub

          }
        });

    b.addMouseListener(
        new MouseListener() {
          @Override
          public void mouseUp(MouseEvent e) {}

          @Override
          public void mouseDown(MouseEvent e) {
            loadNames();
            button.setEnabled(false);
            torename.setText("");
            replacement.setText("");
            listname.setItems(names.toArray(new String[0]));
          }

          @Override
          public void mouseDoubleClick(MouseEvent e) {}
        });

    repl.getTable()
        .addSelectionListener(
            new SelectionListener() {

              @Override
              public void widgetSelected(SelectionEvent e) {
                ReplPair rp =
                    (ReplPair) (((StructuredSelection) repl.getSelection()).getFirstElement());
                if (rp != null) {
                  torename.setText(rp.name);
                  replacement.setText(rp.repl);
                  button.setEnabled(true);
                }
              }

              @Override
              public void widgetDefaultSelected(SelectionEvent e) {}
            });
    updateComposite();
  }

  private void updateComposite() {
    sc
        .setMinSize(isc.getSize().x, isc.computeSize(SWT.DEFAULT, SWT.DEFAULT, true).y);
    sc.layout(true, true);
  }

  protected void addAcronym(String text, String text2) {
    if (text.equals("")) button.setEnabled(false);
    if (replacementMap.containsKey(text)) repl.remove(replacementMap.get(text));
    ReplPair rp = new ReplPair(text, text2);
    repl.add(rp);
    replacementMap.put(text, rp);
    updateComposite();
  }

  @Override
  public void setElement(org.eclipse.core.runtime.IAdaptable element) {
    if (element instanceof IProject) {
      p = (IProject) element;
      TreeDB db = EclipseTracker.getTracker().getTree(p);
      if (db != null) {
        nodeToStore = db.getRootNode();
        Map<String, String> acr = getNamesAcronyms(db);
        for (String s : acr.keySet()) replacementMap.put(s, new ReplPair(s, acr.get(s)));
      }
    }
  };

  private void loadNames() {
    names = new HashSet<>();
    if (nodeToStore == null) return;
    //for(TreeDB db:TreesTracker.getTracker().getTrees()){
    try {
      TreeNode node = nodeToStore;
      DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(node, true, true, false);
      Iterator<TreeNode> it = dftw.iterator();
      while (it.hasNext()) {
        TreeNode nde = it.next();
        for (Attribute a : nde.attrs()) {
          if (a.isReference()) names.add(a.getKey());
        }
      }
    } catch (Exception e) {
      //TODO show error on names
    }
    //}
    java.util.List<Object> l = new ArrayList<>();
    for (String s : names) {
      l.add(s);
    }
//    nodeToStore.attrs().setConstAttribute(PREFERENCE_NAMES, new ListValue(AttributeType.STRING, l),
//        Flag.HIDDEN);
//    nodeToStore.commitAttributes();
  }

  public static Map<String, String> getNamesAcronyms(TreeDB db) {
    Map<String, String> mp = new HashMap<>();
    UUID root = db.getRootNodeUUID();
    if (root != null) {
      Attribute a = db.getRootNode().attrs().getAttribute(PREFERENCE_REPLS);
      if (a != null) {
        //AttributeDefinition a = at.get(PREFERENCE_REPLS);
        ListValue lv = a.getGValue();
        if(lv!=null){
          java.util.List<Object> l = lv.getValues();
          for (int i = 0; i < (l.size() / 2); i++) {
            mp.put(l.get(i * 2).toString(), l.get(i * 2 + 1).toString());
          }
        }
      }
    }
    for (String s : globalRenames.keySet()) if (!mp.containsKey(s)) mp.put(s, globalRenames.get(s));
    //      IPreferenceStore pref = Activator.getDefault().getPreferenceStore();
    //      if( pref.contains(PREFERENCE_REPLS)){
    //          String[]saved = pref.getString(PREFERENCE_REPLS).split("\\|\\|");
    //          for(int i=0;i<saved.length/2;i++)
    //              mp.put(saved[i*2], saved[i*2+1]);
    //      }
    return mp;
  }
  //actions when new selection is performed
  @Override
  public boolean performOk() {
    java.util.List<Object> list = new ArrayList<>();
    for (ReplPair rp : replacementMap.values()) {
      if (!globalRenames.containsKey(rp.name) || !globalRenames.get(rp.name).equals(rp.repl)) {
        list.add(rp.name);
        list.add(rp.repl);
      }
    }
    nodeToStore.attrs().setConstAttribute(PREFERENCE_REPLS, new ListValue(AttributeType.STRING, list),
        Flag.HIDDEN);
    String defText = defRefName.getText();
    if (!defText.equals(nodeToStore.getTreeDB().getDefReferenceName()))
      nodeToStore.attrs().setConstAttribute(TreeDB.DEFAULT_REFERENCE_NAME, defText,
          Flag.HIDDEN);
    nodeToStore.commitAttributes();
    TraceabilityDataSource.setRevertNames(
        nodeToStore.getTreeDB(), getNamesAcronyms(nodeToStore.getTreeDB()));
    fireOutClear(nodeToStore);
    return true;
  }

  private void fireOutClear(TreeNode nodeToStore2) {
    nodeToStore2.clearShownChildren();
    for (UUID uid : nodeToStore2.getActiveWorkChildrenUUIDs()) {
      TreeNode node = nodeToStore2.getTreeDB().getCachedNodeOrVirt(uid);
      if (node != null) fireOutClear(node);
    }
  }

  //Changing settings to default
  @Override
  protected void performDefaults() {
    if (nodeToStore != null) {
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_BY_INDEX);
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_BY_LOCATIONS);
      nodeToStore.attrs().removeAttribute(TreeDB.SORT_SEP_TYPES);
      nodeToStore
          .attrs()
          .setConstAttribute(
              PREFERENCE_REPLS, new ListValue(AttributeType.STRING, new ArrayList<>()),
              Flag.HIDDEN);
      nodeToStore.attrs().removeAttribute(TreeDB.DEFAULT_REFERENCE_NAME);
      nodeToStore.commitAttributes();
      TraceabilityDataSource.setRevertNames(
          nodeToStore.getTreeDB(), getNamesAcronyms(nodeToStore.getTreeDB()));
    }
    repl.getTable().clearAll();
    repl.refresh();
    torename.setText("");
    replacement.setText("");
    defRefName.setText(nodeToStore.getTreeDB().getDefReferenceName());
    button.setEnabled(false);
    replacementMap.clear();
    nodeToStore.getTreeDB().textIsSimple = true;

    super.performDefaults();
  }

  public static void setGlobalRenames(Map<String, String> referenceAcronyms) {
    globalRenames = referenceAcronyms;
  }
}
