/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.fieldassist.FieldDecoration;
import org.eclipse.jface.fieldassist.FieldDecorationRegistry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.operations.GenerateReportHandler;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.editors.panels.SelectRequirementPanel;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.report.Template;

public class ReportSettingsPropertySection extends TreeNodePropertiesTab {
  private static String LBL_ROOT = "Root requirement:";
  private static String LBL_TEMPLATE = "Template:";
  private static String NOT_INSTALLED = "Not avalible";
  // Widgets
  private SelectRequirementPanel rootReqPanel;
  private Combo templates;
  private Text textId;
  private ControlDecoration reqIdDecorator;
  private AttributesPanel attrPanel;
  private Button gen;

  private LinkedHashMap<String, Template> tpls = new LinkedHashMap<>();
  private String selectedTemplateId;
  private ReportSettings report;
  private Composite customGUIpanel;
  private TreeDB db;

  private boolean isData() {
    return node == null || node.getType().equals(ReportData.TYPE_NAME);
  }

  @Override
  protected void setTreeNode(TreeNode node) {
    tpls = new LinkedHashMap<>(Activator.getDefault().getRenewedReportTemplates());
    if (node != null
        && !node.getType().equals(ReportSettings.TYPE_NAME)
        && !node.getType().equals(ReportData.TYPE_NAME)) {
      return;
    }
    report = (ReportSettings) node;
    Requirement req = null;
    if(node!=null){
      db = node.getTreeDB();
      if (report.getStartRequirementQId() == null) {
        req = (Requirement) db.getNode(Requirement.getTypeRootQId(), Path.QID);
      } else {
        req = (Requirement) db.getNode(report.getStartRequirementQId(), Path.QID);
      }
    }
    attrPanel.setNode(node);
    attrPanel.setEnabled(!isData());
    if (req != null) rootReqPanel.setRequirement(req);
    selectedTemplateId = node == null? null : 
        node.attrs().getStringAttributeValue(ReportSettings.ATTR_TEMPLATE);
    if (selectedTemplateId == null) {
      selectedTemplateId = tpls.keySet().iterator().next();
    }
    Iterator<String> it = tpls.keySet().iterator();
    String curId = "";
    int i = 0;
    for (; i < tpls.size() && !selectedTemplateId.equals(curId); i++) {
      curId = it.next();
    }
    templates.select(i - 1);
    if (!selectedTemplateId.equals(curId)) {
      if (!templates.getItem(templates.getItemCount() - 1).equals(NOT_INSTALLED))
        templates.add(NOT_INSTALLED);
      templates.select(templates.getItemCount() - 1);
    }
    templates.setEnabled(!isData());
    rootReqPanel.setTreeDB(db);
    rootReqPanel.setEnabled(!isData());
    gen.setEnabled(!isData());
    textId.setText(node!=null?node.getId():"");
    updateGUIElements();
  }

  private ControlDecoration createDecorator(Text text) {
    ControlDecoration controlDecoration = new ControlDecoration(text, SWT.LEFT | SWT.TOP);
    FieldDecoration fieldDecoration =
        FieldDecorationRegistry.getDefault().getFieldDecoration(FieldDecorationRegistry.DEC_ERROR);
    controlDecoration.setImage(fieldDecoration.getImage());
    return controlDecoration;
  }

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    parent.setLayout(new GridLayout(2, false));
    parent.setLayoutData(new GridData(GridData.FILL_BOTH));

    TabbedPropertySheetWidgetFactory wf = aTabbedPropertySheetPage.getWidgetFactory();
    wf.createLabel(parent, "Id:");

    textId = wf.createText(parent, "");
    textId.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    reqIdDecorator = createDecorator(textId);
    reqIdDecorator.hide();
    textId.addKeyListener(
        new KeyListener() {

          @Override
          public void keyReleased(KeyEvent e) {
            try {
              if (node == null) {
                return;
              }
              if (node.getTreeDB().getTreeLogic().isIdValidExc(node, textId.getText().trim())) {
                reqIdDecorator.hide();
              } else {
                reqIdDecorator.setDescriptionText(
                    "Element Id is not correct and Element Id should not be empty");
                reqIdDecorator.show();
              }
            } catch (SameIdExistsInChildren e1) {
              reqIdDecorator.setDescriptionText("Element with the same Id already exists");
            } catch (IdEmpty e1) {
              reqIdDecorator.setDescriptionText("Element Id is empty");
            } catch (IdContainsWrongChars e1) {
              reqIdDecorator.setDescriptionText(
                  "Element Id contains wrong character - " + e1.getWrongCharacter());
            }
          }

          @Override
          public void keyPressed(KeyEvent e) {}
        });
    textId.addFocusListener(
        new FocusAdapter() {
          public void focusLost(FocusEvent e) {
            if (updating) return;
            if (!textId.getText().equals(textId.getText().trim())) {
              textId.setText(textId.getText().trim());
            }
            if (node == null) {
              return;
            }
            if (node.getId().equals(textId.getText())) {
              return;
            }
            String newId = textId.getText().trim();
            if (node.getTreeDB().getTreeLogic().isIdValid(node, newId)
                && node.getTreeDB().startTransaction("Change Requirement Id", true)) {
              node.setId(newId.trim());
              node.commitAttributes();
              node.getTreeDB().commit();
            } else {
              textId.setText(node.getId());
            }
          }
        });
    wf.createLabel(parent, LBL_ROOT);
    rootReqPanel = new SelectRequirementPanel(parent, SWT.NONE, db);
    rootReqPanel.setCanSelectRoot(true);
    rootReqPanel.addListener(
        SWT.Modify,
        new Listener() {
          @Override
          public void handleEvent(Event event) {
            Requirement req = rootReqPanel.getSelectedRequirement();
            if (report.getTreeDB().startTransaction("Update Report Data", true)) {
              report
                  .attrs()
                  .setConstAttribute(ReportSettings.ATTR_ROOT_REQUIREMENT, req.getQualifiedId(), Flag.HIDDEN);
              report.commitAttributes();
              report.getTreeDB().commit();
            } else {
              if (report.getStartRequirementQId() == null) {
                req = (Requirement) db.getNode(Requirement.getTypeRootQId(), Path.QID);
              } else {
                req = (Requirement) db.getNode(report.getStartRequirementQId(), Path.QID | Path.UV);
              }
              rootReqPanel.setRequirement(req);
            }
          }
        });
    Composite panelmid = new Composite(parent, SWT.NONE);
    GridData gd = new GridData(GridData.FILL_HORIZONTAL);
    gd.horizontalSpan = 2;
    panelmid.setLayoutData(gd);
    GridLayout gl = new GridLayout(5, false);
    gl.horizontalSpacing = gl.verticalSpacing = 0;
    panelmid.setLayout(gl);
    panelmid.setBackground(parent.getBackground());

    wf.createLabel(panelmid, LBL_TEMPLATE);
    createSelectTemplatePanel(panelmid)
        .addListener(
            SWT.Modify,
            new Listener() {
              @Override
              public void handleEvent(Event event) {
                Iterator<String> it = tpls.keySet().iterator();
                for (int i = 0; i < templates.getSelectionIndex() + 1; i++) {
                  selectedTemplateId = it.next();
                }
                if (!selectedTemplateId.equals(NOT_INSTALLED)) {
                  if (templates.getItem(templates.getItemCount() - 1).equals(NOT_INSTALLED))
                    templates.remove(NOT_INSTALLED);
                  if (report.getTreeDB().startTransaction("Update Report Data", true)) {
                    GenerateReportHandler.getOne().clearOldImages(report);
                    report.attrs().setConstAttribute(ReportSettings.ATTR_TEMPLATE, selectedTemplateId, Flag.HIDDEN);
                    report.commitAttributes();
                    report.getTreeDB().commit();
                  }
                  //TODO modify
                  //GenerateReportHandler.generateReport((ReportSettings)ReportSettings);
                  // OpenReportHandler.openReport((ReportSettings)ReportSettings);
                  updateGUIElements();
                }
              }
            });

    customGUIpanel = new Composite(panelmid, SWT.NONE);
    RowLayout rl = new RowLayout();
    rl.center = true;
    customGUIpanel.setLayout(rl);
    customGUIpanel.setBackground(parent.getBackground());
    GridData gridData = new GridData(GridData.FILL_BOTH);
    gridData.grabExcessHorizontalSpace = true;
    gridData.grabExcessVerticalSpace = true;
    customGUIpanel.setLayoutData(gridData);

    gen = new Button(panelmid, SWT.NONE);
    gen.setText("Generate");
    gen.addListener(
        SWT.MouseDown,
        new Listener() {
          @Override
          public void handleEvent(Event event) {
            WorkspaceJob wj = new WorkspaceJob("Generate report ("+report.getTreeDB().getResName()+") "+report.getUserVisibleName()){
            
              @Override
              public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                try{
                  GenerateReportHandler.getOne().generateReport(report,
                    com.unitesk.requality.eclipse.handlers.report.GenerateReportHandler.wrapMonitor(monitor));
                }catch(Exception e){
                  org.eclipse.swt.widgets.Display.getDefault().asyncExec(new Runnable(){
                  
                    @Override
                    public void run() {
                      MessageDialog.openError(org.eclipse.swt.widgets.Display.getDefault().getActiveShell(),
                          "Report generation error","Cannot generate report: "+e.getMessage());
                    }
                  });
                  //RLogger.logError(CoreUtils.getTrace(e));
                }
                return Status.OK_STATUS;
              }
            };
            wj.schedule();
          }
        });
    attrPanel = new AttributesPanel(parent);
    GridData data = new GridData(GridData.FILL_BOTH);
    data.horizontalSpan = 2;
    attrPanel.getControl().setLayoutData(data);
    attrPanel.getControl().setBackground(parent.getBackground());
  }

  public void updateGUIElements() {
    for (Control c : customGUIpanel.getChildren()) c.dispose();
    Template t = getTemplate();
    if (t == null) return;
    rootReqPanel.setEnabled(!t.getIsGlobal() && !isData());
    Map<String, Map<String, String>> mapParams = t.getGUIElements();
    for (String id : mapParams.keySet()) {
      Map<String, String> params = mapParams.get(id);
      String type = params.get("_type");
      if (id.equals("settings")) {
        //TODO global settings - layout etc
      } else if (type != null) {
        if (type.equals("button")) {
          final IHandler[] hs = Activator.getDefault().getCommandHandlers(params.get("_command"));
          if (hs != null && hs.length > 0) {
            Button b = new Button(customGUIpanel, SWT.NONE);
            b.setEnabled(!isData());
            b.setText(params.get("_text"));
            b.addMouseListener(
                new MouseListener() {

                  @Override
                  public void mouseUp(MouseEvent e) {
                    // TODO Auto-generated method stub

                  }

                  @Override
                  public void mouseDown(MouseEvent e) {
                    Map<String, Object> params = new HashMap<>();
                    params.put("target", report);
                    for (IHandler h : hs)
                      try {
                        Activator.executeCommand(h, params);
                        return;
                      } catch (ExecutionException e1) {
                        // TODO Auto-generated catch block
                        e1.printStackTrace();
                      }
                  }

                  @Override
                  public void mouseDoubleClick(MouseEvent e) {
                    // TODO Auto-generated method stub

                  }
                });
          }
        } else if (type.equals("label")) {
          Label label = new Label(customGUIpanel, SWT.NONE);
          label.setBackground(customGUIpanel.getBackground());
          String lab = params.get("_text");
          if (lab != null) label.setText(lab);
        }
      }
    }

    customGUIpanel.layout();
    customGUIpanel.update();

    customGUIpanel.getParent().layout();
    customGUIpanel.getParent().update();
  }

  public Template getTemplate() {
    return tpls.get(selectedTemplateId);
  }

  private Composite createSelectTemplatePanel(Composite parent) {
    final Composite panel = new Composite(parent, SWT.NONE);
    panel.setBackground(parent.getBackground());
    panel.setLayout(new GridLayout());
    templates = new Combo(panel, SWT.NONE | SWT.READ_ONLY);
    tpls = new LinkedHashMap<>(Activator.getDefault().getReportTemplates());
    if (tpls.size() == 0) {
      return panel;
    }
    if (selectedTemplateId == null) {
      selectedTemplateId = tpls.keySet().iterator().next();
    }

    for (Template tpl : tpls.values()) {
      templates.add(tpl.getName());
    }
    Iterator<String> it = tpls.keySet().iterator();
    String curId = "";
    int i = 0;
    for (; i < tpls.size() && !selectedTemplateId.equals(curId); i++) {
      curId = it.next();
    }
    templates.select(i - 1);
    templates.addSelectionListener(
        new SelectionListener() {
          @Override
          public void widgetSelected(SelectionEvent e) {
            panel.notifyListeners(SWT.Modify, new Event());
            updateGUIElements();
          }

          @Override
          public void widgetDefaultSelected(SelectionEvent e) {
            widgetSelected(e);
          }
        });
    return panel;
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return true; //TODO fix;
  }

  private static Set<String> ids = new HashSet<>();
  static{
    ids.add(TreeNode.ATTR_ID);
    ids.add(ReportSettings.ATTR_ROOT_REQUIREMENT);
    ids.add(ReportSettings.ATTR_TEMPLATE);
  }
  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    for(String s:what.getAll())
      if(ids.contains(s)||!s.startsWith("_"))
        return true;
    return false;
  };
}
