/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.nodetypes.TestPurpose;

public class TestDescriptionPropertySection extends TreeNodePropertiesTab {

  private TestPurpose testPurpose;

  private Text description; // Test Case Description
  private Text results; // Expected Results
  private Composite par;

  public TestDescriptionPropertySection() {}

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    TabbedPropertySheetWidgetFactory wf = aTabbedPropertySheetPage.getWidgetFactory();
    par = parent;
    parent.setLayout(new GridLayout());
    //parent.setLayoutData(new GridData(GridData.FILL_BOTH));

    wf.createLabel(parent, "Test purpose description:");

    description = wf.createText(parent, "", SWT.MULTI | SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
    description.setLayoutData(new GridData(GridData.FILL_BOTH));
    description.addFocusListener(
        new FocusAdapter() {
          public void focusLost(FocusEvent e) {
            if (updating) return;
            if (testPurpose == null) return;
            String oldDescr = testPurpose.getDescription();
            if (oldDescr == null) {
              oldDescr = "";
            }
            String newDescr = "";
            newDescr = CoreUtils.htmlEncapsulate(description.getText());
            description.setText(newDescr);
            if (oldDescr.equals(newDescr)) {
              return;
            }
            if (testPurpose.getTreeDB().startTransaction("Update Test Purpose", true)) {
              testPurpose.setDescription(newDescr);
              testPurpose.commitAttributes();
              testPurpose.getTreeDB().commit();
            } else description.setText(testPurpose.getDescription());
          }
        });

    wf.createLabel(parent, "Expected results:");

    results = wf.createText(parent, "", SWT.MULTI | SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
    results.setLayoutData(new GridData(GridData.FILL_BOTH));
    results.addFocusListener(
        new FocusAdapter() {
          public void focusLost(FocusEvent e) {
            if (updating) return;
            if (testPurpose == null) return;
            String oldResults = testPurpose.getExpectedResults(false);
            if (oldResults == null) {
              oldResults = "";
            }
            String newResults = "";
            newResults = CoreUtils.htmlEncapsulate(results.getText());
            results.setText(newResults);
            if (oldResults.equals(newResults)) {
              return;
            }
            if (testPurpose.getTreeDB().startTransaction("Update Test Purpose", true)) {
              if (!newResults.equals("")) {
                testPurpose.setExpectedResults(newResults);
              } else {
                testPurpose.attrs().removeAttribute(TestPurpose.ATTR_EXPECTED_RESULTS);
              }
              testPurpose.commitAttributes();
              testPurpose.getTreeDB().commit();
            } else results.setText(oldResults);
          }
        });
  }

  @Override
  protected void setTreeNode(TreeNode node) {
    if (!(node instanceof TestPurpose) || description == null || description.isDisposed()) {
      return;
    }
    testPurpose = (TestPurpose) node;
    // getDescription() returns null instead of ""
    // if corresponding attribute isn't presented in node
    description.setText(testPurpose.getDescription() != null ? testPurpose.getDescription() : "");
    GridData gd = new GridData(GridData.FILL_BOTH);
    gd.grabExcessHorizontalSpace = true;
    gd.grabExcessVerticalSpace = true;
    if (par.getSize().x > 300) gd.widthHint = par.getSize().x - 300;
    else gd.widthHint = par.getSize().x;
    description.setLayoutData(gd);
    results.setText(testPurpose.getExpectedResults(false));
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return TreeNode.ATTR_DESCRIPTION.equals(a)
        || TestPurpose.ATTR_EXPECTED_RESULTS.equals(a); //TODO fix;
  };
  private static Set<String> ids = new HashSet<>();
  static{
    ids.add(TreeNode.ATTR_DESCRIPTION);
    ids.add(TestPurpose.ATTR_EXPECTED_RESULTS);
  }
  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    return what.getAll().stream().filter(ids::contains).collect(Collectors.toSet()).size()>0;
  };
}
