/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.nodetypes.TestPurpose;

public class TestPredicatePropertySection extends TreeNodePropertiesTab {

  private TestPurpose testPurpose;

  private Text predicate; // Test Predicate

  public TestPredicatePropertySection() {}

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    TabbedPropertySheetWidgetFactory wf = aTabbedPropertySheetPage.getWidgetFactory();

    parent.setLayoutData(new GridData(GridData.FILL_BOTH));
    parent.setLayout(new GridLayout());

    wf.createLabel(parent, "Predicate:");

    predicate = wf.createText(parent, "", SWT.MULTI | SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
    predicate.setLayoutData(new GridData(GridData.FILL_BOTH));
    predicate.addFocusListener(
        new FocusAdapter() {
          public void focusLost(FocusEvent e) {
            if (updating) return;
            if (testPurpose == null) return;
            String oldPredicate = testPurpose.getPredicate();
            if (oldPredicate == null) {
              oldPredicate = "";
            }
            String newPredicate = "";
            newPredicate = predicate.getText();
            if (oldPredicate.equals(newPredicate)) {
              return;
            }
            if (testPurpose.getTreeDB().startTransaction("Update Test Purpose", true)) {
              testPurpose.setPredicate(newPredicate);
              testPurpose.commitAttributes();
              testPurpose.getTreeDB().commit();
            } else predicate.setText(oldPredicate);
          }
        });
  }

  @Override
  protected void setTreeNode(TreeNode node) {
    if (!(node instanceof TestPurpose) || predicate == null || predicate.isDisposed()) {
      return;
    }
    testPurpose = (TestPurpose) node;
    predicate.setText(testPurpose.getPredicate());
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return TreeNode.ATTR_PREDICATE.equals(a); //TODO fix;
  }

  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    // TODO Auto-generated method stub
    return checkNodeChanges(source)||what.contains(TreeNode.ATTR_PREDICATE);
  };
}
