package com.unitesk.requality.eclipse.views.properties;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import org.eclipse.core.resources.IProject;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbenchPropertyPage;
import org.eclipse.ui.dialogs.PropertyPage;
import com.unitesk.requality.core.AttributesVersioningRule;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.SystemNames;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class VersioningPolicy extends PropertyPage implements IWorkbenchPropertyPage {
  private TreeDB db;
  private IProject p;  
  private Button envers, enallver;
  private org.eclipse.swt.widgets.List listtypes, listattrs;
  private Button add;
  private Button remove;
  private Text name;
  private Combo exnames;
  private Button loadnames;
  private Group container4;
  
  private Map<Integer,String> typeMap = new HashMap<>();
  private Map<String, Set<String>> map;
  private String selType = null;
  
  private class ReplPair {
    private String name, repl;

    public ReplPair(String name, String repl) {
      this.name = name;
      this.repl = repl;
    }

    @Override
    public String toString() {
      return "'" + name + "' -> '" + repl + "'";
    }

    @Override
    public boolean equals(Object obj) {
      if (obj instanceof ReplPair) return name.equals(((ReplPair) obj).name);
      return super.equals(obj);
    }
  }

  public VersioningPolicy() {}
  private SelectionListener listAttrList = new SelectionListener(){
  
    @Override
    public void widgetSelected(SelectionEvent e) {
      if(listattrs.getSelectionIndex()>-1)
        name.setText(listattrs.getSelection()[0]);
    }
  
    @Override
    public void widgetDefaultSelected(SelectionEvent e) {
    }
  };
  private void recEnabled(Control c, boolean b){
    c.setEnabled(b);
    if(c instanceof Composite){
      Composite cm = (Composite)c;
      for(Control c2:cm.getChildren())
        recEnabled(c2, b);
    }
  }
  @Override
  protected Control createContents(Composite parent) {

    Composite c = new Composite(parent, SWT.NONE);
    c.setLayout(new GridLayout(1, true));
    c.setLayoutData(new GridData(GridData.FILL_BOTH));

    final Group container3 = new Group(c, SWT.NONE);
    container3.setText("Versioning:");
    container3.setLayout(new GridLayout(1, false));
    container3.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    
    envers = new Button(container3, SWT.CHECK);
    envers.setText("Enable versioning");
    envers.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    
    enallver = new Button(container3, SWT.CHECK);
    enallver.setText("Versioning for all changes");
    enallver.setSelection(true);
    enallver.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

    //initReferenceRenameComposite(container3);
    container4 = new Group(c, SWT.NONE);
    container4.setText("Attributes settings for version control:");
    container4.setLayout(new GridLayout(4, false));
    container4.setLayoutData(new GridData(GridData.FILL_BOTH));

    enallver.addSelectionListener(new SelectionListener(){
    
      @Override
      public void widgetSelected(SelectionEvent e) {
        recEnabled(container4,!enallver.getSelection());
      }
    
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });
    
    envers.addSelectionListener(new SelectionListener(){
    
      @Override
      public void widgetSelected(SelectionEvent e) {
        enallver.setEnabled(false);
        recEnabled(container4,false);
        if(envers.getSelection()){
          enallver.setEnabled(true);
          if(!enallver.getSelection())
            recEnabled(container4,true);
        }
      }
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });
    
    Composite c1 = new Composite(container4, SWT.NONE);
    c1.setLayout(new GridLayout(1,false));
    c1.setLayoutData(new GridData(GridData.FILL_BOTH));
    
    Label l = new Label(c1,SWT.NONE);
    l.setText("Node type:");
    listtypes = new List(c1, SWT.NONE);
    listtypes.setLayoutData(new GridData(GridData.FILL_BOTH));
    listtypes.add("Requirement");
    listtypes.add("Text node");
    typeMap.put(0, Requirement.TYPE_NAME);
    typeMap.put(1, TextNode.TYPE_NAME);
    listtypes.addSelectionListener(new SelectionListener(){
    
      @Override
      public void widgetSelected(SelectionEvent e) {
        int sel = listtypes.getSelectionIndex();
        listattrs.removeAll();
        if(sel>-1){
          selType = typeMap.get(sel);
          setListAttrs();
        }
      }
    
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });

    Composite c2 = new Composite(container4, SWT.NONE);
    c2.setLayout(new GridLayout(1,false));
    c2.setLayoutData(new GridData(GridData.FILL_BOTH));

    l = new Label(c2,SWT.NONE);
    l.setText("Internal attributes:");
    //TODO mouse hover
    listattrs = new List(c2, SWT.NONE);
    listattrs.setLayoutData(new GridData(GridData.FILL_BOTH));
    
    Composite ca = new Composite(container4, SWT.NONE);
    ca.setLayout(new GridLayout(1,true));
    ca.setLayoutData(new GridData(GridData.FILL_VERTICAL));
    
    add = new Button(ca, SWT.NONE);
    add.setText("←");
    add.addListener(SWT.Selection, new Listener() {
      @Override
      public void handleEvent(Event event) {
//        int sel = listattrs.getSelectionIndex();
//        if(sel>-1){
//          String val = listattrs.getItem(sel);
//          listattrs.remove(sel);
//          if(map.containsKey(selType))
//            map.get(selType).remove(val);
//        }
        if(!map.containsKey(selType))
          map.put(selType, new HashSet<>());
        if(!mapContains(name.getText()))
        {
          setMap(name.getText());
        }
      }

      
    });
    remove = new Button(ca, SWT.NONE);
    remove.setText("→");
    remove.addListener(SWT.Selection, new Listener(){
      @Override
      public void handleEvent(Event event) {
        int sel = listattrs.getSelectionIndex();
        String val = null;
        if(sel>-1){
          val = listattrs.getItem(sel);
          listattrs.remove(sel);
          if(map.containsKey(selType))
            map.get(selType).remove(val);
        }
//        else{
//          val = name.getText();
//          if(map.containsKey(selType)&&map.get(selType).contains(val)){
//            listattrs.remove(val);
//          }
//        }
        if(map.containsKey(selType))
          map.get(selType).remove(val);
      }
    });
    
    Composite names = new Composite(container4,SWT.NONE);
    names.setLayout(new GridLayout(1,true));
    names.setLayoutData(new GridData(GridData.FILL_BOTH));
    
    l = new Label(names,SWT.NONE);
    l.setText("Name:");
    
    name = new Text(names, SWT.NONE);
    name.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    name.addModifyListener(new ModifyListener(){

      @Override
      public void modifyText(ModifyEvent e) {
        String n = name.getText();
        add.setEnabled(true);
        remove.setEnabled(true);
        if(n.isEmpty())
        {
          add.setEnabled(false);
          remove.setEnabled(false);
        }
        listattrs.removeSelectionListener(listAttrList);
        listattrs.setSelection(new String[]{n});
        listattrs.addSelectionListener(listAttrList);
      }
    });
    
    l = new Label(names,SWT.NONE);
    l.setText("Attribute name:");
    
    exnames = new Combo(names, SWT.READ_ONLY);
    exnames.setEnabled(false);
    exnames.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    exnames.addSelectionListener(new SelectionListener(){
    
      @Override
      public void widgetSelected(SelectionEvent e) {
        name.setText(exnames.getText());
      }
    
      @Override
      public void widgetDefaultSelected(SelectionEvent e) {
        // TODO Auto-generated method stub
        
      }
    });
    
    loadnames = new Button(names,SWT.NONE);
    loadnames.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    loadnames.setText("Find attr. names");
    loadnames.addListener(SWT.Selection, new Listener(){
      @Override
      public void handleEvent(Event event) {
        // TODO Auto-generated method stub
        DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(db.getRootNode());
        Iterator<TreeNode> it = dftw.iterator();
        SortedSet<String> names = new TreeSet<>();
        while(it.hasNext()){
          TreeNode node = it.next();
          for(Entry<String, Attribute> en:node.attrs().getActiveAttributes().entrySet()){
            Attribute a = en.getValue();
            if(a != null && !a.isVirtual()){
              String name = null;
              if(a.isUserVisible())
                name = a.getKey();
              else
                name = SystemNames.getVisSysName(a.getKey());
              if(name!=null)
                names.add(name);
            }
          }
        }
        exnames.setItems(names.toArray(new String[0]));
        exnames.setEnabled(true);
      }
    });

    AttributesVersioningRule avr = db.getAttributeRules().getARule();
    envers.setSelection(avr!=null);
    if(avr!=null){
      enallver.setSelection(avr.isEnAll());
      map = avr.getTypedNames();
    }else{
      enallver.setSelection(true);
      enallver.setEnabled(false);
    }
    if(map == null)
      map = new HashMap<>();
    initControls();
    recEnabled(container4, avr != null && !avr.isEnAll());
    return c;
  }

  private void initControls() {
    recEnabled(container4, false);
    listtypes.select(0);//sets default selection
    selType = typeMap.get(0);
    setListAttrs();
  }
  private boolean mapContains(String text){
    if(!map.containsKey(selType))
      map.put(selType, new HashSet<>());
    String as = SystemNames.getSysName(text);
    if(as!=null)
      return map.get(selType).contains(as);
    else
      return map.get(selType).contains(text);
  }
  private void setMap(String text) {
    if(!map.containsKey(selType))
      map.put(selType, new HashSet<>());
    String as = SystemNames.getSysName(text);
    if(as!=null){
      map.get(selType).add(as);
      listattrs.add(SystemNames.getVisSysName(as));
    }else{
      map.get(selType).add(text);
      String vtext = SystemNames.getVisSysName(text);
      if(vtext!=null)
        listattrs.add(vtext);
      else
        listattrs.add(text);
    }
  }
  private void setListAttrs() {
    if(!map.containsKey(selType))
      map.put(selType, new HashSet<>());
    SortedSet<String> sset = new TreeSet<>();
    for(String s:map.get(selType)){
      String as = SystemNames.getVisSysName(s);
      if(as!=null)
        sset.add(as);
      else
        sset.add(s);
    }
    listattrs.setItems(sset.toArray(new String[0]));
  }
  @Override
  public void setElement(org.eclipse.core.runtime.IAdaptable element) {
    if (element instanceof IProject) {
      p = (IProject) element;
      db = EclipseTracker.getTracker().getTree(p);
    }
  };

  //actions when new selection is performed
  @Override
  public boolean performOk() {
    if(!envers.getSelection())
      db.getAttributeRules().setARule(null);
    else
      if(enallver.getSelection())
        db.getAttributeRules().setARule(new AttributesVersioningRule(null, true));
      else
        db.getAttributeRules().setARule(new AttributesVersioningRule(map, false));
    db.getAttributeRules().saveRules();
    return true;
  }

  //Changing settings to default
  @Override
  protected void performDefaults() {
    envers.setSelection(false);
    recEnabled(container4,false);
    enallver.setEnabled(false);
    enallver.setSelection(true);
    db.getAttributeRules().setARule(null);
    db.getAttributeRules().saveRules();
    exnames.setEnabled(false);
    exnames.setItems(new String[0]);
    initControls();
    super.performDefaults();
  }
}
