package com.unitesk.requality.eclipse.wizards;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.dialogs.DialogSettings;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.ui.IImportWizard;
import org.eclipse.ui.IWorkbench;
import org.w3c.dom.Document;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.tools.ReqIFConverter;
import com.unitesk.requality.tools.reqif.AttributeDefinition;
import com.unitesk.requality.tools.reqif.ReqIFModel;
import com.unitesk.requality.tools.reqif.SpecObjectType;

public class ImportFromReqIFWizard extends Wizard implements IImportWizard {
  public static String NEW_PROJECT_CONST = "~Add new project~";
  public static String IGNORE_ATTRIBUTE_CONST = "~Ignore attribute~";
  public static String ADD_NEW_ATTRIBUTE_CONST = "~Add new attribute~";
  public static String DOORS_ATTRIBUTES = "~Doors reqif conversion~";

  private IStructuredSelection selection;
  private TreeDB db;
  private WorkspaceFileSelectionPage fp;
  private ReqMappingPage map0, map, map2;

  public ImportFromReqIFWizard() {
    super();
  }

  private Map<IResource, Document> models = new HashMap<>();
  private Map<Document, ReqIFModel> mdls = new HashMap<>();

  protected ReqIFModel getModel() {
    Document doc = getDoc();
    if (doc != null) {
      if (!mdls.containsKey(doc)) {
        IFile f = getRes();
        if (f != null) {
          ReqIFModel mod = new ReqIFModel(f.getParent().getLocationURI().toString()); //TODO
          mod.loadDoors(doc);
          mdls.put(doc, mod);
        }
      }
      return mdls.get(doc);
    }
    return null;
  }

  protected IFile getRes() {
    IFile res = fp.getSelectedResource();
    if (res == null) return null;
    return res;
  }

  protected Document getDoc() {
    IFile res = getRes();
    if (res == null) return null;
    if (!models.containsKey(res))
      try {
        if (res.exists()) {
          res.setCharset("UTF-8", null);
          models.put(res, ReqIFConverter.loadDoc(res.getContents()));
        }
      } catch (CoreException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    return models.get(res);
  }

  @Override
  public void addPages() {
    addPage(fp);
    addPage(map0);
    addPage(map);
    addPage(map2);
  };

  private TreeDB getSelectedTreeDB() {
    if (!(selection instanceof StructuredSelection)) return db;
    TreeDB db = this.db;
    Object o = ((StructuredSelection) selection).getFirstElement();
    if (o instanceof IProject) db = EclipseTracker.getTracker().getTree((IProject) o);
    if (o instanceof TreeNode) db = ((TreeNode) o).getTreeDB();
    if (o instanceof TreeDB) db = (TreeDB) o;
    this.db = db;
    return db;
  }

  @Override
  public void init(IWorkbench workbench, IStructuredSelection selection) {
    setWindowTitle("ReqIF to Requality import wizard");
    this.selection = selection;
    DialogSettings ds = new DialogSettings("importFromReqIf");
    fp = new WorkspaceFileSelectionPage(ds, new String[] {"reqif", "xml"});
    map0 =
        new ReqMappingPage(ds, this) {
          Map<String, String> specs = ReqIFConverter.loadReqIFSpecifications(getModel());

          @Override
          public void updateSecondField() {
            Set<String> names = new HashSet<>();
            names.add(NEW_PROJECT_CONST);
            for (IProject proj : EclipseTracker.getTracker().getProjects())
              names.add(proj.getName());
            map0.initSecondField(names);
          }

          @Override
          public void updateFirstField() {
            specs = ReqIFConverter.loadReqIFSpecifications(getModel());
            if (specs == null || specs.size() < 1) return;
            Set<String> ret = new HashSet<>();
            for (String s : specs.keySet()) ret.add(specs.get(s) + TreeDB.DATA_SEPARATOR + s);
            map0.initFirstField(ret);
            updateSecondField();
          }

          @Override
          protected boolean isClose(String s, String k) {
            return firstPass ? (k.length() > 4 && (s).startsWith(k)) : k.equals(NEW_PROJECT_CONST);
          }
        };
    map0.setTitle("Specification to project mapping");
    map0.setDescription("Map ReqIF specifications to Requality projects.");

    map = new ReqMappingPage(ds, this);

    map2 =
        new ReqMappingPage(ds, this) {
          @Override
          public void updateSecondField() {
            Map<String, String> mapp1 = map.getMapping();
            if (mapp1 == null || mapp1.size() < 1) return;
            Set<String> ret = new HashSet<>();
            ret.add(IGNORE_ATTRIBUTE_CONST);
            ret.add(ADD_NEW_ATTRIBUTE_CONST);
            ret.addAll(ReqIFConverter.renames.keySet());
            map2.initSecondField(ret);
            map2.result.add(DOORS_ATTRIBUTES);
            //TODO load reqs
          }

          @Override
          public void updateFirstField() {
            Map<String, String> mapp1 = map.getMapping();
            if (mapp1 == null || mapp1.size() < 1) return;
            ReqIFModel doc = getModel();
            if (doc != null) {
              Set<String> ret = new HashSet<>();
              for (SpecObjectType sot : doc.getSpecObjectTypes()) {
                for (AttributeDefinition def : sot.getAttributeDefinitions())
                  ret.add(sot.getName() + "." + def.getName());
                //ret.add(s+".LongName");
              }
              map2.initFirstField(ret);
            }
          }

          @Override
          protected boolean isClose(String s, String k) {
            if (map2.firstPass) {
              if (k.length() > 0 && s.endsWith(k)) return true;
              for (String i : ReqIFConverter.renames.keySet())
                for (String vl : ReqIFConverter.renames.get(i))
                  if (s.endsWith(vl) && i.equals(k)) return true;
              return false;
            }
            return k.equals(ADD_NEW_ATTRIBUTE_CONST);
          }
        };
    map2.setTitle("Attributes mapping");
    map2.setDescription("Map ReqIF attributes to Requality attributes.");
  }

  @Override
  public boolean canFinish() {
    return map.getMapping() != null
        && !map.getMapping().isEmpty()
        && fp.getSelectedResource() != null
        && fp.getSelectedResource().exists()
        && map0.getMapping() != null
        && !map0.getMapping().isEmpty();
  };

  @Override
  public boolean performFinish() {
    Map<String, String> mapp1 = map.getMapping();
    Map<String, String> mapp2 = map2.getMapping();
    Map<String, String> mapp0 = map0.getMapping();
    Map<String, TreeDB> nmap = new HashMap<>();
    for (String s : mapp0.keySet()) {
      String pr = mapp0.get(s);
      if (!pr.equals(NEW_PROJECT_CONST)) {
        IProject proj = Activator.getProjectByName(pr);
        if (proj != null && proj.exists()) nmap.put(s, EclipseTracker.getTracker().getTree(proj));
        else nmap.put(s, null);
      } else nmap.put(s, null);
    }
    ReqIFConverter.convertToCatalog(getModel(), mapp1, mapp2, nmap);
    return true;
  }

  public IFile getSelectedFile() {
    return fp.getSelectedResource();
  }
}
