package com.unitesk.requality.eclipse.wizards.massops;

import java.io.IOException;
import java.util.Set;
import org.eclipse.jface.dialogs.DialogSettings;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Text;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.operations.mass.MassOppsArguments;
import com.unitesk.requality.eclipse.views.properties.ProjectSettingsMainSection;

public class AttrTypeChangeExtraSettingsPage extends AbstractExtraSettingsPage {

  private static final String LBL_CHNG_NAME = "Select new attribute name:";
  private static final String LBL_DEF_NAME = "Select new attribute types:";
  private Text newAttrName;
  private Combo type;
  boolean typedNewName = false;
  private boolean choosedNewType = false;

  public AttrTypeChangeExtraSettingsPage(DialogSettings ds, Set<String> names, TreeDB db) {
    super("Attribute type change", ds, names, db);
    setTitle("Attribute type change extra settings page");
  }

  @Override
  public void createControl(Composite parentP) {
    //super.createControl(parent);
    container = new Composite(parentP, SWT.None);
    GridLayout layout = new GridLayout();
    container.setLayout(layout);
    layout.verticalSpacing = 2;
    container.setLayout(layout);
    container.setLayoutData(new GridData(GridData.FILL_BOTH));
    leftright = new Composite(container, SWT.NONE);
    leftright.setLayoutData(new GridData(GridData.FILL_BOTH));
    leftright.setLayout(new GridLayout(2, true));

    left = new Composite(leftright, SWT.NONE);
    left.setLayout(new GridLayout(1, true));
    left.setLayoutData(new GridData(GridData.FILL_BOTH));

    right = new Composite(leftright, SWT.NONE);
    right.setLayout(new GridLayout(1, true));
    right.setLayoutData(new GridData(GridData.FILL_BOTH));

    down = new Composite(container, SWT.NONE);

    (new Label(left, SWT.None)).setText(LBL_ATTR_NAME);
    attrList = new List(left, SWT.V_SCROLL | SWT.H_SCROLL);
    attrList.addSelectionListener(listener);
    GridData gd = new GridData(GridData.FILL_BOTH);
    gd.heightHint = 300;
    attrList.setLayoutData(gd);
    if (attrNames == null || attrNames.isEmpty()) {
      attrList.setItems(new String[]{"There are no attributes"});
      attributesExist = false;
    } else {
      attrList.setItems(attrNames.toArray(new String[0]));
    }

    (new Label(right, SWT.NONE)).setText(LBL_ATTR_VALUE);
    attrValList = new List(right, SWT.V_SCROLL | SWT.H_SCROLL | SWT.READ_ONLY);
    attrValList.setBackground(
        Display.getDefault().getSystemColor(SWT.COLOR_TITLE_INACTIVE_BACKGROUND));
    gd = new GridData(GridData.FILL_BOTH);
    gd.heightHint = 300;
    attrValList.setLayoutData(gd);
    down.setLayout(new GridLayout(2, true));
    attrList.addSelectionListener(listener);
    Label nameLabel1 = new Label(down, SWT.NONE);
    nameLabel1.setText(LBL_CHNG_NAME);
    nameLabel1.setLayoutData(new GridData(GridData.FILL_HORIZONTAL, GridData.CENTER, true, true));
    newAttrName = new Text(down, SWT.BORDER);
    newAttrName.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    newAttrName.addModifyListener(newNameListener);

    Label nameLabel2 = new Label(down, SWT.NONE);
    nameLabel2.setText(LBL_DEF_NAME);
    nameLabel2.setLayoutData(new GridData(GridData.FILL_HORIZONTAL, GridData.CENTER, true, true));
    type = new Combo(down, SWT.BORDER | SWT.READ_ONLY);
    type.add("<old>");
    for (String tp : Attribute.getAvailableTypes(db, ProjectSettingsMainSection.filter)) type.add(tp);
    type.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    type.select(0);
    type.addSelectionListener(typeListener);

    setControl(container);
  }

  private SelectionListener typeListener =
      new SelectionListener() {

        @Override
        public void widgetSelected(SelectionEvent e) {
          int ind = type.getSelectionIndex();
          if (ind > -1) {
            String attrType = type.getItem(ind);
            if (ind == 0) attrType = null;
            dialogSettings.put(MassOppsArguments.Argument.ATTRIBUTE_TYPE.name(), attrType);
            choosedNewType = true;
            setPageComplete(true);
            save();
          }
        }

        @Override
        public void widgetDefaultSelected(SelectionEvent e) {
          // TODO Auto-generated method stub
        }
      };
  private ModifyListener newNameListener =
      new ModifyListener() {
        @Override
        public void modifyText(ModifyEvent e) {
          String value = newAttrName.getText();
          if (value != null && !value.isEmpty()) {
            dialogSettings.put(MassOppsArguments.Argument.NEW_ATTRIBUTE_NAME.name(), value);
            setPageComplete(true);
          }
        }
      };

  private void save() {
    if (getAttributeChoose() && choosedNewType) {
      setPageComplete(true);

      try {
        dialogSettings.save(MassOppsWizard.FILE_NAME);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
  }

  private static String[] dnames = new String[] {};

  @Override
  public String[] getDialogNames() {
    return dnames;
  }

  @Override
  public void reloadSettingsFromDS() {
    // TODO Auto-generated method stub

  }
}
