/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.model;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.mapping.ResourceMapping;
import org.eclipse.core.resources.mapping.ResourceMappingContext;
import org.eclipse.core.resources.mapping.ResourceTraversal;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdapterFactory;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.ui.model.IWorkbenchAdapter;
import org.eclipse.ui.model.IWorkbenchAdapter2;

import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.core.EclipseTracker;

public class TreeNodeAdapterFactory implements IAdapterFactory {
  private Object requalityAdapter = new RequalityWorkbenchAdapter();

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  public Object getAdapter(final Object adaptableObject, Class adapterType) {
    if (adapterType.isInstance(adaptableObject)) {
      return adaptableObject;
    }
    if (adapterType.equals(ResourceMapping.class)) {
      final TreeNode node = (TreeNode) adaptableObject;
      final IFile res = (IFile) EclipseTracker.getResourceByNode(node);
      //if (node.isDisposed()||res==null)
      //  return null;

      ResourceMapping mapper =
          new ResourceMapping() {
            @Override
            public ResourceTraversal[] getTraversals(
                ResourceMappingContext context, IProgressMonitor monitor) throws CoreException {

              List<ResourceTraversal> result = new ArrayList<>();
              // Node
              if (res != null)
                result.add(
                    new ResourceTraversal(
                        new IResource[] {res}, IResource.DEPTH_ZERO, IContainer.EXCLUDE_DERIVED));
              if (res != null && res.getParent() != null) {
                IFolder resources =
                    res.getParent().getFolder(Path.fromOSString(node.getId() + "_resources"));
                if (resources != null && resources.exists()) {
                  result.add(
                      new ResourceTraversal(
                          new IResource[] {resources},
                          IResource.DEPTH_INFINITE,
                          IContainer.EXCLUDE_DERIVED));
                }
                IFolder children = res.getParent().getFolder(Path.fromOSString(node.getId()));
                if (children != null && children.exists()) {
                  result.add(
                      new ResourceTraversal(
                          new IResource[] {children},
                          IResource.DEPTH_INFINITE,
                          IContainer.EXCLUDE_DERIVED));
                }
              }

              return result.toArray(new ResourceTraversal[0]);
            }

            @Override
            public IProject[] getProjects() {
              IProject proj = EclipseTracker.getTracker().getProject(node.getTreeDB());
              if (proj != null) return new IProject[] {proj};
              return new IProject[0];
            }

            @Override
            public String getModelProviderId() {
              return RequalityModelProvider.ID;
            }

            @Override
            public Object getModelObject() {
              return node;
            }
          };
      return mapper;
    }
    if (adapterType == IWorkbenchAdapter.class || adapterType == IWorkbenchAdapter2.class
    // || adapterType == IWorkbenchAdapter3.class
    ) {
      return requalityAdapter;
    }
    if (adapterType.isAssignableFrom(IResource.class)) {
      return EclipseTracker.getResourceByNode((TreeNode) adaptableObject);
    }
    if (adapterType.equals(IFile.class)) {
      return EclipseTracker.getResourceByNode((TreeNode) adaptableObject);
    }
    return null;
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  public Class[] getAdapterList() {
    return new Class[] {
      ResourceMapping.class,
      IWorkbenchAdapter.class,
      IWorkbenchAdapter2.class,
      // IWorkbenchAdapter3.class,
      IResource.class
    };
  }
}
